# -*- coding: utf-8 -*-
"""
:copyright: NOKIA
:author: Cezary Miącz
:contact: cezary.miacz@nokia.com
"""

import base64
import hashlib
import socket
import struct
from threading import Thread


class MsgType(object):

    STREAM = 0x00
    TEXT = 0x01
    BIN = 0x02
    CLOSE = 0x08
    PING = 0x09
    PONG = 0x0A


class WebSocketMessage(object):

    def __init__(self, final, type_id, payload):
        self.final = final
        self.type_id = type_id
        self.payload = payload


class WebSocketMessageBuffer(object):

    def __init__(self, ):
        self.clear()

    def put(self, msg):
        """Put WebSocketMessage into buffer."""
        self.final = bool(msg.final)
        self.payload_buffer.append(msg.payload)
        if msg.type_id != MsgType.STREAM:
            self.type_id = msg.type_id

    def get(self):
        """Get buffered WebSocketMessage."""
        if not self.ready():
            raise Exception("Buffered WebSocket message is not ready.")
        buffered_msg = WebSocketMessage(self.final, self.type_id, "".join(self.payload_buffer))
        self.clear()
        return buffered_msg

    def ready(self):
        """Check if buffered WebSocketMessage is ready."""
        return self.final

    def clear(self):
        """Clear WebSocketMessageBuffer."""
        self.final = False
        self.type_id = None
        self.payload_buffer = []


class WebSocket(object):

    MSG_LENGTH_2B = 126
    MSG_LENGTH_8B = 127

    BUFFSIZE = 4096
    MAGIC_GUID = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
    HANDSHAKE_MSG_TEMPLATE = (
        "HTTP/1.1 101 Web Socket Protocol Handshake\r\n"
        "Upgrade: WebSocket\r\n"
        "Connection: Upgrade\r\n"
        "Sec-WebSocket-Accept: {}\r\n"
        "\r\n"
    )

    def __init__(self, sock):
        self.socket = sock
        self.socket.settimeout(1)
        self.msg_buffer = WebSocketMessageBuffer()
        self.receive_buffer = bytearray()
        self.connected = False
        self.receiver = Thread(target=self._run)
        self.receiver.setDaemon(True)

        self._on_message = self._default_callback
        self._on_close = self._default_callback

    def connect(self):
        """Open WebSocket connection."""
        self._perform_handshake()
        self.connected = True
        self.receiver.start()

    def disconnect(self):
        """Close WebSocket connection."""
        self.connected = False
        self.receiver.join(2)
        self.socket.close()

    def send(self, payload):
        """Send WebSocket message."""
        self.socket.send(self._encode_message(payload))

    def _default_callback(self, msg):
        pass

    def register_callbacks(self, on_message=None, on_close=None):
        """Register callbacks executed on when"""
        self._on_message = on_message or self._default_callback
        self._on_close = on_close or self._default_callback

    def _run(self):
        while self.connected:
            try:
                msg = self._receive_and_decode_message()
            except socket.timeout:
                pass
            else:
                self._handle_received_message(msg)
        self.socket.close()

    def _handle_received_message(self, msg):
        self.msg_buffer.put(msg)
        if self.msg_buffer.ready():
            msg = self.msg_buffer.get()
        else:
            return

        if msg.type_id == MsgType.TEXT:
            self._on_message(msg.payload)

        elif msg.type_id == MsgType.CLOSE:
            self.connected = False
            status = self._decode_close_status(msg.payload)
            self._on_close(status)

    def _receive_and_decode_message(self):
        final, type_id, length, mask = self._receive_and_decode_header()
        payload = self._decode_payload(mask, self._receive_bytes(length))
        return WebSocketMessage(final, type_id, payload)

    def _receive_and_decode_header(self):
        raw = self._receive_bytes(6)

        final = raw[0] & 0x80
        type_id = raw[0] & 0x0F
        length = raw[1] & 0x7F

        if length == self.MSG_LENGTH_8B:
            raw += self._receive_bytes(8)
            length = struct.unpack(">Q", raw[2:10])[0]
            mask = raw[10:14]
        elif length == self.MSG_LENGTH_2B:
            raw += self._receive_bytes(2)
            length = struct.unpack(">H", raw[2:4])[0]
            mask = raw[4:8]
        else:
            mask = raw[2:6]

        return final, type_id, length, mask

    def _receive_bytes(self, length):
        while len(self.receive_buffer) < length:
            self.receive_buffer += bytearray(self.socket.recv(self.BUFFSIZE))

        ret = self.receive_buffer[:length]
        del self.receive_buffer[:length]
        return ret

    def _decode_payload(self, mask, payload):
        for i, char in enumerate(payload):
            payload[i] = char ^ mask[i % 4]
        return str(payload)

    def _decode_close_status(self, raw_status):
        return struct.unpack(">H", raw_status)[0]

    def _encode_message(self, payload):
        if isinstance(payload, unicode):
            payload = payload.encode("UTF8")
        header = self._encode_header(len(payload))
        return header + payload

    def _encode_header(self, msg_length):
        first_byte = 0x80 | MsgType.TEXT
        if msg_length < 126:
            return struct.pack(">BB", first_byte, msg_length)
        elif msg_length < 65536:
            return struct.pack(">BBH", first_byte, self.MSG_LENGTH_2B, msg_length)
        else:
            return struct.pack(">BBQ", first_byte, self.MSG_LENGTH_8B, msg_length)

    def _perform_handshake(self):
        handshake = self.socket.recv(self.BUFFSIZE)
        self.socket.send(self._create_handshake_resp(handshake))

    def _create_handshake_resp(self, handshake):
        web_socket_key = self._get_web_socket_key(handshake)
        web_socket_accept = base64.b64encode(hashlib.sha1(web_socket_key + self.MAGIC_GUID).digest())
        return self.HANDSHAKE_MSG_TEMPLATE.format(web_socket_accept)

    def _get_web_socket_key(self, handshake_msg):
        for line in handshake_msg.splitlines():
            attrib, _, value = line.partition(": ")
            if attrib == "Sec-WebSocket-Key":
                return value
        raise Exception("Cannot find WebSocket key in handshake message:\n{}".format(handshake_msg))


class WebSocketServer(object):

    def __init__(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    def start(self, port=0):
        """Start WebSocket server."""
        self.socket.bind(("localhost", port))
        self.socket.listen(1)

    def stop(self):
        """Stop WebSocket server."""
        self.socket.close()

    def accept(self, timeout):
        """Accept client connection."""
        self.socket.settimeout(timeout)
        client, address = self.socket.accept()
        return WebSocket(client)

    def get_address(self):
        """Get bind port."""
        return self.socket.getsockname()
