const _ = require('lodash'),
    LOG = require('../logger.impl').getLogger('admin.core.alarmsMonitor'),
    events = require('../events'),
    distNameUtils = require('../utils/distNameUtils'),
    alarms = require('./alarms'),
    sequentialExecutorFactory = require('../sequentialExecutorFactory'),
    MetaService = require('../metaService');

module.exports = alarmsMonitor;

function alarmsMonitor(updateService,
                       metaService = MetaService(),
                       decoderService = require('../decoderFactory').newInstance(metaService)) {

    const ALARM_L_CLASS_NAME = 'ALARM_L';
    const current = {
        listener: createCurrentAlarmsListener(),
        event: createCurrentAlarmsEvent(),
        alarms: []
    };
    const historical = {
        listener: createHistoricalAlarmsListener(),
        event: createHistoricalAlarmsEvent(),
        alarms: [],
        processedId: -1,
        processingQueue: Promise.resolve()
    };

    return {
        watchCurrentAlarms: watchCurrentAlarms,
        watchHistoricalAlarms: watchHistoricalAlarms,
        getSeverityCount: getSeverityCount
    };

    function createCurrentAlarmsEvent() {
        var event = events.newEvent();
        event.onFirstAdded(function () {
            updateService.addImTreeListener(current.listener);
        });
        event.onLastRemoved(function () {
            updateService.removeImTreeListener(current.listener);
        });
        return event;
    }

    function watchCurrentAlarms(watcher) {
        var unwatch = current.event.watch(watcher);
        if (current.alarms.length) watcher(current.alarms);
        return unwatch;
    }

    function createCurrentAlarmsListener() {
        return {
            completeImTree: processCompleteImTree
        };
    }

    function processCompleteImTree(tree) {
        var newAlarms = getAlarmsFromImTree(tree);
        alarms.sortUsingObservationTime(newAlarms);
        current.alarms = newAlarms;
        current.event.notify(current.alarms);
    }

    function getAlarmsFromImTree(tree) {
        return tree.findByClassName(ALARM_L_CLASS_NAME).map(addAlarm);

        function addAlarm(alarmNode) {
            return createStandardAlarm(alarmNode.getDecodedParameters(), alarmNode.distName);
        }
    }

    function createStandardAlarm(parameters, distName) {
        return alarms.newAlarm(distName, parameters);
    }

    function createHistoricalAlarmsEvent() {
        var event = events.newEvent();
        event.onFirstAdded(function () {
            updateService.addHistoryListener(historical.listener);
        });
        event.onLastRemoved(function () {
            updateService.removeHistoryListener(historical.listener);
        });
        return event;
    }

    function watchHistoricalAlarms(watcher) {
        var unwatch = historical.event.watch(watcher);
        if (historical.alarms.length) watcher(historical.alarms);
        return unwatch;
    }

    function createHistoricalAlarmsListener() {
        return {
            newEntries: newEntries,
            newEntry: newEntry,
            clearEntries: clearEntries
        };

        function newEntries(entries) {
            if (!entries.length) return;

            historical.processingQueue
                .then(function () {
                    historical.isProcessing = true;
                    processEntry(entries[entries.length - 1]);
                    historical.isProcessing = false;
                })
                .then(notifyHistoricalAlarms);
        }

        function newEntry(entry) {
            if (!entry) return;

            historical.processingQueue
                .then(function () {
                    historical.isProcessing = true;
                    processEntry(entry);
                    historical.isProcessing = false;
                })
                .then(notifyHistoricalAlarms);
        }

        function clearEntries() {
            historical.processingQueue
                .then(function () {
                    historical.alarms = [];
                    historical.processedId = -1;
                })
                .then(notifyHistoricalAlarms);
        }

        function notifyHistoricalAlarms() {
            historical.event.notify(historical.alarms);
        }
    }

    function processEntry(entry) {
        while (entry.id - 1 > historical.processedId) {
            historical.processedId++;
            const entry = updateService.getHistoryEntry(historical.processedId);
            if (!entry) return;

            processAlarmsFromEntry(entry);
        }
        historical.processedId = entry.id;
        processAlarmsFromEntry(entry);
    }

    function processAlarmsFromEntry(entry) {
        if (entry.reset) {
            disappearAllAlarms(entry.timestamp);
            return;
        }
        if (!entry.change) return;

        processUpdatedAlarms(entry.change.updated, entry.timestamp, entry.metaUid);
        processRemovedAlarms(entry.change.removed, entry.timestamp);
    }

    function processRemovedAlarms(removedDistNames, timestamp) {
        if (!removedDistNames) return;

        removedDistNames.forEach(function (removedDistName) {
            if (!alarms.isAlarm(removedDistName)) return;

            const alarm = _.findLast(historical.alarms, {distName: removedDistName});
            if (!alarm) {
                LOG.error('Found alarm that should be removed, but it doesn\'t exist! Dist name: ' + distNameUtils.getName(removedDistName));
                return;
            }
            if (alarm.disappearTime) {
                LOG.info('Found alarm that has already been removed! Dist name: ' + distNameUtils.getName(removedDistName));
                return;
            }
            alarm.updateDisappearTime(timestamp);
        });
    }

    function disappearAllAlarms(timestamp) {
        _.forEachRight(historical.alarms, function (alarm) {
            if (alarm.disappeared()) return;

            alarm.updateDisappearTime(timestamp);
        });
    }

    function processUpdatedAlarms(updatedNodes, timestamp, metaUid) {
        if (!updatedNodes) return;

        const newAlarms = [];
        updatedNodes.forEach(function (updatedNode) {
            if (!alarms.isAlarm(updatedNode.distName)) return;

            const alarm = _.findLast(historical.alarms, {distName: updatedNode.distName});
            if (!alarm || alarm.disappeared()) {
                newAlarms.push(createNewHistoricalAlarm(updatedNode, metaUid, timestamp));
            } else {
                var changedAlarm = createChangedHistoricalAlarm(alarm, updatedNode, metaUid, timestamp);
                if (!changedAlarm.isEqualTo(alarm)) newAlarms.push(changedAlarm);
            }
        });
        alarms.sortUsingEventTime(newAlarms);
        historical.alarms = historical.alarms.concat(newAlarms);
    }

    function createNewHistoricalAlarm(updatedNode, metaUid, timestamp) {
        const parameters = decoderService.decodeByDistName(metaUid, updatedNode.payload, updatedNode.distName);
        return alarms.newHistoricalAlarm(updatedNode.distName, parameters, timestamp);
    }

    function createChangedHistoricalAlarm(changedAlarm, updatedNode, metaUid, timestamp) {
        const parameters = decoderService.decodeByDistName(metaUid, updatedNode.payload, updatedNode.distName);
        return alarms.newChangeAlarm(changedAlarm, updatedNode.distName, parameters, timestamp);
    }

    function getSeverityCount(alarms) {
        return {
            critical: getNumberOf('critical'),
            warning: getNumberOf('warning'),
            major: getNumberOf('major'),
            minor: getNumberOf('minor')
        };

        function getNumberOf(severity) {
            return alarms.filter(function (alarm) {
                return alarm.isNew() && alarm.hasSeverity(severity);
            }).length;
        }
    }

}
