module.exports = reasoningStatuses;

const _ = require('lodash'),
    metaServiceFactory = require('../metaService'),
    decoderFactory = require('../decoderFactory'),
    arrays = require('../arrays'),
    events = require('../events'),
    logger = require('../logger.impl');

const LOG = logger.getLogger('admin.core.reasoningStatuses');
const CLEAR_STATUS_INDICATOR = -1;

function reasoningStatuses({
        updateService,
        metaService = metaServiceFactory(),
        decoderService = decoderFactory.newInstance(metaService)
    }) {
    var abnormalities = new Map(),
    imtrees = [],
    currentEntries = [],
    lastTimeCalculations,
    isProcessing = false,
    visibility = false,
    abortAllUpdates = false,
    maxProcessingTimeMs = 50,
    event = events.newEvent();

    initializeEvent();

    return {
        isUpdating: isUpdating,
        watch: watch,
        calculateObservationTime: calculateObservationTime,
        setProcessing: setProcessing,
        setVisibility: setVisibility
    };

    function watch(watcher) {
        var unwatch = event.watch(watcher);
        if (abnormalities.size) {
            visibility ? watcher({type: 'data', data: abnormalities}) : watcher({type: 'count', data: abnormalities});
        }
        return unwatch;
    }

    function initializeEvent() {
        var historyListener = createHistoryListener(),
        imTreeListener = createImTreeListener(),
        currentEntryListener = createCurrentEntryListener();

        event.onFirstAdded(function () {
            updateService.addHistoryListener(historyListener);
            updateService.addImTreeListener(imTreeListener);
            updateService.addCurrentEntryListener(currentEntryListener);
        });
        event.onLastRemoved(function () {
            updateService.removeHistoryListener(historyListener);
            updateService.removeImTreeListener(imTreeListener);
            updateService.removeCurrentEntryListener(currentEntryListener);
            clearAbnormalities();
        });
    }

    function setProcessing(value) {
        isProcessing = value;
    }

    function setVisibility(value) {
        visibility = value;
        LOG.debug("Visibility is now set to ", visibility);
    }

    function isUpdating() {
        return isProcessing;
    }

    function createHistoryListener() {
        return {
            clearEntries: function () {
                if (!isProcessing) clearAbnormalities();
                else abortAllUpdates = true;
            }
        };
    }

    function createImTreeListener() {
        return {
            completeImTree: function (tree) {
                if (tree) {
                    imtrees.push(tree);
                    if (!isProcessing)
                        updateAbnormalities();
                }
            }
        };
    }

    function createCurrentEntryListener() {
        return {
            changeEntry: function (entry) {
                if (entry) {
                    currentEntries.push(entry);
                    if (!isProcessing)
                        updateAbnormalities();
                }
            }
        };
    }

    function decodeReasoningStatuses(updatedEntry, metaUid) {
        var parameters = decoderService.decodeByDistName(
            metaUid,
            updatedEntry.payload,
            updatedEntry.distName
        );

        var reasoningStatuses = parameters.stateInfo.reasoningStatus;
        if (!Array.isArray(reasoningStatuses))
            reasoningStatuses = Object.keys(reasoningStatuses);

        return reasoningStatuses;
    }

    function hasNoChanges(entry) {
        return entry.execution || entry.numOfChanges === 0 || _.isEmpty(entry.change);
    }

    function clearAbnormalities() {
        isProcessing = false;
        abortAllUpdates = false;
        imtrees = [];
        currentEntries = [];
        abnormalities = new Map();
        event.notify({type: 'clear', data: 0});
    }

    function entryChanged() {
        return currentEntries.length > 0 && (currentEntries.length === 1 ||
            _.first(currentEntries).id !== _.last(currentEntries).id);
    }

    function updateLastTimeCalculations() {
        lastTimeCalculations = {
            entryId: _.first(currentEntries).id,
            abnormalities: abnormalities
        };
    }

    function calculationsWaitingInQueue() {
        return imtrees.length > 1 || currentEntries.length > 1;
    }

    function updateAbnormalities() {
        if (!entryChanged()) return;
        isProcessing = true;
        if (calculationsWaitingInQueue()) {
            updateLastTimeCalculations();
            imtrees = [_.last(imtrees)];
            currentEntries = [_.last(currentEntries)];
        }
        abnormalities = new Map();
        event.notify({type: 'clear', data: CLEAR_STATUS_INDICATOR});
        if (!lastTimeCalculations || _.first(currentEntries).execution || _.first(currentEntries).reset) {
            findAbnormalitiesNormal();
            return;
        }
        findAbnormalities();
    }

    function findAbnormalities() {
        var from = lastTimeCalculations.entryId;
        var to = _.first(currentEntries).id;
        if (smallDifferenceBetweenEntries(from, to)) {
            if (isGoingDown(from, to))
                findAbnormalitiesDown();
            else
                findAbnormalitiesUp();
        }
        else findAbnormalitiesNormal();
    }

    function findAbnormalitiesNormal() {
        arrays.processUntilSuccess(
            Array.from(_.first(imtrees).distNames()),
            maxProcessingTimeMs,
            addAbnormalitiesFromDistName,
            calculateObservationTime
        );
    }

    function addAbnormalitiesFromDistName(distName) {
        if (!isProcessing || abortAllUpdates) return true;
        var node = _.first(imtrees).findByDistName(distName);
        if (!node || !hasReasoningStatus(node.distName, node.metaUid)) return false;
        var reasoningStatuses = getReasoningStatuses(node);
        if (!Array.isArray(reasoningStatuses)) reasoningStatuses = Object.keys(reasoningStatuses);
        addReasoningStatuses();
        return false;

        function addReasoningStatuses() {
            reasoningStatuses.forEach(function (reasoningStatus) {
                var rsForAbn = abnormalities.get(distName) || new Map();
                rsForAbn.set(reasoningStatus, {
                    distName: distName,
                    eventType: reasoningStatus
                });
                abnormalities.set(distName, rsForAbn);
            });
        }
    }

    function findAbnormalitiesUp() {
        arrays.processUntilSuccess(
            Array.from(_.first(imtrees).distNames()),
            maxProcessingTimeMs,
            addAbnormalitiesFromDistName,
            updateActualObservationTime
        );
    }

    function updateActualObservationTime() {
        copyTimeFromLastCalculations(_.first(currentEntries).timestamp);
        calculateObservationTime();
    }

    function copyTimeFromLastCalculations(maxTimestamp) {
        abnormalities.forEach(function (rsMap, distName) {
            if (!lastTimeCalculations.abnormalities.has(distName)) return;

            var oldRsMap = lastTimeCalculations.abnormalities.get(distName);
            rsMap.forEach(function (abnormality, reasoningStatus) {
                if (!oldRsMap.has(reasoningStatus)) return;

                var oldAbnormality = oldRsMap.get(reasoningStatus);
                if (oldAbnormality.observationTime <= maxTimestamp)
                    abnormality.observationTime = oldAbnormality.observationTime;
            });
        });
    }

    function findAbnormalitiesDown() {
        var from = lastTimeCalculations.entryId,
        to = _.first(currentEntries).id;
        abnormalities = lastTimeCalculations.abnormalities;
        arrays.processUntilSuccess(
            _.range(from + 1, to + 1, 1),
            maxProcessingTimeMs,
            addAbnormalitiesFromEntry,
            showAbnormalities
        );
    }

    function getAbnormalitiesSize() {
        var size = 0;
        abnormalities.forEach(function (rsMap) {
            size += rsMap.size;
        });
        return size;
    }

    function showAbnormalities() {
        if (abortAllUpdates) {
            clearAbnormalities();
            return;
        }

        isProcessing = false;
        if (calculationsWaitingInQueue()) {
            updateAbnormalities();
            return;
        }
        event.notify({type: 'data', data: abnormalities});
    }

    function addAbnormalitiesFromEntry(historyEntryId) {
        if (!isProcessing || abortAllUpdates) return true;

        var historyEntry = updateService.getHistoryEntry(historyEntryId);

        if (historyEntry.execution) return false;
        if (historyEntry.reset) {
            abnormalities = new Map();
            return false;
        }
        if (hasNoChanges(historyEntry)) return false;

        historyEntry.change.removed.forEach(function (removedNode) {
            abnormalities.delete(removedNode.distName);
        });
        historyEntry.change.updated.forEach(extractAbnormalitiesFromUpdatedNode);

        return false;

        function extractAbnormalitiesFromUpdatedNode(updatedNode) {
            if (!hasReasoningStatus(updatedNode.distName, historyEntry.metaUid)) return;

            var reasoningStatuses = decodeReasoningStatuses(updatedNode, historyEntry.metaUid);
            var abnForDistName = abnormalities.get(updatedNode.distName) || new Map();
            reasoningStatuses.forEach(function (reasoningStatus) {
                var abn = abnForDistName.get(reasoningStatus) || {
                        distName: updatedNode.distName,
                        eventType: reasoningStatus
                    };
                abn.observationTime = historyEntry.timestamp;
                abnForDistName.set(reasoningStatus, abn);
            });
            filterReasoningStatuses(abnForDistName, reasoningStatuses);
            if (abnForDistName.size === 0) abnormalities.delete(updatedNode.distName);
            else abnormalities.set(updatedNode.distName, abnForDistName);
        }
    }

    function filterReasoningStatuses(abnForDistName, reasoningStatuses) {
        var iterator = abnForDistName.keys(),
        key = iterator.next().value;
        while (key) {
            if (!_.includes(reasoningStatuses, key)) abnForDistName.delete(key);
            key = iterator.next().value;
        }
    }

    function smallDifferenceBetweenEntries(from, to) {
        return Math.abs(from - to) <= _.last(imtrees).size() / 10;
    }

    function smallDifferenceToNextEntry() {
        var from = _.first(currentEntries).id;
        var to = _.last(currentEntries).id;
        return Math.abs(from - to) <= _.last(imtrees).size() / 10;
    }

    function isGoingDown(from, to) {
        return to > from;
    }

    function calculateObservationTime() {
        if (abortAllUpdates) {
            clearAbnormalities();
            return;
        }
        if (calculationsWaitingInQueue() && !smallDifferenceToNextEntry()) {
            isProcessing = false;
            updateAbnormalities();
            return;
        }
        if (!isProcessing) return;

        event.notify({type: 'count', data: getAbnormalitiesSize()});

        if (!visibility || abnormalities.size === 0) {
            isProcessing = false;
            return;
        }

        var toCheck = getElementsWithEmptyTime();
        arrays.processUntilSuccess(
            _.range(_.first(currentEntries).id, -1, -1),
            maxProcessingTimeMs,
            applyTimeFromUpdatedNodes,
            showAbnormalities
        );

        function getElementsWithEmptyTime() {
            var map = new Map();
            abnormalities.forEach(function (rsMap, distName) {
                var newRsMap = new Map();
                rsMap.forEach(function (abnormality, reasoningStatus) {
                    newRsMap.set(reasoningStatus, abnormality);
                });
                if (newRsMap.size > 0) map.set(distName, newRsMap);
            });
            return map;
        }

        function applyTimeFromUpdatedNodes(historyEntryId) {
            if (!isProcessing || abortAllUpdates) return true;

            var historyEntry = updateService.getHistoryEntry(historyEntryId);

            if (historyEntry.reset) return true;
            if (hasNoChanges(historyEntry)) return false;

            historyEntry.change.updated.forEach(onUpdate);
            historyEntry.change.removed.forEach(function (change) {
                if (!toCheck.has(change.distName)) return;

                toCheck.delete(change.distName);
            });

            return toCheck.size === 0;

            function onUpdate(change) {
                if (!toCheck.has(change.distName)) return;

                var reasoningStatuses = decodeReasoningStatuses(change, historyEntry.metaUid);
                var distNameAbnormalities = toCheck.get(change.distName);

                distNameAbnormalities.forEach(function (abnormality) {
                    if (reasoningStatuses.indexOf(abnormality.eventType) === -1) return;

                    abnormality.observationTime = historyEntry.timestamp;
                    distNameAbnormalities.delete(abnormality.eventType);
                });

                if (distNameAbnormalities.size === 0)
                    toCheck.delete(change.distName);
                else
                    toCheck.set(change.distName, distNameAbnormalities);
            }
        }

    }

    function hasReasoningStatus(distName, metaUid) {
        var typeInfo = metaService.getTypeInfo(metaUid, getClassName(distName));
        var stateInfo = typeInfo.getParameterByName('stateInfo');

        return stateInfo && stateInfo.getParameterByName('reasoningStatus');
    }

    function getClassName(distName) {
        var distNameLastPart = _.first(distName.split('/').slice(-1));
        return distNameLastPart.substring(0, distNameLastPart.indexOf('-'));
    }

    function getReasoningStatuses(node) {
        if (!node) return [];

        var stateInfo = node.getDecodedParameters().stateInfo;
        return (stateInfo && stateInfo.reasoningStatus) || [];
    }

}
