const _ = require('lodash');
const imCompatibilityService = require('../../src/utils/imCompatibilityService');
const adjacencyUtilities = require('./adjacencyUtilities');
const ImTree = require('../imtree');

const validationStatus = {
    ERROR: 0,
    WARNING: 1,
    INFO: 2,
    CORRECT: 3
};

module.exports = Object({
    validationStatus: validationStatus,
    validateAdjacencies: validateAdjacencies,
    filterValidationData: filterValidationData
});

function validateAdjacencies(tree) {
    const emptyCorrect = {
        cellGroupStatus: validationStatus.CORRECT,
        localCellValidations: []
    };

    if (tree === null || tree.isEmpty())
        return emptyCorrect;

    const matcher = chooseMatcher(tree);

    tree = filterTreeWithMatcher(tree, matcher);
    if (tree.isEmpty())
        return emptyCorrect;

    return validate(tree, createClassNamesMap(matcher));
}

function chooseMatcher(tree) {
    const LObjectMatcher = {
        matches: function (node, matchCell) {
            return isLObject(node.className);

            function isLObject(className) {
                return isLRelation(className) || isLAdjacency(className) || (matchCell && className === 'LNCEL');
            }

            function isLRelation(className) {
                return className === 'LNREL' || className === 'LNRELW' || className === 'LNRELG';
            }

            function isLAdjacency(className) {
                return className === 'LNADJL' || className === 'LNADJW' || className === 'LNADJG';
            }
        }
    };

    const AObjectMatcher = {
        matches: function (node) {
            if (_.includes(node.distName, this.currentNp))
                return isAObject(node.className);
            return false;

            function isAObject(className) {
                return isARelation(className) || isAAdjacency(className) || className === 'LNCEL_A';
            }

            function isARelation(className) {
                return className === 'LNREL_A' || className === 'LNRELW_A' || className === 'LNRELG_A';
            }

            function isAAdjacency(className) {
                return className === 'LNADJL_A' || className === 'LNADJW_A' || className === 'LNADJG_A';
            }
        },
        currentNp: ''
    };

    if (lObjectsArePresent()) return LObjectMatcher;

    AObjectMatcher.currentNp = findCurrentNp();
    return AObjectMatcher;

    function lObjectsArePresent() {
        let lObjectsPresent = false;
        tree.traverse(node => {
            if (LObjectMatcher.matches(node, false))
                lObjectsPresent = true;
        });
        return lObjectsPresent;
    }

    function findCurrentNp() {
        const BTS_CONF_DN = `${imCompatibilityService.getHardwareDistName(tree)}/BTS_CONF-1`;
        const btsConf = tree.findByDistName(BTS_CONF_DN);
        return btsConf ? `${btsConf.getDecodedParameters().currentNp}/SCF-1` : null;
    }
}

function filterTreeWithMatcher(tree, matcher) {
    const filteredTree = new ImTree(tree.root.decoderService);
    tree.traverse(node => {
        if (matcher.matches(node, true)) {
            const filteredNode = filteredTree.mergeNode(node.distName, node.parameters, node.payload);
            filteredNode.metaUid = node.metaUid;
        }
    });
    return filteredTree;
}

function createClassNamesMap(matcher) {
    const classNames = new Map();
    if (!_.isUndefined(matcher.currentNp)) {
        classNames.set('LNCEL', 'LNCEL_A');
        classNames.set('LNREL', 'LNREL_A');
        classNames.set('LNRELW', 'LNRELW_A');
        classNames.set('LNRELG', 'LNRELG_A');
        classNames.set('LNADJ', 'LNADJ_A');
        classNames.set('LNADJL', 'LNADJL_A');
        classNames.set('LNADJW', 'LNADJW_A');
        classNames.set('LNADJG', 'LNADJG_A');
    } else {
        classNames.set('LNCEL', 'LNCEL');
        classNames.set('LNREL', 'LNREL');
        classNames.set('LNRELW', 'LNRELW');
        classNames.set('LNRELG', 'LNRELG');
        classNames.set('LNADJ', 'LNADJ');
        classNames.set('LNADJL', 'LNADJL');
        classNames.set('LNADJW', 'LNADJW');
        classNames.set('LNADJG', 'LNADJG');
    }
    return classNames;
}

function validate(tree, classNames) {
    const localCells = tree.findByClassName(classNames.get('LNCEL'));
    const externalCells = {
        lteCells: tree.findByClassName(classNames.get('LNADJL')),
        utranCells: tree.findByClassName(classNames.get('LNADJW')),
        geranCells: tree.findByClassName(classNames.get('LNADJG'))
    };

    const localCellValidations = validateLocalCells(localCells, externalCells, classNames);

    return {
        cellGroupStatus: evaluateCellGroupStatus(localCellValidations),
        localCellValidations: localCellValidations
    };
}

function evaluateCellGroupStatus(localCellValidations) {
    let cellGroupStatus = validationStatus.CORRECT;
    for (let i = 0; i < localCellValidations.length; i++) {
        const localCellValidation = localCellValidations[i];
        if (localCellValidation.localCellStatus < cellGroupStatus)
            cellGroupStatus = localCellValidation.localCellStatus;
    }
    return cellGroupStatus;
}

function validateLocalCells(localCells, externalCells, classNames) {
    const localCellValidations = [];
    for (let i = 0; i < localCells.length; i++) {
        const localCell = localCells[i];
        const localCellValidation = validateLocalCell(localCell, externalCells, classNames);
        if (localCellValidation.technologyValidations.length)
            localCellValidations.push(localCellValidation);
    }
    return localCellValidations;
}

function validateLocalCell(localCell, externalCells, classNames) {
    const lteRelations = localCell.findByClassName(classNames.get('LNREL'));
    const utranRelations = localCell.findByClassName(classNames.get('LNRELW'));
    const geranRelations = localCell.findByClassName(classNames.get('LNRELG'));

    const uncheckedValidations = [
        validateTechnologyRelations(lteRelations, externalCells.lteCells, lteRelationMatches, 'LTE'),
        validateTechnologyRelations(utranRelations, externalCells.utranCells, utranRelationMatches, 'UTRAN'),
        validateTechnologyRelations(geranRelations, externalCells.geranCells, geranRelationMatches, 'GERAN')
    ];

    const technologyValidations = [];
    _.each(uncheckedValidations, technologyValidation => {
        if (technologyValidation.relationStatuses.length || technologyValidation.cellStatuses.length)
            technologyValidations.push(technologyValidation);
    });

    return {
        localCellStatus: evaluateLocalCellStatus(technologyValidations),
        localCell: localCell,
        technologyValidations: technologyValidations
    };
}

function evaluateLocalCellStatus(technologyValidations) {
    let localCellStatus = validationStatus.CORRECT;
    for (let i = 0; i < technologyValidations.length; i++) {
        const status = technologyValidations[i].technologyRelationsStatus;
        if (status < localCellStatus)
            localCellStatus = status;
    }
    return localCellStatus;
}

function lteRelationMatches(lteRelation, lteCell) {
    return lteRelation.getDecodedParameters().ecgiLcrId === lteCell.getDecodedParameters().ecgiLcrId &&
        lteRelation.getDecodedParameters().ecgiAdjEnbId === lteCell.getDecodedParameters().ecgiAdjEnbId;
}

function utranRelationMatches(utranRelation, utranCell) {
    return utranRelation.getDecodedParameters().uTargetCid === utranCell.getDecodedParameters().uTargetCid;
}

function geranRelationMatches(geranRelation, geranCell) {
    return geranRelation.getDecodedParameters().gTargetCi === geranCell.getDecodedParameters().gTargetCi;
}

function validateTechnologyRelations(relations, cells, relationMatches, technology) {
    const matching = new Map();
    const missing = [];

    for (let i = 0; i < relations.length; i++) {
        const relation = relations[i];
        validateRelation(relation, matching, missing, cells, relationMatches);
    }

    const relationStatuses = evaluateRelationStatuses(missing, matching);
    const cellStatuses = evaluateCellStatuses(matching, cells);
    return {
        technologyRelationsStatus: evaluateTechnologyRelationsStatus(relationStatuses, cellStatuses),
        relationStatuses: relationStatuses,
        cellStatuses: cellStatuses,
        technology: technology
    };
}

function evaluateTechnologyRelationsStatus(relationStatuses, cellStatuses) {
    let relationsGroupStatus = validationStatus.CORRECT;
    if (relationStatuses.errors.length !== 0)
        relationsGroupStatus = validationStatus.ERROR;
    else if (relationStatuses.warnings.length !== 0)
        relationsGroupStatus = validationStatus.WARNING;

    let cellGroupStatus = validationStatus.CORRECT;
    if (cellStatuses.infos.length !== 0)
        cellGroupStatus = validationStatus.INFO;

    if (relationsGroupStatus < cellGroupStatus)
        return relationsGroupStatus;
    return cellGroupStatus;
}

function validateRelation(relation, matching, missing, cells, relationMatches) {
    let matched = false;

    for (let i = 0; i < cells.length; i++) {
        const cell = cells[i];
        if (relationMatches(relation, cell)) {
            const matchingRelations = matching.get(cell.distName) || [];
            matchingRelations.push(relation);
            matching.set(cell.distName, matchingRelations);
            matched = true;
            break;
        }
    }
    if (!matched)
        missing.push(relation);
}

function evaluateRelationStatuses(missing, matching) {
    const error = [];
    const correct = [];

    matching.forEach((value, key) => {
        const matchingRelations = matching.get(key);
        const correctRelation = matchingRelations.shift();
        correctRelation.matchingCell = key;
        correct.push(correctRelation);

        if (matchingRelations.length !== 0)
            for (let j = 0; j < matchingRelations.length; j++) {
                const errorRelation = matchingRelations[j];
                error.push(errorRelation);
            }
    });

    return {
        length: missing.length + error.length + correct.length,
        warnings: missing,
        errors: error,
        corrects: correct
    };
}

function evaluateCellStatuses(matching, cells) {
    const info = [];
    const correct = [];

    for (let i = 0; i < cells.length; i++) {
        const cell = cells[i];
        if (matching.get(cell.distName))
            correct.push(cell);
        else
            info.push(cell);
    }

    return {
        length: info.length + correct.length,
        infos: info,
        corrects: correct
    };
}

function filterValidationData(filter, validationData) {
    if (!filter) return validationData;

    const filteredValidationData = _.merge({}, validationData); // angular.copy() fails here, idk why

    _.eachRight(filteredValidationData.localCellValidations, (localCellValidation, localCellIndex) => {
        _.each(localCellValidation.technologyValidations, (techValidation) => {
            filterObjects(techValidation.cellStatuses.corrects);
            filterObjects(techValidation.cellStatuses.infos);

            filterLNRELsThatDontHaveRelations(techValidation.relationStatuses, techValidation.cellStatuses);
        });

        if (!cellHasSomeObjectsToShow(localCellValidation))
            if (shouldFilterOutObject(localCellValidation.localCell))
                _.pullAt(filteredValidationData.localCellValidations, localCellIndex);
    });
    return filteredValidationData;

    function filterLNRELsThatDontHaveRelations(relationStatuses, cellStatuses) {
        relationStatuses.warnings = [];
        relationStatuses.errors = [];

        const dnToObjectMap = createMapDnToObject(cellStatuses.corrects);
        _.eachRight(relationStatuses.corrects, (correct, index) => {
            if (!dnToObjectMap[correct.matchingCell])
                _.pullAt(relationStatuses.corrects, index);
        });
    }

    function createMapDnToObject(objects) {
        const map = {};
        _.each(objects, (object) => {
            map[object.distName] = object;
        });
        return map;
    }

    function cellHasSomeObjectsToShow(cellObject) {
        let objectsCount = 0;
        _.each(cellObject.technologyValidations, (technologyValidation) => {
            objectsCount +=
                technologyValidation.cellStatuses.corrects.length +
                technologyValidation.cellStatuses.infos.length +
                technologyValidation.relationStatuses.corrects.length +
                technologyValidation.relationStatuses.warnings.length +
                technologyValidation.relationStatuses.errors.length;
        });
        return objectsCount > 0;
    }

    function filterObjects(objects) {
        _.eachRight(objects, (object, index) => {
            if (shouldFilterOutObject(object))
                _.pullAt(objects, index);
        });
    }

    function shouldFilterOutObject(object) {
        let shouldFilterOut = true;
        const paramsValues = getValuesToSearchIn(object);
        _.each(paramsValues, (param) => {
            if (param && param.value && param.value.toString() === filter) {
                shouldFilterOut = false;
                return;
            }
        });

        return shouldFilterOut;
    }

    function getValuesToSearchIn(object) {
        return [
            adjacencyUtilities.getCellId(object),
            adjacencyUtilities.getDownlinkFrequency(object),
            adjacencyUtilities.getUplinkFrequency(object),
            adjacencyUtilities.getMnc(object),
            adjacencyUtilities.getMcc(object)
        ];
    }
}

