module.exports = abnormalities;

const Promise = require('bluebird');
const ADMIN_API_STATUSES = require('../adminApiStatuses');
const adminMessageValidator = require('../adminMessageValidator');

function abnormalities({
    alarmsMonitor,
    reasoningStatuses,
    onlineImTree,
    connectionService,
    connectionInitializer = require('../connectionInitializer')(onlineImTree, connectionService)
    }) {

    const ACTIONS = {
        currentAlarms: getCurrentAlarms,
        historicalAlarms: getHistoricalAlarms,
        other: getOther
    };

    return {
        execute
    };

    function execute(msgData, completionListener) {
        const validationResult = adminMessageValidator.validate(msgData, getAbnormalitiesSchema());
        if (!validationResult.status) {
            completionListener(ADMIN_API_STATUSES.FAILED, validationResult.message);
            return Promise.reject();
        }
        const type = msgData.parameters.type;
        const action = ACTIONS[type];
        if (!action) {
            completionListener(ADMIN_API_STATUSES.FAILED, `Unrecognized type of abnormalities: ${type}`);
            return Promise.reject();
        }
        const actionWithSeverity = () => action(msgData.parameters.severity);
        return prepareAndRun(actionWithSeverity, completionListener);
    }

    function prepareAndRun(action, completionListener) {
        return connectionInitializer.connectToBts()
            .prepare()
            .then(action)
            .then(result => {
                completionListener(ADMIN_API_STATUSES.COMPLETED, result);
            })
            .catch(reason => {
                completionListener(ADMIN_API_STATUSES.FAILED, reason);
                return Promise.reject(reason);
            });
    }

    function getCurrentAlarms(severity) {
        return getAlarms(alarmsMonitor.watchCurrentAlarms, severity);
    }

    function getHistoricalAlarms(severity) {
        return getAlarms(alarmsMonitor.watchHistoricalAlarms, severity);
    }

    function getAlarms(watchAction, severity) {
        return new Promise(resolve => {
            let alarmsGathered = false;
            const unwatcher = watchAction(onNewAlarms);
            if (alarmsGathered) unwatcher.unwatch();

            function onNewAlarms(alarms) {
                let filteredAlarms = alarms;
                if (severity) filteredAlarms = alarms.filter(alarm => alarm.hasSeverity(severity));
                resolve(filteredAlarms);
                alarmsGathered = true;
                if (unwatcher) unwatcher.unwatch();
            }
        });
    }

    function getOther() {
        return new Promise(resolve => {
            let unwatchNotPerformed = false;
            reasoningStatuses.setVisibility(true);
            const unwatcher = reasoningStatuses.watch(onOthers);
            if (unwatchNotPerformed) unwatcher.unwatch();
            function onOthers(event) {
                if (event.type === 'data') resolve(event.data);
                if (unwatcher) unwatcher.unwatch();
                else unwatchNotPerformed = true;
                reasoningStatuses.setVisibility(false);
            }
        });
    }

    function getAbnormalitiesSchema() {
        return {
            type: 'object',
            id: 'infoModel',
            properties: {
                name: {
                    type: 'string'
                },
                parameters: {
                    type: 'object',
                    id: 'parameters',
                    properties: {
                        type: {
                            type: 'string'
                        },
                        severity: {
                            type: 'string'
                        }
                    },
                    required: ['type']
                }
            },
            required: ['name', 'parameters']
        };
    }

}
