module.exports = adjacencies;

const Promise = require('bluebird');
const getCellsActionFactory = require('./getCellsAction');
const getNbrDataActionFactory = require('./getNbrDataAction');
const adminMessageValidator = require('../../adminMessageValidator');
const ADMIN_API_STATUSES = require('../../adminApiStatuses');

function adjacencies({
    onlineImTree,
    adjacencyValidation,
    connectionService,
    connectionInitializer = require('../../connectionInitializer')(onlineImTree, connectionService)
    }) {

    const getCellsAction = getCellsActionFactory({
        onlineImTree: onlineImTree,
        adjacencyValidation: adjacencyValidation
    });

    const getNbrAction = getNbrDataActionFactory({
        onlineImTree: onlineImTree,
        adjacencyValidation: adjacencyValidation
    });

    const ACTIONS = {
        getCells: getCellsAction.execute,
        getNbrData: getNbrAction.execute
    };

    return Object.freeze({
        execute
    });

    function executor(msgData) {

        return new Promise((resolve, reject) => {
            ACTIONS[msgData.parameters.procedureName](msgData, resolve, reject);
        });
    }

    function execute(msgData, completionListener) {
        const validationResult = adminMessageValidator.validate(msgData, getProcedureSchema());
        if (!validationResult.status) {
            completionListener(ADMIN_API_STATUSES.FAILED, validationResult.message);
            return Promise.reject();
        }
        return connectionInitializer.connectToBts()
            .prepare()
            .then(() => executor(msgData))
            .then((result) => {
                completionListener(ADMIN_API_STATUSES.COMPLETED, result);
            })
            .catch(reason => {
                completionListener(ADMIN_API_STATUSES.FAILED, reason);
                return Promise.reject(reason);
            });
    }

    function getProcedureSchema() {
        return {
            type: 'object',
            id: 'infoModel',
            properties: {
                name: {type: 'string'},
                parameters: {
                    type: 'object',
                    id: 'parameters',
                    properties: {
                        procedureName: {type: 'string'},
                        cellName: {type: 'string'}
                    },
                    required: ['procedureName']
                }
            },
            required: ['name', 'parameters']
        };
    }
}
