module.exports = baseband;

const Promise = require('bluebird');
const ADMIN_API_STATUSES = require('../adminApiStatuses');
const adminMessageValidator = require('../adminMessageValidator');

function baseband({
    dspsService,
    onlineImTree,
    connectionService,
    connectionInitializer = require('../connectionInitializer')(onlineImTree, connectionService)
    }) {

    const BASEBAND_EXECUTORS = {
        names: names,
        baseband: getBaseband
    };

    return {
        execute: execute
    };

    function execute(msgData, completionListener) {
        const validationResult = adminMessageValidator.validate(msgData, getRequestSchema());
        if (!validationResult.status) {
            completionListener(ADMIN_API_STATUSES.FAILED, validationResult.message);
            return;
        }

        prepareAndRun(watchDSPS, completionListener);

        function watchDSPS() {
            dspsService.watch(watcher);
        }

        function watcher(event) {
            var cells = event.cells;
            cells.forEach(function (cell) {
                cell.distName = cell.distName.split('/').pop();
            });

            const filter = msgData.parameters.cells;
            if (filter) {
                cells = cells.filter(function(cell) {
                    return (filter.indexOf(cell.distName) > -1);
                });
            }

            completionListener(ADMIN_API_STATUSES.COMPLETED, dispatch(cells));
            dspsService.unwatch(watcher);
        }

        function dispatch(cells) {
            if (msgData.parameters.onlyNames) return BASEBAND_EXECUTORS.names(cells);
            else return BASEBAND_EXECUTORS.baseband(cells);
        }

        function prepareAndRun(executor, completionListener) {
            return connectionInitializer.connectToBts()
                .prepare()
                .then(executor)
                .catch((reason) => {
                    completionListener(ADMIN_API_STATUSES.FAILED, reason);
                    return Promise.reject(reason);
                });
        }
    }

    function getBaseband(cells) {
        var baseband = [];
        cells.forEach(function (cell) {
            baseband.push(basebandFilter(cell));
        });

        return baseband;
    }

    function basebandFilter(cell) {
        var baseband = {};
        baseband.distName = cell.distName;
        baseband.label = cell.label;
        baseband.channels = [];
        cell.channels.forEach(function (channel) {
            let basebandChannel = {};
            basebandChannel.label = channel.label;
            basebandChannel.rmod = channel.rmod.label;
            basebandChannel.smod = createSmod(channel);
            baseband.channels.push(basebandChannel);
        });
        return baseband;
    }

    function createSmod(channel) {
        return {
            label: channel.smodGroup.label,
            prodName: channel.smodGroup.prodName,
            prodCode: channel.smodGroup.prodCode,
            prodCodePlanned: channel.smodGroup.prodCodePlanned
        }
    }

    function names(cells) {
        var names = [];
        cells.forEach(function (cell) {
            names.push(cell.distName);
        });
        return names;
    }

    function getRequestSchema() {
        return {
            type: 'object',
            id: 'infoModel',
            properties: {
                name: {
                    type: 'string'
                },
                parameters: {
                    type: 'object',
                    id: 'parameters',
                    properties: {
                        cells: {
                            type: 'array',
                            minItems: 1
                        },
                        onlyNames: {
                            type: 'boolean'
                        }
                    },
                    required: ['onlyNames']
                }
            },
            required: ['name', 'parameters']
        }
    }

}


