module.exports = {
    newInstance
};

const Promise = require('bluebird');
const ADMIN_API_STATUSES = require('../adminApiStatuses');
const sfpConnections = require('../../sfp/sfpConnections');
const adminMessageValidator = require('../adminMessageValidator');

function newInstance({
    onlineImTree,
    connectionService,
    connectionInitializer = require('../connectionInitializer')(onlineImTree, connectionService)
    }) {

    const ACTIONS = {
        getLinks,
        getLinksWithParameters,
        getParametersForLink
    };

    return {
        execute
    };

    function getConnections(completeImTree) {
        return sfpConnections.create(completeImTree);
    }

    function createLinkList(completeImTree) {
        return getConnections(completeImTree).map(obj => ({
            cablnkDistName: obj.cablnkDistName,
            source: obj.destination.distName,
            destination: obj.source.distName
        }));
    }

    function parametersForParticularLink(completeImTree, distName) {
        return getConnections(completeImTree).find(obj => obj.cablnkDistName === distName);
    }

    function execute(msgData, completionListener) {
        const validationResult = adminMessageValidator.validate(msgData, getProcedureSchema());
        if (!validationResult.status) {
            completionListener(ADMIN_API_STATUSES.FAILED, validationResult.message);
            return Promise.reject();
        }

        const action = ACTIONS[msgData.parameters.action];
        if (!action) {
            completionListener(ADMIN_API_STATUSES.FAILED, `No "${msgData.parameters.action}" sfp action.`);
            return Promise.reject();
        }

        return connectionInitializer.connectToBts()
            .prepare()
            .then(() => action(msgData, completionListener))
            .then((result) => {
                completionListener(ADMIN_API_STATUSES.COMPLETED, result);
            })
            .catch(reason => {
                completionListener(ADMIN_API_STATUSES.FAILED, reason);
                return Promise.reject(reason);
            });
    }

    function getLinks() {
        return onlineImTree.getImTree()
            .then(completeImTree => Promise.resolve(createLinkList(completeImTree)));
    }

    function getLinksWithParameters() {
        return onlineImTree.getImTree()
            .then(completeImTree => Promise.resolve(getConnections(completeImTree)));
    }

    function getParametersForLink(msgData) {
        if (!msgData.parameters.distName) {
            return Promise.reject('/parameters, Missing required property: distName');
        }

        return new Promise((resolve, reject) => {
            onlineImTree.getImTree().then((completeImTree) => {
                const distName = msgData.parameters.distName;
                const links = parametersForParticularLink(completeImTree, distName);
                if (!links) reject(`No object with distName ${distName}`);
                else resolve(links);
            });
        });
    }

    function getProcedureSchema() {
        return {
            type: 'object',
            id: 'infoModel',
            properties: {
                name: {type: 'string'},
                parameters: {
                    properties: {
                        action: {type: 'string'},
                        distName: {type: 'string'}
                    },
                    required: ['action']
                }
            },
            required: ['name', 'parameters']
        };
    }
}
