module.exports = initCli;

const initializers = require('./initializers');
const metaService = require('../metaService');
const decoderFactory = require('../decoderFactory');
const updateService = require('../updateService');
const sequentialExecutorFactory = require('../sequentialExecutorFactory');
const onlineImTree = require('../onlineImTree');
const imTracker = require('../imTracker/imTracker');
const translatingBtsClient = require('../communication/translatingBtsClient');
const objectIdMapper = require('../communication/objectIdMapper');
const heartBeatHandler = require('../communication/btsClient.heartBeatHandler');
const connectionService = require('../connection/connectionService');
const signInHandlerService = require('../connection/signInHandlerService');
const btsServiceProcedure = require('../procedures/btsServiceProcedure');
const procedureIdFactory = require('../procedures/procedureIdFactory');
const procedureExecutionCleanerService = require('../procedures/procedureExecutionCleanerService');
const procedureExecutor = require('../procedures/procedureExecutor');
const ims2Downloader = require('../procedures/imSnapshot/ims2Downloader');
const snapshotExecutor = require('../procedures/snapshot/snapshotExecutor');
const fileDownloaderFactory = require('../procedures/snapshot/fileDownloaderFactory');
const snapshotStatusHandler = require('../procedures/snapshot/snapshotStatusHandler');
const filesInfoFactory = require('../procedures/snapshot/filesInfoFactory');
const procedureTrackerFactory = require('../procedures/procedureTrackerFactory');
const executionRequestsDispatcher = require('../operations/executionRequestsDispatcher');
const operationsService = require('../operations/operationsService');
const alarmsMonitor = require('../abnormalities/alarmsMonitor');
const btsInfoService = require('../utils/btsInfoService');
const dspsService = require('../dsps/dspsService');
const fileTransferFactory = require('../fileTransfer/fileTransferFactory');
const authService = require('../authentication/authService');
const currentSettingsService = require('../settings/currentSettingsService');
const softwareZipDispatcherFactory = require('../procedures/swUpdate/openSoftwareZip/softwareZipDispatcherFactory');
const FileUploadService = require('../fileTransfer/fileUploadService');
const applicationInfo = require('../applicationInfo/applicationInfo');
const swUpdateExecutorFactory = require('../procedures/swUpdate/swUpdateExecutorFactory');

function initCli({localStorage, sessionStorage}) {
    const metaServiceInst = metaService();
    const objectIdMapperInst = objectIdMapper.newInstance();
    const socket = translatingBtsClient.newInstance({
        metaService: metaServiceInst,
        objectIdMapper: objectIdMapperInst
    });
    const decoderServiceInst = decoderFactory.newInstance(metaServiceInst);
    const sequentialExecutorInst = sequentialExecutorFactory.newInstance();
    const updateServiceInst = updateService({
        socket: socket,
        metaService: metaServiceInst,
        decoderService: decoderServiceInst,
        executor: sequentialExecutorInst
    });
    const onlineImTreeInst = onlineImTree({
        socket: socket,
        metaService: metaServiceInst,
        decoderService: decoderServiceInst,
        executor: sequentialExecutorInst
    });
    const imTrackerInst = imTracker({
        onlineImTree: onlineImTreeInst
    });
    const procedureIdFactoryInst = procedureIdFactory(localStorage);
    const btsServiceProcedureInst = btsServiceProcedure(imTrackerInst, procedureIdFactoryInst);
    const signInHandlerServiceInst = signInHandlerService({
        socket: socket
    });
    const imAuthInst = authService.imAuthFactory({
        socket: socket
    }).newInstance(sessionStorage);
    const currentSettingsServiceInst = currentSettingsService({
        localStorage: localStorage
    });
    const connectionServiceInst = connectionService({
        updateService: updateServiceInst,
        signInHandlerService: signInHandlerServiceInst,
        imAuth: imAuthInst,
        currentSettings: currentSettingsServiceInst
    });
    const executionRequestsDispatcherInst = executionRequestsDispatcher({
        translatingBtsClient: socket,
        metaService: metaServiceInst,
        decoderService: decoderServiceInst
    });
    const operationsServiceInst = operationsService({
        translatingBtsClient: socket,
        metaService: metaServiceInst,
        executionRequestsDispatcher: executionRequestsDispatcherInst,
        onlineImTree: onlineImTreeInst
    });
    const procedureExecutorInst = procedureExecutor(operationsServiceInst, procedureIdFactoryInst);
    const fileTransferInst = fileTransferFactory({
        imAuth: imAuthInst
    });
    const ims2DownloaderInst = ims2Downloader({
        procedureExecutor: procedureExecutorInst,
        btsServiceProcedure: btsServiceProcedureInst,
        fileTransfer: fileTransferInst,
        onlineImTree: onlineImTreeInst,
        procedureIdFactory: procedureIdFactoryInst
    });
    const snapshotExecutorInst = snapshotExecutor(operationsServiceInst);
    const fileDownloaderFactoryInst = fileDownloaderFactory(snapshotExecutorInst, fileTransferInst);
    const snapshotStatusHandlerInst = snapshotStatusHandler(onlineImTreeInst, socket);
    const procedureTrackerInst = procedureTrackerFactory(onlineImTreeInst);
    const filesInfoFactoryInst = filesInfoFactory();
    const adminApiSnapshot = initializers.adminApiSnapshotInitializer(
        snapshotExecutorInst,
        filesInfoFactoryInst,
        snapshotStatusHandlerInst,
        procedureIdFactoryInst,
        procedureTrackerInst,
        fileTransferInst,
        fileDownloaderFactoryInst,
        onlineImTreeInst
    );
    const alarmsMonitorInst = alarmsMonitor(
        updateServiceInst,
        metaServiceInst,
        decoderServiceInst
    );
    const btsInfoServiceInst = btsInfoService.newInstance(updateServiceInst);
    const dspsServiceInst = dspsService({
        updateService: updateServiceInst
    });

    procedureExecutionCleanerService({
        btsServiceProcedure: btsServiceProcedureInst,
        operationsService: operationsServiceInst
    }).enable();

    const AUTO_RECONNECT_INTERVAL_MS = 5000;
    signInHandlerServiceInst.enableAutoReconnect(AUTO_RECONNECT_INTERVAL_MS);

    heartBeatHandler(socket).start();

    const softwareZipDispatcherFactoryInst = softwareZipDispatcherFactory();
    const softwareZipDispatcherInst = softwareZipDispatcherFactoryInst.getDispatcher();
    const fileUploadServiceInst = new FileUploadService(imAuthInst, softwareZipDispatcherInst, applicationInfo);
    const swUpdateExecutorFactoryInst = swUpdateExecutorFactory(operationsServiceInst, fileUploadServiceInst, imTrackerInst);

    initializers.initializeProceduresHandler(
        btsServiceProcedureInst,
        procedureExecutorInst,
        onlineImTreeInst,
        connectionServiceInst,
        procedureIdFactoryInst,
        ims2DownloaderInst,
        adminApiSnapshot,
        imAuthInst,
        softwareZipDispatcherFactoryInst,
        swUpdateExecutorFactoryInst,
        imTrackerInst
    );

    initializers.initializeInfoModelHandler(
        alarmsMonitorInst,
        onlineImTreeInst,
        connectionServiceInst,
        btsInfoServiceInst,
        dspsServiceInst
    );
}
