module.exports = distNameProcedures;

const Promise = require('bluebird');
const ADMIN_API_STATUSES = require('../adminApiStatuses');
const adminMessageValidator = require('../adminMessageValidator');
const distNameUtils = require('../../utils/distNameUtils');

function distNameProcedures({
    btsServiceProcedure,
    procedureExecutor,
    onlineImTree,
    connectionService,
    procedureIdFactory = require('../../procedures/procedureIdFactory')({}),
    remoteProcedureListener = require('./remoteProcedureListener')(btsServiceProcedure),
    connectionInitializer = require('../connectionInitializer')(onlineImTree, connectionService)
    }) {

    const EXECUTORS = {
        block: procedureExecutor.executeBlock,
        unblock: procedureExecutor.executeUnblock,
        reset: procedureExecutor.executeReset
    };

    return {
        executeBlock,
        executeUnblock,
        executeRadioReset
    };

    function executeBlock(msgData, completionListener) {
        return validateAndExecute(msgData, completionListener, EXECUTORS.block);
    }

    function executeUnblock(msgData, completionListener) {
        return validateAndExecute(msgData, completionListener, EXECUTORS.unblock);
    }

    function executeRadioReset(msgData, completionListener) {
        if (!isDataValid(msgData, completionListener)) return Promise.reject();
        const procedureDistName = msgData.parameters.distName;
        if (distNameUtils.getClassName(procedureDistName) !== 'RMOD_L') {
            completionListener(ADMIN_API_STATUSES.FAILED, createUnsupportedDistNameError(procedureDistName, 'reset'));
            return Promise.resolve();
        }

        return execute(msgData, completionListener, EXECUTORS.reset);
    }

    function createUnsupportedDistNameError(distName, procedureName) {
        return `Distname ${distName} is unnsuported for ${procedureName} procedure`;
    }

    function validateAndExecute(msgData, completionListener, executor) {
        const validationResult = adminMessageValidator.validate(msgData, getProcedureSchema());
        if (!validationResult.status) {
            completionListener(ADMIN_API_STATUSES.FAILED, validationResult.message);
            return Promise.resolve();
        }
        return execute(msgData, completionListener, executor);
    }

    function execute(msgData, completionListener, executor) {
        const procedureId = procedureIdFactory.getId().id;
        listenUntilCompleted(completionListener, procedureId);

        const simplifiedExecutorWithId = createSimplifiedExecutorWithId(executor, procedureId);
        return prepareAndRun(simplifiedExecutorWithId, msgData.parameters.distName, completionListener);
    }

    function isDataValid(msgData, completionListener) {
        const validationResult = adminMessageValidator.validate(msgData, getProcedureSchema());
        if (!validationResult.status) {
            completionListener(ADMIN_API_STATUSES.FAILED, validationResult.message);
            return false;
        }
        return true;
    }

    function createSimplifiedExecutorWithId(executor, procedureId) {
        return function (distName) {
            const noop = () => {};
            executor('', distName, noop, {
                clientId: procedureId
            });
        };
    }

    function prepareAndRun(executor, distName, completionListener) {
        return connectionInitializer.connectToBts()
            .checkIfDnExists(distName)
            .prepare()
            .then(() => executor(distName))
            .catch((reason) => {
                completionListener(ADMIN_API_STATUSES.FAILED, reason);
                remoteProcedureListener.unregister();
                return Promise.reject(reason);
            });
    }

    function getProcedureSchema() {
        return {
            type: 'object',
            id: 'procedure',
            properties: {
                name: 'string',
                parameters: {
                    type: 'object',
                    id: 'parameters',
                    properties: {
                        distName: 'string'
                    },
                    required: ['distName']
                }
            },
            required: ['name', 'parameters']
        };
    }

    function listenUntilCompleted(completionListener, procedureId) {
        remoteProcedureListener.create(procedureId, completionListener);
        remoteProcedureListener.register();
    }
}
