module.exports = proceduresDispatcher;

const ADMIN_API_STATUSES = require('../adminApiStatuses');
const adminApi = require('../adminApi');
const adminMessageValidator = require('../adminMessageValidator');
const logger = require('../../logger.impl');
const LOG = logger.getLogger('admin.core.remoteApi.procedures');

function proceduresDispatcher({
    distNameProcedures,
    imSnapshotProcedure,
    snapshotProcedure,
    commissioningProcedure,
    resetProcedure,
    swUpdateProcedure,
    otherProcedure
    }) {

    const EXECUTORS = {
        block: distNameProcedures.executeBlock,
        unblock: distNameProcedures.executeUnblock,
        siteReset: resetProcedure.executeSiteReset,
        btsReset: resetProcedure.executeBtsReset,
        radioReset: distNameProcedures.executeRadioReset,
        imSnapshot: imSnapshotProcedure.execute,
        snapshot: snapshotProcedure.execute,
        downloadPlan: commissioningProcedure.executeDownloadPlan,
        activatePlan: commissioningProcedure.executeActivatePlan,
        swRollback: otherProcedure.executeSwRollback,
        resetToTestDedicated: resetProcedure.executeResetToTestDedicated,
        configurationReset: resetProcedure.executeConfigurationReset,
        softwareUpdate: swUpdateProcedure.execute,
        changeEthernetPortSecurity: otherProcedure.executeChangeEthernetPortSecurity
    };

    adminApi.addHandler('procedure', dispatch);

    return {
        // visible for testing
        dispatch
    };

    function dispatch(msgData, completionListener) {
        const validationResult = adminMessageValidator.validate(msgData, getProcedureSchema());
        if (!validationResult.status) {
            completionListener(ADMIN_API_STATUSES.FAILED, validationResult.message);
            return;
        }

        const remoteExecutor = EXECUTORS[msgData.name];
        if (!remoteExecutor) {
            completionListener(ADMIN_API_STATUSES.FAILED, `No ${msgData.name} procedure in testing API`);
            return;
        }

        remoteExecutor(msgData, completionListener)
            .catch(reason => LOG.info(`Operation failed: ${reason}`));
    }

    function getProcedureSchema() {
        return {
            type: 'object',
            id: 'procedure',
            properties: {
                name: {
                    type: 'string'
                }
            },
            required: ['name']
        };
    }
}
