const Promise = require('bluebird');

const adminApiStatuses = require('../adminApiStatuses');

module.exports = swUpdateProcedure;

const SW_UPDATE_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/SW_UPDATE_L-1';
const SW_UPDATE_NOT_FOUND = `Object not found at: ${SW_UPDATE_DISTNAME}`;
const SW_BUSY_MESSAGE = 'Another update is in progress';
const SW_TIMOEOUTMESSAGE = 'Procedure start time exceeded';
const PROCEDURE_START_MAX_TIME = 60000;

const BUSY = 'Busy';
const SUCCESS = 'Success';

const SOURCE_REMOTE = 'remote';
const SOURCE_ZIP = 'zip';

function swUpdateProcedure({
    softwareZipDispatcherFactory,
    swUpdateExecutorFactory,
    onlineImTree,
    imTracker,
    connectionService,
    connectionInitializer = require('../connectionInitializer')(onlineImTree, connectionService)
    }) {

    let msgParameters;
    let msgCompletionListener;
    let swUpdateListener;

    return {
        execute
    };

    function execute(msgData, completionListener) {
        if (msgCompletionListener)
            return failed(SW_BUSY_MESSAGE);

        msgParameters = msgData.parameters;
        msgCompletionListener = completionListener;

        return connectionInitializer
            .connectToBts()
            .prepare()
            .then(swUpdateCheckState)
            .then(executeAndFinish)
            .catch(failed);
    }

    function swUpdateCheckState() {
        return onlineImTree
                .getImTree()
                .then(imTree => {
                    const node = imTree.findByDistName(SW_UPDATE_DISTNAME);
                    if (!node)
                        return Promise.reject(SW_UPDATE_NOT_FOUND);
                    if (node.getDecodedParameters().stateInfo.swUpdateState !== BUSY)
                        return Promise.resolve();
                    return Promise.reject(SW_BUSY_MESSAGE);
                });
    }

    function executeUpdate() {
        let softwareUpdateConfig;

        if (msgParameters.file) {
            const buffer = new Buffer(msgParameters.file, 'base64');
            const arrayBuffer = new Uint8Array(buffer).buffer;
            const targetBd = softwareZipDispatcherFactory
                .fromZipFile(arrayBuffer)
                .getDispatcher()
                .getTargetBd()
                .name;

            softwareUpdateConfig = {
                shouldActivate: msgParameters.shouldActivate,
                targetBd: targetBd,
                source: SOURCE_ZIP
            };
        } else {
            softwareUpdateConfig = msgParameters;
            softwareUpdateConfig.source = SOURCE_REMOTE;
        }

        return swUpdateExecutorFactory
            .newSoftwareUpdateExecutor(softwareUpdateConfig)
            .execute();
    }

    function expectSuccess() {
        const swUpdatePromise = new Promise((resolve, reject) => {
        const timeoutId = setTimeout(reject, PROCEDURE_START_MAX_TIME, SW_TIMOEOUTMESSAGE);
            swUpdateListener = {
                onParameterChanged: event => {
                    const swUpdateState = event.parameters.stateInfo.swUpdateState;
                    if (swUpdateState === SUCCESS) {
                        msgCompletionListener(adminApiStatuses.COMPLETED);
                        clean();
                        resolve();
                    } else if (swUpdateState === BUSY) {
                        clearTimeout(timeoutId);
                    } else {
                        reject(swUpdateState);
                    }
                }
            };
        });
        const imTrackerPromise = imTracker.trackByDistName(SW_UPDATE_DISTNAME, swUpdateListener, ['stateInfo']);

        return Promise.all([swUpdatePromise, imTrackerPromise]);
    }

    function executeAndFinish() {
        return Promise.all([expectSuccess(), executeUpdate()]);
    }

    function failed(error) {
        msgCompletionListener(adminApiStatuses.FAILED, error && error.toString());
        clean();
        return Promise.reject(error);
    }

    function clean() {
        msgCompletionListener = null;
        imTracker.stopTracking(swUpdateListener, SW_UPDATE_DISTNAME);
    }
}
