module.exports = {
    imAuth,
    imAuthFactory
};

const Rusha = require('rusha');
const _ = require('lodash');
const loggers = require('../logger.impl');
const events = require('../events');
const applicationInfo = require('../applicationInfo/applicationInfo');
const userProfilesRepository = require('../authentication/userProfilesRepository');
const btsClientSessionType = require('../communication/btsClientSessionType');

function imAuth({imAuthFactoryInst, sessionStorage}) {
    return imAuthFactoryInst.newInstance(sessionStorage);
}

function imAuthFactory({socket, routes = {}}) {
    return {
        newInstance: newInstance
    };

    function newInstance(sessionStorage) {
        const log = loggers.getLogger('imc-web-ui.imAuth');
        const event = events.newEvent();

        const AUTH_METHOD = {
            SHA1: {
                id: 1,
                name: 'SHA1',
                hash: (inputStr) => {
                    return new Rusha().digest(inputStr);
                }
            },
            PLAIN_TEXT: {
                id: 2,
                name: 'PLAIN_TEXT',
                hash: (inputStr) => {
                    return inputStr;
                }
            }
        };

        const OFFLINE_USER = {
            userName: 'admin',
            token: 'd033e22ae348aeb5660fc2140aec35850c4da997', //hint: repeat
            authMethod: AUTH_METHOD.SHA1
        };

        return {
            AUTH_METHOD: AUTH_METHOD,
            loginOnlineUser: loginOnlineUser,
            loginOfflineUser: loginOfflineUser,
            loginUserWhenSessionLocked: loginUserWhenSessionLocked,
            isUserSessionLocked: isUserSessionLocked,
            logout: logout,
            getHost: getHost,
            getPort: getPort,
            getLoginInfo: getLoginInfo,
            isSecure: isSecure,
            getUserName: getUserName,
            getProfileName: getProfileName,
            getToken: getToken,
            getAuthMethod: getAuthMethod,
            isOfflineUser: isOfflineUser,
            isOnlineUser: isOnlineUser,
            getDefaultRouteUrl: getDefaultRouteUrl,
            isLoggedIn: isLoggedIn,
            isRouteIdValid: isRouteIdValid,
            hasAccessTo: hasAccessTo,
            watch: watch,
            unwatch: unwatch
        };

        /**
         * @param {object} loginData - {
         *  host: {string},
         *  port: {int},
         *  isSecure: {boolean},
         *  userName: {string},
         *  token: {string},
         *  authMethod: {AUTH_METHOD enum}
         * }.
         * @returns {Promise}.
         * */
        function loginOnlineUser(loginData) {
            return new Promise((resolve, reject) => {
                const errorMessages = [];

                function handleError(error) {
                    log.error('Login error: ', JSON.stringify(error));
                    errorMessages.push(error.message);
                    removeUserFromStorage();
                    socket.disconnect();
                    socket.removeListener(listener);
                    reject(errorMessages);
                }

                const listener = {
                    connected: () => {
                        removeUserFromStorage();
                        const releaseInfo = applicationInfo.getReleaseInfo();
                        socket.login({
                            userName: loginData.userName,
                            token: loginData.token,
                            authMethodId: loginData.authMethod.id,
                            sessionType: btsClientSessionType.PLAIN,
                            clientVersion: releaseInfo.branch,
                            clientId: releaseInfo.hash
                        });
                    },
                    loggedIn: (loginInfo) => {
                        updateLoginInfo(loginInfo);
                        const profileName = loginInfo.profile;
                        const profile = userProfilesRepository.getByName(profileName);
                        if (!profile) {
                            handleError(`Invalid user profile: '${profileName}'`);
                            return;
                        }
                        setUserInStorage(getUser(profile));
                        log.info(`LoggedIn online user as ${profileName}`);
                        socket.disconnect();
                        socket.removeListener(listener);
                        resolve();
                        notify('loginSuccess', loginInfo);
                    },
                    connectionError: handleError,
                    compatibilityError: (error) => {
                        const message = `Invalid gateway version. Available: ${error.data.available}, required: ${error.data.required}`;
                        handleError({ message: message });
                    },
                    loginError: handleError
                };
                socket.addListener(listener);
                socket.connect(loginData.host, loginData.port, loginData.isSecure);
            });

            function getUser(profile) {
                return _.assign({ isOfflineUser: false, profile: profile }, loginData);
            }
        }

        function loginOfflineUser(loginInfo) {
            loginInfo = loginInfo || {};
            const profileName = loginInfo.profile ? loginInfo.profile : 'offline';
            const profile = userProfilesRepository.getByName(profileName);
            setUserInStorage(getOfflineUser(OFFLINE_USER, profile));
            updateLoginInfo(loginInfo);
            notify('loginSuccess', loginInfo);
            log.info(`LoggedIn offline user with profile: ${profileName}`);
            return Promise.resolve();
        }

        function loginUserWhenSessionLocked(authData) {
            if (!authData.token) return Promise.reject('Token not provided');

            return authData.token === getToken() ? Promise.resolve('Token validated') : Promise.reject('Invalid token');
        }

        function getOfflineUser(loginData, profile) {
            return _.assign({ isOfflineUser: true, profile: profile }, loginData);
        }

        function getLoginInfo() {
            return sessionStorage.loginInfo;
        }

        function updateLoginInfo(info) {
            sessionStorage.loginInfo = info;
        }

        function logout() {
            removeUserFromStorage();
            socket.disconnect();
        }

        function removeUserFromStorage() {
            delete sessionStorage.currentUser;
        }

        function getUserFromStorage() {
            return sessionStorage.currentUser;
        }

        function setUserInStorage(user) {
            sessionStorage.currentUser = user;
        }

        function getProfileName() {
            const user = getUserFromStorage();
            return user && user.profile.name;
        }

        function getProfile() {
            const user = getUserFromStorage();
            return user && user.profile;
        }

        function getAuthMethod() {
            if (!getUserFromStorage()) return null;

            const authMethodName = getUserFromStorage().authMethod.name;
            return AUTH_METHOD[authMethodName];
        }

        function isOnlineUser() {
            if (!isLoggedIn()) return false;

            return !isOfflineUser();
        }

        function isOfflineUser() {
            if (!isLoggedIn()) return false;
            return getUserFromStorage().isOfflineUser;
        }

        function getToken() {
            const user = getUserFromStorage();
            return user && user.token;
        }

        function getUserName() {
            const user = getUserFromStorage();
            return user && user.userName;
        }

        function getHost() {
            const user = getUserFromStorage();
            return user && user.host;
        }

        function getPort() {
            const user = getUserFromStorage();
            return user && global.parseInt(user.port, 10);
        }

        function isSecure() {
            const user = getUserFromStorage();
            return user && user.isSecure;
        }

        function getDefaultRouteUrl() {
            const found = _.find(routes, (route) => {
                return route.id === getUserFromStorage().profile.defaultRouteId;
            });
            return found && found.originalPath;
        }

        function isRouteIdValid(id) {
            if (!id) return false;
            if (!isLoggedIn()) return false;

            const profile = getProfile();
            if (profile.accessAllRouteIds) return true;

            if (profile.includedRouteIds && profile.excludedRouteIds)
                throw new Error('Use only includedRouteIds or excludedRouteIds');

            if (profile.includedRouteIds) return _.contains(profile.includedRouteIds, id);
            if (profile.excludedRouteIds) return !_.contains(profile.excludedRouteIds, id);

            return false;
        }

        function isLoggedIn() {
            return !!getUserFromStorage();
        }

        function hasAccessTo(functionality) {
            return isLoggedIn() && !_.contains(getProfile().excludedFunctionality, functionality);
        }

        function isUserSessionLocked() {
            return sessionStorage.lockScreenEnabled;
        }

        function watch(watcher) {
            event.watch(watcher);
        }

        function unwatch(watcher) {
            event.unwatch(watcher);
        }

        function notify(type, data) {
            event.notify({ type: type, data: data });
        }
    }
}
