const _ = require('lodash');

module.exports = {
    newBiMultiMap: newBiMultiMap
};

function newBiMultiMap() {
    const valuesByKey = new Map();
    const keysByValue = new Map();

    return {
        put: put,
        values: values,
        keys: keys,
        countKeys: countKeys,
        countValues: countValues,
        clear: clear,
        removeEntry: removeEntry,
        removeKey: removeKey,
        removeValue: removeValue,
        containsEntry: containsEntry,
        containsKey: containsKey,
        containsValue: containsValue
    };

    function put(key, value) {
        validateKey(key);
        validateValue(value);

        var values = valuesByKey.get(key);
        var keys = keysByValue.get(value);
        var isNewKey = _.isUndefined(values);
        var isNewValue = _.isUndefined(keys);

        values = values || [];
        keys = keys || [];

        if (isNewValue) {
            keys.push(key);
        } else if (!_.includes(keys, key)) {
            keys.push(key);
        }

        if (isNewKey) {
            values.push(value)
        } else if (!_.includes(values, value)) {
            values.push(value);
        }

        keysByValue.set(value, keys);
        valuesByKey.set(key, values);
    }

    function values(key) {
        if (arguments.length === 0) return allValues();

        validateKey(key);

        var values = valuesByKey.get(key);
        if (values) return values.slice();
    }

    function allValues() {
        return Array.from(keysByValue.keys());
    }

    function keys(value) {
        if (arguments.length === 0) return allKeys();

        validateValue(value);

        var keys = keysByValue.get(value);
        if (keys) return keys.slice();
    }

    function allKeys() {
        return Array.from(valuesByKey.keys());
    }

    function countKeys(value) {
        if (arguments.length === 0) return countAllKeys();

        validateValue(value);

        var keys = keysByValue.get(value) || [];
        return keys.length;
    }

    function countAllKeys() {
        return valuesByKey.size;
    }

    function countValues(key) {
        if (arguments.length === 0) return countAllValues();

        validateKey(key);

        var values = valuesByKey.get(key) || [];
        return values.length;
    }

    function countAllValues() {
        return keysByValue.size;
    }


    function clear() {
        valuesByKey.clear();
        keysByValue.clear();
    }

    function removeEntry(key, value) {
        validateKey(key);
        validateValue(value);

        var values = valuesByKey.get(key);
        var keys = keysByValue.get(value);
        var isUnknownKey = _.isUndefined(values);
        var isUnknownValue = _.isUndefined(keys);
        if (isUnknownKey || isUnknownValue) return;

        var valueIndex = _.indexOf(values, value);
        if (valueIndex !== -1) values.splice(valueIndex, 1);
        if (_.isEmpty(values)) {
            valuesByKey.delete(key);
        }

        var keyIndex = _.indexOf(keys, key);
        if (keyIndex !== -1) keys.splice(keyIndex, 1);
        if (_.isEmpty(keys)) {
            keysByValue.delete(value);
        }

        return value;
    }

    function removeKey(key) {
        validateKey(key);

        var values = valuesByKey.get(key);
        var isUnknownKey = _.isUndefined(values);
        if (isUnknownKey) return;

        _.forEach(values, function (value) {
            var keys = keysByValue.get(value);
            var index = _.indexOf(keys, key);
            if (index !== -1) keys.splice(index, 1);

            if (_.isEmpty(keys)) {
                keysByValue.delete(value);
            }
        });
        valuesByKey.delete(key);

        return values;
    }

    function removeValue(value) {
        validateValue(value);

        var keys = keysByValue.get(value);
        var isUnknownValue = _.isUndefined(keys);
        if (isUnknownValue) return;

        _.forEach(keys, function (key) {
            var values = valuesByKey.get(key);
            var index = _.indexOf(values, value);
            if (index !== -1) values.splice(index, 1);

            if (_.isEmpty(values)) {
                valuesByKey.delete(key);
            }
        });
        keysByValue.delete(value);

        return keys;
    }

    function containsEntry(key, value) {
        validateKey(key);
        validateValue(value);

        var values = valuesByKey.get(key);
        var keys = keysByValue.get(value);
        var isUnknownKey = _.isUndefined(values);
        var isUnknownValue = _.isUndefined(keys);

        if (isUnknownKey || isUnknownValue) return false;

        if (values.length < keys.length) {
            return _.includes(values, value);
        } else {
            return _.includes(keys, key);
        }
    }

    function containsKey(key) {
        validateKey(key);

        return !_.isUndefined(valuesByKey.get(key));
    }

    function containsValue(value) {
        validateValue(value);

        return !_.isUndefined(keysByValue.get(value));
    }

    function validateKey(key) {
        if (_.isUndefined(key)) throw 'Key is undefined';
    }

    function validateValue(value) {
        if (_.isUndefined(value)) throw 'Value is undefined';
    }
}
