'use strict';
const logger =  require('../logger.impl');

const log = logger.getLogger('admin.core.btsClient.heartBeatHandler');
module.exports = function newInstance(btsClient) {
    const DEFAULT_HEART_BEAT_INTERVAL_MS = 30 * 1000;
    var heartBeatResponseHandler = null;
    const scheduler = createScheduler();
    var onStopCallback = function () {
    };
    const btsClientListener = {
        loggedIn: function () {
            startHeartBeat();
        },
        disconnected: function () {
            stopHeartBeat();
        }
    };

    return {
        start: registerBtsClientListener,
        stop: unregisterBtsClientListener,
        isStarted: isStarted,
        //for testing
        startHeartBeat: startHeartBeat,
        stopHeartBeat: stopHeartBeat,
        onStop: onStop
    };

    function registerBtsClientListener() {
        btsClient.addListener(btsClientListener);
    }

    function unregisterBtsClientListener() {
        btsClient.removeListener(btsClientListener);
    }

    function isStarted() {
        return scheduler.isScheduled();
    }

    function startHeartBeat(intervalMs) {
        intervalMs = intervalMs || DEFAULT_HEART_BEAT_INTERVAL_MS;
        var kill = false;

        setHeartBeatResponseHandler(function () {
            kill = false;
        });
        scheduler.schedule(function () {
            if (!shouldStart()) {
                stopHeartBeat();
                return;
            }
            if (kill) {
                log.info('Disconnecting because of heartbeat not received on time (' + intervalMs + 'ms)');
                btsClient.disconnect({force: true, reason: 'heart beat failure'});
                stopHeartBeat();
            } else {
                kill = true;
                btsClient.sendHeartBeatRequest();
            }
        }, intervalMs);
    }

    function stopHeartBeat() {
        unsetHeartBeatResponseHandler();
        scheduler.stop();
        onStopCallback();
    }

    function onStop(callback) {
        onStopCallback = callback;
    }

    function shouldStart() {
        return btsClient.isConnected();
    }

    function setHeartBeatResponseHandler(handler) {
        heartBeatResponseHandler = btsClient.addHeartBeatResponseListener(handler);
    }

    function unsetHeartBeatResponseHandler() {
        if (heartBeatResponseHandler) {
            heartBeatResponseHandler.unwatch();
            heartBeatResponseHandler = null;
        }
    }
};

function createScheduler() {
    var shouldStop = false;
    var id;

    function schedule(scheduleFn, scheduleDelayMs) {
        shouldStop = false;
        id = global.setInterval(function () {
            if (!shouldStop) scheduleFn();
        }, scheduleDelayMs);
    }

    function isScheduled() {
        return !shouldStop;
    }

    function stop() {
        shouldStop = true;
        global.clearInterval(id);
        id = null;
    }

    return {
        schedule: schedule,
        stop: stop,
        isScheduled: isScheduled
    };
}

