const sic = require('../sic');
const events = require('../events');
const _ = require('lodash');
const loggers = require('../logger.impl');
const imiConstants = require('../imi/imiConstants');
const SESSION_TYPE = require('./btsClientSessionType.js');
const log = loggers.getLogger('admin.core.btsClient');

const DEFAULT_UPDATE_INTERVAL_MS = 5000;

const webSocketFactory = {
    create: function (url, protocols) {
        const socket = new WebSocket(url, protocols);
        socket.isConnected = function () {
            return socket.readyState === WebSocket.OPEN;
        };
        return socket;
    }
};
module.exports = function createBtsClient(customSocketFactory) {
    const socketFactory = customSocketFactory || webSocketFactory;
    let sicPlusSequenceId = 0;
    const listeners = [];
    const DEFAULT_IMI_MAJOR_VERSION = 9;
    const NORMAL_CLOSE = 1000;
    const FORCE_CLOSE = 4000;
    const GATEWAY_ADDRESS = {
        receiver: 0,
        sender: 0
    };
    socketFactory.create = socketFactory.create || webSocketFactory.create;

    const sicInstance = sic();
    const MAX_SIC_PAYLOAD_LENGTH = sicInstance.MAX_SIC_MESSAGE_SIZE - sicInstance.HEADER_SIZE;
    const MAX_SIC_PLUS_PAYLOAD_LENGTH = MAX_SIC_PAYLOAD_LENGTH - sicInstance.SIC_PLUS_HEADER_SIZE;
    let socket = null;
    let connectionInfo = null;
    let messageBuffer = null;
    let infoModelSicAddress = null;
    let receivedFirstUpdate = false;

    const messageHandlers = {};
    addMessageListener(0x3623, handleUpdateNotification);
    addMessageListener(0x3625, handleOperationExecuted);
    addMessageListener(0x6013, handleMeta);
    addMessageListener(0x6018, handleLoginSuccess);
    addMessageListener(0x6019, handleLoginError);
    addMessageListener(0x601E, handleGetGatewayProtocolVersion);
    addMessageListener(0x601F, handleGatewayInformationResponse);
    addMessageListener(imiConstants.IMGW.SESSIONS_LIST_RESP, handleSessionsList);

    const socketListener = {
        createdMessage: function (msg) {
            log.debug(`received message: 0x ${msg.getId().toString(16)} [${msg.getBuffer().byteLength} bytes]`);
        },
        completedMessage: function (msg) {
            const handler = messageHandlers[msg.getId()];
            if (handler) {
                log.debug(`handling message: ${msg.toString()}`);
                handler.notify(msg);
                return;
            }
            log.debug(`no handler found for message: ${msg.toString()}`);
        }
    };

    return {
        DEFAULT_IMI_VERSION: DEFAULT_IMI_MAJOR_VERSION,
        connect: connect,
        login: login,
        isConnected: isConnected,
        getConnectionInfo: getConnectionInfo,
        disconnect: disconnect,
        addListener: addListener,
        removeListener: removeListener,
        send: send,
        sendChangeRequest: sendChangeRequest,
        sendHeartBeatRequest: sendHeartBeatRequest,
        sendSessionsListRequest,
        addHeartBeatResponseListener: addHeartBeatResponseListener,
        addMessageListener: addMessageListener,
        getSicInstance: getSicInstance,
        //for testing
        createListener: createListener,
        getSocketListener
    };

    function getSocketListener() {
        return socketListener;
    }

    function getSicInstance() {
        return sicInstance;
    }

    function getConnectionInfo() {
        return connectionInfo;
    }

    function sendSicMessage(msg, optional) {
        const payloadBuffer = msg.getBuffer();
        const payloadLength = payloadBuffer.byteLength;
        const header = sicInstance.createHeader();
        header.setId(msg.getId());
        header.setPayloadLength(payloadLength);

        if (optional && _.isFinite(optional.receiver))
            header.setReceiver(optional.receiver);

        if (optional && _.isFinite(optional.sender))
            header.setSender(optional.sender);

        const sicMessage = sicInstance.createMessage(header);
        sicMessage.setPayload(payloadBuffer);
        if (isConnected()) {
            log.debug(`Sending request to BTS: 0x${msg.getId().toString(16)}`);
            socket.send(sicMessage.getBuffer());
        } else {
            log.debug(`Connection is closed! Cannot send request to BTS: 0x${msg.getId().toString(16)}`);
        }
    }

    function send(msg, optional) {
        const payloadBuffer = msg.getBuffer();
        const payloadLength = payloadBuffer.byteLength;
        if (payloadLength <= MAX_SIC_PAYLOAD_LENGTH)
            sendSicMessage(msg, optional);
        else
            sendSicPlusMessage(msg, optional);

        notifyAll('sent', msg);
    }

    function sendSicPlusMessage(msg, optional) {
        const buffer = msg.getBuffer();
        const msgSize = buffer.byteLength;
        const msgId = msg.getId();
        let offset = 0;
        while (offset < msgSize) {
            let bytesToSend = msgSize - offset;
            if (bytesToSend >= MAX_SIC_PLUS_PAYLOAD_LENGTH)
                bytesToSend = MAX_SIC_PLUS_PAYLOAD_LENGTH;

            const message = createSicPlusMessage(new Uint8Array(buffer, offset, bytesToSend));
            sendSicMessage(message, optional);
            offset += bytesToSend;
        }
        sicPlusSequenceId++;

        function createSicPlusMessage(payload) {
            const message = sicInstance.createSicPlusPayload();
            message.setSeqId(sicPlusSequenceId);
            message.setMsgId(msgId);
            message.setMsgSize(msgSize);
            message.setDataOffset(offset);
            message.setData(payload);
            return message;
        }
    }

    function sendChangeRequest(msg) {
        send(msg, {receiver: infoModelSicAddress});
    }

    function sendGetGatewayProtocolVersion() {
        const msg = sicInstance.imgwGetProtocolVersionReq();
        send(msg, GATEWAY_ADDRESS);
    }

    function sendHeartBeatRequest() {
        const heartBeatRequest = sicInstance.aaSysComEcho();
        send(heartBeatRequest, {
            receiver: sicInstance.AASYSCOM_ECHO_CPID
        });
    }

    function sendSessionsListRequest() {
        const msg = sicInstance.imgwSessionsListReq();
        send(msg, GATEWAY_ADDRESS);
    }

    function handleGetGatewayProtocolVersion(imgwGetProtocolVersionResp) {
        const version = imgwGetProtocolVersionResp.getVersion();
        if (sicInstance.GATEWAY_VERSION === version) {
            log.debug(`Using gateway protocol version: ${version}`);
            notifyAll('connected', `Connected to ${connectionInfo.host}:${connectionInfo.port}`);
        } else {
            const error = {
                message: `Invalid gateway version. Available: ${version}, required: ${sicInstance.GATEWAY_VERSION}`,
                data: {
                    available: version,
                    required: sicInstance.GATEWAY_VERSION
                }
            };
            notifyAll('compatibilityError', error);
        }
    }

    function handleLoginSuccess(imgwLoginSuccessResp) {
        const loginSuccessEvent = {
            profile: imgwLoginSuccessResp.getProfile(),
            userName: imgwLoginSuccessResp.getUserName(),
            failedLoginNo: imgwLoginSuccessResp.getFailedLoginNo(),
            lastSuccess: imgwLoginSuccessResp.getLastSuccess()
        };
        notifyAll('loggedIn', loginSuccessEvent);
    }

    function handleLoginError(imgwLoginErrorResp) {
        notifyAll('loginError', {
            errorCode: imgwLoginErrorResp.getErrorCode(),
            message: imgwLoginErrorResp.getDescription()
        });
    }

    function handleUpdateNotification(updateNotification) {
        if (!receivedFirstUpdate) {
            receivedFirstUpdate = true;
            notifyAll('signInCompleted');
        }
        const removedIds = updateNotification.getRemovedObjectIds();
        const managedObjects = updateNotification.getManagedObjects();
        notifyAll('data', {removedIds, managedObjects});
    }

    function handleOperationExecuted(operationExecuted) {
        notifyAll('operation', {
            requestId: operationExecuted.getRequestId(),
            executionStatus: operationExecuted.getExecutionStatus()
        });
    }

    function handleMeta(meta) {
        notifyAll('meta', meta.getMeta());
    }

    function isConnected() {
        return socket && socket.isConnected();
    }

    function handleGatewayInformationResponse(gatewayInformation) {
        const gatewayInformationEvent = {
            endianness: gatewayInformation.getEndianness(),
            embeddedAdminVersion: gatewayInformation.getEmbeddedAdminVersion(),
            embeddedAdminId: gatewayInformation.getEmbeddedAdminId(),
            imiVersionMajor: gatewayInformation.getImiVersionMajor(),
            imiVersionMinor: gatewayInformation.getImiVersionMinor(),
            infoModelSicAddress: gatewayInformation.getInfoModelSicAddress(),
            imProtocolRelease: gatewayInformation.getImProtocolRelease(),
            imProtocolVersionMajor: gatewayInformation.getImProtocolVersionMajor(),
            imProtocolVersionMinor: gatewayInformation.getImProtocolVersionMinor()
        };
        infoModelSicAddress = gatewayInformationEvent.infoModelSicAddress;
        sicInstance.setImiVersion(gatewayInformationEvent.imiVersionMajor);
        notifyAll('gatewayInformation', gatewayInformationEvent);
    }

    function handleSessionsList(sessionsList) {
        notifyAll('sessionsList', sessionsList);
    }

    function notifyAll(eventName, eventData) {
        for (let i = 0; i < listeners.length; i++)
            listeners[i][eventName](eventData);
    }

    function sendLoginRequest(loginData) {
        const msg = sicInstance.imgwLoginReq();
        msg.setLogin(loginData.userName);
        msg.setToken(loginData.token);
        msg.setAuthMethod(loginData.authMethodId);
        msg.setSessionType(loginData.sessionType);
        msg.setClientVersion(loginData.clientVersion);
        msg.setClientId(loginData.clientId);
        if (loginData.sessionType === SESSION_TYPE.INFO_MODEL)
            msg.setUpdateInterval(connectionInfo.updateInterval);
        send(msg, GATEWAY_ADDRESS);
    }

    function connect(host, port, isSecure, updateInterval) {
        if (isConnected()) return;

        if (_.isUndefined(updateInterval))
            updateInterval = DEFAULT_UPDATE_INTERVAL_MS;

        try {
            socket = createSocket();
        } catch (e) {
            log.error('Connection error: ', e);
            notifyAll('connectionError', {errorCode: e.code, message: e.message});
            return;
        }
        connectionInfo = {
            host: host,
            port: port,
            isSecure: isSecure,
            updateInterval: updateInterval
        };
        messageBuffer = sicInstance.createSicMessageBuffer(socketListener);
        socket.onopen = onOpen;
        socket.onmessage = onMessage;
        socket.onclose = onClose;
        socket.onerror = onError;

        function createSocket() {
            const prefix = isSecure ? 'wss://' : 'ws://';
            const newSocket = socketFactory.create(`${prefix}${host}:${port}`, ['binary', 'base64']);
            newSocket.binaryType = 'arraybuffer';
            return newSocket;
        }

        function onOpen() {
            if (!socket.isConnected()) return;

            sendGetGatewayProtocolVersion();
        }

        function onMessage(event) {
            if (event.data && isArrayBuffer(event.data))
                messageBuffer.push(event.data);
        }

        function isArrayBuffer(object) {
            // Second part of the condition is for code run outside of a browser (i.e. nashorn, node)
            // http://stackoverflow.com/questions/15251879/how-to-check-if-a-variable-is-a-typed-array-in-javascript
            return object instanceof ArrayBuffer || Object.prototype.toString.call(object) === '[object ArrayBuffer]';
        }

        function onClose(event) {
            if (event.code !== NORMAL_CLOSE)
                notifyAll('connectionError', {
                    errorCode: event.code,
                    message: `Disconnected with close code: ${event.code}`
                });

            notifyAll('disconnected', {
                source: getCloseSource(event),
                closeEvent: event
            });
            socket.onopen = empty;
            socket.onmessage = empty;
            socket.onclose = empty;
            socket.onerror = empty;
        }

        function onError() {
            //onerror recives no error code!!! A fake error code need to be passed to make code generic
            notifyAll('connectionError', {
                errorCode: -1,
                message: 'Cannot connect to BTS...'
            });
        }
    }

    function login(loginData) {
        if (!isConnected()) throw new Error('Not connected!');
        receivedFirstUpdate = false;
        infoModelSicAddress = null;
        sendLoginRequest(loginData);
    }

    function getCloseSource(closeEvent) {
        if (closeEvent && closeEvent.reason && closeEvent.reason === 'BY_USER') return 'USER';
        return 'BTS';
    }

    function disconnect(options) {
        if (options && options.force)
            forceDisconnect(options.reason);
        else
            plainDisconnect();

        function forceDisconnect(reason) {
            if (!socket) return;

            //send close message to server
            closeSocket(FORCE_CLOSE, reason);
            //and notify about close immediately
            socket.onclose({code: FORCE_CLOSE, reason: reason});
        }

        function plainDisconnect() {
            closeSocket(NORMAL_CLOSE, 'BY_USER');
        }

        function closeSocket(closeCode, closeReason) {
            if (!socket) return;

            try {
                socket.close(closeCode, closeReason);
                log.debug('Socket closed successfully.');
            } catch (error) {
                log.error('close error', error);
            }
        }
    }

    function empty() {
    }

    function createListener(listener) {
        listener = listener || {};
        listener.connected = listener.connected || empty;
        listener.disconnected = listener.disconnected || empty;
        listener.data = listener.data || empty;
        listener.connectionError = listener.connectionError || empty;
        listener.compatibilityError = listener.compatibilityError || empty;
        listener.meta = listener.meta || empty;
        listener.operation = listener.operation || empty;
        listener.loggedIn = listener.loggedIn || empty;
        listener.loginError = listener.loginError || empty;
        listener.signInCompleted = listener.signInCompleted || empty;
        listener.gatewayInformation = listener.gatewayInformation || empty;
        listener.sent = listener.sent || empty;
        listener.sessionsList = listener.sessionsList || empty;
        return listener;
    }

    function addListener(listener) {
        listeners.push(createListener(listener));
    }

    function removeListener(listener) {
        const index = listeners.indexOf(listener);
        if (index > -1) listeners.splice(index, 1);
    }

    function addMessageListener(msgId, watcher) {
        messageHandlers[msgId] = messageHandlers[msgId] || events.newEvent();
        return messageHandlers[msgId].watch(watcher);
    }

    function addHeartBeatResponseListener(watcher) {
        return addMessageListener(sicInstance.aaSysComEcho().getId(), watcher);
    }
};

module.exports.getDefaultUpdateIntervalMs = function () {
    return DEFAULT_UPDATE_INTERVAL_MS;
};
