const collections = require('../collections'),
    SEPARATOR = require('../utils/distNameUtils').SEPARATOR;

module.exports = {newInstance};

function newInstance() {
    const classNamesById = collections.newBiMultiMap();
    const distNamesByUuid = collections.newBiMultiMap();
    const objectInfosByUuid = new Map();

    return {
        addClassName,
        getClassName,
        getClassNameId,
        registerObjectInfos,
        resolveDistName,
        removeUuid,
        resolveObjectInfo,
        clear
    };

    function addClassName(id, className = '') {
        if (classNamesById.containsKey(id))
            throw new Error(`Duplicate className id "${id}"`);
        className = className.toUpperCase();
        if (classNamesById.containsValue(className))
            throw new Error(`Duplicate className "${className}"`);

        classNamesById.put(id, className);
    }

    function getClassName(id) {
        const values = classNamesById.values(id);
        return values && values[0];
    }

    function getClassNameId(className) {
        const keys = classNamesById.keys(className);
        return keys && keys[0];
    }

    function getUuidByDistName(distName) {
        const values = distNamesByUuid.keys(distName);
        return values && values[0];
    }

    function registerObjectInfos(objectInfos) {
        objectInfos.forEach(registerObjectInfo);
        objectInfos.forEach((objectInfo) => {
            resolveDistName(objectInfo.uuid);
        });
    }

    function registerObjectInfo(objectInfo) {
        if (objectInfosByUuid.has(objectInfo.uuid)) return;

        if (typeof objectInfo.instanceId === 'undefined')
            throw new Error(`Missing instanceId in objectInfo "${objectInfo.uuid}"`);

        objectInfosByUuid.set(objectInfo.uuid, objectInfo);
    }

    function resolveDistName(uuid) {
        const distName = resolveDistNameByUuid(uuid);
        if (!distNamesByUuid.containsKey(uuid) && distName)
            distNamesByUuid.put(uuid, distName);

        return distName;

        function resolveDistNameByUuid(objectUuid) {
            if (distNamesByUuid.containsKey(objectUuid)) return distNamesByUuid.values(objectUuid)[0];
            const objectInfo = objectInfosByUuid.get(objectUuid);

            if (!objectInfo)
                return null;

            if (!objectInfo.parentUuid) return getName(objectInfo);
            return resolveDistNameByUuid(objectInfosByUuid.get(objectInfo.parentUuid).uuid) + getName(objectInfo);
        }
    }

    function getName(objectInfo) {
        return `${SEPARATOR}${getClassName(objectInfo.objectType)}-${objectInfo.instanceId}`;
    }

    function removeUuid(uuid) {
        objectInfosByUuid.delete(uuid);
        distNamesByUuid.removeKey(uuid);
    }

    function resolveObjectInfo(distName) {
        checkDistName();
        const uuid = getUuidByDistName(distName);
        return uuid && objectInfosByUuid.get(uuid);

        function checkDistName() {
            if (!distName
                || distName.charAt(0) !== SEPARATOR
                || distName.charAt(distName.length - 1) === SEPARATOR
                || distName !== distName.toUpperCase())
                throw new Error(`Invalid distName format, distName "${distName}"`);
        }
    }

    function clear() {
        classNamesById.clear();
        distNamesByUuid.clear();
        objectInfosByUuid.clear();
    }
}
