const sequentialExecutorFactory = require('../sequentialExecutorFactory'),
    createBtsClient = require('./btsClient'),
    createMetaService = require('../metaService'),
    logger = require('../logger.impl'),
    objectInfoNormalizer = require('./objectInfoNormalizer');

const log = logger.getLogger('admin.core.translatingBtsClient');

module.exports = {
    newInstance
};

function newInstance({
    btsClient = createBtsClient(),
    metaService = createMetaService({}),
    objectIdMapper = require('./objectIdMapper').newInstance()
    } = {}) {

    const listeners = [];
    const executor = sequentialExecutorFactory.newInstance();
    const normalizer = objectInfoNormalizer.newInstance({btsClient});
    const onDataEvent = createNotifyEvent('data');
    const onMetaEvent = createNotifyEvent('meta');
    let mapper = objectIdMapper;
    const translatingBtsClient = Object.assign({}, btsClient, {
        addListener,
        removeListener,
        resolveObjectInfo,
        resolveManagedObjects,
        getMapper,
        sendOperation
    });

    btsClient.addListener({
        connected: createNotifyEvent('connected'),
        disconnected: createNotifyEvent('disconnected'),
        data: executor.promisifyFunction(this, handleOnData),
        connectionError: createNotifyEvent('connectionError'),
        compatibilityError: createNotifyEvent('compatibilityError'),
        meta: executor.promisifyFunction(this, handleOnMeta),
        loggedIn: createNotifyEvent('loggedIn'),
        loginError: createNotifyEvent('loginError'),
        signInCompleted: createNotifyEvent('signInCompleted'),
        gatewayInformation: createNotifyEvent('gatewayInformation'),
        sent: createNotifyEvent('sent'),
        sessionsList: createNotifyEvent('sessionsList'),
        operation: handleOperation
    });

    return translatingBtsClient;

    function sendOperation(msg, listener) {
        translatingBtsClient.addListener(listener);
        try {
            translatingBtsClient.sendChangeRequest(msg);
        } catch (exception) {
            translatingBtsClient.removeListener(listener);
            throw exception;
        }
    }

    function handleOperation(eventData) {
        const requestId = eventData.requestId;
        const listener = findOperationListener(requestId);
        if (!listener) {
            log.warn(`No listener for request id=${requestId}`);
            return;
        }
        listener.operation.callbackFunction(eventData.executionStatus, requestId);
        removeListener(listener);
    }

    function findOperationListener(requestId) {
        for (let i = 0; i < listeners.length; i++) {
            const listener = listeners[i];
            const operation = listener.operation;
            if (operation && operation.requestId === requestId)
                return listener;

        }
        return null;
    }

    function handleOnMeta(metaBuffer) {
        return metaService.setMeta(metaBuffer)
            .then((metaUid) => {
                mapper.clear();
                registerTypes(metaUid);
                onMetaEvent(metaBuffer);
            });
    }

    function registerTypes(metaUid) {
        const classNames = metaService.getMetaRepository().getClassNamesForUid(metaUid);
        classNames.forEach((className) => {
            const id = metaService.getTypeInfo(metaUid, className).getId();
            mapper.addClassName(id, className);
        });
    }

    function handleOnData(eventData) {
        const processedManageObjects = resolveManagedObjects(eventData.managedObjects);
        const removedDistNames = resolveIds(eventData.removedIds);

        onDataEvent({
            managedObjects: processedManageObjects,
            removedDistNames
        });
    }

    function resolveIds(removedIds) {
        const uuids = removedIds.map(removedId => {
            return normalizer.uuidBytesToHex(removedId);
        });
        const distNames = uuids.map(uuid => {
            const distName = mapper.resolveDistName(uuid);
            if (!distName)
                log.warn(`Cannot resolve dist name of removed object id "${uuid}". Object will be ignored.`);

            return distName;
        }).filter(distName => {
            return distName;
        });
        uuids.forEach(mapper.removeUuid);
        return distNames;
    }

    function resolveManagedObjects(managedObjects) {
        const normalizedManagedObjects = managedObjects.map(managedObject => {
            return {
                objectInfo: normalizer.buildObjectInfo(managedObject.object_info),
                payload: convertPayload(managedObject.payload)
            };
        });
        mapper.registerObjectInfos(normalizedManagedObjects.map(mangeObject => {
            return mangeObject.objectInfo;
        }));
        return normalizedManagedObjects.map(managedObject => {
            return {
                distName: mapper.resolveDistName(managedObject.objectInfo.uuid),
                payload: managedObject.payload
            };
        });
        function convertPayload(payload) {
            if (!payload) return new Uint8Array(0);
            return new Uint8Array(payload.toArrayBuffer());
        }
    }

    function addListener(listener = {}) {
        listeners.push(listener);
    }

    function removeListener(listener) {
        const index = listeners.indexOf(listener);
        if (index <= -1) return;
        listeners.splice(index, 1);
    }

    function createNotifyEvent(eventName) {
        return (eventData) => {
            for (let i = 0; i < listeners.length; i++) {
                const eventListener = listeners[i][eventName];
                if (eventListener)
                    eventListener(eventData);
            }
        };
    }

    function resolveObjectInfo(distName) {
        const objectInfo = mapper.resolveObjectInfo(distName);
        return normalizer.buildImiObjectInfo(objectInfo);
    }

    function getMapper() {
        return mapper;
    }
}
