module.exports = signInHandlerService;

const _ = require('lodash');
const loggers = require('../logger.impl');
const btsClientSessionType = require('../communication/btsClientSessionType');

function signInHandlerService({socket}) {
    return signInHandler().newInstance(socket);
}

function signInHandler() {
    const LOG = loggers.getLogger('signin-handler.service');
    const DEFAULT_AUTO_RECONNECT_INTERVAL_MS = 1000;
    return {
        newInstance: newInstance
    };

    function newInstance(socket) {
        const STATE = {
            SIGNED_IN: 'SIGNED_IN',
            SIGNED_OUT: 'SIGNED_OUT',
            IN_PROGRESS: 'IN_PROGRESS'
        };

        let socketListener = null;
        const listeners = [];

        let signInData = null;
        let reconnectIntervalMs = 0;
        let autoReconnect = false;
        let autoReconnectId = null;
        let currentState = STATE.SIGNED_OUT;

        addListener({
            connected: () => {
                LOG.info('connected');
                changeStateTo(STATE.IN_PROGRESS);
            },
            autoReconnect: () => {
                LOG.info('autoReconnect');
                changeStateTo(STATE.IN_PROGRESS);
            },
            signIn: () => {
                LOG.info('signIn');
                changeStateTo(STATE.SIGNED_IN);
            },
            signOut: () => {
                LOG.info('signOut');
                changeStateTo(STATE.SIGNED_OUT);
            },
            disconnected: () => {
                LOG.info('disconnected');
            },
            error: (error) => {
                LOG.error(error);
            },
            info: (info) => {
                LOG.info(info);
            }
        });
        return {
            STATE: STATE,
            addListener: addListener,
            removeListener: removeListener,
            signIn: signIn,
            signOut: signOut,
            enableAutoReconnect: enableAutoReconnect,
            disableAutoReconnect: disableAutoReconnect,
            isSignInInProgress: isSignInInProgress,
            isSignedIn: isSignedIn
        };

        function addListener(listener) {
            listener.connected = listener.connected || _.noop;
            listener.disconnected = listener.disconnected || _.noop;
            listener.signIn = listener.signIn || _.noop;
            listener.signOut = listener.signOut || _.noop;
            listener.autoReconnect = listener.autoReconnect || _.noop;
            listener.error = listener.error || _.noop;
            listener.info = listener.info || _.noop;
            listener.stateChanged = listener.stateChanged || _.noop;
            listeners.push(listener);
        }

        function removeListener(listener) {
            _.remove(listeners, (item) => {
                return item === listener;
            });
        }

        function notifyAll(event, ...args) {
            _.forEach(listeners, (listener) => {
                const fun = listener[event];
                fun.apply(null, args);
            });
        }

        function changeStateTo(newState) {
            if (currentState === newState) return;
            const oldState = currentState;

            currentState = newState;
            notifyAll('stateChanged', newState, oldState);
        }

        function signIn(newSignInData) {
            cancelAutoReconnect();
            changeStateTo(STATE.IN_PROGRESS);
            signInData = newSignInData;
            if (socketListener) socket.removeListener(socketListener);

            socketListener = createSocketListener();
            socket.addListener(socketListener);
            if (socket.isConnected()) socket.disconnect();

            socket.connect(signInData.host, signInData.port, signInData.isSSL, signInData.updateInterval);
        }

        function signOut() {
            if (!signInData) return;

            cancelAutoReconnect();

            if (socket.isConnected()) {
                socketListener.disconnected = cleanUp;
                socketListener.connectionError = cleanUp;
                socketListener.loginError = cleanUp;
                socketListener.compatibilityError = cleanUp;
                socket.disconnect();
            } else {
                cleanUp();
            }

            function cleanUp() {
                socket.removeListener(socketListener);
                notifyAll('disconnected');
                notifyAll('signOut');
                changeStateTo(STATE.SIGNED_OUT);
            }
        }

        function enableAutoReconnect(intervalMs) {
            autoReconnect = true;
            reconnectIntervalMs = intervalMs || DEFAULT_AUTO_RECONNECT_INTERVAL_MS;
        }

        function disableAutoReconnect() {
            autoReconnect = false;
        }

        function cancelAutoReconnect() {
            if (!autoReconnectId) return;

            global.clearTimeout(autoReconnectId);
            autoReconnectId = null;
        }

        function isSignInInProgress() {
            return currentState === STATE.IN_PROGRESS;
        }

        function isSignedIn() {
            return currentState === STATE.SIGNED_IN;
        }

        function createSocketListener() {
            const SOCKET_EVENTS_POSPONE_MS = 50;
            return {
                connected: _.debounce(() => {
                    cancelAutoReconnect();
                    notifyAll('connected');
                    loginUser();
                }, SOCKET_EVENTS_POSPONE_MS),
                disconnected: _.debounce((event) => {
                    notifyAll('disconnected');
                    cancelAutoReconnect();
                    if (shouldAutoReconnect(event)) scheduleAutoReconnect();
                    else changeStateTo(STATE.SIGNED_OUT);
                }, SOCKET_EVENTS_POSPONE_MS),
                loggedIn: () => {
                },
                signInCompleted: (info) => {
                    notifyAll('signIn', info);
                },
                connectionError: _.debounce((error) => {
                    notifyAll('error', error.message);
                    cancelAutoReconnect();
                    if (autoReconnect) scheduleAutoReconnect();
                    else changeStateTo(STATE.SIGNED_OUT);
                }, SOCKET_EVENTS_POSPONE_MS),
                compatibilityError: _.debounce((error) => {
                    notifyAll('error', error.message);
                    cancelAutoReconnect();
                    changeStateTo(STATE.SIGNED_OUT);
                }, SOCKET_EVENTS_POSPONE_MS),
                loginError: _.debounce((error) => {
                    notifyAll('error', error.message);
                    cancelAutoReconnect();
                    const isSMADown = error.errorCode === 1027;
                    if (autoReconnect && isSMADown) {
                        scheduleAutoReconnect();
                    } else {
                        changeStateTo(STATE.SIGNED_OUT);
                        socket.disconnect();
                    }
                }, SOCKET_EVENTS_POSPONE_MS)
            };

            function loginUser() {
                socket.login({
                    userName: signInData.userName,
                    token: signInData.token,
                    authMethodId: signInData.authMethodId,
                    sessionType: btsClientSessionType.INFO_MODEL,
                    clientVersion: signInData.adminVersion,
                    clientId: signInData.clientId
                });
            }

            function shouldAutoReconnect(event) {
                return event.source === 'BTS' && autoReconnect;
            }

            function scheduleAutoReconnect() {
                cancelAutoReconnect();
                changeStateTo(STATE.IN_PROGRESS);
                notifyAll('autoReconnect');
                autoReconnectId = global.setTimeout(makeAutoReconnect, reconnectIntervalMs);

                function makeAutoReconnect() {
                    cancelAutoReconnect();
                    changeStateTo(STATE.IN_PROGRESS);
                    notifyAll('info', 'starting auto reconnect...');
                    if (socket.isConnected()) loginUser();
                    else socket.connect(signInData.host, signInData.port, signInData.isSSL, signInData.updateInterval);
                }
            }
        }
    }
}
