module.exports = dspsService;

const USAGE_LEVEL = {
    FREE: 0.3,
    NORMAL: 0.8,
    CRITICAL: 1
};

const updateServiceProvider = require('../updateService'),
    siteModelFactoryProvider = require('../site/siteModelFactory'),
    distNameService = require('../utils/distNameUtils'),
    events = require('../events'),
    dspsConstants = require('./dspsConstants');

function dspsService({
    updateService = updateServiceProvider(),
    siteModelFactory = siteModelFactoryProvider
    }) {

    const event = events.newEvent();
    var cells = null;
    var imTree = null;

    function Module(modules, nodeAName) {
        this.modules = modules;
        this.nodeAName = nodeAName;
        this.nodeAMap = this.createModAMap();
    }

    Module.prototype = {
        constructor: Module,
        findModule: findModule,
        createModAMap: createModAMap,
        countBBU: countBBU
    };

    event.onFirstAdded(watchUpdateService);
    event.onLastRemoved(function () {
        const imTreeListener = getImTreeListener();
        updateService.removeImTreeListener(imTreeListener);
        cells = null;
        imTree = null;
    });

    return {
        imTree: function () {
            return imTree;
        },
        cells: function () {
            return cells;
        },
        populateDSPs: populateDSPs,
        watch: function (watcher) {
            event.watch(watcher);
            if (cells)
                event.notify({
                    cells: cells,
                    imTree: imTree
                });

        },
        unwatch: event.unwatch
    };

    function getImTreeListener() {
        const imTreeListener = {
            completeImTree: function (completedImTree) {
                imTree = completedImTree;
                cells = populateDSPs(completedImTree);
                event.notify({
                    cells: cells,
                    imTree: imTree
                });
            }
        };
        return imTreeListener;
    }

    function watchUpdateService() {
        const imTreeListener = getImTreeListener();
        updateService.addImTreeListener(imTreeListener);

    }

    function populateDSPs(imTree) {
        const hardware = siteModelFactory.create(imTree);

        const bbmodModules = [];
        const smodModules = [];

        hardware.cabinets.forEach((cabinet) => {
            addToArray(cabinet.bbmods, bbmodModules);
            addToArray(cabinet.smods, smodModules);
        });

        const smods = new Module(smodModules, dspsConstants.CLASS_NAME.SMOD_A);
        const bbmods = new Module(bbmodModules, dspsConstants.CLASS_NAME.BBMOD_A);

        const dspsCells = hardware.cells.map((cell) => {
            const cellLabel = createCellLabel(cell);
            const channels = populateChannels(cell, smods, bbmods, hardware);
            return {
                distName: cell.distName,
                physicalId: cell.phyCellId,
                label: cellLabel,
                channels: channels
            };
        });

        return dspsCells.sort((cell1, cell2) => {
            return cell1.physicalId - cell2.physicalId;
        });

        function addToArray(from, to) {
            from = from || [];
            from.forEach((element) => {
                to.push(element);
            });
        }
    }

    function createCellLabel(cell) {
        return cell.technology ? cell.technology + ' ' + cell.dlChBw + ' ' + dspsConstants.LABELS.CELL.PCI + ' ' + cell.phyCellId : '';
    }

    function populateChannels(cell, smods, bbmods, hardware) {
        const allChannels = cell.channels;
        const channels = [];
        allChannels.forEach(function (channel) {
            const devices = getDevicesDNFromChannelPath(channel);
            let bbu;
            let dsps;
            let smodGroup;
            if (devices.smodGroupDN) {
                bbu = devices.bbuDN ? distNameService.getName(devices.bbuDN) : '';
                dsps = getDspsNames(devices.dspsDN);
                smodGroup = fillModuleData(devices.smodGroupDN, smods, bbmods, devices.connectorsDN, dsps.length);
            }

            const rmod = fillRmodData(hardware, devices.rmodDN, devices.connectorsDN);
            const label = createChannelLabel(channel, hardware, devices.connectorsDN, devices.rmodDN);

            channels.push({
                distName: channel.distName,
                bbu: bbu,
                dsps: dsps || [],
                smodGroup: smodGroup,
                rmod: rmod || {},
                label: label,
                axcAddress: channel.axcAddress
            });
        });
        return groupRepeatedProperties(channels);
    }

    function groupRepeatedProperties(channels) {
        var sortedChannels = channels.sort(function (channel1, channel2) {
            return channel1.axcAddress - channel2.axcAddress;
        });
        groupChannelPropertiesBy(sortedChannels, 'rmod', function (rmod1, rmod2) {
            return rmod1.label === rmod2.label;
        });
        groupChannelPropertiesBy(sortedChannels, 'dsps', function (dsps1, dsps2) {
            return dsps1.length === dsps2.length
                && dsps1.every(function (dsp, index) {
                    return dsps2[index].distName === dsp.distName;
                });

        });
        return sortedChannels;
    }

    function groupChannelPropertiesBy(channels, property, compare) {
        var uniqueData = channels[0][property];
        uniqueData.duplicateCount = 1;
        const channelsTail = channels.slice(1);

        channelsTail.forEach(function (channel) {
            if (compare(channel[property], uniqueData)) {
                uniqueData.duplicateCount++;
                channel[property].duplicateCount = 0;
            } else {
                uniqueData = channel[property];
                uniqueData.duplicateCount = 1;
            }
        });
    }

    function getDevicesDNFromChannelPath(channel) {
        const dspsPaths = channel.devicesPath.filter(function (devicePath) {
            return distNameService.getClassName(devicePath) === dspsConstants.CLASS_NAME.DSP;
        });
        if (dspsPaths.length > 0) {
            var firstDspPath = dspsPaths[0];
            var bbuDN = getBbuFromDspDistName(firstDspPath);
            var smodGroupDN = distNameService.getParent(bbuDN);
        }

        return {
            bbuDN: bbuDN,
            smodGroupDN: smodGroupDN,
            dspsDN: dspsPaths,
            connectorsDN: findConnectorsDN(channel),
            rmodDN: getRmodDn(channel)
        };
    }

    function getBbuFromDspDistName(dspDistName) {
        do {
            var parentDistName = distNameService.getParent(dspDistName);
            var className = distNameService.getClassName(parentDistName);
            if (dspDistName === parentDistName) return parentDistName;

            dspDistName = parentDistName;
        } while (className !== 'BBU_L');
        return dspDistName;
    }

    function getRmodDn(channel) {
        const rmodPath = channel.devicesPath.find(function (devicePath) {
            const dspParent = distNameService.getParent(devicePath);
            return distNameService.getClassName(dspParent) === dspsConstants.CLASS_NAME.RMOD;
        });
        return rmodPath && distNameService.getParent(rmodPath);
    }

    function findConnectorsDN(channel) {
        const connectorsDN = channel.devicesPath.filter(function (devicePath) {
            return distNameService.getClassName(devicePath) === dspsConstants.CLASS_NAME.CONNECTOR;
        });
        return getCorrectConnectorDistName(connectorsDN);
    }

    function getCorrectConnectorDistName(connectors) {
        return connectors.map(function (connector) {
            const parent = distNameService.getParent(connector);
            return (distNameService.getClassName(parent) === dspsConstants.CLASS_NAME.QFSP) ? parent : connector;
        });
    }


    function findModule(distName) {
        var smodGroup = findDevice(this.modules, distName);
        var nodeAData = findNodeAData(smodGroup, this.nodeAMap);
        return {
            smodGroup: smodGroup,
            nodeAData: nodeAData
        };
    }

    function createModAMap() {
        const smodGroupMap = {};
        const btsDataNodes = imTree.findByClassName(dspsConstants.CLASS_NAME.BTS_DATA);
        const activeBtsData = btsDataNodes.find(function (btsData) {
            const parameters = btsData.getDecodedParameters();
            return parameters && parameters.stateInfo && parameters.stateInfo.proceduralState === dspsConstants.STATE_INFO.ACTIVATED;
        });
        var np = activeBtsData && activeBtsData.parent;
        if (np) {
            const smodGroupANodes = np.findByClassName(this.nodeAName);
            smodGroupANodes.forEach(function (node) {
                node.parameters = node.getDecodedParameters();
                if (node.parameters && node.parameters.realDN)
                    smodGroupMap[node.parameters.realDN] = node;
            });
        }
        return smodGroupMap;
    }

    function countBBU(smodGroup) {
        const smodGroupNode = imTree.findByDistName(smodGroup.distName);
        const bbuNodes = smodGroupNode.findByClassName(dspsConstants.CLASS_NAME.BBU);
        const dspsCount = bbuNodes.map(function (bbuNode) {
            const dspsNodes = bbuNode.findByClassName(dspsConstants.CLASS_NAME.DSP);
            return dspsNodes.length;
        });
        return dspsCount.reduce(function (memo, count) {
            return memo + count;
        }, 0);
    }

    function fillModuleData(smodGroupDN, smods, bbmods, connectorsDN, channelDspCount) {
        var className = distNameService.getClassName(smodGroupDN);
        var modules = className === dspsConstants.CLASS_NAME.SMOD ? smods : bbmods;
        var moduleData = modules.findModule(smodGroupDN, connectorsDN);
        var label = createModuleLabel(moduleData.smodGroup, moduleData.nodeAData, connectorsDN);
        var dspUsageRatio = channelDspCount / modules.countBBU(moduleData.smodGroup);
        return {
            distName: smodGroupDN,
            prodCode: moduleData.nodeAData.prodCode,
            prodCodePlanned: moduleData.nodeAData.prodCodePlanned,
            prodName: moduleData.nodeAData.prodName,
            label: label,
            dspUsageClass: getDspUsageClass(dspUsageRatio)
        };
    }

    function getDspUsageClass(ratio) {
        if (ratio < USAGE_LEVEL.FREE)
            return 'free';
        else if (ratio < USAGE_LEVEL.NORMAL)
            return 'normal';
        else if (ratio <= USAGE_LEVEL.CRITICAL)
            return 'critical';
        return '';
    }

    function findDevice(modules, distName) {
        return modules.find(function (smodGroup) {
            return smodGroup.distName === distName;
        });
    }

    function findNodeAData(module, ModMap) {
        if (module.distName && ModMap) {
            var node = ModMap[module.distName];
            if (node) {
                var name = node.className + '-' + node.id;
                var parameters = getParametersFromImTree(module.distName);
                var moduleProdCode = module.className === dspsConstants.CLASS_NAME.SMOD ? parameters.visibleProductCode : module.pc;
                var prodCode = dspsConstants.LABELS.MODULE.PC + ' ' + formatEqualAndDifferentValues(node.parameters.prodCode, moduleProdCode);
                var prodCodePlanned = dspsConstants.LABELS.MODULE.PLANNED_PC + ' ' + node.parameters.prodCodePlanned;
                var prodName = ignoreUnknownValue(formatEqualAndDifferentValues(node.parameters.prodName, module.pn));
            }
        }
        return {
            name: name,
            prodCode: prodCode,
            prodCodePlanned: prodCodePlanned,
            prodName: prodName
        };
    }

    function getParametersFromImTree(distName) {
        var node = imTree.findByDistName(distName);
        node.parameters = node.getDecodedParameters();
        return node.parameters;
    }

    function ignoreUnknownValue(value) {
        return value === 'Unknown' ? '' : value;
    }

    function formatEqualAndDifferentValues(label1, label2) {
        if (label1 === label2)
            return label1 || '';
        if (label1)
            return createDoubleLabel(label1, label2);
        return label2 || '';
    }

    function findChildConnectorDN(connectors, modDistName) {
        return connectors.find(function (devicePath) {
            return distNameService.getParent(devicePath) === modDistName;
        });
    }

    function createModuleLabel(smodGroup, nodeAData, connectorsDN) {
        var connectorDN = findChildConnectorDN(connectorsDN, smodGroup.distName);
        var smodGroupAName = nodeAData.name;
        var smodGroupLabel = findConnectorLabel(smodGroup, connectorDN, dspsConstants.CONNECTORS_SIDE.RIGHT);
        if (smodGroupAName)
            return createDoubleLabel(smodGroupAName, smodGroupLabel);
        return '';
    }

    function createDoubleLabel(label1, label2) {
        return label2 ? label1 + ' / ' + label2 : label1;
    }

    function getDspsNames(dspsPath) {
        const dsps = dspsPath.map(function (device) {
            return {
                name: distNameService.getName(device),
                distName: device
            };
        });
        return dsps.sort(function (dsp1, dsp2) {
            return dsp1.name - dsp2.name;
        });
    }

    function createChannelLabel(channel, hardware, connectorDNs, rmodDN) {
        const rmod = findDevice(hardware.rmods, rmodDN);
        const connectorLabel = connectorDNs.map(function (connectorDN) {
            return findConnectorLabel(rmod, connectorDN, dspsConstants.CONNECTORS_SIDE.RIGHT);
        }).find(function (label) {
            return !!label;
        });
        const label = channel.axcAddress ? channel.direction + ' ' + dspsConstants.LABELS.CHANNEL.AXC + ' ' + channel.axcAddress : channel.direction;
        return connectorLabel ? label + ' / ' + connectorLabel : label;
    }

    function fillRmodData(hardware, rmodDistName, connectors) {
        if (rmodDistName) {
            var rmod = findDevice(hardware.rmods, rmodDistName);
            var portLabel = findRmodPortLabel(connectors, rmod, dspsConstants.CONNECTORS_SIDE.LEFT);
            var name = rmod.className + '-' + rmod.number;
            var label = portLabel ? name + ' / ' + portLabel : portLabel;
            return {
                distName: rmodDistName,
                label: label
            };
        }
        return null;
    }

    function findRmodPortLabel(connectors, rmod, connectorsSide) {
        const rmodConnectors = connectors.filter(function (devicePath) {
            return distNameService.getParent(devicePath) === rmod.distName;
        });
        const rmodPortLabels = rmodConnectors.map(function (connector) {
            return findConnectorLabel(rmod, connector, connectorsSide);
        });
        return rmodPortLabels.find(function (connector) {
            return !!connector;
        });
    }

    function findConnectorLabel(mod, distName, connectorsSide) {
        const connector = mod && ((mod[connectorsSide]).find(
                function (port) {
                    return port.distName === distName;
                }));
        return connector ? connector.label : null;
    }
}

