const sic = require('../sic');
const FileTransferConnection = require('./fileTransferConnection');

let requestId = 0;
const FILE_RESP_MSG = 0x6012;
const FILE_ERROR_MSG = 0x6011;

FileTransferService.prototype = Object.create(FileTransferConnection.prototype);

function FileTransferService(authService) {
    FileTransferService.prototype.getFile = getFile;
    FileTransferConnection.call(this, authService);

    const that = this;
    that.sicInstance = sic();

    function getFile(filePath) {
        return new Promise((resolve, reject) => {
            let errorWatcher = null;
            let responseWatcher = null;
            const currentRequestId = requestId++;
            const webSocketListener = {
                disconnected: disconnected,
                connectionError: connectionError
            };

            try {
                const msg = createMessage(filePath, currentRequestId);
                that.webSocketClient.addListener(webSocketListener);
                errorWatcher = registerErrorListener();
                responseWatcher = registerFileResponseListener();
                that.sendMessage(msg);
            } catch (exception) {
                removeListeners();
                const errorObject = {
                    type: 'File request error',
                    path: filePath,
                    error: exception
                };
                reject(errorObject);
            }

            function registerErrorListener() {
                return that.webSocketClient.addMessageListener(FILE_ERROR_MSG, error);

                function error(errorMessage) {
                    if (errorMessage.getRequestId() !== currentRequestId) return;

                    removeListeners();
                    const errorObject = {
                        type: 'File download error',
                        path: filePath,
                        error: errorMessage.getDescription()
                    };
                    reject(errorObject);
                }
            }

            function registerFileResponseListener() {
                return that.webSocketClient.addMessageListener(FILE_RESP_MSG, responseReceived);

                function responseReceived(message) {
                    if (message.getRequestId() !== currentRequestId) return;

                    removeListeners();
                    try {
                        const file = message.getContent();
                        resolve(file);
                    } catch (error) {
                        reject({
                            type: 'File checkSum incorrect',
                            path: filePath,
                            error: 'Calculated checkSum is different then expected.'
                        });
                    }
                }
            }

            function disconnected() {
                removeListeners();
                reject({
                    type: 'Connection closed error',
                    path: filePath,
                    error: 'Connection closed during file transfer.'
                });
            }

            function connectionError() {
                removeListeners();
                reject({
                    type: 'WebSocket connection error',
                    path: filePath,
                    error: 'WebSocket connection error occurred during file transfer.'
                });
            }

            function removeListeners() {
                errorWatcher.unwatch();
                responseWatcher.unwatch();
                that.webSocketClient.removeListener(webSocketListener);
            }
        }
        );
    }

    function createMessage(path, rId) {
        const msg = that.sicInstance.imgwFileServiceGetReq();
        msg.setRequestId(rId)
            .setUri(path);
        return msg;
    }
}

module.exports = FileTransferService;
