const _ = require('lodash'),
    loggers = require('./logger.impl'),
    log = loggers.getLogger('admin.core.historyFactory');

module.exports = {
    newInstance: newInstance
};

function newInstance(snapshotStep) {
    const SNAPSHOT_STEP = 10000;
    snapshotStep = snapshotStep || SNAPSHOT_STEP;
    var snapshots = [];
    var history = [];
    var allHistoryEntryMerger = createEntryMerger();

    return {
        getEntriesIterator: getEntriesIterator,
        addEntry: addEntry,
        getEntry: getEntryById,
        hasEntry: hasEntryWithId,
        getEntries: getEntries,
        getLastEntry: getLastEntry,
        getFirstEntry: getFirstEntry,
        dropEntriesRange: dropEntriesRange,
        size: size,
        isEmpty: isEmpty,
        clear: clear,
        snapshotForEntryId: snapshotForEntryId,
        snapshotForIndex: snapshotForIndex,
        snapshotEntryForId: snapshotEntryForId,
        snapshotsSize: snapshotsSize,
        mergeEntries: mergeEntries,
        mergeFirstEntries: mergeFirstEntries,
        mergeEntriesRange: mergeEntriesRange
    };

    function dropEntriesRange(fromId, toId) {
        history = history.slice(0, fromId).concat(history.slice(toId));
        log.debug('Entries from ' + fromId + ' to ' + toId + ' dropped.');
        adjustEntries();
        adjustSnapshots();

        function adjustEntries() {
            for (var i = fromId; i < size(); i++) {
                history[i].id = i;
            }
        }

        function adjustSnapshots() {
            log.debug('before drop - history snapshots#' + snapshotsSize());
            const firstEntry = getFirstEntry();
            removeAllSnapshotsFromRange();
            adjustIds();
            log.debug('after drop - history snapshots#' + snapshotsSize());

            function removeAllSnapshotsFromRange() {
                _.remove(snapshots, function (snapshot) {
                    return snapshot.lastEntryId >= fromId && snapshot.lastEntryId < toId;
                });
            }

            function adjustIds() {
                const diff = toId - fromId;
                for (var i = snapshots.length - 1; i >= fromId; i--) {
                    snapshots[i].lastEntryId = snapshots[i].lastEntryId - diff;
                }
            }
        }
    }

    function getEntriesIterator() {
        const copyOfHistory = history.slice(0);
        var index = 0;
        return {
            size: function () {
                return copyOfHistory.length;
            },
            next: function () {
                if (index >= copyOfHistory.length) return {done: true};

                return {
                    value: copyOfHistory[index++],
                    done: false
                };
            }
        };
    }

    function hasAlarmClass(distName) {
        return distName.indexOf('ALARM_L') >= 0;
    }

    function hasAlarm(entry) {
        const change = entry.change;
        if (!change) return false;

        for (var i = 0; i < change.updated.length; i++) {
            var managedObject = change.updated[i];
            if (hasAlarmClass(managedObject.distName)) return true;
        }
        for (var j = 0; j < change.removed.length; j++) {
            var distName = change.removed[j];
            if (hasAlarmClass(distName)) return true;
        }
        return false;
    }

    function addEntry(entry) {
        entry.id = isEmpty() ? 0 : getLastEntry().id + 1;
        entry.hasAlarm = hasAlarm(entry);
        entry.estimatedSizeInBytes = estimateByteSizeOfEntry(entry);
        history.push(entry);

        if (entry.id % snapshotStep === 0) { //includes entry id = 0
            snapshots.push(createSnapshot(entry));
        }

        allHistoryEntryMerger.mergeEntry(entry);
    }

    function estimateByteSizeOfEntry(entry) {
        var size = 1024 + sizeOfObjectKeys(entry);
        if (!entry.change) return size;

        const change = entry.change;
        size += sizeOfObjectKeys(change);
        for (var i = 0; i < change.removed.length; i++) {
            size += sizeOfString(change.removed[i]);
        }
        for (var i = 0; i < change.updated.length; i++) {
            var update = change.updated[i];
            size += 50;
            size += sizeOfString(update.distName);
            size += update.payload ? update.payload.byteLength : 0;
        }
        return size;

        function sizeOfString(s) {
            return s.length * 2;
        }

        function sizeOfObjectKeys(o) {
            return Object.keys(o).length * 10;
        }
    }

    function mergeFirstEntries(n) {
        if (isEmpty() || size() === 1) return history;

        const length = Math.min(n, history.length);
        if (length < 2) {
            log.info('Merging - skipping merge because of [ ' + length + ' < 2 ]');
            return;
        }

        const merger = createEntryMerger();
        log.info('Merging first ' + length + ' entries');
        for (var i = 0; i < length; i++) {
            merger.mergeEntry(history[i]);
        }

        history = history.slice(length - 1);
        history[0] = merger.toEntry();
        adjustEntriesAfterMerge();
        adjustSnapshotsAfterMerge();
        log.info('Merge of first ' + length + ' entries done');
        log.info('Merged entry sizeMB = ' + fromBytesToMBytes(history[0].estimatedSizeInBytes));
        log.info('Entries range after merge: [' + _.first(history).id + ', ' + _.last(history).id + ']');
        return history;

        function adjustEntriesAfterMerge() {
            for (var i = 0; i < size(); i++) {
                history[i].id = i;
            }
        }

        function adjustSnapshotsAfterMerge() {
            var diff = n - 1;
            log.debug('before merge - history snapshots#' + snapshotsSize());
            const firstEntry = getFirstEntry();
            removeAllSnapshotsLessOrEqual(diff);
            adjustIds();
            snapshots.unshift(createSnapshot(firstEntry));
            log.debug('after merge - history snapshots#' + snapshotsSize());

            function removeAllSnapshotsLessOrEqual(entryId) {
                _.remove(snapshots, function (snapshot) {
                    return snapshot.lastEntryId <= entryId;
                });
            }

            function adjustIds() {
                for (var i = 0; i < snapshots.length; i++) {
                    snapshots[i].lastEntryId = snapshots[i].lastEntryId - diff;
                }
            }
        }
    }

    function fromBytesToMBytes(bytes) {
        return (bytes / 1024 / 1024).toFixed(3);
    }

    function mergeEntriesRange(fromId, toId) {
        if (isEmpty() || size() === 1) return history;

        if (toId - fromId < 2) {
            log.info('Merging - skipping merge because number of entries to merge is < 2.');
            return;
        }
        if (fromId < 0 || toId > size()) {
            log.info('Merging - skipping merge because parameters are out of history size boundary.');
            return;
        }

        const merger = createEntryMerger();
        log.info('Merging entries from ' + fromId + ' to ' + toId + ' (excluding).');
        for (var i = fromId; i < toId; i++) {
            merger.mergeEntry(history[i]);
        }

        var beginning = history.slice(0, fromId);
        var ending = history.slice(toId);
        history = beginning.concat([merger.toEntry()]).concat(ending);
        adjustEntriesAfterMerge();
        adjustSnapshotsAfterMerge();
        log.info('Merge entries from ' + fromId + ' to ' + toId + ' (excluding) done.');
        log.info('Merged entries size in MB = ' + fromBytesToMBytes(history[0].estimatedSizeInBytes) + '.');
        log.info('Entries range after merge: [' + _.first(history).id + ', ' + _.last(history).id + ']');
        return history;

        function adjustEntriesAfterMerge() {
            for (var i = fromId; i < size(); i++) {
                history[i].id = i;
            }
        }

        function adjustSnapshotsAfterMerge() {
            log.debug('before merge - history snapshots#' + snapshotsSize());
            const firstEntry = getFirstEntry();
            removeAllSnapshotsFromRange();
            adjustIds();
            snapshots.unshift(createSnapshot(firstEntry));
            log.debug('after merge - history snapshots#' + snapshotsSize());

            function removeAllSnapshotsFromRange() {
                _.remove(snapshots, function (snapshot) {
                    return snapshot.lastEntryId >= fromId && snapshot.lastEntryId < toId;
                });
            }

            function adjustIds() {
                const diff = toId - fromId - 1;
                for (var i = snapshots.length - 1; i >= fromId; i--) {
                    snapshots[i].lastEntryId = snapshots[i].lastEntryId - diff;
                }
            }
        }
    }

    function createSnapshot(entry) {
        const prevSnapshot = findSnapshotForEntryId(entry.id);
        if (!prevSnapshot) return snapshotForCurrentEntry();

        return snapshotWithPrevState(prevSnapshot);

        function snapshotForCurrentEntry() {
            const map = new Map();
            add(map, entry);
            return {
                lastEntryId: entry.id,
                map: map
            };
        }

        function snapshotWithPrevState(prevSnapshot) {
            const map = new Map(prevSnapshot.map);
            const startId = prevSnapshot.lastEntryId;
            const endId = startId + Math.min(snapshotStep, history.length - 1);
            for (var i = startId; i <= endId; i++) {
                add(map, getEntryById(i));
            }
            return {
                lastEntryId: endId,
                map: map
            };
        }
    }

    function findSnapshotForEntryId(entryId) {
        for (var i = snapshotsSize() - 1; i >= 0; i--) {
            var snapshot = snapshots[i];
            if (entryId >= snapshot.lastEntryId) return snapshot;
        }
    }

    function snapshotForEntryId(entryId) {
        const snapshot = findSnapshotForEntryId(entryId);
        if (!snapshot) return;

        if (entryId === snapshot.lastEntryId) return snapshot.map;

        const map = new Map(snapshot.map);
        for (var i = snapshot.lastEntryId; i <= entryId; i++) {
            add(map, getEntryById(i));
        }
        return map;
    }

    function clear() {
        allHistoryEntryMerger = createEntryMerger();
        history = [];
        snapshots = [];
    }

    function add(map, entry) {
        if (entry.change) {
            const change = entry.change;
            for (var i = 0; i < change.removed.length; i++) {
                map.delete(change.removed[i]);
            }
            for (var i = 0; i < change.updated.length; i++) {
                var update = change.updated[i];
                map.set(update.distName, {
                    payload: update.payload,
                    metaUid: entry.metaUid
                });
            }
        }
        if (entry.reset) map.clear();
    }

    function snapshotsSize() {
        return snapshots.length;
    }

    function snapshotForIndex(index) {
        return snapshots[index] && snapshots[index].map;
    }

    function snapshotEntryForId(index) {
        return createEntry(snapshotForEntryId(index));

        function createEntry(snapshotMap) {
            const updated = mapToArray(snapshotMap);
            var entry = {
                numOfChanges: updated.length,
                change: {updated: updated, removed: []}
            };
            entry.hasAlarm = hasAlarm(entry);
            entry.estimatedSizeInBytes = estimateByteSizeOfEntry(entry);
            return entry;

            function mapToArray(map) {
                const array = [];
                map.forEach(function (value, key) {
                    value.distName = key;
                    array.push(value);
                });
                return array;
            }
        }
    }

    function getEntryById(entryId) {
        if (isEmpty()) return;

        const entryIndex = entryId - getFirstEntry().id;
        return history[entryIndex];
    }

    function hasEntryWithId(entryId) {
        return !!getEntryById(entryId);
    }

    function getEntries() {
        return history.slice(0);
    }

    function size() {
        return history.length;
    }

    function isEmpty() {
        return size() === 0;
    }

    function getFirstEntry() {
        return history[0];
    }

    function getLastEntry() {
        return history[size() - 1];
    }

    function mergeEntries() {
        const mergedEntry = allHistoryEntryMerger.toEntry();
        clear();
        addEntry(mergedEntry);
        return mergedEntry;
    }

    function createEntryMerger() {
        var startEntryId = null;
        var endEntryId = null;
        var updatesMap = new Map();
        var metaUid = 'NO META UID';
        var timestamp = -1;

        return {
            mergeEntry: mergeEntry,
            toEntry: toEntry,
            getEntryRange: getEntryRange
        };

        function mergeEntry(entry) {
            if (entry.change) {
                mergeChange(entry);
            } else if (entry.reset) {
                mergeReset(entry);
            }
        }

        function mergeChange(entry) {
            setIds(entry);
            metaUid = entry.metaUid;
            timestamp = entry.timestamp;
            const change = entry.change;

            for (var i = 0; i < change.updated.length; i++) {
                const update = change.updated[i];
                updatesMap.set(update.distName, update);
            }
            for (var i = 0; i < change.removed.length; i++) {
                updatesMap.delete(change.removed[i]);
            }
        }

        function mergeReset(entry) {
            setIds(entry);
            updatesMap = new Map();
            metaUid = 'NO META UID';
            timestamp = entry.timestamp;
        }

        function getEntryRange() {
            return {
                startId: startEntryId,
                endId: endEntryId
            };
        }

        function setIds(entry) {
            if (startEntryId == null) startEntryId = entry.id;
            endEntryId = entry.id;
        }

        function toEntry() {
            const updated = mapToArray(updatesMap);
            var entry = {
                id: endEntryId,
                timestamp: timestamp,
                numOfChanges: updated.length,
                change: {updated: updated, removed: []},
                metaUid: metaUid
            };
            entry.hasAlarm = hasAlarm(entry);
            entry.estimatedSizeInBytes = estimateByteSizeOfEntry(entry);
            return entry;
        }

        function mapToArray(map) {
            const array = [];
            map.forEach(function (value) {
                array.push(value);
            });
            return array;
        }
    }
}

