const _ = require('lodash');
const OnlineImTree = require('../onlineImTree');
const ObjectListenersContainer = require('./objectListenersContainer');
const logger = require('../logger.impl');
const EVENTS = require('./imTrackerEvents');
const distNameUtils = require('../utils/distNameUtils');

const log = logger.getLogger('admin.core.imTracker');

module.exports = imTracker;

function imTracker({
    onlineImTree = OnlineImTree()
    } = {}) {

    const distNameListeners = new Map(),
        classNameListeners = new Map(),
        imTreeListener = imUpdate;

    var imTreeUnwatcher;
    var imTree = new infoModel.ImTree();

    return {
        trackByDistName: trackByDistName,
        trackByClassName: trackByClassName,
        stopTracking: stopTracking,
        //only for testing
        imUpdate: imUpdate
    };


    /**
     * Add listener to track given parameters of the provided class name.
     * @param {String} className
     * @param {Object} listener object containing listening methods:
     * @param {Function=} listener.onListenerRegistered ({eventType, objects}) Callback for handling first state of an object
     * @param {Function=} listener.onObjectAppeared ({eventType, appearedDistNames}) Callback for handling first occurrence of an object
     * @param {Function=} listener.onObjectRemoved ({eventType, removedDistNames}) Callback invoked when object is removed from InfoModel
     * @param {Function=} listener.onParameterChanged ({changedParameters, objects}) Callback invoked when parameters have
     * been changed for the object. Objects is a list of objects that has structure {distName: String, parameters:{}}
     * @param {Function=} listener.onParametersChanged ({eventType, objects}) Callback invoked when something in the object with appropriate
     * classname has changed, it gives all objects. Objects is a list of objects that has structure {distName: String, parameters:{}}
     * @param {Array=} parameters Array with parameters names as strings, if not provided it will track entire object
     * and invoke onChanged method instead of onParameterChanged on given listener
     */

    function trackByClassName(className, listener, parameters) {
        return registerListenerIfNecessary().then(function () {
            var classNameListener = getListenersContainer(classNameListeners, className),
                objectListener = {
                    className: className,
                    callbacks: listener,
                    registeredParameters: parameters
                };
            classNameListener.addListener(objectListener);
            log.info(className + ' class has been registered to tracking');
            var nodes = imTree.findByClassName(className);
            if (!_.isEmpty(nodes)) {
                processOnRegisteredListeners(classNameListener, nodes, function (objects) {
                    return {
                        eventType: EVENTS.LISTENER_REGISTERED, objects: objects
                    }
                });
            }
            return listener;
        });
    }

    /**
     * Add listener to track given parameters of the provided dist name.
     * @param {String} distName
     * @param {Object} listener object containing listening methods:
     * @param {Function=} listener.onListenerRegistered ({eventType, allParameters}) Callback for handling first state of an object
     * @param {Function=} listener.onObjectAppeared ({eventType}) Callback for handling first occurrence of an object
     * @param {Function=} listener.onObjectRemoved ({eventType}) Callback invoked when object is removed from InfoModel
     * @param {Function=} listener.onParameterChanged ({eventType, changedParameters}) Callback invoked when parameters have
     * been changed for the object
     * @param {Function=} listener.onParametersChanged ({eventType, allParameters}) Callback invoked when something in the object has changed,
     * it gives all parameters of the object (allParameters);
     * @param {Function=} listener.onChildAdded ({eventType, childDistNames}) Callback invoked when new child for the object has been added
     * @param {Array=} parameters Array with parameters names as strings, if not provided it will track entire object
     * and invoke onChanged method instead of onParameterChanged on given listener
     */

    function trackByDistName(distName, listener, parameters) {
        return registerListenerIfNecessary().then(function () {
            var distNameListener = getListenersContainer(distNameListeners, distName),
                objectListener = {
                    distName: distName,
                    callbacks: listener,
                    registeredParameters: parameters
                };
            distNameListener.addListener(objectListener);
            log.info(distNameUtils.getName(distName) + ' distName has been registered to tracking');
            var node = imTree.findByDistName(distName);
            if (node) {
                processOnRegisteredListeners(distNameListener, [node], function (objects) {
                    return {
                        eventType: EVENTS.LISTENER_REGISTERED, parameters: _.first(objects).parameters
                    }
                });
            }
            return listener;
        });
    }

    function getListenersContainer(listenersContainerMap, name) {
        var listenersContainer;
        if (listenersContainerMap.has(name)) {
            listenersContainer = listenersContainerMap.get(name);
        } else {
            listenersContainer = new ObjectListenersContainer();
            listenersContainerMap.set(name, listenersContainer);
        }
        return listenersContainer;
    }

    function processOnRegisteredListeners(listenerContainer, nodes, createEventFunction) {
        var objects = [];
        _.each(nodes, function (node) {
            var distName = node.distName,
                parameters = node.getDecodedParameters();
            listenerContainer.addExistedObject(distName);
            objects.push({distName: node.distName, parameters: parameters})
        });

        var event = createEventFunction(objects);

        _.each(listenerContainer.getListeners(), function (listener) {
            notifyAboutEvent(listener, event);
        });
    }

    function processByClassName(listenersContainer, className, updatedDistNames) {
        var nodes = imTree.findByClassName(className);
        if (_.isEmpty(nodes)) {
            preprocessIfNoObjectsExists();
            return;
        }
        var distNames = calculateDistNamesDifferences(nodes),
            currentDistNames = distNames.incoming,
            existedDistNames = distNames.existed,
            removedDistNames = distNames.removed,
            appearedDistNames = distNames.appeared;

        if (!_.isEmpty(removedDistNames)) {
            _.each(removedDistNames, listenersContainer.removeExistedObject, listenersContainer);
            notifyEachListeners(listenersContainer.getListeners(), {
                eventType: EVENTS.OBJECT_REMOVED,
                distNames: removedDistNames
            });
        }

        if (!_.isEmpty(appearedDistNames)) {
            _.each(appearedDistNames, listenersContainer.addExistedObject, listenersContainer);
            notifyEachListeners(listenersContainer.getListeners(), {
                eventType: EVENTS.OBJECT_APPEARED,
                distNames: appearedDistNames
            });
        }

        if (listenersContainer.hasParametersListeners()) {
            const updatePredicate = createUpdatePredicate(updatedDistNames);
            const updatedObjects = _(nodes).filter(updatePredicate).map(function (node) {
                return {
                    distName: node.distName,
                    parameters: node.getDecodedParameters()
                }
            }).value();

            if (!_.isEmpty(updatedObjects)) {
                processParametersListeners(listenersContainer,
                    notifyOnParameterChangeByClassName(updatedObjects, existedDistNames, appearedDistNames, currentDistNames, listenersContainer),
                    notifyOnParametersChangeByClassName(updatedObjects));
                _.each(updatedObjects, function (updatedObject) {
                    listenersContainer.addExistedObject(updatedObject.distName, updatedObject.parameters);
                })
            }
        }


        function calculateDistNamesDifferences(nodes) {
            var incomingDistNames = _(nodes).pluck('distName').value(),
                existedDistNames = listenersContainer.getExistedDistNames(),
                removedDistNames = _.difference(existedDistNames, incomingDistNames),
                appearedDistNames = _.difference(incomingDistNames, existedDistNames);
            return {
                removed: removedDistNames,
                appeared: appearedDistNames,
                incoming: incomingDistNames,
                existed: existedDistNames
            };
        }

        function preprocessIfNoObjectsExists() {
            if (listenersContainer.getExistedObjects().size) {
                notifyEachListeners(listenersContainer.getListeners(), {
                    eventType: EVENTS.OBJECT_REMOVED,
                    distNames: listenersContainer.getExistedDistNames()
                });
                listenersContainer.getExistedObjects().clear();
            }
        }
    }

    function processByDistName(listenersContainer, distName) {
        var node = imTree.findByDistName(distName);

        if (!node) {
            if (listenersContainer.hasExistedObjects()) {
                listenersContainer.removeExistedObject(distName);
                notifyEachListeners(listenersContainer.getListeners(), {eventType: EVENTS.OBJECT_REMOVED});
            }
            return;
        }

        if (!listenersContainer.hasExistedObjects()) {
            listenersContainer.addExistedObject(distName);
            notifyEachListeners(listenersContainer.getListeners(), {eventType: EVENTS.OBJECT_APPEARED});
        }

        if (shouldDecode(node) && listenersContainer.hasParametersListeners()) {
            var updatedParameters = node.getDecodedParameters();
            processParametersListeners(listenersContainer,
                notifyOnParameterChangeByDistName(updatedParameters),
                notifyOnParametersChangeByDistName(updatedParameters));
            listenersContainer.addExistedObject(distName, updatedParameters);
        }

        if (listenersContainer.hasChildListeners()) {
            var childDistNames = [];
            node.forEachChild(function (child) {
                childDistNames.push(child.distName)
            });

            var oldChildDistNames = listenersContainer.children();

            var diff = _.difference(childDistNames, oldChildDistNames);
            if (!_.isEmpty(diff)) {
                notifyEachListeners(listenersContainer.getListeners(), {
                    eventType: EVENTS.CHILD_ADDED,
                    distNames: diff
                });
            }
            listenersContainer.children(childDistNames);
        }
    }

    function notifyAboutChangesInImTree(distNames) {
        distNameListeners.forEach(processByDistName);
        classNameListeners.forEach(function (listenerContainer, className) {
            processByClassName(listenerContainer, className, distNames);
        });
    }

    function registerListenerIfNecessary() {
        if (imTreeUnwatcher) return Promise.resolve();

        return onlineImTree.watchUpdates(imTreeListener)
            .then(function (unwatcher) {
                imTreeUnwatcher = unwatcher;
            });
    }

    function notifyEachListeners(listeners, event) {
        _.each(listeners, function (listener) {
            notifyAboutEvent(listener, event);
        });
    }


    function processParametersListeners(listenersContainer, onParameterCallback, onAllParametersCallback) {
        _.each(listenersContainer.getListeners(), function (listener) {
            if (!hasRegisteredParameters(listener) && hasEventCallback(listener, EVENTS.PARAMETERS_CHANGED)) {
                onAllParametersCallback(listener)
            }
            if (hasRegisteredParameters(listener) && hasEventCallback(listener, EVENTS.PARAMETER_CHANGED)) {
                onParameterCallback(listener, listenersContainer)
            }
        });

        function hasRegisteredParameters(listener) {
            if (!_.isArray(listener.registeredParameters)) return false;
            return !_.isEmpty(listener.registeredParameters)
        }

        function hasEventCallback(listener, eventType) {
            return !_.isUndefined(listener.callbacks[eventType]);
        }
    }

    function notifyOnParametersChangeByDistName(parameters) {
        return function (listener) {
            notifyAboutEvent(listener, {
                eventType: EVENTS.PARAMETERS_CHANGED,
                parameters: parameters
            });
        }
    }

    function notifyOnParametersChangeByClassName(objects) {
        return function (listener) {
            notifyAboutEvent(listener, {
                eventType: EVENTS.PARAMETERS_CHANGED,
                objects: objects
            });
        }
    }

    function notifyOnParameterChangeByClassName(objects, existedDistNames, appearedDistNames, currentDistNames) {
        return function (listener, listenersContainer) {
            var changedParameterObjects = [];
            _.each(objects, function (object) {
                if (areObjectsAppeared(object)) {
                    var parameters = {};
                    _.each(listener.registeredParameters, function (registerParameter) {
                        var registerParameterValue = _.get(object.parameters, registerParameter);
                        if (!_.isUndefined(registerParameterValue)) {
                            parameters[registerParameter] = _.get(object.parameters, registerParameter);
                        }
                    });
                    if (!_.isEmpty(parameters)) {
                        changedParameterObjects.push({distName: object.distName, parameters: object.parameters})
                    }
                } else if (areExistedObjectsUpdated(object)) {
                    var existedObjectsMap = listenersContainer.getExistedObjects(),
                        existedParameters = existedObjectsMap.get(object.distName),
                        changedParameters = getObjectsWithUpdatedParameters(object.parameters, existedParameters, listener.registeredParameters);
                    if (!_.isEmpty(changedParameters)) {
                        changedParameterObjects.push({distName: object.distName, parameters: object.parameters});
                    }
                }
            });

            if (!_.isEmpty(changedParameterObjects)) {
                notifyAboutEvent(listener, {
                    eventType: EVENTS.PARAMETER_CHANGED,
                    objects: changedParameterObjects
                });
            }

            function areObjectsAppeared(object) {
                return _.include(appearedDistNames, object.distName)
            }

            function areExistedObjectsUpdated(object) {
                return _.include(existedDistNames, object.distName) && _.include(currentDistNames, object.distName);
            }
        }
    }

    function notifyOnParameterChangeByDistName(parameters) {
        return function (listener, listenersContainer) {
            var existedObjectsMap = listenersContainer.getExistedObjects();
            var existedParameters = existedObjectsMap.get(listener.distName);
            var changedParameters = getObjectsWithUpdatedParameters(parameters, existedParameters, listener.registeredParameters);
            if (!_.isEmpty(changedParameters)) {
                notifyAboutEvent(listener, {
                    eventType: EVENTS.PARAMETER_CHANGED,
                    parameters: parameters
                });
            }
        }
    }

    function getObjectsWithUpdatedParameters(incomingParameters, existedParameters, registeredParameters) {
        var changedParameters = {};
        _.each(registeredParameters, function (registeredParameter) {
            var newParameterValue = _.get(incomingParameters, registeredParameter);
            if (newParameterValue !== _.get(existedParameters, registeredParameter)) {
                changedParameters[registeredParameter] = newParameterValue;
            }
        });
        return changedParameters;
    }

    function notifyAboutEvent(listener, event) {
        var callback = _.get(listener, "callbacks." + event.eventType, _.noop);
        callback(event);
    }


    /**
     * Removes given listener.
     * @param registeredListener
     * @param  name
     */
    function stopTracking(registeredListener, name) {

        var listenersForName;

        if (distNameListeners.has(name)) {
            listenersForName = distNameListeners;
        } else if (classNameListeners.has(name)) {
            listenersForName = classNameListeners;
        } else {
            log.warn('No listeners is existed for:' + name);
            return;
        }

        removeListeners(listenersForName, name);
        unregisterTrackerIfThereAreNoListeners(listenersForName);

        function removeListeners(nameListeners, name) {
            const listenersContainer = nameListeners.get(name),
                objectListeners = listenersContainer.getListeners();
            _.remove(objectListeners, function (listener) {
                return listener.callbacks === registeredListener;
            });
            if (_.isEmpty(objectListeners)) {
                nameListeners.delete(name);
            }
        }

        function unregisterTrackerIfThereAreNoListeners(nameListeners) {
            if (!nameListeners.size && imTreeListener) {
                imTreeUnwatcher.unwatch();
                imTreeUnwatcher = null;
            }
        }
    }

    function createUpdatePredicate(distNames) {
        return function (node) {
            if (!distNames || !distNames.updated) return false;
            return _.includes(distNames.updated, node.distName);
        }
    }

    function shouldDecode(node) {
        return !node.parameters;
    }

    function imUpdate(updateEvent) {
        imTree = updateEvent.completeImTree;
        const change = updateEvent.change;
        const updated = change && change.updated;
        notifyAboutChangesInImTree({updated: updated});
    }
}
