const ProtoBuf = require('protobufjs');
const IMI_CONSTANTS = require('./imiConstants');
const imiDefinitions = require('./imiDefinitions');
const _ = require('lodash');
const converter = require('../bytesToStringConverter');

function toString(msgName, msgContent) {
    var content = '';
    _.forOwn(msgContent, function (value, key) {
        var isGetter = _.startsWith(key, 'get') && typeof value === 'function';
        if (isGetter) {
            var theValue = value();
            if (_.isNumber(theValue)) theValue += ' [0x' + theValue.toString(16) + ']';
            if (theValue.byteLength) theValue += ' [' + theValue.byteLength + 'bytes]';

            content += '\n' + key + ': ' + theValue;
        }
    });
    return msgName + '{' + content + '\n}';
}

function encodeProto(proto) {
    var buffer = proto.encode();
    return buffer.toArrayBuffer();
}

function decodeProto(protoMessage, protoBuffer) {
    return protoMessage.decode(protoBuffer);
}

function getAvailableVersions() {
    var versions = [];
    imiDefinitions.forEach(function (value, key) {
        versions.push(key);
    });
    return versions.sort();
}

function createMessages(imiVersion) {

    var theDefinition = imiDefinitions.get(imiVersion);
    if (!theDefinition) throw new Error('Unsupported imi version: ' + imiVersion + ' available versions: ' + getAvailableVersions());

    //WTF loadJson changes input parameter!!!
    var imi = ProtoBuf.loadJson(_.cloneDeep(theDefinition)).build('imi.v' + imiVersion);

    function getImiVersion() {
        return imiVersion;
    }

    function aaConfigGetServiceReq(hostLittleEndianness) {
        // name M char[80] Name of service
        // major_version M u32 Major version number of service
        // minor_version M u32 Minor version number of service
        // scope M EAaConfigSvcRegScope (u32) Visibility of service

        var nameSize = 80;
        var name = 0;
        var majorVersion = 80;
        var minorVersion = 84;
        var scope = 88;

        var buffer = new ArrayBuffer(92);
        var dataView = new DataView(buffer);

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.AA_CONFIG_GET_SERVICE_REQ_ID;
        };
        theMsg.setName = function (str) {
            converter.convertStringToBytes(dataView, name, nameSize, str);
        };
        theMsg.getName = () => {
            return converter.convertBytesToString(dataView, name, nameSize);
        };
        theMsg.setMajorVersion = (value) => {
            dataView.setUint32(majorVersion, value, hostLittleEndianness);
        };
        theMsg.getMajorVersion = () => {
            return dataView.getUint32(majorVersion, hostLittleEndianness);
        };
        theMsg.setMinorVersion = (value) => {
            dataView.setUint32(minorVersion, value, hostLittleEndianness);
        };
        theMsg.getMinorVersion = () => {
            return dataView.getUint32(minorVersion, hostLittleEndianness);
        };
        theMsg.setScope = (value) => {
            dataView.setUint32(scope, value, hostLittleEndianness);
        };
        theMsg.getScope = () => {
            return dataView.getUint32(scope, hostLittleEndianness);
        };
        theMsg.getBuffer = () => {
            return buffer;
        };
        theMsg.setBuffer = (value) => {
            buffer = value;
            dataView = new DataView(value);
        };
        theMsg.toString = () => {
            return toString('aaConfigGetServiceReq', theMsg);
        };
        return theMsg;
    }

    function aaConfigGetServiceRsp(hostLittleEndianness) {
        //name M char[] Name of service [size 80 bytes]
        //major_version M u32 Major version number of service
        //minor_version M u32 Minor version number of service
        //scope M EAaConfigSvcRegScope Visibility of service
        //num_services M u32 Number of services
        //service[] M SAaConfigSvcReg/ Minor version u32 + SIC address u32 of service GetResService

        var nameSize = 80;
        var name = 0;
        var majorVersion = 80;
        var minorVersion = 84;
        var scope = 88;
        var numServices = 92;
        var minor = 96;
        var sicAddress = 100;

        var buffer = new ArrayBuffer(104);
        var dataView = new DataView(buffer);

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.AA_CONFIG_GET_SERVICE_RSP_ID;
        };
        theMsg.setName = (str) => {
            converter.convertStringToBytes(dataView, 0, nameSize, str);
        };
        theMsg.getName = () => {
            return converter.convertBytesToString(dataView, 0, nameSize);
        };
        theMsg.setMajorVersion = (value) => {
            dataView.setUint32(majorVersion, value, hostLittleEndianness);
        };
        theMsg.getMajorVersion = () => {
            return dataView.getUint32(majorVersion, hostLittleEndianness);
        };
        theMsg.setMinorVersion = (value) => {
            dataView.setUint32(minorVersion, value, hostLittleEndianness);
        };
        theMsg.getMinorVersion = () => {
            return dataView.getUint32(minorVersion, hostLittleEndianness);
        };
        theMsg.setScope = (value) => {
            dataView.setUint32(scope, value, hostLittleEndianness);
        };
        theMsg.getScope = () => {
            return dataView.getUint32(scope, hostLittleEndianness);
        };
        theMsg.setNumServices = (value) => {
            dataView.setUint32(numServices, value, hostLittleEndianness);
        };
        theMsg.getNumServices = () => {
            return dataView.getUint32(numServices, hostLittleEndianness);
        };
        theMsg.setMinor = (value) => {
            dataView.setUint32(minor, value, hostLittleEndianness);
        };
        theMsg.getMinor = () => {
            return dataView.getUint32(minor, hostLittleEndianness);
        };
        theMsg.setSicAddress = (value) => {
            dataView.setUint32(sicAddress, value, hostLittleEndianness);
        };
        theMsg.getSicAddress = () => {
            return dataView.getUint32(sicAddress, hostLittleEndianness);
        };
        theMsg.getBuffer = () => {
            return buffer;
        };
        theMsg.setBuffer = (value) => {
            buffer = value;
            dataView = new DataView(value);
        };
        theMsg.toString = () => {
            return toString('aaConfigGetServiceRsp', theMsg);
        };
        return theMsg;
    }

    function aaSysComEcho(hostLittleEndianness) {
        var buffer = new ArrayBuffer(4);
        var dataView = new DataView(buffer);
        dataView.setUint32(0, 0xDEADBEEF, hostLittleEndianness);

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.AA_SYS_COM_ECHO_ID;
        };
        theMsg.getBuffer = () => {
            return buffer;
        };
        theMsg.setBuffer = (value) => {
            buffer = value;
            dataView = new DataView(value);
        };
        theMsg.toString = () => {
            return toString('aaSysComEcho', theMsg);
        };
        return theMsg;
    }

    function signIn() {
        var ProtoMessage = imi.IMSignInRequest;

        var proto = new ProtoMessage({
            'rat': 3,
            'update_notification_interval_ms': 1000,
            'platform_im_version': {
                'release': 'main',
                'major': 1,
                'minor': 1
            }
        });

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.SIGN_IN_ID;
        };
        theMsg.setRat = (value) => {
            proto.rat = value;
        };
        theMsg.getRat = () => {
            return proto.rat;
        };
        theMsg.setUpdateIntervalMs = (value) => {
            proto.update_notification_interval_ms = value;
        };
        theMsg.getUpdateIntervalMs = () => {
            return proto.update_notification_interval_ms;
        };
        theMsg.setRelease = (value) => {
            proto.platform_im_version.release = value;
        };
        theMsg.getRelease = () => {
            return proto.platform_im_version.release;
        };
        theMsg.setMajor = (value) => {
            proto.platform_im_version.major = value;
        };
        theMsg.getMajor = () => {
            return proto.platform_im_version.major;
        };
        theMsg.setMinor = (value) => {
            proto.platform_im_version.minor = value;
        };
        theMsg.getMinor = () => {
            return proto.platform_im_version.minor;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return toString('signIn', theMsg);
        };
        return theMsg;
    }

    function signOut() {
        var buffer = new ArrayBuffer(0);

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.SIGN_OUT_ID;
        };
        theMsg.getBuffer = () => {
            return buffer;
        };
        theMsg.setBuffer = (value) => {
            buffer = value;
        };
        theMsg.toString = () => {
            return toString('signOut', theMsg);
        };
        return theMsg;
    }

    function signInResult() {
        var ProtoMessage = imi.IMSignInResult;

        var proto = new ProtoMessage({
            'status': 0,
            'requested_version': {
                'release': 'main',
                'major': 1,
                'minor': 1
            },
            'supported_version': {
                'release': 'main',
                'major': 1,
                'minor': 1
            }
        });

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.SIGN_IN_RESULT;
        };
        theMsg.setStatus = (value) => {
            proto.status = value;
        };
        theMsg.getStatus = () => {
            return proto.status;
        };
        theMsg.setRequestedRelease = (value) => {
            proto.requested_version.release = value;
        };
        theMsg.getRequestedRelease = () => {
            return proto.requested_version.release;
        };
        theMsg.setRequestedMajor = (value) => {
            proto.requested_version.major = value;
        };
        theMsg.getRequestedMajor = () => {
            return proto.requested_version.major;
        };
        theMsg.setRequestedMinor = (value) => {
            proto.requested_version.minor = value;
        };
        theMsg.getRequestedMinor = () => {
            return proto.requested_version.minor;
        };
        theMsg.setSupportedRelease = (value) => {
            proto.supported_version.release = value;
        };
        theMsg.getSupportedRelease = () => {
            return proto.supported_version.release;
        };
        theMsg.setSupportedMajor = (value) => {
            proto.supported_version.major = value;
        };
        theMsg.getSupportedMajor = () => {
            return proto.supported_version.major;
        };
        theMsg.setSupportedMinor = (value) => {
            proto.supported_version.minor = value;
        };
        theMsg.getSupportedMinor = () => {
            return proto.supported_version.minor;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return toString('signInResult', theMsg);
        };
        return theMsg;
    }

    function updateNotification() {
        var ProtoMessage = imi.IMUpdateNotification;

        var proto = new ProtoMessage({});

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.UPDATE_NOTIFICATION_ID;
        };
        theMsg.getRemovedObjectIds = () => {
            return proto.removed_objects_ids || [];
        };
        theMsg.setRemovedObjectIds = (value) => {
            proto.removed_objects_ids = value;
        };
        theMsg.getManagedObjects = () => {
            return proto.objects || [];
        };
        theMsg.setManagedObjects = (value) => {
            proto.objects = value;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return 'updateNotification { \n id: 0x' + theMsg.getId().toString(16) +
                '\n removedObjectIds: ' + theMsg.getRemovedObjectIds().length +
                '\n managedObjects: ' + theMsg.getManagedObjects().length +
                '\n}';
        };
        return theMsg;
    }

    function changeRequest() {
        var ProtoMessage = imi.IMChangeRequest;

        var proto = new ProtoMessage({
            'request_id': 1,
            'operations': []
        });

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.CHANGE_REQUEST_ID;
        };
        theMsg.setRequestId = (value) => {
            proto.request_id = value;
        };
        theMsg.getRequestId = () => {
            return proto.request_id;
        };
        theMsg.getOperations = () => {
            return proto.operations;
        };
        theMsg.setOperations = (value) => {
            proto.operations = value;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return toString('changeRequest', theMsg);
        };
        return theMsg;
    }

    function operationExecuted() {
        var ProtoMessage = imi.IMOperationExecuted;

        var proto = new ProtoMessage({
            'request_id': 1,
            'execution_status': 1
        });

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.OPERATION_EXECUTED_ID;
        };
        theMsg.getRequestId = () => {
            return proto.request_id;
        };
        theMsg.getExecutionStatus = () => {
            return proto.execution_status;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return toString('operationExecuted', theMsg);
        };
        return theMsg;
    }

    function registerExecutor() {
        var ProtoMessage = imi.IMRegisterExecutor;

        var proto = new ProtoMessage({
            'request_id': 1,
            'object_types': [],
            'object_ids': []
        });

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.REGISTER_EXECUTOR_ID;
        };
        theMsg.getRequestId = () => {
            return proto.request_id;
        };
        theMsg.setRequestId = (id) => {
            proto.request_id = id;
        };
        theMsg.getObjectTypes = () => {
            return proto.object_types;
        };
        theMsg.setObjectTypes = (types) => {
            proto.object_types = types;
        };
        theMsg.getObjectIds = () => {
            return proto.object_ids;
        };
        theMsg.setObjectIds = (ids) => {
            proto.object_ids = ids;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return toString('registerExecutor', theMsg);
        };
        return theMsg;
    }

    function executorRegistered() {
        var ProtoMessage = imi.IMExecutorRegistered;

        var proto = new ProtoMessage({
            'request_id': 1,
            'registration_status': 1
        });

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.EXECUTOR_REGISTERED_ID;
        };
        theMsg.getRequestId = () => {
            return proto.request_id;
        };
        theMsg.getRegistrationStatus = () => {
            return proto.registration_status;
        };
        theMsg.setRequestId = function (id) {
            proto.request_id = id;
        };
        theMsg.setRegistrationStatus = function (status) {
            proto.registration_status = status;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return toString('executorRegistered', theMsg);
        };
        return theMsg;
    }

    function unregisterExecutor() {
        var ProtoMessage = imi.IMUnregisterExecutor;

        var proto = new ProtoMessage({
            'object_types': [],
            'object_ids': []
        });

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.UNREGISTER_EXECUTOR_ID;
        };
        theMsg.getObjectTypes = () => {
            return proto.object_types;
        };
        theMsg.setObjectTypes = (types) => {
            proto.object_types = types;
        };
        theMsg.getObjectIds = () => {
            return proto.object_ids;
        };
        theMsg.setObjectIds = (ids) => {
            proto.object_ids = ids;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return toString('unregisterExecutor', theMsg);
        };
        return theMsg;
    }

    function executionRequest() {
        var ProtoMessage = imi.IMExecutionRequest;

        var proto = new ProtoMessage({
            'request_id': 1,
            'operations': []
        });

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.EXECUTION_REQUEST_ID;
        };
        theMsg.getRequestId = () => {
            return proto.request_id;
        };
        theMsg.getOperations = () => {
            return proto.operations;
        };
        theMsg.setRequestId = (id) => {
            proto.request_id = id;
        };
        theMsg.setOperations = (operations) => {
            proto.operations = operations;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return toString('executionRequest', theMsg);
        };
        return theMsg;
    }

    function executionResult() {
        var EExecutionStatus_EXECUTED = 0;
        var ProtoMessage = imi.IMExecutionResult;

        var proto = new ProtoMessage({
            'request_id': 1,
            'execution_status': EExecutionStatus_EXECUTED
        });

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.EXECUTION_RESULT_ID;
        };
        theMsg.getRequestId = () => {
            return proto.request_id;
        };
        theMsg.getExecutionStatus = () => {
            return proto.execution_status;
        };
        theMsg.setRequestId = (id) => {
            proto.request_id = id;
        };
        theMsg.setExecutionStatus = (status) => {
            proto.execution_status = status;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return toString('executionResult', theMsg);
        };
        return theMsg;
    }

    function versionQuery() {
        var buffer = new ArrayBuffer(0);

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.VERSION_QUERY_ID;
        };
        theMsg.getBuffer = () => {
            return buffer;
        };
        theMsg.setBuffer = (value) => {
            buffer = value;
        };
        theMsg.toString = () => {
            return toString('versionQuery', theMsg);
        };
        return theMsg;
    }

    function versionQueryResult() {
        var ProtoMessage = imi.IMVersionQueryResult;

        var proto = new ProtoMessage({
            'supported_version': {
                'release': 'main',
                'major': 1,
                'minor': 1
            }
        });

        var theMsg = {};
        theMsg.getId = () => {
            return IMI_CONSTANTS.VERSION_QUERY_RESULT_ID;
        };
        theMsg.setSupportedRelease = (value) => {
            proto.supported_version.release = value;
        };
        theMsg.getSupportedRelease = () => {
            return proto.supported_version.release;
        };
        theMsg.setSupportedMajor = (value) => {
            proto.supported_version.major = value;
        };
        theMsg.getSupportedMajor = () => {
            return proto.supported_version.major;
        };
        theMsg.setSupportedMinor = (value) => {
            proto.supported_version.minor = value;
        };
        theMsg.getSupportedMinor = () => {
            return proto.supported_version.minor;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.toString = () => {
            return toString('versionQueryResult', theMsg);
        };
        return theMsg;
    }

    function objectInfo() {
        const ProtoMessage = imi.object_info_t;
        var proto = new ProtoMessage({});
        const theMsg = {};

        theMsg.hasObjectCreateInfo = () => {
            return !!proto.object_create_info;
        };
        theMsg.setObjectCreateInfo = (value) => {
            proto.object_create_info = value;
        };
        theMsg.getObjectCreateInfo = () => {
            return proto.object_create_info;
        };
        theMsg.getInstanceId = () => {
            const objectCreateInfo = theMsg.getObjectCreateInfo();
            return objectCreateInfo && objectCreateInfo.instance_id;
        };
        theMsg.setInstanceId = (value) => {
            if (typeof value === 'undefined') return;

            if (!theMsg.hasObjectCreateInfo())
                theMsg.setObjectCreateInfo({});
            theMsg.getObjectCreateInfo().instance_id = value;
        };
        theMsg.getParentId = () => {
            const objectCreateInfo = theMsg.getObjectCreateInfo();
            return objectCreateInfo && objectCreateInfo.parent_id;
        };
        theMsg.setParentId = (value) => {
            if (typeof value === 'undefined') return;

            if (!theMsg.hasObjectCreateInfo())
                theMsg.setObjectCreateInfo({});
            theMsg.getObjectCreateInfo().parent_id = value;
        };
        theMsg.setObjectId = (value) => {
            proto.object_id = value;
        };
        theMsg.getObjectId = () => {
            return proto.object_id;
        };
        theMsg.setObjectType = (value) => {
            proto.object_type = value;
        };
        theMsg.getObjectType = () => {
            return proto.object_type;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.getProto = () => {
            return proto;
        };
        theMsg.setProto = (value) => {
            proto = value;
        };
        theMsg.toString = () => {
            return toString('objectInfo', theMsg);
        };
        return theMsg;
    }

    function operation() {
        const ProtoMessage = imi.operation_t;
        var proto = new ProtoMessage({});
        const theMsg = {};

        theMsg.getOperationType = () => {
            return proto.operation_type;
        };
        theMsg.setOperationType = (value) => {
            proto.operation_type = value;
        };
        theMsg.getObject = () => {
            return proto.object;
        };
        theMsg.setObject = (value) => {
            proto.object = value;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.getProto = () => {
            return proto;
        };
        theMsg.setProto = (value) => {
            proto = value;
        };
        theMsg.toString = () => {
            return toString('operation', theMsg);
        };
        return theMsg;
    }

    function object() {
        const ProtoMessage = imi.object_t;
        var proto = new ProtoMessage({});
        const theMsg = {};
        theMsg.getObjectInfo = () => {
            return theMsg.object_info;
        };
        theMsg.setObjectInfo = (value) => {
            theMsg.object_info = value;
        };
        theMsg.getPayload = () => {
            return theMsg.payload;
        };
        theMsg.setPayload = (value) => {
            theMsg.payload = value;
        };
        theMsg.getBuffer = () => {
            return encodeProto(proto);
        };
        theMsg.setBuffer = (value) => {
            proto = decodeProto(ProtoMessage, value);
        };
        theMsg.getProto = () => {
            return proto;
        };
        theMsg.setProto = (value) => {
            proto = value;
        };
        theMsg.toString = () => {
            return toString('object', theMsg);
        };
        return theMsg;
    }

    return {
        getImiVersion,
        aaConfigGetServiceReq,
        aaConfigGetServiceRsp,
        aaSysComEcho,
        signIn,
        signOut,
        signInResult,
        updateNotification,
        changeRequest,
        operationExecuted,
        registerExecutor,
        executorRegistered,
        unregisterExecutor,
        executionRequest,
        executionResult,
        versionQuery,
        versionQueryResult,
        objectInfo,
        object,
        operation
    };
}

module.exports = {
    IMI_CONSTANTS: IMI_CONSTANTS,
    createMessages: createMessages,
    getAvailableVersions: getAvailableVersions
};
