const Gunzip = require('zlibjs/bin/gunzip.min').Zlib.Gunzip;
const loggers = require('./logger.impl');

const log = loggers.getLogger('admin.core.ims2');

module.exports = {
    readAsync: readAsync
};

var VERSION_OFFSET = 0;
var HEADER_LENGTH_OFFSET = 4;
var HEADER_OFFSET = 8;
var HEADER_INFO_LENGTH_OFFSET = 0;
var HEADER_INFO_OFFSET = 4;

var COMPRESSION = {
    GZIP: 1
};

var FRAME = {
    STARTING_BYTES: [0xE9, 0x11, 0x00, 0xA8, 0x43, 0xA0, 0x41, 0x2D, 0x94, 0xB3, 0x06, 0xDA],
    TYPE_OFFSET: 12,
    COMPRESSION_OFFSET: 16,
    PAYLOAD_LENGTH_OFFSET: 20,
    HEADER_LENGTH: 24
};

var MAX_PROCESSING_TIME_MS = 50;

var shouldCancelReading = false;
var notifyReadBytes;

/**
 * Reads async ims2 buffer and returns ims2 data structure in progressListener.onComplete() function.
 * @param {ArrayBuffer} buffer ims2 buffer
 * @param progressListener Object with methods {
         *       onComplete: function (ims2) {
         *       },
         *       onReadFrame: function (frame) {
         *       },
         *       onReset: function (reset) {
         *       },
         *       onUpdate: function (update) {
         *       },
         *       onMetaBuffer: function (metaBuffer) {
         *       },
         *       onReadBytes: function (procedureTitle, readCount, totalCount) {
         *       }
         *   };
 */
function readAsync(buffer, progressListener) {
    shouldCancelReading = false;

    var ims2 = {
        header: {},
        timeline: [],
        frames: [],
        metaPayload: null,
        progressListener: progressListener
    };
    notifyReadBytes = createReadBytesNotifier(ims2);

    var dataView = new DataView(buffer);
    notifyReadBytes(0, dataView.byteLength);

    setPreHeader(ims2, dataView);
    setHeader(ims2, dataView, HEADER_OFFSET);
    setFrames(ims2, dataView, HEADER_OFFSET + ims2.headerLength);

    return function cancel() {
        shouldCancelReading = true;
    }
}

function createReadBytesNotifier(ims2) {
    return function (loaded, total) {
        ims2.progressListener.onReadBytes('IMS2_FILE_READING', loaded, total);
    }
}

function setPreHeader(ims2, view) {
    ims2.version = view.getUint32(VERSION_OFFSET);
    ims2.headerLength = view.getUint32(HEADER_LENGTH_OFFSET);
    notifyReadBytes(8, view.byteLength);
}

function setHeader(ims2, view, offset) {
    ims2.header.infoLength = view.getUint32(offset + HEADER_INFO_LENGTH_OFFSET);
    ims2.header.info = decodeString(view, offset + HEADER_INFO_OFFSET, ims2.header.infoLength);
    notifyReadBytes(4 + ims2.header.infoLength, view.byteLength);
}

function setFrames(ims2, view, firstFrameOffset) {
    var frameOffset = firstFrameOffset;

    function process() {
        readFramePromise(ims2, view, frameOffset).then(function (frame) {
            if (shouldCancelReading) {
                ims2.progressListener.onComplete(ims2);
            } else if (frame) {
                frameOffset += frame.getLength();
                ims2.frames.push(frame);
                ims2.progressListener.onReadFrame(frame);
                process();
            }
        });
    }

    process();
}

function getUlong(dataView, offset) {
    var highOrderBytes = dataView.getUint32(offset);
    var lowOrderBytes = dataView.getUint32(offset + 4);
    return Math.pow(2, 32) * highOrderBytes + lowOrderBytes;
}

function decodeString(dataView, offset, strLength) {
    var TextDecoder = global['TextDecoder'];
    if (TextDecoder) {
        //Firefox and Chrome
        var unit8Array = new Uint8Array(dataView.buffer, offset, strLength);
        var decoder = new TextDecoder();
        return decoder.decode(unit8Array);
    } else {
        var array = new Array(strLength);
        for (var i = 0; i < strLength; i++) {
            var uint8 = dataView.getUint8(offset + i);
            array.push(String.fromCharCode(uint8));
        }
        return array.join('');
    }
}

function createViewStream(view, start, end) {
    var offset = start;
    var stream = {};

    stream.getUint32 = function () {
        var value = view.getUint32(offset);
        offset += 4;
        return value;
    };

    stream.getUlong = function () {
        var value = getUlong(view, offset);
        offset += 8;
        return value;
    };

    stream.getUint8 = function () {
        var value = view.getUint8(offset);
        offset += 1;
        return value;
    };

    stream.getUint16 = function () {
        var value = view.getUint16(offset);
        offset += 2;
        return value;
    };

    stream.getString = function (strLength) {
        var string = decodeString(view, offset, strLength);
        offset += strLength;
        return string;
    };

    stream.getBytes = function (numOfBytes) {
        var byteBuffer = new Uint8Array(view.buffer, offset, numOfBytes);
        offset += numOfBytes;
        return byteBuffer;
    };

    stream.hasMoreBytes = function () {
        return offset < end;
    };

    stream.offset = function () {
        return offset;
    };

    stream.skipBytes = function (numOfBytes) {
        offset += numOfBytes;
    };

    return stream;
}

function createExecutionRequest(distName, senderSicAddress, requestPayload, senderName, requestId) {
    return {
        distName: distName,
        senderSicAddress: senderSicAddress,
        requestPayload: requestPayload,
        senderName: senderName,
        requestId: requestId
    };
}

function createExecutionStatus(distName, senderSicAddress, requestStatus, senderName, requestId) {
    return {
        distName: distName,
        senderSicAddress: senderSicAddress,
        requestStatus: requestStatus,
        senderName: senderName,
        requestId: requestId
    };
}

var NOTIF_READER = {
    0: function readUpdateChange(distName, entry, stream) {
        var parametersPayloadSize = stream.getUint32();
        var managedObject = {distName: distName, payload: stream.getBytes(parametersPayloadSize)};

        if (!entry.change)
            entry.change = createChangeNotif();
        entry.change.updated.push(managedObject)
    },
    1: function readDeletedChange(distName, entry) {
        if (!entry.change)
            entry.change = createChangeNotif();
        entry.change.removed.push(distName);
    },
    2: function readExecutionRequestingBackwardCompatible(distName, entry, stream) {
        var senderSicAddress = stream.getUint32();
        var changeSize = stream.getUint32();
        var requestPayload = stream.getBytes(changeSize);

        entry.execution = {};
        entry.execution.request = createExecutionRequest(distName, senderSicAddress, requestPayload);
    },
    3: function readExecutionRequestingChange(distName, entry, stream) {
        var senderSicAddress = stream.getUint32();
        var senderNameLength = stream.getUint16();
        var senderName = stream.getString(senderNameLength);
        var requestId = stream.getUint32();
        var changeSize = stream.getUint32();
        var requestPayload = stream.getBytes(changeSize);

        entry.execution = {};
        entry.execution.request = createExecutionRequest(distName, senderSicAddress, requestPayload,
            senderName, requestId);
    },
    4: function readExecutionStatusChange(distName, entry, stream) {
        var senderSicAddress = stream.getUint32();
        var senderNameLength = stream.getUint16();
        var senderName = stream.getString(senderNameLength);

        var requestId = stream.getUint32();
        var statusIdNumber = stream.getUint32();

        entry.execution = {};
        entry.execution.status = createExecutionStatus(distName, senderSicAddress, statusIdNumber,
            senderName, requestId);
    }
};

function readTimelineEntryPromise(stream) {
    return new Promise(function (resolve) {
        var timestamp = stream.getUlong();
        var numOfChanges = stream.getUint32();
        var timelineEntry = createTimelineEntry(timestamp, numOfChanges);

        function process(currentChangeNr) {
            var startTimeMs = Date.now();
            for (var i = currentChangeNr; i < numOfChanges; i++) {
                var distNameLen = stream.getUint16();
                var distName = stream.getString(distNameLen);
                var notifType = stream.getUint8();

                var readNotif = NOTIF_READER[notifType];
                readNotif(distName, timelineEntry, stream);

                if (Date.now() - startTimeMs > MAX_PROCESSING_TIME_MS) {
                    setTimeout(function () {
                        process(++i);
                    }, 0);
                    return;
                }
            }
            var durationMs = Date.now() - startTimeMs;
            resolve([timelineEntry, durationMs]);
        }

        process(0);
    });
}

function frameReaderPromise(ims2, frame) {
    return {
        0: function readUpdatesPayload(view, offset, length) {
            return new Promise(function (resolve) {
                var stream = createViewStream(view, offset, offset + length);

                var elapsedTimeMs = 0;
                var elapsedTimeFactor = 4;

                function process() {
                    if (stream.hasMoreBytes()) {
                        if (shouldCancelReading) {
                            resolve(frame);
                            return;
                        }

                        readTimelineEntryPromise(stream).then(function (result) {
                            var timelineEntry = result[0];
                            var durationMs = result[1];

                            ims2.timeline.push(timelineEntry);
                            notifyReadBytes(stream.offset(), view.byteLength);
                            ims2.progressListener.onUpdate(timelineEntry);

                            elapsedTimeMs += durationMs;
                            if (elapsedTimeMs * elapsedTimeFactor > MAX_PROCESSING_TIME_MS) {
                                elapsedTimeMs = 0;
                                setTimeout(process, 0);
                            } else {
                                process();
                            }
                        });
                    } else {
                        resolve(frame);
                    }
                }

                process();
            });
        },
        1: function readReset(view, offset, length) {
            return new Promise(function (resolve) {
                var stream = createViewStream(view, offset, offset + length);
                var timestamp = stream.getUlong();
                var resetEntry = createTimelineEntry(timestamp, 0);
                resetEntry.reset = true;
                ims2.timeline.push(resetEntry);
                notifyReadBytes(stream.offset(), view.byteLength);
                ims2.progressListener.onReset(resetEntry);
                resolve(frame);
            });
        },
        2: function readMeta(view, offset, length) {
            return new Promise(function (resolve) {
                var metaStart = offset;
                var metaLength = length;
                ims2.metaPayload = new Uint8Array(view.buffer, metaStart, metaLength);
                notifyReadBytes(metaLength, view.byteLength);
                ims2.progressListener.onMetaBuffer(ims2.metaPayload);
                resolve(frame);
            });
        }
    };
}

function readFramePromise(ims2, view, offset) {
    return new Promise(function (resolve) {
        if (offset + FRAME.STARTING_BYTES.length > view.byteLength) {
            ims2.progressListener.onComplete(ims2);
            resolve();
            return;
        }

        var frameStartingBytes = new Uint8Array(view.buffer, offset, FRAME.STARTING_BYTES.length);
        for (var i = 0; i < FRAME.STARTING_BYTES.length; i++) {
            if (frameStartingBytes[i] !== FRAME.STARTING_BYTES[i]) {
                log.debug("Frame starting bytes not equal" + frameStartingBytes);
                resolve();
                return;
            }
        }

        var frame = createFrame();
        frame.header.frameType = view.getUint32(offset + FRAME.TYPE_OFFSET);
        frame.header.compression = view.getUint32(offset + FRAME.COMPRESSION_OFFSET);
        frame.header.payloadLength = view.getUint32(offset + FRAME.PAYLOAD_LENGTH_OFFSET);

        var frameView = view;
        var framePayloadOffset = offset + FRAME.HEADER_LENGTH;
        var framePayloadLength = frame.header.payloadLength;

        if (frame.header.compression === COMPRESSION.GZIP) {
            frameView = decompressGzip(view.buffer, framePayloadOffset, framePayloadLength);
            framePayloadOffset = 0;
            framePayloadLength = frameView.byteLength;
        }

        var readerPromise = frameReaderPromise(ims2, frame)[frame.header.frameType];
        readerPromise(frameView, framePayloadOffset, framePayloadLength).then(function () {
            resolve(frame);
        });
    });
}

function decompressGzip(arrayBuffer, offset, length) {
    var data = new Uint8Array(arrayBuffer, offset, length);
    var gunzip = new Gunzip(data);
    var decompressedData = gunzip.decompress();
    return new DataView(new Uint8Array(decompressedData).buffer);
}

function createTimelineEntry(timestamp, numOfChanges) {
    return {
        timestamp: timestamp,
        numOfChanges: numOfChanges
    };
}

function createChangeNotif() {
    return {
        updated: [],
        removed: []
    };
}

function createFrame() {
    return {
        header: {},

        getLength: function () {
            return FRAME.HEADER_LENGTH + this.header.payloadLength;
        }
    };
}
