const xz = require('./xzDecoder/xzDecoder');
const ims2 = require('./ims2');
const loggers = require('./logger.impl');
const unzipper = require('./unzipper');

var log = loggers.getLogger('admin.core.ims2FromZip');

module.exports = {
    readAsync: readAsync
};

var cancelReadingFn;

/**
 * Reads async ims2 buffer and returns ims2 data structure in progressListener.onComplete() function.
 * @param {ArrayBuffer} buffer ims2 buffer
 * @param progressListener Object with methods {
 *       onComplete: function (ims2) {
 *       },
 *       onReadFrame: function (frame) {
 *       },
 *       onReset: function (reset) {
 *       },
 *       onUpdate: function (update) {
 *       },
 *       onMetaBuffer: function (metaBuffer) {
 *       },
 *       onReadBytes: function (readCount, totalCount) {
 *       },
 *       onError: function () {
 *       }
 *   };
 */

function readAsync(buffer, progressListener) {
    cancelReadingFn = ()=> null;

    setTimeout(()=> {
        var instance = unzipper.newInstance(buffer);
        var fileNames = instance.getFileNames();
        var xzFileName = findXzFileName(fileNames);
        if (!xzFileName) {
            progressListener.onError('NO_IMS_XZ_FILE');
            return;
        }

        try {
            const fileContent = instance.unzipByFileName(xzFileName, {fileContentAsUint8: true});
            cancelReadingFn = xz.decode(fileContent.buffer, createXzProgressListener());
        } catch (error) {
            log.error(error);
            progressListener.onError('ZIP_READER_ERROR');
        }
    });

    return function cancel() {
        cancelReadingFn();
    };

    function findXzFileName(fileNames) {
        return fileNames.find((fileName)=> fileName.endsWith('.ims2.xz'));
    }

    function createXzProgressListener() {
        return {
            onComplete: function (decodedXz) {
                cancelReadingFn = ims2.readAsync(decodedXz, progressListener);
            },
            onReadBytes: function (loaded, total) {
                progressListener.onReadBytes('XZ_FILE_DECODING', loaded, total);
            }
        };
    }
}
