const _ = require('lodash');
const Long = require('long');
const distNameUtils = require('./utils/distNameUtils');
const LOG = loggers.getLogger('admin.core.ims2writer');

module.exports = function (protoDecoder, protoEncoder, metaService) {
    return {
        newInstance: function () {
            return newInstance(protoDecoder, protoEncoder, metaService);
        }
    };
};

let decoder;
let encoder;
let meta;

function newInstance(protoDecoder, protoEncoder, metaService) {
    decoder = protoDecoder;
    encoder = protoEncoder();
    meta = metaService;

    ims2.changes = [];
    ims2.frames = [];
    ims2.fileBuffer = null;

    return {
        addEntry: addEntry,
        addMeta: addMeta,
        addReset: addReset,
        saveBuffer: saveBuffer,
        //only for testing
        saveBufferFromObjects: saveBufferFromObjects,
        getBuffer: getBuffer,
        getHeader: getHeader,
        getFrames: getFrames
    };
}

var IMS2 = {
    INITIAL_SIZE: 8,
    VERSION_OFFSET: 0,
    VERSION: 1,
    ADDITIONAL_FILE_INFO: "AdditionalInfo"
};

var HEADER = {
    INITIAL_SIZE: 4,
    HEADER_SIZE_OFFSET: 4,
    ADDITIONAL_INFO_SIZE_OFFSET: 0,
    ADDITIONAL_INFO_OFFSET: 4
};

var FRAME = {
    STARTING_BYTES: [0xE9, 0x11, 0x00, 0xA8, 0x43, 0xA0, 0x41, 0x2D, 0x94, 0xB3, 0x06, 0xDA],
    INITIAL_SIZE: 8,
    MAGIC_STRING_OFFSET: 0,
    TYPE_OFFSET: 12,
    COMPRESSION_OFFSET: 16,
    PAYLOAD_LENGTH_OFFSET: 20,
    HEADER_LENGTH: 24,
    HEADER_OFFSET: 0,
    PAYLOAD_OFFSET: 24
};

var CHANGESET = {
    INITIAL_SIZE: 12,
    TIMESTAMP_LENGTH: 8,
    NUM_OF_CHANGES_OFFSET: 8,
    CHANGES_OFFSET: 12
};

var CHANGE = {
    NAME_SIZE_LENGTH: 2,
    NAME_OFFSET: 2,
    IS_REMOVED_LENGTH: 1,
    MANAGED_OBJECT_SIZE_LENGTH: 4
};

var CHANGE_TYPE = {
    UPDATE: 0,
    REMOVE: 1,
    EXECUTION_REQUEST_BACKWARD_COMPATIBLE: 2,
    EXECUTION_REQUEST: 3,
    EXECUTION_STATUS: 4
};

var FRAME_TYPE = {
    CHANGESET: 0,
    RESET: 1,
    META: 2
};

var COMPRESSION_TYPE = {
    NO_COMPRESSION: 0,
    GZIP: 1
};


var ims2 = {
    fileBuffer: null,
    header: createFileHeader(),
    frames: [],
    changes: []
};


function addEntry(entry) {
    ims2.changes.push(createChangeSet(entry));
}

function addMeta(meta) {
    checkChanges();
    ims2.frames.push({
        content: createMetaFrame(meta),
        type: FRAME_TYPE.META
    });
}

function addReset(reset) {
    checkChanges();
    ims2.frames.push({
        content: createResetFrame(reset),
        type: FRAME_TYPE.RESET
    });
}

function checkChanges() {
    if (ims2.changes.length)
        addChangeSetsFrame();
}

function addChangeSetsFrame() {
    ims2.frames.push({
        content: createChangeSetsFrame(ims2.changes),
        type: FRAME_TYPE.CHANGESET
    });
    ims2.changes = [];
}

function createChangeSetsFrame(changeSets) {
    var payload = mergeBuffers(changeSets);
    var frameHeaderBuilder = new FrameHeaderBuilder();
    frameHeaderBuilder.setType(FRAME_TYPE.CHANGESET);
    frameHeaderBuilder.setPayloadLength(payload.byteLength);
    frameHeaderBuilder.build();
    var frameHeader = frameHeaderBuilder.getBuffer();
    return createFrame(frameHeader, payload);
}

function createMetaFrame(meta) {
    var frameHeaderBuilder = new FrameHeaderBuilder();
    frameHeaderBuilder.setType(FRAME_TYPE.META);
    frameHeaderBuilder.setPayloadLength(meta.length);
    frameHeaderBuilder.build();
    var header = new Uint8Array(frameHeaderBuilder.getBuffer());
    return createFrame(header, meta);
}

function createResetFrame(entry) {
    var frameHeaderBuilder = new FrameHeaderBuilder(),
        buffer = new ArrayBuffer(CHANGESET.TIMESTAMP_LENGTH),
        stream = new StreamWriter(buffer);

    stream.setUint64(entry.timestamp);
    var payload = stream.getBuffer();
    frameHeaderBuilder.setType(FRAME_TYPE.RESET);
    frameHeaderBuilder.setPayloadLength(payload.byteLength);
    frameHeaderBuilder.build();
    var header = frameHeaderBuilder.getBuffer();
    return createFrame(header, payload);
}

function createFrame(header, payload) {
    var frameBuilder = new FrameBuilder();
    frameBuilder.setHeader(header);
    frameBuilder.setPayload(payload);
    frameBuilder.build();
    return frameBuilder.getBuffer();
}


function createFileHeader() {
    var fileHeaderBuilder = FileHeaderBuilder();
    fileHeaderBuilder.setAdditionalInfo(IMS2.ADDITIONAL_FILE_INFO);
    fileHeaderBuilder.build();
    return fileHeaderBuilder.getBuffer();
}

function createChangeSet(entry) {
    var payload = createChangeSetPayload(entry);
    var changeSetBuilder = new ChangeSetBuilder();
    changeSetBuilder.setTimestamp(entry.timestamp);
    changeSetBuilder.setSize(entry.numOfChanges);
    changeSetBuilder.setPayload(payload);
    changeSetBuilder.build();
    return changeSetBuilder.getBuffer();
}

function mergeFrames(frames) {
    var buffers = _.map(frames, function (n) {
        return n.content;
    });
    return mergeBuffers(buffers);
}

function createFrameListFromObjects(entry, meta) {
    var metaFrame = createMetaFrame(meta);
    var changeSet = createChangeSet(entry);
    var objectFrame = createChangeSetsFrame([changeSet]);
    return mergeBuffers([metaFrame, objectFrame])
}

function createFileBuffer() {
    var fileBuilder = new FileBuilder();
    fileBuilder.setHeader(ims2.header);
    fileBuilder.setFramesList(mergeFrames(ims2.frames));
    fileBuilder.build();
    return fileBuilder.getBuffer();
}

function saveBufferFromObjects(entry, meta, doneCallback) {
    var fileBuilder = new FileBuilder();
    fileBuilder.setHeader(ims2.header);
    fileBuilder.setFramesList(createFrameListFromObjects(entry, meta));
    fileBuilder.build();
    ims2.fileBuffer = fileBuilder.getBuffer();
    if (doneCallback)
        doneCallback(ims2.fileBuffer)
}


function ChangeSetPayload() {
    this.changeType = null;
}

ChangeSetPayload.prototype = {
    setStrategy: function (changeType) {
        this.changeType = changeType;
    },

    getBytes: function (entry) {
        return this.changeType.getBytes(entry);
    }
};

var UpdateChangeStrategy = function () {
    this.getBytes = function (entry) {
        var updatedAsBytes = [],
            updated = entry.change.updated;

        for (var i = 0, len = updated.length; i < len; i++)
            updatedAsBytes.push(changeToBytes(updated[i]))

        function changeToBytes(change) {
            var payload = change.payload,
                name = change.distName,
                className = distNameUtils.getClassName(change.distName);

            const typeInfo = meta.getTypeInfo(entry.metaUid, className);
            if (hasObjectId(typeInfo)) {
                const parameters = decoder.decode(typeInfo, payload);
                try {
                    payload = new Uint8Array(encoder.encode(typeInfo, parameters, true));
                } catch (e) {
                    LOG.error('Object ' + change.distName + ' will be invalid in ims2, due to the error: ' + e);
                }
            }

            const buffer = new ArrayBuffer(CHANGE.NAME_SIZE_LENGTH + name.length +
                    CHANGE.IS_REMOVED_LENGTH + CHANGE.MANAGED_OBJECT_SIZE_LENGTH +
                    payload.length),
                stream = new StreamWriter(buffer);

            addChangeHeader(stream, name, CHANGE_TYPE.UPDATE);
            stream.setUint32(payload.length);
            stream.setUint8Array(payload);

            return stream.getBuffer();
        }

        return mergeBuffers(updatedAsBytes);

        function hasObjectId(typeInfo) {
            return typeInfo.getParameters().some(function (parameterTypeInfo) {
                return parameterTypeInfo.isObjectId();
            });
        }
    }
};

var RemoveChangeStrategy = function () {
    this.getBytes = function (entry) {
        var removedAsBytes = [],
            removed = entry.change.removed;

        for (var i = 0, len = removed.length; i < len; i++)
            removedAsBytes.push(changeToBytes(removed[i]))

        function changeToBytes(distName) {
            var buffer = new ArrayBuffer(CHANGE.NAME_SIZE_LENGTH +
                    distName.length + CHANGE.IS_REMOVED_LENGTH),
                stream = new StreamWriter(buffer);

            addChangeHeader(stream, distName, CHANGE_TYPE.REMOVE);

            return stream.getBuffer();
        }

        return mergeBuffers(removedAsBytes);
    }
};

var ExecutionRequestChangeBackwardCompatibleStrategy = function () {
    this.getBytes = function (entry) {
        var request = entry.execution.request,
            name = request.distName,
            senderSicAddress = request.senderSicAddress,
            payload = request.requestPayload,
            payloadLength = payload.length,
            buffer = new ArrayBuffer(name.length + 11 + payloadLength),
            stream = new StreamWriter(buffer);

        addChangeHeader(stream, name, CHANGE_TYPE.EXECUTION_REQUEST_BACKWARD_COMPATIBLE);

        stream.setUint32(senderSicAddress);
        stream.setUint32(payloadLength);
        stream.setUint8Array(new Uint8Array(payload));

        return stream.getBuffer();
    }
};
var ExecutionRequestChangeStrategy = function () {
    this.getBytes = function (entry) {
        var request = entry.execution.request,
            name = request.distName,
            senderSicAddress = request.senderSicAddress,
            senderName = request.senderName,
            senderNameLength = senderName.length,
            requestId = request.requestId,
            payload = request.requestPayload,
            payloadLength = payload.length,
            buffer = new ArrayBuffer(name.length + 17 + senderNameLength + payloadLength),
            stream = new StreamWriter(buffer);

        addChangeHeader(stream, name, CHANGE_TYPE.EXECUTION_REQUEST);

        stream.setUint32(senderSicAddress);
        stream.setUint16(senderNameLength);
        stream.setString(senderName);
        stream.setUint32(requestId);
        stream.setUint32(payloadLength);
        stream.setUint8Array(payload);

        return stream.getBuffer();
    }
};
var ExecutionStatusChangeStrategy = function () {
    this.getBytes = function (entry) {
        var status = entry.execution.status,
            name = status.distName,
            senderSicAddress = status.senderSicAddress,
            senderName = status.senderName,
            senderNameLength = senderName.length,
            requestId = status.requestId,
            statusIdNumber = status.requestStatus,
            buffer = new ArrayBuffer(17 + senderNameLength),
            stream = new StreamWriter(buffer);

        addChangeHeader(stream, name, CHANGE_TYPE.EXECUTION_STATUS);

        stream.setUint32(senderSicAddress);
        stream.setUint16(senderNameLength);
        stream.setString(senderName);
        stream.setUint32(requestId);
        stream.setUint32(statusIdNumber);

        return stream.getBuffer();
    }
};

function addChangeHeader(stream, name, type) {
    stream.setUint16(name.length);
    stream.setString(name);
    stream.setUint8(type);
}

function createChangeSetPayload(entry) {

    var payload = new ChangeSetPayload(),
        change = entry.change,
        execution = entry.execution,
        changesAsBytes = [],
        updateChangeStrategy = new UpdateChangeStrategy(),
        removeChangeStrategy = new RemoveChangeStrategy(),
        executionStatusChangeStrategy = new ExecutionStatusChangeStrategy(),
        executionRequestChangeBackwardCompatibleStrategy = new ExecutionRequestChangeBackwardCompatibleStrategy(),
        executionRequestChangeStrategy = new ExecutionRequestChangeStrategy();

    if (change && change.updated.length > 0) {
        payload.setStrategy(updateChangeStrategy);
        changesAsBytes.push(payload.getBytes(entry));
    }

    if (change && change.removed.length > 0) {
        payload.setStrategy(removeChangeStrategy);
        changesAsBytes.push(payload.getBytes(entry));
    }

    if (execution && execution.status) {
        payload.setStrategy(executionStatusChangeStrategy);
        changesAsBytes.push(payload.getBytes(entry));
    }

    if (execution && execution.request) {
        if (typeof execution.request.requestId !== "undefined")
            payload.setStrategy(executionRequestChangeStrategy);
        else
            payload.setStrategy(executionRequestChangeBackwardCompatibleStrategy);
        changesAsBytes.push(payload.getBytes(entry));
    }

    return mergeBuffers(changesAsBytes);
}

function saveBuffer(doneCallback) {
    checkChanges();
    ims2.fileBuffer = createFileBuffer();
    if (doneCallback)
        doneCallback(ims2.fileBuffer);
}

function mergeBuffers(objects) {
    var byteLength = 0;
    for (var i = 0, len = objects.length; i < len; i++)
        byteLength += objects[i].byteLength

    var array = new Uint8Array(byteLength);

    var lastPosition = 0;
    for (var i = 0, len = objects.length; i < len; i++) {
        array.set(new Uint8Array(objects[i]), lastPosition)
        lastPosition += objects[i].byteLength;
    }
    return array.buffer
}

function FileBuilder() {

    var stream = null,
        theFile = {},
        header = {},
        framesList = {};

    theFile.setHeader = function (value) {
        header = value;
    };

    theFile.setFramesList = function (value) {
        framesList = value
    };

    theFile.build = function () {
        var buffer = new ArrayBuffer(IMS2.INITIAL_SIZE + header.byteLength + framesList.byteLength);
        stream = new StreamWriter(buffer);

        stream.setUint32(IMS2.VERSION);
        stream.setUint32(header.byteLength);
        stream.setUint8Array(new Uint8Array(header));
        stream.setUint8Array(new Uint8Array(framesList));
    };

    theFile.getBuffer = function () {
        return stream.getBuffer();
    };

    return theFile;
}

function FileHeaderBuilder() {
    var stream = null,
        theHeader = {},
        additionalInfo = {};

    theHeader.setAdditionalInfo = function (value) {
        additionalInfo = value;
    };

    theHeader.build = function () {
        var buffer = new ArrayBuffer(HEADER.INITIAL_SIZE + additionalInfo.length);
        stream = new StreamWriter(buffer);
        stream.setUint32(additionalInfo.length);
        stream.setString(additionalInfo);
    };

    theHeader.getBuffer = function () {
        return stream.getBuffer();
    };

    return theHeader;
}

function FrameBuilder() {
    var stream = null,
        theFrame = {},
        payload = {},
        header = {};

    theFrame.setPayload = function (value) {
        payload = value;
    };

    theFrame.setHeader = function (value) {
        header = value;
    };

    theFrame.build = function () {
        var buffer = new ArrayBuffer(header.byteLength + payload.byteLength);
        stream = new StreamWriter(buffer);
        stream.setUint8Array(new Uint8Array(header));
        stream.setUint8Array(new Uint8Array(payload));

    };

    theFrame.getBuffer = function () {
        return stream.getBuffer();
    };

    return theFrame;
}

function FrameHeaderBuilder() {
    var stream = null,
        theFrameHeader = {},
        type = {},
        compression = COMPRESSION_TYPE.NO_COMPRESSION,
        payloadLength = 0;

    theFrameHeader.setType = function (value) {
        type = value;
    };

    theFrameHeader.setCompression = function (value) {
        compression = value;
    };

    theFrameHeader.setPayloadLength = function (value) {
        payloadLength = value;
    };

    theFrameHeader.build = function () {
        var buffer = new ArrayBuffer(FRAME.HEADER_LENGTH);
        stream = new StreamWriter(buffer);
        stream.setUint8Array(FRAME.STARTING_BYTES);
        stream.setUint32(type);
        stream.setUint32(compression);
        stream.setUint32(payloadLength);
    };

    theFrameHeader.getBuffer = function () {
        return stream.getBuffer();
    };

    return theFrameHeader;
}


function ChangeSetBuilder() {
    var stream = null,
        theChangeSet = {},
        timestamp = {},
        numOfChanges = 0,
        changes = {};

    theChangeSet.setTimestamp = function (value) {
        timestamp = value;
    };

    theChangeSet.setSize = function (value) {
        numOfChanges = value;
    };

    theChangeSet.setPayload = function (value) {
        changes = value;
    };

    theChangeSet.build = function () {
        var length = CHANGESET.INITIAL_SIZE + changes.byteLength;
        var buffer = new ArrayBuffer(length);
        stream = new StreamWriter(buffer);
        stream.setUint64(timestamp);
        stream.setUint32(numOfChanges);
        stream.setUint8Array(new Uint8Array(changes));
    };

    theChangeSet.getBuffer = function () {
        return stream.getBuffer();
    };

    return theChangeSet;
}

function StreamWriter(buffer) {
    var stream = {},
        dataView = new DataView(buffer),
        typedArray = new Uint8Array(buffer),
        offset = 0;

    stream.setUint8 = function (value) {
        dataView.setUint8(offset, value);
        offset++;
    };

    stream.setUint16 = function (value) {
        dataView.setUint16(offset, value);
        offset += 2;
    };


    stream.setUint32 = function (value) {
        dataView.setUint32(offset, value);
        offset += 4
    };

    stream.setUint64 = function (value) {
        var long = Long.fromValue(value);
        var lowBits = long.getLowBits();
        var highBits = long.getHighBits();
        dataView.setUint32(offset, highBits);
        dataView.setUint32(offset + 4, lowBits);
        offset += 8;

    };

    stream.setUint8Array = function (array) {
        typedArray.set(array, offset);
        offset += array.length;
    };

    stream.setString = function (string) {
        var byteArray = stringToBytes(string);
        typedArray.set(byteArray, offset);
        offset += byteArray.length;
    };

    stream.getBuffer = function () {
        return buffer;
    };

    function stringToBytes(str) {
        var byteArray = new Uint8Array(str.length);
        for (var i = 0, len = str.length; i < len; i++)
            byteArray[i] = str.charCodeAt(i);
        return byteArray;
    }

    return stream;
}

function getHeader() {
    return ims2.header;
}

function getFrames() {
    return ims2.frames;
}

function getBuffer() {
    return ims2.fileBuffer;
}
