/**
 * Convention:
 *
 * findBy - deep search
 * findChildrenBy - search only direct descendants
 * findChildBy - search for first direct descendant
 *
 */
function ImTree(decoderService) {
    this.root = new Node("root");
    this.root.decoderService = decoderService;
    this.root.nodeMap = new Map();
    this.root.root = this.root;
}

ImTree.prototype.findByDistName = function (distName) {
    return findByDistName(this.root, distName);
};

ImTree.prototype.findByClassName = function (className) {
    return this.root.findByClassName(className);
};

ImTree.prototype.findAnyByClassName = function (className) {
    return this.root.findAnyByClassName(className);
};

ImTree.prototype.findChildrenByClassName = function (className) {
    return this.root.findChildrenByClassName(className);
};

ImTree.prototype.addNode = function (parentNode, className, id, metaUid) {
    parentNode = parentNode || this.root;
    return parentNode.addNode(className, id, metaUid);
};

ImTree.prototype.mergeNode = function (distName, parameters, payload, metaUid) {
    return mergeNode(this.root, distName, parameters, payload, metaUid);
};

ImTree.prototype.merge = function (update, metaUid) {
    return mergeUpdate(this.root, update, metaUid);
};

ImTree.prototype.size = function () {
    return this.root.size();
};

ImTree.prototype.isEmpty = function () {
    return this.root.nodeMap.size === 0;
};

ImTree.prototype.removeByDistName = function (distName) {
    return removeByDistName(this.root, distName);
};

ImTree.prototype.removeByDistNames = function (distNames) {
    return removeByDistNames(this.root, distNames);
};

ImTree.prototype.traverse = function (callback) {
    this.root.traverse(callback);
};

ImTree.prototype.distNames = function () {
    var distNames = new Set();
    this.root.nodeMap.forEach(function (node, distName) {
        distNames.add(distName);
    });
    return distNames;
};

function Bucket(className) {
    this.className = className;
    this.elements = new Map();
}

Bucket.prototype.add = function (node) {
    this.elements.set(node.id, node);
};

Bucket.prototype.getNodeById = function (id) {
    return this.elements.get(id);
};

Bucket.prototype.removeByNodeId = function (id) {
    this.elements.delete(id);
};

Bucket.prototype.toArray = function () {
    var objects = [];
    this.elements.forEach(function (node) {
        objects.push(node);
    });
    return objects;
};

Bucket.prototype.size = function () {
    return this.elements.size;
};

Bucket.prototype.getAny = function () {
    var elementsIterator = this.elements.values();
    var next = elementsIterator.next();
    return next && next.value;
};

function Node(className, id) {
    this.className = className;
    this.id = id;
    this.buckets = [];
    this.parent = null;
    this.root = null;
    this.payload = null;
    this.parameters = null;
    this.distName = null;
    this.metaUid = null;
}

Node.prototype.addNode = function (className, id, metaUid) {
    var node = new Node(className, parseInt(id));
    node.metaUid = metaUid;
    var parentNode = this;
    parentNode.add(node);
    var parentDistName = parentNode.distName || "";
    node.distName = [parentDistName, "/", node.className, "-", node.id].join('');
    node.parent = parentNode;
    node.root = parentNode.root;
    node.root.nodeMap.set(node.distName, node);
    return node;
};

Node.prototype.isMissing = function () {
    return this.payload === null && this.parameters === null;
};

Node.prototype.getDecodedParameters = function () {
    if (this.parameters) return this.parameters;

    if (this.isMissing()) {
        this.parameters = {};
    } else {
        this.parameters = this.root.decoderService.decodeByClassName(this.metaUid, this.payload, this.className);
    }
    return this.parameters;
};

Node.prototype.getTypeInfo = function () {
    if (this.typeInfo) return this.typeInfo;

    this.typeInfo = this.root.decoderService.getTypeInfo(this.metaUid, this.className);
    return this.typeInfo;
};

Node.prototype.size = function () {
    return size(this);
};

Node.prototype.getName = function () {
    return this.className + "-" + this.id;
};

Node.prototype.getBucket = function (className) {
    for (var i = 0, len = this.buckets.length; i < len; i++) {
        if (className === this.buckets[i].className)
            return this.buckets[i];
    }
};

Node.prototype.getOrCreateBucket = function (className) {
    var bucket = this.getBucket(className);
    if (!bucket) {
        bucket = new Bucket(className);
        this.buckets.push(bucket);
        this.buckets.sort(sortClassNames);
    }
    return bucket;
};

Node.prototype.add = function (childNode) {
    var bucket = this.getOrCreateBucket(childNode.className);
    bucket.add(childNode);
};

Node.prototype.remove = function (childNode) {
    var bucket = this.getBucket(childNode.className);
    if (bucket) {
        bucket.removeByNodeId(childNode.id);
    }
};

Node.prototype.findByDistName = function (distName) {
    return findByDistName(this, distName);
};

Node.prototype.findByClassName = function (className) {
    var matchedNodes = [];
    addNodesWithClassName(this, className, matchedNodes);
    return matchedNodes;
};

Node.prototype.findAnyByClassName = function (className) {
    var queue = [this];
    var index = 0;
    var nodeToCheck;
    var bucket;

    while (index < queue.length) {
        nodeToCheck = queue[index];
        index++;

        for (var i = 0, len = nodeToCheck.buckets.length; i < len; i++) {
            bucket = nodeToCheck.buckets[i];
            if (bucket.className === className && bucket.size())
                return bucket.getAny();

            bucket.elements.forEach(function (node) {
                queue.push(node);
            });
        }
    }
};

Node.prototype.findChildByClassName = function (className) {
    var bucket = this.getBucket(className);
    if (bucket) {
        return bucket.elements.values().next().value;
    }
};

Node.prototype.findChildByName = function (name) {
    var nameParts = splitName(name);
    return this.findChildByClassAndId(nameParts.className, nameParts.id);
};

Node.prototype.findChildByClassAndId = function (className, id) {
    var bucket = this.getBucket(className);
    if (bucket)
        return bucket.getNodeById(id);
};

Node.prototype.findChildrenByClassName = function (className) {
    var bucket = this.getBucket(className);
    if (bucket) {
        return bucket.toArray();
    }
    return [];
};

Node.prototype.getChildren = function () {
    var nodes = [];
    for (var i = 0; i < this.buckets.length; i++) {
        var bucket = this.buckets[i];
        bucket.elements.forEach(function (node) {
            nodes.push(node);
        });
    }
    return nodes;
};

Node.prototype.traverse = function (callback) {
    traverse(this, callback);
};

Node.prototype.forEachChild = function (callback) {
    for (var i = 0, len = this.buckets.length; i < len; i++) {
        this.buckets[i].elements.forEach(function (node) {
            callback(node);
        });
    }
};

function deepCopy(node, nodeCopy) {
    node.forEachChild(function (childNode) {
        var childNodeCopy = nodeCopy.addNode(childNode.className, childNode.id, childNode.metaUid);
        childNodeCopy.payload = childNode.payload;
        deepCopy(childNode, childNodeCopy);
    });
}

Node.prototype.toImTree = function () {
    var imTree = new ImTree(this.root.decoderService);
    var node = this;
    var nodeCopy = imTree.root.addNode(node.className, node.id, node.metaUid);
    nodeCopy.payload = node.payload;
    deepCopy(node, nodeCopy);
    return imTree;
};

function sortClassNames(a, b) {
    return ((a.className < b.className) ? -1 : ((a.className > b.className) ? 1 : 0))
}

function addNodesWithClassName(node, className, matchedNodes) {
    var bucket, elements;

    if (node.className === className) matchedNodes.push(node);

    for (var i = 0, len = node.buckets.length; i < len; i++) {
        bucket = node.buckets[i];
        elements = bucket.elements;
        if (bucket.className !== className) {
            elements.forEach(function (node) {
                addNodesWithClassName(node, className, matchedNodes);
            });
        } else {
            elements.forEach(function (node) {
                matchedNodes.push(node);
            });
        }
    }

}

function splitName(nodeName) {
    var splitOnIndex = nodeName.indexOf('-');
    return {
        className: nodeName.substring(0, splitOnIndex),
        id: parseInt(nodeName.substring(splitOnIndex + 1))
    }
}

function findOrCreateChildNode(parentNode, childName, metaUid) {
    var nameParts = splitName(childName);

    var childNode = parentNode.findChildByClassAndId(nameParts.className, nameParts.id);
    if (!childNode) {
        childNode = parentNode.addNode(nameParts.className, nameParts.id, metaUid);
    }
    return childNode;
}

function mergeUpdate(tree, update, metaUid) {
    return mergeNode(tree, update.distName, null, update.payload, metaUid);
}

function mergeNode(tree, distName, parameters, payload, metaUid) {
    var node = tree.root.nodeMap.get(distName);
    if (!node) {
        var parts = distName.split('/');
        node = tree;
        for (var i = 1; i < parts.length; i++) {
            node = findOrCreateChildNode(node, parts[i], metaUid);
        }
    }

    node.payload = payload;
    node.parameters = parameters;
    node.metaUid = metaUid;
    return node;
}

function removeByDistName(tree, distName) {
    var node = tree.root.nodeMap.get(distName);
    if (node) {
        var parentNode = node.parent;
        if (parentNode) {
            parentNode.remove(node);
            tree.root.nodeMap.delete(distName);
            node.parent = null;
            node.root = null;
            var children = node.getChildren();
            for (var i = 0; i < children.length; i++) {
                removeByDistName(tree, children[i].distName);
            }
        }
    }
}

function removeByDistNames(tree, removedDistNames) {
    for (var i = 0; i < removedDistNames.length; i++) {
        removeByDistName(tree, removedDistNames[i]);
    }
}

function findByDistName(tree, distName) {
    if (!distName || distName.indexOf("/") !== 0) {
        return;
    }

    var parts = distName.split("/");
    var node = tree;
    for (var i = 1, len = parts.length; i < len; i++) {
        var part = parts[i];
        node = node.findChildByName(part);
        if (!node) {
            return;
        }
    }
    return node;
}

function traverse(node, callback) {
    node.forEachChild(function (child) {
        callback(child);
        traverse(child, callback);
    });
}

function size(node) {
    var numberOfNodes = 0;
    traverse(node, function () {
        numberOfNodes++;
    });
    return numberOfNodes;
}

module.exports = ImTree;
