'use strict';

const log4javascript = require('./log4javascript');
const PouchDB = require('pouchdb-browser');
const _ = require('lodash');

var impl = log4javascript;
var logStorage;

var session = Date.now();

/**
 * @param event log4javascript logging event
 * @return {Object} a record to be stored in storage
 * */
function eventRecord(event) {

    function errorAsObject(error) {
        if (error) return {
            name: error.name,
            message: error.message,
            stack: error.stack
        };
    }

    return {
        session: session,
        logger: event.logger.name,
        timestamp: event.timeStampInMilliseconds,
        level: event.level.name,
        messages: event.messages.map(function (message) {
            if (message === null || message === undefined) return '' + message;
            return message.toString();
        }),
        exception: errorAsObject(event.exception)
    };
}

function openLogDatabase(name) {
    return new PouchDB(name, {
        adapter: 'idb',
        auto_compaction: true
    });
}

function createStorage(name) {
    var dbName = name || 'logs';
    var db = openLogDatabase(dbName);
    var dbReady = Promise.resolve();
    var sessionStart = new Date().valueOf();

    var storeCallbacks = [];
    var seqNum = 0;

    /**
     * visible only for testing
     * @returns {Number} the session start timestamp
     */
    function getSessionStart() {
        return sessionStart;
    }

    function id(record) {
        seqNum++;
        return record.timestamp.toString() + '-' + seqNum.toString();
    }

    function prepare(response) {
        return response.rows.map(function (row) {
            var record = row.doc;
            delete record._id;
            delete record._rev;
            return record;
        });
    }

    /**
     * Stores a record using timestamp as a key, with a sequence number appended to avoid collisions.
     * @return {Promise}
     * */
    function store(record) {
        return dbReady.then(function () {
            const stored = db.put(_.merge({}, record, { _id: id(record) }));
            stored.then(function () {
                storeCallbacks.forEach(function (callback) {
                    callback([record]);
                });
            });
            return stored;
        });
    }

    /**
     * @return {Promise} promise that results in record array
     * */
    function all() {
        return db.allDocs({
            include_docs: true
        }).then(function (response) {
            return prepare(response);
        });
    }

    /**
     * @param fromTimestamp {Number} milliseconds, inclusive
     * @param toTimestamp {Number} milliseconds, inclusive
     * @return {Promise} promise that results in record array
     * */
    function byTimestamp(fromTimestamp, toTimestamp) {
        return db.allDocs({
            startkey: fromTimestamp.toString(),
            endkey: (toTimestamp + 1).toString(),
            inclusive_end: false,
            include_docs: true
        }).then(function (response) {
            return prepare(response);
        });
    }

    /**
     * visible only for testing
     * @return {Promise}
     * */
    function currentSession() {
        return byTimestamp(sessionStart, new Date().valueOf());
    }

    /**
     * Callback is first called with all records stored in current session
     * @param {Function} stored - a callback to handle an array of stored records
     */
    function onStore(stored) {
        dbReady.then(currentSession).then(function (records) {
            storeCallbacks.push(stored);

            if (records.length > 0) stored(records);
        });
    }

    /**
     * Returns Blob as promise result.
     * @param fromTimestamp {Number} milliseconds, inclusive
     * @param toTimestamp {Number} milliseconds, inclusive
     * @return {Promise}
     */
    function download(fromTimestamp, toTimestamp) {
        return byTimestamp(fromTimestamp, toTimestamp)
            .then(function (records) {
                var log = JSON.stringify(records, function (key, value) {
                    if (key === 'timestamp')
                        return new Date(value).toISOString();
                    else
                        return value;
                }, '\t');
                return new Blob([log], {
                    type: 'application/json'
                });
            });
    }

    /**
     * @returns {Promise}
     */
    function deleteAll() {
        dbReady = dbReady
            .then(function () {
                return db.destroy();
            })
            .then(function () {
                db = openLogDatabase(dbName);
            });
        return dbReady;
    }

    return {
        store: store,
        fetchAll: all,
        fetchByTimestamp: byTimestamp,
        fetchCurrentSession: currentSession,
        onStore: onStore,
        download: download,
        getSessionStart: getSessionStart,
        deleteAll: deleteAll
    };
}

/**
 * @param dbName {String=} database name, only for testing
 * @return a storage created on given database, or a singleton log storage
 * */
function getLogStorage(dbName) {
    if (dbName) return createStorage(dbName);
    return logStorage || (logStorage = createStorage());
}

/**
 * @param logStorage {Object=} only for testing
 * */
function newDbAppender(logStorage) {
    var storage = logStorage || getLogStorage();

    function DbAppender() {}

    DbAppender.prototype = new impl.Appender();
    DbAppender.prototype.append = function (loggingEvent) {
        storage.store(eventRecord(loggingEvent))
            .catch(function (e) {
                console.error(e);
            });
    };
    return new DbAppender();
}

/**
 * @param {String} pattern
 * @returns {log4javascript.BrowserConsoleAppender}
 */
function newConsoleAppender() {
    var appender = new log4javascript.BrowserConsoleAppender();
    appender.setLayout(new log4javascript.PatternLayout("%d{HH:mm:ss} %-5p %c - %m%n"));
    return appender;
}

function enableLogUncaughtErrors(window) {
    window.addEventListener('error', function (evt) {
        impl.getLogger('UNCAUGHT').error('Application error: ' + evt.error.message, evt.error);
    });
}

function configure(config) {
    if (config.dbName) {
        logStorage = getLogStorage(config.dbName);
        impl.getRootLogger().addAppender(newDbAppender(logStorage));
    }
    impl.getRootLogger().setLevel(config.level);
}

impl.getRootLogger().addAppender(newConsoleAppender());
impl.getRootLogger().setLevel(impl.Level.ALL);

module.exports = {
    getLogger: impl.getLogger,
    getRootLogger: impl.getRootLogger,
    Level: impl.Level,
    getLogStorage: getLogStorage,
    newDbAppender: newDbAppender,
    newConsoleAppender: newConsoleAppender,
    configure: configure,
    enableLogUncaughtErrors: enableLogUncaughtErrors
};
