const logger = require('../../logger.impl');
const xz = require('../../xzDecoder/xzDecoder');
const xml2json = require('../../xml2json');
const log = logger.getLogger('admin.core.pmSnapshot');
const converter = require('../../bytesToStringConverter');

module.exports = pmSnapshot;

function pmSnapshot(fileTransfer = {}) {

    const NE_WBTS_KEY = 'NE-WBTS_1.0';
    const MEASUREMENT_TYPE_KEY = '@measurementType';
    return {
        downloadCompressed,
        convertXMLtoJson,
        formatPmCounters
    };

    function convertXMLtoJson(fileData) {
        return xml2json.toJson(fileData);
    }

    function downloadCompressed(pmLocations) {
        log.info('Downloading compressed pmCounters file from BTS.');

        return startDownloading();

        function startDownloading() {
            return fileTransfer.connect()
                .then(downloadPmFiles)
                .catch(handleErrors)
                .finally(fileTransfer.disconnect);
        }

        function downloadPmFiles() {
            let promises = [];

            pmLocations.forEach((pmLocation) => {
                const promise = getFile(pmLocation)
                    .then(decompressPmFile)
                    .then(byteToString)
                    .then(convertXMLtoJson)
                    .then(getPmCounters);

                promises.push(promise);
            });
            return Promise.all(promises)
                .then((values) => {
                    return [].concat(...values);
                });

            function getFile(pmLocation) {
                return fileTransfer.getFile(pmLocation);
            }
        }

        function getPmCounters(jsonArray) {
            return jsonArray.OMeS.PMSetup;
        }

        function decompressPmFile(fileData) {
            return new Promise((resolve) => {
                xz.decode(fileData.buffer, createXzProgressListener());

                function createXzProgressListener() {
                    return {
                        onComplete: function (decodedXz) {
                            resolve(decodedXz);
                        },
                        onReadBytes: function () {
                        },
                        onError: function (error) {
                            log.info(error);
                        }
                    };
                }
            });
        }

        function byteToString(buffer) {
            return converter.convertBytesToString(new DataView(buffer), 0, buffer.byteLength);
        }

        function handleErrors(error) {
            log.error('File download failed!', error);
        }
    }

    function formatPmCounters(response, existingPmCounters) {
        var values = existingPmCounters.values;
        var startTimes = existingPmCounters.startTimes;
        response.forEach((pm) => {
            const startTime = pm['@startTime'];
            startTimes.push(startTime);
            if (Array.isArray(pm.PMMOResult)) {
                pm.PMMOResult.forEach((pmmoResult) => {
                    const measurementType = pmmoResult[NE_WBTS_KEY][MEASUREMENT_TYPE_KEY];
                    values[measurementType] = getCounters(pmmoResult[NE_WBTS_KEY], startTime, values[measurementType]);
                });
            } else {
                const measurementType = pm.PMMOResult[NE_WBTS_KEY][MEASUREMENT_TYPE_KEY];
                values[measurementType] = getCounters(pm.PMMOResult[NE_WBTS_KEY], startTime, values[measurementType]);
            }
        });
        startTimes = getUniqueArrayValues(startTimes);
        startTimes.sort();
        return existingPmCounters;
    }

    function getCounters(pmCountersMeasurement, startTime, counters = {}) {
        for (let key in pmCountersMeasurement) {
            if (key === MEASUREMENT_TYPE_KEY || key === '#text')
                continue;
            if (typeof counters[key] === 'undefined')
                counters[key] = {};
            counters[key][startTime] = pmCountersMeasurement[key];
        }
        return counters;
    }

    function getUniqueArrayValues(arrayToCheck) {
        var uniqueArrayValues = [];
        arrayToCheck.forEach(function (value) {
            if (uniqueArrayValues.indexOf(value) < -1)
                uniqueArrayValues.push(value);
        });
        return uniqueArrayValues;
    }
}
