const SparkMD5 = require("spark-md5");
const typeInfoV1Constructor = require("./typeInfo");
const typeInfoV2Constructor = require("./typeInfoV2");
const metaVersions = require("./metaReader").versions;

module.exports = function () {
    var repository = new Map();

    function isEmpty() {
        return repository.size === 0;
    }

    function setMetaDefinitions(definitions, metaBuffer) {
        var metaUid = hash(definitions);
        var metaDetails = {definitions: definitions, typeInfos: {}, metaBuffer: metaBuffer};
        repository.set(metaUid, metaDetails);
        return metaUid;
    }

    function getMetaDetails(metaUid) {
        validateUid(metaUid);
        return repository.get(metaUid);
    }

    function getMetaDefinition(metaUid, className) {
        const metaDetails = getMetaDetails(metaUid);
        if (!metaDetails) return;

        const definitions = getDefinitionsFromMeta(metaDetails, className);
        if (!definitions) return;

        return definitions.objectDefinitions[className];
    }

    function getDefinitionsFromMeta(metaDetails, className) {
        for (var i = 0; i < metaDetails.definitions.length; i++) {
            var object = metaDetails.definitions[i].objectDefinitions[className];
            if (object) return metaDetails.definitions[i];
        }
    }

    function getOrCreateTypeInfo(metaUid, className) {
        const metaDetails = getMetaDetails(metaUid);
        var typeInfo = metaDetails.typeInfos[className];
        if (typeInfo) return typeInfo;

        const definitions = getDefinitionsFromMeta(metaDetails, className);
        if (!definitions) return;

        const objectDefinition = definitions.objectDefinitions[className];
        typeInfo = createTypeInfo(objectDefinition, definitions.metaVersion);
        metaDetails.typeInfos[className] = typeInfo;
        return typeInfo;
    }

    function createTypeInfo(objectDefinition, metaVersion) {
        switch (metaVersion) {
            case metaVersions.V2:
                return typeInfoV2Constructor(objectDefinition);
            default:
                return typeInfoV1Constructor(objectDefinition);
        }
    }

    function getMetaBufferForUid(metaUid) {
        return getMetaDetails(metaUid).metaBuffer
    }

    function getClassNamesForUid(metaUid) {
        var metaDefinitions = getMetaDetails(metaUid).definitions;

        return metaDefinitions.map((metaDefinition) => {
            return metaDefinition.objectDefinitions
        }).reduce((classNames, objectDefinitions) => {
            Object.keys(objectDefinitions).forEach((className) => classNames.push(className));
            return classNames;
        }, []);
    }

    function validateUid(metaUid) {
        if (!repository.get(metaUid)) throw new Error("Invalid meta uid!: " + metaUid);
    }

    function hash(object) {
        return SparkMD5.hash(JSON.stringify(object));
    }

    return {
        setMetaDefinitions: setMetaDefinitions,
        getMetaDefinition: getMetaDefinition,
        getTypeInfo: getOrCreateTypeInfo,
        getMetaBufferForUid: getMetaBufferForUid,
        getClassNamesForUid: getClassNamesForUid,
        isEmpty: isEmpty
    };
};

