const _ = require('lodash'),
    log = require('./logger.impl').getLogger('admin.core.metaService'),
    META_STATE = {
        NOT_SET: {},
        SET_IN_PROGRESS: {},
        SET_COMPLETE: {}
    },
    MetaRepository = require('./metaRepository'),
    metaDefinitionFactory = require('./metaDefinitionFactory');

module.exports = metaService;

function metaService({metaReader = require('./metaReader'), metaRepository = MetaRepository } = {}) {

    var state = META_STATE.NOT_SET;
    var metaUid = null;
    var metaReadCounter = 0;
    var repository = metaRepository();

    return {
        setMeta: setMeta,
        addMeta: addMeta,
        isReady: isReady,
        getMetaObject: getMetaObject,
        getTypeInfo: getTypeInfo,
        getLastMetaUid: getLastMetaUid,
        getMetaBufferForUid: getMetaBufferForUid,
        started: hasStarted,
        clear: clear,
        getMetaRepository: getMetaRepository,
        //visible for testing
        setMetaReader: setMetaReader
    };

    function clear() {
        state = META_STATE.NOT_SET;
        repository = metaRepository();
        metaUid = null;
        metaReadCounter = 0;
    }

    function setMeta(metaBuffer) {
        state = META_STATE.SET_IN_PROGRESS;
        return new Promise(function (resolve) {
            readMetaBuffer(metaBuffer, resolve);
        });

        function readMetaBuffer(metaBuffer, resolve) {
            metaReadCounter++;
            metaReader.read(metaBuffer, function (metaEntries) {
                var metaDefinitions = metaDefinitionFactory.create(metaEntries);
                metaUid = repository.setMetaDefinitions(metaDefinitions, metaBuffer);
                metaReadCounter--;
                state = META_STATE.SET_COMPLETE;
                log.debug('meta set #' + metaUid);
                resolve(metaUid);
            });
        }
    }

    function addMeta(metaBuffer) {
        return new Promise(function (resolve) {
            readMetaBuffer(metaBuffer, resolve);
        });

        function readMetaBuffer(metaBuffer, resolve) {
            metaReader.read(metaBuffer, function (metaEntries) {
                var metaDefinitions = metaDefinitionFactory.create(metaEntries);
                var createdMetaUid = repository.setMetaDefinitions(metaDefinitions, metaBuffer);
                log.debug('meta added #' + createdMetaUid);
                resolve(createdMetaUid);
            });
        }
    }

    function isReady() {
        return state === META_STATE.SET_COMPLETE && metaReadCounter === 0;
    }

    function hasStarted() {
        return state === META_STATE.SET_IN_PROGRESS;
    }

    function getMetaObject(metaUid, className) {
        return repository.getMetaDefinition(metaUid, className);
    }

    function getTypeInfo(metaUid, className) {
        return repository.getTypeInfo(metaUid, className);
    }

    function getLastMetaUid() {
        if (!metaUid) throw Error('Meta was not set!');

        return metaUid;
    }

    function getMetaBufferForUid(metaUid) {
        return repository.getMetaBufferForUid(metaUid);
    }

    function getMetaRepository() {
        return repository;
    }

    function setMetaReader(newMetaReader) {
        metaReader = newMetaReader;
    }

}
