const _ = require('lodash'),
    LOG = require('../logger.impl').getLogger('admin.core.onlineHistory'),
    sequentialExecutorFactory = require('../sequentialExecutorFactory'),
    arrays = require('../arrays');

module.exports = onlineHistoryService;

function onlineHistoryService(updateService) {

    var lastMetaUid;
    var onlineHistoryBuffer = [];
    var timestampOffset = 0;
    var entriesAfterHistory;
    var procedureCompletedTimestamp;
    var onlineHistoryBlocks = [];

    const executor = sequentialExecutorFactory.newInstance();
    const TIME_TO_PROCESS = 50;

    return {
        setMeta: executor.promisifyFunction(this, setMeta),
        addEntry: executor.promisifyFunction(this, addEntry),
        apply: executor.promisifyFunction(this, apply),
        getHistoryBlocks: getHistoryBlocks,
        isHistoryFromBts: isHistoryFromBts,
        clear: clear,
        // visible for testing
        getOnlineHistoryBuffer: getOnlineHistoryBuffer,
        getLastMetaUid: getLastMetaUid,
        getTimestampOffset: getTimestampOffset,
        setTimestampOffset: setTimestampOffset
    };

    function isHistoryFromBts(entryId) {
        for (var i = 0; i < onlineHistoryBlocks.length; i++) {
            var block = onlineHistoryBlocks[i];
            if (entryId > block.end) return false;
            if (entryId >= block.start) return true;
        }
        return false;
    }

    function getHistoryBlocks() {
        return onlineHistoryBlocks;
    }

    function setMeta(metaBuffer) {
        return updateService.addMeta(metaBuffer)
            .then(function (createdMetaUid) {
                lastMetaUid = createdMetaUid;
            });
    }

    function addEntry(entry) {
        entry.metaUid = lastMetaUid;
        onlineHistoryBuffer.push(entry);
    }

    function apply(procedureCompletedId, procedureSentId) {
        updateService.performInAsyncQueue(function () {
            return new Promise(function (resolve) {
                const historyStartId = findHistoryStartId(procedureSentId);
                procedureCompletedTimestamp = updateService.SYNC_API.getHistoryEntry(procedureCompletedId).timestamp;
                dropEntriesReplacedByHistory(historyStartId, procedureSentId);
                mergeEntriesUntilProcedureWasCompleted(historyStartId, procedureCompletedId - procedureSentId);
                stashEntriesAfterHistory(historyStartId);
                updateService.SYNC_API.clearUpdateServiceState(true);
                onlineHistoryBlocks.push({start: historyStartId});
                applyHistoryFromBuffer(onlineHistoryBuffer, resolve);
            });

        });
    }

    function dropEntriesReplacedByHistory(historyStartId, procedureSentId) {
        updateService.SYNC_API.dropEntriesRange(historyStartId, procedureSentId);
    }

    function mergeEntriesUntilProcedureWasCompleted(historyStartId, numberOfEntries) {
        updateService.SYNC_API.mergeEntriesRange(historyStartId, historyStartId + numberOfEntries + 1, false);
    }

    function stashEntriesAfterHistory(historyStartId) {
        entriesAfterHistory = updateService.SYNC_API.getHistoryEntries().slice(historyStartId);
        updateService.SYNC_API.dropEntriesRange(historyStartId, historyStartId + entriesAfterHistory.length);
    }

    function findHistoryStartId(procedureSentId) {
        for (var i = procedureSentId; i >= 0; i--) {
            if (updateService.SYNC_API.getHistoryEntry(i).reset) {
                return i + 1;
            }
        }
        return 0;
    }

    function applyHistoryFromBuffer(buffer, resolve) {
        arrays.process(buffer, TIME_TO_PROCESS, function (entry) {
            updateService.SYNC_API.addOnlineEntry(entry);
        }, nextAction);

        function nextAction() {
            unstashEntries(entriesAfterHistory, resolve);
        }
    }

    function unstashEntries(stashedEntries, resolve) {
        var lastEntry = updateService.SYNC_API.getLastHistoryEntry();
        _.last(onlineHistoryBlocks).end = lastEntry.id;
        var timestampOffset = calculateTimestampOffset(lastEntry, procedureCompletedTimestamp);
        updateService.SYNC_API.setTimestampOffset(timestampOffset);
        arrays.process(stashedEntries, TIME_TO_PROCESS, function (entry) {
            entry.timestamp += timestampOffset;
            updateService.SYNC_API.addOnlineEntry(entry);
        }, nextAction);

        function nextAction() {
            notifyListeners(resolve);
        }
    }

    function notifyListeners(resolve) {
        updateService.SYNC_API.createNotifier()
            .imTreeEvents()
            .historyNewEntries()
            .currentEntryChange(updateService.SYNC_API.getLastHistoryEntry())
            .notify();
        resolve();
    }

    function calculateTimestampOffset(lastEntry, oldTimestamp) {
        const TIMESTAMP_DELTA_MS = 100;
        return lastEntry.timestamp - oldTimestamp + TIMESTAMP_DELTA_MS;
    }

    function clear() {
        onlineHistoryBuffer = [];
        onlineHistoryBlocks = [];
        timestampOffset = 0;
        updateService.SYNC_API.setTimestampOffset(timestampOffset);
    }

    function getOnlineHistoryBuffer() {
        return onlineHistoryBuffer;
    }

    function getLastMetaUid() {
        return lastMetaUid;
    }

    function getTimestampOffset() {
        return timestampOffset;
    }

    function setTimestampOffset(newOffset) {
        timestampOffset = newOffset;
    }
}

