const events = require('./events'),
    logger = require('./logger.impl'),
    ImTree = require('./imtree'),
    sequentialExecutorFactory = require('./sequentialExecutorFactory'),
    translatingBtsClient = require('./communication/translatingBtsClient'),
    MetaService = require('./metaService'),
    entryTimestampProvider = require('./entryTimestampProvider')();

const log = logger.getLogger('admin.core.onlineImTree');
module.exports = onlineImTree;

function onlineImTree({
    socket = translatingBtsClient.newInstance(),
    metaService = MetaService(),
    decoderService = require('./decoderFactory').newInstance(metaService),
    executor = sequentialExecutorFactory.newInstance()
    } = {}) {

    const updateEvent = events.newEvent();
    var imTree = new ImTree(decoderService);
    var lastMetaUid;

    socket.addListener({
        disconnected: executor.promisifyFunction(this, handleSocketDisconnect),
        data: executor.promisifyFunction(this, handleSocketData)
    });

    return {
        watchUpdates: executor.promisifyFunction(this, watchUpdates),
        getImTree: getImTree,

        //only for testing
        pushData: handleSocketData,
        pushUpdate: pushUpdate,
        pushRemove: pushRemove,
        pushDisconnect: handleSocketDisconnect
    };

    function getImTree() {
        return new Promise(function (resolve) {
            var imTree;
            var unwatcher;
            unwatcher = watchUpdates(function (entry) {
                imTree = entry.completeImTree;
                if (unwatcher)
                    unwatcher.unwatch();
                resolve(imTree);
            });
            if (imTree)
                unwatcher.unwatch();
        });
    }

    function watchUpdates(listener) {
        const securedListener = exceptionCatcher(listener);
        const updateUnwatcher = updateEvent.watch(securedListener);
        if (!imTree.isEmpty()) {
            const updateEntry = createUpdateEntry(lastMetaUid);
            const updates = filterExistingNodes();
            updateEntry.change.updated = updates;
            updateEntry.numOfChanges = updates.length;
            securedListener(updateEntry);
        }
        return updateUnwatcher;

        function exceptionCatcher(originalFunction) {
            return function () {
                try {
                    originalFunction(arguments[0]);
                } catch (error) {
                    log.error('Exception has been thrown by onlineImTree\'s listener.');
                    log.error(error);
                }
            };
        }

        function filterExistingNodes() {
            const existingNodes = [];
            imTree.traverse(function (node) {
                if (!node.isMissing())
                    existingNodes.push(node.distName);
            });
            return existingNodes;
        }
    }

    function createUpdateEntry(metaUid, updates = [], removed = []) {
        updateImTree();
        const updatedDistNames = [];
        updates.forEach((entry) => updatedDistNames.push(entry.distName));
        return {
            timestamp: entryTimestampProvider.getTimestamp(),
            numOfChanges: updates.length + removed.length,
            change: {updated: updatedDistNames, removed: removed},
            metaUid: metaUid,
            completeImTree: imTree
        };

        function updateImTree() {
            imTree.removeByDistNames(removed);
            for (var i = 0; i < updates.length; i++) {
                imTree.merge(updates[i], metaUid);
            }
        }
    }

    /**
     * data = { managedObjects: [managedObject], removedDistNames: [String] }
     * managedObject = { distName: String, payload: ArrayBuffer }
     * */
    function handleSocketData(data) {
        const updates = data.managedObjects;
        const removed = data.removedDistNames;
        lastMetaUid = metaService.getLastMetaUid();
        const entry = createUpdateEntry(lastMetaUid, updates, removed);
        updateEvent.notify(entry);
        return entry;
    }

    function handleSocketDisconnect(event) {
        if (!isResetFromBTS(event)) return;

        imTree = new ImTree(decoderService);
        const entry = createResetEntry(event.timestamp);
        updateEvent.notify(entry);
        return entry;

        function isResetFromBTS(event) {
            return event.source === "BTS";
        }

        function createResetEntry(timestamp) {
            return {
                timestamp: timestamp || entryTimestampProvider.getTimestamp(),
                numOfChanges: 0,
                completeImTree: imTree,
                reset: true
            };
        }
    }

    function pushUpdate(distName, payload) {
        return handleSocketData({
            removedDistNames: [],
            managedObjects: [{distName: distName, payload: payload}]
        });
    }

    function pushRemove(distNameToRemove) {
        var removes = Array.isArray(distNameToRemove) ? distNameToRemove : [distNameToRemove];
        return handleSocketData({
            removedDistNames: removes,
            managedObjects: []
        });
    }
}
