const operationTypes = require('./operationTypes');
const sic = require('../sic');
const distNameService = require('../utils/distNameUtils');
const _ = require('lodash');

module.exports = executionRequestsDispatcher;

function executionRequestsDispatcher({translatingBtsClient, metaService, decoderService}) {
    const EXECUTION_REQUEST_ID = translatingBtsClient.getSicInstance().executionRequest().getId();
    var operationTypesArray = _.toArray(operationTypes);
    var executionRequestUnwatcher = null;
    var distNameExecutorsMap = new Map();

    return {
        registerExecutorForDistName: registerExecutorForDistName,
        removeExecutorForDistName: removeExecutorForDistName,
        // for tests only
        isListenerRegisteredForDistName: isListenerRegisteredForDistName
    };

    function registerExecutionRequestWatcher() {
        return translatingBtsClient.addMessageListener(EXECUTION_REQUEST_ID, handleExecutionRequest);

        function handleExecutionRequest(message) {
            var imiOperations = message.getOperations();
            if (!imiOperations) return;

            imiOperations.forEach((imiOperation) => {
                var operationMsg = translatingBtsClient.getSicInstance().operation();
                operationMsg.setProto(imiOperation);
                var managedObject = resolveManagedObject(operationMsg);
                var distName = managedObject.distName;
                var executor = distNameExecutorsMap.get(distName);
                if (!executor) return true;

                var operationEvent = {
                    distName: distName,
                    type: operationTypesArray[operationMsg.getOperationType()]
                };

                const metaUid = metaService.getLastMetaUid();
                const object = decoderService.decodeByDistName(metaUid, managedObject.payload, distName);
                if (operationEvent.type !== operationTypes.EXECUTE) {
                    operationEvent.object = object;
                } else {
                    const typeInfo = metaService.getTypeInfo(metaUid, distNameService.getClassName(distName));
                    const operationName = findOperationName(object, typeInfo);
                    operationEvent.operationName = operationName;
                    operationEvent.operationParameters = object[operationName];
                }
                executor(operationEvent);
            });
        }

        function resolveManagedObject(operationMsg) {
            return translatingBtsClient.resolveManagedObjects([operationMsg.getObject()])[0];
        }

        function findOperationName(parameters, typeInfo) {
            const possibleOperationNames = typeInfo.getOperations().map(operation => operation.getName());

            const parameterNames = Object.keys(parameters || {});
            const operationName = parameterNames.find(parameterName => possibleOperationNames.includes(parameterName));

            if (!operationName) {
                throw new Error('Parameters "' + parameterNames + '" don\'t contain any of valid operations ("' + possibleOperationNames + '")')
            }
            return operationName;
        }
    }

    function registerExecutorForDistName(distName, executor) {
        executionRequestUnwatcher = executionRequestUnwatcher || registerExecutionRequestWatcher();
        distNameExecutorsMap.set(distName, executor);
    }

    function removeExecutorForDistName(distName) {
        distNameExecutorsMap.delete(distName);
        if (_.isEmpty(distNameExecutorsMap) && executionRequestUnwatcher) {
            executionRequestUnwatcher.unwatch();
            executionRequestUnwatcher = null;
        }
    }

    function isListenerRegisteredForDistName(distName) {
        return distNameExecutorsMap.has(distName);
    }
}
