const protoEncoder = require('../protoEncoder');
const _ = require('lodash');
const logger = require('../logger.impl');

const log = logger.getLogger('admin.core.operationExecutor');

const EXECUTE = 3;
const EXECUTION_STATUS_TABLE = ['EXECUTED', 'REJECTED', 'FAILED', 'NOTHING_TO_DO', 'OUTDATED',
    'NOT_SUPPORTED', 'NO_EXECUTOR_RESPONSE'];

module.exports = function () {
    var requestId = 1;
    var translatingBtsClient = null;

    return {
        executeOperation,
        setTranslatingBtsClient,
        reset: resetRequestId //visible for testing
    };

    function setTranslatingBtsClient(client) {
        translatingBtsClient = client;
    }

    function resetRequestId() {
        requestId = 1;
    }

    function createImiOperation(operation) {
        var payload = encodeOperation(operation);
        return {
            operation_type: operation.type,
            object: {
                object_info: translatingBtsClient.resolveObjectInfo(operation.managedObject.distName),
                payload: payload
            }
        };
    }

    function encodeOperation(operation) {
        var parameters = _.cloneDeep(operation.managedObject.getDecodedParameters());
        if (operation.type === EXECUTE) {
            parameters[operation.name] = operation.parameters || {};
        }

        return protoEncoder().encode(operation.typeInfo, parameters);
    }

    function createMessage(operation) {
        var msg = translatingBtsClient.getSicInstance().changeRequest();
        msg.setRequestId(requestId);
        msg.setOperations([createImiOperation(operation)]);
        return msg;
    }

    function createOperationListener(callbackFunction) {
        return {
            operation: {
                requestId: requestId,
                callbackFunction: callbackFunction
            }
        };
    }

    /* operation object should has structure like this:
     var operation = {
     name: oper_setSideId - needed if operationType is EXECUTE
     parameters: {sideId: 1, otherParam: false} - needed if operationType is EXECUTE
     managedObject": treeNode_containing_operation,
     typeInfo": typeInfo_of_managed_object,
     type: operation type
     };
     */

    function executeOperation(operation) {
        return new Promise(function (resolve, reject) {
            try {
                var msg = createMessage(operation);
                translatingBtsClient.sendOperation(msg, createOperationListener(operationSent));
                requestId++;
            } catch (exception) {
                log.error('Operation: ' + operationIdentifier() + ' with requestId: ' + requestId +
                    ' was not send, because of error: ' + exception);

                reject({error: exception, requestId: requestId});
            }

            function operationIdentifier() {
                return operation.name || operation.type;
            }

            function operationSent(executionStatus, requestId) {
                var statusString = 'Status of execution operation: ' + operationIdentifier() + ' with requestId: ' +
                    requestId + ' is ' + EXECUTION_STATUS_TABLE[executionStatus];
                log.info(statusString);
                var result = {
                    executionStatus: executionStatus,
                    requestId: requestId,
                    toString: function () {
                        return statusString;
                    }
                };
                if (isExecutedCorrectly()) {
                    return resolve(result);
                }
                return reject(result);

                function isExecutedCorrectly() {
                    return executionStatus === 0;
                }
            }
        });
    }
};
