const _ = require('lodash');
const DefaultOperationExecutor = require('./operationExecutor');
const distNameService = require('../utils/distNameUtils');
const operationTypes = require('./operationTypes');

module.exports = operationsService;

function operationsService({translatingBtsClient, metaService, executionRequestsDispatcher, onlineImTree, operationExecutor = DefaultOperationExecutor}) {
    const executor = operationExecutor();
    let imTree;

    const imTreeDefer = Promise.pending();
    onlineImTree.watchUpdates(imTreeListener);
    executor.setTranslatingBtsClient(translatingBtsClient);

    return {
        executeOperation: executeOperation,
        createObject: createObject,
        createObjectAsOwner: createObjectAsOwner,
        updateObject: updateObject,
        deleteObject: deleteObject,
        isConnected: isConnected,
        //visible for tests:
        createOperation: createOperation,
        deleteObjectOperation: deleteObjectOperation,
        chooseOperationName: chooseOperationName
    };

    function chooseOperationName(typeInfo, nameOrNames) {
        const names = Array.isArray(nameOrNames) ? nameOrNames : [nameOrNames];
        const foundName = names.find((name) => typeInfo.getOperationByName(name));
        if (!foundName) throw new Error(`There is no operation "${names.join(' or ')}" in type "${typeInfo.getClassName()}"`);
        return foundName;
    }

    function imTreeListener(updateEvent) {
        imTree = updateEvent.completeImTree;
        imTreeDefer.resolve();
    }

    function isConnected() {
        return translatingBtsClient.isConnected;
    }

    function getTreeNode(distName) {
        return new Promise((resolve, reject) => {
            const treeNode = imTree.findByDistName(distName);
            if (!treeNode) reject(`Managed object with distName: ${distName} doesn\'t exist`);
            resolve(treeNode);
        });
    }

    function checkIfObjectIsNotCreated(distName) {
        return new Promise((resolve, reject) => {
            const treeNode = imTree.findByDistName(distName);
            if (!treeNode) resolve(distName);
            reject(`Managed object with distname: ${distName} already exists.`);
        });
    }

    function operationForExecute(nameOrNames, parameters, distName) {
        return imTreeDefer.promise
            .then(() => getTreeNode(distName)
                .then(getOperation));

        function getOperation(treeNode) {
            const typeInfo = metaService.getTypeInfo(treeNode.metaUid, treeNode.className);
            const name = chooseOperationName(typeInfo, nameOrNames);

            return {
                name: name,
                parameters: parameters,
                managedObject: treeNode,
                typeInfo: typeInfo,
                type: operationTypes.EXECUTE.value
            };
        }
    }

    function executeOperation(nameOrNames, distName, parameters) {
        return operationForExecute(nameOrNames, parameters, distName).then(executor.executeOperation);
    }

    function treeNodeDecorator(treeNode, parameters) {
        return {
            metaUid: treeNode.metaUid,
            distName: treeNode.distName,
            className: treeNode.className,
            getDecodedParameters() {
                return _.merge(treeNode.getDecodedParameters(), parameters);
            }
        };
    }

    function updateOperation(distName, parameters) {
        return getTreeNode(distName).then(getOperation);

        function getOperation(treeNode) {
            return {
                managedObject: treeNodeDecorator(treeNode, parameters),
                typeInfo: metaService.getTypeInfo(treeNode.metaUid, treeNode.className),
                type: operationTypes.UPDATE.value
            };
        }
    }

    function updateObject(distName, parameters) {
        return imTreeDefer.promise
            .then(() => updateOperation(distName, parameters)
                .then(executor.executeOperation));
    }

    function createOperation(distName, parameters, operationType) {
        const managedObject = {
            metaUid: metaService.getLastMetaUid(),
            distName: distName,
            className: distNameService.getClassName(distName),
            getDecodedParameters() {
                return parameters;
            }
        };
        return {
            managedObject: managedObject,
            typeInfo: metaService.getTypeInfo(managedObject.metaUid, managedObject.className),
            type: operationType.value
        };
    }

    function createObjectAsOwner(distName, parameters, executionListener) {
        return imTreeDefer.promise
             .then(() => checkIfObjectIsNotCreated(distName)
                 .then(() => createOperation(distName, parameters, operationTypes.CREATE_WITH_OWNERSHIP))
                 .then((operation) => executor.executeOperation(operation)))
             .then(registerExecutionListener);

        function registerExecutionListener() {
            executionRequestsDispatcher.registerExecutorForDistName(distName, executionListener);
        }
    }

    function createObject(distName, parameters) {
        const operation = createOperation(distName, parameters, operationTypes.CREATE);
        return executor.executeOperation(operation);
    }

    function deleteObjectOperation(distName) {
        return getTreeNode(distName).then(getOperation);

        function getOperation(treeNode) {
            return {
                managedObject: treeNode,
                typeInfo: metaService.getTypeInfo(treeNode.metaUid, treeNode.className),
                type: operationTypes.DELETE.value
            };
        }
    }

    function deleteObject(distName) {
        function removeExecutionListener() {
            executionRequestsDispatcher.removeExecutorForDistName(distName);
        }

        return imTreeDefer.promise
            .then(() => deleteObjectOperation(distName)
                .then(executor.executeOperation)
                .then(removeExecutionListener));
    }
}
