module.exports = rmodsAvailabilityResponder;

const _ = require('lodash');

function rmodsAvailabilityResponder(imTracker) {
    const PARAMETERS = {
            PROCEDURAL_STATE: 'stateInfo.proceduralState',
            BLOCKING_STATE: 'stateInfo.blockingState',
            SERIAL_NUMBER: 'serialNumber'
        },
        RADIO_STATE = {
            RESETTING: 'resetting',
            BLOCKED: 'Blocked',
            UNBLOCKED: 'Unblocked'
        },
        RU_OBJECT_CLASS = 'RU_L';

    const listeners = [],
        ruObjectsAvailability = new Map(),
        paramatersValues = Object.keys(PARAMETERS).map(function (key) {
            return PARAMETERS[key];
        }),
        ruListener = imTracker.trackByClassName(RU_OBJECT_CLASS, createRuListener(), paramatersValues);

    var isResponderStarted = false;

    return {
        isRmodBlocked: isRmodBlocked,
        isRmodUnblocked: isRmodUnblocked,
        isRmodResetting: isRmodResetting,
        isRmodPresent: isRmodPresent,
        register: register,
        unregister: unregister
    };

    function createRuListener() {
        return {
            onListenerRegistered: function (event) {
                updateAvailability(event.objects);
                notifyListeners();
            },
            onParameterChanged: function (event) {
                updateAvailability(event.objects);
                notifyListeners();
            },
            onObjectRemoved: function (event) {
                var removedDistNames = event.distNames;
                removedDistNames.forEach(function (removedDistName) {
                    ruObjectsAvailability.delete(removedDistName);
                });
                notifyListeners();
            }
        };

        function updateAvailability(ruObjects) {
            ruObjects.forEach(function (ruObject) {
                var parameters = ruObject.parameters,
                    proceduralState = _.get(parameters, PARAMETERS.PROCEDURAL_STATE),
                    blockingState = _.get(parameters, PARAMETERS.BLOCKING_STATE),
                    serialNumber = _.get(parameters, PARAMETERS.SERIAL_NUMBER);

                if (blockingState)
                    ruObjectsAvailability.set(ruObject.distName, {
                        state: getState(blockingState, proceduralState),
                        serialNumber: serialNumber
                    });
            });

            function getState(blockingState, proceduralState) {
                if (proceduralState === RADIO_STATE.RESETTING) return RADIO_STATE.RESETTING;
                if (blockingState === RADIO_STATE.BLOCKED) return RADIO_STATE.BLOCKED;
                if (blockingState === RADIO_STATE.UNBLOCKED) return RADIO_STATE.UNBLOCKED;
                return null;
            }
        }
    }

    function notifyListeners() {
        const aliasMapping = createAliasMapping();
        listeners.forEach(function (listener) {
            listener.update(aliasMapping);
        });
    }

    function createAliasMapping() {
        const aliasMapping = [];
        ruObjectsAvailability.forEach(function (ruObject, distName) {
            aliasMapping.push({
                alias: ruObject.serialNumber,
                distName: distName
            });
        });
        return aliasMapping;
    }

    function register(listener) {
        return new Promise(function (resolve) {
            listeners.push(listener);
            ruListener
                .then(function () {
                    isResponderStarted = true;
                    notifyListeners();
                    resolve(listener);
                });
        });
    }

    function unregister(toUnregister) {
        _.remove(listeners, function (listener) {
            return toUnregister === listener;
        });
        notifyListeners();
    }

    function isRmodBlocked(ruDistName) {
        if (isRmodPresent(ruDistName))
            return getRmodState(ruDistName) === RADIO_STATE.BLOCKED;
        return false;
    }

    function isRmodUnblocked(ruDistName) {
        if (isRmodPresent(ruDistName))
            return getRmodState(ruDistName) === RADIO_STATE.UNBLOCKED;
        return false;
    }

    function isRmodResetting(ruDistName) {
        if (isRmodPresent(ruDistName))
            return getRmodState(ruDistName) === RADIO_STATE.RESETTING;
        return false;
    }

    function isRmodPresent(ruDistName) {
        return isResponderStarted && ruObjectsAvailability.has(ruDistName);
    }

    function getRmodState(ruDistName) {
        return ruObjectsAvailability.get(ruDistName).state;
    }
}