module.exports = siteAvailabilityResponder;

const _ = require('lodash');

function siteAvailabilityResponder(imTracker, resetRequestTracker) {
    const SITE_STATE = {
            BLOCKED: 'Blocked',
            UNBLOCKED: 'Unblocked'
        },
        STATE_INFO = {
            BLOCKING_STATE: 'stateInfo.blockingState',
            OPERATIONAL_STATE: 'stateInfo.operationalState'
        },
        LNBTS_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1';

    var isTrackerStarted = false;
    var listeners = [];

    var availableLnbts = {};

    const stateInfoValues = Object.keys(STATE_INFO).map(function (key) {
        return STATE_INFO[key];
    });
    const trackerPromise = imTracker.trackByDistName(LNBTS_DISTNAME, createLnbtsListener(), stateInfoValues);

    function createLnbtsListener() {
        return {
            onListenerRegistered: parametersChanged,
            onParameterChanged: parametersChanged,
            onObjectRemoved: objectRemoved
        };

        function parametersChanged(event) {
            const parameters = event.parameters;
            checkLnbtsAvailability(parameters);
            notifyListeners();
        }

        function objectRemoved() {
            availableLnbts = {};
            notifyListeners();
        }
    }

    function checkLnbtsAvailability(parameters) {
        const blockingState = _.get(parameters, STATE_INFO.BLOCKING_STATE),
            operationalState = _.get(parameters, STATE_INFO.OPERATIONAL_STATE);
        if (blockingState)
            availableLnbts = {
                distName: LNBTS_DISTNAME,
                state: getState(blockingState),
                isOperational: operationalState === 'Enabled'
            };

        function getState() {
            if (blockingState === SITE_STATE.BLOCKED) return SITE_STATE.BLOCKED;
            if (blockingState === SITE_STATE.UNBLOCKED) return SITE_STATE.UNBLOCKED;
            return null;
        }
    }

    function notifyListeners() {
        listeners.forEach(function (listener) {
            listener.update(availableLnbts);
        });
    }

    function register(listener) {
        if (!isTrackerStarted)
            return new Promise(function (resolve) {
                listeners.push(listener);
                trackerPromise
                    .then(function () {
                        isTrackerStarted = true;
                        notifyListeners();
                        resolve(listener);
                    });
            });
        listeners.push(listener);
        notifyListeners();
        return Promise.resolve();
    }

    function unregister(toUnregister) {
        _.remove(listeners, function (listener) {
            return toUnregister === listener;
        });
        notifyListeners();
    }

    function isSiteBlocked() {
        if (availableLnbts && availableLnbts.state) return availableLnbts.state === SITE_STATE.BLOCKED;

        return false;
    }

    function isSiteUnblocked() {
        if (availableLnbts && availableLnbts.state) return availableLnbts.state === SITE_STATE.UNBLOCKED;

        return false;
    }

    function isBtsOperational() {
        return availableLnbts && availableLnbts.isOperational;
    }

    function isResetting() {
        return resetRequestTracker.isResetExecuting();
    }

    return {
        isSiteBlocked: isSiteBlocked,
        isSiteUnblocked: isSiteUnblocked,
        isBtsOperational: isBtsOperational,
        isResetting: isResetting,
        register: register,
        unregister: unregister
    };
}

