const Gunzip = require('zlibjs/bin/gunzip.min').Zlib.Gunzip;
const ims2 = require('../../ims2');
const logger = require('../../logger.impl');
const btsServiceDefaultListenerFactory = require('../btsServiceDefaultListener');

const log = logger.getLogger('admin.core.ims2Downloader');

module.exports = ims2Downloader;

function ims2Downloader({
    procedureExecutor,
    fileTransfer,
    btsServiceProcedure,
    onlineImTree,
    procedureIdFactory
    }) {

    const FILE_PROTOCOL = 'file:///';

    return {
        downloadDecompressed,
        downloadCompressed
    };

    function downloadDecompressed(procedureId) {
        log.info('Downloading decompressed .ims2 file from BTS.');
        return download(true, procedureId);
    }

    function downloadCompressed(procedureId) {
        log.info('Downloading compressed .ims2 file from BTS.');
        return download(false, procedureId);
    }

    function download(decompress, procedureId) {
        var imsLocation;

        return new Promise(function (resolve, reject) {
            procedureId = procedureId || procedureIdFactory.getId().id;
            trackProcedure();
            procedureExecutor
                .executeImSnapshot({clientId: procedureId})
                .catch(handleErrors);

            function handleErrors(error) {
                fileTransfer.disconnect();
                log.error('File download failed!');
                reject(error);
            }

            function trackProcedure() {
                const procedureListener = btsServiceDefaultListenerFactory.newInstance(procedureId, onSuccess, onFail);

                btsServiceProcedure.registerListener(procedureListener);

                function onSuccess(object) {
                    btsServiceProcedure.removeListener(procedureListener);
                    startDownloading(object.parameters.targetDN);
                }

                function onFail(status) {
                    reject('collectImSnapshot procedure ended with status ', status);
                    btsServiceProcedure.removeListener(procedureListener);
                }

            }

            function startDownloading(targetDN) {
                var imsLocation;

                onlineImTree.getImTree()
                    .then(findImsLocation)
                    .then(fileTransfer.connect)
                    .then(() => downloadIms(imsLocation))
                    .then(decompressIms)
                    .catch(handleErrors);

                function findImsLocation(imTree) {
                    const imSnapshotFile = imTree.findByDistName(targetDN);
                    if (!imSnapshotFile) {
                        reject('Object ' + targetDN + ' is not visible in current online im tree');
                        return;
                    }
                    imsLocation = imSnapshotFile.getDecodedParameters().path;
                }
            }

            function downloadIms(imsLocation) {
                return fileTransfer.getFile(FILE_PROTOCOL + imsLocation);
            }

            function decompressIms(fileData) {
                log.info('.ims2.gzip file downloaded.');
                fileTransfer.disconnect();
                if (!decompress) {
                    resolve({
                        fileData,
                        procedureId
                    });
                    return;
                }

                var decompressed = new Gunzip(fileData).decompress();
                resolve({
                    fileData: decompressed,
                    procedureId
                });
            }

        });
    }
}