const distNameUtils = require('../utils/distNameUtils');

module.exports = procedureExecutor;
const EMPTY_PROCEDURE_NOTIFICATION = {
    addAvailabilityListener: function (distName, listener) {
    }
};

const BTS_SERVICE_DIST_NAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/BTS_SERVICE-1';
const BLOCK_OPERATION_NAME = ['oper_block', 'block'];
const UNBLOCK_OPERATION_NAME = ['oper_unblock', 'unblock'];
const RESET_OPERATION_NAME = ['oper_reset', 'reset'];
const IM_SNAPSHOT_OPERATION_NAME = ['oper_collectImSnapshot', 'collectImSnapshot'];
const DOWNLOAD_PLAN_NAME = ['oper_downloadPlan', 'downloadPlan'];
const ACTIVATE_PLAN_NAME = ['oper_activatePlan', 'activatePlan'];
const SW_ROLLBACK_NAME = ['oper_swRollback', 'swRollback'];
const RESET_TO_TEST_DEDICATED_NAME = ['oper_resetToTestDedicated', 'resetToTestDedicated'];
const CONFIGURATION_RESET_NAME = ['oper_configurationReset', 'configurationReset'];
const CHANGE_ETHERNET_PORT_SECURITY_NAME = ['oper_changeEthernetPortSecurity', 'changeEthernetPortSecurity'];

const LOG = require('../logger.impl').getLogger('admin.core.procedureExecutor');

function procedureExecutor(operationsService, procedureIdFactory = require('./procedureIdFactory'), procedureNotification = EMPTY_PROCEDURE_NOTIFICATION) {

    return {
        executeBlock,
        executeUnblock,
        executeReset,
        executeImSnapshot,
        executeDownloadPlan,
        executeActivatePlan,
        executeSwRollback,
        executeResetToTestDedicated,
        executeConfigurationReset,
        executeChangeEthernetPortSecurity
    };

    function executeBlock(procedureName, distName, procedureStatusListener, parameters = {}) {
        enableNotifications(procedureName, distName, procedureStatusListener);
        parameters.targetDN = distName;
        return executeOperation(BLOCK_OPERATION_NAME, parameters)
            .then(() => {
                logSuccess('block', distName);
            })
            .catch((error) => {
                logFail('block', distName, error);
            });
    }

    function executeUnblock(procedureName, distName, procedureStatusListener, parameters = {}) {
        enableNotifications(procedureName, distName, procedureStatusListener);
        parameters.targetDN = distName;
        return executeOperation(UNBLOCK_OPERATION_NAME, parameters)
            .then(() => {
                logSuccess('unblock', distName);
            })
            .catch((error) => {
                logFail('unblock', distName, error);
            });
    }

    function executeReset(procedureName, distName, procedureStatusListener, parameters = {}) {
        enableNotifications(procedureName, distName, procedureStatusListener);
        parameters.targetDN = distName;
        return executeOperation(RESET_OPERATION_NAME, parameters)
            .then(() => {
                logSuccess('reset', distName);
            })
            .catch((error) => {
                logFail('reset', distName, error);
            });
    }

    function executeImSnapshot(parameters = {}) {
        LOG.info('Executing collectImSnapshot...');
        return executeOperation(IM_SNAPSHOT_OPERATION_NAME, parameters)
            .then(() => {
                LOG.info('Sending collectImSnapshot succeeded!');
            })
            .catch((error) => {
                LOG.info(`Sending collectImSnapshot failed! ${(error && error.toString())}`);
            });
    }

    /**
     * Sends operation over BTS_SERVICE object to download plan.
     * @param {object} parameters - Object that contains properties of operation.
     * @param {number} parameters.planId - ID of plan.
     * @param {string} parameters.uri - URI of plan, e.q. 'http://10.10.10.10:80/scf.xml'
     * @param {number=} parameters.clientId - ID of client.
     * @returns {Promise}
     */
    function executeDownloadPlan(parameters = {}) {
        LOG.info(`Executing downloadPlan... ${JSON.stringify(parameters)}`);
        return executeOperation(DOWNLOAD_PLAN_NAME, parameters);
    }

    /**
     * Sends operation over BTS_SERVICE object to activate plan.
     * @param {object} parameters - Object that contains properties of operation.
     * @param {number} parameters.planId - ID of plan.
     * @param {number=} parameters.clientId - ID of client.
     * @returns {Promise}
     */
    function executeActivatePlan(parameters = {}) {
        LOG.info(`Executing activatePlan... ${JSON.stringify(parameters)}`);
        return executeOperation(ACTIVATE_PLAN_NAME, parameters);
    }

    /**
     * Sends operation over BTS_SERVICE object to trigger software rollback.
     * @param {object} parameters - Object that contains properties of operation.
     * @param {number=} parameters.clientId - ID of client.
     * @returns {Promise}
     */
    function executeSwRollback(parameters = {}) {
        LOG.info(`Executing swRollback... ${JSON.stringify(parameters)}`);
        return executeOperation(SW_ROLLBACK_NAME, parameters);
    }

    /**
     * Sends operation over BTS_SERVICE object to trigger reset to test dedicated state.
     * @param {object} parameters - Object that contains properties of operation.
     * @param {number=} parameters.clientId - ID of client.
     * @returns {Promise}
     */
    function executeResetToTestDedicated(parameters = {}) {
        LOG.info(`Executing resetToTestDedicated... ${JSON.stringify(parameters)}`);
        return executeOperation(RESET_TO_TEST_DEDICATED_NAME, parameters);
    }

    /**
     * Sends operation over BTS_SERVICE object to trigger configuration reset.
     * @param {object} parameters - Object that contains properties of operation.
     * @param {boolean} parameters.removeSiteId - Flag that indicates if site ID should be removed.
     * @param {boolean} parameters.removeSecurityCredentials - Flag that indicates if security credentials should be removed.
     * @param {number=} parameters.clientId - ID of client.
     * @returns {Promise}
     */
    function executeConfigurationReset(parameters = {}) {
        LOG.info(`Executing configurationReset... ${JSON.stringify(parameters)}`);
        return executeOperation(CONFIGURATION_RESET_NAME, parameters);
    }

    /**
     * Sends operation over BTS_SERVICE object to trigger change of Ethernet port security.
     * @param {object} parameters - Object that contains properties of operation.
     * @param {boolean} parameters.enableEthernetPortSec - Flag that indicates if security should be enabled.
     * @param {number=} parameters.clientId - ID of client.
     * @returns {Promise}
     */
    function executeChangeEthernetPortSecurity(parameters = {}) {
        LOG.info(`Executing changeEthernetPortSecurity... ${JSON.stringify(parameters)}`);
        return executeOperation(CHANGE_ETHERNET_PORT_SECURITY_NAME, parameters);
    }

    function executeOperation(operationName, parameters) {
        parameters.clientId = parameters.clientId || procedureIdFactory.getId().id;
        return operationsService.executeOperation(operationName, BTS_SERVICE_DIST_NAME, parameters);
    }

    function enableNotifications(procedureName, targetDistName, procedureStatusListener) {
        LOG.info(`Executing ${procedureName}...`);
        procedureNotification.addAvailabilityListener(targetDistName, procedureStatusListener);
        procedureStatusListener(true);
    }

    function logSuccess(operationName, distName) {
        LOG.info(`Sending ${operationName} on ${distNameUtils.getName(distName)} succeeded!`);
    }

    function logFail(operationName, distName, error) {
        LOG.warn(`Sending ${operationName} on ${distNameUtils.getName(distName)} failed! ${(error && error.toString())}`);
    }
}
