const logger = require('../logger.impl');
const BTS_SERVICE_PROCEDURE_CLASSNAME = 'BTS_SERVICE_PROCEDURE';
const END_POINT_STATES = ['Rejected', 'Completed', 'Aborted', 'Failed'];

const log = logger.getLogger('admin.core.procedureTrackerFactory');
const noop = () => {
};

module.exports = procedureTrackerFactory;
function procedureTrackerFactory(onlineImTree) {

    return {
        createProcedureTracker: createProcedureTracker
    };

    function createProcedureTracker(id, procedureType, procedureListener) {
        const procedureTracker = btsServiceProcedureTracker(id, procedureType, procedureListener);
        return onlineImTree.watchUpdates(procedureTracker.onUpdate)
            .then(function (unwatcher) {
                procedureTracker.setUnwatcher(unwatcher);
                return procedureTracker;
            });
    }
}

function btsServiceProcedureTracker(clientId, procedureType, procedureListener) {
    const listener = createProcedureListener();
    var unwatcher;
    var lastStatus;

    return {
        onUpdate,
        setUnwatcher
    };

    function createProcedureListener() {
        return {
            onRequested: procedureListener.onRequested || noop,
            onAccepted: procedureListener.onAccepted || noop,
            onRejected: procedureListener.onRejected || noop,
            onEnqueued: procedureListener.onEnqueued || noop,
            onCompleted: procedureListener.onCompleted || noop,
            onAborted: procedureListener.onAborted || noop,
            onFailed: procedureListener.onFailed || noop
        };
    }

    function setUnwatcher(_unwatcher) {
        unwatcher = _unwatcher;
    }

    function onUpdate(updateEvent) {
        if (updateEvent.reset) return;

        const updatedImTree = updateEvent.completeImTree;
        const treeNode = findProcedureObject(updatedImTree);
        if (!treeNode) return;

        const parameters = treeNode.getDecodedParameters();
        const status = parameters.status;
        const targetDN = parameters.targetDN;
        if (END_POINT_STATES.indexOf(status) !== -1) {
            unwatcher.unwatch();
        }
        if (status !== lastStatus) {
            lastStatus = status;
            notify(status, targetDN);
        }
    }

    function notify(status, targetDN) {
        switch (status) {
            case 'Requested':
                listener.onRequested(targetDN);
                break;
            case 'Accepted':
                listener.onAccepted(targetDN);
                break;
            case 'Rejected':
                listener.onRejected(targetDN);
                break;
            case 'Enqueued':
                listener.onEnqueued(targetDN);
                break;
            case 'Completed':
                listener.onCompleted(targetDN);
                break;
            case 'Aborted':
                listener.onAborted(targetDN);
                break;
            case 'Failed':
                listener.onFailed(targetDN);
                break;
            default:
                log.error('Unknown procedure status: ' + status);
        }
    }

    function findProcedureObject(updatedImTree) {
        const treeNodes = updatedImTree.findByClassName(BTS_SERVICE_PROCEDURE_CLASSNAME);
        return treeNodes.filter(isRequestedProcedure)[0];
    }

    function isRequestedProcedure(node) {
        const parameters = node.getDecodedParameters();
        return (parameters.clientId === clientId && parameters.procedureType === procedureType);
    }
}
