const loggers = require('../../logger.impl');
const LOG = loggers.getLogger('admin.core.snapshot');
const utils = require('../../utils/distNameUtils');

function fileDownloaderFactory(snapshotExecutor, fileTransfer) {
    const FILE_PREFIX = 'file:///';
    const TRANSPORT_FILE_PATH = 'transport/';

    return {
        createFileDownloader: createFileDownloader,
        //for testing only
        createFileInfo: createFileInfo,
        getZipUri: getZipUri
    };

    function createFileDownloader(distName, zip, filesInfo, errorCallback) {

        return {
            downloadFile: downloadFile
        };

        function downloadFile(parameters) {
            const uri = parameters.uri;
            const presentationUri = getZipUri(uri);
            const fileInfo = createFileInfo(presentationUri);
            filesInfo.push(fileInfo);
            filesInfo.notify();
            LOG.info('Starting download file: ' + uri + ' given by managed object with distName: ' + utils.getName(distName));

            return fileTransfer.getFile(uri)
                .then(zipFile)
                .then(handleSuccessfulDownload)
                .catch(handleError)
                .then(deleteFileFromBTS);

            function zipFile(file) {
                zip.addFile(file, presentationUri);
                return file.length;
            }

            function handleSuccessfulDownload(fileSize) {
                fileInfo.status = 'status-downloaded';
                fileInfo.fileSize = Math.round(fileSize / 1024);
                LOG.info('File: ' + fileInfo.fileName + ' has been successfully downloaded.');
                filesInfo.notify();
            }

            function handleError(error) {
                fileInfo.status = 'status-error';
                fileInfo.details = error.type + ': ' + error.error;
                filesInfo.notify();

                const errorMessage = error.type + ' occurred for file: ' + error.path + '. Details: ' + error.error;
                LOG.error(errorMessage);

                return errorMessage;
            }

            function deleteFileFromBTS(error) {
                snapshotExecutor.deleteFile(distName, error).catch(errorCallback);
            }
        }
    }

    function createFileInfo(uri) {
        return {
            fileName: uri,
            status: 'status-downloading',
            fileSize: '',
            details: ''
        };
    }

    function getZipUri(uri) {
        if (isInternalFile(uri)) return uri.slice(FILE_PREFIX.length);

        return TRANSPORT_FILE_PATH + uri;

        function isInternalFile() {
            return uri.slice(0, FILE_PREFIX.length) === FILE_PREFIX;
        }
    }
}

module.exports = fileDownloaderFactory;
