const snapshotStates = require('./snapshotStates');

const loggers = require('../../logger.impl');
const LOG = loggers.getLogger('admin.core.snapshot');

function snapshot(snapshotExecutor, snapshotListenerFactory, filesInfoFactory, snapshotStatusHandler, procedureIdFactory, procedureTrackerFactory) {
    const STATUS_HANDLER = snapshotStatusHandler;

    const FILES_INFO = filesInfoFactory.createFileInfo();

    const PROCEDURE_LISTENER = {
        onAccepted: accepted,
        onRejected: rejected,
        onCompleted: completed,
        onAborted: aborted
    };

    var selectedCoverage = null;

    return {
        isSnapshotPossible: STATUS_HANDLER.isSnapshotPossible,
        isSnapshotActive: STATUS_HANDLER.isSnapshotActive,
        addStatusListener: STATUS_HANDLER.addStatusListener,
        startSnapshot: startSnapshot,
        cancelSnapshotByUser: cancelSnapshotByUser,
        addFileInfoListener: addFileInfoListener,
        getSelectedCoverage: getSelectedCoverage
    };


    function accepted(snapshotDistName) {
        return registerSnapshotListener(snapshotDistName)
            .then(function () {
                STATUS_HANDLER.setStatus(snapshotStates.ACTIVE);
            })
            .catch(interruptSnapshot);


        function registerSnapshotListener() {
            return snapshotListenerFactory.activateFilesDownload(snapshotDistName, FILES_INFO, interruptSnapshot);
        }

        function interruptSnapshot(error) {
            LOG.error(error);
            return cancelSnapshot(snapshotStates.INTERRUPTED);
        }
    }

    function rejected() {
        STATUS_HANDLER.setStatus(snapshotStates.REJECTED);
    }

    function completed() {
        STATUS_HANDLER.setCompletionStatus(FILES_INFO);
    }

    function aborted() {
        STATUS_HANDLER.setStatus(snapshotStates.CANCELLED);
    }

    function startSnapshot(coverage) {
        selectedCoverage = coverage;
        FILES_INFO.reset();
        const id = procedureIdFactory.getId().id;
        return procedureTrackerFactory.createProcedureTracker(id, 'Snapshot', PROCEDURE_LISTENER)
            .then(markAsRequested)
            .then(triggerSnapshotStartOperation)
            .then(logStart)
            .catch(handleSnapshotRequestCreationError);

        function markAsRequested() {
            STATUS_HANDLER.setStatus(snapshotStates.REQUESTED);
        }

        function triggerSnapshotStartOperation() {
            snapshotExecutor.startSnapshot(coverage, id);
        }

        function logStart() {
            LOG.debug('Operation collectSnapshot has been executed.');
        }

        function handleSnapshotRequestCreationError(error) {
            LOG.error("Operation collectSnapshot hasn't been executed. Reason: " + error);
            STATUS_HANDLER.setStatus(snapshotStates.REJECTED);
        }
    }

    function cancelSnapshotByUser() {
        return cancelSnapshot(snapshotStates.CANCELLED);
    }

    function addFileInfoListener(watcher) {
        return FILES_INFO.addFileInfoListener(watcher);
    }

    function getSelectedCoverage() {
        return selectedCoverage;
    }

    function cancelSnapshot(status) {
        STATUS_HANDLER.setStatus(snapshotStates.CANCELLING);
        return snapshotExecutor.cancel()
            .then(notifyAboutCancel);

        function notifyAboutCancel() {
            STATUS_HANDLER.setStatus(status);
        }
    }
}

module.exports = snapshot;
