const events = require('../../events');
const LOG = require('../../logger.impl').getLogger('admin.core.snapshot');
const snapshotStates = require('./snapshotStates');

const TECHREP_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/TECHREP_L-1';

function snapshotStatusHandler(onlineImTree, socket, snapshotTRAgent) {
    const STATUS_EVENT = events.newEvent();
    var stateInfo = {};
    var status = snapshotStates.AVAILABLE;
    onlineImTree.watchUpdates(onTreeUpdated);

    return {
        isSnapshotPossible: isSnapshotPossible,
        isSnapshotActive: isSnapshotActive,
        addStatusListener: addStatusListener,
        setStatus: setStatus,
        setCompletionStatus: setCompletionStatus,
        isStarted: isStarted
    };

    function onTreeUpdated(updatedEvent) {
        if (updatedEvent.reset) {
            handleReset();
            return;
        }

        const imTree = updatedEvent.completeImTree;
        const treeNode = imTree.findByDistName(TECHREP_DISTNAME);
        if (treeNode) storeStateInfo(treeNode.getDecodedParameters().stateInfo);

        function storeStateInfo(parameters) {
            stateInfo = parameters;
            if (isAvailable() && isSnapshotActive()) setStatus(snapshotStates.BUSY);
            if (canSnapshotStart() && !isTRagentActive()) setStatus(snapshotStates.AVAILABLE);
            if (canSnapshotStart() && isTRagentActive()) setStatus(snapshotStates.AVAILABLE_PARTIALLY);
            if (isAvailable() && !canSnapshotBeDone(stateInfo)) setStatus(snapshotStates.DISABLED);
        }

        function isAvailable() {
            return status === snapshotStates.AVAILABLE;
        }

        function isBusy() {
            return status === snapshotStates.BUSY;
        }

        function canSnapshotStart() {
            return (isBusy() || isDisabled()) && canSnapshotBeDone(stateInfo);
        }

        function isDisabled() {
            return status === snapshotStates.DISABLED;
        }

        function isTRagentActive() {
            return snapshotTRAgent.isSnapshotTRAgentActive();
        }
    }

    function handleReset() {
        if (status === snapshotStates.ACTIVE)
            setStatus(snapshotStates.DISCONNECTED);
        stateInfo = {};
    }

    function isSnapshotPossible(force = 'Yes') {
        const isPossible = !isStarted() && canSnapshotBeDone(stateInfo);
        if (force === 'Yes')
            return isPossible;

        return isPossible && status !== snapshotStates.AVAILABLE_PARTIALLY;
    }

    function isSnapshotActive() {
        return status !== snapshotStates.CANCELLING && stateInfo.usageState === 'Active';
    }

    function addStatusListener(watcher) {
        const unwatcher = STATUS_EVENT.watch(watcher);
        STATUS_EVENT.notify(status);
        return unwatcher;
    }

    function setStatus(state) {
        status = state;
        LOG.info(status.text);
        STATUS_EVENT.notify(status);
    }

    function setCompletionStatus(filesInfo) {
        if (status !== snapshotStates.CANCELLED) {
            const completionStatus = filesInfo.isSuccessful() ? snapshotStates.FINISHED : snapshotStates.FINISHED_WITH_ERRORS;
            setStatus(completionStatus);
        }
    }

    function isStarted() {
        return status === snapshotStates.ACTIVE || status === snapshotStates.REQUESTED || status === snapshotStates.CANCELLING;
    }

    function canSnapshotBeDone() {
        return socket.isConnected() && isTechrepInIdleState(stateInfo);

        function isTechrepInIdleState() {
            return stateInfo.administrativeState === 'Unlocked' && stateInfo.operationalState === 'Enabled' && stateInfo.usageState === 'Idle';
        }
    }
}

module.exports = snapshotStatusHandler;
