const _ = require('lodash');

const LOG = require('../../logger.impl').getLogger('admin.core.softwareUpdateExecutor'),
    SW_UPDATE_SERVICE_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/SW_UPDATE_SERVICE-1',
    OPERATION = {
        REQUEST_UPDATE: ['oper_requestUpdate', 'requestUpdate'],
        ABORT_UPDATE: ['oper_abort', 'abort']
    },
    PROTOCOL = {
        HTTP: 'http'
    },
    SERVER_ADDRESS = 'localhost:12344';

function swUpdateExecutorFactory(operationService, fileUpload, imTracker) {
    return {
        newSoftwareUpdateExecutor: newSoftwareUpdateExecutor
    };

    function newSoftwareUpdateExecutor(softwareUpdateConfig) {
        const sequenceNumber = _.uniqueId();
        var strategy;

        if (isZipSource())
            strategy = createFromZipStrategy();
        else if (isRemoteSource())
            strategy = createRemoteServerStrategy();

        function executeStrategy() {
            return strategy.execute();
        }

        const context = {
            execute: executeStrategy,
            abort: abort
        };

        return context;

        function abort() {
            return executeOperation(OPERATION.ABORT_UPDATE, {});
        }

        function createRemoteServerStrategy() {
            return {
                execute: execute
            };

            function execute() {
                const serverAddress = softwareUpdateConfig.hostname + ':' + softwareUpdateConfig.port;
                const parameters = createUpdateParameters(serverAddress);
                return executeOperation(OPERATION.REQUEST_UPDATE, parameters);
            }

        }

        function createFromZipStrategy() {
            return {
                execute: execute
            };

            function execute() {
                const SW_UPDATE_L_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/SW_UPDATE_L-1';
                const SW_UPDATE_STATE_PARAM = 'stateInfo.swUpdateState';
                const fileUploadService = fileUpload();
                const swUpdateListener = createSwUploadListener();

                return fileUploadService.startService()
                    .then(function () {
                        return imTracker.trackByDistName(SW_UPDATE_L_DISTNAME, swUpdateListener, SW_UPDATE_STATE_PARAM);
                    })
                    .catch(function (error) {
                        fileUploadService.stopService();
                        return Promise.reject(error);
                    })
                    .then(function () {
                        const parameters = createUpdateParameters();
                        return executeOperation(OPERATION.REQUEST_UPDATE, parameters)
                            .catch(failCallBack);
                    });

                function failCallBack() {
                    fileUploadService.stopService();
                    imTracker.stopTracking(swUpdateListener, SW_UPDATE_L_DISTNAME);
                    return Promise.reject();
                }

                function createSwUploadListener() {
                    var swUpdateState;
                    const listener = {
                        onListenerRegistered: onParameterChanged,
                        onParametersChanged: onParameterChanged,
                        onParameterChanged: onParameterChanged,
                        onObjectRemoved: onObjectRemoved
                    };

                    return listener;

                    function onParameterChanged(event) {
                        const newSwUpdateState = _.get(event.parameters, SW_UPDATE_STATE_PARAM);
                        if (swUpdateState === 'Busy' && newSwUpdateState !== 'Busy') {
                            fileUploadService.stopService();
                            imTracker.stopTracking(listener, SW_UPDATE_L_DISTNAME);
                        } else {
                            swUpdateState = newSwUpdateState;
                        }
                    }

                    function onObjectRemoved() {
                        if (swUpdateState === 'Busy') {
                            fileUploadService.stopService();
                            imTracker.stopTracking(listener, SW_UPDATE_L_DISTNAME);
                        }
                    }
                }
            }
        }

        function createUpdateParameters(serverAddress) {
            serverAddress = serverAddress || SERVER_ADDRESS;
            return {
                'seqNumber': sequenceNumber,
                'targetBdFileName': softwareUpdateConfig.targetBd || '',
                'transferProtocol': PROTOCOL.HTTP,
                'serverAddress': serverAddress,
                'login': softwareUpdateConfig.login || '',
                'password': softwareUpdateConfig.password || '',
                'automaticActivation': softwareUpdateConfig.shouldActivate || false
            };
        }

        function executeOperation(name, parameters) {
            return operationService.executeOperation(name, SW_UPDATE_SERVICE_DISTNAME, parameters)
                .then(function () {
                    LOG.info('Operation ' + name + ' with parameter automaticActivation set to: ' + parameters.automaticActivation +
                        ' has sent.');
                })
                .catch(function (errorInfo) {
                    LOG.error('Error has occurred during sending ' + name + ' operation. [' + errorInfo + ']');
                    return Promise.reject(errorInfo);
                });
        }

        function isZipSource() {
            return softwareUpdateConfig.source === 'zip';
        }

        function isRemoteSource() {
            return softwareUpdateConfig.source === 'remote';
        }
    }
}

module.exports = swUpdateExecutorFactory;
