const ByteBuffer = require('bytebuffer');
const LOG = require('./logger.impl').getLogger('admin.core.protoDecoder');
const _ = require('lodash');

var WIRETYPE_VARINT = 0;
var WIRETYPE_FIXED64 = 1;
var WIRETYPE_LENGTH_DELIMITED = 2;
var WIRETYPE_FIXED32 = 5;

module.exports = function (objectIdMapper) {
    function decode(managedObjectTypeInfo, payload) {
        var result = managedObjectTypeInfo.createEmptyParameters();
        if (payload.byteLength <= 0) return result;

        var buffer = toBigEndianByteBuffer(payload);
        const populated = new Set();
        populate(result, populated, managedObjectTypeInfo, buffer, objectIdMapper);
        validate(populated, managedObjectTypeInfo, managedObjectTypeInfo.getClassName());
        return result;
    }

    return {
        decode: decode
    };
};

function populate(result, populated, parentTypeInfo, buffer, objectIdMapper) {
    while (buffer.remaining() > 0) {
        var tag = toTag(readInteger(buffer));
        var typeInfo = getTypeInfoForId(parentTypeInfo, tag.getFieldNumber());
        var name = typeInfo.getName();

        if (typeInfo.isOperation()) {
            result[name] = typeInfo.createEmptyParameters();
        }

        try {
            if (typeInfo.isArray()) {
                if (typeInfo.isSimple()) {
                    var entry = readSimple(buffer, tag, typeInfo, objectIdMapper);
                    result[name].push(entry);
                } else if (typeInfo.isComplex()) {
                    populateComplex(result[name], populated, buffer, tag, typeInfo);
                }
            } else if (typeInfo.isSimple()) {
                result[name] = readSimple(buffer, tag, typeInfo, objectIdMapper);
            } else if (typeInfo.isComplex()) {
                populateComplex(result[name], populated, buffer, tag, typeInfo);
            }
            populated.add(name);
        } catch (e) {
            handleError(name, e);
        }

        function handleError(parameterName, e) {
            if (parentTypeInfo.getClassName) {
                LOG.error('Failed to decode parameter "' + parameterName + '" of object of class "' + parentTypeInfo.getClassName() + '"')
            } else {
                LOG.error('Failed to decode parameter "' + parentTypeInfo.getName() + '.' + parameterName);
            }
            throw e;
        }
    }
}

function populateComplex(result, populated, buffer, tag, typeInfo) {
    var length = readComplexLength(buffer, tag, typeInfo);
    var complexBuffer = buffer.copy(buffer.offset, buffer.offset + length);
    if (typeInfo.isArray()) {
        var item = typeInfo.createEmptyParameters();
        populate(item, populated, typeInfo, complexBuffer);
        result.push(item);
    } else {
        populate(result, populated, typeInfo, complexBuffer);
    }
    buffer.skip(length);
}

function validate(populated, parentTypeInfo, className) {
    if (!parentTypeInfo.getParameters()) return;

    parentTypeInfo.getParameters().forEach(function(parameter) {
        validateMandatory(parameter);
    });

    if (!parentTypeInfo.getOperations) return;

    parentTypeInfo.getOperations().forEach(function(parameter) {
        validateMandatory(parameter);
    });

    function validateMandatory(parameter) {
        if (parameter.isMandatory() && !populated.has(parameter.getName()))
            throw new Error(`Missing mandatory parameter '${parameter.getName()}' of ${className}`);

        if (parameter.isMandatory() || populated.has(parameter.getName()))
            validate(populated, parameter, className);
    }
}

function toTag(byte) {
    function getFieldNumber() {
        return byte >>> 3;
    }

    function getWireType() {
        return byte & 0x07;
    }

    return {
        getFieldNumber: getFieldNumber,
        getWireType: getWireType
    };
}

function readInteger(bigEndianBuffer) {
    return bigEndianBuffer.readVarint64().toNumber();
}

function readString(bigEndianBuffer) {
    return bigEndianBuffer.readVString();
}

function readObjectId(bigEndianBuffer, objectIdMapper) {
    bigEndianBuffer.mark();
    const bytesCount = bigEndianBuffer.readVarint64().toNumber();
    const bufferOffset = bigEndianBuffer.offset;
    const uuid = bigEndianBuffer.toHex(bufferOffset, bufferOffset + bytesCount);
    var distName = objectIdMapper.resolveDistName(uuid);
    if (!distName) {
        bigEndianBuffer.reset();
        distName = bigEndianBuffer.readVString();
    } else {
        bigEndianBuffer.skip(bytesCount);
    }
    return distName;
}

function readFloat32(bigEndianBuffer) {
    bigEndianBuffer.LE();
    try {
        return bigEndianBuffer.readFloat32();
    } finally {
        bigEndianBuffer.BE();
    }
}

function readFloat64(bigEndianBuffer) {
    bigEndianBuffer.LE();
    try {
        return bigEndianBuffer.readFloat64();
    } finally {
        bigEndianBuffer.BE();
    }
}

function readBoolean(bigEndianBuffer) {
    return readInteger(bigEndianBuffer) !== 0;
}

function readEnum(bigEndianBuffer, typeInfo) {
    var enumValue = readInteger(bigEndianBuffer);
    var enumText = typeInfo.getEnumText(enumValue);
    if (!enumText) throw new Error("Unknown enum value![name=" + typeInfo.getName() + ", value=" + enumValue + "]");
    return enumText;
}

function readComplexLength(bigEndianBuffer, tag, typeInfo) {
    var type = tag.getWireType();
    if (typeInfo.isComplex() && type === WIRETYPE_LENGTH_DELIMITED) {
        return readInteger(bigEndianBuffer);
    }
    throw new Error("Invalid type! expected=" + WIRETYPE_LENGTH_DELIMITED + " actual=" + type);
}

function readSimple(bigEndianBuffer, tag, typeInfo, objectIdMapper) {
    var type = tag.getWireType();
    if (typeInfo.isBoolean() && type === WIRETYPE_VARINT) {
        return readBoolean(bigEndianBuffer);
    }
    if (typeInfo.isEnum() && type === WIRETYPE_VARINT) {
        return readEnum(bigEndianBuffer, typeInfo);
    }
    if (typeInfo.isInteger() && type === WIRETYPE_VARINT) {
        return readInteger(bigEndianBuffer);
    }
    if (typeInfo.isString() && type === WIRETYPE_LENGTH_DELIMITED) {
        return readString(bigEndianBuffer);
    }
    if (typeInfo.isObjectId() && type === WIRETYPE_LENGTH_DELIMITED) {
        return readObjectId(bigEndianBuffer, objectIdMapper);
    }
    if (typeInfo.isFloat()) {
        if (type === WIRETYPE_FIXED32) {
            return readFloat32(bigEndianBuffer);
        }
        if (type === WIRETYPE_FIXED64) {
            return readFloat64(bigEndianBuffer);
        }
    }
    throw new Error("Wrong wire type! " + type);
}

function toBigEndianByteBuffer(payload) {
    var buffer = ByteBuffer.wrap(payload);
    buffer.BE();
    return buffer;
}

function getTypeInfoForId(parentTypeInfo, id) {
    var typeInfo = parentTypeInfo.getParameterById(id);
    if (!!typeInfo) return typeInfo;

    if (parentTypeInfo.getOperationById) {
        typeInfo = parentTypeInfo.getOperationById(id);
        if (!!typeInfo) return typeInfo;
    }

    throw new Error("Unknown parameter/operation with id = " + id);
}
