const ByteBuffer = require('bytebuffer');

module.exports = function encoder() {
    const ARRAY = 'array';
    const COMPLEX = 'complex';

    var writeField = {

        INTEGER: function (buffer, tag, value) {
            buffer.writeKey(tag.varInt());
            buffer.writeInteger(value);
        },

        STRING: function (buffer, tag, value) {
            buffer.writeKey(tag.lengthDelimited());
            buffer.writeString(value);
        },

        BOOLEAN: function (buffer, tag, value) {
            buffer.writeKey(tag.varInt());
            buffer.writeBoolean(value);
        },

        FLOAT: function (buffer, tag, value) {
            buffer.writeKey(tag.const64());
            buffer.writeFloat(value);
        },

        ENUM: function (buffer, tag, value, typeInfo) {
            buffer.writeKey(tag.varInt());
            buffer.writeInteger(typeInfo.getEnumValue(value));
        },

        array: function (buffer, tag, value, typeInfo) {
            var elementType = typeInfo.isSimple() ? typeInfo.getBaseType() : COMPLEX;
            value.forEach(function (element) {
                encodeField(buffer, element, typeInfo, elementType);
            });
        },

        complex: function (buffer, tag, value, typeInfo) {
            buffer.writeKey(tag.lengthDelimited());
            var content = encode(typeInfo, value);
            buffer.writeInteger(content.byteLength);
            buffer.append(content);
        }
    };

    var writeObjectId = {
        OBJECTID: function (buffer, tag, value) {
            buffer.writeKey(tag.lengthDelimited());
            buffer.writeString(value);
        }
    }

    var writers = writeField;

    return {
        encode: encode
    };

    /**
     * @returns {!ArrayBuffer}
     */
    function encode(typeInfo, object, convertObjectIdToString = false) {
        writers = writeField;
        if (convertObjectIdToString)
            writers = Object.assign(writers, writeObjectId);

        try {
            var buffer = newBuffer();
            var newArray = [];
            typeInfo.getParameters().forEach(function (parameterTypeInfo) {
                newArray[parameterTypeInfo.getId()] = parameterTypeInfo;
            });
            if (typeInfo.getOperations)
                typeInfo.getOperations().forEach(function (parameterTypeInfo) {
                    newArray[parameterTypeInfo.getId()] = parameterTypeInfo;
                });

            var value;
            newArray.forEach(function (fieldTypeInfo) {
                value = object[fieldTypeInfo.getName()];
                if (_.isNull(value) || _.isUndefined(value))
                    return;
                encodeField(buffer, value, fieldTypeInfo);
            });
            return buffer.toArrayBuffer();
        } catch (e) {
            throw new Error('Cannot encode the object: ' + e.message, e);
        }
    }

    function encodeField(buffer, value, fieldTypeInfo, fieldType) {
        var fieldNumber = fieldTypeInfo.getId();
        var type = fieldType || getType(fieldTypeInfo);

        writers[type](buffer, tag(fieldNumber), value, fieldTypeInfo);
    }

    function getType(fieldTypeInfo) {
        if (fieldTypeInfo.isArray()) return ARRAY;
        if (fieldTypeInfo.isComplex()) return COMPLEX;
        return fieldTypeInfo.getBaseType();
    }

    function newBuffer() {
        var byteBuffer = new ByteBuffer(0).BE();

        return {

            writeKey: function (key) {
                byteBuffer.writeVarint64(key);
            },

            writeInteger: function (value) {
                byteBuffer.writeVarint64(value);
            },

            writeFloat: function (value) {
                try {
                    byteBuffer.LE().writeFloat64(value);
                } finally {
                    byteBuffer.BE();
                }
            },

            writeString: function (value) {
                this.writeInteger(ByteBuffer.calculateUTF8Bytes(value));
                byteBuffer.writeUTF8String(value);
            },

            writeBoolean: function (value) {
                byteBuffer.writeVarint64(value ? 1 : 0);
            },

            append: function (arrayBuffer) {
                byteBuffer.append(arrayBuffer);
            },

            toArrayBuffer: function () {
                trimToDataSize(byteBuffer);
                return byteBuffer.toArrayBuffer();
            }
        }
    }

    function trimToDataSize(byteBuffer) {
        byteBuffer.limit = byteBuffer.offset;
        byteBuffer.offset = 0;
    }

    function tag(fieldNumber) {

        function key(fieldNumber, wireType) {
            return (fieldNumber << 3) | wireType;
        }

        return {
            varInt: function () {
                return key(fieldNumber, 0);
            },
            const64: function () {
                return key(fieldNumber, 1);
            },
            lengthDelimited: function () {
                return key(fieldNumber, 2);
            },
            const32: function () {
                return key(fieldNumber, 5);
            }
        }
    }
};
