const _ = require('lodash');

const STATUSES = {
    OK: 'OK',
    WARNING: 'WARNING',
    ERROR: 'ERROR'
};

module.exports = Object.freeze({
    create,
    STATUSES,
    getFaultDetails,
    getModuleDetails
});

const SFP_CONNECTION = {
    SOURCE: 'source',
    DESTINATION: 'destination'
};
const ACCEPTABLE_MODULES = ['RMOD_L', 'SMOD_L', 'SMOD_EXT_L', 'BBMOD_L'];
const ACCEPTABLE_LINKS = ['RP3_01Link', 'CPRILink'];

function getModuleDetails(distName, imTree) {
    if (!imTree) return null;

    var node = imTree.findByDistName(distName);
    var parameters = node && node.getDecodedParameters();
    if (!parameters) return null;

    var details = {
        ipAddress: parameters.ipAddress,
        productName: parameters.productName,
        serialNumber: parameters.serialNumber
    };

    if (node.className === 'RMOD_L') {
        var ruNode = node.findChildByClassName('RU_L');
        var ruParams = ruNode && ruNode.getDecodedParameters();
        details.productCode = ruParams && (ruParams.visibleProductCode || ruParams.productCode);
    } else {
        details.productCode = parameters.visibleProductCode || parameters.productCode;
    }

    return details;
}

function getFaultDetails(distName, imTree) {
    if (!imTree) return null;

    var node = imTree.findByDistName(distName);
    var parameters = node && node.getDecodedParameters();
    var alarmInformation = parameters && parameters.alarmInformation;

    var details = {};
    if (!alarmInformation) return details;

    details.faultSeverity = alarmInformation.faultSeverity;
    details.eventType = alarmInformation.eventType;
    details.observationTime = alarmInformation.observationTime;
    var alarmText = alarmInformation.alarmText;

    if (!alarmText) return details;

    details.alarmDetail = alarmText.alarmDetail;
    details.alarmDetailNbr = alarmText.alarmDetailNbr;
    details.faultDescription = alarmText.faultDescription;

    return details;
}

function create(imTree) {
    return _.chain(findRP3AndCPRICablnks(imTree))
        .map(function (cablnk) {
            return {
                destination: createModuleObject(cablnk, SFP_CONNECTION.DESTINATION, imTree),
                source: createModuleObject(cablnk, SFP_CONNECTION.SOURCE, imTree),
                alarms: getAlarms(cablnk, imTree),
                cablnkDistName: cablnk.distName
            };
        })
        .filter(function (connection) {
            return connection && connection.source && connection.destination;
        })
        .sortByAll([
            function (connection) {
                return connection.source.name;
            },
            function (connection) {
                return connection.source.port;
            }
        ])
        .value();
}

function createModuleObject(cablnk, sourceOrDest, imTree) {
    const cablnkParams = cablnk.getDecodedParameters();
    const connector = cablnkParams.sourceDistName && imTree.findByDistName(cablnkParams[`${sourceOrDest}DistName`]);
    const module = connector && connector.parent;
    if (!module) return null;
    return {
        name: simplifyClassName(module.className) + module.id,
        distName: module.distName,
        status: getStatus(cablnk, connector),
        port: cablnkParams[`${sourceOrDest}PhysicalPortNo`],
        parameters: getSfpParams(cablnkParams[`${sourceOrDest}DistName`], imTree)
    };
}

function simplifyClassName(className) {
    return className.replace('_L', '_');
}

function getStatus(cablnk, connector) {
    var stateInfo = cablnk.getDecodedParameters().stateInfo;
    if (stateInfo && stateInfo.operationalState === 'Enabled') {
        var connectorStateInfo = connector && connector.getDecodedParameters().stateInfo;
        if (connectorStateInfo && connectorStateInfo.operationalState === 'Enabled') return STATUSES.OK;
    }
    return STATUSES.ERROR;
}

function getSfpParams(connectorDistName, imTree) {
    const connector = imTree.findByDistName(connectorDistName);
    const sfp = connector && connector.findChildByClassName('SFP_L');
    const sfpParams = sfp && sfp.findChildByClassName('SFP_PARAMS_L');
    const decodedSfpParams = sfpParams && sfpParams.getDecodedParameters();

    return {
        tx: getRfPower(decodedSfpParams, 'tx'),
        rx: getRfPower(decodedSfpParams, 'rx'),
        temperature: getTemperature(decodedSfpParams),
        ber: decodedSfpParams && decodedSfpParams.ber,
        lcv: decodedSfpParams && decodedSfpParams.lcv,
        currentTx: decodedSfpParams && decodedSfpParams.sfpTxCurr,
        voltageTx: decodedSfpParams && decodedSfpParams.sfpTxVolt,
        distance: decodedSfpParams && decodedSfpParams.transmissionDistance,
        mode: decodedSfpParams && decodedSfpParams.transmissionMode,
        rate: decodedSfpParams && decodedSfpParams.transmissionRate,
        waveLength: decodedSfpParams && decodedSfpParams.waveLength,
        faults: getFaults(decodedSfpParams),
        distName: (sfpParams && sfpParams.distName) || (sfp && sfp.distName)
    };
}

function getFaults(sfpParams) {
    if (!sfpParams || !sfpParams.stateInfo) return null;

    return sfpParams.stateInfo.reasoningStatus;
}

function getTemperature(sfpParameters) {
    if (!sfpParameters) return null;

    return {
        value: sfpParameters.modTemp,
        alarmThreshold: {
            low: sfpParameters.temperatureLowAlarmThreshold,
            high: sfpParameters.temperatureHighAlarmThreshold
        }
    };
}

function getAlarms(cablnkNode, imTree) {
    if (!cablnkNode) return [];

    const alarms = [];
    var observedNodes = getObservedNodes(cablnkNode, imTree);
    var alarmNodes = imTree.findByClassName('ALARM_L');
    alarmNodes.forEach(function (alarmNode) {
        const alarmDetails = getAlarmDetails(alarmNode);
        if (!alarmDetails.faultId || !hasAlarmingOrReportingResource(alarmDetails)) return;
        alarms.push({
            faultId: alarmDetails.faultId,
            distName: alarmDetails.distName
        });
    });

    return _.chain(alarms)
        .uniq('faultId')
        .sortBy('faultId')
        .value();

    function hasAlarmingOrReportingResource(alarmDetails) {
        return observedNodes.indexOf(alarmDetails.alarmingResource) !== -1 ||
            observedNodes.indexOf(alarmDetails.reportingResource) !== -1;
    }
}

function getAlarmDetails(alarmNode) {
    let details = {};
    if (alarmNode) {
        const alarmParams = alarmNode.getDecodedParameters();
        details.unitInformation = alarmParams && alarmParams.unitInformation;
        if (details.unitInformation) {
            details.alarmingResource = details.unitInformation.alarmingResource;
            details.reportingResource = details.unitInformation.reportingResource;
        }
        details.alarmInformation = alarmParams && alarmParams.alarmInformation;
        details.faultId = details.alarmInformation && details.alarmInformation.faultId;
        details.distName = alarmNode.distName;
    }
    return details;
}

function getObservedNodes(cablnkNode, imTree) {
    var observed = [cablnkNode.distName];
    var cablnkParams = cablnkNode.getDecodedParameters();
    if (cablnkParams.destinationDistName) {
        observed.push(cablnkParams.destinationDistName);
        var connector = imTree.findByDistName(cablnkParams.destinationDistName);
        var sfp = connector && connector.findChildByClassName('SFP_L');
        if (sfp) observed.push(sfp.distName);
    }
    return observed;
}

function getRfPower(sfpParameters, rxOrTx) {
    if (!sfpParameters) return null;

    return {
        value: getPowerInMW(sfpParameters[`${rxOrTx}Power`]),
        warningThreshold: {
            low: sfpParameters[`${rxOrTx}PowerLowWarningThreshold`],
            high: sfpParameters[`${rxOrTx}PowerHighWarningThreshold`]
        },
        alarmThreshold: {
            low: sfpParameters[`${rxOrTx}PowerLowAlarmThreshold`],
            high: sfpParameters[`${rxOrTx}PowerHighAlarmThreshold`]
        }
    };
}

function getPowerInMW(power) {
    return power && (power / 1000);
}

function findRP3AndCPRICablnks(imTree) {
    return imTree.findByClassName('CABLNK_L').filter(function (cablnk) {
        return isAcceptableLink(cablnk) && connectsAcceptableModules(cablnk, imTree);
    });

}

function isAcceptableLink(cablnk) {
    return _.includes(ACCEPTABLE_LINKS, cablnk.getDecodedParameters().cablinkType);
}

function connectsAcceptableModules(cablnk, imTree) {
    var cablnkParams = cablnk.getDecodedParameters();
    var sourceConnector = imTree.findByDistName(cablnkParams.sourceDistName);
    var destinationConnector = imTree.findByDistName(cablnkParams.destinationDistName);

    return _.includes(ACCEPTABLE_MODULES, sourceConnector.parent.className) &&
        _.includes(ACCEPTABLE_MODULES, destinationConnector.parent.className);
}

