const imiMessageFactory = require('./imi/imiMessageFactory');
const gatewayMessagesFactory = require('./gateway/gatewayMessageFactory');
const _ = require('lodash');
const loggers = require('./logger.impl');
const LOG = loggers.getLogger('admin.core.sic');

const CONSTANTS = Object.freeze({
    VERSION: '1.0.0',
    DEFAULT_IMI_VERSION: _.first(imiMessageFactory.getAvailableVersions()),
    GATEWAY_VERSION: 2,
    HEADER_SIZE: 16,
    MAX_SIC_MESSAGE_SIZE: 32768,
    SIC_PLUS_HEADER_SIZE: 20,
    SIC_PLUS_ID: 0x3681,
    AACONFIG_SVC_REG_MINOR_VERSION_WILDCARD: 0xFFFFFFFF,
    AASYSCOM_ADDRESS_NONE: 0xFFFEFFFE,
    AASYSCOM_ADDRESS_INVALID: 0x00000000,
    AASYSCOM_CONTROL_ADDRESS: 0x10110613,
    AACONFIG_SVC_CCS_CPID: 0x10110631,
    AASYSCOM_ECHO_CPID: 0x10110610,
    DEFAULT_SIC_ADDRESS: 100 // This tells the gateway on eNB, CONSTANTS it should place it's own address in the messages,
});

module.exports = function sic() {
    let imiMessages = imiMessageFactory.createMessages(CONSTANTS.DEFAULT_IMI_VERSION);
    let gatewayMessages = gatewayMessagesFactory.createMessages(CONSTANTS.GATEWAY_VERSION);
    const gatewayProtocolMessages = gatewayMessagesFactory.createProtocolMessages();
    const IMI_CONSTANTS = imiMessageFactory.IMI_CONSTANTS;
    const definedMessages = {};
    definedMessages[IMI_CONSTANTS.AA_SYS_COM_ECHO_ID] = aaSysComEcho;
    definedMessages[IMI_CONSTANTS.AA_CONFIG_GET_SERVICE_REQ_ID] = aaConfigGetServiceReq;
    definedMessages[IMI_CONSTANTS.AA_CONFIG_GET_SERVICE_RSP_ID] = aaConfigGetServiceRsp;
    definedMessages[IMI_CONSTANTS.SIGN_IN_ID] = signIn;
    definedMessages[IMI_CONSTANTS.SIGN_OUT_ID] = signOut;
    definedMessages[IMI_CONSTANTS.SIGN_IN_RESULT] = signInResult;
    definedMessages[IMI_CONSTANTS.UPDATE_NOTIFICATION_ID] = updateNotification;
    definedMessages[IMI_CONSTANTS.CHANGE_REQUEST_ID] = changeRequest;
    definedMessages[IMI_CONSTANTS.OPERATION_EXECUTED_ID] = operationExecuted;
    definedMessages[IMI_CONSTANTS.REGISTER_EXECUTOR_ID] = registerExecutor;
    definedMessages[IMI_CONSTANTS.EXECUTOR_REGISTERED_ID] = executorRegistered;
    definedMessages[IMI_CONSTANTS.UNREGISTER_EXECUTOR_ID] = unregisterExecutor;
    definedMessages[IMI_CONSTANTS.EXECUTION_REQUEST_ID] = executionRequest;
    definedMessages[IMI_CONSTANTS.EXECUTION_RESULT_ID] = executionResult;
    definedMessages[IMI_CONSTANTS.VERSION_QUERY_ID] = versionQuery;
    definedMessages[IMI_CONSTANTS.VERSION_QUERY_RESULT_ID] = versionQueryResult;
    definedMessages[IMI_CONSTANTS.IMGW.GET_PROTOCOL_VERSION_REQ_ID] = imgwGetProtocolVersionReq;
    definedMessages[IMI_CONSTANTS.IMGW.GET_PROTOCOL_VERSION_RSP_ID] = imgwGetProtocolVersionResp;
    definedMessages[IMI_CONSTANTS.IMGW.GET_META_REQ_ID] = imgwGetMetaReq;
    definedMessages[IMI_CONSTANTS.IMGW.GET_META_RSP_ID] = imgwGetMetaResp;
    definedMessages[IMI_CONSTANTS.IMGW.FILE_SERVICE_GET_REQ_ID] = imgwFileServiceGetReq;
    definedMessages[IMI_CONSTANTS.IMGW.FILE_SERVICE_ERROR_RSP_ID] = imgwFileServiceErrorResp;
    definedMessages[IMI_CONSTANTS.IMGW.FILE_SERVICE_FILE_CONTENT_RSP_ID] = imgwFileServiceFileContentResp;
    definedMessages[IMI_CONSTANTS.IMGW.LOGIN_REQ_ID] = imgwLoginReq;
    definedMessages[IMI_CONSTANTS.IMGW.LOGIN_SUCCESS_RSP_ID] = imgwLoginSuccessResp;
    definedMessages[IMI_CONSTANTS.IMGW.LOGIN_ERROR_RSP_ID] = imgwLoginErrorResp;
    definedMessages[IMI_CONSTANTS.IMGW.DIAG_SESSION_SETUP_REQ_ID] = imgwDiagSessionSetupReq;
    definedMessages[IMI_CONSTANTS.IMGW.DIAG_SESSION_ACCEPT_RSP_ID] = imgwDiagSessionAcceptResp;
    definedMessages[IMI_CONSTANTS.IMGW.DIAG_SESSION_REJECT_RSP_ID] = imgwDiagSessionRejectResp;
    definedMessages[IMI_CONSTANTS.IMGW.DIAG_SESSION_DATA_ID] = imgwDiagSessionData;
    definedMessages[IMI_CONSTANTS.IMGW.DIAG_SESSION_END_ID] = imgwDiagSessionEnd;
    definedMessages[IMI_CONSTANTS.IMGW.DETECT_ENDIANNESS_REQ_ID] = imgwDetectEndiannessReq;
    definedMessages[IMI_CONSTANTS.IMGW.DETECT_ENDIANNESS_RESP_ID] = imgwDetectEndiannessResp;
    definedMessages[IMI_CONSTANTS.IMGW.GET_GATEWAY_INFORMATION_REQ_ID] = imgwGetGatewayInformationReq;
    definedMessages[IMI_CONSTANTS.IMGW.GET_GATEWAY_INFORMATION_RESP_ID] = imgwGetGatewayInformationResp;
    definedMessages[IMI_CONSTANTS.IMGW.SESSIONS_LIST_REQ] = imgwSessionsListReq;
    definedMessages[IMI_CONSTANTS.IMGW.SESSIONS_LIST_RESP] = imgwSessionsListResp;

    return _.assign({
        createHeader,
        createMessage,
        createSicPlusPayload,
        createSicPlusMessage,
        createSicMessageBuffer,
        toMessage,
        definedMessages,
        setImiVersion,
        getImiVersion,
        aaConfigGetServiceReq,
        aaConfigGetServiceRsp,
        aaSysComEcho,
        setGatewayVersion,
        signIn,
        signInResult: signInResult,
        signOut: signOut,
        updateNotification,
        changeRequest,
        operationExecuted,
        registerExecutor,
        executorRegistered,
        unregisterExecutor,
        executionRequest,
        executionResult,
        versionQuery,
        versionQueryResult,
        operation,
        object,
        objectInfo,
        imgwGetProtocolVersionReq,
        imgwGetProtocolVersionResp,
        imgwGetMetaReq,
        imgwGetMetaResp,
        imgwFileServiceGetReq,
        imgwFileServiceFileContentResp,
        imgwFileServiceErrorResp,
        imgwLoginReq,
        imgwLoginSuccessResp,
        imgwLoginErrorResp,
        imgwDiagSessionSetupReq,
        imgwDiagSessionAcceptResp,
        imgwDiagSessionRejectResp,
        imgwDiagSessionData,
        imgwDiagSessionEnd,
        imgwDetectEndiannessReq,
        imgwDetectEndiannessResp,
        imgwGetGatewayInformationReq,
        imgwGetGatewayInformationResp,
        imgwSessionsListReq,
        imgwSessionsListResp
    }, CONSTANTS);

    function createHeader() {
        let buffer = new ArrayBuffer(CONSTANTS.HEADER_SIZE);
        let dataView = new DataView(buffer);

        const id = 0;
        const receiver = 4;
        const sender = 8;
        const length = 12;
        const flags = 14;

        const theHeader = {};

        theHeader.setId = function (value) {
            dataView.setUint32(id, value);
        };
        theHeader.getId = function () {
            return dataView.getUint32(id);
        };
        theHeader.setReceiver = function (value) {
            dataView.setUint32(receiver, value);
        };
        theHeader.getReceiver = function () {
            return dataView.getUint32(receiver);
        };
        theHeader.setSender = function (value) {
            dataView.setUint32(sender, value);
        };
        theHeader.getSender = function () {
            return dataView.getUint32(sender);
        };
        theHeader.setPayloadLength = function (value) {
            dataView.setUint16(length, value + CONSTANTS.HEADER_SIZE);
        };
        theHeader.getPayloadLength = function () {
            return dataView.getUint16(length) - CONSTANTS.HEADER_SIZE;
        };
        theHeader.getLength = function () {
            return dataView.getUint16(length);
        };
        theHeader.setFlags = function (value) {
            dataView.setUint16(flags, value);
        };
        theHeader.getFlags = function () {
            return dataView.getUint16(flags);
        };
        theHeader.setBuffer = function (value) {
            buffer = value;
            dataView = new DataView(value);
        };
        theHeader.getBuffer = function () {
            return buffer;
        };
        theHeader.toUint8Array = function () {
            return new Uint8Array(theHeader.getBuffer());
        };

        theHeader.isLittleEndian = function () {
            const endiannessFlag = theHeader.getFlags() & 0b1100;
            return endiannessFlag === 0b1100 || endiannessFlag === 0b0100;
        };

        theHeader.setPayloadLength(0);
        theHeader.setSender(CONSTANTS.DEFAULT_SIC_ADDRESS);
        theHeader.setReceiver(CONSTANTS.AACONFIG_SVC_CCS_CPID);
        theHeader.setFlags(8);
        return theHeader;
    }

    function createMessage(header) {
        let payload = new ArrayBuffer(header.getPayloadLength());

        const theMsg = {};
        theMsg.getId = header.getId;
        theMsg.setReceiver = header.setReceiver;
        theMsg.getReceiver = header.getReceiver;
        theMsg.setSender = header.setSender;
        theMsg.getSender = header.getSender;
        theMsg.setPayload = function (value) {
            payload = value;
            header.setPayloadLength(new Uint8Array(value).length);
        };
        theMsg.getPayload = function () {
            return payload;
        };
        theMsg.getBuffer = function () {
            const payloadArray = new Uint8Array(payload);
            const buffer = new ArrayBuffer(CONSTANTS.HEADER_SIZE + payloadArray.length);
            new Uint8Array(buffer).set(header.toUint8Array());
            new Uint8Array(buffer).set(payloadArray, CONSTANTS.HEADER_SIZE);
            return buffer;
        };
        theMsg.setBuffer = function (value) {
            const headerBuffer = value.slice(0, CONSTANTS.HEADER_SIZE);
            header = createHeader();
            header.setBuffer(headerBuffer);
            const payloadBuffer = value.slice(CONSTANTS.HEADER_SIZE, header.getLength());
            theMsg.setPayload(payloadBuffer);
        };
        theMsg.toUint8Array = function () {
            return new Uint8Array(theMsg.getBuffer());
        };
        return theMsg;
    }

    function createSicPlusPayload(hostLittleEndianness) {
        const fieldSize = 20;
        //fields
        let seqId = 0;
        let dataOffset = 0;
        let msgId = 0;
        let msgSize = 0;
        let dataSize = 0;
        let data = new Uint8Array(dataSize);

        const theMsg = {};
        theMsg.getId = getId;
        theMsg.setSeqId = setSeqId;
        theMsg.getSeqId = getSeqId;
        theMsg.setDataOffset = setDataOffset;
        theMsg.getBuffer = getBuffer;
        theMsg.setBuffer = setBuffer;
        theMsg.getDataOffset = getDataOffset;
        theMsg.setMsgId = setMsgId;
        theMsg.getMsgId = getMsgId;
        theMsg.setMsgSize = setMsgSize;
        theMsg.getMsgSize = getMsgSize;
        theMsg.getDataSize = getDataSize;
        theMsg.setData = setData;
        theMsg.getData = getData;
        theMsg.toUint8Array = toUint8Array;

        return theMsg;


        function getId() {
            return CONSTANTS.SIC_PLUS_ID;
        }

        function setSeqId(value) {
            seqId = value;
        }

        function getSeqId() {
            return seqId;
        }

        function setDataOffset(value) {
            dataOffset = value;
        }

        function getDataOffset() {
            return dataOffset;
        }

        function setMsgId(value) {
            msgId = value;
        }

        function getMsgId() {
            return msgId;
        }

        function setMsgSize(value) {
            msgSize = value;
        }

        function getMsgSize() {
            return msgSize;
        }

        function getDataSize() {
            return dataSize;
        }

        function setData(value) {
            dataSize = value.length;
            data = value;
        }

        function getData() {
            return data;
        }

        function toUint8Array() {
            return new Uint8Array(theMsg.getBuffer());
        }

        function getBuffer() {
            const payload = new ArrayBuffer(fieldSize + dataSize);
            const dataView = new DataView(payload);
            dataView.setUint32(0, seqId);
            dataView.setUint32(4, dataOffset);
            dataView.setUint32(8, msgId);
            dataView.setUint32(12, msgSize);
            dataView.setUint32(16, dataSize);
            for (let i = 0; i < dataSize; i++)
                dataView.setUint8(20 + i, data[i]);

            return payload;
        }

        function setBuffer(payload) {
            const dataView = new DataView(payload);
            seqId = dataView.getUint32(0, hostLittleEndianness);
            dataOffset = dataView.getUint32(4, hostLittleEndianness);
            msgId = dataView.getUint32(8, hostLittleEndianness);
            msgSize = dataView.getUint32(12, hostLittleEndianness);
            dataSize = dataView.getUint32(16, hostLittleEndianness);
            data = new Uint8Array(dataSize);
            for (let i = 0; i < dataSize; i++)
                data[i] = dataView.getUint8(20 + i);
        }
    }

    function createSicPlusMessage(header) {
        header = header || createHeader();
        header.setId(CONSTANTS.SIC_PLUS_ID);
        const sicPlusPayload = createSicPlusPayload(header.isLittleEndian());
        const theMsg = {};
        theMsg.getId = header.getId;
        theMsg.setReceiver = header.setReceiver;
        theMsg.getReceiver = header.getReceiver;
        theMsg.setSender = header.setSender;
        theMsg.getSender = header.getSender;
        theMsg.getPayload = getPayload;
        theMsg.setPayload = setPayload;
        theMsg.setSeqId = setSeqId;
        theMsg.getSeqId = getSeqId;
        theMsg.setDataOffset = setDataOffset;
        theMsg.getDataOffset = getDataOffset;
        theMsg.setMsgId = setMsgId;
        theMsg.getMsgId = getMsgId;
        theMsg.setMsgSize = setMsgSize;
        theMsg.getMsgSize = getMsgSize;
        theMsg.getDataSize = getDataSize;
        theMsg.setData = setData;
        theMsg.getData = getData;
        theMsg.getBuffer = getBuffer;
        theMsg.setBuffer = setBuffer;
        theMsg.toUint8Array = toUint8Array;

        function setSeqId(value) {
            sicPlusPayload.setSeqId(value);
        }

        function getSeqId() {
            return sicPlusPayload.getSeqId();
        }

        function setDataOffset(value) {
            sicPlusPayload.setDataOffset(value);
        }

        function getDataOffset() {
            return sicPlusPayload.getDataOffset();
        }

        function setMsgId(value) {
            sicPlusPayload.setMsgId(value);
        }

        function getMsgId() {
            return sicPlusPayload.getMsgId();
        }

        function setMsgSize(value) {
            sicPlusPayload.setMsgSize(value);
        }

        function getMsgSize() {
            return sicPlusPayload.getMsgSize();
        }

        function getDataSize() {
            return sicPlusPayload.getDataSize();
        }

        function setData(value) {
            sicPlusPayload.setData(value);
        }

        function getData() {
            return sicPlusPayload.getData();
        }

        function getBuffer() {
            const payloadBytes = new Uint8Array(getPayload());
            if (header.getPayloadLength() !== payloadBytes.length) {
                LOG.trace('Fixing SIC PLUS createMessage length');
                header.setPayloadLength(payloadBytes.length);
            }
            const buffer = new ArrayBuffer(CONSTANTS.HEADER_SIZE + payloadBytes.length);
            new Uint8Array(buffer).set(header.toUint8Array());
            new Uint8Array(buffer).set(payloadBytes, CONSTANTS.HEADER_SIZE);
            return buffer;
        }

        function setBuffer(value) {
            const headerBuffer = value.slice(0, CONSTANTS.HEADER_SIZE);
            header = createHeader();
            header.setBuffer(headerBuffer);
            const payloadBuffer = value.slice(CONSTANTS.HEADER_SIZE, header.getLength());
            setPayload(payloadBuffer);
        }

        function toUint8Array() {
            return new Uint8Array(theMsg.getBuffer());
        }

        function getPayload() {
            return sicPlusPayload.getBuffer();
        }

        function setPayload(payload) {
            sicPlusPayload.setBuffer(payload);
        }

        return theMsg;
    }

    function setImiVersion(value) {
        imiMessages = imiMessageFactory.createMessages(value);
    }

    function getImiVersion() {
        return imiMessages.getImiVersion();
    }

    function setGatewayVersion(value) {
        gatewayMessages = gatewayMessagesFactory.createMessages(value);
    }

    function imgwGetProtocolVersionReq() {
        return gatewayProtocolMessages.imgwGetProtocolVersionReq();
    }

    function imgwGetProtocolVersionResp() {
        return gatewayProtocolMessages.imgwGetProtocolVersionResp();
    }

    function imgwGetMetaReq() {
        return gatewayMessages.imgwGetMetaReq();
    }

    function imgwGetMetaResp() {
        return gatewayMessages.imgwGetMetaResp();
    }

    function imgwLoginReq() {
        return gatewayMessages.imgwLoginReq();
    }

    function imgwLoginSuccessResp() {
        return gatewayMessages.imgwLoginSuccessResp();
    }

    function imgwLoginErrorResp() {
        return gatewayMessages.imgwLoginErrorResp();
    }

    function imgwFileServiceGetReq() {
        return gatewayMessages.imgwFileServiceGetReq();
    }

    function imgwFileServiceFileContentResp() {
        return gatewayMessages.imgwFileServiceFileContentResp();
    }

    function imgwFileServiceErrorResp() {
        return gatewayMessages.imgwFileServiceErrorResp();
    }

    function imgwDiagSessionSetupReq() {
        return gatewayMessages.imgwDiagSessionSetupReq();
    }

    function imgwDiagSessionAcceptResp() {
        return gatewayMessages.imgwDiagSessionAcceptResp();
    }

    function imgwDiagSessionRejectResp() {
        return gatewayMessages.imgwDiagSessionRejectResp();
    }

    function imgwDiagSessionData() {
        return gatewayMessages.imgwDiagSessionData();
    }

    function imgwDiagSessionEnd() {
        return gatewayMessages.imgwDiagSessionEnd();
    }

    function imgwDetectEndiannessResp() {
        return gatewayMessages.imgwDetectEndiannessResp();
    }

    function imgwDetectEndiannessReq() {
        return gatewayMessages.imgwDetectEndiannessReq();
    }

    function imgwGetGatewayInformationReq() {
        return gatewayMessages.imgwGetGatewayInformationReq();
    }

    function imgwGetGatewayInformationResp() {
        return gatewayMessages.imgwGetGatewayInformationResp();
    }

    function imgwSessionsListReq() {
        return gatewayMessages.imgwSessionsListReq();
    }

    function imgwSessionsListResp() {
        return gatewayMessages.imgwSessionsListResp();
    }

    function aaConfigGetServiceReq(hostLittleEndianness) {
        return imiMessages.aaConfigGetServiceReq(hostLittleEndianness);
    }

    function aaConfigGetServiceRsp(hostLittleEndianness) {
        return imiMessages.aaConfigGetServiceRsp(hostLittleEndianness);
    }

    function aaSysComEcho(hostLittleEndianness) {
        return imiMessages.aaSysComEcho(hostLittleEndianness);
    }

    function signIn() {
        return imiMessages.signIn();
    }

    function signOut() {
        return imiMessages.signOut();
    }

    function signInResult() {
        return imiMessages.signInResult();
    }

    function updateNotification() {
        return imiMessages.updateNotification();
    }

    function changeRequest() {
        return imiMessages.changeRequest();
    }

    function operationExecuted() {
        return imiMessages.operationExecuted();
    }

    function registerExecutor() {
        return imiMessages.registerExecutor();
    }

    function executorRegistered() {
        return imiMessages.executorRegistered();
    }

    function unregisterExecutor() {
        return imiMessages.unregisterExecutor();
    }

    function executionRequest() {
        return imiMessages.executionRequest();
    }

    function executionResult() {
        return imiMessages.executionResult();
    }

    function versionQuery() {
        return imiMessages.versionQuery();
    }

    function versionQueryResult() {
        return imiMessages.versionQueryResult();
    }

    function objectInfo() {
        return imiMessages.objectInfo();
    }

    function operation() {
        return imiMessages.operation();
    }

    function object() {
        return imiMessages.object();
    }

    function toMessage(id, buffer, hostLittleEndianness) {
        const messageConstructor = definedMessages[id];
        if (!messageConstructor)
            throw new Error(`Cannot create createMessage with id = 0x${id.toString(16)}. Message id is unknown!`);

        const message = messageConstructor(hostLittleEndianness);
        message.setBuffer(buffer);
        return message;
    }

    function createSicPlusHandler(listener) {
        const buffer = {};

        return {
            handleMessage: handleMessage
        };

        function handleMessage(plusMessage, hostLittleEndianness) {
            const msgKey = hash(plusMessage);
            LOG.trace(`SIC PLUS: received 1 part for : ${msgKey}`);

            buffer[msgKey] = buffer[msgKey] ||
                {
                    id: plusMessage.getMsgId(),
                    currentSize: 0,
                    totalSize: plusMessage.getMsgSize(),
                    parts: []
                };
            const msgEntry = buffer[msgKey];
            msgEntry.parts.push(plusMessage);
            msgEntry.currentSize += plusMessage.getDataSize();

            if (msgEntry.currentSize === msgEntry.totalSize) {
                const completeDataBuffer = merge(msgEntry.totalSize, msgEntry.parts);
                try {
                    listener.completedMessage(toMessage(msgEntry.id, completeDataBuffer, hostLittleEndianness));
                } catch (e) {
                    LOG.error(`Error while processing SIC PLUS createMessage: ${e}`);
                } finally {
                    delete buffer[msgKey];
                }
            } else {
                LOG.trace(`SIC PLUS: buffering ${msgEntry.currentSize}/${msgEntry.totalSize}`);
            }
        }

        function hash(plusMessage) {
            return `[${plusMessage.getSender()}-${plusMessage.getSeqId()}-${plusMessage.getMsgId()}]`;
        }

        function merge(totalSize, parts) {
            const completeData = new Int8Array(totalSize);
            for (let i = 0; i < parts.length; i++)
                completeData.set(parts[i].getData(), parts[i].getDataOffset());

            return completeData.buffer;
        }
    }

    function createSicMessageBuffer(listener) {
        listener.createdHeader = listener.createdHeader || _.noop;
        listener.createdMessage = listener.createdMessage || _.noop;
        listener.completedMessage = listener.completedMessage || _.noop;

        const STATES = {
            HEADER: 'createHeader',
            PAYLOAD: 'payload'
        };
        let state = STATES.HEADER;
        let currentHeader = null;
        let totalBuffer = new ArrayBuffer(0);
        const sicPlusHandler = createSicPlusHandler(listener);

        return {
            push: function (partBuffer) {
                totalBuffer = concat(totalBuffer, partBuffer);
                process();
            }
        };

        function concat(arrayBuffer1, arrayBuffer2) {
            const a = new Int8Array(arrayBuffer1);
            const b = new Int8Array(arrayBuffer2);

            const result = new Int8Array(a.length + b.length);
            result.set(a);
            result.set(b, a.length);
            return result.buffer;
        }

        function process() {
            while (shouldProcessHeader() || shouldProcessPayload()) {
                if (shouldProcessHeader()) processHeader();
                if (shouldProcessPayload()) processPayload();
            }
        }

        function shouldProcessHeader() {
            return state === STATES.HEADER && totalBuffer.byteLength >= CONSTANTS.HEADER_SIZE;
        }

        function shouldProcessPayload() {
            return state === STATES.PAYLOAD && totalBuffer.byteLength >= currentHeader.getPayloadLength();
        }

        function processHeader() {
            const headerBuffer = totalBuffer.slice(0, CONSTANTS.HEADER_SIZE);
            const header = createHeader();
            header.setBuffer(headerBuffer);
            currentHeader = header;
            state = STATES.PAYLOAD;
            totalBuffer = totalBuffer.slice(CONSTANTS.HEADER_SIZE);

            listener.createdHeader(header);
        }

        function processPayload() {
            const payloadBuffer = totalBuffer.slice(0, currentHeader.getPayloadLength());
            const id = currentHeader.getId();
            const msg = createMessage(currentHeader);
            msg.setPayload(payloadBuffer);
            state = STATES.HEADER;
            totalBuffer = totalBuffer.slice(currentHeader.getPayloadLength());

            listener.createdMessage(msg);

            const knownMessage = definedMessages[id];
            if (knownMessage) {
                listener.completedMessage(toMessage(id, payloadBuffer, currentHeader.isLittleEndian()));
            } else if (id === CONSTANTS.SIC_PLUS_ID) {
                const plusMessage = createSicPlusMessage(currentHeader);
                plusMessage.setPayload(payloadBuffer);
                sicPlusHandler.handleMessage(plusMessage, currentHeader.isLittleEndian());
            }
        }
    }
};
