module.exports = {getCommissioningData};

const distNameService = require('../../utils/distNameUtils'),
    imCompatibilityService = require('../../utils/imCompatibilityService');

const COMMISSIONING_STATUSES = {
    COMMISSIONED: 'COMMISSIONED',
    UNCOMMISSIONED: 'UNCOMMISSIONED',
    NOT_INSTALLED: 'NOT_INSTALLED'
};

const commissioningStatusGetters = {
    RMOD_L: getStatusForRmod,
    TRU_L: getStatusForTru,
    BBMOD_L: getStatusForBbmod,
    MHA_L: getStatusForMha,
    RET_L: getStatusForRet
};

function getStatusForRmod(distName, imTree) {
    const uncommissionedStatus = getStatusIfNoRelationWithAObject(distName, imTree);
    if (uncommissionedStatus) return uncommissionedStatus;

    const ruNode = imTree.findByDistName(distName).findChildByClassName('RU_L');
    if (!ruNode)
        return {
            status: COMMISSIONING_STATUSES.NOT_INSTALLED,
            errorMsg: 'imc.views.site.commissioningStatus.NO_RUL_OBJECT'
        };

    const configurationStatus = getStateInfoParam(ruNode, 'configurationStatus');
    const status = configurationStatus === 'Configured';
    return {
        status: status ? COMMISSIONING_STATUSES.COMMISSIONED : COMMISSIONING_STATUSES.NOT_INSTALLED,
        errorMsg: status ? '' : 'imc.views.site.commissioningStatus.UNCONFIGURED'
    };
}

function getStatusForTru(distName, imTree) {
    const ftmNode = imCompatibilityService.findFtmNode(imTree);
    if (!ftmNode)
        return uncommissionedStatusWithInfo('imc.views.site.commissioningStatus.NO_FTM_OBJECT');

    const proceduralStatus = getStateInfoParam(ftmNode, 'proceduralStatus');
    const isFtmOk = proceduralStatus === 'Configured';

    const trsDataNode = getTrsDataNode(imTree);
    if (!trsDataNode)
        return uncommissionedStatusWithInfo('imc.views.site.commissioningStatus.NO_TRSDATA_OBJECT');

    const proceduralState = getStateInfoParam(trsDataNode, 'proceduralState');
    const isTrsDataOk = proceduralState === 'activated';

    const errorMsgs = [];
    if (!isFtmOk) errorMsgs.push(`FTM.proceduralStatus: ${proceduralStatus}`);
    if (!isTrsDataOk) errorMsgs.push(`TRS_DATA.proceduralState: ${proceduralState}`);

    return {
        status: isFtmOk && isTrsDataOk ?
            COMMISSIONING_STATUSES.COMMISSIONED : COMMISSIONING_STATUSES.UNCOMMISSIONED,
        errorMsg: errorMsgs.join(', ')
    };
}

function getTrsDataNode(imTree) {
    const currentNp = imCompatibilityService.findCurrentNpNode(imTree);
    if (!currentNp) return null;

    return currentNp.findChildByClassName('TRS_DATA');
}

function getStatusForBbmod(distName, imTree) {
    return getStatusByOperationalStateAndRelationWithAObject(distName, imTree);
}

function getStatusForMha(distName, imTree) {
    return getStatusByOperationalStateAndRelationWithAObject(distName, imTree);
}

function getStatusForRet(distName, imTree) {
    return getStatusByOperationalStateAndRelationWithAObject(distName, imTree);
}

function getStatusByOperationalStateAndRelationWithAObject(distName, imTree) {
    const uncommissionedStatus = getStatusIfNoRelationWithAObject(distName, imTree);
    return uncommissionedStatus || getStatusForOperationalState(distName, imTree);
}

function getStatusIfNoRelationWithAObject(distName, imTree) {
    const nodeA = imCompatibilityService.findNodeA(imTree.findByDistName(distName));
    if (!nodeA)
        return uncommissionedStatusWithInfo('imc.views.site.commissioningStatus.UNCOMMISSIONED');

    return false;
}

function uncommissionedStatusWithInfo(msg) {
    return {
        status: COMMISSIONING_STATUSES.UNCOMMISSIONED,
        errorMsg: msg
    };
}

function getStatusForOperationalState(distName, imTree) {
    const operationalState = getStateInfoParam(imTree.findByDistName(distName), 'operationalState');
    return {
        status: operationalState === 'Enabled' ?
            COMMISSIONING_STATUSES.COMMISSIONED : COMMISSIONING_STATUSES.NOT_INSTALLED,
        errorMsg: 'imc.views.site.commissioningStatus.UNCONFIGURED'
    };
}

function getStateInfoParam(node, paramName) {
    return node.getDecodedParameters().stateInfo[paramName] || '';
}

function getCommissioningData(distName, imTree) {
    const statusGetter = commissioningStatusGetters[distNameService.getClassName(distName)];
    return statusGetter(distName, imTree);
}


