module.exports = {
    create: createLinks
};

const imCompatibilityService = require('../../utils/imCompatibilityService'),
    hardwareClasses = require('../../siteCommon/hardwareClassesService'),
    connectorsFactory = require('./siteConnectorsFactory'),
    dnService = require('../../utils/distNameUtils');

const FIBER_LENGTH_UNIT = 'm';
const LINK_CAPACITY_UNIT = 'Mb/s';

function findHwtopWithCablnks(imTree) {
    return imCompatibilityService.findHwtopLNode(imTree);
}

function createLinks(imTree) {
    const regularLinks = [];
    const concentratorLinks = [];
    const hwtop = findHwtopWithCablnks(imTree);
    if (!hwtop) return {links: []};

    const cablnkNodes = hwtop.findChildrenByClassName(hardwareClasses.CABLNK);
    cablnkNodes.forEach((node) => {
        const parameters = node.getDecodedParameters();

        const source = imTree.findByDistName(parameters.sourceDistName);
        const target = imTree.findByDistName(parameters.destinationDistName);

        if (source && target) {
            const sourceDirection = connectorsFactory.getConnectorDirection(source),
                targetDirection = connectorsFactory.getConnectorDirection(target),
                link = createLink(node.distName, parameters, sourceDirection, targetDirection),
                isOptical = isOpticalConnector(source) && isOpticalConnector(target);
            if (isOptical) Object.assign(link, createOpticalLinkProperties(node));
            if (isLinkBetweenConcentrators(link))
                concentratorLinks.push(link);
            else
                regularLinks.push(link);
        }
    });
    return {links: regularLinks.concat(createLinksBetweenModules(concentratorLinks, imTree))};
}

function isLinkBetweenConcentrators(link) {
    return dnService.getParentClassName(link.source.distName) === hardwareClasses.CONCENTRATOR
        && dnService.getParentClassName(link.target.distName) === hardwareClasses.CONCENTRATOR;
}

function createLink(distName, parameters, sourceDirection, targetDirection) {
    return {
        distName,
        source: {distName: parameters.sourceDistName, direction: sourceDirection},
        target: {distName: parameters.destinationDistName, direction: targetDirection},
        stateInfo: parameters.stateInfo,
        linkMode: imCompatibilityService.adaptCablnkLinkMode(parameters.linkMode),
        alarms: []
    };
}

function createLinksBetweenModules(concentratorLinks, imTree) {
    const concentratorMap = createCabinetConnectorToModuleConcentratorMap(concentratorLinks),
        cabinets = imCompatibilityService.findEqmLNode(imTree)
            .findChildrenByClassName(hardwareClasses.MODULES.CABINET);

    const fakeLinks = [];
    cabinets.forEach((cabinet) => {
        const hwlinks = imCompatibilityService.findHwLinks(cabinet.findChildByClassName(hardwareClasses.HWTOP));
        hwlinks.forEach(addFakeLink);
    });

    return Array.from(fakeLinks.reduce((map, currentLink) => {
        const linkKey = currentLink.source.distName + currentLink.target.distName;
        const link = map.get(linkKey);
        if (link)
            link.linkMode = mergeLinkMode([link.linkMode, currentLink.linkMode]);
        else
            map.set(linkKey, currentLink);
        return map;
    }, new Map()).values());

    function addFakeLink(hwlink) {
        const hwlinkParameters = hwlink.getDecodedParameters(),
            firstConcentrator = concentratorMap.get(dnService.getParent(hwlinkParameters.destinationDistName)),
            secondConcentrator = concentratorMap.get(dnService.getParent(hwlinkParameters.sourceDistName));
        if (firstConcentrator && secondConcentrator) {
            const endPoints = getEndPoints(firstConcentrator.distName, secondConcentrator.distName);
            const mergedLinkMode = mergeLinkMode([hwlinkParameters.linkMode, firstConcentrator.linkMode, secondConcentrator.linkMode]);
            fakeLinks.push(createFakeConcentratorLink(endPoints.source, endPoints.target, mergedLinkMode));
        }
    }

    function getEndPoints(firstConcentrator, secondConcentrator) {
        const isFirstConcentratorOnSmod = dnService.getParentClassName(firstConcentrator) === hardwareClasses.MODULES.SMOD;
        return {
            source: isFirstConcentratorOnSmod ? firstConcentrator : secondConcentrator,
            target: isFirstConcentratorOnSmod ? secondConcentrator : firstConcentrator
        };
    }
}

function mergeLinkMode(linkModes) {
    return {
        configuration: some(linkModes, 'configuration'),
        data: some(linkModes, 'data'),
        dc: some(linkModes, 'dc'),
        synchronization: some(linkModes, 'synchronization')
    };

    function some(array, property) {
        return array.some((element) => {
            return !!element[property];
        });
    }
}

function createCabinetConnectorToModuleConcentratorMap(concentratorLinks) {
    const concentratorMap = new Map();
    concentratorLinks.forEach((concentratorLink) => {
        const endPoints = getEndPoints(concentratorLink.target.distName, concentratorLink.source.distName);
        concentratorMap.set(endPoints.source, {
            distName: dnService.getParent(endPoints.target),
            linkMode: concentratorLink.linkMode
        });
    });
    return concentratorMap;

    function getEndPoints(firstDistName, secondDistName) {
        const isFirstConcentratorOnCabinet = isConcentratorOnCabinet(firstDistName);
        return {
            source: isFirstConcentratorOnCabinet ? firstDistName : secondDistName,
            target: isFirstConcentratorOnCabinet ? secondDistName : firstDistName
        };

        function isConcentratorOnCabinet(distName) {
            const secondAncestor = dnService.getAncestor(distName, 2);
            return dnService.getClassName(secondAncestor) === (hardwareClasses.MODULES.CABINET);
        }
    }
}

function createFakeConcentratorLink(sourceDistName, targetDistName, linkMode) {
    return {
        source: {distName: sourceDistName, direction: 'Left'},
        target: {distName: targetDistName, direction: 'Left'},
        linkMode: imCompatibilityService.adaptCablnkLinkMode(linkMode),
        stateInfo: {}
    };
}

function createOpticalLinkProperties(linkNode) {
    const parameters = linkNode.getDecodedParameters();
    return {
        isOptical: true,
        fiberLength: {
            value: parameters.fiberLength,
            unit: FIBER_LENGTH_UNIT
        },
        linkMaxPhysicalCapacity: {
            value: parameters.linkMaxPhysicalCapacity,
            unit: LINK_CAPACITY_UNIT
        }
    };

}

function isOpticalConnector(connectorNode) {
    return connectorNode.getDecodedParameters().physPortType === 'optical';
}
