module.exports = {
    createCellsAndLogLinks
};

const _ = require('lodash'),
    hardwareSorter = require('../siteHardware/siteHardwareSorterService'),
    compatibilityService = require('../../utils/imCompatibilityService'),
    channelDataProvider = require('../../utils/channelDataProviderService'),
    caFactory = require('./siteCarrierAggregationFactory'),
    cellMapsService = require('../cells/siteCellMapsService');


const SORTER = hardwareSorter.SORTER;
const DIRECTION = {
    RX: 'RX',
    TX: 'TX'
};


function createCellsAndLogLinks(imTree) {
    var cells = [];
    var logLinks = [];

    var lnbts = compatibilityService.findLnbtsNode(imTree);
    var cellMaps = cellMapsService.create(imTree);

    var lcellNodes = lnbts.findChildrenByClassName('LCELL');
    _.forEach(lcellNodes, function (node) {
        var cell = createCell(node, cellMaps.lcellDistNameToLncelNodeMap.get(node.distName), cellMaps);
        cells.push(cell);
        logLinks = logLinks.concat(cell.logLinks);
    });

    var btsId = lnbts.getDecodedParameters().btsId;
    caFactory.assignSecondaryCellsToEachCellMappedToRuntime(cells, btsId, imTree, cellMaps);
    var interCa = caFactory.collectInterCarrierAggregationCells(btsId, cells);
    return {cells: cells, logLinks: logLinks, interCells: interCa};
}

function createCell(cellNode, lncelNode, cellMaps) {
    var cell = {
        distName: cellNode.distName,
        name: cellNode.getName(),
        number: cellNode.id,
        subCells: createSubCells(cellNode),
        stateInfo: cellNode.getDecodedParameters().stateInfo,
        alarms: []
    };
    cell.logLinks = createLogLinks(cell.subCells);
    cell.channels = collectSubCellsChannels(cell.subCells);

    if (lncelNode) {
        var lncelParameters = lncelNode.getDecodedParameters();
        cell.lncelDistName = lncelNode.distName;
        cell.dlChBw = lncelParameters.dlChBw;
        cell.ulChBw = lncelParameters.ulChBw;
        cell.phyCellId = lncelParameters.phyCellId;
        cell.technology = lncelParameters.technology ? lncelParameters.technology.toUpperCase() : '';
    }

    var lncelANode = cellMaps.lcellDistNameToLncelANodeMap.get(cell.distName);
    if (lncelANode)
        cell.lcrId = lncelANode.getDecodedParameters().lcrId;

    return cell;
}

function collectSubCellsChannels(subCells) {
    var channels = [];
    _.each(subCells, function (subCell) {
        channels = channels.concat(subCell.channels);
    });
    return channels;
}

function createSubCells(cellNode) {
    return _.map(cellNode.findChildrenByClassName('SUBCELL'), createSubCell);
}

function createSubCell(subCellNode) {
    var channelNodes = getChannelNodes(subCellNode);

    var subCell = {
        name: subCellNode.getName(),
        rxChannels: [],
        txChannels: [],
        logLinks: []
    };
    _.forEach(channelNodes, function (channelNode) {
        assignChannel(createChannel(channelNode), subCell, channelNode.getDecodedParameters().direction);
    });

    subCell.rxChannels.sort(SORTER.BY_NUMBER);
    subCell.txChannels.sort(SORTER.BY_NUMBER);
    subCell.channels = subCell.rxChannels.concat(subCell.txChannels);
    return subCell;
}

function getChannelNodes(cellNode) {
    return cellNode.findByClassName('CHANNEL');
}

function assignChannel(channel, subCell, direction) {
    if (direction === DIRECTION.TX)
        subCell.txChannels.push(channel);
    else if (direction === DIRECTION.RX)
        subCell.rxChannels.push(channel);
    else
        throw Error('Unknown channel direction: ' + direction);
}

function createChannel(channelNode) {
    var parameters = channelNode.getDecodedParameters();
    var number = channelNode.id;

    return {
        distName: channelNode.distName,
        name: parameters.direction + ' #' + number,
        direction: parameters.direction,
        number: number,
        axcAddress: channelDataProvider.getAxcAddress(channelNode),
        devicesPath: channelDataProvider.getDevicesPath(channelNode),
        phyantDN: parameters.AntennaDN
    };
}

function createLogLinks(subCells) {
    var logLinks = [];
    _.forEach(subCells, function (subCell) {
        _.forEach(subCell.rxChannels, function (rxChannel) {
            if (rxChannel.phyantDN)
                logLinks.push(createLogLinkChannelToPhyant(rxChannel.distName, rxChannel.phyantDN));
        });
        _.forEach(subCell.txChannels, function (txChannel) {
            if (txChannel.phyantDN)
                logLinks.push(createLogLinkPhyantToChannel(txChannel.phyantDN, txChannel.distName));
        });
    });
    return logLinks;
}

function createLogLinkPhyantToChannel(phyantDistName, channelDistName) {
    return {
        source: {
            distName: phyantDistName,
            direction: 'Right'
        },
        target: {
            distName: channelDistName,
            direction: 'Left'
        }
    };
}

function createLogLinkChannelToPhyant(channelDistName, phyantDistName) {
    return {
        source: {
            distName: channelDistName,
            direction: 'Left'
        },
        target: {
            distName: phyantDistName,
            direction: 'Right'
        }
    };
}
