module.exports = {
    getConnectorDirection,
    getDirectionFromPosition,
    getDirectionForQsfp,
    isOpticalConnector,
    createConnectors
}

const _ = require('lodash'),
    hardwareSorter = require('../siteHardware/siteHardwareSorterService'),
    hardwareClasses = require('../../siteCommon/hardwareClassesService');

const SORTER = hardwareSorter.SORTER;

const CONNECTOR = {
    POSITION: {
        FRONTEND: 'frontend',
        BACKEND: 'backend'
    }
};

const QSFP_PROPERTIES = {
    POSITION: CONNECTOR.POSITION.BACKEND
};

function isPartOfSmodGroup(className) {
    return _.contains(className, hardwareClasses.MODULES.SMOD)
        || _.contains(className, hardwareClasses.MODULES.SMOD_EXT)
        || _.contains(className, hardwareClasses.MODULES.PWRMOD)
        || _.contains(className, hardwareClasses.MODULES.BBMOD);
}

function getConnectorPosition(connectorNode) {
    var parentClass = connectorNode.parent.className;
    if (isPartOfSmodGroup(parentClass)) return CONNECTOR.POSITION.BACKEND;

    var params = connectorNode.getDecodedParameters();
    if (params.connectorLabel === 'RET') return CONNECTOR.POSITION.BACKEND;

    return params.position;
}

function addConnectorToCorrectPosition(connectors, connector, position) {
    if (position === CONNECTOR.POSITION.FRONTEND) {
        connectors.left.push(connector);
        return;
    }
    if (position === CONNECTOR.POSITION.BACKEND) {
        connectors.right.push(connector);
        return;
    }
    connectors.right.push(connector);
}

function createGenericConnectors(moduleNode, connectors) {
    var connectorNodes = moduleNode.findChildrenByClassName(hardwareClasses.CONNECTOR);

    for (var i = 0; i < connectorNodes.length; i++) {
        var connectorNode = connectorNodes[i];
        var connectorParams = connectorNode.getDecodedParameters();
        var number = connectorNode.id;
        var connector = {
            distName: connectorNode.distName,
            number: number,
            label: connectorParams.connectorLabel,
            alarms: [],
            stateInfo: connectorParams.stateInfo
        };
        addConnectorToCorrectPosition(connectors, connector, getConnectorPosition(connectorNode));
    }
}

function createQsfpConnectors(moduleNode, connectors) {
    var qsfps = moduleNode.findChildrenByClassName(hardwareClasses.QSFP);
    for (var i = 0; i < qsfps.length; i++) {
        var qsfp = qsfps[i];
        var qsfpParams = qsfp.getDecodedParameters();
        var label = qsfpParams.physicalLabel || qsfp.getName();
        var qsfpConnector = {
            distName: qsfp.distName,
            label: label,
            alarms: [],
            stateInfo: qsfpParams.stateInfo
        };
        addConnectorToCorrectPosition(connectors, qsfpConnector, QSFP_PROPERTIES.POSITION);
    }
}

function getConnectorDirection(connectorNode) {
    var position = connectorNode.getDecodedParameters().position;
    return getDirectionFromPosition(position);
}

function getDirectionForQsfp() {
    return getDirectionFromPosition(QSFP_PROPERTIES.POSITION);
}

function getDirectionFromPosition(connectorPosition) {
    if (connectorPosition === CONNECTOR.POSITION.FRONTEND) return 'Left';
    if (connectorPosition === CONNECTOR.POSITION.BACKEND) return 'Right';

    return 'Center';
}

function isOpticalConnector(connectorNode) {
    return connectorNode.getDecodedParameters().physPortType === 'optical';
}

function createConnectors(moduleNode) {
    var connectors = {
        left: [],
        right: []
    };
    createGenericConnectors(moduleNode, connectors);
    createQsfpConnectors(moduleNode, connectors);
    connectors.left.sort(SORTER.BY_LABEL);
    connectors.right.sort(SORTER.BY_LABEL);
    return connectors;
}


