module.exports = {
    createMhaModule,
    createRmodModule,
    createGenericModule,
    createModuleWithCommissioningStatus,
    createRetModule,
    createRaeModule,
    createCabinet
};

const _ = require('lodash'),
    connectorsFactory = require('./siteConnectorsFactory'),
    commissioningStatusResolver = require('../commissioningStatus/commissioningStatusResolverFactory'),
    stateInfoInterpreter = require('../stateInfo/stateInfoInterpreterFactory'),
    measurementsFactory = require('./siteMeasurementsFactory'),
    siteElementCommon = require('../common/siteElementCommon'),
    hardwareClasses = require('../../siteCommon/hardwareClassesService'),
    imCompatibilityService = require('../../utils/imCompatibilityService');

function createCabinet(cabinetNode) {
    const cabinet = createGenericModule(cabinetNode);
    cabinet.smods = [];
    cabinet.bbmods = [];
    return cabinet;
}

function createLnaModule(lnaNode) {
    const parameters = lnaNode.getDecodedParameters();
    return {
        distName: lnaNode.distName,
        className: lnaNode.className,
        name: lnaNode.getName(),
        number: lnaNode.id,
        pn: parameters.productName,
        pc: parameters.productCode,
        sn: parameters.serialNumber,
        stateInfo: parameters.stateInfo,
        leftConnectors: [],
        rightConnectors: [],
        alarms: [],
        children: []
    };
}

function createLnaModules(mhaNode, connectors) {
    const nodes = mhaNode.findChildrenByClassName(hardwareClasses.MODULES.LNA);
    const hwtop = getHwtopFromNode(mhaNode);
    const hwlinks = imCompatibilityService.findHwLinks(hwtop);
    return _.map(nodes, (lnaNode) => {
        const lnaModule = createLnaModule(lnaNode);
        lnaModule.leftConnectors.push(chooseLnaConnector(lnaNode, hwlinks, connectors.left));
        lnaModule.rightConnectors.push(chooseLnaConnector(lnaNode, hwlinks, connectors.right));
        return lnaModule;
    });
}

function getHwtopFromNode(moduleNode) {
    return moduleNode.findChildByClassName(hardwareClasses.HWTOP);
}

function chooseLnaConnector(lnaNode, hwlinks, connectors) {
    return _.find(connectors, (connector) => {
        return !!_.find(hwlinks, (hwlink) => {
            const hwlinkParams = hwlink.getDecodedParameters();
            return _.startsWith(hwlinkParams.sourceDistName, connector.distName) &&
                _.includes(hwlinkParams.destinationDistName, lnaNode.getName());
        });
    });
}

function createMhaModule(mhaNode, imTree) {
    const parameters = mhaNode.getDecodedParameters();
    const connectors = connectorsFactory.createConnectors(mhaNode);
    return {
        distName: mhaNode.distName,
        className: mhaNode.className,
        name: mhaNode.getName(),
        number: mhaNode.id,
        pn: parameters.productName,
        pc: parameters.productCode,
        sn: parameters.serialNumber,
        stateInfo: parameters.stateInfo,
        leftConnectors: [],
        rightConnectors: [],
        alarms: [],
        children: createLnaModules(mhaNode, connectors),
        commissioningData: commissioningStatusResolver.getCommissioningData(mhaNode.distName, imTree)
    };
}

function createRetModule(retNode, imTree) {
    const module = createModuleWithCommissioningStatus(retNode, imTree);
    module.children = createChildrenModules(retNode, hardwareClasses.MODULES.RETU);
    return module;
}

function createRaeModule(raeNode) {
    const module = createGenericModule(raeNode);
    module.children = createChildrenModules(raeNode, hardwareClasses.MODULES.RAEU);
    return module;
}

function createChildrenModules(parentNode, childNodeClass) {
    const nodes = parentNode.findChildrenByClassName(childNodeClass);
    return _.map(nodes, createGenericModule);
}

function createRmodModule(moduleNode, imTree) {
    const module = createGenericModule(moduleNode);
    module.commissioningData = commissioningStatusResolver.getCommissioningData(module.distName, imTree);
    const ruNode = moduleNode.findChildByClassName('RU_L');
    if (!ruNode) return module;

    const ruParams = ruNode.getDecodedParameters();
    module.pc = ruParams.visibleProductCode;
    return module;
}

function createGenericModule(moduleNode) {
    const parameters = moduleNode.getDecodedParameters();
    const connectors = createConnectors(moduleNode);
    const name = siteElementCommon.createModuleName(moduleNode);
    return {
        distName: moduleNode.distName,
        className: moduleNode.className,
        name,
        number: moduleNode.id,
        pn: parameters.productName,
        pc: parameters.visibleProductCode || parameters.productCode,
        sn: parameters.serialNumber,
        stateInfo: parameters.stateInfo,
        leftConnectors: connectors.left,
        rightConnectors: connectors.right,
        concentrators: [],
        alarms: [],
        children: [],
        powerGroups: [],
        fanGroup: createFanGroup(moduleNode),
        measurements: createMeasurements(moduleNode)
    };
}

function createFanGroup(moduleNode) {
    let fanNodes = [];

    if (isSmod(moduleNode)) {
        const smodNode = findSmodForSmodL(moduleNode);
        if (smodNode)
            fanNodes = fanNodes.concat(smodNode.findChildrenByClassName('FAN'));
    }

    fanNodes = fanNodes.concat(moduleNode.findChildrenByClassName(hardwareClasses.FAN));

    if (fanNodes.length)
        return {
            stateInfo: stateInfoInterpreter.mergeStateInfos(_.map(fanNodes, (fanNode) => {
                return fanNode.getDecodedParameters().stateInfo;
            })),
            fans: _.map(fanNodes, (fanNode) => {
                return {
                    distName: fanNode.distName,
                    number: fanNode.id
                };
            })
        };

    return null;
}

function findSmodForSmodL(smodL) {
    const smodLSerialNumber = smodL.getDecodedParameters().serialNumber;
    const smods = findByClassNamePath([smodL.root], ['MRBTS', 'EQM', 'SMOD']);
    return _.find(smods, (smod) => {
        return smod.getDecodedParameters().serialNumber === smodLSerialNumber;
    });
}

function isSmod(module) {
    return module.className === hardwareClasses.MODULES.SMOD;
}

function findByClassNamePath(nodes, classNames) {
    if (!classNames || classNames.length === 0 || nodes.length === 0) return [];

    let foundNodes = [];
    _.forEach(nodes, (node) => {
        foundNodes = foundNodes.concat(node.findChildrenByClassName(classNames[0]));
    });

    if (classNames.length === 1)
        return foundNodes;
    return findByClassNamePath(foundNodes, classNames.splice(1));

}

function createMeasurements(moduleNode) {
    return measurementsFactory.createMeasurements(moduleNode);
}

function createConnectors(moduleNode) {
    return connectorsFactory.createConnectors(moduleNode);
}

function createModuleWithCommissioningStatus(moduleNode, imTree) {
    const module = createGenericModule(moduleNode);
    module.commissioningData = commissioningStatusResolver.getCommissioningData(module.distName, imTree);
    return module;
}
