module.exports = {
    create: createHardware
};

const _ = require('lodash'),
    imCompatibilityService = require('../../utils/imCompatibilityService'),
    dnService = require('../../utils/distNameUtils'),
    cellFactory = require('../factories/siteCellFactory'),
    hardwareSorter = require('./siteHardwareSorterService'),
    siteModulesFactory = require('../factories/siteModulesFactory'),
    lnmmesFactory = require('../factories/lnmmesFactory'),
    linksFactory = require('../factories/linksFactory'),
    controllingRetLinksFactory = require('../factories/controllingRetLinksFactory'),
    alarmsFactory = require('../factories/alarmsFactory'),
    hardwareClasses = require('../../siteCommon/hardwareClassesService');


function createHardware(imTree, moduleFactories) {
    if (imTree.isEmpty() || !imCompatibilityService.findLnbtsNode(imTree)) return createEmptyHardware();

    const hardware = _.assign(
        createEmptyHardware(),
        siteModulesFactory.createModules(imTree, moduleFactories),
        lnmmesFactory.create(imTree),
        cellFactory.createCellsAndLogLinks(imTree),
        linksFactory.create(imTree),
        controllingRetLinksFactory.create(imTree)
    );
    hardware.mapByDistName = createDistNameMapping(hardware);

    hardwareSorter.sortHardware(hardware, hardwareClasses);

    const alarms = alarmsFactory.create(imTree);
    assignAlarms(hardware, alarms);
    hardware.alarms = alarms;

    assignSoftwareVersion(imTree, hardware);

    return hardware;
}

function createEmptyHardware() {
    return {
        cabinets: [],
        lnmmes: [],
        trmods: [],
        smods: [],
        extSmods: [],
        bbmods: [],
        rmods: [],
        mhas: [],
        rets: [],
        raes: [],
        phyants: [],
        cells: [],
        links: [],
        logLinks: [],
        mapByDistName: {},
        alarms: []
    };
}

function createDistNameMapping(hardware) {
    const map = {};

    function putModulesInto(modules) {
        _.forEach(modules, (module) => {
            map[module.distName] = module;
            if (module.children) putModulesInto(module.children);
            if (module.powerGroups) putModulesInto(module.powerGroups);
        });
    }

    function putModulesAndConnectors(modules) {
        _.forEach(modules, (module) => {
            map[module.distName] = module;
            if (module.children) putModulesInto(module.children);
            if (module.powerGroups) putModulesInto(module.powerGroups);
            putConnectors(module.leftConnectors);
            putConnectors(module.rightConnectors);
        });
    }

    function putConnectors(connectors) {
        _.forEach(connectors, (connector) => {
            map[connector.distName] = connector;
        });
    }

    function putCellsInto(cells) {
        putModulesInto(map, cells);
        _.forEach(cells, (cell) => {
            map[cell.lncelDistName] = cell;
        });
    }

    function putLinksInto(links) {
        _.forEach(links, (link) => {
            map[link.distName] = link;
        });
    }

    putModulesAndConnectors(hardware.trmods);
    putModulesAndConnectors(hardware.smods);
    putModulesAndConnectors(hardware.extSmods);
    putModulesAndConnectors(hardware.bbmods);
    putModulesAndConnectors(hardware.rmods);
    putModulesAndConnectors(hardware.mhas);
    putModulesAndConnectors(hardware.rets);
    putModulesAndConnectors(hardware.raes);
    putModulesAndConnectors(hardware.phyants);
    putModulesInto(hardware.lnmmes);
    putCellsInto(hardware.cells);
    putLinksInto(hardware.links);
    return map;
}

function assignAlarms(hardware, alarms) {
    _.forEach(alarms, (alarm) => {
        const alarmingResourceDN = alarm.parameters.unitInformation.alarmingResource;
        const component = findSiteComponentForDistName(hardware, alarmingResourceDN);
        if (!component) {
            if (hardware.smods.length === 1) hardware.smods[0].alarms.push(alarm);
        } else {
            component.alarms.push(alarm);
        }
    });
}

function assignSoftwareVersion(imTree, hardware) {
    const eqmNode = imCompatibilityService.findEqmLNode(imTree);
    if (!eqmNode) return;

    const storages = eqmNode
        .findByClassName(hardwareClasses.STORAGE)
        .filter(hasBuildDataChild);

    _.forEach(storages, (storage) => {
        const module = findSiteComponentForDistName(hardware, storage.distName);
        if (module) {
            const buildData = imCompatibilityService.getActiveBuildData(storage);
            if (!buildData) return;

            const parameters = buildData.getDecodedParameters();
            module.software = {
                moduleName: module.name,
                build: parameters.bldName,
                version: parameters.bldVersion
            };
        }
    });
}

function hasBuildDataChild(node) {
    return node && node.findByClassName(hardwareClasses.BUILD_DATA).length;
}

function findSiteComponentForDistName(hardware, distName) {
    const component = hardware.mapByDistName[distName];
    if (component)
        return component;
    const parent = dnService.getParent(distName);
    if (dnService.isRoot(parent)) return null;

    return findSiteComponentForDistName(hardware, parent);

}


