const SPECIAL_CHARS = [
    {char: "__Space__", substitution: " "},
    {char: "__Dot__", substitution: "."},
    {char: "__Minus__", substitution: "-"},
    {char: "__Plus__", substitution: "+"},
    {char: "__Equals__", substitution: "="},
    {char: "__Percent__", substitution: "%"},
    {char: "__Comma__", substitution: ","},
    {char: "__Asterisk__", substitution: "*"},
    {char: "__DividedBy__", substitution: "/"},
    {char: "__LeftBracket__", substitution: "["},
    {char: "__RightBracket__", substitution: "]"}
];

const TYPES = {
    INTEGER: "INTEGER",
    STRING: "STRING",
    BOOLEAN: "BOOLEAN",
    FLOAT: "FLOAT",
    ENUM: "ENUM"
};

var ITEMS = {};

function createMapping() {
    var mapping = new Map();
    mapping.set(ITEMS, []);
    return mapping;
}

function toArray(object) {
    if (!object) return [];
    return Array.isArray(object) ? object : [object];
}

function isOperation(name) {
    return !!name && name.indexOf("oper_") === 0;
}

function getProductDataProperties(productData) {
    if (!productData) return [];
    return toArray(productData["property"]);
}

function hasLteBtsMOMinOccurs(metaDefinition) {
    var productDatas = toArray(metaDefinition["productData"]);
    for (var i = 0; i < productDatas.length; i++) {
        var productData = productDatas[i];
        if (productData["@name"] === "LTE BTS") {
            var properties = getProductDataProperties(productData);
            for (var j = 0; j < properties.length; j++) {
                var property = properties[j];
                if (property["@name"] === "MO MinOccurs") return true;
            }
        }
    }
    return false;
}

function isHiddenOrHiddenNotVendorSpecific(metaDefinition) {
    var isHidden = metaDefinition['@hidden'] === "true";
    if (!isHidden) return false;

    var productDatas = toArray(metaDefinition["productData"]);
    for (var i = 0; i < productDatas.length; i++) {
        var productData = productDatas[i];
        var properties = getProductDataProperties(productData);
        for (var j = 0; j < properties.length; j++) {
            var property = properties[j];
            if (property["@name"] === "isVendorSpecific") {
                return property["@value"] === "false";
            }
        }
    }
    return true;
}

function isMandatory(metaDefinition) {
    var creation = metaDefinition["creation"];
    if (!creation) return true;
    return creation && creation["@priority"] === "mandatory";
}

function isSimpleType(metaDefinition) {
    return !!metaDefinition['simpleType'];
}

function isDeleted(metaDefinition) {
    return metaDefinition["@status"] === "deleted";
}

function isExcluded(metaDefinition) {
    //voodoo meta magic starts here!
    if (isDeleted(metaDefinition)) return true;

    if (isMandatory(metaDefinition) && hasLteBtsMOMinOccurs(metaDefinition)) return true;

    return isSimpleType(metaDefinition) && isHiddenOrHiddenNotVendorSpecific(metaDefinition);
}

function createParameters(parentTypeInfo) {
    var result = {};
    var parameters = parentTypeInfo.getParameters();
    for (var i = 0; i < parameters.length; i++) {
        var typeInfo = parameters[i];
        var name = typeInfo.getName();
        if (typeInfo.isSimple()) {
            if (typeInfo.isArray()) result[name] = [];
            else result[name] = null;
        }
        if (typeInfo.isComplex()) {
            if (typeInfo.isArray()) result[name] = [];
            else result[name] = createParameters(typeInfo);
        }
    }
    return result;
}

function createTypeInfo(metaDefinition, id) {
    var name = metaDefinition["@name"];
    var maxOccurs = metaDefinition["@maxOccurs"] || "1";
    var simpleType = metaDefinition["simpleType"];
    var complexType = metaDefinition["complexType"];
    var creation = metaDefinition["creation"] || {};
    var hidden = metaDefinition["@hidden"];
    var creationPriority = creation["@priority"];
    var baseType = isSimple() ? getBaseType() : undefined;
    var enumMapping = createEnumMapping();
    var defaultValue = getDefaultValue();
    var array = parseInt(maxOccurs, 10) > 1;
    var range = getRange();
    var parametersMapping = createParametersMapping();
    var bitmap = createBitmap();

    function hasBaseType(expected) {
        if (!isSimple()) return false;
        return simpleType["@base"] === expected;
    }

    function getBaseType() {
        if (!isSimple()) throw new Error('not a simple type');

        if (isInteger()) return TYPES.INTEGER;
        if (isString()) return TYPES.STRING;
        if (isBoolean()) return TYPES.BOOLEAN;
        if (isEnum()) return TYPES.ENUM;
        if (isFloat()) return TYPES.FLOAT;
    }

    function createParametersMapping() {
        if (!isComplex()) return createMapping();
        if (!complexType || !complexType.p) return createMapping();

        var p = toArray(complexType.p);
        var mapping = createMapping();
        var id = 1;
        for (var i = 0; i < p.length; i++) {
            var entry = p[i];
            if (!isExcluded(entry)) {
                var typeInfo = createTypeInfo(entry, id);
                mapping.set(id, typeInfo);
                mapping.set(typeInfo.getName(), typeInfo);
                mapping.get(ITEMS).push(typeInfo);
                id++;
            }
        }
        return mapping;
    }

    function isOperationInternal() {
        return isOperation(getName());
    }

    function getUnits() {
        if (isSimple()) {
            var editing = simpleType["editing"];
            if (editing) {
                return editing["@units"];
            }
        }
    }

    function getRangeProperty() {
        var editing = simpleType["editing"];
        if (!editing) {
            return simpleType["range"];
        } else {
            return editing["range"];
        }
    }

    function createRange(minString, maxString, step) {
        var result = {};
        result.min = minString ? parseInt(minString, 10) : null;
        result.max = maxString ? parseInt(maxString, 10) : null;
        result.step = step ? parseFloat(step) : null;
        return result;
    }

    function getIntegerRange() {
        var rangeProperty = getRangeProperty();
        if (!rangeProperty) return;
        return createRange(rangeProperty["@minIncl"], rangeProperty["@maxIncl"], rangeProperty["@step"]);
    }

    function getStringRange() {
        var length = simpleType["length"];
        if (length) {
            var value = length["@value"];
            if (value) return createRange(value, value);
        }

        var lengthContainer = simpleType.editing || simpleType;
        var minLength = lengthContainer["minLength"] || {};
        var maxLength = lengthContainer["maxLength"] || {};
        return createRange(minLength["@value"], maxLength["@value"]);
    }

    function getIntegerDefaultValue() {
        var editing = simpleType.editing;
        if (!editing) return;
        var defaultProperty = editing["default"] || {};
        var value = defaultProperty["@value"];
        if (value) {
            return parseInt(value, 10);
        }
    }

    function getStringDefaultValue() {
        var defaultProperty = simpleType["default"] || {};
        return defaultProperty["@value"];
    }

    function getBooleanDefaultValue() {
        var defaultProperty = simpleType["default"] || {};
        var value = defaultProperty["@value"];
        if (value) {
            return value === "true";
        }
    }

    function getEnumDefaultValue() {
        var defaultProperty = simpleType["default"] || {};
        var value = defaultProperty["@value"];
        if (value) {
            return parseInt(value, 10);
        }
    }

    function fixSpecialCharsInEnumText(enumText) {
        var result = enumText;
        for (var i = 0; i < SPECIAL_CHARS.length; i++) {
            var specialChar = SPECIAL_CHARS[i];
            result = result.split(specialChar.char).join(specialChar.substitution);
        }
        return result;
    }

    function createEnumMapping() {
        if (isEnum()) {
            var enumeration = toArray(simpleType["enumeration"]);
            var mapping = createMapping();
            for (var i = 0; i < enumeration.length; i++) {
                var item = enumeration[i];
                var value = parseInt(item["@value"], 10);
                var text = fixSpecialCharsInEnumText(item["@text"]);
                mapping.set(value, text);
                mapping.set(text, value);
                mapping.get(ITEMS).push(text);
            }
            return mapping;
        }
    }

    function createBitmap() {
        if (isBitmap()) {
            var bitmap = [];
            var bitProperty = toArray(simpleType["bit"]);
            for (var i = 0; i < bitProperty.length; i++) {
                var item = bitProperty[i];
                var index = parseInt(item["@number"], 10);
                bitmap[index] = item["@text"];
            }
            return bitmap;
        }
    }

    function getName() {
        return name;
    }

    function isArray() {
        return array;
    }

    function isSimple() {
        return metaDefinition.hasOwnProperty("simpleType");
    }

    function isComplex() {
        if (metaDefinition.hasOwnProperty("complexType")) return true;

        //case when complexType property is missing
        var firstP = metaDefinition["p"];
        if (!!firstP) {
            var secondP = firstP["p"];
            if (secondP) return true;
        }
        return false;
    }

    function isMandatoryInternal() {
        return isMandatory(metaDefinition);
    }

    function isSetBySystem() {
        return creationPriority === "Value set by the system";
    }

    function isInteger() {
        if (baseType) return baseType === TYPES.INTEGER;
        if (isEnum()) return false;
        if (isFloat()) return false;
        return hasBaseType("integer") || hasBaseType("decimal");
    }

    function isBitmap() {
        return isInteger() && !!simpleType["bit"];
    }

    function isEnum() {
        if (baseType) return baseType === TYPES.ENUM;
        return hasBaseType("integer") && !!simpleType["enumeration"];
    }

    function isString() {
        if (baseType) return baseType === TYPES.STRING;
        return hasBaseType("string");
    }

    function isObjectId() {
        return false;
    }

    function isBoolean() {
        if (baseType) return baseType === TYPES.BOOLEAN;
        return hasBaseType("boolean");
    }

    function isFloat() {
        if (baseType) return baseType === TYPES.FLOAT;
        // documented in
        // "BTS EP RP Implementation Specification: Info Model Interface specification / 5.2.3 Managed object class parameters"
        var hasValidBaseType = hasBaseType("integer") || hasBaseType("decimal");
        if (!hasValidBaseType) return false;
        var editing = simpleType["editing"];
        if (!editing) return false;
        return !!editing["@divisor"];
    }

    function getRange() {
        if (range) return range;
        if (isInteger()) return getIntegerRange();
        if (isString()) return getStringRange();
    }

    function getDefaultValue() {
        if (defaultValue) return defaultValue;
        if (isString()) return getStringDefaultValue();
        if (isInteger() || isFloat()) return getIntegerDefaultValue();
        if (isBoolean()) return getBooleanDefaultValue();
        if (isEnum()) return getEnumDefaultValue();
    }

    function getEnumText(enumValue) {
        if (isEnum()) return enumMapping.get(enumValue);
    }

    function getEnumValue(enumText) {
        if (isEnum()) return enumMapping.get(enumText);
    }

    function getEnumPossibleText() {
        if (isEnum()) return enumMapping.get(ITEMS);
    }

    function getParameterByName(name) {
        if (isComplex()) return parametersMapping.get(name);
    }

    function getParameterById(id) {
        if (isComplex()) return parametersMapping.get(id);
    }

    function getParameters() {
        if (isComplex()) return parametersMapping.get(ITEMS);
    }

    function getBitmap() {
        return bitmap;
    }

    function getId() {
        return id;
    }

    function isHidden() {
        return hidden;
    }

    function createEmptyParameters() {
        return createParameters(this);
    }

    function getCreationPriority() {
        return creationPriority;
    }

    return {
        getName: getName,
        isOperation: isOperationInternal,
        isSimple: isSimple,
        isComplex: isComplex,
        isMandatory: isMandatoryInternal,
        isSetBySystem: isSetBySystem,
        isArray: isArray,
        isInteger: isInteger,
        isBitmap: isBitmap,
        isString: isString,
        isObjectId: isObjectId,
        isBoolean: isBoolean,
        isFloat: isFloat,
        isEnum: isEnum,
        getBaseType: getBaseType,
        getEnumText: getEnumText,
        getEnumValue: getEnumValue,
        getEnumPossibleText: getEnumPossibleText,
        getDefaultValue: getDefaultValue,
        getRange: getRange,
        getParameterByName: getParameterByName,
        getParameterById: getParameterById,
        getParameters: getParameters,
        getBitmap: getBitmap,
        getId: getId,
        isHidden: isHidden,
        createEmptyParameters: createEmptyParameters,
        getUnits: getUnits,
        getCreationPriority: getCreationPriority
    };
}

module.exports = function createManagedObjectTypeInfo(metaDefinition) {
    var className = metaDefinition["@class"];
    var hidden = metaDefinition["@hidden"];
    var childManagedObjects = createChildManagedObjects();
    var typeInfos = createTypeInfos();

    function createChildManagedObjects() {
        var childProperty = metaDefinition["childManagedObject"];
        if (!childProperty) return [];

        var children = toArray(childProperty);
        var result = [];
        for (var i = 0; i < children.length; i++) {
            result[i] = children[i]["@class"];
        }
        return result;
    }

    function createTypeInfos() {
        var p = metaDefinition["p"];
        var result = {
            parameters: [], parametersMapping: {},
            operations: [], operationsMapping: {}
        };
        if (!p) return result;

        p = toArray(p);
        var id = 1;

        for (var i = 0; i < p.length; i++) {
            var entry = p[i];
            if (!isExcluded(entry, isHidden())) {
                var typeInfo = createTypeInfo(entry, id);
                if (!isOperation(entry["@name"])) {
                    result.parameters.push(typeInfo);
                    result.parametersMapping[id] = typeInfo;
                    result.parametersMapping[entry["@name"]] = typeInfo;
                } else {
                    result.operations.push(typeInfo);
                    result.operationsMapping[id] = typeInfo;
                    result.operationsMapping[entry["@name"]] = typeInfo;
                }
                id++;
            }
        }
        return result;
    }

    function getClassName() {
        return className;
    }

    function isHidden() {
        return hidden === "true";
    }

    function getChildManagedObjects() {
        return childManagedObjects;
    }

    function getParameters() {
        return typeInfos.parameters;
    }

    function getOperations() {
        return typeInfos.operations;
    }

    function getParameterByName(name) {
        return typeInfos.parametersMapping[name];
    }

    function getOperationByName(name) {
        return typeInfos.operationsMapping[name];
    }

    function getParameterById(id) {
        return typeInfos.parametersMapping[id];
    }

    function getOperationById(id) {
        return typeInfos.operationsMapping[id];
    }

    function createEmptyParameters() {
        return createParameters(this);
    }

    return {
        getClassName: getClassName,
        isHidden: isHidden,
        getChildManagedObjects: getChildManagedObjects,
        getParameters: getParameters,
        getOperations: getOperations,
        getParameterByName: getParameterByName,
        createEmptyParameters: createEmptyParameters,
        getOperationByName: getOperationByName,
        getParameterById: getParameterById,
        getOperationById: getOperationById
    }
};
