module.exports = function createManagedObjectTypeInfo(metaDefinition) {
    const className = metaDefinition['@class'];
    const id = metaDefinition['@id'] !== undefined ? parseId(metaDefinition['@id']) : undefined;
    const childManagedObjects = createChildManagedObjects();
    const typeInfos = createTypeInfos();

    return {
        getClassName: getClassName,
        isHidden: isHidden,
        getId: getId,
        getChildManagedObjects: getChildManagedObjects,
        getParameters: getParameters,
        getOperations: getOperations,
        getParameterByName: getParameterByName,
        createEmptyParameters: createEmptyParameters,
        getOperationByName: getOperationByName,
        getParameterById: getParameterById,
        getOperationById: getOperationById
    };

    function createChildManagedObjects() {
        const childProperty = metaDefinition['childManagedObject'];
        if (!childProperty) return [];

        const children = toArray(childProperty);
        return children.map(function (child) {
            return child['@class'];
        });
    }

    function createTypeInfos() {
        const parentDefinition = createParentDefinition(metaDefinition);

        const result = {
            parameters: [], parametersMapping: {},
            operations: [], operationsMapping: {}
        };

        parentDefinition.p.forEach(function (entry) {
            if (isPresent(entry)) {
                const typeInfo = createTypeInfo(entry, parentDefinition);
                const parameterId = parseId(entry['@id']);
                result.parameters.push(typeInfo);
                result.parametersMapping[parameterId] = typeInfo;
                result.parametersMapping[entry['@name']] = typeInfo;
            }
        });

        parentDefinition.action.forEach(function (entry) {
            const typeInfo = createTypeInfo(entry, createParentDefinition(entry, parentDefinition), true);
            result.operations.push(typeInfo);
            result.operationsMapping[parseId(entry['@id'])] = typeInfo;
            result.operationsMapping[entry['@name']] = typeInfo;
        });
        return result;
    }

    function getClassName() {
        return className;
    }

    function getId() {
        return id;
    }

    function isHidden() {
        return id === undefined;
    }

    function getChildManagedObjects() {
        return childManagedObjects;
    }

    function getParameters() {
        return typeInfos.parameters;
    }

    function getOperations() {
        return typeInfos.operations;
    }

    function getParameterByName(name) {
        return typeInfos.parametersMapping[name];
    }

    function getParameterById(id) {
        return typeInfos.parametersMapping[id];
    }

    function getOperationByName(name) {
        return typeInfos.operationsMapping[name];
    }

    function getOperationById(id) {
        return typeInfos.operationsMapping[id];
    }

    function createEmptyParameters() {
        return createParameters(this);
    }
};

function createParentDefinition(metaDefinition, parent) {
    const p = toArray(metaDefinition['p']);
    const action = toArray(metaDefinition['action']);
    const struct = toArray(metaDefinition['struct']);
    const enumeration = toArray(metaDefinition['enumeration']);

    const structsByName = new Map();
    struct.forEach(function (s) {
        const key = s['@name'].toLowerCase();
        structsByName.set(key, s);
    });

    const enumerationsByName = new Map();
    enumeration.forEach(function (e) {
        const key = e['@name'].toLowerCase();
        enumerationsByName.set(key, e);
    });

    return {
        parent: parent,
        p: p,
        action: action,
        struct: struct,
        structsByName: structsByName,
        enumeration: enumeration,
        enumerationsByName: enumerationsByName
    };
}

function createParameters(parentTypeInfo) {
    const result = {};
    const parameters = parentTypeInfo.getParameters();
    for (var i = 0; i < parameters.length; i++) {
        var typeInfo = parameters[i];
        var name = typeInfo.getName();
        if (typeInfo.isSimple()) {
            if (typeInfo.isArray()) result[name] = [];
            else result[name] = null;
        }
        if (typeInfo.isComplex()) {
            if (typeInfo.isArray()) result[name] = [];
            else result[name] = createParameters(typeInfo);
        }
    }
    return result;
}

function extractStructs(parentDefinition) {
    const structsByName = new Map();
    var currentParent = parentDefinition;
    while (currentParent) {
        currentParent.structsByName.forEach(function (value, key) {
            structsByName.set(key, value);
        });
        currentParent = currentParent.parent;
    }
    return structsByName;
}

function extractEnumerations(parentDefinition) {
    const enumerationsByName = new Map();
    var currentParent = parentDefinition;
    while (currentParent) {
        currentParent.enumerationsByName.forEach(function (value, key) {
            enumerationsByName.set(key, value);
        });
        currentParent = currentParent.parent;
    }
    return enumerationsByName;
}

function createTypeInfo(definition, parentDefinition, definesOperation) {
    const id = definition['@id'] !== undefined ? parseId(definition['@id']) : undefined;
    const name = definition['@name'] || '';
    const type = (definition['@type'] || '').toLowerCase();
    const recurrence = (definition['@recurrence'] || '').toLowerCase();

    const structsByName = extractStructs(parentDefinition);
    const enumerationsByName = extractEnumerations(parentDefinition);

    const complex = structsByName.has(type);
    const enumeration = enumerationsByName.has(type);
    const operation = definesOperation ? true : false;
    const baseType = isSimple() ? getBaseType() : undefined;
    const enumMapping = createEnumMapping();
    const defaultValue = getDefaultValue();
    const range = getRange();

    const typeInfo = {
        getName: getName,
        isOperation: isOperation,
        isSimple: isSimple,
        isComplex: isComplex,
        isMandatory: isMandatory,
        isSetBySystem: isSetBySystem,
        isArray: isArray,
        isInteger: isInteger,
        isBitmap: function () {
            return false;
        },
        isString: isString,
        isObjectId: isObjectId,
        isBoolean: isBoolean,
        isFloat: isFloat,
        isEnum: isEnum,
        getBaseType: getBaseType,
        getEnumText: getEnumText,
        getEnumValue: getEnumValue,
        getEnumPossibleText: getEnumPossibleText,
        getDefaultValue: getDefaultValue,
        getRange: getRange,
        getParameterByName: getParameterByName,
        getParameterById: getParameterById,
        getParameters: getParameters,
        getBitmap: function () {
        },
        getId: getId,
        isHidden: isHidden,
        createEmptyParameters: createEmptyParameters,
        getUnits: function () {
        },
        getCreationPriority: getCreationPriority
    };

    const parametersMapping = createParametersMapping();

    return typeInfo;

    function isString() {
        return type === 'string';
    }

    function isObjectId() {
        return type === 'objectid';
    }

    function isInteger() {
        return type === 'integer';
    }

    function isFloat() {
        return type === 'double';
    }

    function isBoolean() {
        return type === 'boolean';
    }

    function isEnum() {
        return enumeration;
    }

    function isSimple() {
        return !isComplex();
    }

    function isComplex() {
        return complex || operation;
    }

    function isArray() {
        return recurrence === 'repeated';
    }

    function getBaseType() {
        if (!isSimple()) throw new Error('not a simple type');

        if (isInteger()) return TYPES.INTEGER;
        if (isString()) return TYPES.STRING;
        if (isObjectId()) return TYPES.OBJECTID;
        if (isBoolean()) return TYPES.BOOLEAN;
        if (isEnum()) return TYPES.ENUM;
        if (isFloat()) return TYPES.FLOAT;
    }

    function getId() {
        return id;
    }

    function isOperation() {
        return operation;
    }

    function getName() {
        return name;
    }

    function getEnumText(enumValue) {
        if (isEnum()) return enumMapping.get(enumValue);
    }

    function getEnumValue(enumText) {
        if (isEnum()) return enumMapping.get(enumText);
    }

    function getEnumPossibleText() {
        if (isEnum()) return enumMapping.get(ITEMS);
    }

    function isMandatory() {
        return recurrence === 'mandatory' || recurrence === 'required';
    }

    function isHidden() {
        return id === undefined;
    }

    function createEmptyParameters() {
        return createParameters(this);
    }

    function getCreationPriority() {
        return recurrence;
    }

    function isSetBySystem() {
        return recurrence === 'value set by the system';
    }

    function getRange() {
        if (range) return range;
        if (isInteger()) return getIntegerRange();
        if (isString()) return getStringRange();
        if (isFloat()) return getFloatRange();
    }

    function getIntegerRange() {
        const rangeTag = definition['range'];
        if (!rangeTag) return;

        return createIntRange(rangeTag['@min'], rangeTag['@max']);
    }

    function getStringRange() {
        const lengthTag = definition['length'];
        if (!lengthTag) return;

        return createIntRange(lengthTag['@min'], lengthTag['@max']);
    }

    function getFloatRange() {
        const rangeTag = definition['range'];
        if (!rangeTag) return;

        return createFloatRange(rangeTag['@min'], rangeTag['@max']);
    }

    function createIntRange(minString, maxString) {
        const result = {};
        result.min = minString ? parseInt(minString, 10) : null;
        result.max = maxString ? parseInt(maxString, 10) : null;
        result.step = null;
        return result;
    }

    function createFloatRange(minString, maxString) {
        const result = {};
        result.min = minString ? parseFloat(minString) : null;
        result.max = maxString ? parseFloat(maxString) : null;
        result.step = null;
        return result;
    }

    function getDefaultValue() {
        if (defaultValue) return defaultValue;

        const value = definition['@default'];
        if (isString()) return value;
        if (isInteger()) return value ? parseInt(value, 10) : undefined;
        if (isFloat()) return value ? parseFloat(value) : undefined;
        if (isBoolean()) return value === 'true';
        if (isEnum()) return value ? parseInt(value, 10) : undefined;
    }

    function getParameterByName(name) {
        if (isComplex()) return parametersMapping.get(name);
    }

    function getParameterById(id) {
        if (isComplex()) return parametersMapping.get(id);
    }

    function getParameters() {
        if (isComplex()) return parametersMapping.get(ITEMS);
    }

    function createEnumMapping() {
        if (isEnum()) {
            const enumDefinition = enumerationsByName.get(type);
            const enumeration = toArray(enumDefinition['enum']);
            const mapping = createMapping();
            enumeration.forEach(function (e) {
                var value = parseInt(e['@value'], 10);
                var text = fixSpecialCharsInEnumName(e['@name']);
                mapping.set(value, text);
                mapping.set(text, value);
                mapping.get(ITEMS).push(text);
            });
            return mapping;
        }
    }

    function createParametersMapping() {
        const mapping = createMapping();
        if (!isComplex()) return mapping;

        const struct = getStruct();
        if (!struct) return mapping;

        const structDefinition = createParentDefinition(struct, parentDefinition);

        for (var i = 0; i < structDefinition.p.length; i++) {
            const entry = structDefinition.p[i];
            const parameterId = parseId(entry['@id']);
            if (isPresent(entry)) {
                const typeInfo = createTypeInfo(entry, structDefinition);
                mapping.set(parameterId, typeInfo);
                mapping.set(typeInfo.getName(), typeInfo);
                mapping.get(ITEMS).push(typeInfo);
            }
        }
        return mapping;
    }

    function getStruct() {
        if (isOperation())
            return parentDefinition;
        else
            return structsByName.get(type);
    }

    function fixSpecialCharsInEnumName(enumText) {
        var result = enumText;
        for (var i = 0; i < SPECIAL_CHARS.length; i++) {
            var specialChar = SPECIAL_CHARS[i];
            result = result.split(specialChar.char).join(specialChar.substitution);
        }
        return result;
    }
}

const SPECIAL_CHARS = [
    {char: '__Space__', substitution: ' '},
    {char: '__Dot__', substitution: '.'},
    {char: '__Minus__', substitution: '-'},
    {char: '__Plus__', substitution: '+'},
    {char: '__Equals__', substitution: '='},
    {char: '__Percent__', substitution: '%'},
    {char: '__Comma__', substitution: ','},
    {char: '__Asterisk__', substitution: '*'},
    {char: '__DividedBy__', substitution: '/'},
    {char: '__LeftBracket__', substitution: '['},
    {char: '__RightBracket__', substitution: ']'}
];

const TYPES = {
    INTEGER: 'INTEGER',
    STRING: 'STRING',
    OBJECTID: 'OBJECTID',
    BOOLEAN: 'BOOLEAN',
    FLOAT: 'FLOAT',
    ENUM: 'ENUM'
};


var ITEMS = {};

function createMapping() {
    var mapping = new Map();
    mapping.set(ITEMS, []);
    return mapping;
}

function toArray(object) {
    if (!object) return [];
    return Array.isArray(object) ? object : [object];
}

function isPresent(param) {
    return param['@id'] !== undefined;
}

function parseId(idString) {
    return parseInt(idString);
}
