const _ = require('lodash'),
    logger = require('./logger.impl'),
    ImTree = require('./imtree'),
    sequentialExecutorFactory = require('./sequentialExecutorFactory'),
    cacheFactory = require('./cacheFactory'),
    historyFactory = require('./historyFactory'),
    btsClient = require('./communication/btsClient'),
    entryTimestampProviderFactory = require('./entryTimestampProvider'),
    metaServiceFactory = require('./metaService'),
    decoderFactory = require('./decoderFactory');

module.exports = updateService;

const LOG = logger.getLogger('admin.core.updateService');

function updateService({
    socket = btsClient(),
    metaService = metaServiceFactory(),
    decoderService = decoderFactory.newInstance(metaService),
    executor = sequentialExecutorFactory.newInstance(),
    entryTimestampProvider = entryTimestampProviderFactory()
    }) {

    const history = historyFactory.newInstance();
    var historyMarker = null;

    var onlineImTree = new ImTree(decoderService);
    var offlineImTree = new ImTree(decoderService);
    const imTreeCache = cacheFactory.newInstance(1);

    const historyListeners = [];
    const imTreeListeners = [];
    const currentEntryListeners = [];
    var currentEntry;
    var listenersEnabled = true;

    socket.addListener({
        disconnected: executor.promisifyFunction(this, function (event) {
            handleSocketDisconnect(event);
        }),
        data: executor.promisifyFunction(this, handleSocketData)
    });

    const SYNC_API = {
        setMeta: setMeta,
        addMeta: addMeta,
        addImTreeListener: addImTreeListener,
        removeImTreeListener: removeImTreeListener,
        addHistoryListener: addHistoryListener,
        getHistoryEntries: getHistoryEntries,
        removeHistoryListener: removeHistoryListener,
        addCurrentEntryListener: addCurrentEntryListener,
        removeCurrentEntryListener: removeCurrentEntryListener,
        setHistoryMarker: setHistoryMarker,
        clearHistoryMarker: clearHistoryMarker,
        clearHistory: clearHistory,
        addHistoryEntry: addHistoryEntry,
        addOnlineEntry: addOnlineEntry,
        notifyHistoryListeners: createHistoryListenersNotification,
        notifyHistoryClearEntries: createHistoryClearEntriesNotification,
        mergeHistory: mergeHistory,
        mergeFirstEntries: mergeFirstEntries,
        mergeFirstEntriesOfSize: mergeFirstEntriesOfSize,
        mergeEntriesRange: mergeEntriesRange,
        dropEntriesRange: dropEntriesRange,
        setTimestampOffset: setTimestampOffset,
        getHistoryEntry: getHistoryEntry,
        getLastHistoryEntry: getLastHistoryEntry,
        disableListeners: disableListeners,
        enableListeners: enableListeners,
        snapshotEntryForId: snapshotEntryForId,
        getHistorySize: getHistorySize,
        createNotifier: createNotifier,
        clearUpdateServiceState: clearUpdateServiceState
    };

    const PUBLIC_API = {
        setMeta: executor.promisifyFunction(this, setMeta),
        addImTreeListener: executor.promisifyFunction(this, addImTreeListener),
        removeImTreeListener: executor.promisifyFunction(this, removeImTreeListener),
        addHistoryListener: executor.promisifyFunction(this, addHistoryListener),
        getHistoryEntries: executor.promisifyFunction(this, getHistoryEntries),
        removeHistoryListener: executor.promisifyFunction(this, removeHistoryListener),
        addCurrentEntryListener: executor.promisifyFunction(this, addCurrentEntryListener),
        removeCurrentEntryListener: executor.promisifyFunction(this, removeCurrentEntryListener),
        setHistoryMarker: executor.promisifyFunction(this, setHistoryMarker),
        clearHistoryMarker: executor.promisifyFunction(this, clearHistoryMarker),
        clearHistory: executor.promisifyFunction(this, clearHistory),
        addHistoryEntry: executor.promisifyFunction(this, addHistoryEntry),
        addOnlineEntry: executor.promisifyFunction(this, addOnlineEntry),
        notifyHistoryListeners: executor.promisifyFunction(this, createHistoryListenersNotification),
        notifyHistoryClearEntries: executor.promisifyFunction(this, createHistoryClearEntriesNotification),
        mergeHistory: executor.promisifyFunction(this, mergeHistory),
        mergeFirstEntries: executor.promisifyFunction(this, mergeFirstEntries),
        mergeFirstEntriesOfSize: executor.promisifyFunction(this, mergeFirstEntriesOfSize),
        mergeEntriesRange: executor.promisifyFunction(this, mergeEntriesRange),
        dropEntriesRange: executor.promisifyFunction(this, dropEntriesRange),
        setTimestampOffset: executor.promisifyFunction(this, setTimestampOffset),
        performInAsyncQueue: executor.promisifyFunction(this, performInAsyncQueue),
        clearUpdateServiceState: executor.promisifyFunction(this, clearUpdateServiceState),
        addMeta: addMeta,
        getHistoryEntry: getHistoryEntry,
        getLastHistoryEntry: getLastHistoryEntry,
        disableListeners: disableListeners,
        enableListeners: enableListeners,
        snapshotEntryForId: snapshotEntryForId,
        getHistorySize: getHistorySize,
        createNotifier: createNotifier
    };

    // visible for testing
    const INTERNAL = {
        pushData: handleSocketData,
        pushUpdate: pushUpdate,
        pushRemove: pushRemove,
        pushDisconnect: handleSocketDisconnect,
        pushDataAsync: executor.promisifyFunction(this, handleSocketData),
        pushDisconnectAsync: executor.promisifyFunction(this, handleSocketDisconnect),
        createUpdateEntry: createUpdateEntry,
        entriesFinder: entriesFinder,
        clear: clearHistory,
        getTimestampOffset: getTimestampOffset
    };

    return _.assign({SYNC_API: SYNC_API}, PUBLIC_API, INTERNAL);

    /**
     * performCallback should only be used with updateService SYNC_API !
     * */
    function performInAsyncQueue(callback) {
        return callback();
    }

    function setTimestampOffset(newOffset) {
        entryTimestampProvider.setOffset(newOffset);
    }

    function getTimestampOffset() {
        return entryTimestampProvider.getOffset();
    }

    function createHistoryListenersNotification() {
        createNotifier()
            .historyNewEntries()
            .notify();
    }

    function createHistoryClearEntriesNotification() {
        createNotifier()
            .historyClearEntries()
            .notify();
    }

    function disableListeners() {
        listenersEnabled = false;
    }

    function enableListeners() {
        listenersEnabled = true;
    }

    function updateImTree(imTree, entry) {
        if (!entry.change) return;

        const change = entry.change;
        imTree.removeByDistNames(change.removed);
        for (var i = 0; i < change.updated.length; i++) {
            imTree.merge(change.updated[i], entry.metaUid);
        }
    }

    function getOrAdjustOfflineImTree() {
        const imTree = imTreeCache.get(historyMarker);
        if (imTree) return imTree;
        imTreeCache.set(historyMarker, adjustImTree(offlineImTree, historyMarker.id));
        return offlineImTree;
    }

    function adjustImTree(imTree, entryId) {
        const map = history.snapshotForEntryId(entryId);
        var newDistNames = new Set();
        var oldDistNames = imTree.distNames();
        map.forEach(function (managedObject, distName) {
            newDistNames.add(distName);
        });
        oldDistNames.forEach(function (distName) {
            if (!newDistNames.has(distName)) {
                imTree.removeByDistName(distName);
            }
        });
        map.forEach(function (managedObject, distName) {
            imTree.mergeNode(distName, null, managedObject.payload, managedObject.metaUid);
        });
        return imTree;
    }

    function isInHistoryMode() {
        return historyMarker !== null;
    }

    function clearHistory() {
        history.clear();
        clearUpdateServiceState();
    }

    function clearUpdateServiceState(onlineHistoryMode = false) {
        var notifier = createNotifier()
            .historyClearEntries(onlineHistoryMode);
        if (historyMarker) {
            historyMarker = null;
            notifier.historyMarkerWasCleared();
        }

        onlineImTree = new ImTree(decoderService);
        offlineImTree = new ImTree(decoderService);
        currentEntry = null;
        imTreeCache.clear();
        notifier.imTreeEvents()
            .notify();
    }

    function pushUpdate(distName, payload) {
        return handleSocketData({
            removedDistNames: [],
            managedObjects: [{distName: distName, payload: payload}]
        });
    }

    function pushRemove(distNameToRemove) {
        var removes = Array.isArray(distNameToRemove) ? distNameToRemove : [distNameToRemove];
        return handleSocketData({
            removedDistNames: removes,
            managedObjects: []
        });
    }

    /**
     * data = { managedObjects: [managedObject], removedDistNames: [String]}
     * managedObject = { distName: String, payload: ArrayBuffer }
     * */
    function handleSocketData(data) {
        const updates = data.managedObjects;
        const removed = data.removedDistNames;
        const entry = createUpdateEntry(updates, removed, data.metaUid);
        entry.timestamp = data.timestamp || entry.timestamp;
        const updatedDistNames = {
            updated: updates.map(function (update) {
                return update.distName;
            })
        };
        history.addEntry(entry);
        var notifier = createNotifier();
        updateImTree(onlineImTree, entry);
        notifier.historyNewEntry(entry)
            .onlineImTree(updatedDistNames);
        if (!isInHistoryMode()) {
            notifier.completeImTree(onlineImTree)
                .updateOfImTree(updates, removed)
                .currentEntryChange(entry);
        }
        notifier.notify();
        return entry;
    }

    function isResetFromBTS(event) {
        return event.source === 'BTS' && history.size() > 0 && !isLastEntryReset();
    }

    function handleSocketDisconnect(event) {
        if (!isResetFromBTS(event)) return;

        const entry = createResetEntry(event.timestamp);
        history.addEntry(entry);
        onlineImTree = new ImTree(decoderService);
        var notifier = createNotifier()
            .historyNewEntry(entry)
            .onlineImTree();
        if (!isInHistoryMode()) {
            notifier.newImTree(onlineImTree)
                .completeImTree(onlineImTree)
                .currentEntryChange(entry);
        }
        notifier.notify();
        return entry;
    }

    function isLastEntryReset() {
        return !!history.getLastEntry().reset;
    }

    function setMeta(metaBuffer) {
        return metaService.setMeta(metaBuffer);
    }

    function addMeta(metaBuffer) {
        return metaService.addMeta(metaBuffer);
    }

    function removeListener(listeners, listener) {
        var index = listeners.indexOf(listener);
        if (index > -1) {
            listeners.splice(index, 1);
        }
    }

    function createUpdateEntry(updates, removed, metaUid) {
        return {
            timestamp: entryTimestampProvider.getTimestamp(),
            numOfChanges: updates.length + removed.length,
            change: {updated: updates, removed: removed},
            metaUid: metaUid || metaService.getLastMetaUid()
        };
    }

    function createResetEntry(timestamp) {
        return {
            timestamp: timestamp || entryTimestampProvider.getTimestamp(),
            numOfChanges: 0,
            reset: true
        };
    }

    function addImTreeListener(listener) {
        listener.completeImTree = exceptionCatcher(listener, listener.completeImTree || _.noop);
        listener.newImTree = exceptionCatcher(listener, listener.newImTree || _.noop);
        listener.updateOfImTree = exceptionCatcher(listener, listener.updateOfImTree || _.noop);
        listener.onlineImTree = exceptionCatcher(listener, listener.onlineImTree || _.noop);

        imTreeListeners.push(listener);
        createNotifier()
            .singleImTreeListener(listener)
            .imTreeEvents()
            .notify();
    }

    function addHistoryListener(listener) {
        listener.newEntry = exceptionCatcher(listener, listener.newEntry || _.noop);
        listener.newEntries = exceptionCatcher(listener, listener.newEntries || _.noop);
        listener.clearEntries = exceptionCatcher(listener, listener.clearEntries || _.noop);
        listener.historyMarkerWasSet = exceptionCatcher(listener, listener.historyMarkerWasSet || _.noop);
        listener.historyMarkerWasCleared = exceptionCatcher(listener, listener.historyMarkerWasCleared || _.noop);

        historyListeners.push(listener);

        var notifier = createNotifier()
            .singleHistoryListener(listener)
            .historyNewEntries();
        if (isInHistoryMode()) {
            notifier.historyMarkerWasSet(historyMarker);
        }
        notifier.notify();
    }

    function addCurrentEntryListener(listener) {
        listener.changeEntry = exceptionCatcher(listener, listener.changeEntry || _.noop);

        currentEntryListeners.push(listener);

        createNotifier()
            .singleCurrentEntryListener(listener)
            .currentEntryChange(currentEntry)
            .notify();
    }

    function removeImTreeListener(listener) {
        removeListener(imTreeListeners, listener);
    }

    function removeHistoryListener(listener) {
        removeListener(historyListeners, listener);
    }

    function removeCurrentEntryListener(listener) {
        removeListener(currentEntryListeners, listener);
    }

    function setHistoryMarker(entry) {
        if (!setHistoryMarkerWithoutNotification(entry)) return;
        notifyAboutSettingHistoryMarker();
        return historyMarker;
    }

    function notifyAboutSettingHistoryMarker(notifier) {
        (notifier || createNotifier())
            .historyMarkerWasSet()
            .imTreeEvents()
            .currentEntryChange(historyMarker)
            .notify();
    }

    function setHistoryMarkerWithoutNotification(entry) {
        if (entry) {
            historyMarker = entry;
        } else if (history.size() > 0) {
            historyMarker = history.getLastEntry();
        } else {
            return;
        }

        return historyMarker;
    }

    function clearHistoryMarker() {
        if (!historyMarker) return;

        historyMarker = null;

        createNotifier()
            .historyMarkerWasCleared()
            .imTreeEvents()
            .currentEntryChange(history.getLastEntry())
            .notify();
    }

    function addHistoryEntry(entry) {
        entry.metaUid = metaService.getLastMetaUid();
        history.addEntry(entry);
        var notifier = createNotifier();
        notifier.notify();
    }

    function addOnlineEntry(entry) {
        history.addEntry(entry);
        updateImTree(onlineImTree, entry);
        var notifier = createNotifier();
        notifier.notify();
    }

    function getHistoryEntry(entryId) {
        return history.getEntry(entryId);
    }

    function getLastHistoryEntry(entryId) {
        return history.getLastEntry(entryId);
    }

    function getHistoryEntries() {
        return history.getEntries();
    }

    function snapshotEntryForId(id) {
        return history.snapshotEntryForId(id);
    }

    function getHistorySize() {
        return history.size();
    }

    function mergeHistory() {
        if (history.size() <= 1) return;

        const wasInHistoryMode = isInHistoryMode();
        clearUpdateServiceState();

        var mergedEntry = history.mergeEntries();

        if (wasInHistoryMode) historyMarker = mergedEntry;

        updateImTree(onlineImTree, mergedEntry);

        const isOnline = !isInHistoryMode();
        const imTree = isOnline ? onlineImTree : getOrAdjustOfflineImTree();
        var notifier = createNotifier()
            .newImTree(imTree)
            .completeImTree(imTree)
            .onlineImTree()
            .historyNewEntry(mergedEntry)
            .currentEntryChange(mergedEntry);
        if (wasInHistoryMode) notifier.historyMarkerWasSet();
        notifier.notify();
    }

    function mergeFirstEntries(numOfEntries) {
        if (numOfEntries < 2) return;

        const prevHistoryMarkerId = historyMarker && historyMarker.id;
        imTreeCache.clear();
        createNotifier()
            .historyMarkerWasCleared()
            .historyClearEntries()
            .notify();

        history.mergeFirstEntries(numOfEntries);

        if (prevHistoryMarkerId !== null) {
            setHistoryMarkerWithoutNotification(chooseNewHistoryMarker());
            var notifier = createNotifier()
                .historyNewEntries();
            notifyAboutSettingHistoryMarker(notifier);
        } else {
            createNotifier()
                .imTreeEvents()
                .historyNewEntries()
                .currentEntryChange(history.getLastEntry())
                .notify();
        }

        function chooseNewHistoryMarker() {
            var newHistoryMarkerId;
            if (prevHistoryMarkerId < numOfEntries) newHistoryMarkerId = 0;
            else newHistoryMarkerId = prevHistoryMarkerId - (numOfEntries - 1);

            return history.getEntry(newHistoryMarkerId);
        }
    }

    function mergeEntriesRange(fromId, toId, shouldNotify = true) {
        if (fromId < 0 || toId > history.size() || toId - fromId < 2) return;

        const prevHistoryMarkerId = historyMarker && historyMarker.id;
        imTreeCache.clear();
        if (shouldNotify) {
            createNotifier()
                .historyMarkerWasCleared()
                .historyClearEntries()
                .notify();
        }

        history.mergeEntriesRange(fromId, toId);

        if (prevHistoryMarkerId !== null) {
            setHistoryMarkerWithoutNotification(chooseNewHistoryMarker());
            var notifier = createNotifier()
                .historyNewEntries();
            notifyAboutSettingHistoryMarker(notifier);
        } else if (shouldNotify) {
            createNotifier()
                .imTreeEvents()
                .historyNewEntries()
                .currentEntryChange(history.getLastEntry())
                .notify();
        }

        function chooseNewHistoryMarker() {
            var newHistoryMarkerId;
            if (prevHistoryMarkerId < fromId) newHistoryMarkerId = prevHistoryMarkerId;
            else if (prevHistoryMarkerId < toId) newHistoryMarkerId = fromId;
            else newHistoryMarkerId = prevHistoryMarkerId - (toId - fromId - 1);

            return history.getEntry(newHistoryMarkerId);
        }
    }

    function mergeFirstEntriesOfSize(sizeInBytes) {
        //discard first entry, because after first merge its the biggest
        //and we want to merge large number of small entries
        var firstEntry = 1;
        var numOfEntriesOfGivenSize = entriesFinder()
            .skipFirst(firstEntry)
            .findOfSize(sizeInBytes);
        mergeFirstEntries(numOfEntriesOfGivenSize + firstEntry);
    }

    function dropEntriesRange(fromId, toId) {
        history.dropEntriesRange(fromId, toId);
    }

    function entriesFinder() {
        var numOfSkipEntries = 0;

        var instance = {
            skipFirst: skipFirst,
            findOfSize: findOfSize
        };

        function skipFirst(n) {
            numOfSkipEntries = n;
            return instance;
        }

        function findOfSize(sizeInBytes) {
            const iterator = history.getEntriesIterator();
            var item = iterator.next();
            if (item.done) return 0;

            for (var i = 0; i < numOfSkipEntries; i++)
                item = iterator.next();

            var size = 0;
            var count = 0;
            while (!item.done) {
                const entry = item.value;
                size += entry.estimatedSizeInBytes;
                item = iterator.next();
                count++;

                if (size >= sizeInBytes) return count;
            }
            return count;
        }

        return instance;
    }

    function exceptionCatcher(thisArg, originalFunction) {
        return function () {
            try {
                return originalFunction.apply(thisArg, arguments);
            } catch (e) {
                LOG.error(e);
            }
        };
    }

    function createNotifier() {

        const notifications = {};

        var notifier = {
            historyNewEntry: historyNewEntryNotification,
            historyNewEntries: historyNewEntriesNotification,
            historyClearEntries: historyClearEntriesNotification,
            historyMarkerWasCleared: historyMarkerWasClearedNotification,
            historyMarkerWasSet: historyMarkerWasSetNotification,
            singleHistoryListener: singleHistoryListener,
            onlineImTree: onlineImTreeNotification,
            completeImTree: completeImTreeNotification,
            updateOfImTree: updateOfImTreeNotification,
            newImTree: newImTreeNotification,
            imTreeEvents: imTreeEventsNotification,
            singleImTreeListener: singleImTreeListener,
            currentEntryChange: currentEntryChangeNotification,
            singleCurrentEntryListener: singleCurrentEntryListener,
            notify: notify
        };

        return notifier;

        function historyNewEntryNotification(entry) {
            notifications.history = notifications.history || {};
            notifications.history.newEntry = {params: entry};
            return notifier;
        }

        function historyNewEntriesNotification() {
            notifications.history = notifications.history || {};
            notifications.history.newEntries = {};
            return notifier;
        }

        function historyClearEntriesNotification(historyMode) {
            notifications.history = notifications.history || {};
            notifications.history.clearEntries = {params: historyMode};
            return notifier;
        }

        function historyMarkerWasClearedNotification() {
            notifications.history = notifications.history || {};
            notifications.history.markerCleared = {};
            return notifier;
        }

        function historyMarkerWasSetNotification() {
            notifications.history = notifications.history || {};
            notifications.history.markerSet = {};
            return notifier;
        }

        function singleHistoryListener(listener) {
            notifications.history = notifications.history || {};
            notifications.history.listener = listener;
            return notifier;
        }

        function onlineImTreeNotification(updatedDistNames) {
            notifications.imTree = notifications.imTree || {};
            notifications.imTree.online = {params: updatedDistNames};
            return notifier;
        }

        function completeImTreeNotification(imTree) {
            notifications.imTree = notifications.imTree || {};
            notifications.imTree.complete = {params: imTree};
            return notifier;
        }

        function updateOfImTreeNotification(updates, removed) {
            notifications.imTree = notifications.imTree || {};
            notifications.imTree.update = {params: [updates, removed]};
            return notifier;
        }

        function newImTreeNotification(imTree) {
            notifications.imTree = notifications.imTree || {};
            notifications.imTree.new = {params: imTree};
            return notifier;
        }

        function imTreeEventsNotification() {
            if (!hasListenersToProcess()) return notifier;
            if (!listenersEnabled) return notifier;

            const isOnline = !isInHistoryMode();
            const imTree = isOnline ? onlineImTree : getOrAdjustOfflineImTree();
            newImTreeNotification(imTree);
            completeImTreeNotification(imTree);
            if (isOnline) onlineImTreeNotification();
            return notifier;

            function hasListenersToProcess() {
                return !_.isEmpty(imTreeListeners) ||
                    (notifications.imTree && notifications.imTree.listener);
            }
        }

        function singleImTreeListener(listener) {
            if (!listenersEnabled) return notifier;

            notifications.imTree = notifications.imTree || {};
            notifications.imTree.listener = listener;
            return notifier;
        }

        function currentEntryChangeNotification(entry) {
            notifications.currentEntry = notifications.currentEntry || {};
            notifications.currentEntry.change = {params: entry};
            return notifier;
        }

        function singleCurrentEntryListener(listener) {
            notifications.currentEntry = notifications.currentEntry || {};
            notifications.currentEntry.listener = listener;
            return notifier;
        }

        function notify() {
            if (notifications.history) notifyHistoryListeners();
            if (notifications.imTree) notifyImTreeListeners();
            if (notifications.currentEntry) notifyCurrentEntryListeners();
        }

        function notifyHistoryListeners() {
            if (notifications.history.newEntry) {
                notifyHistoryNewEntry(notifications.history.newEntry.params);
            }
            if (notifications.history.newEntries) {
                if (notifications.history.listener) {
                    notifyListenerHistoryNewEntries(notifications.history.listener);
                } else {
                    notifyHistoryNewEntries();
                }
            }
            if (notifications.history.clearEntries) {
                notifyHistoryClearEntries(notifications.history.clearEntries.params);
            }
            if (notifications.history.markerCleared) {
                notifyHistoryMarkerWasCleared();
            }
            if (notifications.history.markerSet) {
                if (notifications.history.listener) {
                    notifyListenerHistoryMarkerWasSet(notifications.history.listener);
                } else {
                    notifyHistoryMarkerWasSet();
                }
            }
        }

        function notifyImTreeListeners() {
            if (notifications.imTree.new) {
                if (notifications.imTree.listener) {
                    notifyListenerNewImTree(
                        notifications.imTree.listener,
                        notifications.imTree.new.params
                    );
                } else {
                    notifyNewImTree(notifications.imTree.new.params);
                }
            }
            if (notifications.imTree.complete) {
                if (notifications.imTree.listener) {
                    notifyListenerCompleteImTree(
                        notifications.imTree.listener,
                        notifications.imTree.complete.params
                    );
                } else {
                    notifyCompleteImTree(notifications.imTree.complete.params);
                }
            }
            if (notifications.imTree.online) {
                if (notifications.imTree.listener) {
                    notifyListenerOnlineImTree(
                        notifications.imTree.listener,
                        notifications.imTree.online.params
                    );
                } else {
                    notifyOnlineImTree(notifications.imTree.online.params);
                }
            }
            if (notifications.imTree.update) {
                notifyUpdateOfImTree(
                    notifications.imTree.update.params[0],
                    notifications.imTree.update.params[1]
                );
            }
        }

        function notifyCurrentEntryListeners() {
            if (notifications.currentEntry.change) {
                if (notifications.currentEntry.listener) {
                    notifyListenerCurrentEntryChange(
                        notifications.currentEntry.listener,
                        notifications.currentEntry.change.params
                    );
                } else {
                    notifyCurrentEntryChange(notifications.currentEntry.change.params);
                }
            }
        }

        function notifyHistoryNewEntry(entry) {
            if (!listenersEnabled) return;

            _.forEach(historyListeners, function (listener) {
                listener.newEntry(entry);
            });
        }

        function notifyOnlineImTree(distNames) {
            if (!listenersEnabled) return;

            _.forEach(imTreeListeners, function (listener) {
                notifyListenerOnlineImTree(listener, distNames)
            });
        }

        function notifyCompleteImTree(imTree) {
            if (!listenersEnabled) return;

            _.forEach(imTreeListeners, function (listener) {
                notifyListenerCompleteImTree(listener, imTree)
            });
        }

        function notifyUpdateOfImTree(updates, removed) {
            if (!listenersEnabled) return;

            _.forEach(imTreeListeners, function (listener) {
                listener.updateOfImTree(updates, removed);
            });
        }

        function notifyCurrentEntryChange(entry) {
            if (!listenersEnabled) return;

            currentEntry = entry;
            _.forEach(currentEntryListeners, function (listener) {
                listener.changeEntry(entry);
            });
        }

        function notifyNewImTree(imTree) {
            if (!listenersEnabled) return;

            _.forEach(imTreeListeners, function (listener) {
                notifyListenerNewImTree(listener, imTree);
            });
        }

        function notifyHistoryClearEntries(historyMode) {
            if (!listenersEnabled) return;

            _.forEach(historyListeners, function (listener) {
                listener.clearEntries(historyMode);
            });
        }

        function notifyHistoryMarkerWasCleared() {
            if (!listenersEnabled) return;

            _.forEach(historyListeners, function (listener) {
                listener.historyMarkerWasCleared();
            });
        }

        function notifyListenerCompleteImTree(listener, imTree) {
            listener.completeImTree(imTree);
        }

        function notifyListenerNewImTree(listener, imTree) {
            listener.newImTree(imTree);
        }

        function notifyListenerOnlineImTree(listener, distNames) {
            listener.onlineImTree(onlineImTree, distNames);
        }

        function notifyHistoryNewEntries() {
            if (!listenersEnabled) return;

            _.forEach(historyListeners, function (listener) {
                listener.newEntries(history.getEntries());
            });
        }

        function notifyListenerHistoryNewEntries(listener) {
            if (!listenersEnabled) return;

            listener.newEntries(history.getEntries());
        }

        function notifyListenerHistoryMarkerWasSet(listener) {
            listener.historyMarkerWasSet(historyMarker);
        }

        function notifyHistoryMarkerWasSet() {
            if (!listenersEnabled) return;

            _.forEach(historyListeners, function (listener) {
                listener.historyMarkerWasSet(historyMarker);
            });
        }

        function notifyListenerCurrentEntryChange(listener, entry) {
            if (listenersEnabled) {
                listener.changeEntry(entry);
            }
        }

    }
}
