const SEPARATOR = '/';
module.exports = Object.freeze({
    distNameUpToClassName: distNameUpToClassName,
    distNameStartingFromClassName: distNameStartingFromClassName,
    getParent: getParent,
    getName: getName,
    getId: getId,
    getParsedName: getParsedName,
    getClassName: getClassName,
    getParentClassName: getParentClassName,
    getDistNamePath: getDistNamePath,
    getAncestor: getAncestor,
    isRoot: isRoot,
    isSubtree: isSubtree,
    compareDistNames: compareDistNames,
    SEPARATOR
});

/**
 * return parent distName or root if at root
 */
function getParent(distName) {
    const last = distName.lastIndexOf(SEPARATOR);
    if (last > 0 && last < (distName.length - 1)) {
        return distName.slice(0, last);
    } else {
        return distName;
    }
}

function splitName(nodeName) {
    var splitOnIndex = nodeName.indexOf('-');
    return {
        className: nodeName.substring(0, splitOnIndex),
        id: nodeName.substring(splitOnIndex + 1)
    };
}

/**
 * Example:
 * ("/MRBTS/EQM_L-1/RMOD_L-1", "EQM_L") -> "/MRBTS/EQM_L-1"
 * ("/MRBTS/EQM_L-1/RMOD_L-1", "XXX") -> undefined
 * ("/A-1/B-1/A-1", "A-1") -> "/A-1/B-1/A-1"
 */
function distNameUpToClassName(distName, className) {
    var classIndex = distName.lastIndexOf(className);
    if (classIndex >= 0) {
        var endIndex = distName.indexOf(SEPARATOR, classIndex);
        endIndex = endIndex < 0 ? distName.length : endIndex;
        return distName.substring(0, endIndex);
    }
}

/**
 * Example:
 * ("/MRBTS/EQM_L-1/RMOD_L-1", "EQM_L") -> "EQM_L-1/RMOD_L-1"
 * ("/MRBTS/EQM_L-1/RMOD_L-1", "XXX") -> undefined
 * ("/A-1/B-1/A-1", "A-1") -> "/A-1/B-1/A-1"
 */
function distNameStartingFromClassName(distName, className) {
    var classIndex = distName.indexOf(className);
    if (classIndex >= 0) {
        return distName.substring(classIndex, distName.length);
    }
}

function getName(distName) {
    var last = distName.lastIndexOf(SEPARATOR);
    return distName.substring(last + 1);
}

function getId(distName) {
    var distNameLastPart = distName.split(SEPARATOR).slice(-1)[0];
    return parseInt(distNameLastPart.substring(distNameLastPart.indexOf('-') + 1), 10);
}

function isRoot(distName) {
    return distName === getParent(distName);
}

function getClassName(distName) {
    return splitName(getName(distName)).className;
}

function getParentClassName(distName) {
    return getClassName(getParent(distName));
}

function getParsedName(distName) {
    return splitName(getName(distName));
}

function getDistNamePath(distName) {
    return distName.split(SEPARATOR);
}

function getAncestor(distName, level) {
    var ancestor = distName;
    for (var i = 0; i < level; i++) {
        ancestor = getParent(ancestor);
    }
    return ancestor;
}

function isSubtree(parent, child) {
    return (child === parent) || (child.startsWith(parent + SEPARATOR));
}

function compareDistNames(distName1, distName2) {
    if (isRoot(distName1) && isRoot(distName2))
        return compareNodeNames(distName1, distName2);

    return compareDistNames(getParent(distName1), getParent(distName2)) ||
        compareNodeNames(distName1, distName2);
}

function compareIds(distName1, distName2) {
    var diff = getId(distName1) - getId(distName2);
    if (diff === 0 || isNaN(diff)) return diff;

    return diff > 0 ? 1 : -1;
}

function compareNodeNames(distName1, distName2) {
    return getClassName(distName1).localeCompare(getClassName(distName2)) ||
        compareIds(distName1, distName2);
}
