const distNameUtils = require('./distNameUtils');

module.exports = Object.freeze({
    getHardwareDistName,
    findEqmLNode,
    findEqmANode,
    findHwtopLNode,
    findCabinetLNode,
    findHwtopANode,
    findBbtopLNode,
    findNodeA,
    findCurrentNpNode,
    findMrbtsANode,
    findLnbtsNode,
    findLnbtsANode,
    findLncelANodes,
    findLncelNodes,
    findFtmNode,
    findCranConfig,
    getLnmmePrimaryIpAddress,
    getLnmmeSecondaryIpAddress,
    adaptCablnkLinkMode,
    findHwLinks,
    findBtsConfNode,
    getActiveBuildData
});

function findCabinetLNode(imTree) {
    return findEqmLNode(imTree).findChildByClassName('CABINET_L');
}

function getHardwareDistName(imTree) {
    var node = imTree.findByDistName('/MRBTS-1/RAT-1/BTS_L-1');
    if (node)
        return '/MRBTS-1/RAT-1/BTS_L-1';
    return '/MRBTS-1/RAT-1';
}

function findNodeA(nodeL) {
    var npNode = findCurrentNpNode(nodeL.root);
    if (!npNode) return null;

    var className = distNameUtils.getClassName(nodeL.distName);
    className = className.replace('_L', '');
    className += '_A';

    var nodes = npNode.findByClassName(className);

    return nodes
        && nodes.find(function (node) {
            return node.getDecodedParameters().realDN === nodeL.distName;
        });
}

function findCurrentNpNode(imTree) {
    var btsConfNode = findBtsConfNode(imTree);
    if (!btsConfNode) return null;

    var currentNp = btsConfNode.getDecodedParameters().currentNp;
    return imTree.findByDistName(currentNp);
}

function findBtsConfNode(imTree) {
    return imTree.findByDistName(getHardwareDistName(imTree) + '/BTS_CONF-1');
}

function findEqmLNode(imTree) {
    return imTree.findByDistName(getHardwareDistName(imTree) + '/EQM_L-1');
}

function findHwtopLNode(imTree) {
    return imTree.findByDistName(getHardwareDistName(imTree) + '/HWTOP_L-1');
}

function findBbtopLNode(imTree) {
    return imTree.findByDistName(getHardwareDistName(imTree) + '/BBTOP_L-1');
}

function findLnbtsNode(imTree) {
    return imTree.findByDistName(getHardwareDistName(imTree) + '/LNBTS-1');
}

function findMrbtsANode(imTree) {
    var activeNp = findCurrentNpNode(imTree);
    var scfNode = activeNp && activeNp.findChildByName('SCF-1');
    return scfNode && scfNode.findChildByClassName('MRBTS_A');
}

function findLnbtsANode(imTree) {
    var mrbtsA = findMrbtsANode(imTree);
    return mrbtsA && mrbtsA.findChildByClassName('LNBTS_A');
}

function findEqmANode(imTree) {
    var mrbtsA = findMrbtsANode(imTree);
    return mrbtsA && mrbtsA.findChildByClassName('EQM_A');
}

function findHwtopANode(imTree) {
    var mrbtsA = findMrbtsANode(imTree);
    return mrbtsA && mrbtsA.findChildByClassName('HWTOP_A');
}

function findLncelANodes(imTree) {
    var mrbtsANode = findMrbtsANode(imTree);
    if (!mrbtsANode) return [];

    var lnbtsANode = mrbtsANode.findChildByClassName('LNBTS_A');
    if (!lnbtsANode) return [];

    return lnbtsANode.findChildrenByClassName('LNCEL_A');
}

function findLncelNodes(imTree) {
    var lnbts = findLnbtsNode(imTree);
    if (!lnbts) return [];

    return lnbts.findChildrenByClassName('LNCEL');
}

function findFtmNode(imTree) {
    const FTM_DN = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1/FTM-1';
    return imTree.findByDistName(FTM_DN);
}

function findCranConfig(imTree) {
    const parent = imTree.findByDistName(getHardwareDistName(imTree) + '/FUNC_L-1');
    return parent && parent.findChildByClassName('CRAN_CONFIG');
}

function getLnmmePrimaryIpAddress(lnmmeParams) {
    var ipAddrPrim = lnmmeParams.ipAddrPrim;
    var address;
    var state;
    if (ipAddrPrim && ipAddrPrim.hasOwnProperty('ip_value')) {
        address = ipAddrPrim.ip_value;
        state = ipAddrPrim.faultyIPAddress;
    } else {
        address = ipAddrPrim;
        state = lnmmeParams.ipAddrPrimFromFaultReq;
    }
    return {address: address, state: state};
}

function getLnmmeSecondaryIpAddress(lnmmeParams) {
    var ipAddrSec = lnmmeParams.ipAddrSec;
    var address;
    var state;
    if (ipAddrSec && ipAddrSec.hasOwnProperty('ip_value')) {
        address = ipAddrSec.ip_value;
        state = ipAddrSec.faultyIPAddress;
    } else {
        address = ipAddrSec;
        state = lnmmeParams.ipAddrSecFromFaultReq;
    }
    return {address: address, state: state};
}

function adaptCablnkLinkMode(linkMode) {
    if (typeof linkMode === 'string') return {configuration: linkMode.indexOf('Configuration') !== -1};

    return linkMode;
}

function findHwLinks(hwtop) {
    if (!hwtop) return [];

    const linkNodes = hwtop.findChildrenByClassName('HWLINK_L');
    return linkNodes.concat(hwtop.findChildrenByClassName('HWLINK_NG_L'));
}

function getActiveBuildData(storageNode) {
    const BUILD_DATA_L = 'BUILD_DATA_L';
    const fileSystems = storageNode.findByClassName('FILE_SYSTEM_L');
    if (fileSystems.length === 1) return fileSystems[0].findByClassName(BUILD_DATA_L)[0];

    const activeFileSystem = fileSystems.find(fileSystem => fileSystem.getDecodedParameters().bootPurpose === 'Active');
    return activeFileSystem && activeFileSystem.findByClassName(BUILD_DATA_L)[0];
}
