module.exports = {
    mergeParameters,
    createParametersFinder,
    createQsfpLinkDistName,
    isQsfpLinkDistName
};

const distNameService = require('./distNameUtils'),
    _ = require('lodash');

const QSFP_LINK_DIST_NAME = {
    START: '[',
    END: ']',
    SEPARATOR: ', '
};

const QFSP_PARAMETERS = {
    STATE_INFO: 'stateInfo',
    DESTINATION_DIST_NAME: 'destinationDistName',
    SOURCE_DIST_NAME: 'sourceDistName'
};

const STATE_INFO = {
    OPERATIONAL_STATE: 'operationalState',
    VALUES: {
        ENABLED: 'Enabled',
        DISABLED: 'Disabled'
    }
};

function createQsfpLinkDistName(qsfpLinks, qsfpName) {
    var links = qsfpLinks[qsfpName].links;
    var classNameWithIdList = _.map(links, function (link) {
        return distNameService.getName(link.distName);
    });
    var qsfpLinkNames = QSFP_LINK_DIST_NAME.START +
        classNameWithIdList.join(QSFP_LINK_DIST_NAME.SEPARATOR) +
        QSFP_LINK_DIST_NAME.END;

    var linkParentDistName = distNameService.getParent(qsfpLinks[qsfpName].links[0].distName);
    return linkParentDistName + '/' + qsfpLinkNames;
}

function isQsfpLinkDistName(distName) {
    var objectName = distNameService.getName(distName);
    return (objectName[0] === QSFP_LINK_DIST_NAME.START );
}

function createParametersFinder(parametersFinder) {
    return {
        getNode: function (infoModelTree, distName) {
            return getLinkNodes(infoModelTree, distName);
        },
        nodeExist: function (treeNodes) {
            return _.any(treeNodes, function (node) {
                return !!node;
            });
        },
        getParameters: function (nodes) {
            var parameters = _.map(nodes, parametersFinder.getParameters);
            return mergeParameters(parameters);
        },

        getTypeInfo: function (treeNodes) {
            var node = _.find(treeNodes, function (treeNode) {
                return !!treeNode;
            });
            return parametersFinder.getTypeInfo(node);
        }
    };
}

function getLinkNodes(imTree, distName) {
    var objectName = distNameService.getName(distName);
    var parentDistName = distNameService.getParent(distName);
    var startSymbolLength = QSFP_LINK_DIST_NAME.START.length;
    var endSymbolLength = QSFP_LINK_DIST_NAME.END.length;
    var linkNameList = objectName.substring(startSymbolLength, objectName.length - endSymbolLength).split(QSFP_LINK_DIST_NAME.SEPARATOR);

    return linkNameList.map(function (name) {
        var linkDistName = parentDistName + '/' + name;
        return imTree.findByDistName(linkDistName);
    });
}

function getAllParametersNames(parameters) {
    return _.reduce(parameters, function (parameterNames, linkParameters) {
        return _.union(parameterNames, (_.keys(linkParameters)));
    }, []);
}

function mergeComplexProperties(propertyList, propertyName) {
    return (Array.isArray(propertyList[0])) ?
        _.uniq(_.flatten(propertyList, true)) : mergeParameters(propertyList, propertyName);
}

function mergeParameters(parameters, parentParameter) {
    var allParameterNames = getAllParametersNames(parameters);
    return _.reduce(allParameterNames, createMergedListOfValues, {});

    function createMergedListOfValues(mergedPropertyValues, propertyName) {
        var propertyList = _.pluck(parameters, propertyName);
        if (propertyName === QFSP_PARAMETERS.DESTINATION_DIST_NAME || propertyName === QFSP_PARAMETERS.SOURCE_DIST_NAME) {
            propertyList = distNameService.getParent(propertyList[0]);
        } else if (propertyName === STATE_INFO.OPERATIONAL_STATE && parentParameter === QFSP_PARAMETERS.STATE_INFO) {
            propertyList = _.contains(propertyList, STATE_INFO.VALUES.DISABLED) ? STATE_INFO.VALUES.DISABLED : STATE_INFO.VALUES.ENABLED;
        } else if (typeof propertyList[0] === 'object') {
            propertyList = mergeComplexProperties(propertyList, propertyName);
        } else {
            propertyList = _.uniq(propertyList)
                .filter(function (property) {
                    return (typeof property !== 'undefined' && property !== null);
                });

            if (typeof propertyList[0] === 'boolean')
                propertyList = _.every(propertyList, _.identity);
            else
                propertyList = propertyList.join(', ');

        }
        mergedPropertyValues[propertyName] = propertyList;
        return mergedPropertyValues;
    }
}
