const InputStream = require('./inputStream');
const DataInputStream = require('./dataInputStream');
const DecoderUtil = require('./decoderUtil');
const LZMA2Coder = require('./lzma2Coder');
const CountingInputStream = require('./countingInputStream');

module.exports = BlockInputStream;

function BlockInputStream(input, check, memoryLimit, unpaddedSizeInIndex, uncompressedSizeInIndex) {
    this.check = check;
    this.inData = new DataInputStream(input);
    this.headerSize = 0;
    this.inCounted = null;
    this.filterChain = null;
    this.tempBuf = [0];
    this.endReached = false;
    this.uncompressedSizeInHeader = -1;
    this.uncompressedSize = 0;

    var buf = [];

    this.inData.readNTo(buf, 0, 1);

    if (buf[0] === 0x00) {
        throw 'IndexIndicatorException';
    }
    this.headerSize = 4 * ((buf[0] & 0xFF) + 1);

    this.inData.readNTo(buf, 1, this.headerSize - 1);

    // TODO validate the CRC32

    if ((buf[1] & 0x3C) != 0)
        throw 'Unsupported options in XZ Block Header';

    var filterCount = (buf[1] & 0x03) + 1;
    var filterIDs = [];
    var filterProps = [];

    const fieldsBytes = new Int8Array(buf.slice(2, 2 + this.headerSize - 6));
    var bufStream = new DataInputStream(new InputStream(fieldsBytes.buffer));

    // TODO validate compressed size, decode uncompressed size

    for (var i = 0; i < filterCount; ++i) {
        filterIDs[i] = DecoderUtil.decodeVLI(bufStream);
        var filterPropsSize = DecoderUtil.decodeVLI(bufStream);
        if (filterPropsSize > bufStream.available())
            throw 'CorruptedInputException';

        filterProps[i] = [];
        bufStream.readNTo(filterProps[i], 0, filterPropsSize);
    }

    // TODO validate block header

    var filters = [];
    for (var i = 0; i < filterIDs.length; ++i) {
        // TODO add rest of the filters
        if (filterIDs[i] === LZMA2Coder.FILTER_ID) {
            filters[i] = LZMA2Coder.createDecoder(filterProps[i]);
        } else {
            throw 'Unknown Filter ID' + filterIDs[i];
        }
    }

    // TODO validate filters and memory limit

    this.inCounted = new CountingInputStream(this.inData);

    this.filterChain = this.inCounted;
    for (var i = filters.length - 1; i >= 0; --i)
        this.filterChain = filters[i].getInputStream(this.filterChain);
}

BlockInputStream.prototype.readNTo = function (buf, off, len) {
    if (this.endReached)
        return -1;

    var ret = this.filterChain.readNTo(buf, off, len);
    if (ret > 0) {
        // TODO update this.check - see BlockInputStream.java:209 (xz-java.zip in resources)
        this.uncompressedSize += ret;

        // TODO catch invalid values

        if (ret < len || this.uncompressedSize === this.uncompressedSizeInHeader) {
            if (this.filterChain.read() != -1)
                throw 'CorruptedInputException';

            this.validate();
            this.endReached = true;
        }
    } else if (ret === -1) {
        this.validate();
        this.endReached = true;
    }

    return ret;
};

BlockInputStream.prototype.validate = function () {
    var compressedSize = this.inCounted.getSize();
    // TODO validate compressed size and uncompressed Size

    while ((compressedSize++ & 3) != 0) {
        var i = this.inData.readUByte();
        if (i != 0x00)
            throw 'CorruptedInputException(';
    }

    // TODO validate the integrity check, now this only skips bytes
    this.inData.readN(this.check.getSize());
};

BlockInputStream.prototype.getUnpaddedSize = function () {
    return this.headerSize + this.inCounted.getSize() + this.check.getSize();
};

BlockInputStream.prototype.getUncompressedSize = function () {
    return this.uncompressedSize;
};
