const DecoderUtil = require('./decoderUtil');

module.exports = LZDecoder;

function LZDecoder(dictSize, presetDict) {
    this.buf = new Array(dictSize);
    this.start = 0;
    this.pos = 0;
    this.full = 0;
    this.limit = 0;
    this.pendingLen = 0;
    this.pendingDist = 0;

    if (presetDict != null) {
        this.pos = Math.min(presetDict.length, dictSize);
        this.full = this.pos;
        this.start = this.pos;
        DecoderUtil.arraycopy(presetDict, presetDict.length - this.pos, this.buf, 0, this.pos);
    }
}

LZDecoder.prototype.reset = function () {
    this.start = 0;
    this.pos = 0;
    this.full = 0;
    this.limit = 0;
    this.buf[this.buf.length - 1] = 0x00;
};

LZDecoder.prototype.setLimit = function (outMax) {
    if (this.buf.length - this.pos <= outMax)
        this.limit = this.buf.length;
    else
        this.limit = this.pos + outMax;
};

LZDecoder.prototype.hasSpace = function () {
    return this.pos < this.limit;
};

LZDecoder.prototype.hasPending = function () {
    return this.pendingLen > 0;
};

LZDecoder.prototype.getPos = function () {
    return this.pos;
};

LZDecoder.prototype.getByte = function (dist) {
    var offset = this.pos - dist - 1;
    if (dist >= this.pos)
        offset += this.buf.length;

    return this.buf[offset] & 0xFF;
};

LZDecoder.prototype.putByte = function (b) {
    this.buf[this.pos++] = b;

    if (this.full < this.pos)
        this.full = this.pos;
};

LZDecoder.prototype.repeat = function (dist, len) {
    if (dist < 0 || dist >= this.full)
        throw 'CorruptedInputException';

    var left = Math.min(this.limit - this.pos, len);
    this.pendingLen = len - left;
    this.pendingDist = dist;

    var back = this.pos - dist - 1;
    if (dist >= this.pos)
        back += this.buf.length;

    do {
        this.buf[this.pos++] = this.buf[back++];
        if (back === this.buf.length)
            back = 0;
    } while (--left > 0);

    if (this.full < this.pos)
        this.full = this.pos;
};

LZDecoder.prototype.repeatPending = function () {
    if (this.pendingLen > 0)
        this.repeat(this.pendingDist, this.pendingLen);
};

LZDecoder.prototype.copyUncompressed = function (inData, len) {
    var copySize = Math.min(this.buf.length - this.pos, len);
    inData.readNTo(this.buf, this.pos, copySize);
    this.pos += copySize;

    if (this.full < this.pos)
        this.full = this.pos;
};

LZDecoder.prototype.flush = function (out, outOff) {
    var copySize = this.pos - this.start;
    if (this.pos === this.buf.length)
        this.pos = 0;

    DecoderUtil.arraycopy(this.buf, this.start, out, outOff, copySize);
    this.start = this.pos;
    return copySize;
};
