const DataInputStream = require('./dataInputStream');
const LZDecoder = require('./lzDecoder');
const RangeDecoderFromBuffer = require('./rangeDecoderFromBuffer');
const LZMADecoder = require('./lzmaDecoder');

module.exports = LZMA2InputStream;

function LZMA2InputStream(countingInputStream, dictSize, presetDict) {
    this.COMPRESSED_SIZE_MAX = 1 << 16;

    this.dataInputStream = new DataInputStream(countingInputStream);
    this.dictSize = dictSize;
    this.needProps = true;
    this.endReached = false;
    this.uncompressedSize = 0;
    this.needDictReset = true;
    this.isLZMAChunk = false;
    this.lzma = null;
    this.tempBuf = [0];
    this.exception = null;

    this.lz = new LZDecoder(LZMA2InputStream.getDictSize(dictSize), presetDict);
    this.rc = new RangeDecoderFromBuffer(this.COMPRESSED_SIZE_MAX);

    if (presetDict && presetDict.length > 0)
        this.needDictReset = false;
}

LZMA2InputStream.getMemoryUsage = function (dictSize) {
    return 40 + this.COMPRESSED_SIZE_MAX / 1024 + LZMA2InputStream.getDictSize(dictSize) / 1024;
};

LZMA2InputStream.getDictSize = function (dictSize) {
    const DICT_SIZE_MIN = 4096;
    const DICT_SIZE_MAX = 2147483647 & ~15;

    if (dictSize < DICT_SIZE_MIN || dictSize > DICT_SIZE_MAX)
        throw 'Unsupported dictionary size';

    return (dictSize + 15) & ~15;
};

LZMA2InputStream.prototype.read = function () {
    return this.readNTo(this.tempBuf, 0, 1) === -1 ? -1 : (this.tempBuf[0] & 0xFF);
};

LZMA2InputStream.prototype.readNTo = function (buf, off, len) {
    if (off < 0 || len < 0 || off + len < 0)
        throw 'IndexOutOfBoundsException';

    if (len === 0)
        return 0;

    if (this.exception != null)
        throw this.exception;

    if (this.endReached)
        return -1;

    try {
        var size = 0;

        while (len > 0) {
            if (this.uncompressedSize === 0) {
                this.decodeChunkHeader();
                if (this.endReached)
                    return size === 0 ? -1 : size;
            }

            var copySizeMax = Math.min(this.uncompressedSize, len);
            if (!this.isLZMAChunk) {
                this.lz.copyUncompressed(this.dataInputStream, copySizeMax);
            } else {
                this.lz.setLimit(copySizeMax);
                this.lzma.decode();
                if (!this.rc.isInBufferOK())
                    throw 'CorruptedInputException';
            }

            var copiedSize = this.lz.flush(buf, off);

            off += copiedSize;
            len -= copiedSize;
            size += copiedSize;
            this.uncompressedSize -= copiedSize;
            if (this.uncompressedSize === 0) {
                if (!this.rc.isFinished() || this.lz.hasPending())
                    throw 'CorruptedInputException';
            }
        }

        return size;

    } catch (e) {
        this.exception = e;
        throw e;
    }
};

LZMA2InputStream.prototype.decodeChunkHeader = function () {
    var control = this.dataInputStream.readUByte();
    if (control === 0x00) {
        this.endReached = true;
        return;
    }

    if (control >= 0xE0 || control === 0x01) {
        this.needProps = true;
        this.needDictReset = false;
        this.lz.reset();
    } else if (this.needDictReset) {
        throw 'CorruptedInputException';
    }

    if (control >= 0x80) {
        this.isLZMAChunk = true;
        this.uncompressedSize = (control & 0x1F) << 16;
        this.uncompressedSize += this.dataInputStream.readUShort() + 1;

        var compressedSize = this.dataInputStream.readUShort() + 1;

        if (control >= 0xC0) {
            this.needProps = false;
            this.decodeProps();
        } else if (this.needProps) {
            throw 'CorruptedInputException';

        } else if (control >= 0xA0) {
            this.lzma.reset();
        }
        this.rc.prepareInputBuffer(this.dataInputStream, compressedSize);

    } else if (control > 0x02) {
        throw 'CorruptedInputException';

    } else {
        this.isLZMAChunk = false;
        this.uncompressedSize = this.dataInputStream.readUShort() + 1;
    }
};

LZMA2InputStream.prototype.decodeProps = function () {
    var props = this.dataInputStream.readUByte();

    if (props > (4 * 5 + 4) * 9 + 8)
        throw 'CorruptedInputException';

    var pb = Math.floor(props / (9 * 5));
    props -= pb * 9 * 5;
    var lp = Math.floor(props / 9);
    var lc = props - lp * 9;

    if (lc + lp > 4)
        throw 'CorruptedInputException';

    this.lzma = new LZMADecoder(this.lz, this.rc, lc, lp, pb);
};
