const Coders = require('./lzmaCoder');
const LZMACoder = Coders.LZMACoder;
const LiteralCoder = Coders.LiteralCoder;
const LiteralSubcoder = Coders.LiteralSubcoder;
const LengthCoder = Coders.LengthCoder;

module.exports = LZMADecoder;

function LZMADecoder(lz, rc, lc, lp, pb) {
    LZMACoder.call(this, pb);

    this.lz = lz;
    this.rc = rc;
    this.literalDecoder = null;
    this.matchLenDecoder = new LengthDecoder(rc);
    this.repLenDecoder = new LengthDecoder(rc);
    this.literalDecoder = new LiteralDecoder(rc, lz, this.state, this.reps, lc, lp);
    this.reset();
}

LZMADecoder.prototype = Object.create(LZMACoder.prototype);
LZMADecoder.prototype.constructor = LZMADecoder;

LZMADecoder.prototype.reset = function () {
    LZMACoder.prototype.reset.call(this);
    this.literalDecoder.reset();
    this.matchLenDecoder.reset();
    this.repLenDecoder.reset();
};

LZMADecoder.prototype.endMarkerDetected = function () {
    return this.reps[0] == -1;
};

LZMADecoder.prototype.decode = function () {
    this.lz.repeatPending();

    while (this.lz.hasSpace()) {
        var posState = this.lz.getPos() & this.posMask;
        if (this.rc.decodeBit(this.isMatch[this.state.get()], posState) === 0) {
            this.literalDecoder.decode();
        } else {
            var len = this.rc.decodeBit(this.isRep, this.state.get()) === 0
                ? this.decodeMatch(posState)
                : this.decodeRepMatch(posState);

            this.lz.repeat(this.reps[0], len);
        }
    }

    this.rc.normalize();
};

LZMADecoder.prototype.decodeMatch = function (posState) {
    this.state.updateMatch();

    this.reps[3] = this.reps[2];
    this.reps[2] = this.reps[1];
    this.reps[1] = this.reps[0];

    var len = this.matchLenDecoder.decode(posState);
    var distSlot = this.rc.decodeBitTree(this.distSlots[this.getDistState(len)]);
    if (distSlot < this.DIST_MODEL_START) {
        this.reps[0] = distSlot;
    } else {
        var limit = (distSlot >> 1) - 1;
        this.reps[0] = (2 | (distSlot & 1)) << limit;

        if (distSlot < this.DIST_MODEL_END) {
            this.reps[0] |= this.rc.decodeReverseBitTree(this.distSpecial[distSlot - this.DIST_MODEL_START]);
        } else {
            this.reps[0] |= this.rc.decodeDirectBits(limit - this.ALIGN_BITS) << this.ALIGN_BITS;
            this.reps[0] |= this.rc.decodeReverseBitTree(this.distAlign);
        }
    }

    return len;
};

LZMADecoder.prototype.decodeRepMatch = function (posState) {
    if (this.rc.decodeBit(this.isRep0, this.state.get()) === 0) {
        if (this.rc.decodeBit(this.isRep0Long[this.state.get()], posState) === 0) {
            this.state.updateShortRep();
            return 1;
        }
    } else {
        var tmp;

        if (this.rc.decodeBit(this.isRep1, this.state.get()) === 0) {
            tmp = this.reps[1];
        } else {
            if (this.rc.decodeBit(this.isRep2, this.state.get()) === 0) {
                tmp = this.reps[2];
            } else {
                tmp = this.reps[3];
                this.reps[3] = this.reps[2];
            }

            this.reps[2] = this.reps[1];
        }

        this.reps[1] = this.reps[0];
        this.reps[0] = tmp;
    }

    this.state.updateLongRep();

    return this.repLenDecoder.decode(posState);
};

function LiteralDecoder(rc, lz, state, reps, lc, lp) {
    LiteralCoder.call(this, lc, lp);
    this.lz = lz;

    this.subdecoders = new Array(1 << (lc + lp));
    for (var i = 0; i < this.subdecoders.length; ++i)
        this.subdecoders[i] = new LiteralSubdecoder(rc, lz, state, reps);
}

LiteralDecoder.prototype = Object.create(LiteralCoder.prototype);
LiteralDecoder.prototype.constructor = LiteralDecoder;

LiteralDecoder.prototype.reset = function () {
    for (var i = 0; i < this.subdecoders.length; ++i)
        this.subdecoders[i].reset();
};

LiteralDecoder.prototype.decode = function () {
    var i = this.getSubcoderIndex(this.lz.getByte(0), this.lz.getPos());
    this.subdecoders[i].decode();
};

function LiteralSubdecoder(rc, lz, state, reps) {
    LiteralSubcoder.call(this);
    this.rc = rc;
    this.lz = lz;
    this.state = state;
    this.reps = reps;
}

LiteralSubdecoder.prototype = Object.create(LiteralSubcoder.prototype);
LiteralSubdecoder.prototype.constructor = LiteralSubdecoder;

LiteralSubdecoder.prototype.decode = function () {
    var symbol = 1;

    if (this.state.isLiteral()) {
        do {
            symbol = (symbol << 1) | this.rc.decodeBit(this.probs, symbol);
        } while (symbol < 0x100);

    } else {
        var matchByte = this.lz.getByte(this.reps[0]);
        var offset = 0x100;
        var matchBit;
        var bit;

        do {
            matchByte <<= 1;
            matchBit = matchByte & offset;
            bit = this.rc.decodeBit(this.probs, offset + matchBit + symbol);
            symbol = (symbol << 1) | bit;
            offset &= (0 - bit) ^ ~matchBit;
        } while (symbol < 0x100);
    }
    this.lz.putByte(toByte(symbol));
    this.state.updateLiteral();

    function toByte(n) {
        const view = new DataView(new ArrayBuffer(1));
        view.setInt8(0, n);
        return view.getInt8(0);
    }
};

function LengthDecoder(rc) {
    LengthCoder.call(this);
    this.rc = rc;
    this.MATCH_LEN_MIN = 2; // copied from LZMACoder.js
}

LengthDecoder.prototype = Object.create(LengthCoder.prototype);
LengthDecoder.prototype.constructor = LengthDecoder;

LengthDecoder.prototype.decode = function (posState) {
    if (this.rc.decodeBit(this.choice, 0) == 0) {
        return this.rc.decodeBitTree(this.low[posState]) + this.MATCH_LEN_MIN;
    }

    if (this.rc.decodeBit(this.choice, 1) == 0) {
        return this.rc.decodeBitTree(this.mid[posState]) + this.MATCH_LEN_MIN + this.LOW_SYMBOLS;
    }

    return this.rc.decodeBitTree(this.high) + this.MATCH_LEN_MIN + this.LOW_SYMBOLS + this.MID_SYMBOLS;
};
