describe('alarms', function () {
    var distNameUtils;
    var alarms;

    beforeEach(function () {
        distNameUtils = adminCore.utils.distNameUtils;
        alarms = adminCore.abnormalities.alarms;
    });

    const DIST_NAME = 'MY/DIST-1/NAME-2';
    const ALARM_PARAMETERS = {
        alarmInformation: {
            alarmSeverity: 'ALARM_SEVERITY',
            faultId: 'FAULT_ID',
            alarmNumber: 'NUMBER',
            observationTime: 'BTS_OBSERVATION_TIME',
            eventType: 'EVENT_TYPE',
            alarmText: {
                alarmDetail: 'ALARM_DETAIL'
            }
        },
        unitInformation: {
            alarmingResource: 'ALARMING_RESOURCE'
        }
    };
    const TIMESTAMP = 1234;

    describe('creating alarm', function () {

        it('should create new alarm', function () {
            const alarm = alarms.newAlarm(DIST_NAME, ALARM_PARAMETERS);
            assertHasCorrectAlarmParameters(alarm);
        });

        it('should create new historical alarm', function () {
            const alarm = alarms.newHistoricalAlarm(DIST_NAME, ALARM_PARAMETERS, TIMESTAMP);
            assertHasCorrectHistoricalAlarmParameters(alarm);
        });

        it('should create new change alarm', function () {
            const changedAlarm = alarms.newAlarm(DIST_NAME, ALARM_PARAMETERS);
            const alarm = alarms.newChangeAlarm(changedAlarm, DIST_NAME, ALARM_PARAMETERS, TIMESTAMP);
            assertHasCorrectChangeAlarmParameters(alarm, changedAlarm);
        });

        describe('checking if alarm is new', function () {

            it('should return true for standard alarm', function () {
                const alarm = alarms.newAlarm(DIST_NAME, ALARM_PARAMETERS);
                assert.isTrue(alarm.isNew());
            });

            it('should return true for historical alarm', function () {
                const alarm = alarms.newHistoricalAlarm(DIST_NAME, ALARM_PARAMETERS, TIMESTAMP);
                assert.isTrue(alarm.isNew());
            });

            it('should return false for change alarm', function () {
                const changedAlarm = alarms.newAlarm(DIST_NAME, ALARM_PARAMETERS);
                const alarm = alarms.newChangeAlarm(changedAlarm, DIST_NAME, ALARM_PARAMETERS, TIMESTAMP);
                assert.isFalse(alarm.isNew());
            });

        });

        describe('checking if alarmas are equal', function () {

            it('should return true for alarms with the same parameters and dist name', function () {
                const alarm1 = alarms.newAlarm(DIST_NAME, ALARM_PARAMETERS);
                const alarm2 = alarms.newAlarm(DIST_NAME, ALARM_PARAMETERS);
                assert.isTrue(alarm1.isEqualTo(alarm2));
            });

            it('should return false for alarms with different dist name', function () {
                const alarm1 = alarms.newAlarm(DIST_NAME, ALARM_PARAMETERS);
                const alarm2 = alarms.newAlarm('DIFFERENT/DIST-1/NAME-2', ALARM_PARAMETERS);
                assert.isFalse(alarm1.isEqualTo(alarm2));
            });

            it('should return false for alarms with different parameters', function () {
                const newParameters = _.cloneDeep(ALARM_PARAMETERS);
                newParameters.unitInformation.alarmingResource = 'someOtherAlarmingResource';
                const alarm1 = alarms.newAlarm(DIST_NAME, ALARM_PARAMETERS);
                const alarm2 = alarms.newAlarm(DIST_NAME, newParameters);
                assert.isFalse(alarm1.isEqualTo(alarm2));
            });

        });

        describe('checking if alarm has severity regardless letter case', function () {

            it('should match severity regardless letter case', function () {
                const alarm = alarms.newAlarm(DIST_NAME, createParametersWithSeverity('CriTiCal'));
                assert.isTrue(alarm.hasSeverity('critical'));
                assert.isTrue(alarm.hasSeverity('CRITICAL'));
                assert.isFalse(alarm.hasSeverity('miNoR'));
            });

        });

        describe('updating disappear time', function () {

            it('should update disappear time for historical alarm', function () {
                const alarm = alarms.newHistoricalAlarm(DIST_NAME, ALARM_PARAMETERS, TIMESTAMP);
                alarm.updateDisappearTime(7890);
                assert.deepEqual(alarm.disappearTime, 7890);
            });

            it('should update disappear time for change alarm', function () {
                const changedAlarm = alarms.newHistoricalAlarm(DIST_NAME, ALARM_PARAMETERS);
                const alarm = alarms.newChangeAlarm(changedAlarm, DIST_NAME, ALARM_PARAMETERS, TIMESTAMP);
                alarm.updateDisappearTime(7890);
                assert.deepEqual(alarm.disappearTime, 7890);
                assert.deepEqual(changedAlarm.disappearTime, 7890);
            });

        });

        describe('checking if alarm is disappeared', function () {

            it('should return false if alarm wasn\'t disappeared', function () {
                const alarm = alarms.newHistoricalAlarm(DIST_NAME, ALARM_PARAMETERS, TIMESTAMP);
                assert.isFalse(alarm.disappeared());
            });

            it('should return true if alarm was disappeared', function () {
                const alarm = alarms.newHistoricalAlarm(DIST_NAME, ALARM_PARAMETERS, TIMESTAMP);
                alarm.updateDisappearTime(7890);
                assert.isTrue(alarm.disappeared());
            });

            it('should return true if change alarm was disappeared', function () {
                const changedAlarm = alarms.newHistoricalAlarm(DIST_NAME, ALARM_PARAMETERS);
                const alarm = alarms.newChangeAlarm(changedAlarm, DIST_NAME, ALARM_PARAMETERS, TIMESTAMP);
                alarm.updateDisappearTime(7890);
                assert.isTrue(alarm.disappeared());
                assert.isTrue(changedAlarm.disappeared());
            });

        });

    });

    describe('isAlarm', function () {

        it('should return true if dist name is alarm', function () {
            assert.isTrue(alarms.isAlarm('/MRBTS-1/ALARM_L-1'));
        });

        it('should return false if parent is alarm', function () {
            assert.isFalse(alarms.isAlarm('/MRBTS-1/ALARM_L-1/NOT-ALARM-1'));
        });

        it('should return false if dist name is not alarm', function () {
            assert.isFalse(alarms.isAlarm('/MRBTS-1/NOT-ALARM-1'));
        });

    });

    describe('sorting', function () {

        it('should sort alarms using observation time', function () {
            const alarmsToSort = [
                createAlarmWithObservationTime('SOME/DIST-1/NAME-2', 241),
                createAlarmWithObservationTime('SOME/DIST-1/NAME-2', 122),
                createAlarmWithObservationTime('SOME/DIST-1/NAME-2', 342),
                createAlarmWithObservationTime('SOME/DIST-1/NAME-11', 241),
                createAlarmWithObservationTime('SOME/DIST-1/NAME-1', 241),
                createAlarmWithObservationTime('SOME/DIST-2/NAME-2', 241)
            ];
            alarms.sortUsingObservationTime(alarmsToSort);
            assert.propertyVal(alarmsToSort[0], 'distName', 'SOME/DIST-1/NAME-2');
            assert.propertyVal(alarmsToSort[0], 'btsObservationTime', 122);
            assert.propertyVal(alarmsToSort[1], 'distName', 'SOME/DIST-1/NAME-1');
            assert.propertyVal(alarmsToSort[1], 'btsObservationTime', 241);
            assert.propertyVal(alarmsToSort[2], 'distName', 'SOME/DIST-1/NAME-2');
            assert.propertyVal(alarmsToSort[2], 'btsObservationTime', 241);
            assert.propertyVal(alarmsToSort[3], 'distName', 'SOME/DIST-1/NAME-11');
            assert.propertyVal(alarmsToSort[3], 'btsObservationTime', 241);
            assert.propertyVal(alarmsToSort[4], 'distName', 'SOME/DIST-2/NAME-2');
            assert.propertyVal(alarmsToSort[4], 'btsObservationTime', 241);
            assert.propertyVal(alarmsToSort[5], 'distName', 'SOME/DIST-1/NAME-2');
            assert.propertyVal(alarmsToSort[5], 'btsObservationTime', 342);
        });

        it('should sort alarms using event time', function () {
            const alarmsToSort = [
                createAlarmWithEventTime('SOME/DIST-1/NAME-2', 241),
                createAlarmWithEventTime('SOME/DIST-1/NAME-2', 122),
                createAlarmWithEventTime('SOME/DIST-1/NAME-2', 342),
                createAlarmWithEventTime('SOME/DIST-1/NAME-11', 241),
                createAlarmWithEventTime('SOME/DIST-1/NAME-1', 241),
                createAlarmWithEventTime('SOME/DIST-2/NAME-2', 241)
            ];
            alarms.sortUsingEventTime(alarmsToSort);
            assert.propertyVal(alarmsToSort[0], 'distName', 'SOME/DIST-1/NAME-2');
            assert.propertyVal(alarmsToSort[0], 'eventTime', 122);
            assert.propertyVal(alarmsToSort[1], 'distName', 'SOME/DIST-1/NAME-1');
            assert.propertyVal(alarmsToSort[1], 'eventTime', 241);
            assert.propertyVal(alarmsToSort[2], 'distName', 'SOME/DIST-1/NAME-2');
            assert.propertyVal(alarmsToSort[2], 'eventTime', 241);
            assert.propertyVal(alarmsToSort[3], 'distName', 'SOME/DIST-1/NAME-11');
            assert.propertyVal(alarmsToSort[3], 'eventTime', 241);
            assert.propertyVal(alarmsToSort[4], 'distName', 'SOME/DIST-2/NAME-2');
            assert.propertyVal(alarmsToSort[4], 'eventTime', 241);
            assert.propertyVal(alarmsToSort[5], 'distName', 'SOME/DIST-1/NAME-2');
            assert.propertyVal(alarmsToSort[5], 'eventTime', 342);
        });

    });

    function assertHasCorrectAlarmParameters(alarm) {
        assert.propertyVal(alarm, 'severity', 'ALARM_SEVERITY');
        assert.propertyVal(alarm, 'faultId', 'FAULT_ID');
        assert.propertyVal(alarm, 'number', 'NUMBER');
        assert.propertyVal(alarm, 'distName', DIST_NAME);
        assert.propertyVal(alarm, 'btsObservationTime', 'BTS_OBSERVATION_TIME');
        assert.propertyVal(alarm, 'eventType', 'EVENT_TYPE');
        assert.propertyVal(alarm, 'details', 'ALARM_DETAIL');
        assert.propertyVal(alarm, 'notifier', 'ALARMING_RESOURCE');
    }

    function assertHasCorrectHistoricalAlarmParameters(alarm) {
        assertHasCorrectAlarmParameters(alarm);
        assert.propertyVal(alarm, 'eventTime', TIMESTAMP);
    }

    function assertHasCorrectChangeAlarmParameters(alarm, changedAlarm) {
        assertHasCorrectHistoricalAlarmParameters(alarm);
        assert.deepPropertyVal(alarm, 'stateBefore', changedAlarm);
    }

    function createParametersWithSeverity(severity) {
        return {
            alarmInformation: {
                alarmSeverity: severity,
                faultId: 'FAULT_ID',
                alarmNumber: 'NUMBER',
                observationTime: 'BTS_OBSERVATION_TIME',
                eventType: 'EVENT_TYPE',
                alarmText: {
                    alarmDetail: 'ALARM_DETAIL'
                }
            },
            unitInformation: {
                alarmingResource: 'ALARMING_RESOURCE'
            }
        };
    }

    function createAlarmWithObservationTime(distName, observationTime) {
        return alarms.newAlarm(distName, {
            alarmInformation: {
                observationTime,
                alarmText: {}
            },
            unitInformation: {}
        });
    }

    function createAlarmWithEventTime(distName, eventTime) {
        return alarms.newHistoricalAlarm(distName, {
            alarmInformation: {
                alarmText: {}
            },
            unitInformation: {}
        }, eventTime);
    }

});
