describe('alarmsMonitor', function () {
    var alarmsMonitor;
    var updateService;
    var decoderService;
    var distNameUtils;
    var alarms;

    beforeEach(function () {
        decoderService = createDecoderService();
        const metaService = {
            getLastMetaUid: function () {
                return 0;
            }
        };

        updateService = infoModel.updateService({
            metaService: metaService,
            decoderService: decoderService
        });
        distNameUtils = adminCore.utils.distNameUtils;
        alarmsMonitor = adminCore.abnormalities.alarmsMonitor(updateService, metaService, decoderService);
        alarms = adminCore.abnormalities.alarms;
    });

    describe('current alarms', function () {
        beforeEach(function () {
            sendAlarms([1, 2, 3]);
            sendAlarms([4]);
        });

        describe('history marker not set', function () {
            it('should get all alarms', function (done) {
                alarmsMonitor.watchCurrentAlarms(watcher);

                function watcher(alarms) {
                    assertHasAlarms(alarms, [1, 2, 3, 4]);
                    done();
                }
            });

            it('should not get removed alarms', function (done) {
                removeAlarms([1, 2, 3]);
                alarmsMonitor.watchCurrentAlarms(watcher);

                function watcher(alarms) {
                    assertHasAlarms(alarms, [4]);
                    done();
                }
            });

            it('should get removed alarm that appeared again', function (done) {
                removeAlarms([1, 2, 3, 4]);
                sendAlarms([3]);
                alarmsMonitor.watchCurrentAlarms(watcher);

                function watcher(alarms) {
                    assertHasAlarms(alarms, [3]);
                    done();
                }
            });
        });

        describe('history marker set', function () {

            it('should get alarms from specific history moment', function (done) {
                updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                    .then(function () {
                        alarmsMonitor.watchCurrentAlarms(watcher);

                        function watcher(alarms) {
                            assertHasAlarms(alarms, [1, 2, 3]);
                            done();
                        }

                    })
                    .catch(done);
            });
        });

        it('should inform second watcher about alarms', function (done) {
            alarmsMonitor
                .watchCurrentAlarms(addSecondWatcher)
                .unwatch();

            function addSecondWatcher() {
                alarmsMonitor
                    .watchCurrentAlarms(watcher);
            }

            function watcher(alarms) {
                assertHasAlarms(alarms, [1, 2, 3, 4]);
                done();
            }
        });
    });

    describe('historical alarms', function () {
        const FIRST_UPDATE_TIME = 2222;
        const SECOND_UPDATE_TIME = 3333;

        beforeEach(function () {
            sendAlarms([1, 2, 3], FIRST_UPDATE_TIME);
            sendAlarms([4], SECOND_UPDATE_TIME);
        });

        it('should find all alarms when history marker is not set', function (done) {
            const historicalAlarms = catchEvent(alarmsMonitor.watchHistoricalAlarms);

            historicalAlarms
                .then(function (alarms) {
                    assertHasAlarms(alarms, [1, 2, 3, 4]);
                })
                .then(done)
                .catch(done);
        });

        it('should find all alarms when history marker is set', function (done) {
            const historicalAlarms = updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                .then(function () {
                    return catchEvent(alarmsMonitor.watchHistoricalAlarms);
                });

            historicalAlarms
                .then(function (alarms) {
                    assertHasAlarms(alarms, [1, 2, 3, 4]);
                })
                .then(done)
                .catch(done);
        });

        it('should return correct alarms when watching started, stopped & started again', function (done) {
            alarmsMonitor.watchHistoricalAlarms(noop)
                .unwatch();
            const historicalAlarms = catchEvent(alarmsMonitor.watchHistoricalAlarms);

            historicalAlarms
                .then(function (alarms) {
                    assertHasAlarms(alarms, [1, 2, 3, 4]);
                })
                .then(done)
                .catch(done);
        });

        it('should inform second watcher about alarms', function (done) {
            const alarmsHistory1 = catchEvent(alarmsMonitor.watchHistoricalAlarms);
            const alarmsHistory2 = catchEvent(alarmsMonitor.watchHistoricalAlarms);

            Promise.all([alarmsHistory1, alarmsHistory2])
                .then(function (results) {
                    assertHasAlarms(results[0], [1, 2, 3, 4]);
                    assertHasAlarms(results[1], [1, 2, 3, 4]);
                })
                .then(done)
                .catch(done);
        });

        it('should set disappeared time', function (done) {
            const DISAPPEAR_TIME = 1234;

            const historicalAlarms = catchEvent(alarmsMonitor.watchHistoricalAlarms, 1);

            removeAlarms([2], DISAPPEAR_TIME);
            historicalAlarms
                .then(function (alarms) {
                    assert.lengthOf(alarms, 4);
                    assertIsNewAlarm(alarms[0], 1, FIRST_UPDATE_TIME);
                    assertIsNewAlarm(alarms[1], 2, FIRST_UPDATE_TIME, DISAPPEAR_TIME);
                    assertIsNewAlarm(alarms[2], 3, FIRST_UPDATE_TIME);
                    assertIsNewAlarm(alarms[3], 4, SECOND_UPDATE_TIME);
                })
                .then(done)
                .catch(done);
        });

        it('should find changed alarm', function (done) {
            const CHANGE_TIME = 1234;
            const SUFFIX = '2_2';
            const historicalAlarms = catchEvent(alarmsMonitor.watchHistoricalAlarms, 1);

            sendSpecificAlarm(2, SUFFIX, CHANGE_TIME);
            historicalAlarms
                .then(function (alarms) {
                    assert.lengthOf(alarms, 5);
                    assertIsNewAlarm(alarms[0], 1, FIRST_UPDATE_TIME);
                    assertIsNewAlarm(alarms[1], 2, FIRST_UPDATE_TIME);
                    assertIsNewAlarm(alarms[2], 3, FIRST_UPDATE_TIME);
                    assertIsNewAlarm(alarms[3], 4, SECOND_UPDATE_TIME);
                    assertIsChangedAlarm(alarms[4], 2, undefined, CHANGE_TIME, alarms[1], SUFFIX);
                })
                .then(done)
                .then(done);
        });

        it('should not mark alarm as changed when nothing changed in update', function (done) {
            const CHANGE_TIME = 1234;
            const historicalAlarms = catchEvent(alarmsMonitor.watchHistoricalAlarms, 1);

            sendAlarms([4], CHANGE_TIME);
            historicalAlarms
                .then(function (alarms) {
                    assert.lengthOf(alarms, 4);
                    assertIsNewAlarm(alarms[0], 1, FIRST_UPDATE_TIME);
                    assertIsNewAlarm(alarms[1], 2, FIRST_UPDATE_TIME);
                    assertIsNewAlarm(alarms[2], 3, FIRST_UPDATE_TIME);
                    assertIsNewAlarm(alarms[3], 4, SECOND_UPDATE_TIME);
                })
                .then(done)
                .then(done);
        });

        it('should process entries after watcher unwatchs and new one added', function (done) {
            alarmsMonitor.watchHistoricalAlarms(noop)
                .unwatch();

            sendAlarms([5, 6, 7]);
            const historicalAlarms = catchEvent(alarmsMonitor.watchHistoricalAlarms);

            historicalAlarms
                .then(function (alarms) {
                    assertHasAlarms(alarms, [1, 2, 3, 4, 5, 6, 7]);
                })
                .then(done)
                .then(done);
        });

        it('should inform about empty alarms when history was cleared', function (done) {
            const historicalAlarms = catchEvent(alarmsMonitor.watchHistoricalAlarms, 1);

            updateService.clear();

            historicalAlarms
                .then(function (alarms) {
                    assert.lengthOf(alarms, 0);
                    done();
                })
                .then(done)
                .then(done);
        });

        it('should mark all alarms as disappeared after reset', function (done) {
            const DISAPPEAR_TIME = 1234;
            const historicalAlarms = catchEvent(alarmsMonitor.watchHistoricalAlarms, 1);

            pushReset(DISAPPEAR_TIME);

            historicalAlarms
                .then(function (alarms) {
                    assert.lengthOf(alarms, 4);
                    assertIsNewAlarm(alarms[0], 1, FIRST_UPDATE_TIME, DISAPPEAR_TIME);
                    assertIsNewAlarm(alarms[1], 2, FIRST_UPDATE_TIME, DISAPPEAR_TIME);
                    assertIsNewAlarm(alarms[2], 3, FIRST_UPDATE_TIME, DISAPPEAR_TIME);
                    assertIsNewAlarm(alarms[3], 4, SECOND_UPDATE_TIME, DISAPPEAR_TIME);
                })
                .then(done)
                .then(done);
        });
    });

    describe('severity count', function () {
        it('should find all severities', function () {
            const alarms = [
                newAlarmWithSeverity('critical'),
                newAlarmWithSeverity('critical'),
                newAlarmWithSeverity('warning'),
                newAlarmWithSeverity('minor'),
                newAlarmWithSeverity('minor'),
                newAlarmWithSeverity('major')
            ];

            const severityCount = alarmsMonitor.getSeverityCount(alarms);

            assert.deepEqual(severityCount.minor, 2);
            assert.deepEqual(severityCount.major, 1);
            assert.deepEqual(severityCount.warning, 1);
            assert.deepEqual(severityCount.critical, 2);
        });

        it('should retun 0 for each severity when there is no alarms', function () {
            const alarms = [];

            const severityCount = alarmsMonitor.getSeverityCount(alarms);

            assert.deepEqual(severityCount.minor, 0);
            assert.deepEqual(severityCount.major, 0);
            assert.deepEqual(severityCount.warning, 0);
            assert.deepEqual(severityCount.critical, 0);
        });

        it('should count only NEW alarms', function () {
            const alarms = [
                newAlarmWithSeverity('critical'),
                newAlarmWithSeverity('critical'),
                newAlarmWithSeverity('warning'),
                newAlarmWithSeverity('minor'),
                newAlarmWithSeverity('minor'),
                newAlarmWithSeverity('major')
            ];
            alarms[1].stateBefore = alarms[0];

            const severityCount = alarmsMonitor.getSeverityCount(alarms);

            assert.deepEqual(severityCount.minor, 2);
            assert.deepEqual(severityCount.major, 1);
            assert.deepEqual(severityCount.warning, 1);
            assert.deepEqual(severityCount.critical, 1);
        });
    });

    function newAlarmWithSeverity(alarmSeverity) {
        return alarms.newAlarm('/SOME/DIST-1/NAME-2', {
            alarmInformation: {
                alarmText: {},
                alarmSeverity
            },
            unitInformation: {}
        });
    }

    function assertHasAlarms(alarms, numbers) {
        assert(alarms, numbers.length);
        numbers.forEach(function (number, index) {
            assertAlarmHasCorrectParameters(alarms[index], number);
        });
    }

    function assertAlarmHasCorrectParameters(alarm, number, suffix) {
        const newSuffix = suffix || number;
        const ALARM_PARAMETERS = ['severity', 'faultId', 'number', 'distName',
            'btsObservationTime', 'eventType', 'details', 'notifier'];
        const cleanAlarm = _.pick(alarm, ALARM_PARAMETERS);
        assert.deepEqual(cleanAlarm, {
            severity: 'alarmSeverity' + newSuffix,
            faultId: 'faultId' + newSuffix,
            number: 'alarmNumber' + newSuffix,
            distName: '/A-1/ALARM_L-' + number,
            btsObservationTime: 'observationTime' + newSuffix,
            eventType: 'eventType' + newSuffix,
            details: 'alarmDetail' + newSuffix,
            notifier: 'alarmingResource' + newSuffix
        });
    }

    function assertIsNewAlarm(alarm, number, eventTime, disappearTime, suffix) {
        assertIsChangedAlarm(alarm, number, disappearTime, eventTime, undefined, suffix);
    }

    function assertIsChangedAlarm(alarm, number, disappearTime, eventTime, stateBefore, suffix) {
        assertAlarmHasCorrectParameters(alarm, number, suffix);
        assert.deepEqual(alarm.disappearTime, disappearTime);
        assert.deepEqual(alarm.eventTime, eventTime);
        assert.deepEqual(alarm.stateBefore, stateBefore);
    }

    function removeAlarms(numbers, timestamp) {
        const removedDistNames = [];
        numbers.forEach(function (number) {
            removedDistNames.push('/A-1/ALARM_L-' + number);
        });
        updateService.pushData({
            removedDistNames: removedDistNames,
            managedObjects: [],
            timestamp: timestamp
        });
    }

    function sendAlarms(numbers, timestamp) {
        const managedObjects = [];
        numbers.forEach(function (number) {
            const distName = '/A-1/ALARM_L-' + number;
            decoderService.$data[getClassName(distName) + number] = createAlarmParams(number);
            managedObjects.push({
                distName: distName,
                payload: number
            });
        });
        updateService.pushData({
            removedDistNames: [],
            managedObjects: managedObjects,
            timestamp: timestamp
        });
    }

    function sendSpecificAlarm(number, suffix, timestamp) {
        const distName = '/A-1/ALARM_L-' + number;
        decoderService.$data[getClassName(distName) + suffix] = createAlarmParams(suffix);
        const managedObjects = [{
            distName: distName,
            payload: suffix
        }];
        updateService.pushData({
            removedDistNames: [],
            managedObjects: managedObjects,
            timestamp: timestamp
        });
    }

    function createAlarmParams(number) {
        return {
            alarmInformation: {
                alarmSeverity: 'alarmSeverity' + number,
                faultId: 'faultId' + number,
                alarmNumber: 'alarmNumber' + number,
                observationTime: 'observationTime' + number,
                eventType: 'eventType' + number,
                alarmText: {
                    alarmDetail: 'alarmDetail' + number
                }
            },
            unitInformation: {
                alarmingResource: 'alarmingResource' + number
            }
        };
    }

    function createDecoderService() {
        return {
            $data: {},
            decodeByDistName: function (metaUid, payload, distName) {
                return this.$data[getClassName(distName) + payload];
            },
            decodeByClassName: function (metaUid, payload, className) {
                return this.$data[className + payload];
            }
        };
    }

    function getClassName(distName) {
        return distNameUtils.getClassName(distName);
    }

    function noop() {
    }

    function pushReset(timestamp) {
        updateService.pushDisconnect({
            autoReconnect: false,
            source: 'BTS',
            timestamp: timestamp
        });
    }
});
