const reasoningStatusesFactory = require('../../src/abnormalities/reasoningStatuses');

describe('reasoningStatuses', function () {
    var suite;

    beforeEach(function () {
        suite = {};
        suite.entries = [createEntry(0, 1000, createChanges(['/ABC-1', '/DEF-9', '/GHI-34'], 0))];
        suite.imtrees = [
            createImTree([
                    ['/ABC-1', []],
                    ['/DEF-9', ['reasoning_status_1', 'reasoning_status_2']],
                    ['/GHI-34', []]]
            )
        ];
        const classNamesWithRS = ['DEF', 'GHI'];
        suite.expectedRs = new Map();
        let rs = new Map();
        rs.set('reasoning_status_1', {
            distName: '/DEF-9',
            eventType: 'reasoning_status_1',
            observationTime: 1000
        });
        rs.set('reasoning_status_2', {
            distName: '/DEF-9',
            eventType: 'reasoning_status_2',
            observationTime: 1000
        });
        suite.expectedRs.set('/DEF-9', rs);
        mockServices(classNamesWithRS);
        suite.reasoningStatuses = reasoningStatusesFactory({
            updateService: suite.updateService,
            metaService: suite.metaService,
            decoderService: suite.decoderService
        });
    });

    afterEach(function() {
        suite = null;
    });

    describe('Visibility: false', function() {

        it('should send proper item count', function(done) {
            suite.reasoningStatuses.watch(watcher);
            moveToEntry(0);

            function watcher(event) {
                if (event.type == 'count') {
                    assert.equal(event.data, 2);
                    done();
                }
            }
        });
    });

    describe('Visibility: true', function() {

        it('should send appropriate items', function(done) {
            suite.reasoningStatuses.setVisibility(true);
            suite.reasoningStatuses.watch(watcher);
            moveToEntry(0);

            function watcher(event) {
                if (event.type == 'data') {
                    assert.deepEqual(Array.from(event.data.keys()), Array.from(suite.expectedRs.keys()));
                    assert.deepEqual(Array.from(event.data.values()), Array.from(suite.expectedRs.values()));
                    assert.deepEqual(
                        Array.from(Array.from(event.data.values())[0].keys()),
                        Array.from(Array.from(suite.expectedRs.values())[0].keys()));
                    assert.deepEqual(
                        Array.from(Array.from(event.data.values())[0].values()),
                        Array.from(Array.from(suite.expectedRs.values())[0].values()));
                    done();
                }
            }
        });
    });

    function mockServices(classNamesWithRS) {
        classNamesWithRS = classNamesWithRS || [];

        suite.updateService = {
            addImTreeListener: function (listener) {
                suite.imTreeListener = listener;
            },
            removeImTreeListener: function () {
                suite.imTreeListener = null;
            },
            addCurrentEntryListener: function (listener) {
                suite.currentEntryListener = listener;
            },
            removeCurrentEntryListener: function () {
                suite.currentEntryListener = null;
            },
            addHistoryListener: _.noop,
            removeHistoryListener: _.noop,
            getHistoryEntry: function (historyEntryId) {
                return suite.entries[historyEntryId];
            }
        };

        suite.metaService = {
            getTypeInfo: function (metaUid, distName) {
                return {
                    getParameterByName: function () {
                        return {
                            getParameterByName: function () {
                                return (classNamesWithRS.indexOf(distName) !== -1);
                            }
                        };
                    }
                };
            }
        };

        suite.decoderService = {
            decodeByDistName: function (metaUid, payload, distName) {
                return {
                    stateInfo: {
                        reasoningStatus: getReasoningStatuses(suite.imtrees[getEntryId(payload)], distName)
                    }
                };
            }
        };
        function getEntryId(payload) {
            return payload;
        }

        function getReasoningStatuses(imtree, distName) {
            return imtree.findByDistName(distName).getDecodedParameters().stateInfo.reasoningStatus;
        }

    }

    function createEntry(id, timestamp, updated, removed) {
        updated = updated || [];
        removed = removed || [];
        var numOfChanges = updated.length + removed.length;

        return {
            id: id,
            timestamp: timestamp,
            numOfChanges: numOfChanges,
            change: {updated: updated, removed: removed}
        };
    }

    function createChanges(distNames, entryId) {
        return _.map(distNames, function (distName) {
            return {
                distName: distName,
                payload: entryId
            };
        });
    }

    function createImTree(distNamesWithRS) {
        return {
            distNames: function () {
                return _.map(distNamesWithRS, function (distNameWithRS) {
                    return distNameWithRS[0];
                });
            },
            findByDistName: function (distName) {
                return {
                    distName: distName,
                    getDecodedParameters: function () {
                        return {
                            stateInfo: {
                                reasoningStatus: _.chain(distNamesWithRS)
                                    .find(function (distNameWithRS) {
                                        return distNameWithRS[0] === distName;
                                    })
                                    .result('[1]')
                                    .value()
                            }
                        };
                    }
                };
            }
        };
    }

    function moveToEntry(id) {
        suite.imTreeListener.completeImTree(suite.imtrees[id]);
        suite.currentEntryListener.changeEntry(suite.entries[id]);
    }

});
