const adjacencyUtilities = require('../../src/adjacency/adjacencyUtilities');

describe('adjacency utilities', () => {

    describe('getting downlink frequency', () => {
        it('should get downlink frequency from lnadjl', () => {
            // given
            const lnadjl = getFakeLNADJL();
            // when
            const dlFreq = adjacencyUtilities.getDownlinkFrequency(lnadjl);
            // then
            assert.equal(dlFreq.name, 'fDlEarfcn');
            assert.equal(dlFreq.value, 666);
        });

        it('should get downlink frequency from lnadjw', () => {
            // given
            const lnadjw = getFakeLNADJW();
            // when
            const dlFreq = adjacencyUtilities.getDownlinkFrequency(lnadjw);
            // then
            assert.equal(dlFreq.name, 'uTargetFreq');
            assert.equal(dlFreq.value, 69);
        });

        it('should get downlink frequency from lncel', () => {
            // given
            const lncel = getFakeLNCEL();
            // when
            const dlFreq = adjacencyUtilities.getDownlinkFrequency(lncel);
            // then
            assert.equal(dlFreq.name, 'earfcnDL');
            assert.equal(dlFreq.value, 123);
        });

        it('should return null if object does not have downlink frequency parameter', () => {
            // given
            const someObject = {getDecodedParameters: () => {}};
            // when
            const dlFreq = adjacencyUtilities.getDownlinkFrequency(someObject);
            // then
            assert.isNull(dlFreq);
        });
    });

    describe('getting uplink frequency', () => {
        it('should get uplink frequency from lnadjl', () => {
            // given
            const lnadjl = getFakeLNADJL();
            // when
            const ulFreq = adjacencyUtilities.getUplinkFrequency(lnadjl);
            // then
            assert.equal(ulFreq.name, 'fUlEarfcn');
            assert.equal(ulFreq.value, 666);
        });

        it('should get uplink frequency from lncel', () => {
            // given
            const lncel = getFakeLNCEL();
            // when
            const ulFreq = adjacencyUtilities.getUplinkFrequency(lncel);
            // then
            assert.equal(ulFreq.name, 'earfcnUL');
            assert.equal(ulFreq.value, 123);
        });

        it('should return null if object does not have uplink frequency parameter', () => {
            // given
            const someObject = {getDecodedParameters: () => {}};
            // when
            const ulFreq = adjacencyUtilities.getUplinkFrequency(someObject);
            // then
            assert.isNull(ulFreq);
        });
    });

    describe('getting mnc & mcc', () => {
        it('should get mnc from lnadjl, lnadjw, lnadjg', () => {
            // given
            const lnadjl = getFakeLNADJL();
            const lnadjw = getFakeLNADJW();
            const lnadjg = getFakeLNADJG();
            // when
            const lnadjlMnc = adjacencyUtilities.getMnc(lnadjl);
            const lnadjwMnc = adjacencyUtilities.getMnc(lnadjw);
            const lnadjgMnc = adjacencyUtilities.getMnc(lnadjg);
            // then
            assert.equal(lnadjlMnc.name, 'mnc');
            assert.equal(lnadjwMnc.name, 'mnc');
            assert.equal(lnadjgMnc.name, 'mnc');

            assert.equal(lnadjlMnc.value, 997);
            assert.equal(lnadjwMnc.value, 997);
            assert.equal(lnadjgMnc.value, 997);
        });

        it('should get mcc from lnadjl, lnadjw, lnadjg', () => {
            // given
            const lnadjl = getFakeLNADJL();
            const lnadjw = getFakeLNADJW();
            const lnadjg = getFakeLNADJG();
            // when
            const lnadjlMcc = adjacencyUtilities.getMcc(lnadjl);
            const lnadjwMcc = adjacencyUtilities.getMcc(lnadjw);
            const lnadjgMcc = adjacencyUtilities.getMcc(lnadjg);
            // then
            assert.equal(lnadjlMcc.name, 'mcc');
            assert.equal(lnadjwMcc.name, 'mcc');
            assert.equal(lnadjgMcc.name, 'mcc');

            assert.equal(lnadjlMcc.value, 13);
            assert.equal(lnadjwMcc.value, 13);
            assert.equal(lnadjgMcc.value, 13);
        });

        it('should return null if object does not have mnc or mcc', () => {
            // given
            const someObject = {getDecodedParameters: () => {}};
            // when
            const mnc = adjacencyUtilities.getMnc(someObject);
            const mcc = adjacencyUtilities.getMcc(someObject);
            // then
            assert.isNull(mnc);
            assert.isNull(mcc);
        });
    });

    function getFakeLNADJL() {
        return {
            className: 'LNADJL',
            getDecodedParameters: function () {
                return {
                    fDlEarfcn: 666,
                    fUlEarfcn: 666,
                    ecgiPlmnId: {
                        mnc: 997,
                        mcc: 13
                    }
                };
            }
        };
    }

    function getFakeLNADJW() {
        return {
            className: 'LNADJW',
            getDecodedParameters: function () {
                return {
                    uTargetFreq: 69,
                    primaryPlmnId: {
                        mnc: 997,
                        mcc: 13
                    }
                };
            }
        };
    }

    function getFakeLNADJG() {
        return {
            className: 'LNADJG',
            getDecodedParameters: function () {
                return {
                    uTargetFreq: 69,
                    mnc: 997,
                    mcc: 13
                };
            }
        };
    }

    function getFakeLNCEL() {
        return {
            className: 'LNCEL',
            getDecodedParameters: function () {
                return {
                    earfcnDL: 123,
                    earfcnUL: 123
                };
            }
        };
    }
});

