const adjacencyValidation = require('../../src/adjacency/adjacencyValidation');
const ImTree = require('../../src/imtree');

let validationData;
let tree;

describe('Service: Adjacency validation', () => {

    beforeEach(() => {
        tree = new ImTree();
    });

    describe('Validating adjacencies', () => {
        let correct;
        let error;
        let warning;
        let info;

        beforeEach(() => {
            correct = adjacencyValidation.validationStatus.CORRECT;
            error = adjacencyValidation.validationStatus.ERROR;
            warning = adjacencyValidation.validationStatus.WARNING;
            info = adjacencyValidation.validationStatus.INFO;
        });

        let lncel1;
        let lteValidation;
        let utranValidation;
        let geranValidation;

        const BTS_CONF_DIST_NAME = '/MRBTS-1/RAT-1/BTS_CONF-1';
        const LNBTS_A_DIST_NAME = `${BTS_CONF_DIST_NAME}/NP-1/SCF-1/MRBTS_A-1/LNBTS_A-1`;

        describe('Validation of correct data', () => {
            beforeEach(() => {
                // Given
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-1', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-2', {ecgiLcrId: 2, ecgiAdjEnbId: 2});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-1', {uTargetCid: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-2', {uTargetCid: 2});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-1', {gTargetCi: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-2', {gTargetCi: 2});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-2', {ecgiLcrId: 2, ecgiAdjEnbId: 2});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1', {uTargetCid: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-2', {uTargetCid: 2});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1', {gTargetCi: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-2', {gTargetCi: 2});

                // When
                validationData = adjacencyValidation.validateAdjacencies(tree);
                lncel1 = validationData.localCellValidations[0];
                lteValidation = lncel1.technologyValidations[0];
                utranValidation = lncel1.technologyValidations[1];
                geranValidation = lncel1.technologyValidations[2];
            });

            it('should validate correctly', () => {
                expect(validationData.cellGroupStatus).to.be.equal(correct);
                expect(lncel1.localCellStatus).to.be.equal(correct);
            });

            describe('Correct LTE validation', () => {
                it('should validate the LTE technology correctly', () => {
                    expect(lteValidation.technology).to.be.equal('LTE');
                    expect(lteValidation.technologyRelationsStatus).to.be.equal(correct);
                });

                it('should validate LTE relations correctly', () => {
                    // Given
                    const lteRelations = lteValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(lteRelations, 2, 0, 0);
                    expect(lteRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-1');
                    expect(lteRelations.corrects[0].matchingCell).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1');
                    expect(lteRelations.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-2');
                    expect(lteRelations.corrects[1].matchingCell).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-2');
                });

                it('should validate LTE cells correctly', () => {
                    // Given
                    const lteCells = lteValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(lteCells, 2, 0);
                    expect(lteCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1');
                    expect(lteCells.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-2');
                });

            });

            describe('Correct UTRAN validation', () => {
                it('should validate the UTRAN technology correctly', () => {
                    expect(utranValidation.technology).to.be.equal('UTRAN');
                    expect(utranValidation.technologyRelationsStatus).to.be.equal(correct);
                });

                it('should validate UTRAN relations correctly', () => {
                    // Given
                    const utranRelations = utranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(utranRelations, 2, 0, 0);
                    expect(utranRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-1');
                    expect(utranRelations.corrects[0].matchingCell).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1');
                    expect(utranRelations.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-2');
                    expect(utranRelations.corrects[1].matchingCell).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-2');
                });

                it('should validate UTRAN cells correctly', () => {
                    // Given
                    const utranCells = utranValidation.cellStatuses;

                    //Then
                    expectCellGroupLengths(utranCells, 2, 0);
                    expect(utranCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1');
                    expect(utranCells.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-2');
                });
            });

            describe('Correct GERAN validation', () => {
                it('should validate the GERAN technology correctly', () => {
                    expect(geranValidation.technology).to.be.equal('GERAN');
                    expect(geranValidation.technologyRelationsStatus).to.be.equal(correct);
                });

                it('should validate GERAN relations correctly', () => {
                    // Given
                    const geranRelations = geranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(geranRelations, 2, 0, 0);
                    expect(geranRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-1');
                    expect(geranRelations.corrects[0].matchingCell).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1');
                    expect(geranRelations.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-2');
                    expect(geranRelations.corrects[1].matchingCell).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-2');
                });

                it('should validate GERAN cells correctly', () => {
                    // Given
                    const geranCells = geranValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(geranCells, 2, 0);
                    expect(geranCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1');
                    expect(geranCells.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-2');
                });
            });

        });

        describe('Validation of data with errors', () => {
            beforeEach(() => {
                // Given
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-1', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-2', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-3', {ecgiLcrId: 3, ecgiAdjEnbId: 3});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-1', {uTargetCid: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-2', {uTargetCid: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-3', {uTargetCid: 3});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-1', {gTargetCi: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-2', {gTargetCi: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-3', {gTargetCi: 3});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-2', {ecgiLcrId: 2, ecgiAdjEnbId: 2});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1', {uTargetCid: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-2', {uTargetCid: 2});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1', {gTargetCi: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-2', {gTargetCi: 2});

                // When
                validationData = adjacencyValidation.validateAdjacencies(tree);
                lncel1 = validationData.localCellValidations[0];
                lteValidation = lncel1.technologyValidations[0];
                utranValidation = lncel1.technologyValidations[1];
                geranValidation = lncel1.technologyValidations[2];
            });

            it('should validate with errors', () => {
                // Then
                expect(validationData.cellGroupStatus).to.be.equal(error);
                expect(lncel1.localCellStatus).to.be.equal(error);
            });

            describe('LTE validation with errors', () => {
                it('should validate the LTE technology as having errors', () => {
                    expect(lteValidation.technologyRelationsStatus).to.be.equal(error);
                });

                it('should validate LTE relations as having errors', () => {
                    // Given
                    const lteRelations = lteValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(lteRelations, 1, 1, 1);
                    expect(lteRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-1');
                    expect(lteRelations.errors[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-2');
                    expect(lteRelations.warnings[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-3');
                });

                it('should validate LTE cells as having errors', () => {
                    // Given
                    const lteCells = lteValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(lteCells, 1, 1);
                    expect(lteCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1');
                    expect(lteCells.infos[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-2');
                });

            });

            describe('UTRAN validation with errors', () => {
                it('should validate the UTRAN technology as having errors', () => {
                    expect(utranValidation.technologyRelationsStatus).to.be.equal(error);
                });

                it('should validate UTRAN relations as having errors', () => {
                    // Given
                    const utranRelations = utranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(utranRelations, 1, 1, 1);
                    expect(utranRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-1');
                    expect(utranRelations.errors[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-2');
                    expect(utranRelations.warnings[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-3');
                });

                it('should validate UTRAN cells as having errors', () => {
                    // Given
                    const utranCells = utranValidation.cellStatuses;

                    //Then
                    expectCellGroupLengths(utranCells, 1, 1);
                    expect(utranCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1');
                    expect(utranCells.infos[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-2');
                });
            });

            describe('GERAN validation with errors', () => {
                it('should validate the GERAN technology as having errors', () => {
                    expect(geranValidation.technologyRelationsStatus).to.be.equal(error);
                });

                it('should validate GERAN relations as having errors', () => {
                    // Given
                    const geranRelations = geranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(geranRelations, 1, 1, 1);
                    expect(geranRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-1');
                    expect(geranRelations.errors[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-2');
                    expect(geranRelations.warnings[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-3');
                });

                it('should validate GERAN cells as having errors', () => {
                    // Given
                    const geranCells = geranValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(geranCells, 1, 1);
                    expect(geranCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1');
                    expect(geranCells.infos[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-2');
                });
            });

        });

        describe('Validation of data with warnings', () => {
            beforeEach(() => {
                // Given
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-1', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-2', {ecgiLcrId: 3, ecgiAdjEnbId: 3});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-1', {uTargetCid: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-2', {uTargetCid: 3});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-1', {gTargetCi: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-2', {gTargetCi: 3});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-2', {ecgiLcrId: 2, ecgiAdjEnbId: 2});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1', {uTargetCid: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-2', {uTargetCid: 2});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1', {gTargetCi: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-2', {gTargetCi: 2});

                // When
                validationData = adjacencyValidation.validateAdjacencies(tree);
                lncel1 = validationData.localCellValidations[0];
                lteValidation = lncel1.technologyValidations[0];
                utranValidation = lncel1.technologyValidations[1];
                geranValidation = lncel1.technologyValidations[2];
            });

            it('should validate with warnings', () => {
                // Then
                expect(validationData.cellGroupStatus).to.be.equal(warning);
                expect(lncel1.localCellStatus).to.be.equal(warning);
            });

            describe('LTE validation with warnings', () => {
                it('should validate the LTE technology as having warnings', () => {
                    expect(lteValidation.technologyRelationsStatus).to.be.equal(warning);
                });

                it('should validate LTE relations as having warnings', () => {
                    // Given
                    const lteRelations = lteValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(lteRelations, 1, 1, 0);
                    expect(lteRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-1');
                    expect(lteRelations.warnings[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-2');
                });

                it('should validate LTE cells as having warnings', () => {
                    // Given
                    const lteCells = lteValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(lteCells, 1, 1);
                    expect(lteCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1');
                    expect(lteCells.infos[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-2');
                });

            });

            describe('UTRAN validation with warnings', () => {
                it('should validate the UTRAN technology as having warnings', () => {
                    expect(utranValidation.technologyRelationsStatus).to.be.equal(warning);
                });

                it('should validate UTRAN relations as having warnings', () => {
                    // Given
                    const utranRelations = utranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(utranRelations, 1, 1, 0);
                    expect(utranRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-1');
                    expect(utranRelations.warnings[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-2');
                });

                it('should validate UTRAN cells as having warnings', () => {
                    // Given
                    const utranCells = utranValidation.cellStatuses;

                    //Then
                    expectCellGroupLengths(utranCells, 1, 1);
                    expect(utranCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1');
                    expect(utranCells.infos[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-2');
                });
            });

            describe('GERAN validation with warnings', () => {
                it('should validate the GERAN technology as having warnings', () => {
                    expect(geranValidation.technologyRelationsStatus).to.be.equal(warning);
                });

                it('should validate GERAN relations as having warnings', () => {
                    // Given
                    const geranRelations = geranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(geranRelations, 1, 1, 0);
                    expect(geranRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-1');
                    expect(geranRelations.warnings[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-2');
                });

                it('should validate GERAN cells as having warnings', () => {
                    // Given
                    const geranCells = geranValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(geranCells, 1, 1);
                    expect(geranCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1');
                    expect(geranCells.infos[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-2');
                });
            });

        });

        describe('Validation of data with information', () => {
            beforeEach(() => {
                // Given
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1');

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1', {uTargetCid: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1', {gTargetCi: 1});

                // When
                validationData = adjacencyValidation.validateAdjacencies(tree);
                lncel1 = validationData.localCellValidations[0];
                lteValidation = lncel1.technologyValidations[0];
                utranValidation = lncel1.technologyValidations[1];
                geranValidation = lncel1.technologyValidations[2];
            });

            it('should validate with information', () => {
                // Then
                expect(validationData.cellGroupStatus).to.be.equal(info);
                expect(lncel1.localCellStatus).to.be.equal(info);
            });

            describe('LTE validation with information', () => {
                it('should validate the LTE technology as having information', () => {
                    expect(lteValidation.technologyRelationsStatus).to.be.equal(info);
                });

                it('should validate LTE relations as having information', () => {
                    // Given
                    const lteRelations = lteValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(lteRelations, 0, 0, 0);
                });

                it('should validate LTE cells as having information', () => {
                    // Given
                    const lteCells = lteValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(lteCells, 0, 1);
                    expect(lteCells.infos[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1');
                });

            });

            describe('UTRAN validation with information', () => {
                it('should validate the UTRAN technology as having information', () => {
                    expect(utranValidation.technologyRelationsStatus).to.be.equal(info);
                });

                it('should validate UTRAN relations as having information', () => {
                    // Given
                    const utranRelations = utranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(utranRelations, 0, 0, 0);
                });

                it('should validate UTRAN cells as having information', () => {
                    // Given
                    const utranCells = utranValidation.cellStatuses;

                    //Then
                    expectCellGroupLengths(utranCells, 0, 1);
                    expect(utranValidation.cellStatuses.infos[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1');
                });
            });

            describe('GERAN validation with information', () => {
                it('should validate the GERAN technology as having information', () => {
                    expect(geranValidation.technologyRelationsStatus).to.be.equal(info);
                });

                it('should validate GERAN relations as having information', () => {
                    // Given
                    const geranRelations = geranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(geranRelations, 0, 0, 0);
                });

                it('should validate GERAN cells as having information', () => {
                    // Given
                    const geranCells = geranValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(geranCells, 0, 1);
                    expect(geranValidation.cellStatuses.infos[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1');
                });
            });

        });

        describe('Validation without any external cells', () => {
            beforeEach(() => {
                // Given
                tree.mergeNode('/MRBTS-1/RAT-1/BTS_CONF-1', {currentNp: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1'});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1');
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1`);

                // When
                validationData = adjacencyValidation.validateAdjacencies(tree);
            });

            it('should validate correctly without external cells', () => {
                // Then
                expect(validationData.cellGroupStatus).to.be.equal(correct);
                expect(validationData.localCellValidations.length).to.be.equal(0);
            });

        });

        describe('Validation using A objects', () => {
            beforeEach(() => {
                // Given
                tree.mergeNode(BTS_CONF_DIST_NAME, {currentNp: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1'});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNREL_A-1`, {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNREL_A-2`, {ecgiLcrId: 2, ecgiAdjEnbId: 2});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELW_A-1`, {uTargetCid: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELW_A-2`, {uTargetCid: 2});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELG_A-1`, {gTargetCi: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELG_A-2`, {gTargetCi: 2});

                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJ_A-1/LNADJL_A-1`, {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJ_A-1/LNADJL_A-2`, {ecgiLcrId: 2, ecgiAdjEnbId: 2});

                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJW_A-1`, {uTargetCid: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJW_A-2`, {uTargetCid: 2});

                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJG_A-1`, {gTargetCi: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJG_A-2`, {gTargetCi: 2});

                // When
                validationData = adjacencyValidation.validateAdjacencies(tree);
                lncel1 = validationData.localCellValidations[0];
                lteValidation = lncel1.technologyValidations[0];
                utranValidation = lncel1.technologyValidations[1];
                geranValidation = lncel1.technologyValidations[2];
            });

            it('should validate correctly using A objects', () => {
                // Then
                expect(validationData.cellGroupStatus).to.be.equal(correct);
                expect(lncel1.localCellStatus).to.be.equal(correct);
            });

            describe('LTE validation using A objects', () => {
                it('should validate the LTE technology as being correct', () => {
                    expect(lteValidation.technologyRelationsStatus).to.be.equal(correct);
                });

                it('should validate LTE relations as being correct', () => {
                    // Given
                    const lteRelations = lteValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(lteRelations, 2, 0, 0);
                    expect(lteRelations.corrects[0].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNREL_A-1`);
                    expect(lteRelations.corrects[1].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNREL_A-2`);
                });

                it('should validate LTE cells as being correct', () => {
                    // Given
                    const lteCells = lteValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(lteCells, 2, 0);
                    expect(lteCells.corrects[0].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNADJ_A-1/LNADJL_A-1`);
                    expect(lteCells.corrects[1].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNADJ_A-1/LNADJL_A-2`);

                });

            });

            describe('UTRAN validation using A objects', () => {
                it('should validate the UTRAN technology as being correct', () => {
                    expect(utranValidation.technologyRelationsStatus).to.be.equal(correct);
                });

                it('should validate UTRAN relations as being correct', () => {
                    // Given
                    const utranRelations = utranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(utranRelations, 2, 0, 0);
                    expect(utranRelations.corrects[0].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELW_A-1`);
                    expect(utranRelations.corrects[1].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELW_A-2`);
                });

                it('should validate UTRAN cells as being correct', () => {
                    // Given
                    const utranCells = utranValidation.cellStatuses;

                    //Then
                    expectCellGroupLengths(utranCells, 2, 0);
                    expect(utranCells.corrects[0].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNADJW_A-1`);
                    expect(utranCells.corrects[1].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNADJW_A-2`);
                });
            });

            describe('GERAN validation using A objects', () => {
                it('should validate the GERAN technology as being correct', () => {
                    expect(geranValidation.technologyRelationsStatus).to.be.equal(correct);
                });

                it('should validate GERAN relations as being correct', () => {
                    // Given
                    const geranRelations = geranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(geranRelations, 2, 0, 0);
                    expect(geranRelations.corrects[0].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELG_A-1`);
                    expect(geranRelations.corrects[1].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELG_A-2`);
                });

                it('should validate GERAN cells as being correct', () => {
                    // Given
                    const geranCells = geranValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(geranCells, 2, 0);
                    expect(geranCells.corrects[0].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNADJG_A-1`);
                    expect(geranCells.corrects[1].distName).to.be.equal(`${LNBTS_A_DIST_NAME}/LNADJG_A-2`);
                });
            });

        });

        describe('Validation using L objects when A objects are also present', () => {
            beforeEach(() => {
                // Given
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-1', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-2', {ecgiLcrId: 2, ecgiAdjEnbId: 2});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-1', {uTargetCid: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-2', {uTargetCid: 2});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-1', {gTargetCi: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-2', {gTargetCi: 2});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-2', {ecgiLcrId: 2, ecgiAdjEnbId: 2});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1', {uTargetCid: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-2', {uTargetCid: 2});

                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1', {gTargetCi: 1});
                tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-2', {gTargetCi: 2});

                tree.mergeNode(BTS_CONF_DIST_NAME, {currentNp: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1'});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNREL_A-1`, {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNREL_A-2`, {ecgiLcrId: 2, ecgiAdjEnbId: 2});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELW_A-1`, {uTargetCid: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELW_A-2`, {uTargetCid: 2});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELG_A-1`, {gTargetCi: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNCEL_A-1/LNRELG_A-2`, {gTargetCi: 2});

                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJ_A-1/LNADJL_A-1`, {ecgiLcrId: 1, ecgiAdjEnbId: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJ_A-1/LNADJL_A-2`, {ecgiLcrId: 2, ecgiAdjEnbId: 2});

                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJW_A-1`, {uTargetCid: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJW_A-2`, {uTargetCid: 2});

                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJG_A-1`, {gTargetCi: 1});
                tree.mergeNode(`${LNBTS_A_DIST_NAME}/LNADJG_A-2`, {gTargetCi: 2});

                // When
                validationData = adjacencyValidation.validateAdjacencies(tree);
                lncel1 = validationData.localCellValidations[0];
                lteValidation = lncel1.technologyValidations[0];
                utranValidation = lncel1.technologyValidations[1];
                geranValidation = lncel1.technologyValidations[2];
            });

            it('should validate correctly using L objects', () => {
                // Then
                expect(validationData.cellGroupStatus).to.be.equal(correct);
                expect(lncel1.localCellStatus).to.be.equal(correct);
            });

            describe('LTE validation using L objects', () => {
                it('should validate the LTE technology as being correct', () => {
                    expect(lteValidation.technologyRelationsStatus).to.be.equal(correct);
                });

                it('should validate LTE relations as being correct', () => {
                    // Given
                    const lteRelations = lteValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(lteRelations, 2, 0, 0);
                    expect(lteRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-1');
                    expect(lteRelations.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-2');
                });

                it('should validate LTE cells as being correct', () => {
                    // Given
                    const lteCells = lteValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(lteCells, 2, 0);
                    expect(lteCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1');
                    expect(lteCells.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-2');
                });

            });

            describe('UTRAN validation using L objects', () => {
                it('should validate the UTRAN technology as being correct', () => {
                    expect(utranValidation.technologyRelationsStatus).to.be.equal(correct);
                });

                it('should validate UTRAN relations as being correct', () => {
                    // Given
                    const utranRelations = utranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(utranRelations, 2, 0, 0);
                    expect(utranRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-1');
                    expect(utranRelations.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-2');
                });

                it('should validate UTRAN cells as being correct', () => {
                    // Given
                    const utranCells = utranValidation.cellStatuses;

                    //Then
                    expectCellGroupLengths(utranCells, 2, 0);
                    expect(utranCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1');
                    expect(utranCells.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-2');
                });
            });

            describe('GERAN validation using L objects', () => {
                it('should validate the GERAN technology as being correct', () => {
                    expect(geranValidation.technologyRelationsStatus).to.be.equal(correct);
                });

                it('should validate GERAN relations as being correct', () => {
                    // Given
                    const geranRelations = geranValidation.relationStatuses;

                    // Then
                    expectRelationGroupLengths(geranRelations, 2, 0, 0);
                    expect(geranRelations.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-1');
                    expect(geranRelations.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-2');
                });

                it('should validate GERAN cells as being correct', () => {
                    // Given
                    const geranCells = geranValidation.cellStatuses;

                    // Then
                    expectCellGroupLengths(geranCells, 2, 0);
                    expect(geranCells.corrects[0].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1');
                    expect(geranCells.corrects[1].distName).to.be.equal('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-2');
                });
            });

        });

        it('should not fail without local cells', () => {
            // Given
            tree.mergeNode('/MRBTS-1/RAT-1/BTS_CONF-1', {currentNp: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1'});

            // When
            validationData = adjacencyValidation.validateAdjacencies(tree);

            // Then
            expect(validationData.cellGroupStatus).to.be.deep.equal(correct);
            expect(validationData.localCellValidations.length).to.be.deep.equal(0);
        });

        it('should not fail without objects', () => {
            // When
            validationData = adjacencyValidation.validateAdjacencies(tree);

            // Then
            expect(validationData.cellGroupStatus).to.be.deep.equal(correct);
            expect(validationData.localCellValidations.length).to.be.deep.equal(0);
        });

        function expectRelationGroupLengths(relationGroups, correctsLength, warningsLength, errorsLength) {
            expect(relationGroups.corrects.length).to.be.equal(correctsLength);
            expect(relationGroups.warnings.length).to.be.equal(warningsLength);
            expect(relationGroups.errors.length).to.be.equal(errorsLength);
        }

        function expectCellGroupLengths(cellGroups, correctsLength, infosLength) {
            expect(cellGroups.corrects.length).to.be.equal(correctsLength);
            expect(cellGroups.infos.length).to.be.equal(infosLength);
        }

    });

    describe('Filtering validation data', () => {
        beforeEach(() => {
            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1', {earfcnDL: 666});

            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-1', {ecgiLcrId: 1, ecgiAdjEnbId: 1});
            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-2', {ecgiLcrId: 2, ecgiAdjEnbId: 2});
            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-1', {uTargetCid: 1});
            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELW-2', {uTargetCid: 2});
            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-1', {gTargetCi: 1});
            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNRELG-2', {gTargetCi: 2});

            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1', {
                ecgiLcrId: 1,
                ecgiAdjEnbId: 1,
                fDlEarfcn: 666
            });
            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-2', {
                ecgiLcrId: 2,
                ecgiAdjEnbId: 2,
                fDlEarfcn: 111
            });

            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-1', {uTargetCid: 1, mnc: 666});
            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJW-2', {uTargetCid: 2, mnc: 222});

            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-1', {gTargetCi: 1, mnc: 333});
            tree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJG-2', {gTargetCi: 2, mnc: 333});

            validationData = adjacencyValidation.validateAdjacencies(tree);
        });

        it('should filter data', () => {
            // given
            const filter = '666';
            // when
            const filteredData = adjacencyValidation.filterValidationData(filter, validationData);
            // then
            assert.equal(filteredData.localCellValidations[0].localCell.distName, '/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1');
            const lteTech = filteredData.localCellValidations[0].technologyValidations[0];
            assert.equal(lteTech.cellStatuses.corrects.length, 1);
            assert.equal(lteTech.cellStatuses.corrects[0].distName, '/MRBTS-1/RAT-1/LNBTS-1/CPLANE-1/LNADJ-1/LNADJL-1');
            assert.equal(lteTech.relationStatuses.corrects[0].distName, '/MRBTS-1/RAT-1/LNBTS-1/LNCEL-1/LNREL-1');
        });
    });
});
