const ADMIN_API_STATUSES = require('../../src/adminApi/adminApiStatuses');
const intervalNotifier = require('../../src/adminApi/intervalNotifier');

describe('ADMIN API: adminApi', function () {
    var socketMock;
    var adminApi, sandbox;
    beforeEach(function () {
        sandbox = sinon.sandbox.create();
        adminApi = require('../../src/adminApi/adminApi');
        mockIntervalNotifierMethods();

        socketMock = {
            send: sinon.spy()
        };
    });

    afterEach(function () {
        restoreIntervalNotifier();
    });

    function mockIntervalNotifierMethods() {
        sandbox.stub(intervalNotifier, 'start');
        sandbox.stub(intervalNotifier, 'restart');
        sandbox.stub(intervalNotifier, 'stop');
    }

    function restoreIntervalNotifier() {
        sandbox.restore();
    }

    it('should not run any command', function () {
        // given
        const remoteHandler = addHandlerToRemoteApi(sinon.stub());
        const data = createProcedureMsg(1, 'UnknownProcedure', 'UnknownProcedure', 'dist-1/name-1');
        // when
        adminApi.handle(JSON.stringify(data), socketMock);
        // then
        assert.equal(remoteHandler.prepare.callCount, 0);
    });

    it('should run procedure command - request and response must have the same requestId', function () {
        // given
        const requestId = 1;
        const requestStatus = 'requestStatus';
        const payload = 'payload';
        const data = createProcedureMsg(requestId, 'procedure', 'SomeProcedure', 'dist-1/name-1');
        const remoteHandler = addHandlerToRemoteApi(function (dataMsg, completionListener) {
            completionListener(requestStatus, payload);
        });
        adminApi.addHandler('procedure', remoteHandler.prepare);

        // when
        adminApi.handle(JSON.stringify(data), socketMock);

        // then
        assert.equal(remoteHandler.prepare.callCount, 1);
        assert(remoteHandler.prepare.calledWith(data.parameters, sinon.match.func));
        assert(socketMock.send.calledWith(JSON.stringify({
            requestId: requestId,
            requestStatus: requestStatus,
            payload: payload
        })));
    });

    it('should start and stop interval notifier when completed status occured', function () {
        testIntervalNotifierWithStatus(ADMIN_API_STATUSES.COMPLETED);
    });

    it('should start and stop interval notifier when failed status occured', function () {
        testIntervalNotifierWithStatus(ADMIN_API_STATUSES.FAILED);
    });

    function testIntervalNotifierWithStatus(requestStatus) {
        // given
        const data = createProcedureMsg(1, 'procedure', 'SomeProcedure', 'dist-1/name-1');
        const remoteHandler = addHandlerToRemoteApi(function (dataMsg, completionListener) {
            completionListener(requestStatus, 'payload');
        });
        adminApi.addHandler('procedure', remoteHandler.prepare);

        // when
        adminApi.handle(JSON.stringify(data), socketMock);

        // then
        assert.equal(intervalNotifier.start.callCount, 1);
        assert.equal(intervalNotifier.restart.callCount, 0);
        assert.equal(intervalNotifier.stop.callCount, 1);
    }

    it('should start, restart and stop interval notifier', function () {
        // given
        const data = createProcedureMsg(1, 'procedure', 'SomeProcedure', 'dist-1/name-1');
        addHandlerToRemoteApi(function (dataMsg, completionListener) {
            completionListener(ADMIN_API_STATUSES.ONGOING, 'payload');
            completionListener(ADMIN_API_STATUSES.COMPLETED, 'payload');
        });

        // when
        adminApi.handle(JSON.stringify(data), socketMock);
        // then
        assert.equal(intervalNotifier.start.callCount, 1);
        assert.equal(intervalNotifier.restart.callCount, 1);
        assert.equal(intervalNotifier.stop.callCount, 1);
    });

    it('should let interval notifier notify once', function (done) {
        // given
        restoreIntervalNotifier();
        const requestId = 1;
        const notifyTimeMs = 10;
        const data = createProcedureMsg(requestId, 'procedure', 'SomeProcedure', 'dist-1/name-1');
        addHandlerToRemoteApi(function (dataMsg, completionListener) {
            setTimeout(function () {
                // then
                assert(socketMock.send.calledOnce);
                assert(socketMock.send.calledWith(JSON.stringify({
                    requestId: requestId,
                    requestStatus: ADMIN_API_STATUSES.ONGOING,
                    payload: 'Waiting for response'
                })));

                completionListener(ADMIN_API_STATUSES.COMPLETED, 'payload');
                assert(socketMock.send.calledTwice);
                assert(socketMock.send.calledWith(JSON.stringify({
                    requestId: requestId,
                    requestStatus: ADMIN_API_STATUSES.COMPLETED,
                    payload: 'payload'
                })));
                done();
            }, notifyTimeMs + 1);
        });

        // when
        adminApi.setNotifierInterval(notifyTimeMs);
        adminApi.handle(JSON.stringify(data), socketMock);
    });

    it('should sent failure when received invalid message format - no all necessary data', function() {
        // given
        const notEnoughData = {
            requestId: 1,
            requestType: 'someType'
        };
        // when
        adminApi.handle(JSON.stringify(notEnoughData), socketMock);
        // then
        assert(socketMock.send.calledWith(JSON.stringify({
            requestId: null,
            requestStatus: ADMIN_API_STATUSES.FAILED ,
            payload: 'Missing required property: parameters'
        })));
    });

    it('should sent failure when received invalid message format - invalid data', function() {
        // given
        const notEnoughData = {
            requestId: 'invalidRequestId',
            requestType: 'someType',
            parameters: {}
        };
        // when
        adminApi.handle(JSON.stringify(notEnoughData), socketMock);
        // then
        assert(socketMock.send.calledWith(JSON.stringify({
            requestId: null,
            requestStatus: ADMIN_API_STATUSES.FAILED ,
            payload: '/requestId, Invalid type: string (expected number)'
        })));
    });

    function addHandlerToRemoteApi(prepareFunction) {
        const remoteHandler = {
            prepare: sinon.spy(prepareFunction)
        };
        adminApi.addHandler('procedure', remoteHandler.prepare);
        return remoteHandler;
    }

    function createProcedureMsg(requestId, requestType, procedureName, distName) {
        return {
            requestId: requestId,
            requestType: requestType,
            parameters: {
                name: procedureName,
                parameters: {
                    distName: distName
                }
            }
        };
    }
});
