const Promise = require('bluebird');
const ADMIN_API_STATUSES = require('../../../src/adminApi/adminApiStatuses');
const abnormalitiesFactory = require('../../../src/adminApi/infoModel/abnormalities');
const testUtils = require('../adminApi.test.util');
const alarms = require('../../../src/abnormalities/alarms');

describe('ADMIN API: abnormalities', () => {

    let abnormalities,
    connectionInitializerMock,
    alarmsMonitorMock,
    reasoningStatusesMock;

    beforeEach(() => {
        connectionInitializerMock = testUtils.createConnectionInitializerMock();
        abnormalities = abnormalitiesFactory({
            connectionInitializer: connectionInitializerMock
        });
    });

    it('should notify when message format is invalid - no type', done => {
        // given
        const completionListener = sinon.spy();
        const msgData = {
            name: 'Block',
            parameters: {}
        };
        // when
        abnormalities.execute(msgData, completionListener)
            .then(() => done(new Error('This promise should not be resolved')))
            .catch(() => {
                // then
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    '/parameters, Missing required property: type'
                ).calledOnce);
                done();
            });
    });

    it('should notify when message format is invalid - type as number', done => {
        // given
        const completionListener = sinon.spy();
        const msgData = createMsgData(1234);
        // when
        abnormalities.execute(msgData, completionListener)
            .then(() => done(new Error('This promise should not be resolved')))
            .catch(() => {
                // then
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    '/parameters/type, Invalid type: number (expected string)'
                ).calledOnce);
                done();
            });
    });

    it('should notify when message format is invalid - severity as number', done => {
        // given
        const completionListener = sinon.spy();
        const msgData = createMsgData('currentAlarms', 1234);
        // when
        abnormalities.execute(msgData, completionListener)
            .then(() => done(new Error('This promise should not be resolved')))
            .catch(() => {
                // then
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    '/parameters/severity, Invalid type: number (expected string)'
                ).calledOnce);
                done();
            });
    });

    describe('currentAlarms', () => {
        let existingAlarms;

        beforeEach(() => {
            existingAlarms = [
                createNewAlarm('critical', 1),
                createNewAlarm('major', 2),
                createNewAlarm('minor', 3),
                createNewAlarm('warning', 4)
            ];
            alarmsMonitorMock = createAlarmsMonitorMock({
                currentAlarms: existingAlarms
            });
            abnormalities = abnormalitiesFactory({
                alarmsMonitor: alarmsMonitorMock,
                connectionInitializer: connectionInitializerMock
            });
            connectionInitializerMock.prepare.returns(Promise.resolve());
        });

        it('should execute currentAlarms without severity', done => {
            // given
            const completionListener = sinon.spy();
            const msgData = createMsgData('currentAlarms');
            // when
            abnormalities.execute(msgData, completionListener)
                .then(() => {
                    // then
                    assert(connectionInitializerMock.connectToBts.calledOnce);
                    assert(completionListener.withArgs(ADMIN_API_STATUSES.COMPLETED, existingAlarms).calledOnce);
                }).then(done).catch(done);
        });

        it('should execute currentAlarms with severity', done => {
            // given
            const completionListener = sinon.spy();
            const msgData = createMsgData('currentAlarms', 'major');
            const majorAlarms = [existingAlarms[1]];
            // when
            abnormalities.execute(msgData, completionListener)
                .then(() => {
                    // then
                    assert(connectionInitializerMock.connectToBts.calledOnce);
                    assert(completionListener.withArgs(ADMIN_API_STATUSES.COMPLETED, majorAlarms).calledOnce);
                }).then(done).catch(done);
        });

        it('should execute currentAlarms with unrecognized severity', done => {
            // given
            const completionListener = sinon.spy();
            const msgData = createMsgData('currentAlarms', 'unrecognizedSeverity');
            // when
            abnormalities.execute(msgData, completionListener)
                .then(() => {
                    // then
                    assert(connectionInitializerMock.connectToBts.calledOnce);
                    assert(completionListener.withArgs(ADMIN_API_STATUSES.COMPLETED, []).calledOnce);
                }).then(done).catch(done);
        });

    });

    describe('historicalAlarms', () => {
        let existingAlarms;

        beforeEach(() => {
            existingAlarms = [
                createNewAlarm('critical', 1),
                createNewAlarm('major', 2),
                createNewAlarm('minor', 3),
                createNewAlarm('warning', 4)
            ];
            alarmsMonitorMock = createAlarmsMonitorMock({
                historicalAlarms: existingAlarms
            });
            abnormalities = abnormalitiesFactory({
                alarmsMonitor: alarmsMonitorMock,
                connectionInitializer: connectionInitializerMock
            });
            connectionInitializerMock.prepare.returns(Promise.resolve());
        });

        it('should execute historicalAlarms without severity', done => {
            // given
            const completionListener = sinon.spy();
            const msgData = createMsgData('historicalAlarms');
            // when
            abnormalities.execute(msgData, completionListener)
                .then(() => {
                    // then
                    assert(connectionInitializerMock.connectToBts.calledOnce);
                    assert(completionListener.withArgs(ADMIN_API_STATUSES.COMPLETED, existingAlarms).calledOnce);
                }).then(done).catch(done);
        });

        it('should execute historicalAlarms with severity', done => {
            // given
            const completionListener = sinon.spy();
            const msgData = createMsgData('historicalAlarms', 'major');
            const majorAlarms = [existingAlarms[1]];
            // when
            abnormalities.execute(msgData, completionListener)
                .then(() => {
                    // then
                    assert(connectionInitializerMock.connectToBts.calledOnce);
                    assert(completionListener.withArgs(ADMIN_API_STATUSES.COMPLETED, majorAlarms).calledOnce);
                }).then(done).catch(done);
        });

        it('should execute historicalAlarms with unrecognized severity', done => {
            // given
            const completionListener = sinon.spy();
            const msgData = createMsgData('historicalAlarms', 'unrecognizedSeverity');
            // when
            abnormalities.execute(msgData, completionListener)
                .then(() => {
                    // then
                    assert(connectionInitializerMock.connectToBts.calledOnce);
                    assert(completionListener.withArgs(ADMIN_API_STATUSES.COMPLETED, []).calledOnce);
                }).then(done).catch(done);
        });

    });

    describe('otherAbnormalities', () => {
        let exisitingAbnormalities;

        beforeEach(() => {
            exisitingAbnormalities = [
                createNewAbnormality(1),
                createNewAbnormality(2),
                createNewAbnormality(3),
                createNewAbnormality(4)
            ];
            reasoningStatusesMock = createReasoningStatusesMock({
                abnormalities: exisitingAbnormalities
            });
            abnormalities = abnormalitiesFactory({
                reasoningStatuses: reasoningStatusesMock,
                connectionInitializer: connectionInitializerMock
            });
            connectionInitializerMock.prepare.returns(Promise.resolve());
        });

        it('should execute other', done => {
            // given
            const completionListener = sinon.spy();
            const msgData = createMsgData('other');
            // when
            abnormalities.execute(msgData, completionListener)
                .then(() => {
                    // then
                    assert(connectionInitializerMock.connectToBts.calledOnce);
                    assert(completionListener.withArgs(ADMIN_API_STATUSES.COMPLETED, exisitingAbnormalities).calledOnce);
                }).then(done).catch(done);
        });

    });

    function createMsgData(abnormalitiesType, severity) {
        const message = {
            name: 'abnormalities',
            parameters: {
                type: abnormalitiesType
            }
        };
        if (severity) message.parameters.severity = severity;
        return message;
    }

    function createAlarmsMonitorMock({
        currentAlarms = [],
        historicalAlarms = []
        } = {}) {
        return {
            watchCurrentAlarms: (listener) => {
                listener(currentAlarms);
                return Promise.resolve();
            },
            watchHistoricalAlarms: (listener) => {
                listener(historicalAlarms);
                return Promise.resolve();
            }
        };
    }

    function createReasoningStatusesMock({
        abnormalities = [],
        } = {}) {
        var visibility = false;
        return {
            watch: (listener) => {
                if(visibility)
                    listener({type: 'data', data: abnormalities});
                else
                    listener({type: 'count', data: abnormalities});
                return Promise.resolve();
            },
            setVisibility: (value) => {
                visibility = value;
            }
        };
    }

    function createNewAlarm(severity, number) {
        const parameters = {
            alarmInformation: {
                alarmSeverity: severity,
                alarmText: {}
            },
            unitInformation: {}
        };
        return alarms.newAlarm(`SOME/DIST_NAME-${number}`, parameters);
    }

    function createNewAbnormality(number) {
        return {
            distName: `SOME/DIST_NAME-${number}`,
            observationTime: Date.now(),
            eventType: `EventType-${number}`
        }
    }

})
;
