const ADMIN_API_STATUSES = require('../../../../src/adminApi/adminApiStatuses');
const adjacenciesFactory = require('../../../../src/adminApi/infoModel/adjacencies/adjacencies');
const adjacencyValidation = require('../../../../src/adjacency/adjacencyValidation');
const ImTree = require('../../../../src/imtree');
const Promise = require('bluebird');

describe('getNbrDataAction', () => {
    let imTree;
    let adjacencies;
    let connectToBtsSpy;
    let connectionInitializerMock;

    function createConnectionInitializerMock() {
        connectToBtsSpy = sinon.spy(() => ({prepare: () => Promise.resolve()}));
        return {
            connectToBts: connectToBtsSpy
        };
    }

    beforeEach(() => {
        imTree = new ImTree();
        connectionInitializerMock = createConnectionInitializerMock();
        adjacencies = adjacenciesFactory({
            onlineImTree: {
                getImTree: () => Promise.resolve(imTree)
            },
            adjacencyValidation: adjacencyValidation,
            connectionInitializer: connectionInitializerMock
        });
        sinon.stub(adjacencyValidation, 'validateAdjacencies', () => Object.create({
            localCellValidations: [{
                localCell: {
                    className: 'LNCEL',
                    id: '303',
                    distName: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-64755/LNBTS_A-64755/LNCEL-303',
                    getDecodedParameters: () => Object.create({phyCellId: '203'})
                },
                technologyValidations: []
            },
                {
                    localCell: {
                        className: 'LNCEL',
                        id: '302',
                        distName: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-64755/LNBTS_A-64755/LNCEL-302',
                        getDecodedParameters: () => Object.create({phyCellId: '202'})
                    },
                    technologyValidations: [
                        {
                            technology: 'LTE',
                            cellStatuses: {
                                corrects: [{
                                    className: 'LNADJL_A',
                                    id: 23,
                                    distName: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-64755/LNBTS_A-64755/LNADJ_A-218/LNADJL_A-23',
                                    getDecodedParameters: () => Object.create({
                                        ecgiAdjEnbId: 74504,
                                        ecgiLcrId: 1,
                                        fDlEarfcn: 2325,
                                        fUlEarfcn: 20325,
                                        phyCellId: 410,
                                        sourceOfData: 'X2',
                                        ecgiPlmnId: {
                                            mcc: 310,
                                            mnc: 260
                                        }
                                    })
                                }],
                                infos: [{
                                    className: 'LNADJL_A',
                                    id: 24,
                                    distName: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-64755/LNBTS_A-64755/LNADJ_A-218/LNADJL_A-24',
                                    getDecodedParameters: () => Object.create({
                                        ecgiAdjEnbId: 74504,
                                        ecgiLcrId: 1,
                                        fDlEarfcn: 2325,
                                        fUlEarfcn: 20325,
                                        phyCellId: 410,
                                        sourceOfData: 'X2',
                                        ecgiPlmnId: {
                                            mcc: 310,
                                            mnc: 260
                                        }
                                    })
                                }]
                            },
                            relationStatuses: {
                                corrects: [{
                                    className: 'LNREL_A',
                                    matchingCell: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-64755/LNBTS_A-64755/LNADJ_A-218/LNADJL_A-23',
                                    id: 318
                                }],
                                errors: [],
                                warnings: [{
                                    className: 'LNREL_A',
                                    matchingCell: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-64755/LNBTS_A-64755/LNADJ_A-218/LNADJL_A-213',
                                    id: 318
                                }]
                            }
                        },
                        {
                            technology: 'UTRAN',
                            cellStatuses: {
                                corrects: [{
                                    className: 'LNADJW_A',
                                    id: 0,
                                    distName: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-64755/LNBTS_A-64755/LNADJW_A-0',
                                    getDecodedParameters: () => Object.create({
                                        uTargetCid: 5091,
                                        uTargetFreq: 662,
                                        uTargetLac: 10188,
                                        uTargetRac: 88,
                                        uTargetScFdd: 122,
                                        primaryPlmnId: {
                                            mcc: 310,
                                            mnc: 260
                                        }
                                    })
                                }],
                                infos: []
                            },
                            relationStatuses: {
                                corrects: [{
                                    className: 'LNRELW_A',
                                    matchingCell: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-64755/LNBTS_A-64755/LNADJW_A-0',
                                    id: 0
                                }],
                                errors: [],
                                warnings: []
                            }
                        },
                        {
                            technology: 'GERAN',
                            cellStatuses: {
                                corrects: [],
                                infos: [{
                                    className: 'LNADJG_A',
                                    id: 0,
                                    distName: '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-64755/LNBTS_A-64755/LNADJG_A-0',
                                    getDecodedParameters: () => Object.create({
                                        gTargetCi: 5091,
                                        gTargetLac: 662,
                                        gTargetRac: 10188,
                                        mcc: 88,
                                        mnc: 122
                                    })
                                }]
                            },
                            relationStatuses: {
                                corrects: [],
                                errors: [],
                                warnings: [{
                                    className: 'LNRELG_A',
                                    id: 0
                                }]
                            }
                        }
                    ]
                }]
        }));
    });

    const nbrDataFull = {
        'LNCEL-303': {},
        'LNCEL-302': {
            'LTE': [
                {
                    'STATUS': 'correct',
                    'LNREL': 'LNREL_A-318',
                    'LNADJL': 'LNADJL_A-23',
                    'CELL': {
                        'ecgiAdjEnbId': 74504,
                        'ecgiLcrId': 1,
                        'mcc': 310,
                        'mnc': 260,
                        'sourceOfData': 'X2',
                        'phyCellId': 'PCI-410',
                        'fDlEarfcn': 2325,
                        'fUlEarfcn': 20325
                    }
                },
                {
                    'STATUS': 'unused',
                    'LNREL': '',
                    'LNADJL': 'LNADJL_A-24',
                    'CELL': {
                        'ecgiAdjEnbId': 74504,
                        'ecgiLcrId': 1,
                        'mcc': 310,
                        'mnc': 260,
                        'sourceOfData': 'X2',
                        'phyCellId': 'PCI-410',
                        'fDlEarfcn': 2325,
                        'fUlEarfcn': 20325
                    }
                },
                {
                    'STATUS': 'loose end',
                    'LNREL': 'LNREL_A-318',
                    'LNADJL': '',
                    'CELL': ''
                }
            ],
            'UTRAN': [
                {
                    'STATUS': 'correct',
                    'LNREL': 'LNRELW_A-0',
                    'LNADJL': 'LNADJW_A-0',
                    'CELL': {
                        'uTargetCid': 5091,
                        'uTargetFreq': 662,
                        'uTargetLac': 10188,
                        'uTargetRac': 88,
                        'uTargetScFdd': 122,
                        'mcc': 310,
                        'mnc': 260
                    }
                }
            ],
            'GERAN': [
                {
                    'STATUS': 'unused',
                    'LNREL': '',
                    'LNADJL': 'LNADJG_A-0',
                    'CELL': {
                        'gTargetCi': 5091,
                        'gTargetLac': 662,
                        'gTargetRac': 10188,
                        'mcc': 88,
                        'mnc': 122
                    }
                },
                {
                    'STATUS': 'loose end',
                    'LNREL': 'LNRELG_A-0',
                    'LNADJL': '',
                    'CELL': ''
                }
            ]
        }
    };

    afterEach(() => {
        adjacencyValidation.validateAdjacencies.restore();
    });

    it('should return adjacent cells', done => {

        // given
        const msgData = createMsgData('getNbrData');
        // when
        adjacencies.execute(msgData, (requestStatus, payload) => {
            assert.equal(requestStatus, ADMIN_API_STATUSES.COMPLETED);
            assert.equal(JSON.stringify(payload), JSON.stringify(nbrDataFull));
            done();
        });
    });

    it('should return adjacent cells for cell given in parameter', done => {

        // given
        const msgData = createMsgDataWithParams({
            name: 'getNbrData',
            cellName: 'LNCEL-303'
        });
        // when
        adjacencies.execute(msgData, (requestStatus, payload) => {
            assert.equal(requestStatus, ADMIN_API_STATUSES.COMPLETED);
            assert.equal(JSON.stringify(payload), JSON.stringify({'LNCEL-303': {}}));
            done();
        });
    });

    it('should fail for missing cell', done => {

        // given
        const completionListener = sinon.spy();
        const msgData = createMsgDataWithParams({
            name: 'getNbrData',
            cellName: 'LNCEL_A-2'
        });
        // when
        adjacencies.execute(msgData, completionListener)
            .then(() => done(new Error('This promise should not be resolved')))
            .catch(() => {
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    'No local cells with any relations were found.'
                ).calledOnce);
                done();
            });
    });

    function createMsgData(procedureName) {
        return {
            name: 'adjacencies',
            parameters: {
                procedureName: procedureName
            }
        };
    }

    function createMsgDataWithParams(params) {
        return {
            name: 'adjacencies',
            parameters: {
                procedureName: params.name,
                cellName: params.cellName
            }
        };
    }
});
