const Promise = require('bluebird');
const ADMIN_API_STATUSES = require('../../../src/adminApi/adminApiStatuses');
const basebandFactory = require('../../../src/adminApi/infoModel/baseband');

describe('baseband', () => {

    let baseband;
    let dspsService;

    beforeEach(() => {
        dspsService = createDspsServiceMock();
        baseband = basebandFactory({
            connectionInitializer: createConnectionInitializerMock(),
            dspsService: dspsService
        });
    });

    it('should get entire baseband', (done) => {
        const msgData = createMsgData(false);
        const expectedBaseband = [
            {
                distName: 'CELL-1',
                label: 'C1',
                channels: []
            },
            {
                distName: 'CELL-2',
                label: 'C2',
                channels: []
            }
        ];
        var status, basebandInfo;
        new Promise(resolve => {
            baseband.execute(msgData, (requestStatus, payload) => {
                    status = requestStatus;
                    basebandInfo = payload;
                    resolve();
                }
            );
        })
            .then(() => {
                assert.equal(status, ADMIN_API_STATUSES.COMPLETED);
                assert.deepEqual(basebandInfo, expectedBaseband);
            })
            .then(done)
            .catch(done);


    });

    it('should get all cell names', done => {
        var status, basebandInfo;
        const msgData = createMsgData(true);
        let expectedNames = [
            'CELL-1',
            'CELL-2'
        ];

        new Promise((resolve, reject) => {
            baseband.execute(msgData, (requestStatus, payload) => {
                status = requestStatus;
                basebandInfo = payload;
                resolve();
            });
        })
            .then(() => {
                assert.equal(status, ADMIN_API_STATUSES.COMPLETED);
                assert.deepEqual(basebandInfo, expectedNames);
            })
            .then(done)
            .catch(done);

    });

    it('should get baseband for one cell', done => {
        const msgData = createMsgData(false, ['CELL-1']);
        var status, basebandInfo;
        let expectedBaseband = [
            {
                distName: 'CELL-1',
                label: 'C1',
                channels: []
            }
        ];

        new Promise((resolve, reject) => {
            baseband.execute(msgData, (requestStatus, payload) => {
                status = requestStatus;
                basebandInfo = payload;
                resolve();
            });
        })
            .then(() => {
                assert.equal(status, ADMIN_API_STATUSES.COMPLETED);
                assert.deepEqual(basebandInfo[0], expectedBaseband[0]);
                assert.deepEqual(basebandInfo, expectedBaseband);
            })
            .then(done)
            .catch(done);
    });

    it('should fail due to missing parameter', done => {
        const msgData = createMsgData();
        var status, response;
        baseband.execute(msgData, (requestStatus, payload) => {
            status = requestStatus;
            response = payload;
        });
        assert.equal(status, ADMIN_API_STATUSES.FAILED);
        assert.equal(response, '/parameters, Missing required property: onlyNames');
        done();
    });


    function createMsgData(onlyNames, cells) {
        var msg = {
            name: 'baseband',
            parameters: {
                onlyNames: onlyNames,
            }
        };

        if (cells) msg.parameters.cells = cells;

        return msg;
    }

    function createConnectionInitializerMock() {
        const connectToBtsSpy = (function () {
            return {
                prepare: () => {
                    return new Promise((resolve, reject) => {
                        resolve();
                    });
                }
            };
        });
        return {
            connectToBts: connectToBtsSpy
        };
    }

    function createDspsServiceMock() {
        var event = events.newEvent()
        return {
            watch: function (watcher) {
                event.watch(watcher);
                this.notify();
            },
            unwatch: function (watcher) {
                event.unwatch(watcher)
            },
            notify: function () {
                event.notify({
                    cells: [
                        {
                            distName: '/LONG/DIST/NAME/FOR/TESTING/CELL-1',
                            label: 'C1',
                            channels: []
                        },
                        {
                            distName: '/LONG/DIST/NAME/FOR/TESTING/CELL-2',
                            label: 'C2',
                            channels: []
                        }
                    ]
                })
            }
        };
    }
});
