const ADMIN_API_STATUSES = require('../../../src/adminApi/adminApiStatuses');
const parametersGetterFactory = require('../../../src/adminApi/infoModel/parametersGetter');
const ImTree = require('../../../src/imtree');
const objects = require('../../../src/utils/objects');
const _ = require('lodash');
const Promise = require('bluebird');

describe('ADMIN API: parametersGetter', () => {
    const MRBTS_DIST_NAME = '/MRBTS-1';
    const msgData = {
        name: 'getParameters',
        parameters: {
            distName: MRBTS_DIST_NAME,
            pattern: ''
        }
    };

    let onlineImTreeMock,
        parametersGetter,
        imTree,
        prepareStub,
        connectToBtsSpy,
        checkIfDnExistsSpy,
        completionListener;

    beforeEach(() => {
        imTree = new ImTree();
        onlineImTreeMock = {
            getImTree: sinon.stub()
        };
        const connectionInitializerMock = createConnectionInitializerMock();
        parametersGetter = parametersGetterFactory({
            onlineImTree: onlineImTreeMock,
            connectionInitializer: connectionInitializerMock
        });
        completionListener = sinon.spy();
    });

    function createConnectionInitializerMock() {
        prepareStub = sinon.stub();
        checkIfDnExistsSpy = sinon.spy(() => ({prepare: prepareStub}));
        connectToBtsSpy = sinon.spy(() => ({checkIfDnExists: checkIfDnExistsSpy}));
        return {
            connectToBts: connectToBtsSpy
        };
    }

    it('should notify when message format is invalid', (done) => {
        //given
        prepareStub.returns(Promise.resolve());
        const msgDataCorrupted = _.cloneDeep(msgData);
        msgDataCorrupted.parameters.distName = 123;
        onlineImTreeMock.getImTree.returns(Promise.resolve(imTree));
        //when
        parametersGetter.execute(msgDataCorrupted, completionListener)
            .then(() => {
                //then
                sinon.assert.notCalled(connectToBtsSpy);
                sinon.assert.calledOnce(completionListener);
                sinon.assert.calledWith(
                    completionListener,
                    ADMIN_API_STATUSES.FAILED,
                    '/parameters/distName, Invalid type: number (expected string)'
                );
            }).then(done).catch(done);
    });

    it('should notify when there is no object with specified distName', (done) => {
        //given
        prepareStub.returns(Promise.reject());
        onlineImTreeMock.getImTree.returns(Promise.resolve(imTree));
        //when
        parametersGetter.execute(msgData, completionListener)
            .then(() => {
                //then
                sinon.assert.calledOnce(connectToBtsSpy);
                sinon.assert.calledOnce(checkIfDnExistsSpy);
                sinon.assert.calledWith(checkIfDnExistsSpy, MRBTS_DIST_NAME);
                sinon.assert.calledOnce(completionListener);
                sinon.assert.calledWith(completionListener, ADMIN_API_STATUSES.FAILED);
            }).then(done).catch(done);
    });

    it('should notify when filtering is complete', (done) => {
        //given
        const parametersFixture = {
            foo: 123,
            bar: {}
        };
        prepareStub.returns(Promise.resolve());
        const node = imTree.addNode(null, 'MRBTS', 1);
        node.parameters = parametersFixture;
        onlineImTreeMock.getImTree.returns(Promise.resolve(imTree));
        const filterObjectsSpy = sinon.spy(objects, 'filterObject');
        //when
        parametersGetter.execute(msgData, completionListener)
            .then(() => {
                //then
                sinon.assert.calledOnce(connectToBtsSpy);
                sinon.assert.calledOnce(checkIfDnExistsSpy);
                sinon.assert.calledWith(checkIfDnExistsSpy, MRBTS_DIST_NAME);
                sinon.assert.calledOnce(filterObjectsSpy);
                sinon.assert.calledWith(filterObjectsSpy, parametersFixture);
                sinon.assert.calledOnce(completionListener);
                sinon.assert.calledWith(completionListener, ADMIN_API_STATUSES.COMPLETED);
            }).then(done).catch(done);
    });
});
