const ADMIN_API_STATUSES = require('../../../src/adminApi/adminApiStatuses');
const sfpMonitoringFactory = require('../../../src/adminApi/infoModel/sfpMonitoring');
const ImTree = require('../../../src/imtree');
const Promise = require('bluebird');

describe('ADMIN API: sfpMonitoring', () => {
    let sfpMonitoring, imTree, eqmL, hwtopL;
    let connectToBtsSpy;
    let prepareStub;
    let connectionInitializerMock;

    function createConnectionInitializerMock() {
        prepareStub = sinon.stub();
        connectToBtsSpy = sinon.spy(() => ({prepare: prepareStub}));
        return {
            connectToBts: connectToBtsSpy
        };
    }

    const onlineImTreeMock = {
        getImTree: sinon.stub()
    };

    beforeEach(() => {
        connectionInitializerMock = createConnectionInitializerMock();
        imTree = new ImTree();
        eqmL = imTree.mergeNode('/MRBTS-1/RAT-1/EQM_L-1', {});
        hwtopL = imTree.mergeNode('/MRBTS-1/RAT-1/BTS_L-1/HWTOP_L-1', {});
        const rmod = createModule('RMOD_L', 1);
        const smodConnector = createConnectorNode(createModule('SMOD_L', 1), 1);
        const rmodConnector1 = createConnectorNode(rmod, 1);
        const rmodConnector2 = createConnectorNode(rmod, 2);
        const bbmodConnector = createConnectorNode(createModule('BBMOD_L', 1), 1);
        createCablnkNode(rmodConnector1, smodConnector, 1, 1, 1);
        createCablnkNode(rmodConnector2, bbmodConnector, 2, 2, 1);
        createSfpNodeWithParams(smodConnector, 1, 1);
        createSfpNodeWithParams(rmodConnector1, 1, 1);
        createSfpNodeWithParams(rmodConnector2, 1, 1);
        createSfpNodeWithParams(bbmodConnector, 1, 1);
        sfpMonitoring = sfpMonitoringFactory.newInstance({
            onlineImTree: onlineImTreeMock,
            connectionInitializer: connectionInitializerMock
        });
        onlineImTreeMock.getImTree.returns(Promise.resolve(imTree));
        prepareStub.returns(Promise.resolve());
    });

    it('getLinks should return list of links', (done) => {
        const completionListener = sinon.spy();
        const msgData = createMsgData('getLinks');
        const response = [{
            cablnkDistName: '/MRBTS-1/RAT-1/BTS_L-1/HWTOP_L-1/CABLNK_L-1',
            source: '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1',
            destination: '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1'
        }, {
            cablnkDistName: '/MRBTS-1/RAT-1/BTS_L-1/HWTOP_L-1/CABLNK_L-2',
            source: '/MRBTS-1/RAT-1/EQM_L-1/BBMOD_L-1',
            destination: '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1'
        }];

        sfpMonitoring.execute(msgData, completionListener)
            .then(() => {
                assert(connectToBtsSpy.calledOnce);
                assert(completionListener.withArgs(ADMIN_API_STATUSES.COMPLETED, response).calledOnce);
            }).then(done).catch(done);
    });

    it('getLinksWithParameters should return all links with parameters', (done) => {
        const completionListener = sinon.spy();
        const msgData = createMsgData('getLinksWithParameters');

        sfpMonitoring.execute(msgData, completionListener)
            .then(() => {
                assert(connectToBtsSpy.calledOnce);
                assert(completionListener.calledOnce);
                const status = completionListener.args[0][0];
                const response = completionListener.args[0][1];

                assert.equal(status, ADMIN_API_STATUSES.COMPLETED);

                assert.isDefined(response);
                const cablink1 = response[0];
                const cablink2 = response[1];

                assert.lengthOf(cablink1.alarms, 0);
                assert.equal(cablink1.destination.name, 'SMOD_1');
                assert.equal(cablink1.destination.distName, '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1');
                assert.equal(cablink1.destination.status, 'OK');
                assert.equal(cablink1.destination.port, 1);
                assert.equal(cablink1.destination.parameters.tx.value, 0.001);
                assert.equal(cablink1.destination.parameters.rx.value, 0.001);
                assert.equal(cablink1.destination.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                assert.equal(cablink1.source.name, 'RMOD_1');
                assert.equal(cablink1.source.distName, '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1');
                assert.equal(cablink1.source.status, 'OK');
                assert.equal(cablink1.source.port, 1);
                assert.equal(cablink1.source.parameters.tx.value, 0.001);
                assert.equal(cablink1.source.parameters.rx.value, 0.001);
                assert.equal(cablink1.source.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');


                assert.equal(cablink2.destination.name, 'BBMOD_1');
                assert.equal(cablink2.destination.distName, '/MRBTS-1/RAT-1/EQM_L-1/BBMOD_L-1');
                assert.equal(cablink2.destination.status, 'OK');
                assert.equal(cablink2.destination.port, 1);
                assert.equal(cablink2.destination.parameters.tx.value, 0.001);
                assert.equal(cablink2.destination.parameters.rx.value, 0.001);
                assert.equal(cablink2.destination.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/BBMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                assert.equal(cablink2.source.name, 'RMOD_1', 'rrh name');
                assert.equal(cablink2.source.distName, '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1');
                assert.equal(cablink2.source.status, 'OK');
                assert.equal(cablink2.source.port, 2);
                assert.equal(cablink2.source.parameters.tx.value, 0.001);
                assert.equal(cablink2.source.parameters.rx.value, 0.001);
                assert.equal(cablink2.source.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1/CONNECTOR_L-2/SFP_L-1/SFP_PARAMS_L-1');
            }).then(done).catch(done);
    });

    it('getParametersForLink should return parameters for particular link', (done) => {
        const completionListener = sinon.spy();
        const msgData = createMsgData('getParametersForLink', '/MRBTS-1/RAT-1/BTS_L-1/HWTOP_L-1/CABLNK_L-1');

        sfpMonitoring.execute(msgData, completionListener)
            .then(() => {
                assert(connectToBtsSpy.calledOnce);
                assert(completionListener.calledOnce);
                const status = completionListener.args[0][0];
                const connection = completionListener.args[0][1];
                assert.equal(status, ADMIN_API_STATUSES.COMPLETED);
                assert.isDefined(connection);
                assert.lengthOf(connection.alarms, 0);
                assert.equal(connection.destination.name, 'SMOD_1');
                assert.equal(connection.destination.distName, '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1');
                assert.equal(connection.destination.status, 'OK');
                assert.equal(connection.destination.port, 1);
                assert.equal(connection.destination.parameters.tx.value, 0.001);
                assert.equal(connection.destination.parameters.rx.value, 0.001);
                assert.equal(connection.destination.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                assert.equal(connection.source.name, 'RMOD_1');
                assert.equal(connection.source.distName, '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1');
                assert.equal(connection.source.status, 'OK');
                assert.equal(connection.source.port, 1);
                assert.equal(connection.source.parameters.tx.value, 0.001);
                assert.equal(connection.source.parameters.rx.value, 0.001);
                assert.equal(connection.source.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
            }).then(done).catch(done);
    });

    it('should fail for invalid distName', (done) => {
        const completionListener = sinon.spy();
        const invalidDistName = '/MRBTS-1/RAT-1/BTS_L-1/HWTOP_L-1/CABLNK_L-11';
        const msgData = createMsgData('getParametersForLink', invalidDistName);

        sfpMonitoring.execute(msgData, completionListener)
            .then(() => done(new Error('This promise shouldn\'t be resolved')))
            .catch(() => {
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    `No object with distName ${invalidDistName}`
                ).calledOnce);
                done();
            });
    });

    it('should fail for wrong type of distName', (done) => {
        const completionListener = sinon.spy();
        const invalidDistName = 2;
        const msgData = createMsgData('getParametersForLink', invalidDistName);

        sfpMonitoring.execute(msgData, completionListener)
            .then(() => done(new Error('This promise shouldn\'t be resolved')))
            .catch(() => {
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    '/parameters/distName, Invalid type: number (expected string)'
                ).calledOnce);
                done();
            });
    });

    it('should notify when message format is invalid - no distName', (done) => {
        const completionListener = sinon.spy();
        const msgData = {
            name: 'sfp',
            parameters: {
                action: 'getParametersForLink'
            }
        };

        sfpMonitoring.execute(msgData, completionListener)
            .then(() => done(new Error('This promise shouldn\'t be resolved')))
            .catch(() => {
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    '/parameters, Missing required property: distName'
                ).calledOnce);
                done();
            });
    });

    it('should notify when message format is invalid - no action', (done) => {
        const completionListener = sinon.spy();
        const msgData = {
            name: 'sfp',
            parameters: {}
        };

        sfpMonitoring.execute(msgData, completionListener)
            .then(() => done(new Error('This promise shouldn\'t be resolved')))
            .catch(() => {
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    '/parameters, Missing required property: action'
                ).calledOnce);
                done();
            });
    });

    it('should notify when message format is invalid - no parameters', (done) => {
        const completionListener = sinon.spy();
        const msgData = {
            name: 'sfp'
        };

        sfpMonitoring.execute(msgData, completionListener)
            .then(() => done(new Error('This promise shouldn\'t be resolved')))
            .catch(() => {
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    'Missing required property: parameters'
                ).calledOnce);
                done();
            });
    });

    function createMsgData(action, distName) {
        return {
            name: 'sfp',
            parameters: {
                action,
                distName
            }
        };
    }

    function createModule(className, id, parameters, parentNode = eqmL) {
        const moduleNode = imTree.addNode(parentNode, className, id);
        moduleNode.parameters = parameters || {};
        return moduleNode;
    }

    function createConnectorNode(module, id, state = 'Enabled') {
        const connector = imTree.addNode(module, 'CONNECTOR_L', id);
        connector.parameters = {
            stateInfo: {operationalState: state}
        };
        return connector;
    }

    function createCablnkNode(rmodConnector, destModuleConnector, id, rmodPort, destinationPort, state = 'Enabled', type = 'RP3_01Link') {
        const cablnk = imTree.addNode(hwtopL, 'CABLNK_L', id);
        cablnk.parameters = {
            cablinkType: type,
            destinationDistName: destModuleConnector.distName,
            sourceDistName: rmodConnector.distName,
            sourcePhysicalPortNo: rmodPort,
            destinationPhysicalPortNo: destinationPort,
            stateInfo: {operationalState: state}
        };
        return cablnk;
    }

    function createSfpNode(connector) {
        return imTree.addNode(connector, 'SFP_L', 1);
    }

    function createSfpParamsNode(sfp, rxPower, txPower) {
        const sfpParamsNode = imTree.addNode(sfp, 'SFP_PARAMS_L', 1);
        sfpParamsNode.parameters = {
            rxPower: rxPower,
            txPower: txPower
        };
        return sfpParamsNode;
    }

    function createSfpNodeWithParams(connector, rxPower, txPower) {
        const sfp = createSfpNode(connector);
        createSfpParamsNode(sfp, rxPower, txPower);
    }
});

