const Promise = require('bluebird');
const assert = require('assert');

const softwareInfoProcedure = require('../../../src/adminApi/infoModel/stateInfo.js');
const adminApiStatuses = require('../../../src/adminApi/adminApiStatuses');

const errorMessage = 'error message';
const error = new Error(errorMessage);

describe('softwareInfoProcedure', function () {
    'use strict';

    const msgData = {
        parameters: {
            distName: '/DIST-1/NAME-1'
        }
    };

    var completionListenerStub;
    var decodedParametersStub;
    var nodeStub;
    var imTreeStub;
    var onlineImTreeStub;
    var connectionInitializerStub;

    var softwareInfoInstance;

    beforeEach(function () {
        completionListenerStub = sinon.spy();

        decodedParametersStub = {
            stateInfo: {}
        };

        nodeStub = {
            getDecodedParameters: sinon.stub().returns(decodedParametersStub)
        };

        imTreeStub = {
            findByDistName: sinon.stub().returns(nodeStub)
        };

        onlineImTreeStub = {
            getImTree: sinon.stub().returns(Promise.resolve(imTreeStub))
        };

        connectionInitializerStub = {
            prepare: sinon.stub().returns(Promise.resolve())
        };
        connectionInitializerStub.connectToBts = sinon.stub().returns(connectionInitializerStub);

        softwareInfoInstance = softwareInfoProcedure({
            onlineImTree: onlineImTreeStub,
            connectionInitializer: connectionInitializerStub
        });
    });

    it('should notify completion listener with requested stateInfo after connecting to BTS', function (done) {
        softwareInfoInstance.execute(msgData, completionListenerStub)
            .then(function () {
                assert(connectionInitializerStub.connectToBts.calledOnce);
                assert(connectionInitializerStub.prepare.calledOnce);

                assert(onlineImTreeStub.getImTree.calledOnce);
                assert(imTreeStub.findByDistName.calledWith(msgData.parameters.distName));

                assert(completionListenerStub.calledWith(adminApiStatuses.COMPLETED, decodedParametersStub.stateInfo));

                done();
            });
    });

    it('should notify completion listener about failure if BTS connection cannot be established', function (done) {
        connectionInitializerStub.prepare = sinon.stub().returns(Promise.reject(error));

        softwareInfoInstance.execute(msgData, completionListenerStub)
            .catch(function () {
                assert(completionListenerStub.calledWith(adminApiStatuses.FAILED, error.toString()));

                done();
            });
    });

    it('should notify completion listener about failure if im tree cannot be received', function (done) {
        onlineImTreeStub.getImTree = sinon.stub().returns(Promise.reject(error));

        softwareInfoInstance.execute(msgData, completionListenerStub)
            .catch(function () {
                assert(completionListenerStub.calledWith(adminApiStatuses.FAILED, error.toString()));

                done();
            });
    });

    it('should notify completion listener about failure if requested node was not found', function (done) {
        imTreeStub.findByDistName = sinon.stub().returns();

        softwareInfoInstance.execute(msgData, completionListenerStub)
            .catch(function () {
                assert(completionListenerStub.calledWith(adminApiStatuses.FAILED));

                done();
            });
    });

    it('should notify completion listener about failure if requested node does not have state info', function (done) {
        nodeStub.getDecodedParameters = sinon.stub().returns({});

        softwareInfoInstance.execute(msgData, completionListenerStub)
            .catch(function () {
                assert(completionListenerStub.calledWith(adminApiStatuses.FAILED));

                done();
            });
    });
});
