const timingsFactory = require('../../../src/adminApi/infoModel/timings');
const timingServiceForOnlineImTree = require('../../../src/timings/timingServiceForOnlineImTree');
const ADMIN_API_STATUSES = require('../../../src/adminApi/adminApiStatuses');
const ImTree = require('../../../src/imtree');
const Promise = require('bluebird');

describe('ADMIN API: timings', () => {
    let timings;
    let connectionInitializerMock;
    let connectToBtsSpy;
    let prepareStub;
    let imTree;

    function createConnectionInitializerMock() {
        prepareStub = sinon.stub();
        connectToBtsSpy = sinon.spy(() => ({prepare: prepareStub}));
        return {
            connectToBts: connectToBtsSpy
        };
    }

    const onlineImTreeMock = {
        getImTree: sinon.stub()
    };

    beforeEach(() => {
        imTree = new ImTree();
        connectionInitializerMock = createConnectionInitializerMock();
        timings = timingsFactory({
            onlineImTree: onlineImTreeMock,
            connectionInitializer: connectionInitializerMock,
            timingServiceForOnlineImTree: timingServiceForOnlineImTree
        });
        onlineImTreeMock.getImTree.returns(Promise.resolve(imTree));
        prepareStub.returns(Promise.resolve());
    });

    it('should complete when given correct request schema', (done) => {
        const req = createRequestWithoutParam('findAllTimings');
        timings.execute(req, (requestStatus) => {
            assert.equal(requestStatus, ADMIN_API_STATUSES.COMPLETED);
            done();
        });
    });

    it('should complete when given correct request schema', (done) => {
        const req = createRequestWithParam('findTimingsForProcedureName', 'abc');
        timings.execute(req, (requestStatus) => {
            assert.equal(requestStatus, ADMIN_API_STATUSES.COMPLETED);
            done();
        });
    });

    it('should fail when given incorrect request schema', (done) => {
        const completionListener = sinon.spy();
        const req = createRequestWithParam('findAllTimings', 'xyz');
        timings.execute(req, completionListener)
            .then(() => done(new Error('This promise should not be resolved')))
            .catch(() => {
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    '/parameters/param, Additional properties not allowed'
                ).calledOnce);
                done();
            });
    });

    it('should fail when given incorrect request schema', (done) => {
        const completionListener = sinon.spy();
        const req = createRequestWithoutParam('findTimingsForProcedureName', 'xyz');
        timings.execute(req, completionListener)
            .then(() => done(new Error('This promise should not be resolved')))
            .catch(() => {
                assert(completionListener.withArgs(
                    ADMIN_API_STATUSES.FAILED,
                    '/parameters, Missing required property: param'
                ).calledOnce);
                done();
            });
    });

    it('should work consistently across many request to timings api', (done) => {
        imTree.mergeNode('/MRBTS-1/TIMINGS_L-1', {});
        imTree.mergeNode('/MRBTS-1/XYZ-2/TIMINGS_L-2', {});
        const req = createRequestWithoutParam('findAllTimings');
        timings.execute(req, (requestStatus, payload) => {
            assert.equal(requestStatus, ADMIN_API_STATUSES.COMPLETED);
            assert.equal(payload.length, 2);
            imTree.mergeNode('/MRBTS-1/ABC-2/TIMINGS_L-2', {});
            timings.execute(req, (requestStatus2, payload2) => {
                assert.equal(requestStatus2, ADMIN_API_STATUSES.COMPLETED);
                assert.equal(payload2.length, 3);
                done();
            });
        });
    });
});

function createRequestWithoutParam(procedureName) {
    return {
        name: 'timings',
        parameters: {
            procedure: procedureName
        }
    };
}

function createRequestWithParam(procedureName, param) {
    return {
        'name': 'timings',
        'parameters': {
            'procedure': procedureName,
            'param': param
        }
    };
}
