const ADMIN_API_STATUSES = require('../../../src/adminApi/adminApiStatuses');
const distNameProceduresFactory = require('../../../src/adminApi/procedures/distNameProcedures');
const testUtils = require('../adminApi.test.util');
const Promise = require('bluebird');

describe('ADMIN API: distNameProcedures', function () {

    const LNCEL_DIST_NAME = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1/LNCEL-1';

    var distNameRemoteProcedures,
        connectionInitializerMock,
        procedureExecutorMock;

    beforeEach(function () {
        procedureExecutorMock = testUtils.createProcedureExecutorMock();
        connectionInitializerMock = testUtils.createConnectionInitializerMock();
        distNameRemoteProcedures = distNameProceduresFactory({
            procedureExecutor: procedureExecutorMock,
            btsServiceProcedure: testUtils.createBtsServiceProcedureMock(),
            connectionInitializer: connectionInitializerMock
        });
    });

    it('should execute block', function (done) {
        // given
        connectionInitializerMock.prepare.returns(Promise.resolve());
        const msgData = createMsgData('block', LNCEL_DIST_NAME);
        // when
        distNameRemoteProcedures.executeBlock(msgData, testUtils.noop)
            .then(function () {
                // then
                assert(connectionInitializerMock.connectToBts.calledOnce);
                assert(connectionInitializerMock.checkIfDnExists.withArgs(LNCEL_DIST_NAME).calledOnce);
                assert(procedureExecutorMock.executeBlock.calledOnce);
            }).then(done).catch(done);
    });

    it('should execute unblock', function (done) {
        // given
        connectionInitializerMock.prepare.returns(Promise.resolve());
        const msgData = createMsgData('unblock', LNCEL_DIST_NAME);
        // when
        distNameRemoteProcedures.executeUnblock(msgData, testUtils.noop)
            .then(function () {
                // then
                assert(connectionInitializerMock.connectToBts.calledOnce);
                assert(connectionInitializerMock.checkIfDnExists.withArgs(LNCEL_DIST_NAME).calledOnce);
                assert(procedureExecutorMock.executeUnblock.calledOnce);
            }).then(done).catch(done);
    });

    describe('reset', function () {

        it('should notify when reset is executed on incorrect object', function (done) {
            // given
            connectionInitializerMock.prepare.returns(Promise.resolve());
            const msgData = createMsgData('radioReset', '/MRBTS-1/RAT-1');
            // when
            distNameRemoteProcedures.executeRadioReset(msgData, function (requestStatus, payload) {
                // then
                assert.equal(requestStatus, ADMIN_API_STATUSES.FAILED);
                assert.equal(payload, 'Distname /MRBTS-1/RAT-1 is unnsuported for reset procedure');
                assert.equal(procedureExecutorMock.executeReset.callCount, 0);
                done();
            });
        });
    });

    it('should notify when failed at existing procedure preparation', function (done) {
        connectionInitializerMock.prepare.returns(Promise.reject('some error'));
        const msgData = createMsgData('unblock', LNCEL_DIST_NAME);
        const completionListener = sinon.spy();
        // when
        distNameRemoteProcedures.executeUnblock(msgData, completionListener)
            .then(() => done(new Error('This promise should not be resolved')))
            .catch(() => {
                // then
                assert(completionListener.withArgs(ADMIN_API_STATUSES.FAILED, 'some error').calledOnce);
                done();
            });
    });

    it('should notify when message format is invalid - no distName', function (done) {
        // given
        const msgData = {
            name: 'block',
            parameters: {}
        };
        // when
        distNameRemoteProcedures.executeBlock(msgData, function (requestStatus, payload) {
            // then
            assert.equal(requestStatus, ADMIN_API_STATUSES.FAILED);
            assert.equal(payload, '/parameters, Missing required property: distName');
            assert.equal(procedureExecutorMock.executeBlock.callCount, 0);
            done();
        });
    });

    function createMsgData(procedureName, distName) {
        return {
            name: procedureName,
            parameters: {
                distName: distName
            }
        };
    }
});
