const ADMIN_API_STATUSES = require('../../../src/adminApi/adminApiStatuses');
const proceduresDispatcherFactory = require('../../../src/adminApi/procedures/proceduresDispatcher');
const testUtils = require('../adminApi.test.util');
const Promise = require('bluebird');

describe('ADMIN API: proceduresDispatcher', function () {

    const LNCEL_DIST_NAME = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1/LNCEL-1';
    const RMOD_DIST_NAME = '/MRBTS-1/RAT-1/BTS_L-1/EQM_L-1/RMOD_L-1';

    let proceduresDispatcher,
        distNameProceduresMock,
        imSnapshotProcedureMock,
        snapshotProcedureMock,
        commissioningProcedureMock,
        resetProcedureMock,
        swUpdateProcedureMock,
        otherProcedureMock;

    beforeEach(function () {
        distNameProceduresMock = createDistNameProceduresMock();
        imSnapshotProcedureMock = createProcedureMock();
        snapshotProcedureMock = createProcedureMock();
        commissioningProcedureMock = createCommissioningProcedureMock();
        resetProcedureMock = createResetProcedureMock();
        swUpdateProcedureMock = createProcedureMock();
        otherProcedureMock = createOtherProcedureMock();

        proceduresDispatcher = proceduresDispatcherFactory({
            distNameProcedures: distNameProceduresMock,
            imSnapshotProcedure: imSnapshotProcedureMock,
            snapshotProcedure: snapshotProcedureMock,
            commissioningProcedure: commissioningProcedureMock,
            resetProcedure: resetProcedureMock,
            swUpdateProcedure: swUpdateProcedureMock,
            otherProcedure: otherProcedureMock
        });
    });

    function createResetProcedureMock() {
        return {
            executeBtsReset: sinon.spy(Promise.resolve),
            executeSiteReset: sinon.spy(Promise.resolve),
            executeResetToTestDedicated: sinon.spy(Promise.resolve),
            executeConfigurationReset: sinon.spy(Promise.resolve)
        };
    }

    function createOtherProcedureMock() {
        return {
            executeSwRollback: sinon.spy(Promise.resolve),
            executeChangeEthernetPortSecurity: sinon.spy(Promise.resolve)
        };
    }

    function createDistNameProceduresMock() {
        return {
            executeBlock: sinon.spy(Promise.resolve),
            executeUnblock: sinon.spy(Promise.resolve),
            executeRadioReset: sinon.spy(Promise.resolve)
        };
    }

    function createProcedureMock() {
        return {
            execute: sinon.spy(Promise.resolve)
        };
    }

    function createCommissioningProcedureMock() {
        return {
            executeDownloadPlan: sinon.spy(Promise.resolve),
            executeActivatePlan: sinon.spy(Promise.resolve)
        };
    }

    it('should not forward unknown procedure', function (done) {
        // given
        const procedureName = 'unknownProcedure';
        const msgData = createMsgData(procedureName, LNCEL_DIST_NAME);
        // when
        proceduresDispatcher.dispatch(msgData, (requestStatus, payload) => {
            // then
            assert.equal(requestStatus, ADMIN_API_STATUSES.FAILED);
            assert.equal(payload, 'No ' + procedureName + ' procedure in testing API');
            done();
        });
    });

    it('should forward btsReset procedure', function () {
        // given
        const msgData = createSiteOrBtsResetData('btsReset');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(resetProcedureMock.executeBtsReset);
    });

    it('should forward siteReset procedure', function () {
        // given
        const msgData = createSiteOrBtsResetData('siteReset');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(resetProcedureMock.executeSiteReset);
    });

    it('should forward radioReset procedure', function () {
        // given
        const msgData = createMsgData('radioReset', RMOD_DIST_NAME);
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(distNameProceduresMock.executeRadioReset);
    });

    it('should forward block procedure', function () {
        // given
        const msgData = createMsgData('block', LNCEL_DIST_NAME);
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(distNameProceduresMock.executeBlock);
    });

    it('should forward unblock procedure', function () {
        // given
        const msgData = createMsgData('unblock', LNCEL_DIST_NAME);
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(distNameProceduresMock.executeUnblock);
    });

    it('should forward imSnapshot procedure', function () {
        // given
        const msgData = createMsgData('imSnapshot');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(imSnapshotProcedureMock.execute);
    });

    it('should forward snapshot procedure', function () {
        // given
        const msgData = createMsgData('snapshot');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(snapshotProcedureMock.execute);
    });

    it('should forward downloadPlan procedure', function () {
        // given
        const msgData = createMsgData('downloadPlan');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(commissioningProcedureMock.executeDownloadPlan);
    });

    it('should forward activatePlan procedure', function () {
        // given
        const msgData = createMsgData('activatePlan');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(commissioningProcedureMock.executeActivatePlan);
    });

    it('should forward swRollback procedure', function () {
        // given
        const msgData = createMsgData('swRollback');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(otherProcedureMock.executeSwRollback);
    });

    it('should forward resetToTestDedicated procedure', function () {
        // given
        const msgData = createMsgData('resetToTestDedicated');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(resetProcedureMock.executeResetToTestDedicated);
    });

    it('should forward configurationReset procedure', function () {
        // given
        const msgData = createMsgData('configurationReset');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(resetProcedureMock.executeConfigurationReset);
    });

    it('should forward changeEthernetPortSecurity procedure', function () {
        // given
        const msgData = createMsgData('changeEthernetPortSecurity');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(otherProcedureMock.executeChangeEthernetPortSecurity);
    });

    it('should forward softwareUpdate procedure', function () {
        // given
        const msgData = createMsgData('softwareUpdate');
        // when
        proceduresDispatcher.dispatch(msgData, testUtils.noop);
        // then
        sinon.assert.calledOnce(swUpdateProcedureMock.execute);
    });

    it('should fail when msgData doesn\'t contain required name parameter', function (done) {
        // given
        const msgData = {
            parameters: {
                distName: LNCEL_DIST_NAME
            }
        };
        // when
        proceduresDispatcher.dispatch(msgData, (requestStatus, payload) => {
            // then
            assert.equal(requestStatus, ADMIN_API_STATUSES.FAILED);
            assert.equal(payload, 'Missing required property: name');
            done();
        });
    });

    function createMsgData(procedureName, distName) {
        return {
            name: procedureName,
            parameters: {
                distName
            }
        };
    }

    function createSiteOrBtsResetData(procedureName) {
        return {
            name: procedureName
        };
    }
});
