const resetProcedure = require('../../../src/adminApi/procedures/resetProcedure');
const testUtils = require('../adminApi.test.util');
const ADMIN_API_STATUSES = require('../../../src/adminApi/adminApiStatuses');

describe('resetProcedure', () => {
    let suit;

    beforeEach(() => {
        suit = {};
        suit.completionListener = sinon.spy();
        suit.procedureExecutorMock = testUtils.createProcedureExecutorMock();
        suit.connectionServiceStub = {
            isSignedIn: () => true
        };
        suit.resetProcedure = resetProcedure.newInstance({
            procedureExecutor: suit.procedureExecutorMock,
            btsServiceProcedure: testUtils.createBtsServiceProcedureMock(),
            connectionService: suit.connectionServiceStub,
            onlineImTree: suit.onlineImTreeStub
        });
    });
    afterEach(() => (suit = null));

    describe('reset', () => {

        it('should execute bts reset', (done) => {
            // given
            const msgData = createResetMsgData('btsReset');
            // when
            suit.resetProcedure.executeBtsReset(msgData, suit.completionListener)
                .then(() => {
                    // then
                    sinon.assert.calledOnce(suit.procedureExecutorMock.executeReset);
                }).then(done).catch(done);
        });

        it('should execute site reset', (done) => {
            // given
            const msgData = createResetMsgData('siteReset');
            // when
            suit.resetProcedure.executeSiteReset(msgData, suit.completionListener)
                .then(() => {
                    // then
                    sinon.assert.calledOnce(suit.procedureExecutorMock.executeReset);

                }).then(done).catch(done);
        });

        it('should execute reset to test dedicated', (done) => {
            // given
            const msgData = createResetMsgData('resetToTestDedicated');
            // when
            suit.resetProcedure.executeResetToTestDedicated(msgData, suit.completionListener)
                .then(() => {
                    // then
                    sinon.assert.calledOnce(suit.procedureExecutorMock.executeResetToTestDedicated);
                }).then(done).catch(done);
        });

        function createResetMsgData(procedureName) {
            return {
                name: procedureName
            };
        }

    });
    describe('#configurationReset', () => {
        it('should not execute configurationReset when name is wrong', (done) => {
            // given
            const msgData = {
                name: 'configurationResetWrong',
                parameters: {
                    removeSiteId: true,
                    removeSecurityCredentials: true
                }
            };
            // when
            suit.resetProcedure.executeConfigurationReset(msgData, suit.completionListener)
                .then(testUtils.failAssert)
                .catch((error) => {
                    // then
                    assert.equal(error, 'Invalid schema');
                    sinon.assert.calledOnce(suit.completionListener);
                    sinon.assert.calledWith(suit.completionListener, ADMIN_API_STATUSES.FAILED, '/name, String does not match pattern: ^configurationReset$');
                    sinon.assert.notCalled(suit.procedureExecutorMock.executeConfigurationReset);
                    done();
                });
        });

        it('should not execute configurationReset when \'parameters\' are missing', (done) => {
            // given
            const msgData = {
                name: 'configurationReset'
            };
            // when
            suit.resetProcedure.executeConfigurationReset(msgData, suit.completionListener)
                .then(testUtils.failAssert)
                .catch((error) => {
                    // then
                    assert.equal(error, 'Invalid schema');
                    sinon.assert.calledOnce(suit.completionListener);
                    sinon.assert.calledWith(suit.completionListener, ADMIN_API_STATUSES.FAILED, 'Missing required property: parameters');
                    sinon.assert.notCalled(suit.procedureExecutorMock.executeConfigurationReset);
                    done();
                });
        });
        it('should not execute configurationReset when \'removeSiteId\' is missing', (done) => {
            // given
            const msgData = {
                name: 'configurationReset',
                parameters: {
                    removeSecurityCredentials: true
                }
            };
            // when
            suit.resetProcedure.executeConfigurationReset(msgData, suit.completionListener)
                .then(testUtils.failAssert)
                .catch((error) => {
                    // then
                    assert.equal(error, 'Invalid schema');
                    sinon.assert.calledOnce(suit.completionListener);
                    sinon.assert.calledWith(suit.completionListener, ADMIN_API_STATUSES.FAILED, '/parameters, Missing required property: removeSiteId');
                    sinon.assert.notCalled(suit.procedureExecutorMock.executeConfigurationReset);
                    done();
                });
        });
        it('should not execute configurationReset when \'removeSecurityCredentials\' is missing', (done) => {
            // given
            const msgData = {
                name: 'configurationReset',
                parameters: {
                    removeSiteId: true
                }
            };
            // when
            suit.resetProcedure.executeConfigurationReset(msgData, suit.completionListener)
                .then(testUtils.failAssert)
                .catch((error) => {
                    // then
                    assert.equal(error, 'Invalid schema');
                    sinon.assert.calledOnce(suit.completionListener);
                    sinon.assert.calledWith(suit.completionListener, ADMIN_API_STATUSES.FAILED, '/parameters, Missing required property: removeSecurityCredentials');
                    sinon.assert.notCalled(suit.procedureExecutorMock.executeConfigurationReset);
                    done();
                });
        });

        it('should execute configuration reset', (done) => {
            // given
            const msgData = {
                name: 'configurationReset',
                parameters: {
                    removeSiteId: true,
                    removeSecurityCredentials: true
                }
            };
            // when
            suit.resetProcedure.executeConfigurationReset(msgData, suit.completionListener)
                .then(() => {
                    // then
                    sinon.assert.calledOnce(suit.procedureExecutorMock.executeConfigurationReset);
                })
                .then(done).catch(done);
        });

    })
});
