const snapshotProcedureFactory = require('../../../src/adminApi/procedures/snapshotProcedure');
const snapshotStates = require('../../../src/procedures/snapshot/snapshotStates');
const Promise = require('bluebird');

describe('ADMIN API: snapshotProcedure', function () {
    'use strict';

    var unwatcher;
    var snapshotMock;
    var snapshotFileSaverMock;
    var connectionInitializerMock;

    var completionListenerMock;

    var snapshotProcedure;

    var msgData = {
        name: 'snapshot',
        parameters: {
            coverage: 'minor'
        }
    };

    beforeEach(function () {
        unwatcher = {
            unwatch: sinon.spy()
        };

        snapshotMock = {
            startSnapshot: sinon.stub().returns(Promise.resolve()),
            addStatusListener: sinon.stub().returns(unwatcher),
            isSnapshotPossible: sinon.stub().returns(true)
        };

        snapshotFileSaverMock = {
            setListener: sinon.spy(),
            failure: sinon.spy(),
            setAborted: sinon.spy()
        };

        connectionInitializerMock = {
            prepare: sinon.stub().returns(Promise.resolve())
        };

        connectionInitializerMock.connectToBts = sinon.stub().returns(connectionInitializerMock);

        completionListenerMock  = sinon.spy();

        snapshotProcedure = snapshotProcedureFactory({
            snapshot: snapshotMock,
            snapshotFileSaver: snapshotFileSaverMock,
            connectionInitializer: connectionInitializerMock
        });
    });
    describe('execute method', function () {
        it('should attach listeners to connected BTS', function (done) {
            snapshotProcedure.execute(msgData, completionListenerMock)
                .then(() => {
                    assert(snapshotMock.isSnapshotPossible.calledOnce);
                    assert(snapshotFileSaverMock.setListener.calledOnce);
                    assert(snapshotMock.startSnapshot.calledWith(msgData.parameters.coverage));
                    assert(snapshotMock.addStatusListener.calledOnce);
                    done();
                });
        });

        it('should pass the file to the completion listener and remove active status listener', function (done) {
            var status = 'status';
            var message = 'message';

            snapshotFileSaverMock.setListener = function (listener) {
                listener(status, message);

                assert(completionListenerMock.calledWith(status, message));
                assert(unwatcher.unwatch.calledOnce);

                done();
            };
            snapshotProcedure.execute(msgData, completionListenerMock);
        });

        it('should not perform snapshot if BTS connection cannot be established', function (done) {
            connectionInitializerMock.prepare = sinon.stub().returns(Promise.reject());
            snapshotProcedure.execute(msgData, completionListenerMock)
                .catch(() => {
                    assert(snapshotFileSaverMock.setListener.notCalled);
                    assert(snapshotMock.addStatusListener.notCalled);
                    assert(snapshotMock.startSnapshot.notCalled);

                    done();
                });
        });

        it('should reject if snapshot not possible', function (done) {
            snapshotMock.isSnapshotPossible = sinon.stub().returns(false);
            snapshotProcedure.execute(msgData, completionListenerMock)
                .catch((state) => {
                    assert.equal(snapshotStates.REJECTED, state);

                    done();
                });
        });

        it('added status listener should react appropriately to the received status', function (done) {
            snapshotMock.addStatusListener = function (statusListener) {
                statusListener(snapshotStates.DISCONNECTED);
                assert(snapshotFileSaverMock.failure.calledWith(snapshotStates.DISCONNECTED.text));

                statusListener(snapshotStates.REJECTED);
                assert(snapshotFileSaverMock.failure.calledWith(snapshotStates.REJECTED.text));

                statusListener(snapshotStates.CANCELLED);
                assert(snapshotFileSaverMock.failure.calledWith(snapshotStates.CANCELLED.text));

                statusListener(snapshotStates.DISABLED);
                assert(snapshotFileSaverMock.failure.calledWith(snapshotStates.DISABLED.text));

                statusListener(snapshotStates.INTERRUPTED);
                assert(snapshotFileSaverMock.failure.calledWith(snapshotStates.INTERRUPTED.text));

                statusListener(snapshotStates.CANCELLING);
                assert(snapshotFileSaverMock.setAborted.calledOnce);


                function notCalled(key) {
                    snapshotFileSaverMock.failure.reset();
                    snapshotFileSaverMock.setAborted.reset();
                    statusListener(key);
                    assert(snapshotFileSaverMock.failure.notCalled);
                    assert(snapshotFileSaverMock.setAborted.notCalled);
                }

                notCalled(snapshotStates.REQUESTED);
                notCalled(snapshotStates.ACTIVE);
                notCalled(snapshotStates.BUSY);
                notCalled(snapshotStates.FINISHED);
                notCalled(snapshotStates.FINISHED_WITH_ERRORS);
                notCalled(snapshotStates.AVAILABLE);

                done();
            };
            snapshotProcedure.execute(msgData, completionListenerMock);

        });
    });
});
