const adminApiStatuses = require('../../../src/adminApi/adminApiStatuses');
const assert = require('assert');
const Promise = require('bluebird');

const swUpdateProcedure = require('../../../src/adminApi/procedures/swUpdateProcedure');

describe('swUpdateProcedure', function () {
    const SW_UPDATE_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/SW_UPDATE_L-1';
    const BUSY = 'Busy';
    const SUCCESS = 'Success';
    const FAILURE = 'Failure';
    const NOT_STARTED = 'Not_started';
    const error = 'error text';

    const targetBd = {
        name: 'TargetBd.xml'
    };

    var softwareZipDispatcherFactoryStub;
    var swUpdateExecutorStub;
    var swUpdateExecutorFactoryStub;
    var connectionInitializerStub;

    var swUpdateProcedureInstance;

    var completionListenerMock;

    var nodeStub;
    var imTreeStub;
    var onlineImTreeStub;

    var swUpdateListener;
    var imTrackerStub;
    var stateSequence;

    var swUpdateState = {
        stateInfo: {
            swUpdateState: NOT_STARTED
        }
    };

    const msgData = {
        parameters: {
            file: 'WklQIENPTlRFTlQ=',
            shouldActivate: false
        }
    };

    const softwareUpdateConfig = {
        shouldActivate: msgData.parameters.shouldActivate,
        targetBd: targetBd.name,
        source: 'zip'
    };

    const uint8Array = new Uint8Array([ 90, 73, 80, 32, 67, 79, 78, 84, 69, 78, 84 ]);

    //Busy, Failure, Success
    function changeState(state) {
        var event = {
            parameters: {
                stateInfo: {
                    swUpdateState: state
                }
            }
        };
        swUpdateListener.onParameterChanged(event);
    }

    beforeEach(function () {
        softwareZipDispatcherFactoryStub = {
            getDispatcher: sinon.stub().returnsThis(),
            getTargetBd: sinon.stub().returns(targetBd),
            fromZipFile: sinon.stub().returnsThis()
        };

        swUpdateExecutorStub = {
            execute: sinon.stub().returns(Promise.resolve())
        };
        swUpdateExecutorFactoryStub = {
            newSoftwareUpdateExecutor: sinon.stub().returns(swUpdateExecutorStub)
        };

        connectionInitializerStub = {
            prepare: sinon.stub().returns(Promise.resolve())
        };
        connectionInitializerStub.connectToBts = sinon.stub().returns(connectionInitializerStub);

        nodeStub = {
            getDecodedParameters: sinon.stub().returns(swUpdateState)
        };

        imTreeStub = {
            findByDistName: sinon.stub().returns(nodeStub)
        };

        onlineImTreeStub = {
            getImTree: sinon.stub().returns(Promise.resolve(imTreeStub))
        };

        stateSequence = function () {
            return Promise.resolve()
                .then(() => changeState(BUSY))
                .then(() => changeState(SUCCESS));
        };

        function trackByDistName(distName, listener) {
            swUpdateListener = listener;
            return stateSequence();
        }

        imTrackerStub = {
            stopTracking: sinon.spy(),
            trackByDistName: trackByDistName
        };
        sinon.stub(imTrackerStub, 'trackByDistName', trackByDistName);

        swUpdateProcedureInstance = swUpdateProcedure({
            softwareZipDispatcherFactory: softwareZipDispatcherFactoryStub,
            swUpdateExecutorFactory: swUpdateExecutorFactoryStub,
            connectionInitializer: connectionInitializerStub,
            onlineImTree: onlineImTreeStub,
            imTracker: imTrackerStub
        });

        completionListenerMock = sinon.mock();
    });

    describe('execute method', function () {
        it('should serve files and execute software update after connecting to bts', function (done) {
            softwareZipDispatcherFactoryStub.fromZipFile = function (arrayBuffer) {
                var zip = new Uint8Array(arrayBuffer);
                assert.deepEqual(zip, uint8Array);

                return softwareZipDispatcherFactoryStub;
            };

            swUpdateProcedureInstance.execute(msgData, completionListenerMock)
                .then(function () {
                    assert(connectionInitializerStub.connectToBts.calledOnce);
                    assert(connectionInitializerStub.prepare.calledOnce);
                    assert(onlineImTreeStub.getImTree.calledOnce);
                    assert(imTreeStub.findByDistName.calledWith(SW_UPDATE_DISTNAME));
                    assert(nodeStub.getDecodedParameters.calledOnce);

                    assert(imTrackerStub.trackByDistName.calledWith(SW_UPDATE_DISTNAME));

                    assert(softwareZipDispatcherFactoryStub.getDispatcher.calledOnce);

                    assert(swUpdateExecutorFactoryStub.newSoftwareUpdateExecutor.calledWith(softwareUpdateConfig));
                    assert(swUpdateExecutorStub.execute.calledOnce);

                    assert(imTrackerStub.stopTracking.calledOnce);
                    assert(completionListenerMock.calledWith(adminApiStatuses.COMPLETED));

                    done();
                });
        });

        it('should execute software update with config after connecting to bts', function (done) {
            const hostname = 'hostname.com';
            const port = 80;
            const targetBdName = 'trgetBd.xml';

            const msg = {
                parameters: {
                    hostname: hostname,
                    port: port,
                    targetBd: targetBdName,
                    shouldActivate: true
                }
            };

            const config = {
                hostname: hostname,
                port: port,
                targetBd: targetBdName,
                shouldActivate: msg.parameters.shouldActivate,
                source: 'remote'
            };

            swUpdateProcedureInstance.execute(msg, completionListenerMock)
                .then(function () {
                    assert(connectionInitializerStub.connectToBts.calledOnce);
                    assert(connectionInitializerStub.prepare.calledOnce);

                    assert(swUpdateExecutorFactoryStub.newSoftwareUpdateExecutor.calledWith(config));
                    assert(swUpdateExecutorStub.execute.calledOnce);

                    assert(completionListenerMock.calledWith(adminApiStatuses.COMPLETED));

                    done();
                });
        });

        it('should notify completion listener, if unable connect to bts', function (done) {
            connectionInitializerStub.prepare = sinon.stub().returns(Promise.reject(error));

            swUpdateProcedureInstance.execute(msgData, completionListenerMock)
                .catch(function () {
                    assert(completionListenerMock.calledWith(adminApiStatuses.FAILED, error));

                    done();
                });
        });

        it('should notify completion listener, when can not get imTree', function (done) {
            onlineImTreeStub.getImTree = sinon.stub().returns(Promise.reject(error));
            swUpdateProcedureInstance.execute(msgData, completionListenerMock)
                .catch(function () {
                    assert(completionListenerMock.calledWith(adminApiStatuses.FAILED, error));

                    done();
                });
        });

        it('should notify completion listener, when can not find node at dist name', function (done) {
            imTreeStub.findByDistName = sinon.stub().returns(null);
            swUpdateProcedureInstance.execute(msgData, completionListenerMock)
                .catch(function () {
                    done();
                });
        });

        it('should notify completion listener, when cannot track im tree', function (done) {
            imTrackerStub.trackByDistName = sinon.stub().returns(Promise.reject(error));
            swUpdateProcedureInstance.execute(msgData, completionListenerMock)
                .catch(function () {
                    assert(completionListenerMock.calledWith(adminApiStatuses.FAILED, error));

                    done();
                });
        });

        it('should notify completion listener on update executor failure', function (done) {
            swUpdateExecutorStub.execute = sinon.stub().returns(Promise.reject(error));
            swUpdateProcedureInstance.execute(msgData, completionListenerMock)
                .catch(function () {
                    assert(completionListenerMock.calledWith(adminApiStatuses.FAILED, error));

                    done();
                });
        });

        it('should notify completion listener when update failure state occurs', function (done) {
            stateSequence = function () {
                return Promise.resolve()
                    .then(() => changeState(BUSY))
                    .then(() => changeState(FAILURE));
            };
            swUpdateProcedureInstance.execute(msgData, completionListenerMock)
                .catch(function () {
                    assert(completionListenerMock.calledWith(adminApiStatuses.FAILED));

                    done();
                });
        });

        it('should notify completion listener on update failure state', function (done) {
            stateSequence = function () {
                return Promise.resolve()
                    .then(() => changeState(FAILURE));
            };
            swUpdateProcedureInstance.execute(msgData, completionListenerMock)
                .catch(function () {
                    assert(completionListenerMock.calledWith(adminApiStatuses.FAILED));

                    done();
                });
        });

        it('should notify completion listener when update is in busy state', function (done) {
            swUpdateState.stateInfo.swUpdateState = BUSY;
            swUpdateProcedureInstance.execute(msgData, completionListenerMock)
                .catch(function () {
                    assert(completionListenerMock.calledWith(adminApiStatuses.FAILED));

                    done();
                });
        });
    });
});
