const sic = require('../../src/sic');
const authService = require('../../src/authentication/authService');

describe('Service: auth', () => {

    const USERS = {
        OPERATOR: {
            userName: 'operator',
            profile: 'operator'
        },
        DEVELOPER: {
            userName: 'developer',
            profile: 'developer'
        },
        TESTER: {
            userName: 'tester',
            profile: 'tester'
        }
    };

    let auth;
    let sicInst;
    let sessionStorage;
    let btsClientMock;

    beforeEach(() => {
        sicInst = sic();
        btsClientMock = createBtsClientMock(USERS);
        sessionStorage = {};

        auth = authService.imAuth({
            imAuthFactoryInst: authService.imAuthFactory({socket: btsClientMock}),
            sessionStorage
        });
    });

    describe('login online user', () => {

        it('on start user should be logout', () => {
            assertIsLogout();
        });

        it('should login user', (done) => {
            auth.loginOnlineUser(validOnlineLoginData())
                .then(() => {
                    assert.isTrue(auth.isLoggedIn());
                })
                .then(done).catch(done);
        });

        it('should store user data after successful login', (done) => {
            const loginData = validOnlineLoginData();
            auth.loginOnlineUser(loginData)
                .then(() => {
                    assert.isTrue(auth.isOnlineUser());
                    assert.isFalse(auth.isOfflineUser());
                    assert.equal(auth.getHost(), loginData.host);
                    assert.equal(auth.getPort(), loginData.port);
                    assert.equal(auth.isSecure(), loginData.isSecure);
                    assert.equal(auth.getToken(), loginData.token);
                    assert.equal(auth.getAuthMethod(), loginData.authMethod);
                })
                .then(done).catch(done);
        });

        it('should remove listener after login', (done) => {
            auth.loginOnlineUser(validOnlineLoginData())
                .then(() => {
                    assert.isFalse(btsClientMock.hasListener());
                })
                .then(done).catch(done);
        });

        it('should login and logout user', (done) => {
            auth.loginOnlineUser(validOnlineLoginData())
                .then(() => {
                    assert.isTrue(auth.isLoggedIn());
                    auth.logout();
                    assertIsLogout();
                })
                .then(done).catch(done);
        });

        it('should not login invalid user', (done) => {
            auth.loginOnlineUser(invalidOnlineLoginData())
                .then(() => {
                    assert.fail();
                })
                .catch(() => {
                    assertIsLogout();
                })
                .then(done).catch(done);
        });

        it('should remove listener after login failure', (done) => {
            auth.loginOnlineUser(invalidOnlineLoginData())
                .then(() => {
                    assert.fail();
                })
                .catch(() => {
                    assert.isFalse(btsClientMock.hasListener());
                })
                .then(done).catch(done);
        });

        it('should not login because of connection error', (done) => {
            btsClientMock.connect = () => {
                btsClientMock.listener.connectionError({errorCode: -1, message: 'Cannot connect!'});
            };

            auth.loginOnlineUser(validOnlineLoginData())
                .then(() => {
                    assert.fail();
                })
                .catch((errors) => {
                    assert.deepEqual(errors, ['Cannot connect!']);
                    assertIsLogout();
                })
                .then(done).catch(done);
        });

        it('should not login because of compatibility error', (done) => {
            const error = {message: 'Compatibility error', data: {available: 1, required: 2}};
            btsClientMock.connect = () => {
                btsClientMock.listener.compatibilityError(error);
            };

            auth.loginOnlineUser(validOnlineLoginData())
                .then(() => {
                    assert.fail();
                })
                .catch((errors) => {
                    assert.deepEqual(errors, ['Invalid gateway version. Available: 1, required: 2']);
                    assertIsLogout();
                })
                .then(done).catch(done);
        });

    });

    describe('login offline user', () => {

        beforeEach((done) => {
            auth.loginOfflineUser()
                .then(done).catch(done);
        });

        it('should not use btsClient when login offline user', (done) => {
            btsClientMock.addListener = () => {
                throw Error('Should be offline!');
            };

            auth.loginOfflineUser()
                .then(() => {
                    assert.isTrue(auth.isOfflineUser());
                    assert.isFalse(auth.isOnlineUser());
                })
                .then(done).catch(done);
        });

        it('should limit access to functionality', () => {
            assert.isFalse(auth.hasAccessTo('CONNECT_TO_BTS'));
            assert.isFalse(auth.hasAccessTo('USER_ACCOUNT'));
            assert.isFalse(auth.hasAccessTo('SHOW_PARAMETERS'));
            assert.isFalse(auth.hasAccessTo('SHOW_PARAMETER_HISTORY'));
            assert.isFalse(auth.hasAccessTo('DOWNLOAD_OFFLINE_VERSION'));
            assert.isFalse(auth.hasAccessTo('SETTINGS_UPDATE_INTERVAL'));
            assert.isFalse(auth.hasAccessTo('OTHER_ABNORMALITIES'));
            assert.isFalse(auth.hasAccessTo('SHOW_TIMELINE_DETAILS'));
            assert.isFalse(auth.hasAccessTo('SHOW_SITE_PARAMETERS'));
            assert.isFalse(auth.hasAccessTo('STATE_INFO_CHART'));
            assert.isFalse(auth.hasAccessTo('SOFTWARE_DETAILS'));
        });
    });

    describe('login users based on profile', () => {

        it('should login as developer', (done) => {
            auth.loginOnlineUser(validOnlineLoginData('developer'))
                .then(() => {
                    assert.isTrue(auth.isLoggedIn());
                    assert.equal(auth.getProfileName(), 'developer');
                })
                .then(done).catch(done);
        });

        it('should login as operator', (done) => {
            auth.loginOnlineUser(validOnlineLoginData('operator'))
                .then(() => {
                    assert.isTrue(auth.isLoggedIn());
                    assert.equal(auth.getProfileName(), 'operator');
                })
                .then(done).catch(done);
        });

        it('should login as tester', (done) => {
            auth.loginOnlineUser(validOnlineLoginData('tester'))
                .then(() => {
                    assert.isTrue(auth.isLoggedIn());
                    assert.equal(auth.getProfileName(), 'tester');
                })
                .then(done).catch(done);
        });

    });

    describe('operator access rights', () => {

        beforeEach((done) => {
            auth.loginOnlineUser(validOnlineLoginData('operator'))
                .then(() => {
                    assert.equal(auth.getProfileName(), 'operator');
                })
                .then(done).catch(done);
        });

        it('should limit access to views', () => {
            assert.isTrue(auth.isRouteIdValid('site-runtime'));

            assert.isFalse(auth.isRouteIdValid('info-model'));
            assert.isFalse(auth.isRouteIdValid('procedures-editor'));
            assert.isFalse(auth.isRouteIdValid('process-analysis-overview'));
            assert.isFalse(auth.isRouteIdValid('process-analysis-editor'));
        });

        it('should limit access to functionality', () => {
            assert.isTrue(auth.hasAccessTo('CONNECT_TO_BTS'));

            assert.isFalse(auth.hasAccessTo('SHOW_PARAMETERS'));
            assert.isFalse(auth.hasAccessTo('SHOW_PARAMETER_HISTORY'));
            assert.isFalse(auth.hasAccessTo('DOWNLOAD_OFFLINE_VERSION'));
            assert.isFalse(auth.hasAccessTo('SETTINGS_UPDATE_INTERVAL'));
            assert.isFalse(auth.hasAccessTo('OTHER_ABNORMALITIES'));
            assert.isFalse(auth.hasAccessTo('SHOW_TIMELINE_DETAILS'));
            assert.isFalse(auth.hasAccessTo('SHOW_SITE_PARAMETERS'));
            assert.isFalse(auth.hasAccessTo('STATE_INFO_CHART'));
            assert.isFalse(auth.hasAccessTo('SOFTWARE_DETAILS'));
        });

    });

    describe('developer access rights', () => {

        beforeEach((done) => {
            auth.loginOnlineUser(validOnlineLoginData('developer'))
                .then(() => {
                    assert.equal(auth.getProfileName(), 'developer');
                })
                .then(done).catch(done);
        });

        it('should have access to limited functionality', () => {
            assert.isTrue(auth.hasAccessTo('CONNECT_TO_BTS'));

            assert.isTrue(auth.hasAccessTo('SHOW_PARAMETERS'));
            assert.isTrue(auth.hasAccessTo('SHOW_PARAMETER_HISTORY'));
            assert.isTrue(auth.hasAccessTo('DOWNLOAD_OFFLINE_VERSION'));
            assert.isTrue(auth.hasAccessTo('SETTINGS_UPDATE_INTERVAL'));
            assert.isTrue(auth.hasAccessTo('OTHER_ABNORMALITIES'));
            assert.isTrue(auth.hasAccessTo('SHOW_TIMELINE_DETAILS'));
            assert.isTrue(auth.hasAccessTo('SHOW_SITE_PARAMETERS'));
            assert.isTrue(auth.hasAccessTo('STATE_INFO_CHART'));
            assert.isTrue(auth.hasAccessTo('SOFTWARE_DETAILS'));
        });

    });

    describe('screen lock login', () => {
        it('should accept valid token on session locked', (done) => {
            auth.loginOnlineUser(validOnlineLoginData('developer'));
            const authData = {token: 'any'};

            auth.loginUserWhenSessionLocked(authData)
                .then((eventMsg) => {
                    assert(eventMsg === 'Token validated');
                    done();
                });
        });

        it('should not accept invalid token on session locked', (done) => {
            auth.loginOnlineUser(validOnlineLoginData('developer'));
            const authData = {token: 'invalid_token'};

            auth.loginUserWhenSessionLocked(authData)
                .catch((eventMsg) => {
                    assert(eventMsg === 'Invalid token');
                    done();
                });
        });

        it('should not accept empty token on session locked', (done) => {
            auth.loginOnlineUser(validOnlineLoginData('developer'));
            const authData = {token: ''};

            auth.loginUserWhenSessionLocked(authData)
                .catch((eventMsg) => {
                    assert(eventMsg === 'Token not provided');
                    done();
                });
        });

    });

    function validOnlineLoginData(userName) {
        return {
            userName: userName || USERS.DEVELOPER.userName,
            token: 'any',
            authMethod: auth.AUTH_METHOD.SHA1,
            host: '127.0.0.1',
            port: 8080,
            isSecure: false
        };
    }

    function invalidOnlineLoginData() {
        const data = validOnlineLoginData();
        data.userName = 'any user name';
        return data;
    }

    function createBtsClientMock(users) {
        return {
            listener: null,
            connect() {
                if (this.listener && this.listener.connected) this.listener.connected();
            },
            disconnect() {
                if (this.listener && this.listener.disconnected) this.listener.disconnected();
            },
            login(loginData) {
                const user = users[loginData.userName.toUpperCase()];
                if (user) {
                    const loginInfo = {
                        profile: user.profile,
                        username: user.userName
                    };
                    this.listener.loggedIn(loginInfo);
                } else {
                    this.listener.loginError({
                        errorCode: 1025,
                        message: 'User authentication error [invalidLogin]'
                    });
                }
            },
            addListener(value) {
                this.listener = value;
            },
            removeListener() {
                this.listener = null;
            },
            hasListener() {
                return this.listener !== null;
            },
            getSicInstance() {
                return sicInst;
            }
        };
    }

    function assertIsLogout() {
        assert.isFalse(auth.isLoggedIn(), 'is loggedIn!');
        assert.isFalse(auth.isOnlineUser(), 'is online user!');
        assert.isFalse(auth.isOfflineUser(), 'is offline user!');

        assert.notOk(auth.getHost(), 'has host!');
        assert.notOk(auth.getPort(), 'has port!');
        assert.notOk(auth.getUserName(), 'has userName!');
        assert.notOk(auth.getProfileName(), 'has profileName!');
        assert.notOk(auth.isSecure(), 'has isSecure!');
        assert.notOk(auth.getToken(), 'has token!');
        assert.notOk(auth.getAuthMethod(), 'has authMethod!');
    }

});
