(function (global) {
    'use strict';

    var assert = chai.assert;
    var collections = global.collections;

    describe('bi multi map', function () {
        var map;

        beforeEach(function () {
            map = collections.newBiMultiMap();
        });

        it('should put values', function () {
            // when
            map.put('key 1', 1);
            map.put('key 1', 2);
            map.put('key 1', 3);
            map.put('key 1', 1);
            map.put('key 1', 3);

            map.put(2, 'a');
            map.put(2, 'b');
            map.put(2, 'a');

            const objectKey = {key: 'key'};
            const objectValue = {value: 'value'};
            map.put(objectKey, objectValue);
            map.put(objectKey, {value: 'value'});
            map.put(objectKey, objectValue);

            // then
            const values1 = map.values('key 1');
            const values2 = map.values(2);
            const values3 = map.values(objectKey);
            const values4 = map.values('undefined key');

            expect(values1).to.have.length(3);
            assert.include(values1, 1);
            assert.include(values1, 2);
            assert.include(values1, 3);

            expect(values2).to.have.length(2);
            assert.include(values2, 'a');
            assert.include(values2, 'b');

            expect(values3).to.have.length(2);

            expect(values4).to.be.undefined;
        });

        it('should put value to multiple keys', function () {
            // given
            const value = 'some value';

            // when
            map.put('key 1', value);
            map.put(2, value);
            const objectKey = {};
            map.put(objectKey, value);
            map.put('key 1', value);

            // then
            const keys = map.keys(value);
            expect(keys).to.have.length(3);
            assert.include(keys, 'key 1');
            assert.include(keys, 2);
            assert.include(keys, objectKey);
        });

        it('should return all keys', function () {
            // when
            map.put(1, 2);
            map.put('a', 'b');
            const objectKey = {};
            map.put(objectKey, {});

            // then
            const keys = map.keys();
            expect(keys).to.have.length(3);
            assert.include(keys, 1);
            assert.include(keys, 'a');
            assert.include(keys, objectKey);
        });

        it('should return all values', function () {
            // when
            map.put(1, 2);
            map.put('a', 'b');
            const objectValue = {};
            map.put({}, objectValue);

            // then
            const values = map.values();
            expect(values).to.have.length(3);
            assert.include(values, 2);
            assert.include(values, 'b');
            assert.include(values, objectValue);
        });

        it('should remove key', function () {
            // given
            map.put('key 1', 1);
            map.put('key 1', 2);

            map.put('key 2', 1);

            // when
            const values = map.removeKey('key 1');

            // then
            expect(values).to.have.length(2);
            assert.include(values, 1);
            assert.include(values, 2);

            const allKeys = map.keys();
            const allValues = map.values();

            expect(allKeys).to.have.length(1);
            assert.include(allKeys, 'key 2');

            expect(allValues).to.have.length(1);
            assert.include(allValues, 1);

            const keys = map.keys(1);
            expect(keys).to.have.length(1);
            assert.include(keys, 'key 2');
        });

        it('should remove value', function () {
            // given
            map.put('key 1', 1);
            map.put('key 2', 1);

            map.put('key 1', 2);

            // when
            const keys = map.removeValue(1);

            // then
            expect(keys).to.have.length(2);
            assert.include(keys, 'key 1');
            assert.include(keys, 'key 2');

            const allKeys = map.keys();
            const allValues = map.values();

            expect(allKeys).to.have.length(1);
            assert.include(allKeys, 'key 1');

            expect(allValues).to.have.length(1);
            assert.include(allValues, 2);

            const values = map.values('key 1');
            expect(values).to.have.length(1);
            assert.include(values, 2);
        });

        it('should remove entry', function () {
            // given
            map.put('key 1', 1);
            map.put('key 2', 1);
            map.put('key 2', 2);

            // when
            map.removeEntry('key 2', 1);

            // then
            const values = map.values('key 2');
            expect(values).to.have.length(1);
            assert.include(values, 2);

            const allKeys = map.keys();
            const allValues = map.values();

            expect(allKeys).to.have.length(2);
            assert.include(allKeys, 'key 1');
            assert.include(allKeys, 'key 2');

            expect(allValues).to.have.length(2);
            assert.include(allValues, 1);
            assert.include(allValues, 2);
        });

        it('should remove entry with removal of unused keys', function () {
            // given
            map.put('key 1', 1);
            map.put('key 2', 1);

            // when
            map.removeEntry('key 2', 1);

            // then
            const allKeys = map.keys();
            const allValues = map.values();

            expect(allKeys).to.have.length(1);
            assert.include(allKeys, 'key 1');

            expect(allValues).to.have.length(1);
            assert.include(allValues, 1);
        });

        it('should remove entry with removal of unused values', function () {
            // given
            map.put('key 1', 1);
            map.put('key 2', 2);

            // when
            map.removeEntry('key 2', 2);

            // then
            const allKeys = map.keys();
            const allValues = map.values();

            expect(allKeys).to.have.length(1);
            assert.include(allKeys, 'key 1');

            expect(allValues).to.have.length(1);
            assert.include(allValues, 1);
        });

        it('should check if map contains entry', function () {
            // given
            map.put('key 1', 1);
            map.put('key 2', 2);

            // then
            expect(map.containsEntry('key 1', 1)).to.be.true;
            expect(map.containsEntry('key 2', 2)).to.be.true;

            expect(map.containsEntry('key 1', 2)).to.be.false;
            expect(map.containsEntry('key 2', 1)).to.be.false;
            expect(map.containsEntry('key 2', null)).to.be.false;
        });

        it('should check if map contains key', function () {
            // given
            map.put('key 1', 1);
            map.put('key 2', 2);

            // then
            expect(map.containsKey('key 1')).to.be.true;
            expect(map.containsKey('key 2')).to.be.true;

            expect(map.containsKey('key 3')).to.be.false;
            expect(map.containsKey(null)).to.be.false;
        });

        it('should check if map contains value', function () {
            // given
            map.put('key 1', 1);
            map.put('key 2', 2);

            // then
            expect(map.containsValue(1)).to.be.true;
            expect(map.containsValue(2)).to.be.true;

            expect(map.containsValue(3)).to.be.false;
            expect(map.containsValue(null)).to.be.false;
        });

        it('should count keys', function () {
            // given
            map.put('key 1', 1);
            map.put('key 1', 2);
            map.put('key 1', 3);
            map.put('key 2', 2);
            map.put('key 2', 3);

            // then
            expect(map.countKeys()).to.be.equal(2);
            expect(map.countKeys(1)).to.be.equal(1);
            expect(map.countKeys(2)).to.be.equal(2);
            expect(map.countKeys(999)).to.be.equal(0);
        });

        it('should count values', function () {
            // given
            map.put('key 1', 1);
            map.put('key 1', 2);
            map.put('key 1', 3);
            map.put('key 2', 2);
            map.put('key 2', 3);

            // then
            expect(map.countValues()).to.be.equal(3);
            expect(map.countValues('key 1')).to.be.equal(3);
            expect(map.countValues('key 2')).to.be.equal(2);
            expect(map.countValues(999)).to.be.equal(0);
            expect(map.countValues(null)).to.be.equal(0);
        });
    });
})(this);
