describe('btsClient', function () {

    const LOGIN_DATA = {
        userName: 'any login',
        token: 'any token',
        clientVersion: 'any client version',
        clientId: '1234',
        authMethodId: 1,
        sessionType: 2
    };

    var btsClient, socket, sic;

    beforeEach(function () {
        sic = infoModel.sic(9);

        var sentMessages = [];
        var sicBuffer = sic.createSicMessageBuffer({
            completedMessage: function (message) {
                sentMessages.push(message);
            }
        });
        socket = {
            isConnected: function () {
                return true;
            },
            send: function (arrayBuffer) {
                sicBuffer.push(arrayBuffer);
            },
            close: function (code, reason) {
                if (this.onclose) this.onclose({code: code, reason: reason});
            },
            //methods for testing
            pullMessage: function () {
                return _.pullAt(sentMessages, 0)[0];
            },
            pushMessage: function (message) {
                var event = {
                    data: toSicMessageBytes(message)
                };
                this.onmessage(event);
            }
        };

        var socketFactory = {
            create: function () {
                return socket;
            }
        };
        btsClient = infoModel.btsClient(socketFactory);
        btsClient.connect('127.0.0.1', 123, false);
    });

    it('should get default update interval', function () {
        assert.equal(infoModel.btsClient.getDefaultUpdateIntervalMs(), 5000);
    });

    describe('connect', function () {

        it('on socket open should do send get gateway protocol version request', function () {
            socket.onopen();

            assertSentGetGatewayProtocolVersionRequest();
        });

        it('should send login request', function () {
            btsClient.login(LOGIN_DATA);

            assertSentLoginRequest();
        });

    });

    describe('listener', function () {
        var listener;

        beforeEach(function () {
            listener = createBtsClientListener();
            btsClient.addListener(listener);
            listener.events.assertSize(0);
        });

        it('should notify about connection error when cannot create socket', function () {
            var socketFactory = {
                create: function () {
                    const e = new Error();
                    e.code = 1000;
                    e.message = 'Connection error!';
                    throw e;
                }
            };
            btsClient = infoModel.btsClient(socketFactory);
            btsClient.addListener(listener);
            btsClient.connect('127.0.0.1', 123, false);

            listener.events.assertEvents([
                {name: 'connectionError', args: {errorCode: 1000, message: 'Connection error!'}}
            ]);
        });

        it('should notify about compatibility error when gateway protocol versions are different', () => {
            socket.onopen();

            var versionResp = sic.imgwGetProtocolVersionResp();
            versionResp.setVersion(1);
            socket.pushMessage(versionResp);

            listener.events.assertEvents([
                {name: 'compatibilityError'}
            ]);
        });

        it('should notify about compatibility error when gateway protocol versions are different', () => {
            socket.onopen();

            var versionResp = sic.imgwGetProtocolVersionResp();
            versionResp.setVersion(1);
            socket.pushMessage(versionResp);

            listener.events.assertEvents([
                {name: 'compatibilityError'}
            ]);
        });

        it('should notify listener about connected', function () {
            socket.onopen();

            var versionResp = sic.imgwGetProtocolVersionResp();
            versionResp.setVersion(2);
            socket.pushMessage(versionResp);

            listener.events.assertEvents([
                {name: 'connected', args: 'Connected to 127.0.0.1:123'}
            ]);
        });


        it('should notify listener about loggedIn', function () {
            const loginSuccessEvent = {
                profile: 'any profile',
                userName: 'name',
                failedLoginNo: 3,
                lastSuccess: '3 days ago'
            };

            btsClient.login(LOGIN_DATA);
            assertSentLoginRequest();

            const loginSuccessResp = sic.imgwLoginSuccessResp();
            loginSuccessResp.setProfile(loginSuccessEvent.profile);
            loginSuccessResp.setUserName(loginSuccessEvent.userName);
            loginSuccessResp.setFailedLoginNo(loginSuccessEvent.failedLoginNo);
            loginSuccessResp.setLastSuccess(loginSuccessEvent.lastSuccess);
            socket.pushMessage(loginSuccessResp);

            listener.events.assertEvents([
                {
                    name: 'loggedIn',
                    args: loginSuccessEvent
                }
            ]);
        });

        it('should notify listener about login error', function () {
            btsClient.login(LOGIN_DATA);
            assertSentLoginRequest();

            var loginErrorResp = sic.imgwLoginErrorResp();
            loginErrorResp.setDescription('login error!');
            loginErrorResp.setErrorCode(1);
            socket.pushMessage(loginErrorResp);

            listener.events.assertEvents([
                {
                    name: 'loginError',
                    args: {
                        errorCode: loginErrorResp.getErrorCode(),
                        message: loginErrorResp.getDescription()
                    }
                }
            ]);
        });

        it('should notify listener about gateway information', function () {
            const gatewayInformationEvent = {
                embeddedAdminId: '#1',
                embeddedAdminVersion: 'ADMIN-release-2016.06.01-1',
                endianness: 1,
                imProtocolRelease: 'IB1604029',
                imProtocolVersionMajor: 38,
                imProtocolVersionMinor: 0,
                imiVersionMajor: 9,
                imiVersionMinor: 0,
                infoModelSicAddress: 269555139
            };
            const gatewayInformationResp = sic.imgwGetGatewayInformationResp();

            gatewayInformationResp.setEmbeddedAdminId(gatewayInformationEvent.embeddedAdminId);
            gatewayInformationResp.setEmbeddedAdminVersion(gatewayInformationEvent.embeddedAdminVersion);
            gatewayInformationResp.setEndianness(gatewayInformationEvent.endianness);
            gatewayInformationResp.setImProtocolRelease(gatewayInformationEvent.imProtocolRelease);
            gatewayInformationResp.setImProtocolVersionMajor(gatewayInformationEvent.imProtocolVersionMajor);
            gatewayInformationResp.setImProtocolVersionMinor(gatewayInformationEvent.imProtocolVersionMinor);
            gatewayInformationResp.setImiVersionMajor(gatewayInformationEvent.imiVersionMajor);
            gatewayInformationResp.setImiVersionMinor(gatewayInformationEvent.imiVersionMinor);
            gatewayInformationResp.setInfoModelSicAddress(gatewayInformationEvent.infoModelSicAddress);

            socket.pushMessage(gatewayInformationResp);

            listener.events.assertEvents([
                {
                    name: 'gatewayInformation',
                    args: gatewayInformationEvent
                }
            ]);
        });

        it('should notify listener about signInCompleted after first data', function () {
            socket.pushMessage(sic.updateNotification());

            listener.events.assertEvents([
                {name: 'signInCompleted'},
                {name: 'data'}
            ]);
        });

        it('should notify listener about incoming data', function () {
            socket.pushMessage(sic.updateNotification());
            socket.pushMessage(sic.updateNotification());

            var expectedData = {
                removedIds: [],
                managedObjects: []
            };

            listener.events.assertEvents([
                {name: 'signInCompleted'},
                {name: 'data', args: expectedData},
                {name: 'data', args: expectedData}
            ]);
        });

        it('should notify listener about error and close connection', function () {
            socket.onerror(Error('Bum!'));

            listener.events.assertEvents([
                {name: 'connectionError', args: {errorCode: -1, message: 'Cannot connect to BTS...'}}
            ]);
        });

    });

    describe('custom message listener', function () {
        const FILE_ERROR_MSG = 0x6011;

        it('should be called when his message has been received', function () {
            var receivedMessageId;
            btsClient.addMessageListener(FILE_ERROR_MSG, function (message) {
                receivedMessageId = message.getId();
            });
            socket.onopen();
            socket.pushMessage(sic.imgwFileServiceErrorResp());
            assert.strictEqual(receivedMessageId, FILE_ERROR_MSG);
        });
    });


    function assertSentLoginRequest(requestParams) {
        assertSentRequest(sic.imgwLoginReq().getId(), requestParams);
    }

    function assertSentGetGatewayProtocolVersionRequest() {
        assertSentRequest(sic.imgwGetProtocolVersionReq().getId());
    }

    function toSicMessageBytes(msg) {
        var payloadBuffer = msg.getBuffer();
        var payloadLength = payloadBuffer.byteLength;
        var header = sic.createHeader();
        header.setId(msg.getId());
        header.setPayloadLength(payloadLength);
        var sicMsg = sic.createMessage(header);
        sicMsg.setPayload(payloadBuffer);
        return sicMsg.getBuffer();
    }

    function assertSentRequest(requestId, requestParams) {
        var msg = socket.pullMessage();
        assert.strictEqual(msg.getId(), requestId);

        if (!requestParams) return;

        var params = Object.keys(requestParams);
        _.forEach(params, function (param) {
            assert.strictEqual(msg[param](), requestParams[param]);
        });
    }

    function createBtsClientListener() {
        var listener = btsClient.createListener();
        var events = [];
        var keys = Object.keys(listener);
        _.forEach(keys, function (key) {
            if (key === 'sent') return;

            listener[key] = function () {
                events.push({name: key, args: arguments[0]});
            }
        });
        listener.events = {
            clear: function () {
                events = [];
            },
            assertSize: function (expectedSize) {
                assert.lengthOf(events, expectedSize);
            },
            assertEvents: function (expectedEvents) {
                this.assertSize(expectedEvents.length);
                for (var i = 0; i < events.length; i++) {
                    var expectedEvent = expectedEvents[i];
                    var event = events[i];

                    assert.strictEqual(event.name, expectedEvent.name);
                    if (expectedEvent.args) {
                        assert.deepEqual(event.args, expectedEvent.args);
                    }
                }
            }
        };
        return listener;
    }
});
