describe('objectIdMapper', () => {
    var objectIdMapper = require('../../src/communication/objectIdMapper');
    var mapper;
    beforeEach(function () {
        mapper = objectIdMapper.newInstance();
    });

    describe('#addClassName', ()=> {
        it('should return added types', () => {
            // when
            mapper.addClassName(0, 'CLASS_A');
            mapper.addClassName(1, 'CLASS_B');

            // then
            assert.equal(mapper.getClassName(0), 'CLASS_A');
            assert.equal(mapper.getClassName(1), 'CLASS_B');
            assert.isUndefined(mapper.getClassName(3));

            assert.equal(mapper.getClassNameId('CLASS_A'), 0);
            assert.equal(mapper.getClassNameId('CLASS_B'), 1);
            assert.isUndefined(mapper.getClassNameId('CLASS_C'));
        });

        it('should throw when adding duplicated type id', ()=> {
            mapper.addClassName(0, 'CLASS_A');

            assert.throws(()=> {
                mapper.addClassName(0, 'CLASS_B');
            });
        });

        it('should throw when adding duplicated type', ()=> {
            mapper.addClassName(0, 'CLASS_A');

            assert.throws(()=> {
                mapper.addClassName(1, 'CLASS_A');
            });
        });

        it('should register uppercase class', ()=> {
            mapper.addClassName(0, 'class_a');

            assert.equal(mapper.getClassName(0), 'CLASS_A');
        });
    });

    describe('#resolveDistName', ()=> {
        it('should resolve dist name', ()=> {
            // given
            mapper.addClassName(0, 'MRBTS');
            mapper.addClassName(1, 'EQM');
            mapper.registerObjectInfos([{
                uuid: '0x0',
                objectType: 0,
                instanceId: 0
            }, {
                uuid: '0x1',
                objectType: 1,
                instanceId: 1,
                parentUuid: '0x0'
            }]);
            // when
            var mrbtsDistName = mapper.resolveDistName('0x0');
            var eqmDistName = mapper.resolveDistName('0x1');

            // then
            assert.equal(mrbtsDistName, '/MRBTS-0');
            assert.equal(eqmDistName, '/MRBTS-0/EQM-1');
        });

        it('should not resolve distname when object info is not already registered', ()=> {
            //given when
            var distName = mapper.resolveDistName('0x0');

            //then
            assert.isNull(distName);
        });
    });


    describe('#removeUuid', ()=> {
        it('should remove uuid', ()=> {
            // given
            mapper.addClassName(0, 'MRBTS');
            var uuid = '0x0';
            mapper.registerObjectInfos([{
                uuid: uuid,
                objectType: 0,
                instanceId: 0

            }]);

            //when
            mapper.removeUuid(uuid);
            var distName = mapper.resolveDistName(uuid);

            //then
            assert.isNull(distName);
        });
    });

    describe('#resolveObjectInfo', ()=> {
        it('should return objectInfo for previously resolved distName', ()=> {
            // given
            mapper.addClassName(0, 'MRBTS');
            var expectedObjectInfo = {
                uuid: '0x1',
                objectType: 0,
                instanceId: 1

            };
            mapper.registerObjectInfos([{
                uuid: '0x0',
                objectType: 0,
                instanceId: 0

            }, expectedObjectInfo]);
            var distName = '/MRBTS-1';

            // when
            var objectInfo = mapper.resolveObjectInfo(distName);

            // then
            assert.deepEqual(expectedObjectInfo, objectInfo);
        });

        it('should return undefined when object was removed', ()=> {
            // given
            mapper.addClassName(0, 'MRBTS');
            var uuid = '0x0';
            mapper.registerObjectInfos([{
                uuid: uuid,
                objectType: 0,
                instanceId: 0
            }]);
            var distName = '/MRBTS-1';

            // when
            mapper.removeUuid(uuid);
            var objectInfo = mapper.resolveObjectInfo(distName);

            // then
            assert.isUndefined(objectInfo);
        });

        it('should throw error if distName is undefined', ()=> {
            assert.throws(mapper.resolveObjectInfo);
        });
        it('should throw error if distName is empty', ()=> {
            assert.throws(()=> {
                mapper.resolveObjectInfo('');
            });
        });
        it('should throw error if distName does not start with slash', ()=> {
            assert.throws(()=> {
                mapper.resolveObjectInfo('MRBTS-1');
            });
        });
        it('should throw error if distName does ends with slash', ()=> {
            assert.throws(()=> {
                mapper.resolveObjectInfo('/MRBTS-1/');
            });
        });
        it('should throw error if distName is not in upperCase', ()=> {
            assert.throws(()=> {
                mapper.resolveObjectInfo('/mrbts-1');
            });
        });
    });
});
