const _ = require('lodash');
const signInHandlerService = require('../../src/connection/signInHandlerService');
const btsClientSessionType = require('../../src/communication/btsClientSessionType');

describe('signInHandler', () => {
    const CLOSE_EVENT_BY_BTS = {source: 'BTS'};
    let signInHandler, socketMock, signInListener, signInData;

    beforeEach(() => {
        signInData = createSignInData();
        signInListener = createListener();
        socketMock = createSocketMock();
        signInHandler = signInHandlerService({socket: socketMock, socketSessionType: btsClientSessionType.INFO_MODEL});
        signInHandler.addListener(signInListener);
    });
    afterEach(() => {
        signInHandler.removeListener(signInListener);
        signInData = null;
        signInListener = null;
        socketMock = null;
        signInHandler = null;
    });

    describe('signIn', () => {

        it('should signIn', (done) => {
            signInListener.onEvents([
                'stateChanged:SIGNED_OUT->IN_PROGRESS',
                'connected',
                'stateChanged:IN_PROGRESS->SIGNED_IN',
                'signIn'
            ], () => {
                assert.isFalse(signInHandler.isSignInInProgress());
                assert.isTrue(signInHandler.isSignedIn());
                done();
            });

            signInHandler.signIn(signInData);
        });

        it('should notify about connection error', (done) => {
            signInListener.onEvents([
                'stateChanged:SIGNED_OUT->IN_PROGRESS',
                'error:cannot connect',
                'stateChanged:IN_PROGRESS->SIGNED_OUT'
            ], () => {
                assert.isFalse(signInHandler.isSignInInProgress());
                assert.isFalse(signInHandler.isSignedIn());
                done();
            });

            socketMock.connect = function () {
                this.notify('connectionError', {message: 'cannot connect'});
            };

            signInHandler.signIn(signInData);
        });

        it('should notify about compatibility error', (done) => {
            signInListener.onEvents([
                'stateChanged:SIGNED_OUT->IN_PROGRESS',
                'error:compatibility error',
                'stateChanged:IN_PROGRESS->SIGNED_OUT'
            ], () => {
                assert.isFalse(signInHandler.isSignInInProgress());
                assert.isFalse(signInHandler.isSignedIn());
                done();
            });

            socketMock.connect = function () {
                this.notify('compatibilityError', {message: 'compatibility error'});
            };

            signInHandler.signIn(signInData);
        });

        it('should notify about login error', (done) => {
            signInListener.onEvents([
                'stateChanged:SIGNED_OUT->IN_PROGRESS',
                'connected',
                'error:cannot login',
                'stateChanged:IN_PROGRESS->SIGNED_OUT',
                'disconnected'
            ], () => {
                assert.isFalse(signInHandler.isSignInInProgress());
                assert.isFalse(signInHandler.isSignedIn());
                done();
            });

            socketMock.login = function () {
                this.notify('loginError', {message: 'cannot login'});
            };

            signInHandler.signIn(signInData);
        });

        it('should reconnect after login error - SMA unreachable', (done) => {
            const expectedNumberOfFailedLogin = 3;

            const originalLogin = socketMock.login;
            socketMock.login = function () {
                this.notify('loginError', {errorCode: 1027, message: 'SMA unreachable'});
            };
            signInListener.onEvent('error:SMA unreachable', (occurrence) => {
                assert.isTrue(signInHandler.isSignInInProgress());
                if (occurrence === expectedNumberOfFailedLogin) socketMock.login = originalLogin;
            });

            signInListener.onEvent('signIn', () => {
                assert.deepEqual(signInListener.events(), [
                    'stateChanged:SIGNED_OUT->IN_PROGRESS',
                    'connected',
                    'error:SMA unreachable',
                    'autoReconnect',
                    'info:starting auto reconnect...',
                    'error:SMA unreachable',
                    'autoReconnect',
                    'info:starting auto reconnect...',
                    'error:SMA unreachable',
                    'autoReconnect',
                    'info:starting auto reconnect...',
                    'stateChanged:IN_PROGRESS->SIGNED_IN',
                    'signIn'
                ]);
                assert.isFalse(signInHandler.isSignInInProgress());
                assert.isTrue(signInHandler.isSignedIn());
                done();
            });

            signInHandler.enableAutoReconnect(1);
            signInHandler.signIn(signInData);
        });

        it('should auto reconnect after disconnect', (done) => {
            signInListener.onEvent('signIn', (occurrence) => {
                if (occurrence === 1) socketMock.notify('disconnected', CLOSE_EVENT_BY_BTS);

                if (occurrence === 2) {
                    assert.deepEqual(signInListener.events(), [
                        'stateChanged:SIGNED_OUT->IN_PROGRESS',
                        'connected',
                        'stateChanged:IN_PROGRESS->SIGNED_IN',
                        'signIn',
                        'disconnected',
                        'stateChanged:SIGNED_IN->IN_PROGRESS',
                        'autoReconnect',
                        'info:starting auto reconnect...',
                        'stateChanged:IN_PROGRESS->SIGNED_IN',
                        'signIn'
                    ]);
                    assert.isFalse(signInHandler.isSignInInProgress());
                    assert.isTrue(signInHandler.isSignedIn());
                    done();
                }
            });

            signInHandler.enableAutoReconnect(1);
            signInHandler.signIn(signInData);
        });

        it('should schedule reconnect after disconnect', (done) => {
            signInListener.onEvent('signIn', () => {
                socketMock.notify('disconnected', CLOSE_EVENT_BY_BTS);
            });
            signInListener.onEvent('autoReconnect', () => {
                assert.isTrue(signInHandler.isSignInInProgress());
                assert.isFalse(signInHandler.isSignedIn());
                done();
            });

            signInHandler.enableAutoReconnect(Number.MAX_SAFE_INTEGER);
            signInHandler.signIn(signInData);
        });

    });

    describe('signOut', () => {

        it('should signOut after signIn', (done) => {
            signInListener.onEvent('signIn', signInHandler.signOut);
            signInListener.onEvent('signOut', () => {
                assert.deepEqual(signInListener.events(), [
                    'stateChanged:SIGNED_OUT->IN_PROGRESS',
                    'connected',
                    'stateChanged:IN_PROGRESS->SIGNED_IN',
                    'signIn',
                    'disconnected',
                    'stateChanged:SIGNED_IN->SIGNED_OUT',
                    'signOut'
                ]);
                assert.isFalse(signInHandler.isSignInInProgress());
                assert.isFalse(signInHandler.isSignedIn());
                done();
            });

            signInHandler.signIn(signInData);
        });

        it('should do nothing when not signIn', () => {
            signInHandler.signOut();

            assert.isFalse(signInListener.hasEvents());
            assert.isFalse(signInHandler.isSignedIn());
        });

        it('after signOut should not react on socket listener', (done) => {
            signInListener.onEvent('signIn', signInHandler.signOut);
            signInListener.onEvent('signOut', () => {
                socketMock.notify('connectionError', {message: 'bum!'})
                    .then(() => {
                        assert.deepEqual(signInListener.events(), [
                            'stateChanged:SIGNED_OUT->IN_PROGRESS',
                            'connected',
                            'stateChanged:IN_PROGRESS->SIGNED_IN',
                            'signIn',
                            'disconnected',
                            'stateChanged:SIGNED_IN->SIGNED_OUT',
                            'signOut'
                        ]);
                        assert.isFalse(signInHandler.isSignedIn());
                    })
                    .then(done).catch(done);
            });

            signInHandler.signIn(signInData);
        });

    });

    function createSocketMock() {
        let listener = null;
        let connected = false;
        return {
            isConnected() {
                return connected;
            },
            connect() {
                this.notify('connected');
                connected = true;
            },
            disconnect() {
                this.notify('disconnected', {source: 'USER'});
                connected = false;
            },
            login() {
                this.notify('loggedIn');
                this.notify('signInCompleted');
            },
            addListener(newListener) {
                listener = newListener;
            },
            removeListener() {
                listener = null;
            },
            //for testing
            notify(event, data) {
                return new Promise((resolve) => {
                    global.setTimeout(() => {
                        if (listener)
                            listener[event](data);

                        resolve();
                    });
                });
            }
        };
    }

    function createSignInData() {
        return {
            host: 'any host',
            port: 123,
            isSSL: false,
            userName: 'ANY USERNAME',
            token: 'ANY TOKEN',
            authMethodId: 1
        };
    }

    function createListener() {
        const onEvent = new Map();
        const allEvents = [];

        function notify() {
            const eventsCallback = onEvent.get(JSON.stringify(allEvents));
            if (eventsCallback) eventsCallback();

            const event = _.last(allEvents);
            const eventCallback = onEvent.get(event);
            if (eventCallback) {
                const occurrence = _.countBy(allEvents)[event];
                eventCallback(occurrence);
            }
        }

        return {
            connected: () => {
                allEvents.push('connected');
                notify();
            },
            disconnected: () => {
                allEvents.push('disconnected');
                notify();
            },
            signIn: () => {
                allEvents.push('signIn');
                notify();
            },
            signOut: () => {
                allEvents.push('signOut');
                notify();
            },
            autoReconnect: () => {
                allEvents.push('autoReconnect');
                notify();
            },
            error: (error) => {
                allEvents.push(`error:${error}`);
                notify();
            },
            info: (info) => {
                allEvents.push(`info:${info}`);
                notify();
            },
            stateChanged: (newState, oldState) => {
                const e = `stateChanged:${oldState}->${newState}`;
                allEvents.push(e.toString());
                notify();
            },
            //for testing
            hasEvents: () => {
                return allEvents.length !== 0;
            },
            events: () => {
                return allEvents;
            },
            onEvents: (events, callback) => {
                onEvent.set(JSON.stringify(events), callback);
            },
            onEvent: (event, callback) => {
                onEvent.set(event, callback);
            }
        };
    }
});
