const dspsService = require('../../src/dsps/dspsService'),
    distNameService = require('../../src/utils/distNameUtils');


(function () {
    'use strict';

    describe('Baseband', function () {

        var dspsTable;

        var mockUpdateService = createMockUpdateService();
        var mockSiteModelFactory = createMockSiteModelFactory();

        beforeEach(function () {
            dspsTable = dspsService({
                updateService: mockUpdateService,
                siteModelFactory: mockSiteModelFactory
            });
            dspsTable.watch(function () {
            });
        });

        afterEach(function () {
            mockSiteModelFactory.clean();

        });

        describe('channel', function () {
            var imTree, cells;

            beforeEach(function () {
                imTree = mockInfoModelTree([]);
                cells = createCells(1);
            });

            it('label should have AxC address', function () {
                run(cells, imTree);
                expect(dspsTable.cells()[0].channels[0].label).to.be.equal('RX AxC 100');
            });

            it('label should  consist of RMOD connector lable if cell is active', function () {
                var RMOD_DN = '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1';
                var CONNECTOR_DN = '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1/CONNECTOR_L-0';

                var channel = createChannel('RX', 0, [RMOD_DN, CONNECTOR_DN]);
                cells[0].channels[0] = channel;

                var rmod = new ModuleBuilder(RMOD_DN).addConnector('ANT1', 'right').get();
                mockSiteModelFactory.addRmod(rmod);

                run(cells, imTree);

                expect(dspsTable.cells()[0].channels[0].label).to.be.equal('RX AxC 100 / ANT1');
            });

            it('label should  be display properly when AxC address is null', function () {
                cells[0].channels[0].axcAddress = null;

                run(cells, imTree);

                expect(dspsTable.cells()[0].channels[0].label).to.be.equal('RX');
            });
        });

        describe('cell', function () {
            var imTree;

            beforeEach(function () {
                imTree = mockInfoModelTree([]);
            });

            it('label should have technology, dlChBw and pci', function () {
                var cells = [];
                var cell = createCell(1, 'FDD', '10 MHz');
                cells.push(cell);
                run(cells, imTree);

                expect(dspsTable.cells()[0].label).to.be.equal('FDD 10 MHz PCI 51');
            });
        });

        describe('RMOD', function () {
            var imTree;

            beforeEach(function () {
                imTree = mockInfoModelTree([]);
            });

            it('label should have rmod distName and connector to smod group', function () {
                var cells = createCells(1);
                var RMOD_DN = '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1';
                var CONNECTOR_DN = '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1/CONNECTOR_L-0';

                var channel = createChannel('RX', 0, [RMOD_DN, CONNECTOR_DN]);
                cells[0].channels[0] = channel;

                var rmod = new ModuleBuilder(RMOD_DN).addConnector('OPT_IF1', 'left').get();
                mockSiteModelFactory.addRmod(rmod);

                run(cells, imTree);

                expect(dspsTable.cells()[0].channels[0].rmod.label).to.be.equal('RMOD_L-1 / OPT_IF1');
            });
        });

        describe('Module', function () {
            var NP_DN = '/MRBTS-1/RAT-1/BTS_CONF-1/NP-1';
            var BTS_DATA_DN = NP_DN + '/BTS_DATA-1';

            describe('SMOD', function () {
                var imTree, smodNode, channel, smod, cells;
                var SMOD_DN = '/MRBTS-1/RAT-1/EQM_L-1/CABINET_L-1/SMOD_L-1';
                var DSP_DN = SMOD_DN + '/BBU_L-1/SOC_L-1/DSP_L-1';
                var CONNECTOR_DN = SMOD_DN + '/CONNECTOR_L-0';

                var SMOD_A_DN = NP_DN + '/SCF-1/MRBTS_A-8/EQM_A-1/SMOD_A-1';

                beforeEach(function () {
                    smodNode = mockTreeNode('SMOD_L', SMOD_DN, []);
                    smod = new ModuleBuilder(SMOD_DN).addConnector('RF_EXT2', 'right').get();
                    mockSiteModelFactory.addSmod(smod);

                    channel = createChannel('RX', 0, [SMOD_DN, CONNECTOR_DN, DSP_DN]);
                    cells = createCells(1);
                    cells[0].channels[0] = channel;
                });


                it('should have empty label if connected SMOD_A does not exist', function () {
                    imTree = mockInfoModelTree([smodNode]);

                    run(cells, imTree);

                    expect(dspsTable.cells()[0].channels[0].smodGroup.label).to.be.equal('');
                });

                it('label shoud have SMOD_A name', function () {
                    var btsDataNode = createBtsDataNode();
                    imTree = mockInfoModelTree([btsDataNode, smodNode]);

                    run(cells, imTree);

                    expect(dspsTable.cells()[0].channels[0].smodGroup.label).to.be.equal('SMOD_A-1 / RF_EXT2');
                });

                it('should have visible product code', function () {
                    var btsDataNode = createBtsDataNode({prodCode: '1234'});
                    imTree = mockInfoModelTree([btsDataNode, smodNode]);

                    run(cells, imTree);

                    expect(dspsTable.cells()[0].channels[0].smodGroup.prodCode).to.be.equal('PC: 1234');
                });

                it('should have visible planned product code', function () {
                    var btsDataNode = createBtsDataNode({prodCodePlanned: '4321'});
                    imTree = mockInfoModelTree([btsDataNode, smodNode]);

                    run(cells, imTree);

                    expect(dspsTable.cells()[0].channels[0].smodGroup.prodCodePlanned).to.be.equal('PLANNED PC: 4321');
                });

                it('should have visible product name', function () {
                    var btsDataNode = createBtsDataNode({prodName: 'name'});
                    imTree = mockInfoModelTree([btsDataNode, smodNode]);

                    run(cells, imTree);

                    expect(dspsTable.cells()[0].channels[0].smodGroup.prodName).to.be.equal('name');
                });

                it('should have correct dsp count', function () {
                    var btsDataNode = createBtsDataNode({prodName: 'name'});
                    imTree = mockInfoModelTree([btsDataNode, smodNode]);

                    run(cells, imTree);

                    expect(dspsTable.cells()[0].channels[0].dsps).length.to.be(1);
                });

                it('should have correct dsp count when there is no SOC_L objects', function () {
                    var DSP_WITHOUT_SOC_DN = SMOD_DN + '/BBU_L-1/DSP_L-1';
                    var channelWithDsp = createChannel('RX', 0, [SMOD_DN, CONNECTOR_DN, DSP_WITHOUT_SOC_DN]);
                    cells[0].channels[0] = channelWithDsp;

                    var btsDataNode = createBtsDataNode({prodName: 'name'});
                    imTree = mockInfoModelTree([btsDataNode, smodNode]);

                    run(cells, imTree);

                    expect(dspsTable.cells()[0].channels[0].dsps).length.to.be(1);
                });


                it('should have correct dsp count when there is POWER_GROUP_L objects', function () {
                    var DSP_WITHOUT_SOC_DN = SMOD_DN + '/BBU_L-1/POWER_GROUP_L-1/SOC_L-1/DSP_L-1';
                    var channelWithDsp = createChannel('RX', 0, [SMOD_DN, CONNECTOR_DN, DSP_WITHOUT_SOC_DN]);
                    cells[0].channels[0] = channelWithDsp;

                    var btsDataNode = createBtsDataNode({prodName: 'name'});
                    imTree = mockInfoModelTree([btsDataNode, smodNode]);

                    run(cells, imTree);

                    expect(dspsTable.cells()[0].channels[0].dsps).length.to.be(1);
                });


                it('dsp should have correct name and distName', function () {
                    var btsDataNode = createBtsDataNode({prodName: 'name'});
                    imTree = mockInfoModelTree([btsDataNode, smodNode]);

                    run(cells, imTree);

                    expect(dspsTable.cells()[0].channels[0].dsps[0].name).to.be.equal('DSP_L-1');
                    expect(dspsTable.cells()[0].channels[0].dsps[0].distName).to.be.equal(DSP_DN);
                });

                function createBtsDataNode(smodAProperties) {
                    smodAProperties = Object.assign({realDN: SMOD_DN}, smodAProperties);
                    var smodANode = mockTreeNode('SMOD_A', SMOD_A_DN, [], smodAProperties);
                    var npNode = mockTreeNode('NP', NP_DN, [smodANode]);

                    var btsDataNode = mockTreeNode('BTS_DATA', BTS_DATA_DN, [], {stateInfo: {proceduralState: 'activated'}});
                    btsDataNode.parent = npNode;
                    return btsDataNode;
                }
            });
        });

        function run(cells, imTree) {
            mockSiteModelFactory.setCells(cells);
            mockUpdateService.notifyImTreeCompleted(imTree);
        }


        function mockInfoModelTree(nodes) {
            var infoModelTree = {};
            infoModelTree.findByClassName = function (className) {
                return nodes.filter(function (node) {
                    return node.className === className;
                });
            };
            infoModelTree.findByDistName = function (distName) {
                return nodes.find(function (node) {
                    return node.distName === distName;
                });
            };
            return infoModelTree;
        }

        function getIdFromDistName(distName) {
            var distNameParts = distName.split('/');
            return distNameParts[distNameParts.length - 1].split('-')[1];
        }

        function mockTreeNode(className, distName, children, parameters) {
            var treeNodeMock = {};
            treeNodeMock.className = className;
            treeNodeMock.distName = distName;
            treeNodeMock.id = getIdFromDistName(distName);
            treeNodeMock.findByClassName = function (searchedClassName) {
                return children.filter(function (child) {
                    return child.className === searchedClassName;
                });
            };
            treeNodeMock.parameters = parameters || {};
            treeNodeMock.getDecodedParameters = function () {
                return treeNodeMock.parameters;
            };
            return treeNodeMock;
        }

        function createMockSiteModelFactory() {
            var mockFactory = {};
            var cells = [];
            var logLinks = [];
            var rmods = [];
            var smods = [];
            var cabinet = {
                smods: smods
            };

            mockFactory.create = function () {
                return {cells: cells, logLinks: logLinks, rmods: rmods, cabinets: [cabinet]};
            };

            mockFactory.setCells = function (_cells) {
                cells = _cells;
            };

            mockFactory.addRmod = function (rmod) {
                rmods.push(rmod);
            };

            mockFactory.addSmod = function (smod) {
                smods.push(smod);
            };

            mockFactory.clean = function () {
                cells = [];
                logLinks = [];
                rmods = [];
                smods = [];
                cabinet = {
                    smods: smods
                };
            };


            return mockFactory;
        }

        function doNothing() {

        }

        function createMockUpdateService() {
            var listener = [];
            var updateService = {};
            updateService.addImTreeListener = function (treeListener) {
                listener.push(treeListener.completeImTree || doNothing);
            };
            updateService.removeImTreeListener = function () {
            };

            updateService.notifyImTreeCompleted = function (imTree) {
                listener.map(function (fun) {
                    fun(imTree);
                });

            };
            return updateService;
        }

        function createCells(num) {
            var cells = [];
            for (var i = 0; i < num; i++)
                cells.push(createCell(i));

            return cells;
        }

        function createCell(id, technology, dlChBw) {
            return {
                distName: 'cell-${id}',
                name: 'cell-${id}',
                number: id,
                channels: createChannels(1, 'RX'),
                logLinks: [],
                stateInfo: [],
                alarms: [],
                technology: technology,
                dlChBw: dlChBw,
                phyCellId: 50 + id
            };
        }

        function createChannels(num, direction) {
            var channels = [];
            for (var i = 0; i < num; i++)
                channels.push(createChannel(direction, i));
            return channels;
        }

        function createChannel(direction, id, devicesPath) {
            var number = id;
            var addressStart = 100;
            return {
                distName: 'undef',
                name: direction + ' #' + number,
                direction: direction,
                number: number,
                axcAddress: '' + (addressStart + id),
                devicesPath: devicesPath || []
            };
        }

        function createConnector(parentDistName, id, label) {
            return {
                distName: parentDistName + '/CONNECTOR_L-' + id,
                number: id,
                label: label
            };
        }

        function ModuleBuilder(distName) {
            var rightConnectors = [];
            var leftConnectors = [];

            this.addConnector = function (label, direction) {
                var connectorId = rightConnectors.length;
                var connector = createConnector(distName, connectorId, label);
                direction === 'right' ? rightConnectors.push(connector) : leftConnectors.push(connector);
                return this;
            };

            this.get = function () {
                return {
                    distName: distName,
                    className: distNameService.getClassName(distName),
                    number: getIdFromDistName(distName),
                    rightConnectors: rightConnectors,
                    leftConnectors: leftConnectors
                };
            };
        }
    });

})();
