(function () {
    'use strict';

    var assert = chai.assert;

    describe('events', function () {

        var event;
        beforeEach(function () {
                event = events.newEvent();
            }
        );

        it('should notify all event watchers', function () {
            var count = 0;
            event.watch(function () {
                count++;
            });

            event.watch(function () {
                count++;
            });

            event.notify('foo');
            event.notify('bar');

            assert.equal(count, 4);
        });

        it('should unsubscribe via subscription', function () {
            var events = [];
            var watcher = function (event) {
                events.push(event);
            };
            var watching = event.watch(watcher);

            event.notify('foo');

            watching.unwatch();

            event.notify('42');
            event.watch(watcher);
            event.notify('bar');

            assert.deepEqual(events, ['foo', 'bar']);
        });

        it('should unsubscribe via unwatch', function () {
            var events = [];
            var watcher = function (event) {
                events.push(event);
            };
            event.watch(watcher);

            event.notify('foo');

            event.unwatch(watcher);

            event.notify('42');
            event.watch(watcher);
            event.notify('bar');

            assert.deepEqual(events, ['foo', 'bar']);
        });

        it('should throw when removing watcher that was not added', function () {
            assert.throws(function () {
                event.unwatch(function () {
                });
            });
        });

        it('should notify all event watchers and return resolved future', function (done) {
            var count = 0;
            event.watch(function () {
                count++;
                return 'watcher 1';
            });

            event.watch(function () {
                count++;
                return 'watcher 2';
            });

            var promise1 = event.notify('foo');
            var promise2 = event.notify('bar');

            assert.equal(count, 4);
            Promise.all([promise1, promise2]).then(function (value) {
                assert.equal(value.length, 2);
                assert.deepEqual(value[0], ['watcher 1', 'watcher 2']);
                assert.deepEqual(value[1], ['watcher 1', 'watcher 2']);
            }).then(done).catch(done);
        });

        it('should notify all event watchers and return unresolved future', function (done) {
            var resolvePromise;

            event.watch(function () {
                return new Promise(function (resolve, reject) {
                    resolvePromise = resolve;
                });
            });

            event.watch(function () {
                return Promise.resolve('watcher 2');
            });

            var promise = event.notify('foo');
            var testPromise = testablePromise(promise);

            promise.then(function (value) {
                assert.isTrue(testPromise.isSettled());
                assert.isTrue(testPromise.isResolved());

                assert.deepEqual(value, ['watcher 1', 'watcher 2']);
            }).then(done).catch(done);

            resolvePromise('watcher 1');
        });

        it('should notify all event watchers and return unresolved future that will be rejected', function (done) {
            var rejectPromise;

            event.watch(function () {
                return new Promise(function (resolve, reject) {
                });
            });

            event.watch(function () {
                return new Promise(function (resolve, reject) {
                    rejectPromise = reject;
                });
            });

            var promise = event.notify('foo');
            var testPromise = testablePromise(promise);

            testPromise.catch(function (reason) {
                assert.isTrue(testPromise.isSettled());
                assert.isTrue(testPromise.isRejected());
                assert.equal(reason, 'test error');
            }).then(done).catch(done);

            rejectPromise('test error');
        });

        it('should call callback when first watcher is added', function () {
            var count = 0;

            event.onFirstAdded(function () {
                count++;
            });

            event.watch(function () {
            });
            assert.equal(count, 1);

            event.watch(function () {
            });

            assert.equal(count, 1);
        });

        it('should call callback when last watcher is removed', function () {
            var count = 0;

            event.onLastRemoved(function () {
                count++;
            });

            var watcher1 = function () {
            };
            var watcher2 = function () {
            };

            event.watch(watcher1);
            event.watch(watcher2);

            assert.equal(count, 0);

            event.unwatch(watcher2);
            assert.equal(count, 0);

            event.unwatch(watcher1);

            assert.equal(count, 1);
        });

        function testablePromise(promise) {
            var value;
            var error;
            var settled = false;
            var resolved = false;
            var rejected = false;

            const p = promise.then(function (v) {
                value = v;
                settled = true;
                resolved = true;
                return v;
            }, function (err) {
                error = err;
                settled = true;
                rejected = true;
                throw err;
            });

            p.isSettled = function () {
                return settled;
            };
            p.isResolved = function () {
                return resolved;
            };
            p.isRejected = function () {
                return rejected;
            };
            p.value = function () {
                return value;
            };
            p.error = function () {
                return error;
            };

            return p;
        }
    });
})();
