const _ = require('lodash');
const FileTransferService = require('../../src/fileTransfer/fileTransferService');
const fileTransferTestUtil = require('./fileTransferTestUtil');

describe('fileTransferService', () => {

    let requestId = 0;
    let socket, sic, fileTransferClient, fileTransferService, testUtil;

    beforeEach(() => {
        testUtil = fileTransferTestUtil();
        sic = testUtil.sic;
        socket = testUtil.socket;
        const authService = testUtil.authService;
        const applicationInfo = testUtil.applicationInfo;
        fileTransferClient = testUtil.fileTransferClient;
        fileTransferService = new FileTransferService(authService, applicationInfo);
        fileTransferService.setWebSocketClient(fileTransferClient);
    });

    it('should ask about gateway protocol version during connection', (done) => {
        fileTransferService.connect()
            .then(assertSentGetGatewayProtocolVersionRequest)
            .then(done).catch(done);

        testUtil.performConnect();

        function assertSentGetGatewayProtocolVersionRequest() {
            assertSentRequest(sic.imgwGetProtocolVersionReq(), 0);
        }
    });

    it('should send login request during connection', (done) => {
        fileTransferService.connect()
            .then(assertSentLoginRequest)
            .then(done).catch(done);

        testUtil.performConnect();

        function assertSentLoginRequest(requestParams) {
            assertSentRequest(sic.imgwLoginReq(), 1, requestParams);
        }
    });


    describe('getFile', () => {
        const URI = '/rom/config.txt';
        const EXPECTED_FILE_CONTENT = new Uint8Array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]);

        beforeEach((done) => {
            fileTransferService.connect().then(done).catch(done);
            testUtil.performConnect();
        });

        it('should throw error because of incorrect checkSum', (done) => {
            const errorStatus = {
                type: 'File checkSum incorrect',
                path: URI,
                error: 'Calculated checkSum is different then expected.'
            };
            const incorrectCheckSum = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20];
            fileTransferService.getFile(URI)
                .then(verifyArrayAsFail)
                .catch(handleError)
                .then(done).catch(done);

            sendFile(incorrectCheckSum);

            function verifyArrayAsFail() {
                assert.fail(true, "File shouldn't be downloaded");
            }

            function handleError(receivedErrorStatus) {
                assert.equal(receivedErrorStatus.type, errorStatus.type);
                assert.equal(receivedErrorStatus.path, errorStatus.path);
                assert.equal(receivedErrorStatus.error, errorStatus.error);
            }
        });

        it('should get file as Uint8Array', (done) => {
            fileTransferService.getFile(URI)
                .then(verifyArray)
                .catch(handleError)
                .then(done).catch(done);

            sendFile();

            function verifyArray(resultArray) {
                assert.deepEqual(EXPECTED_FILE_CONTENT, resultArray);
            }

            function handleError(error) {
                assert.fail(true, `Error occured ${error}`);
            }
        });

        it('should handle error correctly', (done) => {
            const errorStatus = {
                type: 'File download error',
                path: URI,
                error: 'some error'
            };

            fileTransferService.getFile(URI)
                .then(verifyArrayAsFail)
                .catch(handleError)
                .then(done).catch(done);

            socket.receive(createErrorMessage());

            function verifyArrayAsFail() {
                assert.fail(true, "File shouldn't be downloaded");
            }

            function handleError(receivedErrorStatus) {
                assert.equal(receivedErrorStatus.type, errorStatus.type);
                assert.equal(receivedErrorStatus.path, errorStatus.path);
                assert.equal(receivedErrorStatus.error, errorStatus.error);
            }

            function createErrorMessage() {
                const message = sic.imgwFileServiceErrorResp();
                message.setRequestId(requestId++);
                message.setErrorCode(4);
                message.setDescription(errorStatus.error);
                message.isSicPlus = false;
                return message;
            }
        });

        it('should handle socket disconnected correctly', (done) => {
            const errorStatus = {
                type: 'Connection closed error',
                path: URI,
                error: 'Connection closed during file transfer.'
            };

            fileTransferService.getFile(URI)
                .then(fail)
                .catch(handleError)
                .then(done).catch(done);

            fileTransferClient.disconnect();

            function fail() {
                assert.fail(true, "File shouldn't be downloaded");
            }

            function handleError(receivedErrorStatus) {
                assert.equal(receivedErrorStatus.type, errorStatus.type);
                assert.equal(receivedErrorStatus.path, errorStatus.path);
                assert.equal(receivedErrorStatus.error, errorStatus.error);
            }
        });

        function sendFile(checkSum) {
            _.forEach(createSicPlusMessages(checkSum), (message) => {
                socket.receive(message);
            });
            requestId++;
        }

        function createSicPlusMessages(checkSum) {
            const fileResponseMessage = sic.imgwFileServiceFileContentResp();
            fileResponseMessage.setRequestId(requestId);
            fileResponseMessage.setContent(EXPECTED_FILE_CONTENT);
            if (checkSum) fileResponseMessage.setHash(checkSum);
            const buffer = fileResponseMessage.getBuffer();
            const msgSize = buffer.byteLength;
            const bufferArray = new Uint8Array(buffer);
            const messages = [];
            for (let i = 0; i < msgSize; i++) {
                const message = createMessage(msgSize, new Uint8Array([bufferArray[i]]), i);
                messages.push(message);
            }
            return messages;
        }

        function createMessage(msgSize, payload, offset) {
            const message = sic.createSicPlusMessage();
            message.setSeqId(1);
            message.setSender(0);
            message.setMsgId(0x6012);
            message.setMsgSize(msgSize);
            message.setDataOffset(offset);
            message.setData(payload);
            message.isSicPlus = true;
            return message;
        }
    });

    function assertSentRequest(message, positionInBuffer, requestParams) {
        const msg = socket.getMessage(positionInBuffer);
        assert.strictEqual(msg.getId(), message.getId());
        if (!requestParams) return;
        const params = Object.keys(requestParams);
        _.forEach(params, (param) => {
            assert.strictEqual(msg[param](), requestParams[param]);
        });
    }
});
