const FileUploadService = require('../../src/fileTransfer/fileUploadService');
const fileTransferTestUtil = require('./fileTransferTestUtil');

describe('fileUploadService', () => {

    const SESSION_TYPE_SW_UPDATE = 3;
    const REQUEST_ID = 1;
    const URI = '/rom/config 1.txt';
    const INCORRECT_URI = 'rom/config.zip';
    const EXPECTED_FILE_CONTENT = new Uint8Array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]);

    const fileDispatcher = () => {
        const files = new Map([[URI, EXPECTED_FILE_CONTENT]]);
        return {
            getFile: getFile
        };
        function getFile(uri) {
            return files.get(uri);
        }
    };
    let testUtil, socket, sic, fileTransferClient, fileUploadService;

    beforeEach(() => {
        testUtil = fileTransferTestUtil();
        sic = testUtil.sic;
        socket = testUtil.socket;
        const authService = testUtil.authService;
        const applicationInfo = testUtil.applicationInfo;
        fileTransferClient = testUtil.fileTransferClient;
        fileUploadService = new FileUploadService(authService, fileDispatcher(), applicationInfo);
        fileUploadService.setWebSocketClient(fileTransferClient);
    });

    it('should ask about gateway protocol version during start', (done) => {
        const ignoredMessages = 0;
        fileUploadService
            .startService()
            .then(assertSentGetGatewayProtocolVersionRequest)
            .then(done)
            .catch(done);

        performConnect();

        function assertSentGetGatewayProtocolVersionRequest() {
            assertSentRequest(sic.imgwGetProtocolVersionReq(), ignoredMessages);
        }
    });

    it('should send login request with sw_update session type during start', (done) => {
        const ignoredMessages = 1;
        fileUploadService
            .startService()
            .then(assertSentLoginRequest)
            .then(done)
            .catch(done);

        performConnect();

        function assertSentLoginRequest() {
            const loginRequestId = sic.imgwLoginReq().getId();
            const msg = socket.getMessage(ignoredMessages);
            assert.strictEqual(msg.getId(), loginRequestId);
            assert.strictEqual(msg.getSessionType(), SESSION_TYPE_SW_UPDATE);
        }
    });

    it('should send file if requested', (done) => {
        const ignoredMessages = 2;
        const expectedResponse = sic.imgwFileServiceFileContentResp()
            .setRequestId(REQUEST_ID)
            .setContent(EXPECTED_FILE_CONTENT);

        fileUploadService
            .startService()
            .then(sendFileRequest)
            .then(assertSentFileContentResponse)
            .then(done)
            .catch(done);

        performConnect();

        function sendFileRequest() {
            const message = sic.imgwFileServiceGetReq()
                .setRequestId(REQUEST_ID)
                .setUri(encodeURIComponent(URI));
            socket.receive(message);
        }

        function assertSentFileContentResponse() {
            const msg = socket.getMessage(ignoredMessages);
            assert.strictEqual(msg.getId(), expectedResponse.getId());
            assert.strictEqual(msg.getRequestId(), expectedResponse.getRequestId());
            assert.deepEqual(msg.getContent(), expectedResponse.getContent());
            assert.deepEqual(msg.getHash(), expectedResponse.getHash());
        }
    });

    it('should send error response if file not exists', (done) => {
        const ignoredMessages = 2;
        const expectedErrorStatus = {
            type: 'File download error',
            path: INCORRECT_URI,
            error: 'File: rom/config.zip is not present in a given location.'
        };

        const expectedMessage = createErrorMessage();

        fileUploadService
            .startService()
            .then(sendFileRequestWithNotExistingUri)
            .then(assertSentErrorResponse)
            .then(done)
            .catch(done);

        performConnect();

        function createErrorMessage() {
            const message = sic.imgwFileServiceErrorResp()
                .setRequestId(REQUEST_ID)
                .setErrorCode(1)
                .setDescription(expectedErrorStatus.error);
            message.isSicPlus = false;
            return message;
        }

        function sendFileRequestWithNotExistingUri() {
            const message = sic.imgwFileServiceGetReq()
                .setRequestId(REQUEST_ID)
                .setUri(encodeURIComponent(INCORRECT_URI));
            socket.receive(message);
        }

        function assertSentErrorResponse() {
            const msg = socket.getMessage(ignoredMessages);
            assert.strictEqual(msg.getId(), expectedMessage.getId());
            assert.strictEqual(msg.getRequestId(), expectedMessage.getRequestId());
            assert.strictEqual(msg.getErrorCode(), expectedMessage.getErrorCode());
            assert.strictEqual(msg.getDescription(), expectedMessage.getDescription());
        }
    });

    it('should disconnect socket if stopService called', (done) => {
        fileUploadService
            .startService()
            .then(assertSocketConnected)
            .then(fileUploadService.stopService)
            .then(assertSocketDisconnected)
            .then(done)
            .catch(done);

        performConnect();

        function assertSocketConnected() {
            assert.ok(fileTransferClient.isConnected());
        }

        function assertSocketDisconnected() {
            assert.ok(!fileTransferClient.isConnected());
        }

    });

    function assertSentRequest(message, ignoredMessagesCount) {
        const msg = socket.getMessage(ignoredMessagesCount);
        assert.strictEqual(msg.getId(), message.getId());
    }

    function performConnect() {
        testUtil.performConnect();
    }
});
