describe('gatewayMessageFactory', function () {

    const REQUEST_ID = 1;
    const URI = "file:///ram/infoModel";
    const FILE_GET_REQ_ID = 0x6000;
    const FILE_GET_RESP_ID = 0x6012;
    const FILE_ERROR_ID = 0x6011;
    const ERROR_CODE = 4;
    const ERROR_DESCRIPTION = 'unknown problem';

    var messageFactory;

    beforeEach(function () {
        messageFactory = infoModel.gateway.messageFactory;
    });

    describe('protocol messages', function () {
        var messages;

        beforeEach(function () {
            messages = messageFactory.createProtocolMessages();
        });

        describe('imgwGetProtocolVersionReq', function () {
            var message;

            beforeEach(function () {
                message = messages.imgwGetProtocolVersionReq();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), 0x600E);
            });
        });

        describe('imgwGetProtocolVersionResp', function () {
            var message;

            beforeEach(function () {
                message = messages.imgwGetProtocolVersionResp();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), 0x601E);
            });

            it('should set version', function () {
                message.setVersion(666);

                var messageFromBuffer = messages.imgwGetProtocolVersionResp();
                messageFromBuffer.setBuffer(message.getBuffer());

                assert.deepEqual(messageFromBuffer.getVersion(), 666);
            });
        });

    });

    describe('v2', function () {
        var messages;

        beforeEach(function () {
            messages = messageFactory.createMessages(2);
        });

        describe('imgwLoginReq', function () {
            const LOGIN = 'john';
            const AUTH_METHOD_SHA1 = 1;
            const TOKEN = 'abc';
            const SESSION_TYPE_PLAIN = 1;
            const CLIENT_VERSION = 'FL_16_A';
            const CLIENT_ID = '#111111';
            const UPDATE_INTERVAL = 4;

            var message;

            beforeEach(function () {
                message = messages.imgwLoginReq();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), 0x6008);
            });

            it('should set login', function () {
                message.setLogin(LOGIN);

                assert.deepEqual(message.getLogin(), LOGIN);
            });

            it('should set auth method', function () {
                message.setAuthMethod(AUTH_METHOD_SHA1);

                assert.deepEqual(message.getAuthMethod(), AUTH_METHOD_SHA1);
            });

            it('should set token', function () {
                message.setToken(TOKEN);

                assert.deepEqual(message.getToken(), TOKEN);
            });

            it('should set session type', function () {
                message.setSessionType(SESSION_TYPE_PLAIN);

                assert.deepEqual(message.getSessionType(), SESSION_TYPE_PLAIN);
            });

            it('should set client version', function () {
                message.setClientVersion(CLIENT_VERSION);

                assert.deepEqual(message.getClientVersion(), CLIENT_VERSION);
            });

            it('should set client id', function () {
                message.setClientId(CLIENT_ID);

                assert.deepEqual(message.getClientId(), CLIENT_ID);
            });

            it('should set update interval', function () {
                message.setUpdateInterval(UPDATE_INTERVAL);

                assert.deepEqual(message.getUpdateInterval(), UPDATE_INTERVAL);
            });

            it('should correctly decode and encode message', function () {
                message.setLogin(LOGIN)
                    .setAuthMethod(AUTH_METHOD_SHA1)
                    .setToken(TOKEN)
                    .setSessionType(SESSION_TYPE_PLAIN)
                    .setClientVersion(CLIENT_VERSION)
                    .setClientId(CLIENT_ID)
                    .setUpdateInterval(UPDATE_INTERVAL);

                const messageBuffer = message.getBuffer();
                const newMessage = messages.imgwLoginReq();
                newMessage.setBuffer(messageBuffer);

                assert.deepEqual(message.getLogin(), LOGIN);
                assert.deepEqual(newMessage.getAuthMethod(), AUTH_METHOD_SHA1);
                assert.deepEqual(newMessage.getToken(), TOKEN);
                assert.deepEqual(newMessage.getSessionType(), SESSION_TYPE_PLAIN);
                assert.deepEqual(newMessage.getClientVersion(), CLIENT_VERSION);
                assert.deepEqual(newMessage.getClientId(), CLIENT_ID);
                assert.deepEqual(newMessage.getUpdateInterval(), UPDATE_INTERVAL);
            });
        });

        describe('imgwLoginSuccessResp', function () {
            const PROFILE = 'userProfile';
            const USER_NAME = 'userName';
            const FAILED_LOGIN_NO = 4;
            const LAST_SUCCESS = '3 days ago';

            var message;

            beforeEach(function () {
                message = messages.imgwLoginSuccessResp();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), 0x6018);
            });

            it('should set profile', function () {
                message.setProfile(PROFILE);

                assert.deepEqual(message.getProfile(), PROFILE);
            });

            it('should set user name', function () {
                message.setUserName(USER_NAME);

                assert.deepEqual(message.getUserName(), USER_NAME);
            });

            it('should set failed login number', function () {
                message.setFailedLoginNo(FAILED_LOGIN_NO);

                assert.deepEqual(message.getFailedLoginNo(), FAILED_LOGIN_NO);
            });

            it('should set last success info', function () {
                message.setLastSuccess(LAST_SUCCESS);

                assert.deepEqual(message.getLastSuccess(), LAST_SUCCESS);
            });

            it('should correctly decode and encode message', function () {
                message.setProfile(PROFILE)
                    .setUserName(USER_NAME)
                    .setFailedLoginNo(FAILED_LOGIN_NO)
                    .setLastSuccess(LAST_SUCCESS);
                const messageBuffer = message.getBuffer();
                const newMessage = messages.imgwLoginSuccessResp();
                newMessage.setBuffer(messageBuffer);

                assert.deepEqual(message.getProfile(), PROFILE);
                assert.deepEqual(newMessage.getUserName(), USER_NAME);
                assert.deepEqual(newMessage.getFailedLoginNo(), FAILED_LOGIN_NO);
                assert.deepEqual(newMessage.getLastSuccess(), LAST_SUCCESS);
            });
        });

        describe('imgwLoginErrorResp', function () {
            var message;

            beforeEach(function () {
                message = messages.imgwLoginErrorResp();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), 0x6019);
            });
        });

        describe('imgwGetMetaReq', function () {
            var message;

            beforeEach(function () {
                message = messages.imgwGetMetaReq();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), 0x6001);
            });
        });

        describe('imgwGetMetaResp', function () {
            var message;

            beforeEach(function () {
                message = messages.imgwGetMetaResp();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), 0x6013);
            });
            var checkSum = new Uint8Array([226, 74, 87, 139, 57, 254, 23, 210, 23, 209, 106, 42, 247, 120, 4, 39, 133, 81, 173, 187]);
            var fileData = new Uint8Array([12, 9, 103, 4, 1]);

            it('should set meta', function () {
                message.setMeta(fileData);
                assert.deepEqual(new Uint8Array(message.getMeta()), fileData);
                assert.deepEqual(message.getHash(), checkSum);
            });

            it('should throw exception if meta has invalid checkSum', function () {
                var invalidCheckSum = new Uint8Array([226, 74, 87, 139, 57, 254, 23, 210, 23, 209, 106, 42, 247, 120, 4, 39, 133, 81, 173, 184]);
                message
                    .setMeta(fileData)
                    .setHash(invalidCheckSum);
                assert.deepEqual(message.getHash(), invalidCheckSum);
                try {
                    assert.fail(new Uint8Array(message.getMeta()), fileData, 'Exception was not thrown even if meta has incorrect checkSum.');
                } catch (error) {
                    assert.deepEqual(error, "Meta checkSum incorrect. Meta wasn't loaded.");
                }
            });

            it('should correctly decode and encode message', function () {
                message.setMeta(fileData);
                var newMessage = messages.imgwGetMetaResp();
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(new Uint8Array(newMessage.getMeta()), fileData);
                assert.deepEqual(newMessage.getHash(), checkSum);
            });
        });

        describe('imgwFileServiceGetReq', function () {
            var message;

            beforeEach(function () {
                message = messages.imgwFileServiceGetReq();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), FILE_GET_REQ_ID);
            });

            it('should set request id', function () {
                message.setRequestId(REQUEST_ID);

                assert.deepEqual(message.getRequestId(), REQUEST_ID);
            });

            it('should set uri', function () {
                message.setUri(URI);

                assert.deepEqual(message.getUri(), URI);
            });

            it('should correctly decode and encode message', function () {
                message
                    .setRequestId(REQUEST_ID)
                    .setUri(URI);
                var messageBuffer = message.getBuffer();
                var newMessage = messages.imgwFileServiceGetReq();
                newMessage.setBuffer(messageBuffer);

                assert.deepEqual(newMessage.getRequestId(), REQUEST_ID);
                assert.deepEqual(newMessage.getUri(), URI);
            });
        });

        describe('imgwFileServiceFileContentResp', function () {
            var message;
            const FILE_CONTENT = [0, 1, 2, 3, 5, 6];
            const CHECKSUM_CONTENT = [165, 71, 212, 39, 60, 111, 113, 238, 78, 96, 82, 82, 84, 112, 175, 87, 124, 191, 100, 92];

            beforeEach(function () {
                message = messages.imgwFileServiceFileContentResp();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), FILE_GET_RESP_ID);
            });

            it('should set request id', function () {
                message.setRequestId(REQUEST_ID);

                assert.deepEqual(message.getRequestId(), REQUEST_ID);
            });

            it('should set content', function () {
                message.setContent(FILE_CONTENT);
                assert.deepEqual(message.getContent(), new Uint8Array(FILE_CONTENT));
            });

            it('should throw exception if content has invalid checkSum', function () {
                const invalidCheckSum = new Uint8Array([226, 74, 87, 139, 57, 254, 23, 210, 23, 209, 106, 42, 247, 120, 4, 39, 133, 81, 173, 184]);
                message
                    .setContent(FILE_CONTENT)
                    .setHash(invalidCheckSum);
                assert.deepEqual(message.getHash(), invalidCheckSum);
                try {
                    assert.fail(new Uint8Array(message.getContent()), FILE_CONTENT, 'Exception was not thrown even if meta has incorrect checkSum.');
                } catch (error) {
                    assert.deepEqual(error, "Downloaded file checkSum is incorrect.");
                }
            });

            it('should set hash', function () {
                message.setHash(CHECKSUM_CONTENT);
                assert.deepEqual(message.getHash(), new Uint8Array(CHECKSUM_CONTENT));
            });

            it('should correctly decode and encode message', function () {
                message
                    .setRequestId(REQUEST_ID)
                    .setContent(FILE_CONTENT)
                    .setHash(CHECKSUM_CONTENT);

                const newMessage = messages.imgwFileServiceFileContentResp();
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getRequestId(), REQUEST_ID);
                assert.deepEqual(newMessage.getContent(), new Uint8Array(FILE_CONTENT));
                assert.deepEqual(newMessage.getHash(), new Uint8Array(CHECKSUM_CONTENT));
            });
        });

        describe('imgwFileServiceErrorResp', function () {
            var message;

            beforeEach(function () {
                message = messages.imgwFileServiceErrorResp();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), FILE_ERROR_ID);
            });

            it('should set request id', function () {
                message.setRequestId(REQUEST_ID);

                assert.deepEqual(message.getRequestId(), REQUEST_ID);
            });

            it('should set error code', function () {
                message.setErrorCode(ERROR_CODE);

                assert.deepEqual(message.getErrorCode(), ERROR_CODE);
            });

            it('should set description', function () {
                message.setDescription(ERROR_DESCRIPTION);

                assert.deepEqual(message.getDescription(), ERROR_DESCRIPTION);
            });

            it('should correctly decode and encode message', function () {
                message.setRequestId(REQUEST_ID)
                    .setErrorCode(ERROR_CODE)
                    .setDescription(ERROR_DESCRIPTION);
                const messageBuffer = message.getBuffer();
                const newMessage = messages.imgwFileServiceErrorResp();
                newMessage.setBuffer(messageBuffer);

                assert.deepEqual(message.getRequestId(), REQUEST_ID);
                assert.deepEqual(newMessage.getErrorCode(), ERROR_CODE);
                assert.deepEqual(newMessage.getDescription(), ERROR_DESCRIPTION);
            });
        });

        describe('imgwFileServiceErrorResp', function () {
            var message;

            beforeEach(function () {
                message = messages.imgwFileServiceErrorResp();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), FILE_ERROR_ID);
            });

            it('should set request id', function () {
                message.setRequestId(REQUEST_ID);

                assert.deepEqual(message.getRequestId(), REQUEST_ID);
            });

            it('should set error code', function () {
                message.setErrorCode(ERROR_CODE);

                assert.deepEqual(message.getErrorCode(), ERROR_CODE);
            });

            it('should set description', function () {
                message.setDescription(ERROR_DESCRIPTION);

                assert.deepEqual(message.getDescription(), ERROR_DESCRIPTION);
            });

            it('should correctly decode and encode message', function () {
                message.setRequestId(REQUEST_ID)
                    .setErrorCode(ERROR_CODE)
                    .setDescription(ERROR_DESCRIPTION);
                const messageBuffer = message.getBuffer();
                const newMessage = messages.imgwFileServiceErrorResp();
                newMessage.setBuffer(messageBuffer);

                assert.deepEqual(message.getRequestId(), REQUEST_ID);
                assert.deepEqual(newMessage.getErrorCode(), ERROR_CODE);
                assert.deepEqual(newMessage.getDescription(), ERROR_DESCRIPTION);
            });
        });

        describe('imgwGetGatewayInformationResp', function () {
            const ENDIANNESS_BIG = 1;
            const ADMIN_VERSION = 'FL_16_A';
            const ADMIN_ID = '#111111';
            const IMI_VERSION_MAJOR = 9;
            const IMI_VERSION_MINOR = 8;
            const IM_SIC_ADDRESS = 1059;
            const IM_PROTOCOL_RELEASE = 'v15';
            const IM_PROTOCOL_VERSION_MINOR = 37;
            const IM_PROTOCOL_VERSION_MAJOR = 35;

            var message;

            beforeEach(function () {
                message = messages.imgwGetGatewayInformationResp();
            });

            it('should have valid id', function () {
                assert.deepEqual(message.getId(), 0x601F);
            });

            it('should set endianness', function () {
                message.setEndianness(ENDIANNESS_BIG);

                assert.deepEqual(message.getEndianness(), ENDIANNESS_BIG);
            });

            it('should set embedded admin version', function () {
                message.setEmbeddedAdminVersion(ADMIN_VERSION);

                assert.deepEqual(message.getEmbeddedAdminVersion(), ADMIN_VERSION);
            });

            it('should set embedded admin id', function () {
                message.setEmbeddedAdminId(ADMIN_ID);

                assert.deepEqual(message.getEmbeddedAdminId(), ADMIN_ID);
            });

            it('should set IMI major version', function () {
                message.setImiVersionMajor(IMI_VERSION_MAJOR);

                assert.deepEqual(message.getImiVersionMajor(), IMI_VERSION_MAJOR);
            });

            it('should set IMI minor version', function () {
                message.setImiVersionMinor(IMI_VERSION_MINOR);

                assert.deepEqual(message.getImiVersionMinor(), IMI_VERSION_MINOR);
            });

            it('should set IM sic address', function () {
                message.setInfoModelSicAddress(IM_SIC_ADDRESS);

                assert.deepEqual(message.getInfoModelSicAddress(), IM_SIC_ADDRESS);
            });

            it('should set IM protocol release', function () {
                message.setImProtocolRelease(IM_PROTOCOL_RELEASE);

                assert.deepEqual(message.getImProtocolRelease(), IM_PROTOCOL_RELEASE);
            });

            it('should set IM protocol major version', function () {
                message.setImProtocolVersionMajor(IM_PROTOCOL_VERSION_MAJOR);

                assert.deepEqual(message.getImProtocolVersionMajor(), IM_PROTOCOL_VERSION_MAJOR);
            });

            it('should set IM protocol minor version', function () {
                message.setImProtocolVersionMinor(IM_PROTOCOL_VERSION_MINOR);

                assert.deepEqual(message.getImProtocolVersionMinor(), IM_PROTOCOL_VERSION_MINOR);
            });

            it('should correctly decode and encode message', function () {
                message.setEndianness(ENDIANNESS_BIG)
                    .setEmbeddedAdminVersion(ADMIN_VERSION)
                    .setEmbeddedAdminId(ADMIN_ID)
                    .setImiVersionMajor(IMI_VERSION_MAJOR)
                    .setImiVersionMinor(IMI_VERSION_MINOR)
                    .setInfoModelSicAddress(IM_SIC_ADDRESS)
                    .setImProtocolRelease(IM_PROTOCOL_RELEASE)
                    .setImProtocolVersionMajor(IM_PROTOCOL_VERSION_MAJOR)
                    .setImProtocolVersionMinor(IM_PROTOCOL_VERSION_MINOR);

                const messageBuffer = message.getBuffer();
                const newMessage = messages.imgwGetGatewayInformationResp();
                newMessage.setBuffer(messageBuffer);

                assert.deepEqual(message.getEndianness(), ENDIANNESS_BIG);
                assert.deepEqual(newMessage.getEmbeddedAdminVersion(), ADMIN_VERSION);
                assert.deepEqual(newMessage.getEmbeddedAdminId(), ADMIN_ID);
                assert.deepEqual(newMessage.getImiVersionMajor(), IMI_VERSION_MAJOR);
                assert.deepEqual(newMessage.getImiVersionMinor(), IMI_VERSION_MINOR);
                assert.deepEqual(newMessage.getInfoModelSicAddress(), IM_SIC_ADDRESS);
                assert.deepEqual(newMessage.getImProtocolRelease(), IM_PROTOCOL_RELEASE);
                assert.deepEqual(newMessage.getImProtocolVersionMajor(), IM_PROTOCOL_VERSION_MAJOR);
                assert.deepEqual(newMessage.getImProtocolVersionMinor(), IM_PROTOCOL_VERSION_MINOR);
            });
        });
        describe('imgwSessionsListResp', () => {
            let message;
            const SESSIONS = [{
                userLogin: 'login',
                clientVersion: 'ver1'
            }];

            beforeEach(() => {
                message = messages.imgwSessionsListResp();
            });

            it('should have valid id', () => {
                assert.deepEqual(message.getId(), 0x6024);
            });
            it('should set sessions', () => {
                message.setSessions(SESSIONS);

                assert.deepEqual(message.getSessions(), SESSIONS);
            });
        });
        describe('imgwSession', () => {
            let message;
            const USER_LOGIN = 'login';
            const PROFILE_NAME = 'profile';
            const CLIENT_VERSION = 'ver';
            const CONNECTION_TYPE = 1;
            const SESSION_TYPE = 1;
            const AGE = 10;
            const IP = '10.10.10.10';
            const SESSION_ID = 'sessionId';

            beforeEach(() => {
                message = messages.imgwSession();
            });
            it('should set user login', () => {
                message.setUserLogin(USER_LOGIN);

                assert.deepEqual(message.getUserLogin(), USER_LOGIN);
            });
            it('should set profile name', () => {
                message.setProfileName(PROFILE_NAME);

                assert.deepEqual(message.getProfileName(), PROFILE_NAME);
            });
            it('should set client version', () => {
                message.setClientVersion(CLIENT_VERSION);

                assert.deepEqual(message.getClientVersion(), CLIENT_VERSION);
            });
            it('should set connection type', () => {
                message.setConnectionType(CONNECTION_TYPE);

                assert.deepEqual(message.getConnectionType(), CONNECTION_TYPE);
            });
            it('should set session type', () => {
                message.setSessionType(SESSION_TYPE);

                assert.deepEqual(message.getSessionType(), SESSION_TYPE);
            });
            it('should set age', () => {
                message.setAge(AGE);

                assert.deepEqual(message.getAge(), AGE);
            });
            it('should set ip', () => {
                message.setIp(IP);

                assert.deepEqual(message.getIp(), IP);
            });
            it('should set session id', () => {
                message.setSessionId(SESSION_ID);

                assert.deepEqual(message.getSessionId(), SESSION_ID);
            });
            it('should display proper string', () => {
                message.setUserLogin(USER_LOGIN);
                message.setProfileName(PROFILE_NAME);
                message.setClientVersion(CLIENT_VERSION);
                message.setConnectionType(CONNECTION_TYPE);
                message.setSessionType(SESSION_TYPE);
                message.setAge(AGE);
                message.setIp(IP);
                message.setSessionId(SESSION_ID);

                assert.deepEqual(message.toString(), 'imgwSession\n' +
                    '{\n' +
                    ' userLogin: login,\n' +
                    ' profileName: profile,\n' +
                    ' clientVersion: ver,\n' +
                    ' connectionType: 1,\n' +
                    ' sessionType: 1,\n' +
                    ' age: 10,\n' +
                    ' ip: 10.10.10.10,\n' +
                    ' sessionId: sessionId\n' +
                    '}');
            });
        });
    });
});
