describe('historyFactory', function () {

    var history,
        historyFactory = infoModel.historyFactory,
        assertions = infoModel.historyEntryAssertions;

    describe('addEntry', function () {
        beforeEach(function () {
            history = historyFactory.newInstance(1);
        });

        it('should add entry', function () {
            var entry = createChangeEntry(0, '/MRBTS-1');
            history.addEntry(entry);

            assert.isFalse(entry.hasAlarm);
        });

        it('should add entry with alarm', function () {
            var entry = createChangeEntry(0, ['/MRBTS-1', '/MRBTS-1/ALARM_L-1']);
            history.addEntry(entry);

            assert.isTrue(entry.hasAlarm);
        });
    });

    describe('Method snapshotEntryForId', function () {
        beforeEach(function () {
            history = historyFactory.newInstance(1);
        });

        it('should return entry with updates for first index', function () {
            const entry = createChangeEntry(0, ['/MRBTS-1', '/MRBTS-1/RAT-1']);
            history.addEntry(entry);

            const mergedEntry = history.snapshotEntryForId(0);

            assert.equal(mergedEntry.change.updated[0].distName, entry.change.updated[0].distName);
            assert.equal(mergedEntry.change.updated[0].payload, entry.change.updated[0].payload);
            assert.equal(mergedEntry.change.updated[1].distName, entry.change.updated[1].distName);
            assert.equal(mergedEntry.change.updated[1].payload, entry.change.updated[1].payload);
        });

        it('should return entry with updates for given indexes', function () {
            const entry1 = createChangeEntry(0, '/MRBTS-1'),
                entry2 = createChangeEntry(1, '/MRBTS-1/RAT-1');
            history.addEntry(entry1);
            history.addEntry(entry2);

            const mergedEntry1 = history.snapshotEntryForId(0);
            const mergedEntry2 = history.snapshotEntryForId(1);

            assert.equal(mergedEntry1.change.updated[0].distName, entry1.change.updated[0].distName);
            assert.equal(mergedEntry2.change.updated[0].payload, entry1.change.updated[0].payload);
            assert.equal(mergedEntry2.change.updated[1].distName, entry2.change.updated[0].distName);
            assert.equal(mergedEntry2.change.updated[1].payload, entry2.change.updated[0].payload);
        });
    });

    describe('1 step', function () {

        beforeEach(function () {
            history = historyFactory.newInstance(1);
        });

        it('should return null or undefined when getting unknown index', function () {
            var s = history.snapshotForEntryId(0);

            assert.notOk(s);
        });

        it('should return map with updates for first index', function () {
            var entry = {
                id: 0,
                metaUid: '1',
                change: {
                    removed: [],
                    updated: [
                        {distName: '/MRBTS-1', payload: new ArrayBuffer(0)}
                    ]
                }
            };
            history.addEntry(entry);
            assert.equal(history.snapshotsSize(), 1);

            var s = history.snapshotForEntryId(0);

            assert.equal(s.size, 1);
            assert.deepEqual(s.get('/MRBTS-1'), {
                payload: new ArrayBuffer(0),
                metaUid: '1'
            });
        });

        it('should return map with updates for first index', function () {
            var entry = createChangeEntry(0, ['/MRBTS-1', '/MRBTS-1/RAT-1']);
            history.addEntry(entry);
            var s = history.snapshotForEntryId(0);

            assert.equal(s.size, 2);
            assert.isTrue(s.has('/MRBTS-1'));
            assert.isTrue(s.has('/MRBTS-1/RAT-1'));
        });

        it('should return maps with updates for given indexes', function () {
            history.addEntry(createChangeEntry(0, '/MRBTS-1'));
            history.addEntry(createChangeEntry(1, '/MRBTS-1/RAT-1'));
            assert.equal(history.snapshotsSize(), 2);

            var s0 = history.snapshotForEntryId(0);
            var s1 = history.snapshotForEntryId(1);

            assert.equal(s0.size, 1);
            assert.isTrue(s0.has('/MRBTS-1'));

            assert.equal(s1.size, 2);
            assert.isTrue(s1.has('/MRBTS-1'));
            assert.isTrue(s1.has('/MRBTS-1/RAT-1'));
        });

    });

    describe('n step', function () {

        beforeEach(function () {
            history = historyFactory.newInstance(3);

            history.addEntry(createChangeEntry(0, '/A-1'));//0
            history.addEntry(createChangeEntry(1, '/B-1'));
            history.addEntry(createChangeEntry(2, '/C-1'));
            history.addEntry(createChangeEntry(3, '/D-1'));//1
            history.addEntry(createChangeEntry(4, '/E-1'));
            history.addEntry(createChangeEntry(5, '/F-1'));
            history.addEntry(createChangeEntry(6, '/G-1'));//2
        });

        it('should have 3 snapshots - for 0, 3, 6', function () {
            assert.equal(history.snapshotsSize(), 3);
        });

        it('first snapshot ', function () {
            var s = history.snapshotForIndex(0);

            assert.equal(s.size, 1);
            assert.isTrue(s.has('/A-1'));
        });

        it('second snapshot contains entries from first ', function () {
            var s = history.snapshotForIndex(1);

            assert.equal(s.size, 4);
            assert.isTrue(s.has('/A-1'));//first
            assert.isTrue(s.has('/B-1'));//second
            assert.isTrue(s.has('/C-1'));//second
            assert.isTrue(s.has('/D-1'));//second
        });

        it('third snapshot contains entries from first & second', function () {
            var s = history.snapshotForIndex(2);

            assert.equal(s.size, 7);
            assert.isTrue(s.has('/A-1'));//first
            assert.isTrue(s.has('/B-1'));//second
            assert.isTrue(s.has('/C-1'));//second
            assert.isTrue(s.has('/D-1'));//second
            assert.isTrue(s.has('/E-1'));//third
            assert.isTrue(s.has('/F-1'));//third
            assert.isTrue(s.has('/G-1'));//third
        });

        it('should get snapshots for given indexes', function () {
            for (var i = 0; i < history.size(); i++) {
                var s = history.snapshotForEntryId(i);
                assert.equal(s.size, i + 1);
            }
        });
    });

    describe('handle reset', function () {

        beforeEach(function () {
            history = historyFactory.newInstance(2);

            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createResetEntry(1));
            history.addEntry(createChangeEntry(2, '/B-1'));
            history.addEntry(createChangeEntry(3, '/C-1'));
        });

        it('should have valid size', function () {
            assert.equal(history.snapshotsSize(), 2);
        });

        it('should have first entry', function () {
            var s = history.snapshotForIndex(0);

            assert.equal(s.size, 1);
            assert.isTrue(s.has('/A-1'));
        });

        it('should have only second entry', function () {
            var s = history.snapshotForIndex(1);

            assert.equal(s.size, 1);
            assert.isTrue(s.has('/B-1'));
        });

        it('should have valid entries for history indexes', function () {
            var s;

            s = history.snapshotForEntryId(0);
            assert.equal(s.size, 1);
            assert.isTrue(s.has('/A-1'));

            s = history.snapshotForEntryId(1);//reset
            assert.equal(s.size, 0);

            s = history.snapshotForEntryId(2);
            assert.equal(s.size, 1);
            assert.isTrue(s.has('/B-1'));

            s = history.snapshotForEntryId(3);
            assert.equal(s.size, 2);
            assert.isTrue(s.has('/B-1'));
            assert.isTrue(s.has('/C-1'));
        });

    });

    describe('merge first n', function () {

        beforeEach(function () {
            history = historyFactory.newInstance(2);
        });

        it('should merge first entry', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));

            const newEntries = history.mergeFirstEntries(1);

            assert.lengthOf(newEntries, 1);
            const first = newEntries[0];
            assertContainsOnly(first, {updates: ['/A-1']});
        });

        it('should merge into empty array', function () {
            const newEntries = history.mergeFirstEntries(1000);

            assert.lengthOf(newEntries, 0);
        });

        it('should merge first 2 entries', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createChangeEntry(1, '/B-1'));

            const newEntries = history.mergeFirstEntries(2);

            assert.lengthOf(newEntries, 1);
            const first = newEntries[0];
            assertContainsOnly(first, {updates: ['/A-1', '/B-1']});
        });

        it('should merge first n entries', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createChangeEntry(1, '/B-1'));
            history.addEntry(createChangeEntry(2, '/C-1'));
            history.addEntry(createChangeEntry(3, '/D-1'));

            const newEntries = history.mergeFirstEntries(2);

            assert.lengthOf(newEntries, 3);
            const first = newEntries[0];
            assertContainsOnly(first, {updates: ['/A-1', '/B-1']});
            const second = newEntries[1];
            assertContainsOnly(second, {updates: ['/C-1']});
            const third = newEntries[2];
            assertContainsOnly(third, {updates: ['/D-1']});
        });

        it('should merge with alarm', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createChangeEntry(1, '/C-1'));
            history.addEntry(createChangeEntry(2, '/ALARM_L-1'));

            const newEntries = history.mergeFirstEntries(3);

            assert.lengthOf(newEntries, 1);
            const first = newEntries[0];
            assert.isTrue(first.hasAlarm);
        });

        it('merged entry should have last entry timestamp', function () {
            const firstEntry = createChangeEntry(0, '/A-1');
            firstEntry.timestamp = 1;
            history.addEntry(firstEntry);
            const lastEntry = createChangeEntry(1, '/B-1');
            lastEntry.timestamp = 2;
            history.addEntry(lastEntry);

            const newEntries = history.mergeFirstEntries(2);

            assert.lengthOf(newEntries, 1);
            assert.equal(newEntries[0].timestamp, lastEntry.timestamp);
        });

        it('merged entry should have id 0', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createChangeEntry(1, '/B-1'));
            history.addEntry(createChangeEntry(2, '/C-1'));

            const newEntries = history.mergeFirstEntries(3);

            assert.lengthOf(newEntries, 1);
            assert.equal(newEntries[0].id, 0);
        });

        describe('snapshots', function () {

            beforeEach(function () {
                history = historyFactory.newInstance(2);
            });

            it('should get snapshot for merged one entry', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0

                assert.equal(history.snapshotsSize(), 1);
                history.mergeFirstEntries(1);
                assert.equal(history.snapshotsSize(), 1);

                const s = history.snapshotForIndex(0);
                assert.isTrue(s.has('/A-1'));
            });

            it('should get snapshot for merged entry', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0
                history.addEntry(createChangeEntry(1, '/B-1'));

                assert.equal(history.snapshotsSize(), 1);
                history.mergeFirstEntries(2);
                assert.equal(history.snapshotsSize(), 1);

                const s = history.snapshotForIndex(0);
                assert.isTrue(s.has('/A-1'));
                assert.isTrue(s.has('/B-1'));
            });

            it('should leave only one snapshot', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0
                history.addEntry(createChangeEntry(1, '/B-1'));
                history.addEntry(createChangeEntry(2, '/C-1'));//snapshot 1
                history.addEntry(createChangeEntry(3, '/D-1'));
                history.addEntry(createChangeEntry(4, '/E-1'));//snapshot 2

                assert.equal(history.snapshotsSize(), 3);
                history.mergeFirstEntries(5);
                assert.equal(history.snapshotsSize(), 1);

                const s = history.snapshotForIndex(0);
                assert.isTrue(s.has('/A-1'));
                assert.isTrue(s.has('/B-1'));
                assert.isTrue(s.has('/C-1'));
                assert.isTrue(s.has('/D-1'));
                assert.isTrue(s.has('/E-1'));
            });

            it('should get snapshot for merged entry - more snapshots', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0
                history.addEntry(createChangeEntry(1, '/B-1'));
                history.addEntry(createChangeEntry(2, '/C-1'));//snapshot 1
                history.addEntry(createChangeEntry(3, '/D-1'));
                history.addEntry(createChangeEntry(4, '/E-1'));//snapshot 2
                history.addEntry(createChangeEntry(5, '/F-1'));

                assert.equal(history.snapshotsSize(), 3);
                history.mergeFirstEntries(2);
                assert.equal(history.snapshotsSize(), 3);

                const s = history.snapshotForIndex(0);
                assert.isTrue(s.has('/A-1'));
                assert.isTrue(s.has('/B-1'));
            });

            it('should remove unused snapshots', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0
                history.addEntry(createChangeEntry(1, '/B-1'));
                history.addEntry(createChangeEntry(2, '/C-1'));//snapshot 1
                history.addEntry(createChangeEntry(3, '/D-1'));
                history.addEntry(createChangeEntry(4, '/E-1'));//snapshot 2
                history.addEntry(createChangeEntry(5, '/F-1'));

                assert.equal(history.snapshotsSize(), 3);
                history.mergeFirstEntries(4);
                assert.equal(history.snapshotsSize(), 2);

                const s0 = history.snapshotForIndex(0);
                assert.isTrue(s0.has('/A-1'));
                assert.isTrue(s0.has('/B-1'));
                assert.isTrue(s0.has('/C-1'));
                assert.isTrue(s0.has('/D-1'));

                const s1 = history.snapshotForIndex(1);
                assert.isTrue(s1.has('/E-1'));
            });

            it('should get correct snapshots for each entryId', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0
                history.addEntry(createChangeEntry(1, '/B-1'));
                history.addEntry(createChangeEntry(2, '/C-1'));//snapshot 1
                history.addEntry(createChangeEntry(3, '/D-1'));
                history.addEntry(createChangeEntry(4, '/E-1'));//snapshot 2
                history.addEntry(createChangeEntry(5, '/F-1'));

                assert.equal(history.snapshotsSize(), 3);
                history.mergeFirstEntries(2);
                assert.equal(history.snapshotsSize(), 3);
                assert.equal(history.size(), 5);

                const s0 = history.snapshotForEntryId(0);
                assert.isTrue(s0.has('/A-1'));
                assert.isTrue(s0.has('/B-1'));
                const s1 = history.snapshotForEntryId(1);
                assert.isTrue(s1.has('/C-1'));
                const s2 = history.snapshotForEntryId(2);
                assert.isTrue(s2.has('/D-1'));
                const s3 = history.snapshotForEntryId(3);
                assert.isTrue(s3.has('/E-1'));
                const s4 = history.snapshotForEntryId(4);
                assert.isTrue(s4.has('/F-1'));
            });

        });

    });

    describe('merge range', function () {

        beforeEach(function () {
            history = historyFactory.newInstance(2);
        });

        it('should merge first entry', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));

            const newEntries = history.mergeEntriesRange(0, 1);

            assert.lengthOf(newEntries, 1);
            const first = newEntries[0];
            assertContainsOnly(first, {updates: ['/A-1']});
        });

        it('should merge into empty array', function () {
            const newEntries = history.mergeEntriesRange(0, 1000);

            assert.lengthOf(newEntries, 0);
        });

        it('should merge only first 2 entries', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createChangeEntry(1, '/B-1'));
            history.addEntry(createChangeEntry(2, '/C-1'));
            history.addEntry(createChangeEntry(3, '/D-1'));

            const newEntries = history.mergeEntriesRange(0, 2);

            assert.lengthOf(newEntries, 3);
            const first = newEntries[0];
            assertContainsOnly(first, {updates: ['/A-1', '/B-1']});
            const second = newEntries[1];
            assertContainsOnly(second, {updates: ['/C-1']});
            const third = newEntries[2];
            assertContainsOnly(third, {updates: ['/D-1']});
        });

        it('should merge only last 2 entries', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createChangeEntry(1, '/B-1'));
            history.addEntry(createChangeEntry(2, '/C-1'));
            history.addEntry(createChangeEntry(3, '/D-1'));

            const newEntries = history.mergeEntriesRange(2, 4);

            assert.lengthOf(newEntries, 3);
            const first = newEntries[0];
            assertContainsOnly(first, {updates: ['/A-1']});
            const second = newEntries[1];
            assertContainsOnly(second, {updates: ['/B-1']});
            const third = newEntries[2];
            assertContainsOnly(third, {updates: ['/C-1', '/D-1']});
        });

        it('should merge only middle 2 entries', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createChangeEntry(1, '/B-1'));
            history.addEntry(createChangeEntry(2, '/C-1'));
            history.addEntry(createChangeEntry(3, '/D-1'));

            const newEntries = history.mergeEntriesRange(1, 3);

            assert.lengthOf(newEntries, 3);
            const first = newEntries[0];
            assertContainsOnly(first, {updates: ['/A-1']});
            const second = newEntries[1];
            assertContainsOnly(second, {updates: ['/B-1', '/C-1']});
            const third = newEntries[2];
            assertContainsOnly(third, {updates: ['/D-1']});
        });

        it('should merge with alarm', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createChangeEntry(1, '/C-1'));
            history.addEntry(createChangeEntry(2, '/ALARM_L-1'));

            const newEntries = history.mergeEntriesRange(0, 3);

            assert.lengthOf(newEntries, 1);
            const first = newEntries[0];
            assert.isTrue(first.hasAlarm);
        });

        it('merged entry should have last entry timestamp', function () {
            const firstEntry = createChangeEntry(0, '/A-1');
            const lastEntry = createChangeEntry(1, '/B-1');
            firstEntry.timestamp = 1;
            lastEntry.timestamp = 2;
            history.addEntry(firstEntry);
            history.addEntry(lastEntry);

            const newEntries = history.mergeEntriesRange(0, 2);

            assert.lengthOf(newEntries, 1);
            assert.equal(newEntries[0].timestamp, lastEntry.timestamp);
        });

        it('entries after merge should have sequential ids starting from 0', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createChangeEntry(1, '/B-1'));
            history.addEntry(createChangeEntry(2, '/C-1'));
            history.addEntry(createChangeEntry(3, '/D-1'));

            const newEntries = history.mergeEntriesRange(1, 3);

            assert.lengthOf(newEntries, 3);
            assertContainsOnly(newEntries[0], {updates: ['/A-1']});
            assert.deepEqual(newEntries[0].id, 0);
            assertContainsOnly(newEntries[1], {updates: ['/B-1', '/C-1']});
            assert.deepEqual(newEntries[1].id, 1);
            assertContainsOnly(newEntries[2], {updates: ['/D-1']});
            assert.deepEqual(newEntries[2].id, 2);
        });

        describe('snapshots', function () {

            beforeEach(function () {
                history = historyFactory.newInstance(2);
            });

            it('should get snapshot for merged one entry', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0

                assert.equal(history.snapshotsSize(), 1);
                history.mergeEntriesRange(0, 1);
                assert.equal(history.snapshotsSize(), 1);

                const s = history.snapshotForIndex(0);
                assert.isTrue(s.has('/A-1'));
            });

            it('should get snapshot for merged entry', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0
                history.addEntry(createChangeEntry(1, '/B-1'));

                assert.equal(history.snapshotsSize(), 1);
                history.mergeEntriesRange(0, 2);
                assert.equal(history.snapshotsSize(), 1);

                const s = history.snapshotForIndex(0);
                assert.isTrue(s.has('/A-1'));
                assert.isTrue(s.has('/B-1'));
            });

            it('should leave only one snapshot', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0
                history.addEntry(createChangeEntry(1, '/B-1'));
                history.addEntry(createChangeEntry(2, '/C-1'));//snapshot 1
                history.addEntry(createChangeEntry(3, '/D-1'));
                history.addEntry(createChangeEntry(4, '/E-1'));//snapshot 2

                assert.equal(history.snapshotsSize(), 3);
                history.mergeEntriesRange(0, 5);
                assert.equal(history.snapshotsSize(), 1);

                const s = history.snapshotForIndex(0);
                assert.isTrue(s.has('/A-1'));
                assert.isTrue(s.has('/B-1'));
                assert.isTrue(s.has('/C-1'));
                assert.isTrue(s.has('/D-1'));
                assert.isTrue(s.has('/E-1'));
            });

            it('should get snapshot for merged entry - more snapshots', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0
                history.addEntry(createChangeEntry(1, '/B-1'));
                history.addEntry(createChangeEntry(2, '/C-1'));//snapshot 1
                history.addEntry(createChangeEntry(3, '/D-1'));
                history.addEntry(createChangeEntry(4, '/E-1'));//snapshot 2
                history.addEntry(createChangeEntry(5, '/F-1'));

                assert.equal(history.snapshotsSize(), 3);
                history.mergeEntriesRange(0, 2);
                assert.equal(history.snapshotsSize(), 3);

                const s = history.snapshotForIndex(0);
                assert.isTrue(s.has('/A-1'));
                assert.isTrue(s.has('/B-1'));
            });

            it('should remove unused snapshots', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0
                history.addEntry(createChangeEntry(1, '/B-1'));
                history.addEntry(createChangeEntry(2, '/C-1'));//snapshot 1
                history.addEntry(createChangeEntry(3, '/D-1'));
                history.addEntry(createChangeEntry(4, '/E-1'));//snapshot 2
                history.addEntry(createChangeEntry(5, '/F-1'));

                assert.equal(history.snapshotsSize(), 3);
                history.mergeEntriesRange(0, 4);
                assert.equal(history.snapshotsSize(), 2);

                const s0 = history.snapshotForIndex(0);
                assert.isTrue(s0.has('/A-1'));
                assert.isTrue(s0.has('/B-1'));
                assert.isTrue(s0.has('/C-1'));
                assert.isTrue(s0.has('/D-1'));

                const s1 = history.snapshotForIndex(1);
                assert.isTrue(s1.has('/E-1'));
            });

            it('should get correct snapshots for each entryId', function () {
                history.addEntry(createChangeEntry(0, '/A-1'));//snapshot 0
                history.addEntry(createChangeEntry(1, '/B-1'));
                history.addEntry(createChangeEntry(2, '/C-1'));//snapshot 1
                history.addEntry(createChangeEntry(3, '/D-1'));
                history.addEntry(createChangeEntry(4, '/E-1'));//snapshot 2
                history.addEntry(createChangeEntry(5, '/F-1'));

                assert.equal(history.snapshotsSize(), 3);
                history.mergeEntriesRange(2, 5);
                assert.equal(history.snapshotsSize(), 2);
                assert.equal(history.size(), 4);

                const s0 = history.snapshotForEntryId(0);
                assert.isTrue(s0.has('/A-1'));
                const s1 = history.snapshotForEntryId(1);
                assert.isTrue(s1.has('/B-1'));
                const s2 = history.snapshotForEntryId(2);
                assert.isTrue(s2.has('/C-1'));
                assert.isTrue(s2.has('/D-1'));
                assert.isTrue(s2.has('/E-1'));
                const s3 = history.snapshotForEntryId(3);
                assert.isTrue(s3.has('/F-1'));
            });

        });

    });

    describe('merge history', function () {

        beforeEach(function () {
            history = historyFactory.newInstance(2);
        });

        it('should merge one entry', function () {
            history.addEntry({
                    metaUid: 'META_1',
                    change: {
                        updated: [
                            createEmptyUpdate('/A-1'),
                            createEmptyUpdate('/B-1'),
                            createEmptyUpdate('/C-1')
                        ],
                        removed: []
                    }
                }
            );

            history.mergeEntries();

            assertHistoryHasOneEntry(history);
            var entry = history.getLastEntry();
            assertContainsOnly(entry, {updates: ['/A-1', '/B-1', '/C-1'], removes: []});
        });

        it('should merge updates', function () {
            history.addEntry({
                    metaUid: 'META_1',
                    change: {
                        updated: [createEmptyUpdate('/A-1')],
                        removed: []
                    }
                }
            );
            history.addEntry({
                    metaUid: 'META_1',
                    change: {
                        updated: [createEmptyUpdate('/A-1')],
                        removed: []
                    }
                }
            );

            history.mergeEntries();

            assertHistoryHasOneEntry(history);
            var entry = history.getLastEntry();
            assertContainsOnly(entry, {updates: ['/A-1'], removes: []});
        });

        it('should update and remove cancel each other', function () {
            history.addEntry({
                    metaUid: 'META_1',
                    change: {
                        updated: [createEmptyUpdate('/A-1')],
                        removed: []
                    }
                }
            );
            history.addEntry({
                    metaUid: 'META_1',
                    change: {
                        updated: [],
                        removed: ['/A-1']
                    }
                }
            );

            history.mergeEntries();

            assertHistoryHasOneEntry(history);
            assertIsEmptyChangeEntry(history.getLastEntry());
        });

        it('should merge updates only from last reset', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createResetEntry(1));
            history.addEntry(createChangeEntry(2, '/B-1'));

            history.mergeEntries();

            assertHistoryHasOneEntry(history);
            var entry = history.getLastEntry();
            assertContainsOnly(entry, {updates: ['/B-1'], removes: []});
        });

        it('should have metaUid', function () {
            history.addEntry({
                    metaUid: 'META_1',
                    change: {
                        updated: [createEmptyUpdate('/A-1')],
                        removed: []
                    }
                }
            );
            history.addEntry({
                    metaUid: 'META_1',
                    change: {
                        updated: [createEmptyUpdate('/A-1')],
                        removed: []
                    }
                }
            );

            history.mergeEntries();

            assertHistoryHasOneEntry(history);
            var entry = history.getLastEntry();
            assert.equal(entry.metaUid, 'META_1');
        });

        it('merged entry should have id 0', function () {
            history.addEntry(createChangeEntry(0, '/A-1'));
            history.addEntry(createChangeEntry(1, '/B-1'));
            history.addEntry(createChangeEntry(2, '/C-1'));

            history.mergeEntries();

            assertHistoryHasOneEntry(history);
            assert.deepEqual(history.getLastEntry().id, 0);
        });

    });

    function createChangeEntry(id, distNames) {
        return assertions.createChangeEntry(id, distNames);
    }

    function createEmptyUpdate(distName) {
        return assertions.createEmptyUpdate(distName);
    }

    function createResetEntry(id) {
        return assertions.createResetEntry(id);
    }

    function assertIsEmptyChangeEntry(entry) {
        assertions.assertIsEmptyChangeEntry(entry);
    }

    function assertContainsUpdate(entry, distName) {
        assertions.assertContainsUpdate(entry, distName);
    }

    function assertContainsOnly(entry, expected) {
        assertions.assertContainsOnly(entry, expected);
    }

    function assertHistoryHasOneEntry(history) {
        assert.equal(history.snapshotsSize(), 1);
        assert.equal(history.size(), 1);
    }

});
