describe('Service: imTracker', function () {

    const EVENTS = require('../../src/imTracker/imTrackerEvents');

    var DEFAULT_PARAMETERS = {'param-1': 'value', 'param-2': {'param-2-1': 'value-2', 'param-2-2': 'value-2'}},
        NEW_PARAMETERS = {'param-1': 'newValue', 'param-2': {'param-2-1': 'newValue-2', 'param-2-2': 'newValue-2'}};

    var imTracker = {},
        onlineImTree = {};

    var decoderServiceStub;
    beforeEach(function () {
        var metaServiceStub = {
            getLastMetaUid: function () {
                return 0;
            }
        };

        decoderServiceStub = (function () {
            var firstUpdate = true;
            return {
                setToFirstUpdate: function () {
                    firstUpdate = true;
                },
                setToSecondUpdate: function () {
                    firstUpdate = false;
                },
                decodeByDistName: function () {
                    if (firstUpdate)
                        return DEFAULT_PARAMETERS;
                    return NEW_PARAMETERS;
                },
                decodeByClassName: function () {
                    if (firstUpdate)
                        return DEFAULT_PARAMETERS;
                    return NEW_PARAMETERS;
                }
            };
        })();

        onlineImTree = adminCore.onlineImTree({
            metaService: metaServiceStub,
            decoderService: decoderServiceStub
        });

        imTracker = adminCore.imTracker({
            onlineImTree: onlineImTree
        });
    });

    describe('listeners testing', function () {
        var objectListener, anotherObjectListener;

        beforeEach(function () {
            objectListener = {
                removeEvent: function (event) {
                    delete objectListener[event];
                },
                useWith: function (event) {
                    objectListener[event] = sinon.spy();
                    return objectListener;
                }
            };

            anotherObjectListener = {
                removeEvent: function (event) {
                    delete objectListener[event];
                },
                useWith: function (event) {
                    anotherObjectListener[event] = sinon.spy();
                    return objectListener;
                }
            };
        });

        describe('tracking by distName', function () {
            var DISTNAME = '/MRBTS-1/RAT-1',
                CHILDREN_DISTNAMES = {CHILD_1: '/MRBTS-1/RAT-1/CHILD-1', CHILD_2: '/MRBTS-1/RAT-1/CHILD-2'};

            describe('onObjectAppeared testing', function () {

                beforeEach(function () {
                    objectListener.useWith('onObjectAppeared');
                    anotherObjectListener.useWith('onObjectAppeared');
                });

                it('should not call onObjectAppeared on listener if object is not added to tracking', function () {
                    pushFirstUpdate(DISTNAME);

                    assert(objectListener.onObjectAppeared.notCalled);
                });

                it('should call onObjectAppeared on listener if object is added to tracking and appeared first time', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        var event = {eventType: EVENTS.OBJECT_APPEARED};

                        assert(objectListener.onObjectAppeared.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectAppeared.calledOnce);
                        done();
                    });
                });

                it('should call onObjectAppeared on listener if object is added to tracking and just has been in infomodel', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushFirstUpdate(DISTNAME);
                        var event = {eventType: EVENTS.OBJECT_APPEARED};

                        assert(objectListener.onObjectAppeared.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectAppeared.calledOnce);
                        done();
                    });
                });


                it('should call onObjectAppeared on listener for two another objects in one pushFirstUpdate', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        return imTracker.trackByDistName(CHILDREN_DISTNAMES.CHILD_1, anotherObjectListener);
                    }).then(function () {
                        pushFirstUpdate(DISTNAME, CHILDREN_DISTNAMES.CHILD_1);

                        var event = {eventType: EVENTS.OBJECT_APPEARED};

                        assert(objectListener.onObjectAppeared.calledOnce);
                        assert(objectListener.onObjectAppeared.calledWith(sinon.match(event)));

                        assert(anotherObjectListener.onObjectAppeared.calledOnce);
                        assert(anotherObjectListener.onObjectAppeared.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should call onObjectAppeared on listener for two another objects in the following updates', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        return imTracker.trackByDistName(CHILDREN_DISTNAMES.CHILD_1, anotherObjectListener);
                    }).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushFirstUpdate(CHILDREN_DISTNAMES.CHILD_1);

                        var event = {eventType: EVENTS.OBJECT_APPEARED};

                        assert(objectListener.onObjectAppeared.calledOnce);
                        assert(objectListener.onObjectAppeared.calledWith(sinon.match(event)));

                        assert(anotherObjectListener.onObjectAppeared.calledOnce);
                        assert(anotherObjectListener.onObjectAppeared.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should not call onObjectAppeared on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onObjectAppeared');
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);

                        assert.isUndefined(objectListener.onObjectAppeared);
                        done();
                    });
                });

                it('should not call onObjectAppeared on listener if object was in Infomodel', function (done) {
                    pushFirstUpdate(DISTNAME);
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);

                        assert(objectListener.onObjectAppeared.notCalled);
                        done();
                    });
                });

                it("should call onObjectAppeared on listener if object appears in Infomodel after reset", function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushReset();
                        pushFirstUpdate(DISTNAME);

                        var event = {eventType: EVENTS.OBJECT_APPEARED};

                        assert(objectListener.onObjectAppeared.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectAppeared.calledTwice);
                        done();
                    });
                })
            });

            describe('onListenerRegistered testing', function () {

                beforeEach(function () {
                    objectListener.useWith('onListenerRegistered');
                    anotherObjectListener.useWith('onListenerRegistered');
                });

                it('should not call onListenerRegistered on listener if object was not in infomodel and is not added to tracking', function () {
                    assert(objectListener.onListenerRegistered.notCalled);
                });

                it('should not call onListenerRegistered on listener if object was not in infomodel and is added to tracking', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);

                        assert(objectListener.onListenerRegistered.notCalled);
                        done();
                    });
                });

                it('should not call onListenerRegistered on listener if object was in infomodel and is not added to tracking', function () {
                    pushFirstUpdate(DISTNAME);

                    assert(objectListener.onListenerRegistered.notCalled);
                });

                it('should call onListenerRegistered on listener if object was in infomodel and is added to tracking', function (done) {
                    pushFirstUpdate(DISTNAME);
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        var event = {
                            eventType: 'onListenerRegistered',
                            parameters: DEFAULT_PARAMETERS
                        };

                        assert(objectListener.onListenerRegistered.calledOnce);
                        assert(objectListener.onListenerRegistered.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should not call onListenerRegistered on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onListenerRegistered');
                    pushFirstUpdate(DISTNAME);
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        assert.isUndefined(objectListener.onListenerRegistered);
                        done();
                    });
                });
            });

            describe('onParameterChanged testing', function () {

                beforeEach(function () {
                    objectListener.useWith('onParameterChanged');
                    anotherObjectListener.useWith('onParameterChanged');
                });

                it('should call onParameterChanged on listener if object with tracked parameters is appeared', function (done) {
                    var parametersToTracking = ['param-1'];
                    imTracker.trackByDistName(DISTNAME, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAME);
                        var event = {
                            eventType: EVENTS.PARAMETER_CHANGED,
                            parameters: DEFAULT_PARAMETERS
                        };

                        assert(objectListener.onParameterChanged.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should call onParameterChanged on listener if tracked parameters were changed', function (done) {
                    var parametersToTracking = ['param-1'];
                    imTracker.trackByDistName(DISTNAME, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushSecondUpdate(DISTNAME);
                        var event = {
                            eventType: EVENTS.PARAMETER_CHANGED,
                            parameters: NEW_PARAMETERS
                        };

                        assert(objectListener.onParameterChanged.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should call onParameterChanged on listener after reset', function (done) {
                    var parametersToTracking = ['param-1'];

                    imTracker.trackByDistName(DISTNAME, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushReset();
                        pushFirstUpdate(DISTNAME);
                        var event = {
                            eventType: EVENTS.PARAMETER_CHANGED,
                            parameters: DEFAULT_PARAMETERS
                        };

                        assert(objectListener.onParameterChanged.calledTwice);
                        assert(objectListener.onParameterChanged.firstCall.calledWith(sinon.match(event)));
                        assert(objectListener.onParameterChanged.secondCall.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should call onParameterChanged on listener if tracked parameters were changed with complex parameter', function (done) {
                    var parametersToTracking = ['param-2.param-2-1'];

                    imTracker.trackByDistName(DISTNAME, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushSecondUpdate(DISTNAME);
                        var event = {
                            eventType: EVENTS.PARAMETER_CHANGED,
                            parameters: NEW_PARAMETERS
                        };

                        assert(objectListener.onParameterChanged.secondCall.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should not call onParameterChanged on listener if tracked parameters were not changed', function (done) {
                    var parametersToTracking = ['param-1'];
                    imTracker.trackByDistName(DISTNAME, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushFirstUpdate(DISTNAME);

                        assert(objectListener.onParameterChanged.calledOnce);
                        done();
                    });
                });

                it('should not call onParameterChanged on listener if tracked parameters do not exist', function (done) {
                    var parametersToTracking = ['param-3'];

                    imTracker.trackByDistName(DISTNAME, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAME);

                        assert(objectListener.onParameterChanged.notCalled);
                        done();
                    });
                });

                it('should not call onParameterChanged on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onParameterChanged');
                    var parametersToTracking = ['param-1'];
                    pushFirstUpdate(DISTNAME);
                    imTracker.trackByDistName(DISTNAME, objectListener, parametersToTracking).then(function () {
                        pushSecondUpdate(DISTNAME);

                        assert.isUndefined(objectListener.onParameterChanged);
                        done();
                    });
                });
            });

            describe('onObjectRemoved testing', function () {

                beforeEach(function () {
                    objectListener.useWith('onObjectRemoved');
                    anotherObjectListener.useWith('onObjectRemoved');
                });

                it('should call onObjectRemoved on listener if tracked object was removed', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushRemove(DISTNAME);
                        var event = {eventType: EVENTS.OBJECT_REMOVED};

                        assert(objectListener.onObjectRemoved.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectRemoved.calledOnce);
                        done();
                    });
                });

                it('should call onObjectRemoved on listener only once if tracked object was removed', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushRemove(DISTNAME);
                        pushRemove(DISTNAME);
                        var event = {eventType: EVENTS.OBJECT_REMOVED};

                        assert(objectListener.onObjectRemoved.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectRemoved.calledOnce);
                        done();
                    });
                });

                it('should not call onObjectRemoved on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onObjectRemoved');

                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushRemove(DISTNAME);

                        assert.isUndefined(objectListener.onObjectRemoved);
                        done();
                    });
                });

                it('should call onObjectRemoved on listener after reset', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushReset();
                        var event = {eventType: EVENTS.OBJECT_REMOVED};

                        assert(objectListener.onObjectRemoved.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectRemoved.calledOnce);
                        done();
                    });

                });

                it('should not call onObjectRemoved on listener if object does not exist', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(CHILDREN_DISTNAMES.CHILD_1);
                        pushFirstUpdate(CHILDREN_DISTNAMES.CHILD_2);
                        assert(objectListener.onObjectRemoved.notCalled);
                        done();
                    });

                });

            });

            describe('onParametersChanged', function () {

                beforeEach(function () {
                    objectListener.useWith('onParametersChanged');
                    anotherObjectListener.useWith('onParametersChanged');
                });

                it('should call onParametersChanged on listener if tracked object was changed', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        var event = {
                            eventType: EVENTS.PARAMETERS_CHANGED,
                            parameters: DEFAULT_PARAMETERS
                        };

                        assert(objectListener.onParametersChanged.calledOnce);
                        assert(objectListener.onParametersChanged.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should call two times onParametersChanged on listener if tracked object was updated two times ', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushFirstUpdate(DISTNAME);

                        assert(objectListener.onParametersChanged.calledTwice);
                        done();
                    });
                });

                it('should not call onParametersChanged on listener if tracked object was not updated ', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushFirstUpdate(CHILDREN_DISTNAMES.CHILD_1);

                        assert(objectListener.onParametersChanged.calledOnce);
                        done();
                    });
                });

                it('should not call onParametersChanged on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onParametersChanged');

                    pushFirstUpdate(DISTNAME);
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushSecondUpdate(DISTNAME);

                        assert.isUndefined(objectListener.onParametersChanged);
                        done()
                    });
                });

                it('should call onParametersChanged on listener after reset', function (done) {
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(DISTNAME);
                        pushReset();
                        pushFirstUpdate(DISTNAME);
                        var event = {
                            parameters: DEFAULT_PARAMETERS,
                            eventType: EVENTS.PARAMETERS_CHANGED
                        };

                        assert(objectListener.onParametersChanged.calledTwice);
                        assert(objectListener.onParametersChanged.calledWith(sinon.match(event)));
                        done();
                    });
                })

            });

            describe('onChildAdded testing', function () {

                beforeEach(function () {
                    objectListener.useWith('onChildAdded');
                    anotherObjectListener.useWith('onChildAdded');
                });

                it('should call onChildAdded', function (done) {
                    pushFirstUpdate(DISTNAME);
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(CHILDREN_DISTNAMES.CHILD_1);
                        pushSecondUpdate(CHILDREN_DISTNAMES.CHILD_1);
                        var event = {
                            eventType: EVENTS.CHILD_ADDED,
                            distNames: [CHILDREN_DISTNAMES.CHILD_1]
                        };

                        assert(objectListener.onChildAdded.calledOnce);
                        assert(objectListener.onChildAdded.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should call onChildAdded two times', function (done) {
                    pushFirstUpdate(DISTNAME);
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(CHILDREN_DISTNAMES.CHILD_1);
                        pushSecondUpdate(CHILDREN_DISTNAMES.CHILD_2);

                        assert(objectListener.onChildAdded.calledTwice);

                        done();
                    });
                });

                it('should call onChildAdded two times if objects appeared in single pushFirstUpdate', function (done) {
                    pushFirstUpdate(DISTNAME);
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushFirstUpdate(CHILDREN_DISTNAMES.CHILD_1, CHILDREN_DISTNAMES.CHILD_2);
                        var event = {
                            eventType: EVENTS.CHILD_ADDED,
                            distNames: [CHILDREN_DISTNAMES.CHILD_1, CHILDREN_DISTNAMES.CHILD_2]
                        };

                        assert(objectListener.onChildAdded.calledOnce);
                        assert(objectListener.onChildAdded.calledWith(sinon.match(event)));

                        done();
                    });
                });

                it('should not call onChildAdded', function (done) {
                    pushFirstUpdate(DISTNAME);
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushSecondUpdate(DISTNAME);

                        assert(objectListener.onChildAdded.notCalled);
                        done();
                    });
                });

                it('should not call onChildAdded on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onChildAdded');

                    pushFirstUpdate(DISTNAME);
                    imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                        pushSecondUpdate(CHILDREN_DISTNAMES.CHILD_1);

                        assert.isUndefined(objectListener.onChildAdded);
                        done();
                    });
                });
            });

            it('should stop tracking', function (done) {
                objectListener.useWith('onParametersChanged');
                anotherObjectListener.useWith('onParametersChanged');

                pushFirstUpdate(DISTNAME);
                imTracker.trackByDistName(DISTNAME, objectListener).then(addAnotherObjectListener).then(function () {
                    imTracker.stopTracking(objectListener, DISTNAME);
                    imTracker.stopTracking(anotherObjectListener, DISTNAME);
                    pushSecondUpdate(DISTNAME);
                    assert(objectListener.onParametersChanged.notCalled);
                    assert(anotherObjectListener.onParametersChanged.notCalled);
                }).done(done);

                function addAnotherObjectListener() {
                    return imTracker.trackByDistName(DISTNAME, anotherObjectListener);
                }
            });

            it('should not stop tracking if distName is not specified', function (done) {
                objectListener.useWith('onParametersChanged');

                pushFirstUpdate(DISTNAME);
                imTracker.trackByDistName(DISTNAME, objectListener).then(function () {
                    imTracker.stopTracking(objectListener);
                    pushSecondUpdate(DISTNAME);
                    assert(objectListener.onParametersChanged.calledOnce);
                }).done(done);

            });

        });

        describe('tracking by className', function () {

            var CLASSNAMES = {
                CHILD: "CHILD",
                GRANDCHILD: "GRANDCHILD"
            };

            var DISTNAMES = {
                CHILD_1: '/ROOT-1/DISTNAME-1/CHILD-1',
                CHILD_2: '/ROOT-1/DISTNAME-1/CHILD-2',
                GRANDCHILD_1: '/ROOT-1/DISTNAME-1/CHILD-1/GRANDCHILD-1',
                GRANDCHILD_2: '/ROOT-1/DISTNAME-1/CHILD-1/GRANDCHILD-2'
            };

            describe('onObjectAppeared testing', function () {

                beforeEach(function () {
                    objectListener.useWith('onObjectAppeared');
                    anotherObjectListener.useWith('onObjectAppeared');
                });

                it('should call onObjectAppeared on listener if object is added to tracking and appeared first time', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        var event = {
                            eventType: EVENTS.OBJECT_APPEARED,
                            distNames: [DISTNAMES.CHILD_1, DISTNAMES.CHILD_2]
                        };

                        assert(objectListener.onObjectAppeared.calledOnce);
                        assert(objectListener.onObjectAppeared.calledWith(sinon.match(event)));
                        done();
                    });
                });


                it('should call onObjectAppeared on listener for two another objects in one pushFirstUpdate', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        return imTracker.trackByClassName(CLASSNAMES.GRANDCHILD, anotherObjectListener);
                    }).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.GRANDCHILD_1);

                        var event = {
                                eventType: EVENTS.OBJECT_APPEARED,
                                distNames: [DISTNAMES.CHILD_1]
                            },
                            anotherEvent = {
                                eventType: EVENTS.OBJECT_APPEARED,
                                distNames: [DISTNAMES.GRANDCHILD_1]
                            };

                        assert(objectListener.onObjectAppeared.calledOnce);
                        assert(objectListener.onObjectAppeared.calledWith(sinon.match(event)));

                        assert(anotherObjectListener.onObjectAppeared.calledOnce);
                        assert(anotherObjectListener.onObjectAppeared.calledWith(sinon.match(anotherEvent)));
                        done();
                    });
                });

                it('should call onObjectAppeared on listener for two another objects in the following updates', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        return imTracker.trackByClassName(CLASSNAMES.GRANDCHILD, anotherObjectListener);
                    }).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1);
                        pushFirstUpdate(DISTNAMES.GRANDCHILD_1);

                        var event = {
                                eventType: EVENTS.OBJECT_APPEARED,
                                distNames: [DISTNAMES.CHILD_1]
                            },
                            anotherEvent = {
                                eventType: EVENTS.OBJECT_APPEARED,
                                distNames: [DISTNAMES.GRANDCHILD_1]
                            };

                        assert(objectListener.onObjectAppeared.calledOnce);
                        assert(objectListener.onObjectAppeared.calledWith(sinon.match(event)));

                        assert(anotherObjectListener.onObjectAppeared.calledOnce);
                        assert(anotherObjectListener.onObjectAppeared.calledWith(sinon.match(anotherEvent)));
                        done();
                    });
                });


                it('should not call onObjectAppeared on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onObjectAppeared');
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(CLASSNAMES.CHILD);

                        assert.isUndefined(objectListener.onObjectAppeared);
                        done();
                    });
                });

                it('should not call onObjectAppeared on listener if object was in Infomodel', function (done) {
                    pushFirstUpdate(DISTNAMES.CHILD_1);
                    imTracker.trackByDistName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1);

                        assert(objectListener.onObjectAppeared.notCalled);
                        done();
                    });
                });

                it("should call onObjectAppeared on listener if object appears in Infomodel after reset", function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1);
                        pushReset();
                        pushFirstUpdate(DISTNAMES.CHILD_1);

                        var event = {
                            eventType: EVENTS.OBJECT_APPEARED,
                            distNames: [DISTNAMES.CHILD_1]
                        };

                        assert(objectListener.onObjectAppeared.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectAppeared.calledTwice);
                        done();
                    });
                });

                it("should call onObjectAppeared on listener if object appears if same class object exists", function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1);
                        pushFirstUpdate(DISTNAMES.CHILD_2);

                        var event = {
                            eventType: EVENTS.OBJECT_APPEARED,
                            distNames: [DISTNAMES.CHILD_1]
                        };

                        var secondEvent = {
                            eventType: EVENTS.OBJECT_APPEARED,
                            distNames: [DISTNAMES.CHILD_2]
                        };


                        assert(objectListener.onObjectAppeared.calledTwice);
                        assert(objectListener.onObjectAppeared.firstCall.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectAppeared.secondCall.calledWith(sinon.match(secondEvent)));
                        done();
                    });
                })
            });

            describe('onListenerRegistered testing', function () {

                beforeEach(function () {
                    objectListener.useWith('onListenerRegistered');
                    anotherObjectListener.useWith('onListenerRegistered');
                });

                it('should not call onListenerRegistered on listener if object was not in infomodel and is added to tracking', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1);

                        assert(objectListener.onListenerRegistered.notCalled);
                        done();
                    });
                });

                it('should call onListenerRegistered on listener if object was in infomodel and is added to tracking', function (done) {
                    pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        var event = {
                            eventType: 'onListenerRegistered',
                            objects: [
                                {distName: DISTNAMES.CHILD_1, parameters: DEFAULT_PARAMETERS},
                                {distName: DISTNAMES.CHILD_2, parameters: DEFAULT_PARAMETERS}]
                        };

                        assert(objectListener.onListenerRegistered.calledOnce);
                        assert(objectListener.onListenerRegistered.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should not call onListenerRegistered on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onListenerRegistered');
                    pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        assert.isUndefined(objectListener.onListenerRegistered);
                        done();
                    });
                });
            });

            describe('onObjectRemoved testing', function () {

                beforeEach(function () {
                    objectListener.useWith('onObjectRemoved');
                    anotherObjectListener.useWith('onObjectRemoved');
                });

                it('should call onObjectRemoved on listener if tracked single object was removed', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushRemove(DISTNAMES.CHILD_1);
                        var event = {eventType: EVENTS.OBJECT_REMOVED, distNames: [DISTNAMES.CHILD_1]};

                        assert(objectListener.onObjectRemoved.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectRemoved.calledOnce);
                        done();
                    });
                });

                it('should call onObjectRemoved on listener if all tracked objects were removed', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushRemove(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        var event = {
                            eventType: EVENTS.OBJECT_REMOVED,
                            distNames: [DISTNAMES.CHILD_1, DISTNAMES.CHILD_2]
                        };

                        assert(objectListener.onObjectRemoved.calledOnce);
                        assert(objectListener.onObjectRemoved.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should call onObjectRemoved on listener only once if tracked object was removed', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushRemove(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushRemove(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        var event = {
                            eventType: EVENTS.OBJECT_REMOVED,
                            distNames: [DISTNAMES.CHILD_1, DISTNAMES.CHILD_2]
                        };

                        assert(objectListener.onObjectRemoved.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectRemoved.calledOnce);
                        done();
                    });
                });


                it('should call onObjectRemoved on listener two time if tracked object was removed in the following updates', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushRemove(DISTNAMES.CHILD_1);
                        pushRemove(DISTNAMES.CHILD_2);
                        var event = {eventType: EVENTS.OBJECT_REMOVED, distNames: [DISTNAMES.CHILD_1]};
                        var secondEvent = {eventType: EVENTS.OBJECT_REMOVED, distNames: [DISTNAMES.CHILD_2]};

                        assert(objectListener.onObjectRemoved.firstCall.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectRemoved.secondCall.calledWith(sinon.match(secondEvent)));
                        assert(objectListener.onObjectRemoved.calledTwice);
                        done();
                    });
                });

                it('should not call onObjectRemoved on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onObjectRemoved');

                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushRemove(DISTNAMES.CHILD_1);

                        assert.isUndefined(objectListener.onObjectRemoved);
                        done();
                    });
                });

                it('should call onObjectRemoved on listener after reset', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushReset();
                        var event = {
                            eventType: EVENTS.OBJECT_REMOVED,
                            distNames: [DISTNAMES.CHILD_1, DISTNAMES.CHILD_2]
                        };

                        assert(objectListener.onObjectRemoved.calledWith(sinon.match(event)));
                        assert(objectListener.onObjectRemoved.calledOnce);
                        done();
                    });

                });

                it('should not call onObjectRemoved on listener if object was not removed', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        assert(objectListener.onObjectRemoved.notCalled);
                        done();
                    });

                });

            });

            describe('onParametersChanged', function () {

                beforeEach(function () {
                    objectListener.useWith('onParametersChanged');
                    anotherObjectListener.useWith('onParametersChanged');
                });

                it('should call onParametersChanged on listener if tracked object was changed', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        var event = {
                            eventType: EVENTS.PARAMETERS_CHANGED,
                            objects: [
                                {distName: DISTNAMES.CHILD_1, parameters: DEFAULT_PARAMETERS},
                                {distName: DISTNAMES.CHILD_2, parameters: DEFAULT_PARAMETERS}
                            ]
                        };
                        assert(objectListener.onParametersChanged.calledOnce);
                        assert(objectListener.onParametersChanged.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should call two times onParametersChanged on listener if tracked object was updated two times ', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushSecondUpdate(DISTNAMES.CHILD_1);
                        var event = {
                            eventType: EVENTS.PARAMETERS_CHANGED,
                            objects: [
                                {distName: DISTNAMES.CHILD_1, parameters: DEFAULT_PARAMETERS},
                                {distName: DISTNAMES.CHILD_2, parameters: DEFAULT_PARAMETERS}
                            ]
                        };

                        var secondEvent = {
                            eventType: EVENTS.PARAMETERS_CHANGED,
                            objects: [{distName: DISTNAMES.CHILD_1, parameters: NEW_PARAMETERS}]
                        };

                        assert(objectListener.onParametersChanged.firstCall.calledWith(sinon.match(event)));
                        assert(objectListener.onParametersChanged.secondCall.calledWith(sinon.match(secondEvent)));

                        assert(objectListener.onParametersChanged.calledTwice);
                        done();
                    });
                });

                it('should not call onParametersChanged on listener if tracked object was not updated ', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1);
                        pushFirstUpdate(DISTNAMES.GRANDCHILD_1);

                        assert(objectListener.onParametersChanged.calledOnce);
                        done();
                    });
                });

                it('should not call onParametersChanged on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onParametersChanged');

                    pushFirstUpdate(DISTNAMES.CHILD_1);
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushSecondUpdate(DISTNAMES.CHILD_1);

                        assert.isUndefined(objectListener.onParametersChanged);
                        done()
                    });
                });

                it('should call onParametersChanged on listener after reset', function (done) {
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushReset();
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        var event = {
                            eventType: EVENTS.PARAMETERS_CHANGED,
                            objects: [
                                {distName: DISTNAMES.CHILD_1, parameters: DEFAULT_PARAMETERS},
                                {distName: DISTNAMES.CHILD_2, parameters: DEFAULT_PARAMETERS}
                            ]
                        };

                        assert(objectListener.onParametersChanged.calledTwice);
                        assert(objectListener.onParametersChanged.calledWith(sinon.match(event)));
                        done();
                    });
                })

            });

            describe('onParameterChanged testing', function () {

                beforeEach(function () {
                    objectListener.useWith('onParameterChanged');
                    anotherObjectListener.useWith('onParameterChanged');
                });

                it('should call onParameterChanged on listener if object with tracked parameters is appeared', function (done) {
                    var parametersToTracking = ['param-1'];
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        var event = {
                            eventType: EVENTS.PARAMETER_CHANGED,
                            objects: [
                                {distName: DISTNAMES.CHILD_1, parameters: DEFAULT_PARAMETERS},
                                {distName: DISTNAMES.CHILD_2, parameters: DEFAULT_PARAMETERS}]
                        };

                        assert(objectListener.onParameterChanged.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should call onParameterChanged on listener if tracked parameters were changed', function (done) {
                    var parametersToTracking = ['param-1'];
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushSecondUpdate(DISTNAMES.CHILD_1);
                        var event = {
                            eventType: EVENTS.PARAMETER_CHANGED,
                            objects: [
                                {distName: DISTNAMES.CHILD_1, parameters: DEFAULT_PARAMETERS},
                                {distName: DISTNAMES.CHILD_2, parameters: DEFAULT_PARAMETERS}]
                        };

                        var secondEvent = {
                            eventType: EVENTS.PARAMETER_CHANGED,
                            objects: [{distName: DISTNAMES.CHILD_1, parameters: NEW_PARAMETERS}]
                        };

                        assert(objectListener.onParameterChanged.calledTwice);
                        assert(objectListener.onParameterChanged.firstCall.calledWith(sinon.match(event)));
                        assert(objectListener.onParameterChanged.secondCall.calledWith(sinon.match(secondEvent)));
                        done();
                    });
                });

                it('should call onParameterChanged on listener after reset', function (done) {
                    var parametersToTracking = ['param-1'];

                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushReset();
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        var event = {
                            eventType: EVENTS.PARAMETER_CHANGED,
                            objects: [
                                {distName: DISTNAMES.CHILD_1, parameters: DEFAULT_PARAMETERS},
                                {distName: DISTNAMES.CHILD_2, parameters: DEFAULT_PARAMETERS}]
                        };

                        assert(objectListener.onParameterChanged.calledTwice);
                        assert(objectListener.onParameterChanged.firstCall.calledWith(sinon.match(event)));
                        assert(objectListener.onParameterChanged.secondCall.calledWith(sinon.match(event)));
                        done();
                    });
                });

                it('should call onParameterChanged on listener if tracked parameters were changed with complex parameter', function (done) {
                    var parametersToTracking = ['param-2.param-2-1'];

                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushSecondUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        var event = {
                            eventType: EVENTS.PARAMETER_CHANGED,
                            objects: [
                                {distName: DISTNAMES.CHILD_1, parameters: DEFAULT_PARAMETERS},
                                {distName: DISTNAMES.CHILD_2, parameters: DEFAULT_PARAMETERS}]
                        };

                        var secondEvent = {
                            eventType: EVENTS.PARAMETER_CHANGED,
                            objects: [
                                {distName: DISTNAMES.CHILD_1, parameters: NEW_PARAMETERS},
                                {distName: DISTNAMES.CHILD_2, parameters: NEW_PARAMETERS}]
                        };

                        assert(objectListener.onParameterChanged.calledTwice);
                        assert(objectListener.onParameterChanged.firstCall.calledWith(sinon.match(event)));
                        assert(objectListener.onParameterChanged.secondCall.calledWith(sinon.match(secondEvent)));
                        done();
                    });
                });

                it('should not call onParameterChanged again on listener if tracked parameters were not changed', function (done) {
                    var parametersToTracking = ['param-1'];
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);

                        assert(objectListener.onParameterChanged.calledOnce);
                        done();
                    });
                });

                it('should not call onParameterChanged on listener if tracked parameters do not exist', function (done) {
                    var parametersToTracking = ['param-3'];

                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener, parametersToTracking).then(function () {
                        pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);

                        assert(objectListener.onParameterChanged.notCalled);
                        done();
                    });
                });

                it('should not call onParameterChanged on listener if event function does not exist', function (done) {
                    objectListener.removeEvent('onParameterChanged');
                    var parametersToTracking = ['param-1'];
                    pushFirstUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);
                    imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener, parametersToTracking).then(function () {
                        pushSecondUpdate(DISTNAMES.CHILD_1, DISTNAMES.CHILD_2);

                        assert.isUndefined(objectListener.onParameterChanged);
                        done();
                    });
                });
            });

            it('should stop tracking', function (done) {
                objectListener.useWith('onParametersChanged');
                anotherObjectListener.useWith('onParametersChanged');

                pushFirstUpdate(DISTNAMES.CHILD_1);
                imTracker.trackByClassName(CLASSNAMES.CHILD, objectListener)
                    .then(function () {
                        imTracker.stopTracking(objectListener, CLASSNAMES.CHILD);
                        pushSecondUpdate(DISTNAMES.CHILD_1);
                        assert(objectListener.onParametersChanged.notCalled);
                        assert(anotherObjectListener.onParametersChanged.notCalled);
                    }).done(done);
            });
        });
    });


    function pushFirstUpdate() {
        decoderServiceStub.setToFirstUpdate();
        update(arguments)
    }

    function pushSecondUpdate() {
        decoderServiceStub.setToSecondUpdate();
        update(arguments)
    }

    function update(distNames) {
        var updated = _.map(distNames, function (distName) {
            return {distName: distName, payload: 1};
        });
        onlineImTree.pushData({removedDistNames: [], managedObjects: updated});
    }

    function pushRemove() {
        remove.call(this, arguments)
    }

    function remove(distNames) {
        onlineImTree.pushData({removedDistNames: distNames, managedObjects: []});
    }

    function pushReset() {
        pushDisconnect("BTS");
    }

    function pushDisconnect(source) {
        onlineImTree.pushDisconnect({autoReconnect: false, source: source});
    }

});
