describe('imiMessageFactory', function () {
    var messageFactory;

    beforeEach(function () {
        messageFactory = infoModel.imi.messageFactory;
    });

    it('should get available versions',() => {
        var versions = messageFactory.getAvailableVersions();

        assert.deepEqual(versions, [9]);
    });

    it('should create messages for v9',() => {
        var ver9Messages = messageFactory.createMessages(9);

        assert.ok(ver9Messages);
    });

    it('should not create messages for invalid version',() => {
        expect(function () {
            messageFactory.createMessages(666);
        }).to.throw('Unsupported imi version: 666');
    });

    describe('imi v9',() => {
        var imiMessages;

        beforeEach(function () {
            imiMessages = messageFactory.createMessages(9);
        });

        describe('aaConfigGetServiceReq',() => {
            var message;

            beforeEach(function () {
                message = imiMessages.aaConfigGetServiceReq();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x1F24);
            });

            it('should set name',() => {
                message.setName("info model");

                assert.deepEqual(message.getName(), "info model");
            });

            it('should set major version',() => {
                message.setMajorVersion(2);

                assert.deepEqual(message.getMajorVersion(), 2);
            });

            it('should set minor version',() => {
                message.setMinorVersion(3);

                assert.deepEqual(message.getMinorVersion(), 3);
            });

            it('should set scope',() => {
                message.setScope(4);

                assert.deepEqual(message.getScope(), 4);
            });

            it('should convert to and from buffer',() => {
                message.setName("info model");
                message.setMajorVersion(2);
                message.setMinorVersion(3);
                message.setScope(4);

                var buffer = message.getBuffer();
                var actual = imiMessages.aaConfigGetServiceReq();
                actual.setBuffer(buffer);

                assert.deepEqual(actual.getName(), "info model");
                assert.deepEqual(actual.getMajorVersion(), 2);
                assert.deepEqual(actual.getMinorVersion(), 3);
                assert.deepEqual(actual.getScope(), 4);
            });

        });

        describe('aaConfigGetServiceRsp',() => {
            var message;

            beforeEach(function () {
                message = imiMessages.aaConfigGetServiceRsp();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x1F25);
            });

            it('should set name',() => {
                message.setName("info model");

                assert.deepEqual(message.getName(), "info model");
            });

            it('should set major version',() => {
                message.setMajorVersion(2);

                assert.deepEqual(message.getMajorVersion(), 2);
            });

            it('should set minor version',() => {
                message.setMinorVersion(3);

                assert.deepEqual(message.getMinorVersion(), 3);
            });

            it('should set scope',() => {
                message.setScope(4);

                assert.deepEqual(message.getScope(), 4);
            });

            it('should set num services',() => {
                message.setNumServices(5);

                assert.deepEqual(message.getNumServices(), 5);
            });

            it('should set minor',() => {
                message.setMinor(6);

                assert.deepEqual(message.getMinor(), 6);
            });

            it('should set sic address',() => {
                message.setSicAddress(7);

                assert.deepEqual(message.getSicAddress(), 7);
            });

        });

        describe('aaSysComEcho',() => {
            var message;

            beforeEach(function () {
                message = imiMessages.aaSysComEcho();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x0123);
            });
        });

        describe('versionQuery',() => {
            var message;

            beforeEach(function () {
                message = imiMessages.versionQuery();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x362F);
            });

            it('should have 0 bytes payload',() => {
                assert.deepEqual(new DataView(message.getBuffer()).byteLength, 0);
            });
        });

        describe('versionQueryResult',() => {
            var message;

            beforeEach(function () {
                message = imiMessages.versionQueryResult();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x3630);
            });

            it('should set supported release',() => {
                message.setSupportedRelease('main');

                assert.deepEqual(message.getSupportedRelease(), 'main');
            });

            it('should set supported major version',() => {
                message.setSupportedMajor(2);

                assert.deepEqual(message.getSupportedMajor(), 2);
            });

            it('should set supported minor version',() => {
                message.setSupportedMinor(3);

                assert.deepEqual(message.getSupportedMinor(), 3);
            });

            it('should change protobuf payload',() => {
                message.setSupportedMinor(3);

                var newMessage = imiMessages.versionQueryResult();
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getSupportedMinor(), 3);
            });

        });

        describe('signIn',() => {
            var message;

            beforeEach(function () {
                message = imiMessages.signIn();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x3621);
            });

            it('should set rat',() => {
                message.setRat(1);

                assert.deepEqual(message.getRat(), 1);
            });

            it('should set release',() => {
                message.setRelease("main");

                assert.deepEqual(message.getRelease(), "main");
            });

            it('should set major',() => {
                message.setMajor(666);

                assert.deepEqual(message.getMajor(), 666);
            });

            it('should set minor',() => {
                message.setMinor(666);

                assert.deepEqual(message.getMinor(), 666);
            });

            it('should set update interval',() => {
                message.setUpdateIntervalMs(5);

                assert.deepEqual(message.getUpdateIntervalMs(), 5);
            });

            it('should change protobuf payload',() => {
                message.setUpdateIntervalMs(999);

                var newMessage = imiMessages.signIn();
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getUpdateIntervalMs(), 999);
            });
        });

        describe('signInResult',() => {
            var message;

            beforeEach(function () {
                message = imiMessages.signInResult();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x3626);
            });

            it('should set status',() => {
                message.setStatus(0);

                assert.deepEqual(message.getStatus(), 0);
            });

            it('should set supported release',() => {
                message.setSupportedRelease('main');

                assert.deepEqual(message.getSupportedRelease(), 'main');
            });

            it('should set supported major version',() => {
                message.setSupportedMajor(2);

                assert.deepEqual(message.getSupportedMajor(), 2);
            });

            it('should set supported minor version',() => {
                message.setSupportedMinor(3);

                assert.deepEqual(message.getSupportedMinor(), 3);
            });

            it('should set requested release',() => {
                message.setRequestedRelease('main');

                assert.deepEqual(message.getRequestedRelease(), 'main');
            });

            it('should set requested major version',() => {
                message.setRequestedMajor(2);

                assert.deepEqual(message.getRequestedMajor(), 2);
            });

            it('should set requested minor version',() => {
                message.setRequestedMinor(3);

                assert.deepEqual(message.getRequestedMinor(), 3);
            });

            it('should change protobuf payload',() => {
                message.setStatus(1);

                var newMessage = imiMessages.signInResult();
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getStatus(), 1);
            });
        });

        describe('signOut',() => {
            var message;

            beforeEach(function () {
                message = imiMessages.signOut();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x3622);
            });

            it('should have 0 bytes payload',() => {
                assert.deepEqual(new DataView(message.getBuffer()).byteLength, 0);
            });
        });

        describe('operationExecuted',() => {
            var message;

            beforeEach(function () {
                message = imiMessages.operationExecuted();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x3625);
            });

            it('should return valid request id',() => {
                assert.deepEqual(message.getRequestId(), 1);
            });

            it('should return valid execution status',() => {
                assert.deepEqual(message.getExecutionStatus(), 1);
            });

            it('should correctly decode and encode message',() => {
                var newMessage = imiMessages.operationExecuted();
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getRequestId(), message.getRequestId());
                assert.deepEqual(newMessage.getExecutionStatus(), message.getExecutionStatus());
            });
        });

        describe('updateNotification',() => {
            var message;
            beforeEach(function () {
                message = imiMessages.updateNotification();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x3623);
            });
        });

        describe('changeRequest',() => {
            var message,
                operations = [{
                    "operation_type": 3,
                    "object": {
                        "object_info": {
                            object_id: '0x0',
                            object_type: 0,
                            object_create_info: {
                                instance_id: 0
                            }
                        },
                        "payload": null
                    }
                }, {
                    "operation_type": 3,
                    "object": {
                        "object_info": {
                            object_id: '0x1',
                            object_type: 1,
                            object_create_info: {
                                instance_id: 1,
                                parent_id: '0x0'
                            }
                        },
                        "payload": null
                    }
                }];
            beforeEach(function () {
                message = imiMessages.changeRequest();
            });
            it('should be represented by valid string',() => {
                message.setRequestId(1);
                message.setOperations(operations);
                var expectedString = 'changeRequest{' +
                    '\ngetId: 13860 [0x3624]' +
                    '\ngetRequestId: 1 [0x1]' +
                    '\ngetOperations: [object Object],[object Object]' +
                    '\ngetBuffer: [object ArrayBuffer] [42bytes]' +
                    '\n}';
                assert.deepEqual(message.toString(), expectedString);
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x3624);
            });

            it('should set request id',() => {
                message.setRequestId(7);

                assert.deepEqual(message.getRequestId(), 7);
            });

            it('should set operations',() => {
                message.setOperations(operations);

                assert.deepEqual(message.getOperations(), operations);
            });

            it('should correctly decode and encode message',() => {
                message.setRequestId(1);
                message.setOperations(operations);
                var newMessage = imiMessages.changeRequest();
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getRequestId(), message.getRequestId());
                operationsEqual(newMessage.getOperations(), message.getOperations());
            });

            function operationsEqual(operations1, operations2) {
                assert.deepEqual(operations1.length, operations2.length);
                for (var i = 0; i < operations1.length; i++) {
                    assert.deepEqual(operations1[i].distname, operations2[i].distname);
                }
            }
        });

        describe('registerExecutor',() => {
            var message;
            var objectTypes = [0, 1];
            var objectIds = [12, 13];
            beforeEach(function () {
                message = imiMessages.registerExecutor();
            });
            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x3627);
            });

            it('should return valid request id',() => {
                assert.deepEqual(message.getRequestId(), 1);
            });
            it('should return valid array of object types',() => {
                message.setObjectTypes(objectTypes);
                assert.deepEqual(message.getObjectTypes(), objectTypes);
            });

            it('should return valid array of ids',() => {
                message.setObjectIds(objectIds);
                assert.deepEqual(message.getObjectIds(), objectIds);
            });

            it('should correctly decode and encode message',() => {
                var newMessage = imiMessages.registerExecutor();
                newMessage.setObjectTypes(objectTypes);
                newMessage.setObjectIds(objectIds);
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getRequestId(), message.getRequestId());
                assert.deepEqual(newMessage.getObjectTypes(), message.getObjectTypes());
                assert.deepEqual(newMessage.getObjectIds(), message.getObjectIds());
            });
        });

        describe('executorRegistered',() => {
            var message;
            var REGISTRATION_STATUS_SUCCESS = 0;

            beforeEach(function () {
                message = imiMessages.executorRegistered();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x3628);
            });

            it('should return valid request id',() => {
                assert.deepEqual(message.getRequestId(), 1);
            });

            it('should return valid registration status',() => {
                message.setRegistrationStatus(REGISTRATION_STATUS_SUCCESS);
                assert.deepEqual(message.getRegistrationStatus(), REGISTRATION_STATUS_SUCCESS);
            });

            it('should correctly decode and encode message',() => {
                var newMessage = imiMessages.executorRegistered();
                message.setRegistrationStatus(REGISTRATION_STATUS_SUCCESS);
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getRequestId(), message.getRequestId());
                assert.deepEqual(newMessage.getRegistrationStatus(), message.getRegistrationStatus());
            });
        });

        describe('unregisterExecutor',() => {
            var message;
            var objectTypes = [0, 1];
            var objectIds = [12, 13];
            beforeEach(function () {
                message = imiMessages.unregisterExecutor();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x3629);
            });

            it('should return valid array of managed object classes',() => {
                message.setObjectTypes(objectTypes);
                assert.deepEqual(message.getObjectTypes(), objectTypes);
            });

            it('should return valid array of managed object distNames',() => {
                message.setObjectIds(objectIds);
                assert.deepEqual(message.getObjectIds(), objectIds);
            });

            it('should correctly decode and encode message',() => {
                var newMessage = imiMessages.unregisterExecutor();
                newMessage.setObjectTypes(objectTypes);
                newMessage.setObjectIds(objectIds);
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getObjectTypes(), message.getObjectTypes());
                assert.deepEqual(newMessage.getObjectIds(), message.getObjectIds());
            });
        });

        describe('executionRequest',() => {
            var message;
            var operations = [{
                "operation_type": 3,
                "object": {
                    "object_info": {
                        object_id: '0x0',
                        object_type: 0,
                        object_create_info: {
                            instance_id: 0
                        }
                    },
                    "payload": null
                }
            }, {
                "operation_type": 3,
                "object": {
                    "object_info": {
                        object_id: '0x1',
                        object_type: 1,
                        object_create_info: {
                            instance_id: 1,
                            parent_id: '0x0'
                        }
                    },
                    "payload": null
                }
            }];

            beforeEach(function () {
                message = imiMessages.executionRequest();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x362A);
            });

            it('should set request id',() => {
                message.setRequestId(7);
                assert.deepEqual(message.getRequestId(), 7);
            });

            it('should set operations',() => {
                message.setOperations(operations);
                assert.deepEqual(message.getOperations(), operations);
            });


            it('should correctly decode and encode message',() => {
                message.setRequestId(1);
                message.setOperations(operations);
                var newMessage = imiMessages.executionRequest();
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getRequestId(), message.getRequestId());
                operationsEqual(newMessage.getOperations(), message.getOperations());
            });

            function operationsEqual(operations1, operations2) {
                assert.deepEqual(operations1.length, operations2.length);
                for (var i = 0; i < operations1.length; i++) {
                    assert.deepEqual(operations1[i].distname, operations2[i].distname);
                }
            }
        });

        describe('executionResult',() => {
            var message;
            var EExecutionStatus_EXECUTED = 0;

            beforeEach(function () {
                message = imiMessages.executionResult();
            });

            it('should have valid id',() => {
                assert.deepEqual(message.getId(), 0x362B);
            });

            it('should return valid request id',() => {
                assert.deepEqual(message.getRequestId(), 1);
            });

            it('should return valid execution status',() => {
                message.setExecutionStatus(EExecutionStatus_EXECUTED);
                assert.deepEqual(message.getExecutionStatus(), EExecutionStatus_EXECUTED);
            });

            it('should correctly decode and encode message',() => {
                var newMessage = imiMessages.executionResult();
                newMessage.setExecutionStatus(EExecutionStatus_EXECUTED);
                newMessage.setBuffer(message.getBuffer());

                assert.deepEqual(newMessage.getRequestId(), message.getRequestId());
                assert.deepEqual(newMessage.getExecutionStatus(), message.getExecutionStatus());
            });
        });
    });
});
