const testUtil = require('./test.util');
const ims2Handler = require('../src/ims2');

describe('ims2', function () {

    var FRAME_TYPE = {
        CHANGESET: 0,
        RESET: 1,
        META: 2
    };

    describe('read async', function () {
        var progressListener;

        beforeEach(function () {
            progressListener = {
                onComplete: function (ims2) {
                },
                onReadFrame: function (frame) {
                    progressListener.frames = progressListener.frames || [];
                    progressListener.frames.push(frame);
                },
                onReset: function (reset) {
                    progressListener.onResetInvoked = progressListener.onResetInvoked || 0;
                    progressListener.onResetInvoked++;
                },
                onUpdate: function (update) {
                    progressListener.onUpdateInvoked = progressListener.onUpdateInvoked || 0;
                    progressListener.onUpdateInvoked++;
                },
                onMetaBuffer: function (metaBuffer) {
                    progressListener.onMetaInvoked = progressListener.onMetaInvoked || 0;
                    progressListener.onMetaInvoked++;
                },
                onReadBytes: function (title, readCount, totalCount) {
                    progressListener.readBytes = progressListener.readBytes || [];
                    progressListener.readBytes.push(readCount);
                    progressListener.totalCount = totalCount;
                }
            };
        });

        it('should fill ims2 structure from arrayBuffer', function (done) {
            loadBuffer('base/resources/ims2/faults.dat', function (arrayBuffer) {
                progressListener.onComplete = function (ims2) {
                    assert.equal(ims2.version, 1);
                    assert.equal(ims2.headerLength, 23);
                    assert.equal(ims2.header.infoLength, 19);
                    assert.equal(ims2.header.info, 'TODO: add info here');

                    var frameWithMeta = ims2.frames[0];
                    assert.equal(frameWithMeta.header.frameType, FRAME_TYPE.META);
                    assert.equal(frameWithMeta.header.compression, 0);
                    assert.equal(frameWithMeta.header.payloadLength, 178433);

                    var frameWithUpdates = ims2.frames[1];
                    assert.equal(frameWithUpdates.header.frameType, FRAME_TYPE.CHANGESET);
                    assert.equal(frameWithUpdates.header.compression, 0);
                    assert.equal(frameWithUpdates.header.payloadLength, 531452);

                    assert.equal(ims2.timeline.length, 215);

                    var firstEntry = ims2.timeline[0];
                    var firstNotif = firstEntry.change;
                    assert.equal(firstEntry.timestamp, 1408708730301);
                    assert.equal(firstEntry.numOfChanges, 2147);
                    assert.equal(firstNotif.removed.length, 0);
                    assert.equal(firstNotif.updated.length, 2147);

                    var lastNotif = ims2.timeline[ims2.timeline.length - 1].change;
                    assert.equal(lastNotif.removed.length, 0);
                    assert.equal(lastNotif.updated.length, 1);
                    assert.equal(lastNotif.updated[0].distName, "/MRBTS-1/EQM-1/SMOD-1/CCU-1/OSCILLATOR-1");
                    assert.equal(lastNotif.updated[0].payload.length, 20);

                    done();
                };

                ims2Handler.readAsync(arrayBuffer, progressListener);
            });
        });

        it('should invoke complete when read all ims2 data', function (done) {
            loadBuffer('base/resources/ims2/faults.dat', function (arrayBuffer) {
                progressListener.onComplete = function (ims2) {
                    assert.equal(ims2.timeline.length, 215);
                    done();
                };

                ims2Handler.readAsync(arrayBuffer, progressListener);
            });
        });

        it('should notify about each read frame', function (done) {
            loadBuffer('base/resources/ims2/faults.dat', function (arrayBuffer) {
                progressListener.onComplete = function (ims2) {
                    assert.deepEqual(progressListener.frames, ims2.frames);
                    done();
                };

                ims2Handler.readAsync(arrayBuffer, progressListener);
            });
        });

        it('should notify about each meta buffer', function (done) {
            loadBuffer('base/resources/ims2/faults.dat', function (arrayBuffer) {
                progressListener.onComplete = function () {
                    assert.equal(progressListener.onMetaInvoked, 1);
                    done();
                };

                ims2Handler.readAsync(arrayBuffer, progressListener);
            });
        });

        it('should notify about each reset', function (done) {
            loadBuffer('base/resources/ims2/two_resets.dat', function (arrayBuffer) {
                progressListener.onComplete = function () {
                    assert.deepEqual(progressListener.onResetInvoked, 2);
                    done();
                };

                ims2Handler.readAsync(arrayBuffer, progressListener);
            });
        });

        it('should notify about update', function (done) {
            loadBuffer('base/resources/ims2/two_resets.dat', function (arrayBuffer) {
                progressListener.onComplete = function () {
                    assert.deepEqual(progressListener.onUpdateInvoked, 2746);
                    done();
                };

                ims2Handler.readAsync(arrayBuffer, progressListener);
            });
        });

        it('should notify about read bytes', function (done) {
            loadBuffer('base/resources/ims2/two_resets.dat', function (arrayBuffer) {
                progressListener.onComplete = function () {
                    var actualTotal = progressListener.readBytes[progressListener.readBytes.length - 1];
                    assert.equal(actualTotal, progressListener.totalCount);

                    done();
                };

                ims2Handler.readAsync(arrayBuffer, progressListener);
            });
        });

        it('should cancel reading', function (done) {
            loadBuffer('base/resources/ims2/two_resets.dat', function (arrayBuffer) {
                var cancelIndex = 3;
                var cancelFn;

                progressListener.onUpdate = function () {
                    progressListener.onUpdateInvoked = progressListener.onUpdateInvoked || 0;
                    progressListener.onUpdateInvoked++;

                    if (progressListener.onUpdateInvoked >= cancelIndex) {
                        cancelFn();
                    }
                };

                progressListener.onComplete = function (ims2) {
                    assert.equal(ims2.timeline.length, cancelIndex);

                    done();
                };

                cancelFn = ims2Handler.readAsync(arrayBuffer, progressListener);
            });
        });

        it('should read ims2 with execution requests', function (done) {
            loadBuffer('base/resources/ims2/executionRequests.dat', function (arrayBuffer) {
                progressListener.onComplete = function (ims2) {
                    assert.equal(ims2.timeline.length, 778);
                    done();
                };

                ims2Handler.readAsync(arrayBuffer, progressListener);
            });
        });

        it('should read ims2 with compressed frames', function (done) {
            loadBuffer('base/resources/ims2/compressedFrames.dat', function (arrayBuffer) {
                progressListener.onComplete = function (ims2) {
                    assert.equal(ims2.timeline.length, 1556);
                    done();
                };

                ims2Handler.readAsync(arrayBuffer, progressListener);
            });
        });
    });

    function loadBuffer(file, fnOnBuffer) {
        testUtil.loadBuffer(file)
            .then(fnOnBuffer);
    }
});
