const testUtil = require('./test.util');
const ims2Handler = require('../src/ims2');
const ims2writer = require('../src/ims2writer');

describe('ims2writer', function () {

    var HEADER = {
        ADDITIONAL_INFO_OFFSET: 4,
        ADDITIONAL_INFO: "AdditionalInfo"
    };

    var FRAME = {
        STARTING_BYTES: [0xE9, 0x11, 0x00, 0xA8, 0x43, 0xA0, 0x41, 0x2D, 0x94, 0xB3, 0x06, 0xDA],
        TYPE_OFFSET: 12,
        COMPRESSION_OFFSET: 16,
        PAYLOAD_OFFSET: 24
    };

    var FRAME_TYPE = {
        CHANGESET: 0,
        RESET: 1,
        META: 2
    };

    var COMPRESSION_TYPE = {
        NO_COMPRESSION: 0,
        GZIP: 1
    };

    describe('write ims2', function () {
        var readFileListener;
        var readWriterBufferListener;
        var writer;
        var metaService, typeInfo;
        var decode, encode

        beforeEach(function () {
            typeInfo = createTypeInfo('string');
            metaService = {
                getTypeInfo: function (uuid, className) {
                    return typeInfo;
                }
            };

            decode = sinon.spy();
            encode = sinon.spy();
            writer = ims2writer({decode: decode},
                function () {
                    return {
                        encode: encode
                    }
                }, metaService).newInstance();

            readFileListener = {
                onComplete: function (ims2) {
                },
                onReadFrame: function (frame) {
                },
                onReset: function (reset) {
                    writer.addReset(reset);
                },
                onUpdate: function (update) {
                    writer.addEntry(update);
                },
                onMetaBuffer: function (metaBuffer) {
                    writer.addMeta(metaBuffer);
                },
                onReadBytes: function (readCount, totalCount) {
                }
            };

            readWriterBufferListener = {
                onComplete: function (ims2) {
                },
                onReadFrame: function (frame) {
                },
                onReset: function (reset) {

                },
                onUpdate: function (update) {

                },
                onMetaBuffer: function (metaBuffer) {

                },
                onReadBytes: function (readCount, totalCount) {
                }
            };
        });

        it('should not writer equal to null', function (done) {
            assert.isNotNull(writer);
            done();
        });

        it('should create new instance of writer', function (done) {
            loadBuffer('base/resources/ims2/two_resets.dat', function (arrayBuffer) {
                readFileListener.onComplete = function (ims2) {
                    assert.notEqual(writer.getFrames().length, 0);
                    writer = ims2writer({decode: decode},
                        function () {
                            return {
                                encode: encode
                            }
                        }, metaService).newInstance();
                    assert.equal(writer.getFrames().length, 0);
                    done();
                };

                ims2Handler.readAsync(arrayBuffer, readFileListener);
            });
        });

        it('should create file header', function (done) {
            var header = writer.getHeader();
            assert.isNotNull(header);
            var dataView = new DataView(header);
            assert.equal(dataView.getInt32(0), HEADER.ADDITIONAL_INFO.length);
            assert.equal(decodeString(dataView, HEADER.ADDITIONAL_INFO_OFFSET, HEADER.ADDITIONAL_INFO.length), HEADER.ADDITIONAL_INFO);
            done();
        });

        it('should add appropriate frames', function (done) {
            loadBuffer('base/resources/ims2/two_resets.dat', function (arrayBuffer) {
                readFileListener.onComplete = function (ims2) {
                    writer.saveBuffer();
                    var frames = writer.getFrames();
                    assert.equal(frames.length, 6);
                    assertFrameNumber(frames, 2, FRAME_TYPE.RESET);
                    assertFrameNumber(frames, 1, FRAME_TYPE.META);
                    assertFrameNumber(frames, 3, FRAME_TYPE.CHANGESET);
                    done();
                };

                ims2Handler.readAsync(arrayBuffer, readFileListener);
            });
        });

        it('should frames have structure with header', function (done) {
            loadBuffer('base/resources/ims2/two_resets.dat', function (arrayBuffer) {
                readFileListener.onComplete = function (ims2) {
                    writer.saveBuffer();
                    var frames = writer.getFrames();
                    assertFrame(frames[0], FRAME_TYPE.META);
                    assertFrame(frames[1], FRAME_TYPE.CHANGESET);
                    assertFrame(frames[2], FRAME_TYPE.RESET);
                    assertFrame(frames[3], FRAME_TYPE.CHANGESET);
                    assertFrame(frames[4], FRAME_TYPE.RESET);
                    assertFrame(frames[5], FRAME_TYPE.CHANGESET);
                    done();
                };

                ims2Handler.readAsync(arrayBuffer, readFileListener);
            });
        });

        it('should write ims2 with executions requests and status', function (done) {
            loadBuffer('base/resources/ims2/executionRequestsAndStatus.dat', function (arrayBuffer) {
                var expectedIms2 = null;

                readWriterBufferListener.onComplete = function (ims2) {
                    assertTimeline(ims2.timeline, expectedIms2.timeline);
                    done();
                };

                readFileListener.onComplete = function (ims2) {
                    writer.saveBuffer();
                    expectedIms2 = ims2;
                    ims2Handler.readAsync(writer.getBuffer(), readWriterBufferListener);
                };

                ims2Handler.readAsync(arrayBuffer, readFileListener);
            });
        });

        it('should write ims2 with resets', function (done) {
            loadBuffer('base/resources/ims2/two_resets.dat', function (arrayBuffer) {
                var expectedIms2 = null;

                readWriterBufferListener.onComplete = function (ims2) {
                    assertTimeline(ims2.timeline, expectedIms2.timeline);
                    done();
                };

                readFileListener.onComplete = function (ims2) {
                    writer.saveBuffer();
                    expectedIms2 = ims2;
                    ims2Handler.readAsync(writer.getBuffer(), readWriterBufferListener);
                };

                ims2Handler.readAsync(arrayBuffer, readFileListener);
            });
        });

        it('should write ims2 with only one entry with objects', function (done) {
            loadBuffer('base/resources/ims2/executionRequestsAndStatus.dat', function (arrayBuffer) {
                var meta = null;
                var writtenObjects = null;
                var expectedEntry = null;
                readWriterBufferListener.onComplete = function (ims2) {
                    assert.equal(ims2.timeline.length, 1);
                    assert.deepEqual(ims2.timeline[0], expectedEntry);
                    done();
                };

                readFileListener.onMetaBuffer = function (metaBuffer) {
                    meta = metaBuffer;
                };

                readFileListener.onComplete = function (ims2) {
                    expectedEntry = ims2.timeline[5];
                    writer.saveBufferFromObjects(expectedEntry, meta, function (buffer) {
                        writtenObjects = buffer;
                    });
                    ims2Handler.readAsync(writtenObjects, readWriterBufferListener);
                };

                ims2Handler.readAsync(arrayBuffer, readFileListener);
            });
        });

        it('should write ims2 with ObjectId parameter', function () {
            // given
            const payload = ['0x6', '0x6', '0x6'];
            typeInfo = createTypeInfo('objectid');

            // when
            writer.addEntry({
                change: {
                    updated: [{
                        distName: 'distname',
                        payload: payload
                    }],
                    removed: []
                },
                numOfChanges: 1,
                timestamp: 0
            });

            // then
            assert(decode.calledOnce);
            assert(encode.calledOnce);
            sinon.assert.calledWith(decode, metaService.getTypeInfo(), payload);
            sinon.assert.calledWith(encode, metaService.getTypeInfo(), sinon.match.any, true);
        });

        function createTypeInfo(type) {
            return infoModel.typeInfoV2({
                p: [
                    {'@id': 0, '@type': type, '@name': 'parameter1'}
                ]
            });
        }

        function assertTimeline(timeline, expectedTimeline) {
            assert.equal(timeline.length, expectedTimeline.length);
            var INDEX_OF_FIRST_ELEMENT = 0;
            var INDEX_OF_THE_MIDDLE_ELEMENT = Math.floor(expectedTimeline.length / 2);
            var INDEX_OF_LAST_ELEMENT = expectedTimeline.length - 1;
            assert.deepEqual(timeline[INDEX_OF_FIRST_ELEMENT], expectedTimeline[INDEX_OF_FIRST_ELEMENT]);
            assert.deepEqual(timeline[INDEX_OF_THE_MIDDLE_ELEMENT], expectedTimeline[INDEX_OF_THE_MIDDLE_ELEMENT]);
            assert.deepEqual(timeline[INDEX_OF_LAST_ELEMENT], expectedTimeline[INDEX_OF_LAST_ELEMENT]);
        }

        function assertFrame(frame, type) {
            var dataView = new DataView(frame.content);
            var magicStringAsBytes = getBytes(dataView, FRAME.MAGIC_STRING_OFFSET, 12);
            assert.deepEqual(new Uint8Array(FRAME.STARTING_BYTES), magicStringAsBytes);
            assert.equal(dataView.getInt32(FRAME.TYPE_OFFSET), type);
            assert.equal(dataView.getInt32(FRAME.COMPRESSION_OFFSET), COMPRESSION_TYPE.NO_COMPRESSION);
        }

        function assertFrameNumber(frames, expected, type) {
            var number = 0;
            for (var i = 0, len = frames.length; i < len; i++)
                if (frames[i].type === type)
                    number++;
            assert.equal(expected, number);
        }

    });

});

function decodeString(dataView, offset, strLength) {
    if (TextDecoder) {
        //Firefox and Chrome
        var unit8Array = new Uint8Array(dataView.buffer, offset, strLength);
        var decoder = new TextDecoder();
        return decoder.decode(unit8Array);
    } else {
        var array = new Array(strLength);
        for (var i = 0; i < strLength; i++) {
            var uint8 = dataView.getUint8(offset + i);
            array.push(String.fromCharCode(uint8));
        }
        return array.join('');
    }
}

function getBytes(view, offset, numOfBytes) {
    var buffer = view.buffer;
    return new Uint8Array(buffer, offset, numOfBytes);
}

function loadBuffer(file, fnOnBuffer) {
    testUtil.loadBuffer(file).then(fnOnBuffer);
}
