const ImTree = require('../src/imtree');

describe('ImTree', function () {
    var tree, sandbox;

    beforeEach(function () {
        tree = new ImTree();
        sandbox = sinon.sandbox.create();
    });

    afterEach(function () {
        sandbox.restore();
    });

    function createUpdate(distName) {
        return {
            distName: distName,
            payload: new ArrayBuffer(0)
        };
    }

    describe('#merge()', function () {

        it('should merge update', function () {
            var update = createUpdate('/MRBTS-1');

            tree.merge(update);

            assert.ok(tree.findByDistName('/MRBTS-1'));
        });

        it('should merge with missing node', function () {
            var metaUid = 123;
            tree.merge(createUpdate('/MRBTS-1/EQM_L-1'), metaUid);

            var node = tree.findByDistName('/MRBTS-1');

            assert.isTrue(node.isMissing());
            assert.equal(node.distName, '/MRBTS-1');
            assert.equal(node.metaUid, metaUid);
            assert.isNull(node.payload);
        });

        it('should merge missing node', function () {
            var metaUid = 123;
            tree.merge(createUpdate('/MRBTS-1/EQM_L-1'), metaUid);
            tree.merge(createUpdate('/MRBTS-1'), metaUid);

            var node = tree.findByDistName('/MRBTS-1');

            assert.isFalse(node.isMissing());
            assert.ok(node.payload);
        });

        it('should merge missing node with different meta uid', function () {
            tree.merge(createUpdate('/MRBTS-1/EQM_L-1'), 123);
            var missingNode = tree.findByDistName('/MRBTS-1');
            var existingNode = tree.findByDistName('/MRBTS-1/EQM_L-1');
            assert.equal(missingNode.metaUid, 123);
            assert.equal(existingNode.metaUid, 123);

            tree.merge(createUpdate('/MRBTS-1'), 321);
            var updatedNode = tree.findByDistName('/MRBTS-1');
            assert.equal(updatedNode.metaUid, 321);
            assert.equal(existingNode.metaUid, 123);
        });

        it('should merge node', function () {
            var distName = '/MRBTS-1/EQM_L-1';
            var parameters = {};
            var payload = new ArrayBuffer(0);

            var node = tree.mergeNode(distName, parameters, payload);

            assert.equal(node.distName, distName);
            assert.equal(node.parameters, parameters);
            assert.equal(node.payload, payload);
        });

        it('should merge updates', function () {
            var updateRoot = createUpdate('/MRBTS-1');
            var updateChild = createUpdate('/MRBTS-1/RAT-1');

            tree.merge(updateRoot);
            tree.merge(updateChild);

            assert.ok(tree.findByDistName('/MRBTS-1/RAT-1'));
        });

        it('should merge update payload', function () {
            var update1 = createUpdate('/MRBTS-1');
            var update2 = createUpdate('/MRBTS-1');

            tree.merge(update1);
            tree.merge(update2);
            assert.equal(tree.findByDistName('/MRBTS-1').payload, update2.payload);
        });

        it('should merge with missing nodes update', function () {
            var update = createUpdate('/MRBTS-1/RAT-1');

            tree.merge(update);

            assert.isTrue(tree.findByDistName('/MRBTS-1').isMissing());
            assert.equal(tree.findByDistName('/MRBTS-1/RAT-1').payload, update.payload);
        })

    });

    describe('#remove()', function () {

        it('should remove by distName', function () {
            var update = createUpdate('/MRBTS-1');

            var node = tree.merge(update);

            tree.removeByDistName('/MRBTS-1');

            assert.notOk(tree.findByDistName('/MRBTS-1'));
            assert.notOk(node.parent);
            assert.notOk(node.root);
        });

        it('should remove root with children by distName', function () {
            var update = createUpdate('/MRBTS-1/RAT-1');

            tree.merge(update);

            tree.removeByDistName('/MRBTS-1');

            assert.notOk(tree.findByDistName('/MRBTS-1'));
            assert.notOk(tree.findByDistName('/MRBTS-1/RAT-1'));
        });

        it('should clear node map when removing root', function () {
            var update = createUpdate('/MRBTS-1/RAT-1/EQM_L-1');
            tree.merge(update);

            tree.removeByDistName('/MRBTS-1');

            assert.equal(tree.root.nodeMap.size, 0);
        });

        it('should not remove by missing distName', function () {
            var update = createUpdate('/MRBTS-1');

            tree.merge(update);

            tree.removeByDistName('/MRBTS-1/XXX-1');

            assert.ok(tree.findByDistName('/MRBTS-1'));
        });

        it('should remove by distNames', function () {
            var update = createUpdate('/MRBTS-1/RAT-1');

            tree.merge(update);

            tree.removeByDistNames(['/MRBTS-1/RAT-1', '/MRBTS-1']);

            assert.notOk(tree.findByDistName('/MRBTS-1'));
            assert.notOk(tree.findByDistName('/MRBTS-1/RAT-1'));
        })

    });

    describe('#findByClassName()', function () {

        it('should find by class name', function () {
            var update = createUpdate('/MRBTS-1/RAT-1');

            tree.merge(update);

            var node = tree.findByDistName('/MRBTS-1');
            assert.deepEqual(tree.findByClassName('MRBTS'), [node]);
        });

        it('should return itself when class name matches', function () {
            var update = createUpdate('/MRBTS-1/RAT-1/RAT-2');

            tree.merge(update);

            var node = tree.findByDistName('/MRBTS-1/RAT-1');
            var child = tree.findByDistName('/MRBTS-1/RAT-1/RAT-2');
            assert.deepEqual(node.findByClassName('RAT'), [node, child]);
        });

        it('should return empty array when no such class name', function () {
            var update = createUpdate('/MRBTS-1');

            tree.merge(update);

            assert.deepEqual(tree.findByClassName('XXX'), []);
        })

    });

    describe('#findAnyByClassName()', function () {

        it('should find by className using BFS', function () {
            tree.merge(createUpdate('/MRBTS-1/EQM-1'));
            tree.merge(createUpdate('/MRBTS-1/RAT-1'));
            tree.merge(createUpdate('/MRBTS-1/EQM-1/RAT-1'));

            assert.equal(tree.findAnyByClassName('RAT').distName, '/MRBTS-1/RAT-1')
        });

    });

    describe('#findChildrenByClassName()', function () {
        it('should return array with only direct children', function () {
            tree.merge(createUpdate('/MRBTS-1'));
            tree.merge(createUpdate('/BTS-1/MRBTS-1'));

            assert.deepEqual(tree.findChildrenByClassName('MRBTS'), [tree.findByDistName('/MRBTS-1')]);
        });

        it('Should return empty array if there is no matching classes', function () {
            tree.merge(createUpdate('/BTS-1/MRBTS-1'));

            assert.deepEqual(tree.findChildrenByClassName('MRBTS'), []);
        })
    });

    describe('#traverse()', function () {

        it('should do nothing on empty tree', function () {
            var visited = [];
            tree.traverse(function (node) {
                visited.push(node);
            });

            assert.deepEqual(visited.length, 0);
        });

        it('should visit each node', function () {
            var update = createUpdate('/MRBTS-1/RAT-1');

            tree.merge(update);

            var visited = [];
            tree.traverse(function (node) {
                visited.push(node);
            });

            assert.deepEqual(visited.length, 2);
        })

    });

    describe('#addNode()', function () {

        it('should add node to root', function () {
            var node = tree.addNode(tree.root, 'MRBTS', 1);

            assert.ok(tree.findByDistName('/MRBTS-1'));
            assert.deepEqual(node.distName, '/MRBTS-1');
            assert.deepEqual(node.className, 'MRBTS');
            assert.deepEqual(node.id, 1);
        });

        it('should add node to root when parant node is null', function () {
            var node = tree.addNode(null, 'MRBTS', 1);

            assert.ok(tree.findByDistName('/MRBTS-1'));
            assert.deepEqual(node.distName, '/MRBTS-1');
            assert.deepEqual(node.className, 'MRBTS');
            assert.deepEqual(node.id, 1);
        });

        it('should add node to parent node', function () {
            var parent = tree.addNode(tree.root, 'MRBTS', 1);
            var child = tree.addNode(parent, 'RAT', 1);

            assert.ok(tree.findByDistName('/MRBTS-1/RAT-1'));
            assert.deepEqual(child.distName, '/MRBTS-1/RAT-1');
        })

    });

    describe('#decoderService()', function () {
        var decoderService;
        var parameters = {}, typeInfo = {};
        var decodeByClassNameStub, getTypeInfoStub;

        beforeEach(function () {
            decodeByClassNameStub = sandbox.stub().returns(parameters);
            getTypeInfoStub = sandbox.stub().returns(typeInfo);

            decoderService = {
                decodeByClassName: decodeByClassNameStub,
                getTypeInfo: getTypeInfoStub
            };

            tree = new ImTree(decoderService);
        });

        it('should get decoder service from root', function () {
            var update = createUpdate('/MRBTS-1/RAT-1');

            tree.merge(update);

            assert.equal(tree.findByDistName('/MRBTS-1').root.decoderService, decoderService);
            assert.equal(tree.findByDistName('/MRBTS-1/RAT-1').root.decoderService, decoderService);
        });

        it('should lazy decode parameters using decoder service', function () {
            var update = createUpdate('/MRBTS-1/RAT-1');

            tree.merge(update);

            var node = tree.findByDistName('/MRBTS-1/RAT-1');

            node.getDecodedParameters();
            node.getDecodedParameters();
            node.getDecodedParameters();

            assert.equal(decodeByClassNameStub.callCount, 1);
        });

        it('should lazy get typeinfo using decoder service', function () {
            var update = createUpdate('/MRBTS-1/RAT-1');

            tree.merge(update);

            var node = tree.findByDistName('/MRBTS-1/RAT-1');

            node.getTypeInfo();
            node.getTypeInfo();
            node.getTypeInfo();

            assert.equal(getTypeInfoStub.callCount, 1);
        });

        it('should invalidate parameters after new payload', function () {
            var update1 = createUpdate('/MRBTS-1/RAT-1');
            var update2 = createUpdate('/MRBTS-1/RAT-1');

            tree.merge(update1);

            var node = tree.findByDistName('/MRBTS-1/RAT-1');
            node.getDecodedParameters();

            tree.merge(update2);
            node.getDecodedParameters();

            assert.equal(decodeByClassNameStub.callCount, 2);
        })
    });

    describe('#size()', function () {

        it('should get size of empty tree', function () {
            assert.equal(tree.size(), 0);
        });

        it('should get size of non empty tree', function () {
            var update = createUpdate('/MRBTS-1/RAT-1/EQM_L-1');

            tree.merge(update);

            assert.equal(tree.size(), 3);
        })
    });

    describe('#isEmpty()', function () {

        it('should return true when tree is empty', function () {
            assert.isTrue(tree.isEmpty());
        });

        it('should return false when tree is not empty', function () {
            tree.mergeNode('/MRBTS-1');

            assert.isFalse(tree.isEmpty());
        });

        it('should return true when removed node from tree', function () {
            tree.mergeNode('/MRBTS-1');
            tree.removeByDistName('/MRBTS-1');

            assert.isTrue(tree.isEmpty());
        });

        it('should return true when removed all nodes from tree', function () {
            tree.mergeNode('/MRBTS-1/RAT-1/EQM_L-1');
            tree.removeByDistName('/MRBTS-1');

            assert.isTrue(tree.isEmpty());
        });

    });

    describe('#getDecodedParameters()', function () {

        var tree;
        var decoderService;
        beforeEach(function () {
            var decoderMapping = new Map();
            decoderService = {
                decodeByClassName: function (metaUid, payload, className) {
                    return decoderMapping.get(key(metaUid, className));
                },
                addDecoderResult: function (metaUid, className, parameters) {
                    decoderMapping.set(key(metaUid, className), parameters);
                }
            };

            tree = new ImTree(decoderService);
        });

        it('should get decoded parameters by meta uid', function () {
            var metaUid = 123;
            var parameters = {};
            decoderService.addDecoderResult(metaUid, 'RAT', parameters);

            var node = tree.merge(createUpdate('/RAT-1'), metaUid);

            assert.equal(node.getDecodedParameters(), parameters);
        });

        it('should get decoded parameters by different meta uids', function () {
            var parameters1 = {};
            var parameters2 = {};
            decoderService.addDecoderResult(123, 'RAT', parameters1);
            decoderService.addDecoderResult(321, 'EQM_L', parameters2);

            var node1 = tree.merge(createUpdate('/RAT-1'), 123);
            var node2 = tree.merge(createUpdate('/RAT-1/EQM_L-1'), 321);

            assert.equal(node1.getDecodedParameters(), parameters1);
            assert.equal(node2.getDecodedParameters(), parameters2);
        });

        it('should get undefined when meta uid is invalid', function () {
            var parameters = {};
            decoderService.addDecoderResult(123, 'RAT', parameters);

            var node = tree.merge(createUpdate('/RAT-1'), 321);

            assert.isUndefined(node.getDecodedParameters());
        });

        function key(metaUid, className) {
            return metaUid + '-' + className;
        }

    });

    describe('#getTypeInfo()', function () {

        var tree;
        var decoderService;

        beforeEach(function () {
            var typeInfoMapping = new Map();
            decoderService = {
                getTypeInfo: function (metaUid, className) {
                    return typeInfoMapping.get(key(metaUid, className));
                },
                addTypeInfo: function (metaUid, className, parameters) {
                    typeInfoMapping.set(key(metaUid, className), parameters);
                }
            };

            tree = new ImTree(decoderService);
        });

        it('should get typeInfo by meta uid', function () {
            var metaUid = 123;
            var typeInfo = {};
            decoderService.addTypeInfo(metaUid, 'RAT', typeInfo);

            var node = tree.merge(createUpdate('/RAT-1'), metaUid);

            assert.equal(node.getTypeInfo(), typeInfo);
        });

        it('should get typeInfo by different meta uids', function () {
            var typeInfo1 = {};
            var typeInfo2 = {};
            decoderService.addTypeInfo(123, 'RAT', typeInfo1);
            decoderService.addTypeInfo(321, 'EQM_L', typeInfo2);

            var node1 = tree.merge(createUpdate('/RAT-1'), 123);
            var node2 = tree.merge(createUpdate('/RAT-1/EQM_L-1'), 321);

            assert.equal(node1.getTypeInfo(), typeInfo1);
            assert.equal(node2.getTypeInfo(), typeInfo2);
        });

        it('should get undefined when meta uid is invalid', function () {
            var typeInfo = {};
            decoderService.addTypeInfo(123, 'RAT', typeInfo);

            var node = tree.merge(createUpdate('/RAT-1'), 321);

            assert.isUndefined(node.getTypeInfo());
        });

        function key(metaUid, className) {
            return metaUid + '-' + className;
        }

    });

    describe('#isMissing', function () {

        it('should be true if no payload AND no parameters', function () {
            assert.isTrue(tree.mergeNode('/obj1-1', null, null).isMissing());

            assert.isFalse(tree.mergeNode('/obj1-1', {foo: 'bar'}, null).isMissing());
            assert.isFalse(tree.mergeNode('/obj1-1', null, new ArrayBuffer(0)).isMissing());
        });
    });
});
