const loggers = require('../src/logger.impl');

describe('logging', function () {
    'use strict';

    const assert = chai.assert;

    describe('storage appender', function () {

        var logger, loggerName;

        beforeEach(function () {
            loggerName = 'imc.lib.test.' + new Date().valueOf().toString() + Math.random();
            logger = loggers.getLogger(loggerName);
            logger.setLevel(loggers.Level.ALL);
        });

        it('should store message', function (done) {
            logger.addAppender(loggers.newDbAppender({
                store: function (record) {
                    assert.isNumber(record.session);
                    assert.isNumber(record.timestamp);
                    assert.deepEqual(record.messages, ['a message', 'one more message']);
                    assert.equal(record.logger, loggerName);
                    assert.equal(record.level, 'DEBUG');
                    done();
                }
            }));

            logger.debug('a message', 'one more message');
        });

        it('should store error message', function (done) {
            logger.addAppender(loggers.newDbAppender({
                store: function (event) {
                    assert.equal(event.exception.name, 'Error');
                    assert.equal(event.exception.message, 'foo');
                    assert.isTrue(event.exception.stack.length > 0);
                    assert.equal(event.level, 'ERROR');
                    done();
                }
            }));

            logger.error('an error!', new Error('foo'));
        });

        it('should store undefined', function (done) {
            logger.addAppender(loggers.newDbAppender({
                store: function (record) {
                    assert.deepEqual(record.messages, ['undefined']);
                    done();
                }
            }));

            logger.info(undefined);
        });

        it('should store null', function (done) {
            logger.addAppender(loggers.newDbAppender({
                store: function (record) {
                    assert.deepEqual(record.messages, ['null']);
                    done();
                }
            }));

            logger.info(null);
        });
    });

    describe('log storage', function () {
        var storage, dbName;

        beforeEach(function () {
            dbName = 'TEST_LOGS' + new Date().valueOf().toString();
            storage = loggers.getLogStorage(dbName);
        });

        afterEach(function (done) {
            new PouchDB(dbName).destroy()
                .then(function () {
                    done();
                }).catch(done);
        });

        function storeAll(records) {
            return Promise.all(records.map(function (record) {
                return storage.store(record);
            }));
        }

        it('should save and read one record', function (done) {
            var timestamp = 123456789;
            var record = {
                timestamp: timestamp,
                level: 'ERROR',
                logger: 'test.logger',
                messages: ['a message', 'another message'],
                exception: {
                    name: 'Error',
                    message: 'error message'
                }
            };

            storage.store(record)
                .then(function () {
                    return storage.fetchByTimestamp(timestamp, timestamp);
                })
                .then(function (records) {
                    assert.deepEqual(records, [record]);
                })
                .then(done).catch(error);
        });

        it('should fetch all records', function (done) {
            var records = [
                {
                    timestamp: new Date().valueOf(),
                    level: 'INFO'
                },
                {
                    timestamp: new Date().valueOf(),
                    level: 'DEBUG'
                },
                {
                    timestamp: new Date().valueOf(),
                    level: 'WARN'
                },
                {
                    timestamp: new Date().valueOf() + 1,
                    level: 'TRACE'
                },
                {
                    timestamp: new Date().valueOf() + 1,
                    level: 'ERROR'
                }
            ];

            storeAll(records).then(function () {
                return storage.fetchAll();
            }).then(function (result) {
                assert.deepEqual(result, records);
            }).then(done).catch(error);
        });

        it('should select records by timestamp', function (done) {
            var timestamp = new Date().valueOf();
            var records = [
                {
                    timestamp: timestamp,
                    level: 'INFO'
                },
                {
                    timestamp: timestamp + 1,
                    level: 'DEBUG'
                },
                {
                    timestamp: timestamp + 2,
                    level: 'WARN'
                },
                {
                    timestamp: timestamp + 2,
                    level: 'TRACE'
                },
                {
                    timestamp: timestamp + 3,
                    level: 'TRACE'
                }
            ];

            storeAll(records).then(function () {
                return storage.fetchByTimestamp(timestamp + 1, timestamp + 2);
            }).then(function (result) {
                assert.deepEqual(result, [records[1], records[2], records[3]]);
            }).then(done).catch(error);
        });

        it('should fetch logs for current session', function (done) {
            var start = storage.getSessionStart();
            var records = [
                {
                    timestamp: start - 1,
                    level: 'DEBUG'
                },
                {
                    timestamp: start,
                    level: 'WARN'
                },
                {
                    timestamp: start + 1,
                    level: 'TRACE'
                }
            ];

            storeAll(records).then(function () {
                return storage.fetchCurrentSession();
            }).then(function (result) {
                assert.deepEqual(result, [records[1], records[2]]);
            }).then(done).catch(error);
        });

        it('should notify store listener', function (done) {
            var record = {
                timestamp: new Date().valueOf(),
                level: 'TRACE'
            };

            storage.onStore(function (records) {
                assert.deepEqual(records, [record]);
                done();
            });

            storage.store(record);
        });

        it('should notify new listener with all records from current session', function (done) {
            var start = storage.getSessionStart();
            var oldRecords = [
                {
                    timestamp: start - 1,
                    level: 'DEBUG'
                },
                {
                    timestamp: start,
                    level: 'WARN'
                },
                {
                    timestamp: start + 1,
                    level: 'TRACE'
                }
            ];

            storeAll(oldRecords).then(function () {
                storage.onStore(function (records) {
                    assert.deepEqual(records, [oldRecords[1], oldRecords[2]]);
                    done();
                });
            });
        });

        it('should delete all records', function (done) {
            storeAll([
                {
                    timestamp: new Date().valueOf(),
                    level: 'DEBUG'
                },
                {
                    timestamp: new Date().valueOf(),
                    level: 'WARN'
                },
                {
                    timestamp: new Date().valueOf(),
                    level: 'TRACE'
                }
            ]).then(function () {
                return storage.deleteAll();
            }).then(function () {
                return storage.fetchAll();
            }).then(function (result) {
                assert(result.length === 0, 'storage should be empty');
            }).then(done).catch(error);
        });

        it('should store new records after deleting all', function (done) {
            var newMessage = {
                timestamp: new Date().valueOf(),
                level: 'ERROR'
            };
            storage.store({
                timestamp: new Date().valueOf(),
                level: 'DEBUG'
            })
                .then(function () {
                    return storage.deleteAll();
                }).then(function () {
                    return storage.store(newMessage);
                }).then(function () {
                    return storage.fetchAll();
                }).then(function (result) {
                    assert.deepEqual(result, [newMessage]);
                }).then(done).catch(error);
        });
    });

    function error(e) {
        console.error(e.toString() + '\n' + e.stack);
    }
});
