const metaRepository = require('../src/metaRepository');
const metaReader = require('../src/metaReader');
const metaDefinitionFactory = require('../src/metaDefinitionFactory');

describe('MetaRepository', function () {

    var theMetaRepository;

    beforeEach(function () {
        theMetaRepository = metaRepository();
    });

    describe('set meta definitions', function () {
        var definitions = [{
            metaVersion: '1',
            objectDefinitions: {"SMOD_L": {}}
        }, {
            metaVersion: '1',
            objectDefinitions: {"RMOD_L": {}}
        }];

        it('should be empty when no definitions are set', function () {
            assert.isTrue(theMetaRepository.isEmpty());
        });

        it('should not be empty when definitions are set', function () {
            theMetaRepository.setMetaDefinitions(definitions);

            assert.isFalse(theMetaRepository.isEmpty());
        });

        it('should set meta definitions', function () {
            var uid = theMetaRepository.setMetaDefinitions(definitions);

            assert.deepEqual(theMetaRepository.getMetaDefinition(uid, "SMOD_L"), {});
        });

        it('should get null when getting not existing class name', function () {
            var uid = theMetaRepository.setMetaDefinitions(definitions);

            assert.notOk(theMetaRepository.getMetaDefinition(uid, "NOT_EXIST"));
        });

        it('should get meta definition for given uid', function () {
            var uid1 = theMetaRepository.setMetaDefinitions([{
                objectDefinitions: {'SMOD_L': 1}
            }]);
            var uid2 = theMetaRepository.setMetaDefinitions([{
                objectDefinitions: {'SMOD_L': 2}
            }]);

            assert.deepEqual(theMetaRepository.getMetaDefinition(uid1, "SMOD_L"), 1);
            assert.deepEqual(theMetaRepository.getMetaDefinition(uid2, "SMOD_L"), 2);
        });

        it('should get meta blob for given uid', function () {
            var metaBlob = new Blob();
            var uid = theMetaRepository.setMetaDefinitions(definitions, metaBlob);

            assert.equal(theMetaRepository.getMetaBufferForUid(uid), metaBlob);
        });

        it('should throw when getting meta blob for invalid uid', function () {
            assert.throws(function () {
                theMetaRepository.getMetaBufferForUid(-666);
            });
        });
    });

    describe('getTypeInfo', function () {

        it('should return typeInfo for class name', function () {
            var uid = theMetaRepository.setMetaDefinitions([{
                objectDefinitions: {'SMOD_L': 1}
            }]);

            var typeInfo = theMetaRepository.getTypeInfo(uid, "SMOD_L");

            assert.ok(typeInfo);
        });

        it('should return undefined when not exist getTypeInfo for class name', function () {
            var uid = theMetaRepository.setMetaDefinitions([{
                objectDefinitions: {'SMOD_L': 1}
            }]);

            var typeInfo = theMetaRepository.getTypeInfo(uid, "SMOD_X");

            assert.isUndefined(typeInfo);
        });

    });
        describe('getClassNamesForUid', function () {
            it('should concat class names from different definitions', function () {
                var uid1 = theMetaRepository.setMetaDefinitions([{
                    objectDefinitions: {'SMOD_L': 1, 'RMOD_L': 2}
                }, {
                    objectDefinitions: {'TRMOD_L': 3}
                }]);
                var uid2 = theMetaRepository.setMetaDefinitions([{
                    objectDefinitions: {'BBMOD_L': 1}
                }]);

                var classNamesFor1 = theMetaRepository.getClassNamesForUid(uid1);
                var classNamesFor2 = theMetaRepository.getClassNamesForUid(uid2);

                assert.lengthOf(classNamesFor1, 3);
                assert.include(classNamesFor1, 'SMOD_L');
                assert.include(classNamesFor1, 'RMOD_L');
                assert.include(classNamesFor1, 'TRMOD_L');
                assert.lengthOf(classNamesFor2, 1);
                assert.include(classNamesFor2, 'BBMOD_L');
            });

        });

    describe('areUidsOfSameMeta', function () {

        it('should return true when the same meta', function () {
            var uid1 = theMetaRepository.setMetaDefinitions({'SMOD_L': 1});
            var uid2 = theMetaRepository.setMetaDefinitions({'SMOD_L': 1});

            assert.equal(uid1, uid2);
        });

        it('should return false when not the same meta', function () {
            var uid1 = theMetaRepository.setMetaDefinitions({'SMOD_L': 1});
            var uid2 = theMetaRepository.setMetaDefinitions({'SMOD_L': 2});

            assert.notEqual(uid1, uid2);
        });
    });

    describe('reading from file', function () {

        it('should create repository from file', function (done) {
            loadBlob('base/resources/meta/meta_v17.zip', function (buffer) {
                metaReader.read(buffer, function (entries) {
                    var definitions = metaDefinitionFactory.create(entries);
                    var uid = theMetaRepository.setMetaDefinitions(definitions, buffer);

                    var typeInfo = theMetaRepository.getTypeInfo(uid, "SMOD_L");
                    assert.ok(typeInfo);

                    done();
                });
            })
        });

        function loadBlob(filePath, doneCallback) {
            var request = new XMLHttpRequest();
            request.open('GET', filePath, true);
            request.responseType = 'arraybuffer';

            request.onload = function () {
                if (this.status == 200) {
                    var buffer = request.response;
                    doneCallback(buffer);
                }
            };
            request.send();
        }
    });

});
