describe('Service: metaService', function () {

    var defaultMetaBuffer;
    var metaReader;
    var metaService;

    beforeEach(function () {
        defaultMetaBuffer = new ArrayBuffer(0);
        var metaMapping = new Map();
        metaMapping.set(defaultMetaBuffer, []);
        metaReader = {
            read: function (buffer, callback) {
                callback(metaMapping.get(buffer));
            },
            addReadResult: function (buffer, result) {
                metaMapping.set(buffer, result);
            }
        };
        metaService = infoModel.metaService({metaReader: metaReader});
        metaService.setMetaReader(metaReader);
    });

    it('should exist', function () {
        assert.isTrue(!!metaService);
    });

    describe('isReady', function () {

        it('should be false when meta was not set', function () {
            assert.isFalse(metaService.isReady());
        });

        it('should be true after setting meta', function (done) {
            metaService.setMeta(defaultMetaBuffer).then(function () {
                assert.isTrue(metaService.isReady());
                done();
            });
        });

    });

    describe('started', function () {

        it('should be false when meta was not set', function () {
            assert.isFalse(metaService.started());
        });

        it('should be false after setting meta', function () {
            metaService.setMeta(defaultMetaBuffer, function () {
                assert.isFalse(metaService.started());
            });
        });

        it('should be true while reading meta', function () {
            var originalRead = metaReader.read;
            metaReader.read = function (buffer, callback) {
                assert.isTrue(metaService.started());
                originalRead(buffer, callback);
            };

            metaService.setMeta(defaultMetaBuffer, function () {
                assert.isFalse(metaService.started());
            });
        });

    });

    describe('getLastMetaUid', function () {

        it('should throw when meta is not set', function () {
            assert.throws(function () {
                metaService.getLastMetaUid();
            });
        });

        it('should get meta uid when meta was set', function (done) {
            metaService.setMeta(defaultMetaBuffer).then(function () {
                assert.ok(metaService.getLastMetaUid());
                done();
            });
        });

    });

    describe('getMetaObject', function () {

        it('should get meta definition when meta was set', function (done) {
            var metaBuffer = new ArrayBuffer(0);
            var definition = {'@class': 'SMOD_L'};
            metaReader.addReadResult(metaBuffer, [{objects: [definition]}]);

            metaService.setMeta(metaBuffer).then(function (metaUid) {
                assert.deepEqual(metaService.getMetaObject(metaUid, "SMOD_L"), definition);
                done();
            });
        });

        it('should get meta definition for given meta uid', function (done) {
            var smodMetaBuffer = new ArrayBuffer(0);
            var smodDefinition = {'@class': 'SMOD_L'};

            var rmodMetaBuffer = new ArrayBuffer(0);
            var rmodDefinition = {'@class': 'RMOD_L'};

            metaReader.addReadResult(smodMetaBuffer, [{objects: [smodDefinition]}]);
            metaReader.addReadResult(rmodMetaBuffer, [{objects: [rmodDefinition]}]);

            Promise.join(metaService.setMeta(smodMetaBuffer), metaService.setMeta(rmodMetaBuffer))
                .then(function (result) {
                    var smodUid = result[0];
                    var rmodUid = result[1];
                    assert.deepEqual(metaService.getMetaObject(smodUid, "SMOD_L"), smodDefinition);
                    assert.deepEqual(metaService.getMetaObject(rmodUid, "RMOD_L"), rmodDefinition);
                }).then(done).catch(done);
        });

        it('should get undefined when getting not existing class name', function (done) {
            metaService.setMeta(defaultMetaBuffer).then(function (metaUid) {
                assert.notOk(metaService.getMetaObject(metaUid, "NOT_EXIST"));
                done();
            });
        });

    });
});
