describe('Service: onlineHistoryService', function () {

    var updateService;
    var onlineHistoryService;
    var assertions;

    beforeEach(function () {
        var metaService = {
            getLastMetaUid: function () {
                return 0;
            },
            addMeta: function (meta) {
                return Promise.resolve(meta.uid);
            }
        };
        updateService = infoModel.updateService({metaService: metaService});
        onlineHistoryService = adminCore.onlineHistoryService(updateService);
        onlineHistoryService.clear();
        assertions = infoModel.historyEntryAssertions;
    });

    describe('addEntry', function () {

        it('should add online history entries to buffer', function (done) {
            var entry1 = createEntryWithUpdate(0, '/A-1'),
                entry2 = createEntryWithUpdate(1, '/A-2'),
                entry3 = createEntryWithUpdate(2, '/A-3');

            onlineHistoryService.addEntry(entry1)
                .then(function () {
                    return onlineHistoryService.addEntry(entry2);
                })
                .then(function () {
                    return onlineHistoryService.addEntry(entry3);
                })
                .then(function () {
                    const buffer = onlineHistoryService.getOnlineHistoryBuffer();
                    assert.lengthOf(buffer, 3);
                    assertions.assertContainsOnlyUpdates(buffer[0], ['/A-1']);
                    assertions.assertContainsOnlyUpdates(buffer[1], ['/A-2']);
                    assertions.assertContainsOnlyUpdates(buffer[2], ['/A-3']);
                })
                .then(done)
                .catch(done);
        });

        it('should set metaUid in added entry', function (done) {
            var entry = createEntryWithUpdate(0, '/A-1');
            var meta = createMeta(1234);

            onlineHistoryService.setMeta(meta)
                .then(function () {
                    return onlineHistoryService.addEntry(entry);
                })
                .then(function () {
                    const buffer = onlineHistoryService.getOnlineHistoryBuffer();
                    assert.lengthOf(buffer, 1);
                    assert.deepEqual(buffer[0].metaUid, 1234);
                })
                .then(done)
                .catch(done);
        });

    });

    describe('setMeta', function () {

        it('should set lastMetaUid when new meta added', function (done) {
            var meta = createMeta(1234);

            onlineHistoryService.setMeta(meta)
                .then(function () {
                    assert.deepEqual(onlineHistoryService.getLastMetaUid(), 1234);
                })
                .then(done)
                .catch(done);
        });

        it('should reset lastMetaUid when another meta added', function (done) {
            var meta1 = createMeta(1234),
                meta2 = createMeta(4321);

            onlineHistoryService.setMeta(meta1)
                .then(function () {
                    assert.deepEqual(onlineHistoryService.getLastMetaUid(), 1234);
                })
                .then(function () {
                    return onlineHistoryService.setMeta(meta2);
                })
                .then(function () {
                    assert.deepEqual(onlineHistoryService.getLastMetaUid(), 4321);
                })
                .then(done)
                .catch(done);
        });

    });

    describe('clear', function () {

        it('should reset onlineHistoryBuffer', function (done) {
            var entry = createEntryWithUpdate(0, '/A-1');

            onlineHistoryService.addEntry(entry)
                .then(function () {
                    onlineHistoryService.clear();
                    assert.deepEqual(onlineHistoryService.getOnlineHistoryBuffer(), []);
                })
                .then(done)
                .catch(done);
        });

        it('should reset history blocks', function (done) {
            var entry = createEntryWithUpdate(0, '/A-1');

            onlineHistoryService.addEntry(entry)
                .then(function () {
                    onlineHistoryService.clear();
                    assert.deepEqual(onlineHistoryService.getHistoryBlocks(), []);
                })
                .then(done)
                .catch(done);
        });

        it('should reset timestampOffset internally', function () {
            onlineHistoryService.setTimestampOffset(1000);
            onlineHistoryService.clear();
            assert.deepEqual(onlineHistoryService.getTimestampOffset(), 0);
        });

        it('should reset timestampOffset in updateService', function () {
            updateService.SYNC_API.setTimestampOffset(1000);
            onlineHistoryService.clear();
            assert.deepEqual(updateService.getTimestampOffset(), 0);
        });

    });

    describe('getHistoryBlocks', function () {

        it('should set historyBlocks to [] when there is no history', function () {
            assert.deepEqual(onlineHistoryService.getHistoryBlocks(), []);
        });

    });

    describe('apply', function () {
        var historyEntries = [],
            onlineEntries = [],
            meta;

        beforeEach(function (done) {
            historyEntries = [
                createEntryWithUpdate(0, '/A-1'),
                createEntryWithUpdate(1, '/A-2'),
                createEntryWithUpdate(2, '/A-3')
            ];
            onlineEntries = [
                createEntryWithUpdate(0, '/A-4'),
                createEntryWithUpdate(1, '/A-5'),
                createEntryWithUpdate(2, '/A-6')
            ];
            meta = createMeta(1234);

            pushUpdateEntries(onlineEntries);
            onlineHistoryService.setMeta(meta)
                .then(function () {
                    return onlineHistoryService.addEntry(historyEntries[0]);
                })
                .then(function () {
                    return onlineHistoryService.addEntry(historyEntries[1]);
                })
                .then(function () {
                    return onlineHistoryService.addEntry(historyEntries[2]);
                })
                .then(done)
                .catch(done);
        });

        it('should set one history block', function (done) {
            const procedureCompleteId = 1,
                procedureSentId = 0;

            onlineHistoryService.apply(procedureCompleteId, procedureSentId)
                .then(function () {
                    assert.deepEqual(onlineHistoryService.getHistoryBlocks(), [{start: 0, end: 2}]);
                })
                .then(done)
                .catch(done);
        });

        it('should set new history history block after reset', function (done) {
            const procedureCompleteId = 5,
                procedureSentId = 4;

            pushReset();
            pushUpdateEntries([createEntryWithUpdate(4, '/A-7'), createEntryWithUpdate(5, '/A-8')]);
            onlineHistoryService.apply(procedureCompleteId, procedureSentId)
                .then(function () {
                    assert.lengthOf(updateService.SYNC_API.getHistoryEntries(), 8);
                    assert.deepEqual(onlineHistoryService.getHistoryBlocks(), [{start: 4, end: 6}]);
                })
                .then(done)
                .catch(done);
        });

        it('should drop all entries before procedureSentId', function (done) {
            const procedureCompleteId = 1,
                procedureSentId = 1;

            onlineHistoryService.apply(procedureCompleteId, procedureSentId)
                .then(function () {
                    var entries = updateService.SYNC_API.getHistoryEntries();
                    assert.lengthOf(entries, 5);
                    assertions.assertContainsOnlyUpdates(entries[0], ['/A-1']);
                    assertions.assertContainsOnlyUpdates(entries[1], ['/A-2']);
                    assertions.assertContainsOnlyUpdates(entries[2], ['/A-3']);
                    assertions.assertContainsOnlyUpdates(entries[3], ['/A-5']);
                    assertions.assertContainsOnlyUpdates(entries[4], ['/A-6']);
                })
                .then(done)
                .catch(done);
        });

        it('should merge all entries from procedureSentId to procedureCompleteId', function (done) {
            const procedureCompleteId = 2,
                procedureSentId = 1;

            onlineHistoryService.apply(procedureCompleteId, procedureSentId)
                .then(function () {
                    var entries = updateService.SYNC_API.getHistoryEntries();
                    assert.lengthOf(entries, 4);
                    assertions.assertContainsOnlyUpdates(entries[0], ['/A-1']);
                    assertions.assertContainsOnlyUpdates(entries[1], ['/A-2']);
                    assertions.assertContainsOnlyUpdates(entries[2], ['/A-3']);
                    assertions.assertContainsOnlyUpdates(entries[3], ['/A-5', '/A-6']);
                })
                .then(done)
                .catch(done);
        });

        it('should notify listeners', function (done) {
            var listener = createListener(),
                procedureCompleteId = 2,
                procedureSentId = 1;

            const expectedEvents = [
                // clear history notifications
                'clearEntries',
                'newImTree',
                'completeImTree',
                'onlineImTree',
                // online history notifications
                'newEntries',
                'newImTree',
                'completeImTree',
                'onlineImTree',
                'changeEntry'
            ];

            addListener(listener)
                .then(function () {
                    listener.clearEvents();
                    return onlineHistoryService.apply(procedureCompleteId, procedureSentId);
                })
                .then(function () {
                    assert.deepEqual(listener.$events, expectedEvents);
                })
                .then(done)
                .catch(done);
        });

    });

    function createEntryWithUpdate(id, distName) {
        return assertions.createChangeEntry(id, [distName], []);
    }

    function pushUpdateEntries(entries) {
        entries.forEach(function (entry) {
            var update = entry.change.updated[0];
            updateService.pushUpdate(update.distName, update.payload);
        });
    }

    function pushReset() {
        updateService.pushDisconnect({autoReconnect: false, source: 'BTS'});
    }

    function createMeta(metaUid) {
        return {
            uid: metaUid
        };
    }

    function createListener() {
        return {
            $events: [],
            clearEvents: function () {
                this.$events = [];
            },
            // imTreeListener part
            completeImTree: function () {
                this.$events.push('completeImTree');
            },
            newImTree: function () {
                this.$events.push('newImTree');
            },
            updateOfImTree: function () {
                this.$events.push('updateOfImTree');
            },
            onlineImTree: function () {
                this.$events.push('onlineImTree');
            },
            // historyListener part
            newEntry: function () {
                this.$events.push('newEntry');
            },
            newEntries: function () {
                this.$events.push('newEntries');
            },
            clearEntries: function () {
                this.$events.push('clearEntries');
            },
            historyMarkerWasSet: function () {
                this.$events.push('historyMarkerWasSet');
            },
            historyMarkerWasCleared: function () {
                this.$events.push('historyMarkerWasCleared');
            },
            // currentEntryListener part
            changeEntry: function () {
                this.$events.push('changeEntry');
            },
            // validationListener part
            timestampDecreasing: function () {
                this.$events.push('timestampDecreasing');
            },
            clear: function () {
                this.$events.push('clear');
            }
        };
    }

    function addListener(listener) {
        return updateService.addImTreeListener(listener)
            .then(function () {
                return updateService.addHistoryListener(listener);
            })
            .then(function () {
                return updateService.addCurrentEntryListener(listener);
            });
    }

});
