describe('Service: onlineImTree', function () {

    const DISTNAME = '/A-1';
    const META_UID = 0;
    var onlineImTree;
    var listener;

    beforeEach(function () {
        var metaService = {
            getLastMetaUid: function () {
                return META_UID;
            }
        };
        onlineImTree = adminCore.onlineImTree({metaService: metaService});
    });


    describe('onlineImTree', function () {
        var receivedEvent;
        beforeEach(function () {
            receivedEvent = null;
        });

        it('should notify imTree listener on registration if imTree is not empty', function (done) {
            pushUpdate(DISTNAME, 2);

            onlineImTree.watchUpdates(listener)
                .then(function () {
                    assertTreeContainsNode(DISTNAME);
                    assertNumberOfChanges(1);
                    assertEventHasUpdated(DISTNAME);
                    assertEventHasEmptyRemoved();
                    assertEventHasCorrectMetaUid();
                })
                .then(done).catch(done);
        });

        it('should not notify imTree listener on registration if imTree is empty', function (done) {
            onlineImTree.watchUpdates(listener)
                .then(function () {
                    assert.isNull(receivedEvent);
                })
                .then(done).catch(done);
        });

        it('should add and remove given distName from imTree', function (done) {
            onlineImTree.watchUpdates(listener)
                .then(function () {
                    pushUpdate(DISTNAME, 2);
                    assertTreeContainsNode(DISTNAME);
                    assertNumberOfChanges(1);
                    assertEventHasUpdated(DISTNAME);
                    assertEventHasEmptyRemoved();
                    assertEventHasCorrectMetaUid();

                    pushRemove(DISTNAME);
                    assertTreeIsEmpty();
                    assertNumberOfChanges(1);
                    assertEventHasEmptyUpdated();
                    assertEventHasRemoved(DISTNAME);
                    assertEventHasCorrectMetaUid();
                })
                .then(done).catch(done);
        });

        function listener(updateEvent) {
            receivedEvent = updateEvent;
        }

        it('should notify about reset', function (done) {
            onlineImTree.watchUpdates(listener)
                .then(function () {
                    pushUpdate(DISTNAME);

                    pushReset();
                    assertNumberOfChanges(0);
                    assertTreeIsEmpty();
                    assert.ok(receivedEvent.reset);
                })
                .then(done).catch(done);
        });

        it('should get imTree instantly', function (done) {
            pushUpdate(DISTNAME);
            assertImTree(done);
        });

        it('should get imTree after some time', function (done) {
            assertImTree(done);
            pushUpdate(DISTNAME);
        });

        function assertImTree(done) {
            onlineImTree.getImTree()
                .then(function (imTree) {
                    assert.isFalse(imTree.isEmpty());
                    assert.equal(imTree.findByDistName(DISTNAME).distName, DISTNAME);
                })
                .then(done).catch(done);
        }

        function pushUpdate(distName, byteSize) {
            onlineImTree.pushUpdate(distName, new Uint8Array(byteSize || 0));
        }

        function pushRemove(distName) {
            onlineImTree.pushRemove(distName, 1);
        }

        function pushReset() {
            onlineImTree.pushDisconnect({autoReconnect: false, source: "BTS"});
        }

        function assertTreeContainsNode(distName) {
            const distNames = Array.from(receivedEvent.completeImTree.distNames());
            assert.lengthOf(distNames, 1);
            assert.equal(distNames[0], distName);
        }

        function assertTreeIsEmpty() {
            assert.ok(receivedEvent.completeImTree.isEmpty());
        }

        function assertEventHasCorrectMetaUid() {
            assert.equal(receivedEvent.metaUid, META_UID);
        }

        function assertEventHasUpdated(distName) {
            const updatedDistNames = receivedEvent.change.updated;
            assert.lengthOf(updatedDistNames, 1);
            assert.equal(updatedDistNames[0], distName);
        }

        function assertEventHasRemoved(distName) {
            const removedDistNames = receivedEvent.change.removed;
            assert.lengthOf(removedDistNames, 1);
            assert.equal(removedDistNames[0], distName);
        }

        function assertEventHasEmptyUpdated() {
            assert.lengthOf(receivedEvent.change.updated, 0);
        }

        function assertEventHasEmptyRemoved() {
            assert.lengthOf(receivedEvent.change.removed, 0);
        }

        function assertNumberOfChanges(expectedNumber) {
            assert.equal(receivedEvent.numOfChanges, expectedNumber);
        }
    });
});
