'use strict';
const sic = require('../../src/sic');
const TranslatingBtsClient = require('../../src/communication/translatingBtsClient');
const ProtoByteBuffer = require('protobufjs').ByteBuffer;


describe('executionRequestDispatcher', function () {

    var metaServiceStub, decoderServiceStub, socket, executionRequestsDispatcher;
    var SNAPSHOT_REQ_DIST_NAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/TECHREP_L-1/SNAPSHOT_REQ-0';
    var TECHREP_DIST_NAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/TECHREP_L-1';
    var EXECUTOR = _.noop;
    var META_UID = 1;
    var SNAPSHOT_REQ_PARAMS = {
        stateInfo: {},
        oper_accepted: {
            description: "Status ok"
        }
    };
    var EXECUTION_REQUEST_WITH_SNAPSHOT_DISTNAME;
    const operationTypes = adminCore.operations.operationTypes;

    function createMessage(operationType) {

        const message = socket.getSicInstance().executionRequest();
        var objectT = {
            object_info: socket.resolveObjectInfo(SNAPSHOT_REQ_DIST_NAME),
            payload: new ProtoByteBuffer()
        };

        var operationT = {
            operation_type: operationType.value,
            object: objectT
        };
        var operations = [operationT];
        message.setOperations(operations);
        return message;
    }

    beforeEach(function () {
            decoderServiceStub = createDecoderService();

            const typeInfos = new Map();
            typeInfos.set('SNAPSHOT_REQ', createSnapshotReqTypeInfo());

            metaServiceStub = createMetaService();
            socket = createTranslatingBtsClient();
            initMapper();

            executionRequestsDispatcher = adminCore.operations.executionRequestsDispatcher({
                translatingBtsClient: socket,
                metaService: metaServiceStub,
                decoderService: decoderServiceStub
            });

            EXECUTION_REQUEST_WITH_SNAPSHOT_DISTNAME = createMessage(operationTypes.CREATE);

            function createMetaService() {
                return {
                    getLastMetaUid: function () {
                        return META_UID;
                    },
                    getTypeInfo: function (metaUid, className) {
                        return typeInfos.get(className);
                    }
                }
            }

            function createDecoderService() {
                return {
                    decodeByDistName: function (metaUid, payload, distName) {
                        return SNAPSHOT_REQ_PARAMS;
                    }
                }
            }

            function createTranslatingBtsClient() {
                var listener;

                return new Proxy(TranslatingBtsClient.newInstance({}), {
                    get: function (target, property, receiver) {
                        if (property === 'addMessageListener') return addMessageListener;
                        if (property === 'receive') return receive;

                        return target[property] || receiver[property];
                    }
                });

                function addMessageListener(msgId, msgListener) {
                    listener = msgListener;
                    return {
                        unwatch: function () {
                            listener = null;
                        }
                    };
                }

                function receive(message) {
                    listener(message);
                }
            }

            function initMapper() {
                const mapper = socket.getMapper();

                mapper.addClassName(0, 'MRBTS');
                mapper.addClassName(1, 'RAT');
                mapper.addClassName(2, 'BTS_L');
                mapper.addClassName(3, 'FUNC_L');
                mapper.addClassName(4, 'TECHREP_L');
                mapper.addClassName(5, 'SNAPSHOT_REQ');

                mapper.registerObjectInfos([
                    createObjectInfo('00', 0, 1),
                    createObjectInfo('01', 1, 1, '00'),
                    createObjectInfo('02', 2, 1, '01'),
                    createObjectInfo('03', 3, 1, '02'),
                    createObjectInfo('04', 4, 1, '03'),
                    createObjectInfo('05', 5, 0, '04')
                ]);

                function createObjectInfo(uuid, objectType, instanceId, parentUuid) {
                    return {
                        uuid,
                        objectType,
                        instanceId,
                        parentUuid
                    }
                }
            }

            function createSnapshotReqTypeInfo() {
                const operations = [{
                    getName: function () {
                        return 'oper_rejected';
                    }
                }, {
                    getName: function () {
                        return 'oper_accepted';
                    }
                }, {
                    getName: function () {
                        return 'oper_failed';
                    }
                }];
                return {
                    getOperations: function () {
                        return operations;
                    }
                }
            }
        }
    );

    afterEach(function () {
        executionRequestsDispatcher.removeExecutorForDistName(SNAPSHOT_REQ_DIST_NAME);
    });

    describe('registerExecutorForDistName', function () {
        it('should register executor in dispatcher correctly', function () {

            executionRequestsDispatcher.registerExecutorForDistName(SNAPSHOT_REQ_DIST_NAME, EXECUTOR);
            assert.isTrue(executionRequestsDispatcher.isListenerRegisteredForDistName(SNAPSHOT_REQ_DIST_NAME));
        });
    });

    describe('removeExecutorForDistName', function () {
        it('should register executor in dispatcher correctly', function () {

            executionRequestsDispatcher.registerExecutorForDistName(SNAPSHOT_REQ_DIST_NAME, EXECUTOR);
            executionRequestsDispatcher.removeExecutorForDistName(SNAPSHOT_REQ_DIST_NAME, EXECUTOR);
            assert.isFalse(executionRequestsDispatcher.isListenerRegisteredForDistName(SNAPSHOT_REQ_DIST_NAME));
        });
    });

    it('should provide event to correct executor', function () {
        var wasCorrectExecutorNotified = false;

        executionRequestsDispatcher.registerExecutorForDistName(SNAPSHOT_REQ_DIST_NAME, snapshotExecutor);
        executionRequestsDispatcher.registerExecutorForDistName(TECHREP_DIST_NAME, techrepExecutor);

        socket.receive(EXECUTION_REQUEST_WITH_SNAPSHOT_DISTNAME);

        assert.ok(wasCorrectExecutorNotified);

        function snapshotExecutor(event) {
            wasCorrectExecutorNotified = event;
        }

        function techrepExecutor(event) {
            wasCorrectExecutorNotified = false;
        }
    });

    it('should provide event with correct structure for EXECUTE operation request', function () {
        function snapshotExecutor(event) {
            assert.strictEqual(event.distName, SNAPSHOT_REQ_DIST_NAME);
            assert.strictEqual(event.type, operationTypes.EXECUTE);
            assert.strictEqual(event.operationName, 'oper_accepted');
            assert.deepEqual(event.operationParameters, {
                description: "Status ok"
            });
        }

        executionRequestsDispatcher.registerExecutorForDistName(SNAPSHOT_REQ_DIST_NAME, snapshotExecutor);
        socket.receive(createMessage(operationTypes.EXECUTE));
    });


    it('should provide event with correct structure for non EXECUTE operation request', function () {
        function snapshotExecutor(event) {
            assert.strictEqual(event.distName, SNAPSHOT_REQ_DIST_NAME);
            assert.strictEqual(event.type, operationTypes.CREATE);
            assert.strictEqual(event.object, SNAPSHOT_REQ_PARAMS);
        }

        executionRequestsDispatcher.registerExecutorForDistName(SNAPSHOT_REQ_DIST_NAME, snapshotExecutor);
        socket.receive(EXECUTION_REQUEST_WITH_SNAPSHOT_DISTNAME);
    });
});
