const testUtil = require('../test.util');
const OperationExecutor = require('../../src/operations/operationExecutor');
const TranslatingBtsClient = require('../../src/communication/translatingBtsClient');
const ProtoByteBuffer = require('protobufjs').ByteBuffer;


describe('OperationExecutor', function () {
    var operationExecutor;
    var EXECUTE = 3;
    var operation;
    var JSON_PATH = 'base/resources/meta/mrbts_simple.json';
    var managedObject = {
        distName: '/MRBTS-1',
        parameters: {
            siteId: '4'
        },
        getDecodedParameters: function () {
            return this.parameters;
        }
    };
    var defaultObjectInfo = {
        uuid: '011F',
        objectType: 0,
        instanceId: 1

    };

    beforeEach(function () {
        operationExecutor = OperationExecutor();
        operationExecutor.reset();
        operation = {
            name: 'setSiteId',
            parameters: {
                siteId: '1'
            },
            managedObject: managedObject,
            type: EXECUTE
        };
    });

    function createTranslatingBtsClient(executionCode, messageVerifyFunction) {

        var translatingBtsClient = TranslatingBtsClient.newInstance();
        var listener;
        translatingBtsClient.sendChangeRequest = sendChangeRequest;
        translatingBtsClient.addListener = addListener;
        translatingBtsClient.removeListener = removeListener;
        translatingBtsClient.getListener = getListener;

        return translatingBtsClient;
        function sendChangeRequest(msg) {
            if (messageVerifyFunction) {
                messageVerifyFunction(msg);
            }
            setTimeout(function () {
                listener.operation.callbackFunction(executionCode, msg.getRequestId());
            }, 50);
        }

        function addListener(operationListener) {
            listener = operationListener;
        }

        function removeListener() {
            listener = null;
        }

        function getListener() {
            return listener;
        }
    }

    function createExecuteOperationFunction(translatingBtsClient) {
        return (operation) => {
            var operationExecutor = OperationExecutor();
            translatingBtsClient.getMapper().registerObjectInfos([defaultObjectInfo]);
            operationExecutor.setTranslatingBtsClient(translatingBtsClient);
            return operationExecutor.executeOperation(operation);
        };
    }

    function createRegisterTypesFunction(translatingBtsClient) {
        return (typeInfo) => {
            translatingBtsClient.getMapper().addClassName(typeInfo.getId(), typeInfo.getClassName());
            return typeInfo;
        };
    }

    function customizeOperation(typeInfo) {
        operation.typeInfo = typeInfo;
        return operation;
    }

    function customizeOperationUsingIncorrectParameter(typeInfo) {
        operation.typeInfo = typeInfo;
        operation.parameters = {badParam: 1};
        return operation;
    }


    describe('executeOperation', function () {

        it('should not change managed object', function (done) {
            var translatingBtsClient = createTranslatingBtsClient(0);
            loadTypeInfo(JSON_PATH)
                .then(createRegisterTypesFunction(translatingBtsClient))
                .then(customizeOperation)
                .then(createExecuteOperationFunction(translatingBtsClient))
                .then(verifyResult)
                .then(done)
                .catch(done);


            function verifyResult(result) {
                assert.strictEqual(result.executionStatus, 0);
                assert.strictEqual(result.requestId, 1);
                assert.strictEqual(managedObject.parameters['setSiteId'], undefined);
            }
        });
    });

    it('should perform resolve when send ok and execution status is executed', function (done) {
        var translatingBtsClient = createTranslatingBtsClient(0);

        loadTypeInfo(JSON_PATH)
            .then(createRegisterTypesFunction(translatingBtsClient))
            .then(customizeOperation)
            .then(createExecuteOperationFunction(translatingBtsClient))
            .then(resolveAsSuccess)
            .then(done)
            .catch(done);
    });

    it('should perform reject when send ok and execution status is different then executed', function (done) {
        var translatingBtsClient = createTranslatingBtsClient(1);

        loadTypeInfo(JSON_PATH)
            .then(createRegisterTypesFunction(translatingBtsClient))
            .then(customizeOperation)
            .then(createExecuteOperationFunction(translatingBtsClient))
            .then(resolveAsFailure)
            .catch(rejectAsSuccess)
            .then(done)
            .catch(done);
    });

    it('should send message with correct format', function (done) {
        var translatingBtsClient = createTranslatingBtsClient(0, verifyMessage);

        loadTypeInfo(JSON_PATH)
            .then(createRegisterTypesFunction(translatingBtsClient))
            .then(customizeOperation)
            .then(createExecuteOperationFunction(translatingBtsClient))
            .then(resolveAsSuccess)
            .then(() => done())
            .catch(done);

        function verifyMessage(msg) {
            assert.strictEqual(msg.getRequestId(), 1);
            assert.strictEqual(msg.getOperations().length, 1);
            var operation = msg.getOperations()[0];

            assert.strictEqual(operation.operation_type, 3);
            assert.strictEqual(operation.object.object_info.object_create_info.instance_id, 1);
            assert.deepEqual(operation.object.object_info.object_id, ProtoByteBuffer.fromHex('011F'));
            assert.strictEqual(operation.object.object_info.object_type, 0);
            assert.instanceOf(operation.object.payload, ArrayBuffer);
        }
    });

    it('should remove listener when exception during sending message occurs', function (done) {
        var translatingBtsClient = failingBtsClient();

        loadTypeInfo(JSON_PATH)
            .then(createRegisterTypesFunction(translatingBtsClient))
            .then(customizeOperation)
            .then(createExecuteOperationFunction(translatingBtsClient))
            .then(resolveAsFailure)
            .catch(() => {
                assert.isNull(translatingBtsClient.getListener());
            }).then(done).catch(done);

        function failingBtsClient() {
            var socket = createTranslatingBtsClient(0);

            function sendChangeRequest(msg) {
                throw {};
            }

            socket.sendChangeRequest = sendChangeRequest;
            return socket;
        }
    });

    function resolveAsSuccess(result) {
        assert.strictEqual(result.executionStatus, 0);
        assert.strictEqual(result.requestId, 1);
    }

    function resolveAsFailure() {
        assert.notOk(true);
    }

    function rejectAsSuccess(result) {
        assert.strictEqual(result.executionStatus, 1);
        assert.strictEqual(result.requestId, 1);
    }

    function loadTypeInfo(jsonPath) {
        return testUtil.loadTypeInfoV2(jsonPath);
    }
});
