const _ = require('lodash');
const TranslatingBtsClient = require('../../src/communication/translatingBtsClient');
const onlineImTree = require('../../src/onlineImTree');
const operationsService = require('../../src/operations/operationsService');
const operationTypes = require('../../src/operations/operationTypes');

describe('operationsService', () => {
    'use strict';

    let metaServiceStub, operationExecutorStub, executionRequestsDispatcherStub, onlineImTreeInst,
        executionRequestsDispatcher, metaService, operationsServiceInst;
    const CLASS_NAME = 'SNAPSHOT_REQ';
    const DIST_NAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/TECHREP_L-1/SNAPSHOT_REQ-0';
    const NOT_EXISTING_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/TECHREP_L-1/SNAPSHOT_L-0';
    const META_UID = 1;
    const ERROR_MESSAGE = `Managed object with distName: ${NOT_EXISTING_DISTNAME} doesn't exist`;
    const EXECUTOR = _.noop;
    const OPERATION_NAME = 'oper_start';

    const typeInfo = mockTypeInfo();
    const OPERATION_PARAMETERS = {siteId: 3};
    const EMPTY_PARAMETERS = {};
    const TREE_NODE_PARAMETERS = {
        'id': 0,
        'coverage': 'Minor',
        'reason': 'NDT'
    };
    const SNAPSHOT_REQ_TYPE_ID = 5;

    const managedObject = {
        distName: DIST_NAME,
        payload: null
    };

    const SOCKET_DATA = {
        managedObjects: [managedObject],
        removedDistNames: [],
        metaUid: META_UID
    };

    function mockTypeInfo() {
        const operations = new Map();
        operations.set(OPERATION_NAME, {});

        return {
            name: CLASS_NAME,
            getClassName: () => {
                return CLASS_NAME;
            },
            getOperationByName: (name) => {
                return operations.get(name);
            },
            getId: () => {
                return SNAPSHOT_REQ_TYPE_ID;
            }
        };
    }

    beforeEach(() => {

        metaServiceStub = {
            getTypeInfo() {
                return typeInfo;
            },
            getLastMetaUid() {
                return META_UID;
            }
        };

        operationExecutorStub = () => {
            function executeOperation(operation) {
                return new Promise((resolve) => resolve(operation));
            }

            return {
                executeOperation,
                setTranslatingBtsClient: _.noop
            };
        };

        executionRequestsDispatcherStub = function () {
            let listener;

            function registerExecutorForDistName(distName, executor) {
                listener = executor;
            }

            function removeExecutorForDistName() {
                listener = null;
            }

            function hasRegisteredListener() {
                return listener;
            }

            return {
                registerExecutorForDistName,
                removeExecutorForDistName,
                hasRegisteredListener
            };
        };

        metaService = metaServiceStub;
        executionRequestsDispatcher = executionRequestsDispatcherStub();

        const translatingBtsClient = TranslatingBtsClient.newInstance();
        initMapper(translatingBtsClient);
        onlineImTreeInst = onlineImTree({
            socket: translatingBtsClient,
            metaService
        });
        operationsServiceInst = operationsService({
            operationExecutor: operationExecutorStub,
            translatingBtsClient,
            metaService,
            executionRequestsDispatcher,
            onlineImTree: onlineImTreeInst
        });
        onlineImTreeInst.pushData(SOCKET_DATA);
    });

    function initMapper(socket) {
        const mapper = socket.getMapper();
        mapper.addClassName(0, 'MRBTS');
        mapper.addClassName(1, 'RAT');
        mapper.addClassName(2, 'BTS_L');
        mapper.addClassName(3, 'FUNC_L');
        mapper.addClassName(4, 'TECHREP_L');
        mapper.addClassName(SNAPSHOT_REQ_TYPE_ID, 'SNAPSHOT_REQ');

        mapper.registerObjectInfos([
            createObjectInfo('00', 0, 1),
            createObjectInfo('01', 1, 1, '00'),
            createObjectInfo('02', 2, 1, '01'),
            createObjectInfo('03', 3, 1, '02'),
            createObjectInfo('04', 4, 1, '03'),
            createObjectInfo('05', 5, 0, '04')
        ]);

        function createObjectInfo(uuid, objectType, instanceId, parentUuid) {
            return {
                uuid,
                objectType,
                instanceId,
                parentUuid
            };
        }
    }

    function validateError(error) {
        assert.strictEqual(error, ERROR_MESSAGE);
    }

    function incorrectResult() {
        assert.fail(true);
    }

    describe('choose operation name', () => {
        it('should choose correct operation', () => {
            const names = ['something', 'else', OPERATION_NAME, 'different'];
            const operationName = operationsServiceInst.chooseOperationName(typeInfo, names);
            assert.equal(operationName, OPERATION_NAME);
        });

        it('should return the same name if it is correct', () => {
            const operationName = operationsServiceInst.chooseOperationName(typeInfo, OPERATION_NAME);
            assert.equal(operationName, OPERATION_NAME);
        });

        it('should throw when none of names is correct', () => {
            const names = ['something', 'else', 'different'];
            expect(() => {
                operationsServiceInst.chooseOperationName(typeInfo, names);
            }).to.throw();
        });

        it('should throw when name is incorrect', () => {
            expect(() => {
                operationsServiceInst.chooseOperationName(typeInfo, 'my_oper');
            }).to.throw();
        });

    });

    describe('executeOperation', () => {
        it('should create correct operation', (done) => {
            operationsServiceInst.executeOperation(OPERATION_NAME, DIST_NAME, OPERATION_PARAMETERS)
                .then(verifyOperation)
                .done(done);

            function verifyOperation(operation) {
                assert.strictEqual(operation.name, OPERATION_NAME);
                assert.strictEqual(operation.parameters, OPERATION_PARAMETERS);
                assert.strictEqual(operation.managedObject.distName, DIST_NAME);
                assert.deepEqual(operation.managedObject.getDecodedParameters(), EMPTY_PARAMETERS);
                assert.strictEqual(operation.typeInfo, typeInfo);
                assert.strictEqual(operation.type, operationTypes.EXECUTE.value);
            }
        });

        it('should reject when managed object not exists', (done) => {
            operationsServiceInst.executeOperation(OPERATION_NAME, NOT_EXISTING_DISTNAME, OPERATION_PARAMETERS)
                .then(incorrectResult)
                .catch(validateError)
                .done(done);
        });
    });

    describe('createObject', () => {
        it('should not register execution listener', (done) => {
            operationsServiceInst.createObject(DIST_NAME, TREE_NODE_PARAMETERS, EXECUTOR).then(verifyListenerRegistration).done(done);

            function verifyListenerRegistration() {
                assert.notOk(executionRequestsDispatcher.hasRegisteredListener());
            }
        });
    });

    describe('createObjectAsOwner', () => {
        it('should register execution listener', (done) => {
            operationsServiceInst.createObjectAsOwner(NOT_EXISTING_DISTNAME, TREE_NODE_PARAMETERS, EXECUTOR).then(verifyListenerRegistration).done(clear);

            function verifyListenerRegistration() {
                assert.ok(executionRequestsDispatcher.hasRegisteredListener());
            }

            function clear() {
                executionRequestsDispatcher.removeExecutorForDistName(NOT_EXISTING_DISTNAME);
                done();
            }
        });
    });

    describe('createOperation', () => {
        it('should create correct operation', () => {
            const operationType = operationTypes.CREATE;

            const createdOperation = operationsServiceInst.createOperation(DIST_NAME, TREE_NODE_PARAMETERS, operationType);
            verifyOperation(createdOperation);

            function verifyOperation(operation) {
                assert.isUndefined(operation.name);
                assert.isUndefined(operation.parameters);
                assert.strictEqual(operation.managedObject.metaUid, META_UID);
                assert.strictEqual(operation.managedObject.className, CLASS_NAME);
                assert.strictEqual(operation.managedObject.distName, DIST_NAME);
                assert.strictEqual(operation.managedObject.getDecodedParameters(), TREE_NODE_PARAMETERS);
                assert.strictEqual(operation.typeInfo, typeInfo);
                assert.strictEqual(operation.type, operationType.value);
            }
        });
    });

    describe('updateObject', () => {
        const CHANGED_PARAMETERS = {
            'id': 1,
            'coverage': 'Major',
            'reason': 'IMPORTANT'
        };

        it('should create operation with correct operation type', (done) => {
            operationsServiceInst.updateObject(DIST_NAME, CHANGED_PARAMETERS).then(verifyOperation).done(done);

            function verifyOperation(operation) {
                assert.isUndefined(operation.name);
                assert.isUndefined(operation.parameters);
                assert.strictEqual(operation.managedObject.metaUid, META_UID);
                assert.strictEqual(operation.managedObject.className, CLASS_NAME);
                assert.strictEqual(operation.managedObject.distName, DIST_NAME);
                assert.deepEqual(operation.managedObject.getDecodedParameters(), CHANGED_PARAMETERS);
                assert.strictEqual(operation.typeInfo, typeInfo);
                assert.strictEqual(operation.type, operationTypes.UPDATE.value);
            }
        });

        it('should reject when managed object not exists', (done) => {
            operationsServiceInst.updateObject(NOT_EXISTING_DISTNAME, CHANGED_PARAMETERS).then(incorrectResult).catch(validateError).done(done);
        });
    });

    describe('deleteObjectOperation', () => {
        it('should create operation with correct operation type', (done) => {
            operationsServiceInst.deleteObjectOperation(DIST_NAME).then(verifyOperation).done(done);

            function verifyOperation(operation) {
                assert.isUndefined(operation.name);
                assert.isUndefined(operation.parameters);
                assert.strictEqual(operation.managedObject.metaUid, META_UID);
                assert.strictEqual(operation.managedObject.className, CLASS_NAME);
                assert.strictEqual(operation.managedObject.distName, DIST_NAME);
                assert.deepEqual(operation.managedObject.getDecodedParameters(), EMPTY_PARAMETERS);
                assert.strictEqual(operation.typeInfo, typeInfo);
                assert.strictEqual(operation.type, operationTypes.DELETE.value);
            }
        });
    });

    describe('deleteObject', () => {
        it('should reject when managed object not exists', (done) => {
            operationsServiceInst.deleteObject(NOT_EXISTING_DISTNAME).then(incorrectResult).catch(validateError).done(done);
        });

        it('should unregister execution listener', (done) => {
            executionRequestsDispatcher.registerExecutorForDistName(DIST_NAME, EXECUTOR);
            operationsServiceInst.deleteObject(DIST_NAME).then(verifyListenerUnregistration).catch(validateError).done(done);
            function verifyListenerUnregistration() {
                assert.ok(!executionRequestsDispatcher.hasRegisteredListener());
            }
        });
    });
});
