const RmodsAvailabilityResponder = require('../../../src/procedures/availabilityResponders/rmodsAvailabilityResponder');
const ImTracker = require('../../../src/imTracker/imTracker');
const ImTree = require('../../../src/imtree');

describe('Service: rmodsAvailabilityResponder', function () {
    'use strict';
    var imTracker, rmodsAvailabilityResponder, rmodsListener;

    beforeEach(function () {
        imTracker = ImTracker();
        rmodsAvailabilityResponder = RmodsAvailabilityResponder(imTracker);
    });

    describe('listening on radios tests', function () {
        const RADIO_STATE = {
            RESETTING: 'resetting',
            BLOCKED: 'Blocked',
            UNBLOCKED: 'Unblocked'
        };
        const RADIO_UNIT = {
            FIRST: {
                RU_DISTNAME: '/MRBTS-1/RAT-1/BTS_L-1/EQM_L-1/RMOD_L-1/RU_L-1',
                RMOD_DISTNAME: '/MRBTS-1/RAT-1/BTS_L-1/EQM_L-1/RMOD_L-1',
                SERIAL_NUMBER: 'XXX'
            },
            SECOND: {
                RU_DISTNAME: '/MRBTS-1/RAT-1/BTS_L-1/EQM_L-1/RMOD_L-2/RU_L-1',
                RMOD_DISTNAME: '/MRBTS-1/RAT-1/BTS_L-1/EQM_L-1/RMOD_L-2',
                SERIAL_NUMBER: 'YYY'
            }
        };
        var imTree;

        beforeEach(function () {
            imTree = new ImTree();
            rmodsListener = rmodsAvailabilityResponder
                .register({
                    update: function () {
                    }
                });
        });

        it('should radio be in resetting state ', function (done) {
            const radioUnit = RADIO_UNIT.FIRST;
            update(radioUnit, RADIO_STATE.UNBLOCKED, true);
            rmodsListener
                .then(function () {
                    assert.ok(rmodsAvailabilityResponder.isRmodResetting(radioUnit.RU_DISTNAME));
                })
                .then(done).catch(done);
        });

        it('should radio be in unblocked state ', function (done) {
            const radioUnit = RADIO_UNIT.FIRST;
            update(radioUnit, RADIO_STATE.UNBLOCKED);
            rmodsListener
                .then(function () {
                    assert.ok(rmodsAvailabilityResponder.isRmodUnblocked(radioUnit.RU_DISTNAME));
                })
                .then(done).catch(done);
        });

        it('should radio be in blocked state ', function (done) {
            const radioUnit = RADIO_UNIT.FIRST;
            update(radioUnit, RADIO_STATE.BLOCKED);
            rmodsListener
                .then(function () {
                    assert.ok(rmodsAvailabilityResponder.isRmodBlocked(radioUnit.RU_DISTNAME));
                })
                .then(done).catch(done);
        });

        it('should execute listener callback for one radio after update', function (done) {
            const radioUnit = RADIO_UNIT.FIRST;
            update(radioUnit, RADIO_STATE.BLOCKED);
            rmodsListener
                .then(function () {
                    assert.ok(rmodsAvailabilityResponder.isRmodPresent(radioUnit.RU_DISTNAME));
                    assert.notOk(rmodsAvailabilityResponder.isRmodUnblocked(radioUnit.RU_DISTNAME));
                    assert.notOk(rmodsAvailabilityResponder.isRmodResetting(radioUnit.RU_DISTNAME));
                })
                .then(done).catch(done);
        });

        it('should execute listener callback for one radio after update and remove', function (done) {
            const radioUnit = RADIO_UNIT.FIRST;
            update(radioUnit, RADIO_STATE.BLOCKED);
            remove(radioUnit);
            rmodsListener
                .then(function () {
                    assert.notOk(rmodsAvailabilityResponder.isRmodPresent(radioUnit.RU_DISTNAME));
                })
                .then(done).catch(done);
        });

        it('should execute listener callback for zero radio after reset', function (done) {
            const radioUnit = RADIO_UNIT.FIRST;
            update(radioUnit, RADIO_STATE.BLOCKED);
            reset();
            rmodsListener
                .then(function () {
                    assert.notOk(rmodsAvailabilityResponder.isRmodPresent(radioUnit.RU_DISTNAME));
                })
                .then(done).catch(done);

        });

        it('should execute listener callback for two radios after update', function (done) {
            const radioUnit = RADIO_UNIT.FIRST,
                secondRadioUnit = RADIO_UNIT.SECOND;
            update(radioUnit, RADIO_STATE.BLOCKED);
            update(secondRadioUnit, RADIO_STATE.BLOCKED);
            rmodsListener
                .then(function () {
                    assert.ok(rmodsAvailabilityResponder.isRmodPresent(radioUnit.RU_DISTNAME));
                    assert.notOk(rmodsAvailabilityResponder.isRmodUnblocked(radioUnit.RU_DISTNAME));
                    assert.notOk(rmodsAvailabilityResponder.isRmodResetting(radioUnit.RU_DISTNAME));
                    assert.ok(rmodsAvailabilityResponder.isRmodPresent(secondRadioUnit.RU_DISTNAME));
                    assert.notOk(rmodsAvailabilityResponder.isRmodUnblocked(secondRadioUnit.RU_DISTNAME));
                    assert.notOk(rmodsAvailabilityResponder.isRmodResetting(secondRadioUnit.RU_DISTNAME));
                })
                .then(done).catch(done);
        });

        it('should execute listener callback if two radios were added and one was removed', function (done) {
            const radioUnit = RADIO_UNIT.FIRST,
                secondRadioUnit = RADIO_UNIT.SECOND;
            update(radioUnit, RADIO_STATE.BLOCKED);
            update(secondRadioUnit, RADIO_STATE.BLOCKED);
            remove(secondRadioUnit);
            rmodsListener
                .then(function () {
                    assert.ok(rmodsAvailabilityResponder.isRmodPresent(radioUnit.RU_DISTNAME));
                    assert.notOk(rmodsAvailabilityResponder.isRmodUnblocked(radioUnit.RU_DISTNAME));
                    assert.notOk(rmodsAvailabilityResponder.isRmodResetting(radioUnit.RU_DISTNAME));
                    assert.notOk(rmodsAvailabilityResponder.isRmodPresent(secondRadioUnit.RU_DISTNAME));
                })
                .then(done).catch(done);
        });

        function update(radioUnit, blockingState, isResetting) {
            const serialNumber = radioUnit.SERIAL_NUMBER;
            const parameters = createRuParameters(serialNumber, blockingState, isResetting);
            imTree.mergeNode(radioUnit.RU_DISTNAME, parameters);
            const updateEvent = {
                completeImTree: imTree,
                change: {
                    updated: [radioUnit.RU_DISTNAME]
                }
            };
            imTracker.imUpdate(updateEvent);
        }

        function remove(radioUnit) {
            imTree.removeByDistName(radioUnit.RU_DISTNAME);
            const updateEvent = {
                completeImTree: imTree
            };
            imTracker.imUpdate(updateEvent);
        }

        function reset() {
            imTree = new ImTree();
            const updateEvent = {
                completeImTree: imTree
            };
            imTracker.imUpdate(updateEvent);
        }

        function createRuParameters(serialNumber, blockingState, isResetting) {
            return {
                serialNumber: serialNumber,
                stateInfo: {
                    blockingState: blockingState,
                    proceduralState: isResetting ? RADIO_STATE.RESETTING : ''
                }
            };
        }
    });
});
