(function () {
    'use strict';

    const SiteAvailabilityResponder = require('../../../src/procedures/availabilityResponders/siteAvailabilityResponder');
    const ImTracker = require('../../../src/imTracker/imTracker');
    const ImTree = require('../../../src/imtree');
    const ResetRequestTracker = require('../../../src/procedures/resetRequestTracker');

    describe('SiteAvailabilityResponder', function () {

        var imTracker, siteAvailabilityResponder, siteListener;
        const LNBTS_STATE = {
            BLOCKED: 'Blocked',
            UNBLOCKED: 'Unblocked',
            ENABLED: 'Enabled',
            DISABLED: 'Disabled'
        };
        const LNBTS = {distName: '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1'};
        var imTree, resetRequestTracker;

        beforeEach(function () {
            imTracker = ImTracker();
            resetRequestTracker = ResetRequestTracker.newInstance();
            siteAvailabilityResponder = SiteAvailabilityResponder(imTracker, resetRequestTracker);
        });

        describe('listening on site tests', function () {
            beforeEach(function () {
                imTree = new ImTree();
                siteListener = siteAvailabilityResponder
                    .register({
                        update: function () {
                        }
                    });
                update();
            });

            it('should site be in unblocked state', function (done) {
                siteListener
                    .then(function () {
                        update(LNBTS_STATE.UNBLOCKED, LNBTS_STATE.ENABLED);
                    })
                    .then(function () {
                        assert.ok(siteAvailabilityResponder.isSiteUnblocked());
                    })
                    .then(done)
                    .catch(done);
            });

            it('should site be in blocked state', function (done) {
                siteListener
                    .then(function () {
                        update(LNBTS_STATE.BLOCKED, LNBTS_STATE.ENABLED);
                    })
                    .then(function () {
                        assert.ok(siteAvailabilityResponder.isSiteBlocked());
                    })
                    .then(done)
                    .catch(done);
            });

            it('should site be in operational state', function (done) {
                siteListener
                    .then(function () {
                        update(LNBTS_STATE.UNBLOCKED, LNBTS_STATE.ENABLED);
                    })
                    .then(function () {
                        assert.ok(siteAvailabilityResponder.isBtsOperational());
                    })
                    .then(done)
                    .catch(done);
            });

            it('should site not be in operational state', function (done) {
                siteListener
                    .then(function () {
                        update(LNBTS_STATE.BLOCKED, LNBTS_STATE.DISABLED);
                    })
                    .then(function () {
                        assert.notOk(siteAvailabilityResponder.isBtsOperational());
                    })
                    .then(done)
                    .catch(done);
            });
        });

        describe('on LNBTS removed', function () {
            it('isSiteUnblocked should return false', function (done) {
                siteListener
                    .then(function () {
                        update(LNBTS_STATE.UNBLOCKED, LNBTS_STATE.RESETTING, LNBTS_STATE.ENABLED);
                        remove(LNBTS);
                    })
                    .then(function () {
                        assert.notOk(siteAvailabilityResponder.isSiteUnblocked());
                    })
                    .then(done)
                    .catch(done);
            });

            it('isSiteBlocked should return false', function (done) {
                siteListener
                    .then(function () {
                        update(LNBTS_STATE.BLOCKED, LNBTS_STATE.ENABLED);
                        remove(LNBTS);
                    })
                    .then(function () {
                        assert.notOk(siteAvailabilityResponder.isSiteBlocked());
                    })
                    .then(done)
                    .catch(done);
            });

            it('isBtsOperational should return false', function (done) {
                siteListener
                    .then(function () {
                        update(LNBTS_STATE.UNBLOCKED, LNBTS_STATE.RESETTING, LNBTS_STATE.ENABLED);
                        remove(LNBTS);
                    })
                    .then(function () {
                        assert.notOk(siteAvailabilityResponder.isBtsOperational());
                    })
                    .then(done)
                    .catch(done);
            });
        });


        describe('on reset', function () {
            it('isSiteUnblocked should return false', function (done) {
                siteListener
                    .then(function () {
                        update(LNBTS_STATE.UNBLOCKED, LNBTS_STATE.ENABLED);
                        reset();
                    })
                    .then(function () {
                        assert.notOk(siteAvailabilityResponder.isSiteUnblocked());
                    })
                    .then(done)
                    .catch(done);
            });

            it('isSiteBlocked should return false', function (done) {
                siteListener
                    .then(function () {
                        update(LNBTS_STATE.BLOCKED, LNBTS_STATE.ENABLED);
                        reset();
                    })
                    .then(function () {
                        assert.notOk(siteAvailabilityResponder.isSiteBlocked());
                    })
                    .then(done)
                    .catch(done);
            });

            it('isBtsOperational should return false', function (done) {
                siteListener
                    .then(function () {
                        update(LNBTS_STATE.UNBLOCKED, LNBTS_STATE.ENABLED);
                        reset();
                    })
                    .then(function () {
                        assert.notOk(siteAvailabilityResponder.isBtsOperational());
                    })
                    .then(done)
                    .catch(done);
            });
        });

        function update(blockingState, operationalState) {
            const parameters = createLnbtsParameters(blockingState, operationalState);
            const node = imTree.mergeNode(LNBTS.distName, parameters);
            node.parameters = null;
            node.getDecodedParameters = function () {
                return parameters;
            };
            const updateEvent = {
                completeImTree: imTree,
                change: {updated: [LNBTS.distName]}
            };
            imTracker.imUpdate(updateEvent);
        }

        function remove(lnbts) {
            imTree.removeByDistName(lnbts.distName);
            imTracker.imUpdate({completeImTree: imTree});
        }

        function reset() {
            imTree = new ImTree();
            imTracker.imUpdate({completeImTree: imTree});
        }

        function createLnbtsParameters(blockingState, operationalState) {
            return {
                stateInfo: {
                    blockingState: blockingState,
                    operationalState: operationalState
                }
            };
        }
    });
})();
