(function () {
    'use strict';

    describe('ChangeBlockingStateReq', function () {

        const DISTNAME_1 = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/CHANGE_BLOCKING_STATE_REQ-1';
        const DISTNAME_2 = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/CHANGE_BLOCKING_STATE_REQ-2';
        const OBJECT_DN = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1/LNCEL-21711';
        const OBJECT_DN_2 = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1/LNCEL-21713';
        const OPERATION_TYPE_BLOCK = 'Block';
        const OPERATION_TYPE_UNBLOCK = 'Unblock';
        const OPERATION_STATUS_ONGOING = 'Ongoing';
        const OPERATION_STATUS_COMPLETED = 'Completed';
        const ChangeBlockingStateReq = adminCore.procedures.ChangeBlockingStateReq;
        const parameters = {
            objectDN: OBJECT_DN,
            operationType: 'Block',
            operationStatus: OPERATION_STATUS_ONGOING
        };

        describe('Method: hasDistName', function () {
            it('should return true if request object has given distName', function () {
                const requestObject = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                assert.ok(requestObject.hasDistName(DISTNAME_1));
            });

            it('should return false if request object hasn\'t given distName', function () {
                const requestObject = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                assert.notOk(requestObject.hasDistName(DISTNAME_2));
            });

        });

        describe('Method: setOperationStatus', function () {

            it('should set operation status of request object to given', function () {
                const requestObject = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                assert.ok(requestObject.isExecuting());
                requestObject.setOperationStatus(OPERATION_STATUS_COMPLETED);
                assert.notOk(requestObject.isExecuting());
            });
        });

        describe('Method: isExecuting', function () {

            it('should return true if operationStatus of request object is other than Completed', function () {
                const requestObject = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                assert.ok(requestObject.isExecuting());
            });

            it('should return false if operationStatus of request object is Completed', function () {
                const requestObject = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                requestObject.setOperationStatus(OPERATION_STATUS_COMPLETED);
                assert.notOk(requestObject.isExecuting());
            });
        });

        describe('Method: isForOperation', function () {

            it('should return true if request object is dedicated for given operation', function () {
                const requestObject = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                assert.ok(requestObject.isForOperation(OPERATION_TYPE_BLOCK, OBJECT_DN));
            });

            it('should return false if request object is dedicated for different operation type', function () {
                const parameters2 = {
                    objectDN: OBJECT_DN,
                    operationType: OPERATION_TYPE_UNBLOCK,
                    operationStatus: OPERATION_STATUS_ONGOING
                };
                const requestObject = new ChangeBlockingStateReq(DISTNAME_1, parameters2);
                assert.notOk(requestObject.isForOperation(OPERATION_TYPE_BLOCK, OBJECT_DN));
            });

            it('should return false if request object is dedicated for different managed object', function () {
                const requestObject = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                assert.notOk(requestObject.isForOperation(OPERATION_TYPE_BLOCK, OBJECT_DN_2));
            });
        });

        describe('Method: isEqual', function () {

            it('should return true if request objects have the same objectDN, operationType and distName', function () {
                const requestObject1 = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                const requestObject2 = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                assert.ok(requestObject1.isEqual(requestObject2));
            });

            it('should return false if request objects have different objectDN', function () {
                const parameters2 = {
                    objectDN: OBJECT_DN_2,
                    operationType: OPERATION_TYPE_BLOCK,
                    operationStatus: OPERATION_STATUS_ONGOING
                };
                const requestObject1 = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                const requestObject2 = new ChangeBlockingStateReq(DISTNAME_1, parameters2);
                assert.notOk(requestObject1.isEqual(requestObject2));
            });

            it('should return false if request objects have different operationType', function () {
                const parameters2 = {
                    objectDN: OBJECT_DN,
                    operationType: OPERATION_TYPE_UNBLOCK,
                    operationStatus: OPERATION_STATUS_ONGOING
                };
                const requestObject1 = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                const requestObject2 = new ChangeBlockingStateReq(DISTNAME_1, parameters2);
                assert.notOk(requestObject1.isEqual(requestObject2));
            });

            it('should return false if request objects have different distName', function () {
                const requestObject1 = new ChangeBlockingStateReq(DISTNAME_1, parameters);
                const requestObject2 = new ChangeBlockingStateReq(DISTNAME_2, parameters);
                assert.notOk(requestObject1.isEqual(requestObject2));
            });
        });
    });
})();

