(function () {
    'use strict';

    describe('ChangeBlockingStateReqTracker', function () {

        const DISTNAME_1 = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/CHANGE_BLOCKING_STATE_REQ-1';
        const DISTNAME_2 = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/CHANGE_BLOCKING_STATE_REQ-2';
        const OBJECT_DN = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1/LNCEL-21711';
        const OBJECT_DN_2 = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1/LNCEL-21713';
        const OPERATION_TYPE_BLOCK = 'Block';
        const OPERATION_TYPE_UNBLOCK = 'Unblock';
        const OPERATION_STATUS_ONGOING = 'Ongoing';
        const OPERATION_STATUS_COMPLETED = 'Completed';
        const ChangeBlockingStateReqTracker = adminCore.procedures.ChangeBlockingStateReqTracker;
        const request1 = {
            distName: DISTNAME_1,
            parameters: {
                objectDN: OBJECT_DN,
                operationStatus: OPERATION_STATUS_ONGOING,
                operationType: OPERATION_TYPE_BLOCK
            }
        };
        const request2 = {
            distName: DISTNAME_2,
            parameters: {
                objectDN: OBJECT_DN_2,
                operationStatus: OPERATION_STATUS_ONGOING,
                operationType: OPERATION_TYPE_UNBLOCK
            }
        };

        var imTracker;

        beforeEach(function () {
            imTracker = createDummyTracker();
        });

        function createDummyTracker() {

            var imListener;
            return {
                trackByClassName: trackByClassName,
                getListener: getListener
            };

            function trackByClassName(className, listener) {
                imListener = listener;
            }

            function getListener() {
                return imListener;
            }
        }

        describe('On event: onListenerRegistered', function () {
            it('should add to tracking two requests objects', function () {
                const event = {
                    objects: [request1, request2]
                };
                const changeBlockingStateReqTracker = new ChangeBlockingStateReqTracker({imTracker: imTracker});
                imTracker.getListener().onListenerRegistered(event);
                assert.ok(changeBlockingStateReqTracker.isRequestExecuting(OPERATION_TYPE_BLOCK, OBJECT_DN));
                assert.ok(changeBlockingStateReqTracker.isRequestExecuting(OPERATION_TYPE_UNBLOCK, OBJECT_DN_2));
            });
        });

        describe('On event: onObjectRemoved', function () {
            it('should remove from tracking object with given distName', function () {
                const event = {
                    objects: [request1, request2]
                };
                const changeBlockingStateReqTracker = new ChangeBlockingStateReqTracker({imTracker: imTracker});
                imTracker.getListener().onListenerRegistered(event);
                assert.ok(changeBlockingStateReqTracker.isRequestExecuting(OPERATION_TYPE_BLOCK, OBJECT_DN));
                assert.ok(changeBlockingStateReqTracker.isRequestExecuting(OPERATION_TYPE_UNBLOCK, OBJECT_DN_2));

                const removeEvent = {
                    distNames: [DISTNAME_2]
                };
                imTracker.getListener().onObjectRemoved(removeEvent);
                assert.ok(changeBlockingStateReqTracker.isRequestExecuting(OPERATION_TYPE_BLOCK, OBJECT_DN));
                assert.notOk(changeBlockingStateReqTracker.isRequestExecuting(OPERATION_TYPE_UNBLOCK, OBJECT_DN_2));
            });
        });


        describe('On event: onParameterChanged', function () {
            it('should add to tracking new request object', function () {
                const request = {
                    distName: DISTNAME_1,
                    parameters: {
                        objectDN: OBJECT_DN,
                        operationStatus: OPERATION_STATUS_ONGOING,
                        operationType: OPERATION_TYPE_BLOCK
                    }
                };

                const event = {
                    objects: [request]
                };
                const changeBlockingStateReqTracker = new ChangeBlockingStateReqTracker({imTracker: imTracker});
                imTracker.getListener().onParameterChanged(event);
                assert.ok(changeBlockingStateReqTracker.isRequestExecuting(OPERATION_TYPE_BLOCK, OBJECT_DN));
            });

            it('should update existing request object', function () {
                const request = {
                    distName: DISTNAME_1,
                    parameters: {
                        objectDN: OBJECT_DN,
                        operationStatus: OPERATION_STATUS_COMPLETED,
                        operationType: OPERATION_TYPE_BLOCK
                    }
                };

                const event = {
                    objects: [request]
                };
                const changeBlockingStateReqTracker = new ChangeBlockingStateReqTracker({imTracker: imTracker});
                imTracker.getListener().onParameterChanged(event);
                assert.ok(changeBlockingStateReqTracker.isRequestExecuting(OPERATION_TYPE_BLOCK, OBJECT_DN));
            });
        });
    });
})();
