const _ = require('lodash');
const procedureExecutionCleanerFunc = require('../../src/procedures/procedureExecutionCleanerService');

describe('Service: procedure-execution-cleaner', () => {
    let procedureExecutionCleaner, executeOperationSpy;
    const DELETE_OPERATION = ['oper_delete', 'delete'];

    beforeEach(() => {
        executeOperationSpy = sinon.spy(Promise.resolve);
        const operationsService = {executeOperation: executeOperationSpy};
        const btsServiceProcedureFake = {registerListener: _.noop};
        procedureExecutionCleaner = procedureExecutionCleanerFunc({
            btsServiceProcedure: btsServiceProcedureFake,
            operationsService: operationsService
        });
    });

    it('should execute \'oper_delete\' when procedure is \'Rejected\'', () => {
        const distName = 'BTS_SERVICE_PROCEDURE-1';
        procedureExecutionCleaner.update(new EventBuilder()
            .addObject(distName, {status: 'Rejected'})
            .build());

        assert.isTrue(executeOperationSpy.calledOnce);
        assert.isTrue(executeOperationSpy.calledWith(DELETE_OPERATION, distName));
    });

    it('should execute \'oper_delete\' when procedure is \'Completed\'', () => {
        const distName = 'BTS_SERVICE_PROCEDURE-1';
        procedureExecutionCleaner.update(new EventBuilder()
            .addObject(distName, {status: 'Completed'})
            .build());

        assert.isTrue(executeOperationSpy.calledOnce);
        assert.isTrue(executeOperationSpy.calledWith(DELETE_OPERATION, distName));
    });

    it('should execute \'oper_delete\' when procedure is \'Failed\'', () => {
        const distName = 'BTS_SERVICE_PROCEDURE-1';
        procedureExecutionCleaner.update(new EventBuilder()
            .addObject(distName, {status: 'Failed'})
            .build());

        assert.isTrue(executeOperationSpy.calledOnce);
        assert.isTrue(executeOperationSpy.calledWith(DELETE_OPERATION, distName));
    });

    it('should execute \'oper_delete\' when procedure is \'Aborted\'', () => {
        const distName = 'BTS_SERVICE_PROCEDURE-1';
        procedureExecutionCleaner.update(new EventBuilder()
            .addObject(distName, {status: 'Aborted'})
            .build());

        assert.isTrue(executeOperationSpy.calledOnce);
        assert.isTrue(executeOperationSpy.calledWith(DELETE_OPERATION, distName));
    });

    it('should not execute \'oper_delete\' when procedure is \'Requested\'', () => {
        const distName = 'BTS_SERVICE_PROCEDURE-1';
        procedureExecutionCleaner.update(new EventBuilder()
            .addObject(distName, {status: 'Requested'})
            .build());

        assert.isFalse(executeOperationSpy.called);
    });

    it('should not execute \'oper_delete\' when procedure is \'Accepted\'', () => {
        const distName = 'BTS_SERVICE_PROCEDURE-1';
        procedureExecutionCleaner.update(new EventBuilder()
            .addObject(distName, {status: 'Accepted'})
            .build());

        assert.isFalse(executeOperationSpy.called);
    });

    it('should not execute \'oper_delete\' when procedure is \'Enqueued\'', () => {
        const distName = 'BTS_SERVICE_PROCEDURE-1';
        procedureExecutionCleaner.update(new EventBuilder()
            .addObject(distName, {status: 'Enqueued'}).build());

        assert.isFalse(executeOperationSpy.called);
    });

    it('should call execute \'oper_delete\' for all final procedures', () => {
        const distName1 = 'BTS_SERVICE_PROCEDURE-1';
        const distName2 = 'BTS_SERVICE_PROCEDURE-2';
        const distName3 = 'BTS_SERVICE_PROCEDURE-3';

        procedureExecutionCleaner.update(new EventBuilder()
            .addObject(distName1, {status: 'Completed'})
            .addObject(distName2, {status: 'Enqueued'})
            .addObject(distName3, {status: 'Rejected'})
            .build());

        assert.isTrue(executeOperationSpy.calledTwice);
        assert.isTrue(executeOperationSpy.firstCall.calledWith(DELETE_OPERATION, distName1));
        assert.isTrue(executeOperationSpy.secondCall.calledWith(DELETE_OPERATION, distName3));
    });

    function EventBuilder() {
        const event = {
            objects: []
        };

        this.addObject = (distName, parameters) => {
            event.objects.push({distName: distName, parameters: parameters});
            return this;
        };

        this.build = () => {
            return event;
        };
    }
});
