describe('procedureExecutor', function () {

    const TARGET_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1',
        BTS_SERVICE_DIST_NAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/BTS_SERVICE-1',
        OPERATION_PARAMETERS = {
            clientId: 1,
            targetDN: TARGET_DISTNAME
        },
        BLOCK_OPERATION_NAME = 'Site block',
        BLOCK_OPERATION_TYPE = ['oper_block', 'block'],
        UNBLOCK_OPERATION_NAME = 'Site unblock',
        UNBLOCK_OPERATION_TYPE = ['oper_unblock', 'unblock'],
        RESET_OPERATION_NAME = 'Site reset',
        RESET_OPERATION_TYPE = ['oper_reset', 'reset'];


    var procedureExecutor, procedureIdFactory, executeOperationSpy, addAvailabilityListenerSpy, procedureStatusListener, getIdSpy;

    beforeEach(() => {
        getIdSpy = sinon.stub().returns({id: 1});
        procedureStatusListener = sinon.spy();
        executeOperationSpy = sinon.spy(Promise.resolve);
        addAvailabilityListenerSpy = sinon.spy();

        procedureIdFactory = {getId: getIdSpy};
        var procedureNotification = {addAvailabilityListener: addAvailabilityListenerSpy},
            operationsService = {executeOperation: executeOperationSpy};
        procedureExecutor = adminCore.procedures.procedureExecutor(operationsService, procedureIdFactory, procedureNotification);
    });

    describe('Method: executeBlock', () => {

        it('should add availability listener to procedureNotification service', () => {
            procedureExecutor.executeBlock(BLOCK_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener);

            assertAddAvailabilityListenerCalledOnceWithTargetAndProcedureStatusListener();
        });

        it('should execute operation using operationsService', () => {
            procedureExecutor.executeBlock(BLOCK_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener);
            assertOperationExecutedOnceWithParams(BLOCK_OPERATION_TYPE, OPERATION_PARAMETERS);

        });

        it('should execute procedureStatusListener once', () => {
            procedureExecutor.executeBlock(BLOCK_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener);

            assertProcedureStatusListenerCalledOnceWith(true);
        });

        it('should execute operation with given id', () => {
            const id = 666;
            procedureExecutor.executeBlock(BLOCK_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener, {
                clientId: id
            });

            assertOperationExecutedOnceWithParams(BLOCK_OPERATION_TYPE, {
                clientId: id,
                targetDN: TARGET_DISTNAME
            });
        });
    });

    describe('Method: executeUnblock', () => {

        it('should add availability listener to procedureNotification service', () => {
            procedureExecutor.executeUnblock(UNBLOCK_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener);

            assertAddAvailabilityListenerCalledOnceWithTargetAndProcedureStatusListener();
        });

        it('should execute operation using operationsService', () => {
            procedureExecutor.executeUnblock(UNBLOCK_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener);

            assertOperationExecutedOnceWithParams(UNBLOCK_OPERATION_TYPE, OPERATION_PARAMETERS);
        });

        it('should execute procedureStatusListener once', () => {
            procedureExecutor.executeUnblock(UNBLOCK_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener);

            assertProcedureStatusListenerCalledOnceWith(true);
        });

        it('should execute operation with given id', () => {
            const id = 666;
            procedureExecutor.executeUnblock(UNBLOCK_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener, {
                clientId: id
            });

            assertOperationExecutedOnceWithParams(UNBLOCK_OPERATION_TYPE, {
                clientId: id,
                targetDN: TARGET_DISTNAME
            });
        });
    });

    describe('Method: executeReset', () => {

        it('should add availability listener to procedureNotification service', () => {
            procedureExecutor.executeReset(RESET_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener);

            assertAddAvailabilityListenerCalledOnceWithTargetAndProcedureStatusListener();
        });

        it('should execute operation using operationsService', () => {
            procedureExecutor.executeReset(RESET_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener);

            assertOperationExecutedOnceWithParams(RESET_OPERATION_TYPE, OPERATION_PARAMETERS);
        });

        it('should execute procedureStatusListener once', () => {
            procedureExecutor.executeReset(RESET_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener);

            assertProcedureStatusListenerCalledOnceWith(true);
        });

        it('should execute operation with given id', () => {
            const id = 666;
            procedureExecutor.executeReset(RESET_OPERATION_NAME, TARGET_DISTNAME, procedureStatusListener, {
                clientId: id
            });

            assertOperationExecutedOnceWithParams(RESET_OPERATION_TYPE, {
                clientId: id,
                targetDN: TARGET_DISTNAME
            });
        });
    });

    describe('#executeDownloadPlan', () => {

        const DOWNLOAD_PLAN_PARAMETERS = {uri: 'http://10.10.10.10:80/scf.xml', planId: 1},
            DOWNLOAD_PLAN_NAME = ['oper_downloadPlan', 'downloadPlan'];

        it('should execute operation using operationsService', () => {
            procedureExecutor.executeDownloadPlan(DOWNLOAD_PLAN_PARAMETERS);

            assertOperationExecutedOnceWithParams(DOWNLOAD_PLAN_NAME, {
                uri: 'http://10.10.10.10:80/scf.xml',
                planId: 1,
                clientId: 1
            });
        });

        it('should execute operation with given id', () => {
            DOWNLOAD_PLAN_PARAMETERS.clientId = 666;
            procedureExecutor.executeDownloadPlan(DOWNLOAD_PLAN_PARAMETERS);

            assertOperationExecutedOnceWithParams(DOWNLOAD_PLAN_NAME, {
                uri: 'http://10.10.10.10:80/scf.xml',
                planId: 1,
                clientId: 666
            });
        });
    });
    describe('#executeActivatePlan', () => {

        const ACTIVATE_PLAN_PARAMETERS = {planId: 1},
            ACTIVATE_PLAN_NAME = ['oper_activatePlan', 'activatePlan'];

        it('should execute operation using operationsService', () => {
            procedureExecutor.executeActivatePlan(ACTIVATE_PLAN_PARAMETERS);

            assertOperationExecutedOnceWithParams(ACTIVATE_PLAN_NAME, {
                clientId: 1,
                planId: 1
            });
        });

        it('should execute operation with given id', () => {
            ACTIVATE_PLAN_PARAMETERS.clientId = 666;
            procedureExecutor.executeActivatePlan(ACTIVATE_PLAN_PARAMETERS);

            assertOperationExecutedOnceWithParams(ACTIVATE_PLAN_NAME, {
                clientId: 666,
                planId: 1
            });
        });
    });
    describe('#executeSwRollback', () => {

        const SW_ROLLBACK_NAME = ['oper_swRollback', 'swRollback'],
            SW_ROLLBACK_PARAMETERS = {};
        it('should execute operation using operationsService', () => {
            procedureExecutor.executeSwRollback();

            assertOperationExecutedOnceWithParams(SW_ROLLBACK_NAME, {
                clientId: 1
            });
        });

        it('should execute operation with given id', () => {
            SW_ROLLBACK_PARAMETERS.clientId = 666;
            procedureExecutor.executeSwRollback(SW_ROLLBACK_PARAMETERS);

            assertOperationExecutedOnceWithParams(SW_ROLLBACK_NAME, {
                clientId: 666
            });
        });
    });
    describe('#executeResetToTestDedicated', () => {

        const RESET_TO_TEST_DEDICATED_NAME = ['oper_resetToTestDedicated', 'resetToTestDedicated'],
            RESET_TO_TEST_DEDICATED_PARAMETERS = {};

        it('should execute operation using operationsService', () => {
            procedureExecutor.executeResetToTestDedicated();

            assertOperationExecutedOnceWithParams(RESET_TO_TEST_DEDICATED_NAME, {
                clientId: 1
            });
        });

        it('should execute operation with given id', () => {
            RESET_TO_TEST_DEDICATED_PARAMETERS.clientId = 666;
            procedureExecutor.executeResetToTestDedicated(RESET_TO_TEST_DEDICATED_PARAMETERS);

            assertOperationExecutedOnceWithParams(RESET_TO_TEST_DEDICATED_NAME, {
                clientId: 666
            });
        });
    });
    describe('#executeConfigurationReset', () => {

        const CONFIGURATION_RESET_NAME = ['oper_configurationReset', 'configurationReset'],
            CONFIGURATION_PARAMETERS = {removeSiteId: true, removeSecurityCredentials: true};

        it('should execute operation using operationsService', () => {
            procedureExecutor.executeConfigurationReset(CONFIGURATION_PARAMETERS);

            assertOperationExecutedOnceWithParams(CONFIGURATION_RESET_NAME, {
                clientId: 1,
                removeSiteId: true,
                removeSecurityCredentials: true
            });
        });

        it('should execute operation with given id', () => {
            CONFIGURATION_PARAMETERS.clientId = 666;
            procedureExecutor.executeConfigurationReset(CONFIGURATION_PARAMETERS);

            assertOperationExecutedOnceWithParams(CONFIGURATION_RESET_NAME, {
                clientId: 666,
                removeSiteId: true,
                removeSecurityCredentials: true
            });

        });
    });
    describe('#executeChangeEthernetPortSecurity', () => {
        const CHANGE_ETHERNET_PORT_SECURITY_NAME = ['oper_changeEthernetPortSecurity', 'changeEthernetPortSecurity'],
            CHANGE_PARAMETERS = {enableEthernetPortSec: true};

        it('should execute operation using operationsService', () => {
            procedureExecutor.executeChangeEthernetPortSecurity(CHANGE_PARAMETERS);

            assertOperationExecutedOnceWithParams(CHANGE_ETHERNET_PORT_SECURITY_NAME, {
                clientId: 1,
                enableEthernetPortSec: true
            });
        });

        it('should execute operation with given id', () => {
            CHANGE_PARAMETERS.clientId = 666;
            procedureExecutor.executeChangeEthernetPortSecurity(CHANGE_PARAMETERS);

            assertOperationExecutedOnceWithParams(CHANGE_ETHERNET_PORT_SECURITY_NAME, {
                clientId: 666,
                enableEthernetPortSec: true
            });
        });
    });
    function assertAddAvailabilityListenerCalledOnceWithTargetAndProcedureStatusListener() {
        sinon.assert.calledOnce(addAvailabilityListenerSpy);
        sinon.assert.calledWith(addAvailabilityListenerSpy, TARGET_DISTNAME, procedureStatusListener);
    }

    function assertOperationExecutedOnceWithParams(names, parameters) {
        sinon.assert.calledOnce(executeOperationSpy);
        sinon.assert.calledWith(executeOperationSpy, names, BTS_SERVICE_DIST_NAME, parameters);
    }

    function assertProcedureStatusListenerCalledOnceWith(parameter) {
        sinon.assert.calledOnce(procedureStatusListener);
        sinon.assert.calledWith(procedureStatusListener, parameter);
    }
});
