'use strict';

const ProcedureTrackerFactory = require('../../src/procedures/procedureTrackerFactory');

describe('Factory: procedureTrackerFactory', function () {

    var procedureTrackerFactory, updatedImTree;
    const SNAPSHOT_DISTNAME_1 = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/TECHREP_L-1/SNAPSHOT_L-1';
    const SNAPSHOT_DISTNAME_3 = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/TECHREP_L-1/SNAPSHOT_L-3';

    const CLIENT_ID_1 = 1;
    const CLIENT_ID_2 = 2;
    const CLIENT_ID_3 = 3;
    const CLIENT_ID_4 = 4;

    const PROCEDURE_TYPE_SNAPSHOT = 'Snapshot';
    const PROCEDURE_TYPE_BLOCK = 'Block';

    const STATUS_ACCEPTED = 'Accepted';
    const STATUS_COMPLETED = 'Completed';


    const BTS_SERVICE_PROCEDURE_1 = {
        className: 'BTS_SERVICE_PROCEDURE',
        distName: 'BTS_SERVICE_PROCEDURE-1',
        getDecodedParameters: function () {
            return {
                clientId: CLIENT_ID_1,
                procedureType: PROCEDURE_TYPE_SNAPSHOT,
                targetDN: SNAPSHOT_DISTNAME_1
            };
        }
    };

    const BTS_SERVICE_PROCEDURE_2 = {
        className: 'BTS_SERVICE_PROCEDURE',
        distName: 'BTS_SERVICE_PROCEDURE-2',
        getDecodedParameters: function () {
            return {
                clientId: CLIENT_ID_2,
                procedureType: PROCEDURE_TYPE_BLOCK
            };
        }
    };

    const BTS_SERVICE_PROCEDURE_3 = {
        className: 'BTS_SERVICE_PROCEDURE',
        distName: 'BTS_SERVICE_PROCEDURE-3',
        getDecodedParameters: function () {
            return {
                clientId: CLIENT_ID_3,
                procedureType: PROCEDURE_TYPE_SNAPSHOT,
                targetDN: SNAPSHOT_DISTNAME_3,
                status: STATUS_ACCEPTED
            };
        }
    };

    const BTS_SERVICE_PROCEDURE_4 = {
        className: 'BTS_SERVICE_PROCEDURE',
        distName: 'BTS_SERVICE_PROCEDURE-4',
        getDecodedParameters: function () {
            return {
                clientId: CLIENT_ID_4,
                procedureType: PROCEDURE_TYPE_SNAPSHOT,
                targetDN: SNAPSHOT_DISTNAME_3,
                status: STATUS_COMPLETED
            };
        }
    };

    beforeEach(function () {
        updatedImTree = {
            findByClassName: function (className) {
                return [BTS_SERVICE_PROCEDURE_1, BTS_SERVICE_PROCEDURE_2, BTS_SERVICE_PROCEDURE_3, BTS_SERVICE_PROCEDURE_4];
            },
            isEmpty: function () {
                return false;
            }
        };
        const metaService = {
            getLastMetaUid: function () {
                return 1;
            }
        };
        const onlineImTree = adminCore.onlineImTree({metaService: metaService});
        procedureTrackerFactory = ProcedureTrackerFactory(onlineImTree);
    });

    describe('createProcedureTracker method', function () {
        it('should create BtsServiceProcedureTracker ', function (done) {
            procedureTrackerFactory.createProcedureTracker(CLIENT_ID_1, PROCEDURE_TYPE_SNAPSHOT, {})
                .then(assert.ok)
                .then(done)
                .catch(done);
        });
    });

    describe('Tracker: BtsServiceProcedureTracker', function () {

        describe('onlineImTree method', function () {

            it('should call on procedure listener method "onAccepted" with SNAPSHOT_DISTNAME_3', function (done) {
                const procedureListener = {
                    onAccepted: sinon.spy()
                };
                procedureTrackerFactory.createProcedureTracker(CLIENT_ID_3, PROCEDURE_TYPE_SNAPSHOT, procedureListener)
                    .then(function (procedureTracker) {
                        procedureTracker.onUpdate({completeImTree: updatedImTree});
                        assert.isTrue(procedureListener.onAccepted.calledWith(SNAPSHOT_DISTNAME_3));
                    })
                    .then(done)
                    .catch(done);
            });

            it('should call on procedure listener method "Accepted" with SNAPSHOT_DISTNAME_3 only once', function (done) {
                const procedureListener = {
                    onAccepted: sinon.spy()
                };
                procedureTrackerFactory.createProcedureTracker(CLIENT_ID_3, PROCEDURE_TYPE_SNAPSHOT, procedureListener)
                    .then(function (procedureTracker) {
                        procedureTracker.onUpdate({completeImTree: updatedImTree});
                        procedureTracker.onUpdate({completeImTree: updatedImTree});
                        assert.isTrue(procedureListener.onAccepted.calledWith(SNAPSHOT_DISTNAME_3));
                        assert.isTrue(procedureListener.onAccepted.calledOnce);
                    })
                    .then(done)
                    .catch(done);
            });
        });
    });
});
