(function () {
    'use strict';

    const ResetRequestTracker = require('../../src/procedures/resetRequestTracker');

    describe('ResetRequestTracker', function () {

        const DISTNAME_1 = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/RESET_REQ-1';
        const DISTNAME_2 = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/RESET_REQ-2';
        const OBJECT_DN = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1';
        const OBJECT_DN_2 = '/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1/LNCEL-21713';

        const request1 = {
            distName: DISTNAME_1,
            parameters: {
                objectDN: OBJECT_DN
            }
        };
        const request2 = {
            distName: DISTNAME_2,
            parameters: {
                objectDN: OBJECT_DN_2
            }
        };

        var imTracker;

        beforeEach(function () {
            imTracker = createFakeTracker();
        });

        function createFakeTracker() {

            var imListener;
            return {
                trackByClassName: trackByClassName,
                getListener: getListener
            };

            function trackByClassName(className, listener) {
                imListener = listener;
            }

            function getListener() {
                return imListener;
            }
        }

        describe('On event: onListenerRegistered', function () {
            it('should add to tracking requests objects with objectDN set to LNBTS', function () {
                //given
                const event = {
                    objects: [request1]
                };
                const resetRequestTracker = ResetRequestTracker.newInstance({imTracker: imTracker});

                //when
                imTracker.getListener().onListenerRegistered(event);

                //then
                assert.ok(resetRequestTracker.isResetExecuting());
            });

            it('should not add to tracking requests objects with objectDN not set to LNBTS', function () {
                //given
                const event = {
                    objects: [request2]
                };
                const resetRequestTracker = ResetRequestTracker.newInstance({imTracker: imTracker});

                //when
                imTracker.getListener().onListenerRegistered(event);

                //then
                assert.notOk(resetRequestTracker.isResetExecuting());
            });
        });

        describe('On event: onObjectRemoved', function () {
            it('should remove from tracking object with given distName', function () {
                //given
                const event = {
                    objects: [request1]
                };
                const removeEvent = {
                    distNames: [DISTNAME_1]
                };
                const resetRequestTracker = ResetRequestTracker.newInstance({imTracker: imTracker});
                imTracker.getListener().onListenerRegistered(event);
                assert.ok(resetRequestTracker.isResetExecuting());

                //when
                imTracker.getListener().onObjectRemoved(removeEvent);

                //then
                assert.notOk(resetRequestTracker.isResetExecuting());
            });
        });


        describe('On event: onParameterChanged', function () {
            it('should add to tracking new request object if has LNBTS as objectDN', function () {
                //given
                const request = {
                    distName: DISTNAME_1,
                    parameters: {
                        objectDN: OBJECT_DN
                    }
                };
                const event = {
                    objects: [request]
                };
                const resetRequestTracker = ResetRequestTracker.newInstance({imTracker: imTracker});

                //when
                imTracker.getListener().onParameterChanged(event);

                //then
                assert.ok(resetRequestTracker.isResetExecuting());
            });

            it('should not add to tracking new request object if hasn\'t LNBTS as objectDN', function () {
                //given
                const request = {
                    distName: DISTNAME_1,
                    parameters: {
                        objectDN: OBJECT_DN_2
                    }
                };
                const event = {
                    objects: [request]
                };
                const resetRequestTracker = ResetRequestTracker.newInstance({imTracker: imTracker});

                //when
                imTracker.getListener().onParameterChanged(event);

                //then
                assert.notOk(resetRequestTracker.isResetExecuting());
            });

            it('should not add existing request object again', function () {
                //given
                const request = {
                    distName: DISTNAME_1,
                    parameters: {
                        objectDN: OBJECT_DN
                    }
                };
                const event = {
                    objects: [request]
                };
                const removeEvent = {
                    distNames: [DISTNAME_1]
                };
                const resetRequestTracker = ResetRequestTracker.newInstance({imTracker: imTracker});
                imTracker.getListener().onParameterChanged(event);
                assert.ok(resetRequestTracker.isResetExecuting());


                // when
                imTracker.getListener().onParameterChanged(event);
                assert.ok(resetRequestTracker.isResetExecuting());
                imTracker.getListener().onObjectRemoved(removeEvent);

                //then
                assert.notOk(resetRequestTracker.isResetExecuting());
            });
        });
    });
})();
