const assert = require('chai').assert;
const Promise = require('bluebird');

const fileDownloaderFactory = require('../../../src/procedures/snapshot/fileDownloaderFactory');

describe('Factory: fileDownloaderFactory', function () {
    'use strict';

    var fileDownloaderFactoryInstance;
    var fileTransferStub;
    var snapshotExecutorStub;

    const ZIP = {
        addFile: sinon.spy()
    };
    const FILE = {};

    const INTERNAL_URI = 'file:///ram/BTS6249_1011_pm_1_syslog.zip';
    const EXPECTED_INTERNAL_URI = 'ram/BTS6249_1011_pm_1_syslog.zip';
    const EXTERNAL_URI = 'ExtendedNeLogs.tar.gz';
    const EXPECTED_EXTERNAL_URI = 'transport/ExtendedNeLogs.tar.gz';

    function createFileInfoService() {
        return {
            push: sinon.spy(),
            notify: sinon.spy()
        };
    }

    beforeEach(function () {

        fileTransferStub = {
            getFile: sinon.stub().returns(new Promise(function (resolve) {
                return resolve(FILE);
            }))
        };

        snapshotExecutorStub = {
            deleteFile: sinon.stub().returns(new Promise(function (resolve) {
                return resolve();
            }))
        };

        fileDownloaderFactoryInstance = fileDownloaderFactory(snapshotExecutorStub, fileTransferStub);
    });

    describe('getZipUri method', function () {
        it('should slice file:/// prefix from URI, when uri contains full path', function () {
            assert.strictEqual(fileDownloaderFactoryInstance.getZipUri(INTERNAL_URI), EXPECTED_INTERNAL_URI);
        });

        it('should add transport/ prefix to URI, when uri contains only file name', function () {
            assert.strictEqual(fileDownloaderFactoryInstance.getZipUri(EXTERNAL_URI), EXPECTED_EXTERNAL_URI);
        });
    });

    describe('createFileInfo method', function () {
        it('should create correct fileInfo', function () {
            const EMPTY_STRING = '';
            const fileInfo = fileDownloaderFactoryInstance.createFileInfo(EXPECTED_EXTERNAL_URI);
            assert.strictEqual(fileInfo.fileName, EXPECTED_EXTERNAL_URI);
            assert.strictEqual(fileInfo.status, 'status-downloading');
            assert.strictEqual(fileInfo.fileSize, EMPTY_STRING);
            assert.strictEqual(fileInfo.details, EMPTY_STRING);
        });
    });

    describe('createFileDownloader method', function () {
        var fileInfoService = createFileInfoService();
        const DIAGN_FILE_DISTNAME = 'DIAGN_FILE_L-1';
        const errorCallback = function () {
        };
        var fileDownloader = null;

        it('should return file downloader', function () {
            assert.ok(fileDownloaderFactoryInstance.createFileDownloader(DIAGN_FILE_DISTNAME, ZIP, fileInfoService, errorCallback));
        });

        describe('downloadFile method', function () {
            beforeEach(function () {
                fileInfoService = createFileInfoService();
                fileDownloader = fileDownloaderFactoryInstance.createFileDownloader(DIAGN_FILE_DISTNAME, ZIP, fileInfoService, errorCallback);
            });

            const parameters = {
                uri: INTERNAL_URI
            };

            it('should download file with unchanged uri then delete it From BTS', function (done) {
                fileDownloader.downloadFile(parameters)
                    .then(function () {
                        assert.isTrue(fileTransferStub.getFile.calledWith(INTERNAL_URI));
                        assert.isTrue(snapshotExecutorStub.deleteFile.calledWith(DIAGN_FILE_DISTNAME, undefined));
                    })
                    .done(done);
            });

            it('should add file to zip archive', function (done) {
                fileDownloader.downloadFile(parameters)
                    .then(function () {
                        assert.isTrue(ZIP.addFile.calledWith(FILE));
                    })
                    .done(done);
            });

            it('should create and fulfill fileInfo after successful file downloading', function (done) {
                fileDownloader.downloadFile(parameters)
                    .then(function () {
                        assert.isTrue(fileInfoService.push.calledOnce);
                        assert.isTrue(fileInfoService.notify.calledTwice);
                    })
                    .done(done);
            });

            it('should create and fulfill fileInfo after failed file downloading and should delete the file from BTS', function (done) {
                const error = {
                    type: 'error-type',
                    error: 'error-error'
                };

                fileTransferStub.getFile = sinon.stub().returns(new Promise(function (resolve, reject) {
                    return reject(error);
                }));

                fileDownloader.downloadFile(parameters)
                    .then(function () {
                        assert.isTrue(fileInfoService.push.calledOnce);
                        assert.isTrue(fileInfoService.notify.calledTwice);

                        assert.isTrue(snapshotExecutorStub.deleteFile.calledWith(DIAGN_FILE_DISTNAME));
                    })
                    .done(done);
            });
        });
    });
});
