
const assert = require('chai').assert;
const Promise = require('bluebird');

const snapshot = require('../../../src/procedures/snapshot/snapshot.js');
const snapshotStates = require('../../../src/procedures/snapshot/snapshotStates');

describe('Service: snapshot', function () {
    'use strict';

    const ID = 1;

    var filesInfoFactoryStub, snapshotExecutorStub, snapshotStatusHandlerStub,
        snapshotListenerFactoryStub, procedureIdFactoryStub, procedureTrackerFactoryStub;

    var snapshotInstance;

    const COVERAGE = 'Minor';

    const FILES_INFO = {
        reset: sinon.spy()
    };

    beforeEach(function () {

        snapshotStatusHandlerStub = {
            setStatus: sinon.spy(),
            isStarted: sinon.stub().returns(true)
        };

        snapshotExecutorStub = {
            startSnapshot: sinon.stub().returns(new Promise(function (resolve) {
                return resolve();
            })),
            cancel: sinon.stub().returns(new Promise(function (resolve) {
                return resolve();
            }))
        };

        filesInfoFactoryStub = {
            createFileInfo: function () {
                return FILES_INFO;
            }
        };

        snapshotListenerFactoryStub = {
            activateFilesDownload: sinon.stub().returns(Promise.resolve())
        };

        procedureIdFactoryStub = {
            getId: sinon.stub().returns({id: ID})
        };

        procedureTrackerFactoryStub = {
            createProcedureTracker: sinon.stub().returns(Promise.resolve())
        };

        snapshotInstance = snapshot(snapshotExecutorStub, snapshotListenerFactoryStub, filesInfoFactoryStub, snapshotStatusHandlerStub, procedureIdFactoryStub, procedureTrackerFactoryStub);
    });

    after(function () {
        FILES_INFO.reset.reset();
        snapshotStatusHandlerStub.setStatus.reset();
    });

    describe('startSnapshot method', function () {

        it('should reset files info', function () {
            snapshotInstance.startSnapshot(COVERAGE);
            assert.isTrue(procedureIdFactoryStub.getId.calledOnce);
            assert.isTrue(procedureTrackerFactoryStub.createProcedureTracker.calledOnce);
            assert.isTrue(FILES_INFO.reset.calledOnce);
        });

        it('should set status to REQUESTED', function (done) {
            snapshotInstance.startSnapshot(COVERAGE)
                .then(function () {
                    assert.isTrue(snapshotStatusHandlerStub.setStatus.calledWith(snapshotStates.REQUESTED));
                })
                .done(done);
        });

        it('should inform listener about activated files download', function (done) {
            var procedureTrackerFactoryMock = {
                createProcedureTracker: function (id, procedure, PROCEDURE_LISTENER) {
                    PROCEDURE_LISTENER.onAccepted();

                    assert.isTrue(snapshotListenerFactoryStub.activateFilesDownload.calledOnce);
                    done();

                    return Promise.resolve();
                }
            };

            snapshotInstance = snapshot(snapshotExecutorStub, snapshotListenerFactoryStub, filesInfoFactoryStub, snapshotStatusHandlerStub, procedureIdFactoryStub, procedureTrackerFactoryMock);
            snapshotInstance.startSnapshot(COVERAGE);
        });
    });

    describe('cancelSnapshotByUser method', function () {

        it('should set correct statuses', function (done) {
            snapshotInstance.cancelSnapshotByUser().then(function () {
                assert.isTrue(snapshotStatusHandlerStub.setStatus.calledTwice);
                assert.isTrue(snapshotStatusHandlerStub.setStatus.calledWith(snapshotStates.CANCELLING));
                assert.isTrue(snapshotStatusHandlerStub.setStatus.calledWith(snapshotStates.CANCELLED));
            }).done(done);
        });

        it('should cancel snapshot procedure', function (done) {
            snapshotInstance.cancelSnapshotByUser()
                .then(function () {
                    assert.isTrue(snapshotExecutorStub.cancel.calledOnce);
                })
                .done(done);
        });
    });
});
