const assert = require('chai').assert;

const snapshotExecutor = require('../../../src/procedures/snapshot/snapshotExecutor');

describe('Service: snapshotExecutor', function () {
    'use strict';

    var operationsServiceStub, snapshotExecutorInstance;
    const TECHREP_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/TECHREP_L-1';
    const BTS_SERVICE_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/BTS_SERVICE-1';
    const COVERAGE_ENUM = {Minor: 'Minor', Full: 'Full'};
    const SNAPSHOT_REQUEST_PARAMETERS = {
        'clientId': 0,
        'coverage': COVERAGE_ENUM.Minor
    };
    const FILE_DISTNAME = TECHREP_DISTNAME + '/SNAPSHOT_L-1/DIAGN_FILE_L-1';
    const FAILURE_REASON = 'File transfer error';
    const DELETE_FILE_OPERATION_NAME = ['oper_deleteFile', 'deleteFile'];
    const CANCEL_OPERATION_NAME = ['oper_cancel', 'cancel'];
    const COLLECT_SNAPSHOT_OPERATION_NAME = ['oper_collectSnapshot', 'collectSnapshot'];

    beforeEach(function () {

        operationsServiceStub = {
            createObjectAsOwner: sinon.spy(),
            executeOperation: sinon.spy(),
            deleteObject: sinon.spy()
        };

        snapshotExecutorInstance = snapshotExecutor(operationsServiceStub);
    });

    describe('startSnapshot method', function () {
        it('should call operationsService.executeOperation method with correct parameters', function () {
            snapshotExecutorInstance.startSnapshot(COVERAGE_ENUM.Minor, SNAPSHOT_REQUEST_PARAMETERS.clientId);
            assertIsOperationCalledOnceOnOperationsService(operationsServiceStub.executeOperation);
            assert.isTrue(operationsServiceStub.executeOperation.calledWith(COLLECT_SNAPSHOT_OPERATION_NAME, BTS_SERVICE_DISTNAME, SNAPSHOT_REQUEST_PARAMETERS));
        });
    });

    describe('deleteFile method', function () {
        it('should call operationsService.executeOperation method with correct parameters', function () {
            const operationParameters = {
                'DIAGN_FILE_L_DN': FILE_DISTNAME,
                'transferFailureReason': FAILURE_REASON
            };
            snapshotExecutorInstance.deleteFile(FILE_DISTNAME, FAILURE_REASON);
            assertIsOperationCalledOnceOnOperationsService(operationsServiceStub.executeOperation);
            assert.isTrue(operationsServiceStub.executeOperation.calledWith(DELETE_FILE_OPERATION_NAME, TECHREP_DISTNAME, operationParameters));
        });

        it('should call operationsService.executeOperation method with correct parameters - default failure reason', function () {
            const operationParameters = {
                'DIAGN_FILE_L_DN': FILE_DISTNAME,
                'transferFailureReason': ''
            };
            snapshotExecutorInstance.deleteFile(FILE_DISTNAME);
            assertIsOperationCalledOnceOnOperationsService(operationsServiceStub.executeOperation);
            assert.isTrue(operationsServiceStub.executeOperation.calledWith(DELETE_FILE_OPERATION_NAME, TECHREP_DISTNAME, operationParameters));
        });
    });

    describe('cancel method', function () {
        it('should call operationsService.executeOperation method', function () {
            snapshotExecutorInstance.cancel();
            assertIsOperationCalledOnceOnOperationsService(operationsServiceStub.executeOperation);
            assert.isTrue(operationsServiceStub.executeOperation.calledWith(CANCEL_OPERATION_NAME, TECHREP_DISTNAME));
        });
    });

    function assertIsOperationCalledOnceOnOperationsService(operation) {
        assert.isTrue(operation.calledOnce);
        assert.isTrue(operation.calledOn(operationsServiceStub));
    }
});
