const assert = require('chai').assert;
const Promise = require('bluebird');

const snapshotListenerFactory = require('../../../src/procedures/snapshot/snapshotListenerFactory');

describe('Factory: snapshotListenerFactory', function () {
    'use strict';

    var snapshotListenerFactoryInstance;

    var onlineImTreeStub, fileTransferStub, zipFactoryStub, fileDownloaderFactoryStub, fileDownloaderStub, updateUnwatcherStub;

    const SNAPSHOT_DISTNAME = '/MRBTS-1/RAT-1/BTS_L-1/FUNC_L-1/TECHREP_L-1/SNAPSHOT_L-0';

    const ZIP = {saveZip: sinon.spy(), setName: sinon.spy()};
    const NOT_EMPTY_FILE_INFO_SERVICE = {
        isEmpty: function () {
            return false;
        }
    };
    const EMPTY_FILE_INFO_SERVICE = {
        isEmpty: function () {
            return true;
        }
    };
    const INTERRUPT_SNAPSHOT_METHOD = function () {
    };
    const FILE_INFO_SERVICE = {};

    beforeEach(function () {

        fileTransferStub = {
            connect: sinon.stub().returns(new Promise(function (resolve) {
                return resolve();
            })),
            disconnect: sinon.spy()
        };

        zipFactoryStub = {
            newZip: sinon.stub().returns(ZIP)
        };

        fileDownloaderStub = {
            downloadFile: sinon.stub()
        };

        fileDownloaderFactoryStub = {
            createFileDownloader: sinon.stub().returns(fileDownloaderStub)
        };

        updateUnwatcherStub = {
            unwatch: sinon.spy()
        };

        onlineImTreeStub = {
            watchUpdates: sinon.stub().returns(new Promise(function (resolve) {
                return resolve(updateUnwatcherStub);
            }))
        };

        snapshotListenerFactoryInstance = snapshotListenerFactory(fileTransferStub, zipFactoryStub, fileDownloaderFactoryStub, onlineImTreeStub);
    });

    describe('createSnapshotListener method', function () {
        it('should create snapshot listener object', function () {
            var snapshotListener = snapshotListenerFactoryInstance.createSnapshotListener(SNAPSHOT_DISTNAME, EMPTY_FILE_INFO_SERVICE, INTERRUPT_SNAPSHOT_METHOD);
            assert.ok(snapshotListener);
        });

        describe('onSnapshotLRemoved method', function () {
            beforeEach(function () {
                ZIP.saveZip.reset();
            });

            it('should unregister snapshot listener from onlineImTree', function () {
                const snapshotListener = snapshotListenerFactoryInstance.createSnapshotListener(SNAPSHOT_DISTNAME, EMPTY_FILE_INFO_SERVICE, INTERRUPT_SNAPSHOT_METHOD);
                snapshotListener.setUnwatcher(updateUnwatcherStub);
                snapshotListener.onSnapshotLRemoved();
                assert.isTrue(updateUnwatcherStub.unwatch.calledOnce);
            });

            it('should save zip file if file info service is not empty', function () {
                const snapshotListener = snapshotListenerFactoryInstance.createSnapshotListener(SNAPSHOT_DISTNAME, NOT_EMPTY_FILE_INFO_SERVICE, INTERRUPT_SNAPSHOT_METHOD);
                snapshotListener.setUnwatcher(updateUnwatcherStub);
                snapshotListener.onSnapshotLRemoved();
                assert.isTrue(ZIP.saveZip.calledOnce);
            });

            it('should not save zip file if file info service is empty', function () {
                const snapshotListener = snapshotListenerFactoryInstance.createSnapshotListener(SNAPSHOT_DISTNAME, EMPTY_FILE_INFO_SERVICE, INTERRUPT_SNAPSHOT_METHOD);
                snapshotListener.setUnwatcher(updateUnwatcherStub);
                snapshotListener.onSnapshotLRemoved();
                assert.isFalse(ZIP.saveZip.calledOnce);
            });

            it('should disconnect file transfer service', function () {
                const snapshotListener = snapshotListenerFactoryInstance.createSnapshotListener(SNAPSHOT_DISTNAME, EMPTY_FILE_INFO_SERVICE, INTERRUPT_SNAPSHOT_METHOD);
                snapshotListener.setUnwatcher(updateUnwatcherStub);
                snapshotListener.onSnapshotLRemoved();
                assert.isTrue(fileTransferStub.disconnect.calledOnce);
            });
        });

        describe('onUpdate method', function () {
            it('should download files given by event', function () {
                const snapshotListener = snapshotListenerFactoryInstance.createSnapshotListener(SNAPSHOT_DISTNAME, EMPTY_FILE_INFO_SERVICE, INTERRUPT_SNAPSHOT_METHOD);

                const parameters = {};

                const file = {
                    distName: 'test file name',
                    getDecodedParameters: sinon.stub().returns(parameters)
                };

                const treeNode = {
                    getChildren: sinon.stub().returns([file])
                };

                const updateEvent = {
                    completeImTree: {
                        findByDistName: sinon.stub().returns(treeNode)
                    }
                };

                snapshotListener.onUpdate(updateEvent);

                assert(updateEvent.completeImTree.findByDistName.calledWith(SNAPSHOT_DISTNAME));
                assert(treeNode.getChildren.calledOnce);
                assert(fileDownloaderFactoryStub.createFileDownloader.calledWith(file.distName, ZIP, EMPTY_FILE_INFO_SERVICE, INTERRUPT_SNAPSHOT_METHOD));
                assert(file.getDecodedParameters.calledOnce);
                assert(fileDownloaderStub.downloadFile.calledWith(parameters));
            });
        });
    });

    describe('activateSnapshotListener method', function () {
        it('should connect to file transfer', function (done) {
            snapshotListenerFactoryInstance.activateFilesDownload(FILE_INFO_SERVICE, INTERRUPT_SNAPSHOT_METHOD)
                .then(function () {
                    assert.isTrue(fileTransferStub.connect.calledOnce);
                })
                .done(done);
        });

        it('should register snapshot listener to updateService', function (done) {
            onlineImTreeStub.watchUpdates.reset();
            snapshotListenerFactoryInstance.activateFilesDownload(FILE_INFO_SERVICE, INTERRUPT_SNAPSHOT_METHOD)
                .then(function () {
                    assert.isTrue(onlineImTreeStub.watchUpdates.calledOnce);
                })
                .done(done);
        });
    });
});
