const assert = require('assert');
const swUpdateExecutorFactory = require('../../../src/procedures/swUpdate/swUpdateExecutorFactory');
const Promise = require('bluebird');

describe('Factory: swUpdateExecutorFactory', function () {
    var configZip = {
        source: 'zip'
    };

    var configRemote = {
        source: 'remote'
    };

    var operationServiceStub;
    var fileUploadServiceStub;
    var fileUploadStub;
    var imTrackerStub;

    var swUpdateExecutor;

    beforeEach(function () {
        operationServiceStub = {
            executeOperation: sinon.stub().returns(Promise.resolve())
        };

        fileUploadServiceStub = {
            startService: sinon.stub().returns(Promise.resolve()),
            stopService: sinon.spy()
        };

        fileUploadStub = sinon.stub().returns(fileUploadServiceStub);

        imTrackerStub = {
            trackByDistName: sinon.stub().returns(Promise.resolve()),
            stopTracking: sinon.spy()
        };

        swUpdateExecutor = swUpdateExecutorFactory(operationServiceStub, fileUploadStub, imTrackerStub);
    });

    describe('newSoftwareUpdateExecutor', function () {
        it('execute create from zip strategy resolve', function (done) {
            var executor = swUpdateExecutor.newSoftwareUpdateExecutor(configZip);

            executor.execute()
                .then(() => {
                    assert(fileUploadServiceStub.startService.calledOnce);
                    assert(imTrackerStub.trackByDistName.calledOnce);
                    assert(operationServiceStub.executeOperation.calledOnce);

                    done();
                });
            assert(fileUploadStub.calledOnce);
        });

        it('execute create from zip strategy file upload rejection', function (done) {
            var executor = swUpdateExecutor.newSoftwareUpdateExecutor(configZip);
            fileUploadServiceStub.startService = sinon.stub().returns(Promise.reject());

            executor.execute()
                .catch(() => {
                    done();
                });
        });

        it('execute create from zip strategy im tracker track by dist name rejection', function (done) {
            var executor = swUpdateExecutor.newSoftwareUpdateExecutor(configZip);
            imTrackerStub.trackByDistName = sinon.stub().returns(Promise.reject());

            executor.execute()
                .catch(() => done());
        });

        it('execute create from zip strategy im tracker operation rejected', function (done) {
            var executor = swUpdateExecutor.newSoftwareUpdateExecutor(configZip);
            operationServiceStub.executeOperation = sinon.stub().returns(Promise.reject());

            executor.execute()
                .catch(() => {
                    assert(fileUploadServiceStub.stopService.calledOnce);
                    assert(imTrackerStub.stopTracking.calledOnce);

                    done();
                });
        });

        it('execute create remote server strategy resolve', function (done) {
            var executor = swUpdateExecutor.newSoftwareUpdateExecutor(configRemote);

            executor.execute()
                .then(() => {
                    assert(operationServiceStub.executeOperation.calledOnce);

                    done();
                });
        });

        it('execute create remote server strategy operation rejected', function (done) {
            var executor = swUpdateExecutor.newSoftwareUpdateExecutor(configRemote);
            operationServiceStub.executeOperation = sinon.stub().returns(Promise.reject());

            executor.execute()
                .catch(() => done());
        });
    });
});
