const testUtil = require('./test.util');
const protoDecoder = require('../src/protoDecoder');
const typeInfo = require('../src/typeInfo');
const typeInfoV2 = require('../src/typeInfoV2');
const objectIdMapper = require('../src/communication/objectIdMapper');

describe('protoDecoder', function () {

    var WIRETYPE_VARINT = 0;
    var WIRETYPE_FIXED64 = 1;
    var WIRETYPE_LENGTH_DELIMITED = 2;

    var theProtoDecoder;
    var idMapper;

    beforeEach(function () {
        idMapper = objectIdMapper.newInstance();

        idMapper.addClassName(0, 'MRBTS');
        idMapper.registerObjectInfos([{
            uuid: '00',
            objectType: 0,
            instanceId: 0
        }]);

        theProtoDecoder = protoDecoder(idMapper);
    });

    describe('simple type', function () {

        it('should decode integer', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var payload = bufferOf([tag(id(1), WIRETYPE_VARINT), 0x96, 0x01]);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), payload);

                assert.deepEqual(parameters.integer, 150);
                done();
            });
        });

        it('should decode string', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var payload = bufferOf(toProtoBytes(id(2), 'Wroclaw'));

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), payload);

                assert.deepEqual(parameters.string, 'Wroclaw');
                done();
            });
        });

        it('should decode objectid with typeinfo V2', function (done) {
            const meta = {
                p: [
                    {'@id': 0, '@type': 'objectid', '@name': 'parameter1'}
                ]
            };
            const payload = bufferOf([tag(id(0), WIRETYPE_LENGTH_DELIMITED), 0x1, 0x0]);

            const parameters = theProtoDecoder.decode(typeInfoV2(meta), payload);

            assert.deepEqual(parameters.parameter1, '/MRBTS-0');
            done();
        });

        it('should decode float', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var payload = bufferOf([
                    tag(id(3), WIRETYPE_FIXED64),
                    0x9A, 0x99, 0x99, 0x99, 0x99, 0x99, 0x28, 0x40
                ]);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), payload);

                assert.deepEqual(parameters.float, 12.3);
                done();
            });
        });

        it('should decode boolean', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var payload = bufferOf([tag(id(4), WIRETYPE_VARINT), 0x1]);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), payload);

                assert.deepEqual(parameters.boolean, true);
                done();
            });
        });

        it('should decode enum', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var payload = bufferOf([tag(id(5), WIRETYPE_VARINT), 0x3]);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), payload);

                assert.deepEqual(parameters.enum, 'Z');
                done();
            });
        });

        it('should decode array', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var payload = bufferOf([
                    tag(id(6), WIRETYPE_VARINT), 0x3,
                    tag(id(6), WIRETYPE_VARINT), 0x9
                ]);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), payload);

                assert.deepEqual(parameters.array, [0x3, 0x9]);
                done();
            });
        });

        it('should decode array without consecutively items', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var payload = bufferOf([
                    tag(id(6), WIRETYPE_VARINT), 0x3,//array item
                    tag(id(1), WIRETYPE_VARINT), 0x7,//other item
                    tag(id(6), WIRETYPE_VARINT), 0x9 //array item
                ]);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), payload);

                assert.deepEqual(parameters.array, [0x3, 0x9]);
                assert.deepEqual(parameters.integer, 0x7);
                done();
            });
        });

        it('should decode sequence of simple types', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var bytes = [];
                bytes.push(tag(id(1), WIRETYPE_VARINT), 0x3);
                bytes = bytes.concat(toProtoBytes(id(2), 'Dabrowa Gornicza'));
                bytes.push(tag(id(6), WIRETYPE_VARINT), 0x7);
                bytes.push(tag(id(6), WIRETYPE_VARINT), 0x9);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), bufferOf(bytes));

                assert.deepEqual(parameters.integer, 0x3);
                assert.deepEqual(parameters.string, 'Dabrowa Gornicza');
                assert.deepEqual(parameters.array, [0x7, 0x9]);
                done();
            });
        });

    });

    describe('complex', function () {

        it('should decode inner simple', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var bytes = [];
                bytes.push(tag(id(7), WIRETYPE_LENGTH_DELIMITED), length(4));
                bytes.push(tag(id(1), WIRETYPE_VARINT), 0x5);
                bytes.push(tag(id(2), WIRETYPE_VARINT), 0x6);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), bufferOf(bytes));

                assert.deepEqual(parameters.complex.integer1, 0x5);
                assert.deepEqual(parameters.complex.integer2, 0x6);
                done();
            });
        });

        it('should decode inner simple + integer', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var bytes = [];
                //complex
                bytes.push(tag(id(7), WIRETYPE_LENGTH_DELIMITED), length(4));
                bytes.push(tag(id(1), WIRETYPE_VARINT), 0x5);
                bytes.push(tag(id(2), WIRETYPE_VARINT), 0x7);
                //integer
                bytes.push(tag(id(1), WIRETYPE_VARINT), 0x8);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), bufferOf(bytes));

                assert.deepEqual(parameters.complex, {integer1: 0x5, integer2: 0x7});
                assert.deepEqual(parameters.integer, 0x8);
                done();
            });
        });

        it('should decode array of inner one parameter', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var bytes = [];
                bytes.push(tag(id(8), WIRETYPE_LENGTH_DELIMITED), length(2));
                bytes.push(tag(id(1), WIRETYPE_VARINT), 0x5);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), bufferOf(bytes));

                assert.deepEqual(parameters.complexArrayInnerOne, [{integer: 0x5}]);
                done();
            });
        });

        it('should decode array of inner two parameters', function (done) {
            loadJson('base/resources/meta/mrbts.json', function (mrbtsMeta) {
                var bytes = [];
                //item 1
                bytes.push(tag(id(9), WIRETYPE_LENGTH_DELIMITED), length(4));
                bytes.push(tag(id(1), WIRETYPE_VARINT), 0x1);
                bytes.push(tag(id(2), WIRETYPE_VARINT), 0x2);
                //item 2
                bytes.push(tag(id(9), WIRETYPE_LENGTH_DELIMITED), length(4));
                bytes.push(tag(id(1), WIRETYPE_VARINT), 0x3);
                bytes.push(tag(id(2), WIRETYPE_VARINT), 0x4);

                var parameters = theProtoDecoder.decode(typeInfo(mrbtsMeta), bufferOf(bytes));

                assert.deepEqual(parameters.complexArrayInnerTwo, [
                    {'integer1': 1, 'integer2': 2},
                    {'integer1': 3, 'integer2': 4}
                ]);

                done();
            });
        });

    });

    describe('validation', function () {

        it('should throw exception when no required simple type', function (done) {
            loadJson('base/resources/meta/mrbts_with_required.json', function (mrbtsMeta) {
                var bytes = [];
                bytes.push(tag(id(1), WIRETYPE_VARINT), 0x5); // integer_optional

                theProtoDecoder.decode(typeInfo(mrbtsMeta), bufferOf(bytes));
            }).catch(function (e) {
                assert.equal(e.message, 'Missing mandatory parameter \'integer_required\' of MRBTS');
                done();
            });
        });

        it('should throw exception when no required simple type in complex type', function (done) {
            loadJson('base/resources/meta/mrbts_with_required.json', function (mrbtsMeta) {
                var bytes = [];
                bytes.push(tag(id(2), WIRETYPE_VARINT), 0x5); // integer_required
                bytes.push(tag(id(3), WIRETYPE_LENGTH_DELIMITED), length(2)); // complex_optional
                bytes.push(tag(id(1), WIRETYPE_VARINT), 0x5); // integer1_inner_required

                theProtoDecoder.decode(typeInfo(mrbtsMeta), bufferOf(bytes));
            }).catch(function (e) {
                assert.equal(e.message, 'Missing mandatory parameter \'integer2_inner_required\' of MRBTS');
                done();
            });
        });

    });

    describe('integration tests', function () {

        it('should decode operations from payload', function (done) {
            loadJson('base/resources/meta/build_data_l.json', function (meta) {
                var uint8 = new Uint8Array([
                    0xa, 0x8, 0x8, 0x1, 0x10, 0x0, 0x18,
                    0x0, 0x20, 0x2, 0x12, 0xc, 0x56,
                    0x45, 0x47, 0x32, 0x34, 0x2e,
                    0x31, 0x30, 0x2e, 0x52, 0x30,
                    0x37, 0x1a, 0xc, 0x56, 0x45,
                    0x47, 0x32, 0x34, 0x2e, 0x31,
                    0x30, 0x2e, 0x52, 0x30, 0x37,
                    0x22, 0x0
                ]);

                var decoded = theProtoDecoder.decode(typeInfo(meta), uint8.buffer);

                assert.ok(decoded.oper_swValid);
                done();
            });
        });

    });

    function loadJson(filePath, doneCallback) {
        return testUtil.loadJsonFile(filePath).then(doneCallback);
    }

    function toProtoBytes(tagId, text) {
        var result = [];
        result.push(tag(tagId, WIRETYPE_LENGTH_DELIMITED));
        result.push(text.length);

        for (var i = 0; i < text.length; i++) {
            result.push(text.charCodeAt(i));
        }
        return result;
    }

    function tag(id, type) {
        return (id << 3) | type;
    }

    function id(number) {
        return number;
    }

    function bufferOf(array) {
        return new Uint8Array(array).buffer;
    }

    function length(number) {
        return number;
    }

});
