const testUtil = require('./test.util');
const protoEncoder = require('../src/protoEncoder');
const typeInfo = require('../src/typeInfo');
const typeInfoV2 = require('../src/typeInfoV2');

describe('protoEncoder', function () {
    var assert = chai.assert;
    var theEncoder = protoEncoder();

    describe('simple types', function () {

        it('should encode empty object', function (done) {
            assertEncodes(
                {},
                []
            ).then(done).catch(done);
        });

        it('should encode integer', function (done) {
            assertEncodes(
                {integer: 150},
                [tag(1).varInt(), 0x96, 0x01]
            ).then(done).catch(done);
        });

        it('should encode float', function (done) {
            assertEncodes(
                {float: 12.3},
                [tag(3).const64(), 0x9A, 0x99, 0x99, 0x99, 0x99, 0x99, 0x28, 0x40]
            ).then(done).catch(done);
        });

        describe('string', function () {

            it('should encode empty', function (done) {
                assertEncodes(
                    {string: ''},
                    [tag(2).lengthDelimited(), length(0x00)]
                ).then(done).catch(done);
            });

            it('should encode', function (done) {
                assertEncodes(
                    {string: 'a Foo'},
                    [tag(2).lengthDelimited(), length(0x05), 0x61, 0x20, 0x46, 0x6F, 0x6F]
                ).then(done).catch(done);
            });

            it('should handle unicode', function (done) {
                assertEncodes(
                    {string: 'ą Fóó'},
                    [tag(2).lengthDelimited(), length(0x08), 0xC4, 0x85, 0x20, 0x46, 0xC3, 0xB3, 0xC3, 0xB3]
                ).then(done).catch(done);
            });
        });

        describe('objectid', function () {

            it('should encode', function () {
                const meta = {p: [
                    {'@id': 0, '@type': 'objectid', '@name': 'parameter1'}
                ]}
                const expected = [tag(0).lengthDelimited(), length(0x08), 0x2F, 0x4D, 0x52, 0x42, 0x54, 0x53, 0x2D, 0x30];
                var encoded = theEncoder.encode(typeInfoV2(meta), {parameter1: '/MRBTS-0'}, true);
                assert.deepEqual(new Uint8Array(encoded), new Uint8Array(expected));
            });

        });

        describe('boolean', function () {

            it('should encode true', function (done) {
                assertEncodes(
                    {boolean: true},
                    [tag(4).varInt(), 0x01]
                ).then(done).catch(done);
            });

            it('should encode false', function (done) {
                assertEncodes(
                    {boolean: false},
                    [tag(4).varInt(), 0x00]
                ).then(done).catch(done);
            });
        });

        it('should encode enum', function (done) {
            assertEncodes(
                {enum: 'Y'},
                [tag(5).varInt(), 0x02]
            ).then(done).catch(done);
        });

        describe('array', function () {

            it('should encode empty', function (done) {
                assertEncodes(
                    {array: []},
                    []
                ).then(done).catch(done);
            });

            it('should encode integers', function (done) {
                var key = tag(6).varInt();
                assertEncodes(
                    {array: [3, 9, 1]},
                    [
                        key, 0x03,
                        key, 0x09,
                        key, 0x01
                    ]
                ).then(done).catch(done);
            });

            it('should encode strings', function (done) {
                var key = tag(11).lengthDelimited();
                assertEncodes(
                    {stringArray: ['a', ' ', 'F', 'o', 'o']},
                    [
                        key, length(0x01), 0x61,
                        key, length(0x01), 0x20,
                        key, length(0x01), 0x46,
                        key, length(0x01), 0x6F,
                        key, length(0x01), 0x6F
                    ]
                ).then(done).catch(done);
            });
        });

        it('should encode all present simple fields', function (done) {
            assertEncodes(
                {
                    integer: 150,
                    float: 12.3,
                    string: 'a Foo',
                    array: [1, 2, 3],
                    enum: 'Y'
                },
                [
                    tag(1).varInt(), 0x96, 0x01,
                    tag(2).lengthDelimited(), length(0x05), 0x61, 0x20, 0x46, 0x6F, 0x6F,
                    tag(3).const64(), 0x9A, 0x99, 0x99, 0x99, 0x99, 0x99, 0x28, 0x40,
                    tag(5).varInt(), 0x02,
                    tag(6).varInt(), 0x01, tag(6).varInt(), 0x02, tag(6).varInt(), 0x03
                ]
            ).then(done).catch(done);
        });
    });

    describe('complex types', function () {

        it('should not encode empty', function (done) {
            assertEncodes(
                {complexOptional: {}},
                [tag(12).lengthDelimited(), length(0x00)]
            ).then(done).catch(done);
        });

        it('should encode partial value', function (done) {
            assertEncodes(
                {
                    complexOptional: {
                        integer2: 42
                    }
                },
                [
                    tag(12).lengthDelimited(), length(0x02),
                    tag(2).varInt(), 0x2A
                ]
            ).then(done).catch(done);
        });

        it('should encode full value', function (done) {
            assertEncodes(
                {
                    complex: {
                        integer1: 4,
                        integer2: 2
                    }
                },
                [
                    tag(7).lengthDelimited(), length(0x04),
                    tag(1).varInt(), 0x04,
                    tag(2).varInt(), 0x02
                ]
            ).then(done).catch(done);
        });

        it('should handle mixed complex and simple values', function (done) {
            assertEncodes(
                {
                    float: 12.3,
                    complex: {
                        integer1: 4,
                        integer2: 2
                    },
                    integer: 42
                },
                [
                    tag(1).varInt(), 0x2A,
                    tag(3).const64(), 0x9A, 0x99, 0x99, 0x99, 0x99, 0x99, 0x28, 0x40,
                    tag(7).lengthDelimited(), length(0x04),
                    tag(1).varInt(), 0x04,
                    tag(2).varInt(), 0x02
                ]
            ).then(done).catch(done);
        });

        it('should encode array', function (done) {
            assertEncodes(
                {
                    complexArrayInnerTwo: [
                        {"integer1": 1, "integer2": 2},
                        {"integer1": 3, "integer2": 4}
                    ]
                },
                [
                    tag(9).lengthDelimited(), length(0x04),
                    tag(1).varInt(), 0x01,
                    tag(2).varInt(), 0x02,

                    tag(9).lengthDelimited(), length(0x04),
                    tag(1).varInt(), 0x03,
                    tag(2).varInt(), 0x04
                ]
            ).then(done).catch(done);
        });

        it('should encode operation', function (done) {
            assertEncodes(
                {
                    oper_setSiteId: {
                        integer: 42
                    }
                },
                [
                    tag(10).lengthDelimited(), length(0x02),
                    tag(1).varInt(), 0x2A
                ]
            ).then(done).catch(done);
        });
    });

    describe('regression', function () {

        it('should be compatible with decoder', function (done) {
            loadTypeInfo('base/resources/meta/mrbts.json').then(function (typeInfo) {
                var decoder = infoModel.protoDecoder();
                var object = {
                    integer: 42,
                    string: "forty two",
                    float: 42.42,
                    boolean: true,
                    enum: 'Z',
                    array: [4, 2],
                    complex: {integer1: 42, integer2: 24},
                    complexArrayInnerOne: [
                        {integer: 101},
                        {integer: 102}
                    ],
                    complexArrayInnerTwo: [
                        {"integer1": 1, "integer2": 2},
                        {"integer1": 3, "integer2": 4}
                    ],
                    oper_setSiteId: {integer: 11},
                    stringArray: ['one', 'two', 'three'],
                    complexOptional: {
                        integer1: 1,
                        integer2: 2
                    }
                };
                var encoded = theEncoder.encode(typeInfo, object);
                assert.deepEqual(
                    object,
                    decoder.decode(typeInfo, encoded)
                );
                done();
            });
        });
    });

    function assertEncodes(object, expected) {
        return loadTypeInfo('base/resources/meta/mrbts.json').then(function (typeInfo) {
            var encoded = theEncoder.encode(typeInfo, object);
            assert.deepEqual(new Uint8Array(encoded), new Uint8Array(expected));
        });
    }

    var tag = function (id) {
        function tag(id, wireType) {
            return (id << 3) | wireType;
        }

        return {
            varInt: function () {
                return tag(id, 0);
            },
            const64: function () {
                return tag(id, 1);
            },
            lengthDelimited: function () {
                return tag(id, 2);
            }
        }
    };

    function loadTypeInfo(file) {
        return testUtil.loadTypeInfo(file);
    }

    function length(length) {
        return length;
    }
});
