const sfpConnectionsFactory = require('../../src/sfp/sfpConnectionsService');
const ImTree = require('../../src/imtree');
const _ = require('lodash');

describe('Service: sfpConnectionsService', () => {

    let imTree, eqmL, hwtopL, updateService, sfpConnectionsService;

    function updateServiceStub() {
        let imTreeListener = null;
        let data = null;

        return {
            addImTreeListener: listener => {
                imTreeListener = listener;
                if (data) imTreeListener.completeImTree(data);
            },
            removeImTreeListener: () => {
                imTreeListener = null;
            },
            pushData: tree => {
                data = tree;
                imTreeListener.completeImTree(tree);
            }
        };
    }

    beforeEach(() => {
        updateService = updateServiceStub();
        sfpConnectionsService = sfpConnectionsFactory.newInstance(updateService);
        imTree = new ImTree();
        eqmL = imTree.mergeNode('/MRBTS-1/RAT-1/EQM_L-1', {});
        hwtopL = imTree.mergeNode('/MRBTS-1/RAT-1/BTS_L-1/HWTOP_L-1', {});
    });

    describe('RMOD-SMOD & RMOD-BBMOD connections', () => {

        beforeEach(() => {
            const rmod = createModule('RMOD_L', 1);
            const smodConnector = createConnectorNode(createModule('SMOD_L', 1), 1);
            const rmodConnector1 = createConnectorNode(rmod, 1);
            const rmodConnector2 = createConnectorNode(rmod, 2);
            const bbmodConnector = createConnectorNode(createModule('BBMOD_L', 1), 1);
            createCablnkNode(rmodConnector1, smodConnector, 1, 1, 1);
            createCablnkNode(rmodConnector2, bbmodConnector, 2, 2, 1);
            createSfpNodeWithParams(smodConnector, 1, 1);
            createSfpNodeWithParams(rmodConnector1, 1, 1);
            createSfpNodeWithParams(rmodConnector2, 1, 1);
            createSfpNodeWithParams(bbmodConnector, 1, 1);

        });

        it('should get exact number of connections', (done) => {
            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                assert.lengthOf(sfpConnectionsService.createConnections(), 2);
                done();
            }
        });

        it('should get RMOD-SMOD connection', (done) => {

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connection = _.find(sfpConnectionsService.createConnections(), {
                    cablnkDistName: '/MRBTS-1/RAT-1/BTS_L-1/HWTOP_L-1/CABLNK_L-1'
                });

                assert.isDefined(connection);
                assert.lengthOf(connection.alarms, 0);
                assert.deepEqual(connection.destination.name, 'SMOD_1');
                assert.deepEqual(connection.destination.distName, '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1');
                assert.deepEqual(connection.destination.status, sfpConnectionsService.STATUSES.OK);
                assert.deepEqual(connection.destination.port, 1);
                assert.deepEqual(connection.destination.parameters.tx.value, 0.001);
                assert.deepEqual(connection.destination.parameters.rx.value, 0.001);
                assert.deepEqual(connection.destination.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                assert.deepEqual(connection.source.name, 'RMOD_1');
                assert.deepEqual(connection.source.distName, '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1');
                assert.deepEqual(connection.source.status, sfpConnectionsService.STATUSES.OK);
                assert.deepEqual(connection.source.port, 1);
                assert.deepEqual(connection.source.parameters.tx.value, 0.001);
                assert.deepEqual(connection.source.parameters.rx.value, 0.001);
                assert.deepEqual(connection.source.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                done();
            }
        });

        it('should get RMOD-BBMOD connection', (done) => {

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connection = _.find(sfpConnectionsService.createConnections(), {
                    cablnkDistName: '/MRBTS-1/RAT-1/BTS_L-1/HWTOP_L-1/CABLNK_L-2'
                });

                assert.isDefined(connection);
                assert.lengthOf(connection.alarms, 0);
                assert.deepEqual(connection.destination.name, 'BBMOD_1');
                assert.deepEqual(connection.destination.distName, '/MRBTS-1/RAT-1/EQM_L-1/BBMOD_L-1');
                assert.deepEqual(connection.destination.status, sfpConnectionsService.STATUSES.OK);
                assert.deepEqual(connection.destination.port, 1);
                assert.deepEqual(connection.destination.parameters.tx.value, 0.001);
                assert.deepEqual(connection.destination.parameters.rx.value, 0.001);
                assert.deepEqual(connection.destination.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/BBMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                assert.deepEqual(connection.source.name, 'RMOD_1', 'rrh name');
                assert.deepEqual(connection.source.distName, '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1');
                assert.deepEqual(connection.source.status, sfpConnectionsService.STATUSES.OK);
                assert.deepEqual(connection.source.port, 2);
                assert.deepEqual(connection.source.parameters.tx.value, 0.001);
                assert.deepEqual(connection.source.parameters.rx.value, 0.001);
                assert.deepEqual(connection.source.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1/CONNECTOR_L-2/SFP_L-1/SFP_PARAMS_L-1');
                done();
            }
        });

    });

    describe('radio chaining', () => {

        beforeEach(() => {
            const rmod1 = createModule('RMOD_L', 1);
            const rmod2 = createModule('RMOD_L', 2);
            const rmod3 = createModule('RMOD_L', 3);
            const rmod1Connector1 = createConnectorNode(rmod1, 1);
            const rmod2Connector1 = createConnectorNode(rmod2, 1);
            const rmod2Connector2 = createConnectorNode(rmod2, 2);
            const rmod3Connector1 = createConnectorNode(rmod3, 1);
            createCablnkNode(rmod1Connector1, rmod2Connector1, 1, 1, 1);
            createCablnkNode(rmod2Connector2, rmod3Connector1, 2, 2, 1);
            createSfpNodeWithParams(rmod1Connector1, 1, 1);
            createSfpNodeWithParams(rmod2Connector1, 1, 1);
            createSfpNodeWithParams(rmod2Connector2, 1, 1);
            createSfpNodeWithParams(rmod3Connector1, 1, 1);
        });

        it('should get exact number of connections', (done) => {
            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                assert.lengthOf(sfpConnectionsService.createConnections(), 2);
                done();
            }
        });

        it('should get first RMOD-RMOD connection', (done) => {

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connection = _.find(sfpConnectionsService.createConnections(), {
                    cablnkDistName: '/MRBTS-1/RAT-1/BTS_L-1/HWTOP_L-1/CABLNK_L-1'
                });

                assert.isDefined(connection);
                assert.lengthOf(connection.alarms, 0);
                assert.deepEqual(connection.destination.name, 'RMOD_2');
                assert.deepEqual(connection.destination.distName, '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-2');
                assert.deepEqual(connection.destination.status, sfpConnectionsService.STATUSES.OK);
                assert.deepEqual(connection.destination.port, 1);
                assert.deepEqual(connection.destination.parameters.tx.value, 0.001);
                assert.deepEqual(connection.destination.parameters.rx.value, 0.001);
                assert.deepEqual(connection.destination.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-2/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                assert.deepEqual(connection.source.name, 'RMOD_1');
                assert.deepEqual(connection.source.distName, '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1');
                assert.deepEqual(connection.source.status, sfpConnectionsService.STATUSES.OK);
                assert.deepEqual(connection.source.port, 1);
                assert.deepEqual(connection.source.parameters.tx.value, 0.001);
                assert.deepEqual(connection.source.parameters.rx.value, 0.001);
                assert.deepEqual(connection.source.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                done();
            }

        });

        it('should get second RMOD-RMOD connection', (done) => {

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connection = _.find(sfpConnectionsService.createConnections(), {
                    cablnkDistName: '/MRBTS-1/RAT-1/BTS_L-1/HWTOP_L-1/CABLNK_L-2'
                });

                assert.isDefined(connection);
                assert.lengthOf(connection.alarms, 0);
                assert.deepEqual(connection.destination.name, 'RMOD_3');
                assert.deepEqual(connection.destination.distName, '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-3');
                assert.deepEqual(connection.destination.status, sfpConnectionsService.STATUSES.OK);
                assert.deepEqual(connection.destination.port, 1);
                assert.deepEqual(connection.destination.parameters.tx.value, 0.001);
                assert.deepEqual(connection.destination.parameters.rx.value, 0.001);
                assert.deepEqual(connection.destination.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-3/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                assert.deepEqual(connection.source.name, 'RMOD_2');
                assert.deepEqual(connection.source.distName, '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-2');
                assert.deepEqual(connection.source.status, sfpConnectionsService.STATUSES.OK);
                assert.deepEqual(connection.source.port, 2);
                assert.deepEqual(connection.source.parameters.tx.value, 0.001);
                assert.deepEqual(connection.source.parameters.rx.value, 0.001);
                assert.deepEqual(connection.source.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-2/CONNECTOR_L-2/SFP_L-1/SFP_PARAMS_L-1');
                done();
            }

        });
    });

    describe('error status', () => {
        let rmodConnector, smodConnector, cablnk;

        beforeEach(() => {
            smodConnector = createConnectorNode(createModule('SMOD_L', 1), 1);
            rmodConnector = createConnectorNode(createModule('RMOD_L', 1), 1);
            cablnk = createCablnkNode(rmodConnector, smodConnector, 1, 1, 1);
            createSfpNodeWithParams(smodConnector, 1, 1);
            createSfpNodeWithParams(rmodConnector, 1, 1);
        });

        it('should get error status when source (RMOD) connector is disabled', (done) => {
            rmodConnector.parameters.stateInfo.operationalState = 'Disabled';

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].destination.status, sfpConnectionsService.STATUSES.OK);
                assert.deepEqual(connections[0].source.status, sfpConnectionsService.STATUSES.ERROR);
                done();
            }
        });

        it('should get error status when destination (SMOD) connector is disabled', (done) => {
            smodConnector.parameters.stateInfo.operationalState = 'Disabled';

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].destination.status, sfpConnectionsService.STATUSES.ERROR);
                assert.deepEqual(connections[0].source.status, sfpConnectionsService.STATUSES.OK);
                done();
            }
        });

        it('should get error status when cablnk is disabled', (done) => {
            cablnk.parameters.stateInfo.operationalState = 'Disabled';
            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].destination.status, sfpConnectionsService.STATUSES.ERROR);
                assert.deepEqual(connections[0].source.status, sfpConnectionsService.STATUSES.ERROR);
                done();
            }
        });

    });

    describe('sfp parameters', () => {
        let smodSfp, rmodSfp;

        beforeEach(() => {
            const smodConnector = createConnectorNode(createModule('SMOD_L', 1), 1);
            const rmodConnector = createConnectorNode(createModule('RMOD_L', 1), 1);
            createCablnkNode(rmodConnector, smodConnector, 1, 1, 1);
            smodSfp = createSfpNode(smodConnector);
            rmodSfp = createSfpNode(rmodConnector);
        });

        it('should return SFP_PARAMS as sfp distName', (done) => {
            createSfpParamsNode(rmodSfp, 1, 1);
            createSfpParamsNode(smodSfp, 1, 1);

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].source.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                assert.deepEqual(connections[0].destination.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/CONNECTOR_L-1/SFP_L-1/SFP_PARAMS_L-1');
                done();
            }
        });

        it('should return SFP as sfp distName when SFP_PARAMS are not available', (done) => {
            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].source.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/RMOD_L-1/CONNECTOR_L-1/SFP_L-1');
                assert.deepEqual(connections[0].destination.parameters.distName,
                    '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/CONNECTOR_L-1/SFP_L-1');
                done();
            }
        });
    });

    describe('connection link type', () => {
        let cablnk;

        beforeEach(() => {
            const smodConnector = createConnectorNode(createModule('SMOD_L', 1), 1);
            const rmodConnector = createConnectorNode(createModule('RMOD_L', 1), 1);
            cablnk = createCablnkNode(rmodConnector, smodConnector, 1, 1, 1);
            createSfpNodeWithParams(smodConnector, 1, 1);
            createSfpNodeWithParams(rmodConnector, 1, 1);
        });

        it('should find RP3_01Link connection', (done) => {
            cablnk.parameters.cablinkType = 'RP3_01Link';
            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].source.name, 'RMOD_1');
                assert.deepEqual(connections[0].destination.name, 'SMOD_1');
                done();
            }

        });

        it('should find CPRILink connection', (done) => {
            cablnk.parameters.cablinkType = 'CPRILink';

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].source.name, 'RMOD_1');
                assert.deepEqual(connections[0].destination.name, 'SMOD_1');
                done();
            }
        });

        it('should not find RP3Link connection', (done) => {
            cablnk.parameters.cablinkType = 'RP3Link';
            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 0);
                done();
            }
        });

    });

    describe('modules types', () => {

        it('should find connection with SMOD_L module', (done) => {
            createConnectionWithModule('SMOD_L');

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].destination.name, 'SMOD_1');
                done();
            }
        });

        it('should find connection with RMOD_L module', (done) => {
            createConnectionWithModule('RMOD_L');

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].destination.name, 'RMOD_1');
                done();
            }
        });

        it('should find connection with SMOD_EXT_L module', (done) => {
            createConnectionWithModule('SMOD_EXT_L');

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].destination.name, 'SMOD_EXT_1');
                done();
            }
        });


        it('should find connection with BBMOD_L module', (done) => {
            createConnectionWithModule('BBMOD_L');

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 1);
                assert.deepEqual(connections[0].destination.name, 'BBMOD_1');
                done();
            }
        });

        it('should not find connection with SOME_OTHER_MODULE_L module', (done) => {
            createConnectionWithModule('SOME_OTHER_MODULE_L');

            sfpConnectionsService.watchSfpConnections(sfpConnectionsServiceListener);
            updateService.pushData(imTree);

            function sfpConnectionsServiceListener() {
                const connections = sfpConnectionsService.createConnections();
                assert.lengthOf(connections, 0);
                done();
            }
        });

        function createConnectionWithModule(moduleName) {
            const moduleConnector = createConnectorNode(createModule(moduleName, 1), 1);
            const rmodConnector = createConnectorNode(createModule('RMOD_L', 1), 1);
            createCablnkNode(rmodConnector, moduleConnector, 1, 1, 1);
            createSfpNodeWithParams(moduleConnector, 1, 1);
            createSfpNodeWithParams(rmodConnector, 1, 1);
        }

    });

    function createModule(className, id, parameters, parentNode) {
        parentNode = parentNode || eqmL;
        const moduleNode = imTree.addNode(parentNode, className, id);
        moduleNode.parameters = parameters || {};
        return moduleNode;
    }

    function createConnectorNode(module, id, state) {
        if (!state) state = 'Enabled';
        const connector = imTree.addNode(module, 'CONNECTOR_L', id);
        connector.parameters = {
            stateInfo: {operationalState: state}
        };
        return connector;
    }

    function createCablnkNode(rmodConnector, destModuleConnector, id, rmodPort, destinationPort, state, type) {
        if (!type) type = 'RP3_01Link';
        if (!state) state = 'Enabled';
        const cablnk = imTree.addNode(hwtopL, 'CABLNK_L', id);
        cablnk.parameters = {
            cablinkType: type,
            destinationDistName: destModuleConnector.distName,
            sourceDistName: rmodConnector.distName,
            sourcePhysicalPortNo: rmodPort,
            destinationPhysicalPortNo: destinationPort,
            stateInfo: {operationalState: state}
        };
        return cablnk;
    }

    function createSfpNode(connector) {
        return imTree.addNode(connector, 'SFP_L', 1);
    }

    function createSfpParamsNode(sfp, rxPower, txPower) {
        const sfpParamsNode = imTree.addNode(sfp, 'SFP_PARAMS_L', 1);
        sfpParamsNode.parameters = {
            rxPower: rxPower,
            txPower: txPower
        };
        return sfpParamsNode;
    }

    function createSfpNodeWithParams(connector, rxPower, txPower) {
        const sfp = createSfpNode(connector);
        createSfpParamsNode(sfp, rxPower, txPower);
        return sfp;
    }

});

