describe('sic', function () {
    var sic;

    var log = loggers.getLogger('imc.lib.sic');

    beforeEach(function () {
        sic = infoModel.sic();
    });

    describe('header', function () {
        var header;

        beforeEach(function () {
            header = sic.createHeader();
        });

        it('should set id', function () {
            header.setId(7);

            assert.deepEqual(header.getId(), 7);
        });

        it('should default falgs be set to 8 which is big endian', function () {
            assert.deepEqual(header.getFlags(), 8);
        });

        it('should default sender be set to 100', function () {
            assert.deepEqual(header.getSender(), 100);
        });

        it('should convert header to big endian buffer', function () {
            header.setId(1);
            header.setReceiver(2);
            header.setSender(3);
            header.setPayloadLength(0);
            header.setFlags(5);

            var expected = new Uint8Array(16);
            //id
            expected[0] = 0;
            expected[1] = 0;
            expected[2] = 0;
            expected[3] = 1;
            //receiver
            expected[4] = 0;
            expected[5] = 0;
            expected[6] = 0;
            expected[7] = 2;
            //sender
            expected[8] = 0;
            expected[9] = 0;
            expected[10] = 0;
            expected[11] = 3;
            //length
            expected[12] = 0;
            expected[13] = sic.HEADER_SIZE + header.getPayloadLength();
            //flags
            expected[14] = 0;
            expected[15] = 5;

            assert.deepEqual(header.toUint8Array(), expected);
        });

        it('should convert big endian buffer to header', function () {
            var buffer = new ArrayBuffer(16);
            var array = new Uint8Array(buffer);
            //id
            array[0] = 0;
            array[1] = 0;
            array[2] = 0;
            array[3] = 1;
            //receiver
            array[4] = 0;
            array[5] = 0;
            array[6] = 0;
            array[7] = 2;
            //sender
            array[8] = 0;
            array[9] = 0;
            array[10] = 0;
            array[11] = 3;
            //length
            array[12] = 0;
            array[13] = 4;
            //flags
            array[14] = 0;
            array[15] = 5;

            header.setBuffer(buffer);

            assert.deepEqual(header.getId(), 1);
            assert.deepEqual(header.getReceiver(), 2);
            assert.deepEqual(header.getSender(), 3);
            assert.deepEqual(header.getLength(), 4);
            assert.deepEqual(header.getFlags(), 5);
        });

    });

    describe('message', function () {
        var header;
        var message;

        beforeEach(function () {
            header = sic.createHeader();
            header.setPayloadLength(0);

            message = sic.createMessage(header);
        });

        it('should get message buffer', function () {
            assert(message.getBuffer());
        });

        it('should get header size when message length is 0', function () {
            header.setPayloadLength(0);

            assert.deepEqual(message.toUint8Array().length, sic.HEADER_SIZE);
        });

        it('message size be equal to header size + payload size', function () {
            message.setPayload(new ArrayBuffer(8));

            assert.deepEqual(message.toUint8Array().length, sic.HEADER_SIZE + 8);
        });

        it('setting payload should also change header.length = payload.length', function () {
            message.setPayload(new ArrayBuffer(8));

            assert.deepEqual(header.getLength(), sic.HEADER_SIZE + 8);
        });

    });

    describe('sicPlusMessage', function () {
        var header;
        var message;

        beforeEach(function () {
            header = sic.createHeader();
            header.setPayloadLength(0);

            message = sic.createSicPlusMessage(header);
        });

        it('should get message buffer', function () {
            assert(message.getBuffer());
        });

        it('should get sic plus message id', function () {
            assert.deepEqual(message.getId(), sic.SIC_PLUS_ID);
        });

        it('should set seq id', function () {
            message.setSeqId(1);

            assert.deepEqual(message.getSeqId(), 1);
        });

        it('should set data offset', function () {
            message.setDataOffset(2);

            assert.deepEqual(message.getDataOffset(), 2);
        });

        it('should set msg id', function () {
            message.setMsgId(3);

            assert.deepEqual(message.getMsgId(), 3);
        });

        it('should set msg size', function () {
            message.setMsgSize(4);

            assert.deepEqual(message.getMsgSize(), 4);
        });

        it('should set data', function () {
            var data = new Uint8Array(2);
            data[0] = 1;
            data[1] = 1;

            message.setData(data);

            assert.deepEqual(message.getDataSize(), data.length);
            assert.deepEqual(message.getData(), data);
        });

        it('should set buffer', function () {
            message.setSeqId(1);
            message.setDataOffset(2);
            message.setMsgId(3);
            message.setMsgSize(4);

            var data = new Uint8Array(2);
            data[0] = 1;
            data[1] = 1;
            message.setData(data);

            log.debug('buffer length: ' + message.getBuffer().byteLength);

            var newMessage = sic.createSicPlusMessage(header);
            newMessage.setBuffer(message.getBuffer());

            assert.deepEqual(message.getSeqId(), newMessage.getSeqId());
            assert.deepEqual(message.getDataOffset(), newMessage.getDataOffset());
            assert.deepEqual(message.getMsgId(), newMessage.getMsgId());
            assert.deepEqual(message.getMsgSize(), newMessage.getMsgSize());
            assert.deepEqual(message.getDataSize(), newMessage.getDataSize());
            assert.deepEqual(message.getData(), newMessage.getData());
        });

    });

    describe('imi messages', function () {

        it('should have aaConfigGetServiceReq', function () {
            assert.equal(sic.aaConfigGetServiceReq().getId(), 0x1F24);
        });

        it('should have aaConfigGetServiceRsp', function () {
            assert.equal(sic.aaConfigGetServiceRsp().getId(), 0x1F25);
        });

        it('should have aaSysComEcho', function () {
            assert.equal(sic.aaSysComEcho().getId(), 0x123);
        });

        it('should have meta', function () {
            assert.equal(sic.imgwGetMetaResp().getId(), 0x6013);
        });

        it('should have signIn', function () {
            assert.equal(sic.signIn().getId(), 0x3621);
        });

        it('should have signOut', function () {
            assert.equal(sic.signOut().getId(), 0x3622);
        });

        it('should have changeRequest', function () {
            assert.equal(sic.changeRequest().getId(), 0x3624);
        });

        it('should have operationExecuted', function () {
            assert.equal(sic.operationExecuted().getId(), 0x3625);
        });
    });

    describe('createMessage', function () {

        it('should create message from array buffer', function () {
            var message = sic.aaConfigGetServiceReq();
            message.setName("info model");
            var buffer = message.getBuffer();

            var createdMessage = sic.toMessage(message.getId(), buffer);

            assert.deepEqual(createdMessage.getId(), message.getId());
            assert.deepEqual(createdMessage.getName(), message.getName());
        });

        it('should create defined messages from array buffer', function () {
            for (var key in sic.definedMessages) {
                var messageConstructor = sic.definedMessages[key];
                var message = messageConstructor();

                var buffer = message.getBuffer();
                var createdMessage = sic.toMessage(key, buffer);

                assert.deepEqual(createdMessage.getId(), message.getId());
            }
        });

        it('should throw exc when message id is not known', function () {
            var fakeHeader = sic.createHeader();
            fakeHeader.setId(0xDEAD);
            var fakeMessage = sic.createMessage(fakeHeader);
            var buffer = fakeMessage.getBuffer();

            var createdMessage;
            try {
                createdMessage = sic.createMessage(buffer);
            } catch (e) {
            }

            assert(!createdMessage);
        });
    });

    describe('sic message buffer', function () {
        var messageBuffer;
        var listener;

        beforeEach(function () {
            listener = {};
            listener.headers = [];
            listener.messages = [];
            listener.completedMessages = [];
            listener.createdHeader = function (header) {
                listener.headers.push(header);
            };
            listener.createdMessage = function (message) {
                listener.messages.push(message);
            };
            listener.completedMessage = function (message) {
                listener.completedMessages.push(message);
            };

            messageBuffer = sic.createSicMessageBuffer(listener);
        });

        it('should notify about created header', function () {
            var header = sic.createHeader();
            header.setPayloadLength(0);

            messageBuffer.push(header.getBuffer());

            assert.deepEqual(listener.headers.length, 1);
        });

        it('should notify about message', function () {
            var header = sic.createHeader();
            var message = sic.createMessage(header);
            message.setPayload(new ArrayBuffer(2));

            messageBuffer.push(message.getBuffer());

            assert.deepEqual(listener.headers.length, 1);
            assert.deepEqual(listener.messages.length, 1);
        });

        it('should notify about Nth messages', function () {
            var n = 5;
            for (var i = 0; i < n; i++) {
                var header = sic.createHeader();
                var message = sic.createMessage(header);
                message.setPayload(new ArrayBuffer(2));
                messageBuffer.push(message.getBuffer());
            }

            assert.deepEqual(listener.headers.length, n);
            assert.deepEqual(listener.messages.length, n);
        });

        it('should notify about Nth messages from large buffer', function () {
            var n = 5;
            var buffer = new ArrayBuffer(0);
            for (var i = 0; i < n; i++) {
                var header = sic.createHeader();
                var message = sic.createMessage(header);
                message.setPayload(new ArrayBuffer(2));
                buffer = concat(buffer, message.getBuffer());
            }

            messageBuffer.push(buffer);

            assert.deepEqual(listener.headers.length, n);
            assert.deepEqual(listener.messages.length, n);
        });

        it('should notify about completed message', function () {
            var header = sic.createHeader();
            var message = sic.createMessage(header);
            var signIn = sic.signIn();
            header.setId(signIn.getId());
            message.setPayload(signIn.getBuffer());

            messageBuffer.push(message.getBuffer());

            assert.deepEqual(listener.headers.length, 1);
            assert.deepEqual(listener.messages.length, 1);
            assert.deepEqual(listener.completedMessages.length, 1);
        });

        it('should notify about completed sic message', function () {
            const seqId = 1;
            const sender = 52654;
            const meta = sic.imgwGetMetaResp();
            const metaBuffer = meta.getBuffer();
            const msgSize = metaBuffer.byteLength;
            const payload = new Uint8Array(metaBuffer);

            for (var i = 0; i < msgSize; i++) {
                var m = sic.createSicPlusMessage();
                m.setSeqId(seqId);
                m.setSender(sender);
                m.setMsgId(meta.getId());
                m.setMsgSize(msgSize);
                m.setDataOffset(i);
                var data = new Uint8Array(1);
                data.set([payload[i]]);
                m.setData(data);
                messageBuffer.push(m.getBuffer());
            }
            assert.deepEqual(listener.messages.length, msgSize);
            assert.deepEqual(listener.completedMessages.length, 1);
        });
    });

    function concat(arrayBuffer1, arrayBuffer2) {
        var a = new Int8Array(arrayBuffer1);
        var b = new Int8Array(arrayBuffer2);

        var result = new Int8Array(a.length + b.length);
        result.set(a);
        result.set(b, a.length);
        return result.buffer;
    }
});
