const siteCellFactoryService = require('../../../src/site/factories/siteCellFactory'),
    ImTree = require('../../../src/imtree'),
    imTreeNodeFactoryService = require('../../../src/site/imTreeNodeFactory');

(function () {
    'use strict';
    describe('siteCellFactory', function () {

        var factory,
            imTree,
            imTreeNodeFactory;

        var lnbts;

        beforeEach(function () {
            factory = siteCellFactoryService;
            imTreeNodeFactory = imTreeNodeFactoryService;

            imTree = new ImTree();
            lnbts = imTree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1', {});
        });

        it('should create cell', function () {
            createCell(1);

            const cells = factory.createCellsAndLogLinks(imTree).cells;

            assert.deepEqual(cells.length, 1);

            assert.deepEqual(cells[0].distName, '/MRBTS-1/RAT-1/LNBTS-1/LCELL-1');
            assert.deepEqual(cells[0].name, 'LCELL-1');
            assert.deepEqual(cells[0].number, 1);
        });

        describe('channels', function () {

            it('should create cell with empty channels', function () {
                createCell(1);

                const cell = factory.createCellsAndLogLinks(imTree).cells[0];

                assert.deepEqual(cell.subCells[0].rxChannels, []);
                assert.deepEqual(cell.subCells[0].txChannels, []);
            });

            it('should create rx channel', function () {
                const cellNo = 1;
                createCell(cellNo);
                createChannel(cellNo, 1, {direction: 'RX'});

                const cell = factory.createCellsAndLogLinks(imTree).cells[0];
                const rxChannel = cell.subCells[0].rxChannels[0];

                assert.deepEqual(rxChannel.distName, '/MRBTS-1/RAT-1/LNBTS-1/LCELL-1/SUBCELL-0/CHANNEL-1');
                assert.deepEqual(rxChannel.name, 'RX #1');
                assert.deepEqual(rxChannel.number, 1);
                assert.deepEqual(cell.subCells[0].txChannels, []);
            });

            it('should create tx channel', function () {
                const cellNo = 1;
                createCell(cellNo);
                createChannel(cellNo, 1, {direction: 'TX'});

                const cell = factory.createCellsAndLogLinks(imTree).cells[0];

                assert.deepEqual(cell.subCells[0].txChannels.length, 1);
                assert.deepEqual(cell.subCells[0].rxChannels, []);
            });

            it('should sort channels by number', function () {
                const cellNo = 1;
                createCell(cellNo);
                createChannel(cellNo, 3, {direction: 'TX'});
                createChannel(cellNo, 1, {direction: 'TX'});
                createChannel(cellNo, 2, {direction: 'TX'});

                const channels = factory.createCellsAndLogLinks(imTree).cells[0].subCells[0].txChannels;

                assert.deepEqual(channels[0].number, 1);
                assert.deepEqual(channels[1].number, 2);
                assert.deepEqual(channels[2].number, 3);
            });

            it('should create log link from phyant to channel', function () {
                const cellNo = 1;
                createCell(cellNo);
                createChannel(cellNo, 1, {direction: 'RX', AntennaDN: '/MRBTS-1/RAT-1/EQM_L-1/PHYANT_L-1'});

                const logLink = factory.createCellsAndLogLinks(imTree).logLinks[0];

                assert.deepEqual(logLink.source.distName, '/MRBTS-1/RAT-1/LNBTS-1/LCELL-1/SUBCELL-0/CHANNEL-1');
                assert.deepEqual(logLink.source.direction, 'Left');
                assert.deepEqual(logLink.target.distName, '/MRBTS-1/RAT-1/EQM_L-1/PHYANT_L-1');
                assert.deepEqual(logLink.target.direction, 'Right');
            });

            it('should create log link from channel to phyant', function () {
                const cellNo = 1;
                createCell(cellNo);
                createChannel(cellNo, 1, {direction: 'TX', AntennaDN: '/MRBTS-1/RAT-1/EQM_L-1/PHYANT_L-1'});

                const logLink = factory.createCellsAndLogLinks(imTree).logLinks[0];

                assert.deepEqual(logLink.source.distName, '/MRBTS-1/RAT-1/EQM_L-1/PHYANT_L-1');
                assert.deepEqual(logLink.source.direction, 'Right');
                assert.deepEqual(logLink.target.distName, '/MRBTS-1/RAT-1/LNBTS-1/LCELL-1/SUBCELL-0/CHANNEL-1');
                assert.deepEqual(logLink.target.direction, 'Left');
            });

        });

        function createCell(cellNumber, parameters) {
            return imTreeNodeFactory.createCell(imTree, lnbts, cellNumber, parameters);
        }

        function createChannel(parentCellNumber, channelNumber, parameters) {
            return imTreeNodeFactory.createChannel(imTree, lnbts, parentCellNumber, channelNumber, parameters);
        }
    });

})();

