const siteConnectorsFactory = require('../../../src/site/factories/siteConnectorsFactory'),
    ImTree = require('../../../src/imtree'),
    imTreeNodeFactoryService = require('../../../src/site/imTreeNodeFactory');

(function () {
    'use strict';
    describe('siteConnectorsFactory', function () {

        var factory,
            imTree,
            imTreeNodeFactory;

        var eqmL;

        beforeEach(function () {
            factory = siteConnectorsFactory;
            imTreeNodeFactory = imTreeNodeFactoryService;

            imTree = new ImTree();
            eqmL = imTree.mergeNode('/MRBTS-1/RAT-1/EQM_L-1', {});
        });


        it('should create empty connectors', function () {
            const smodNode = createModule('SMOD_L', 1);

            const connectors = factory.createConnectors(smodNode);

            assert.deepEqual(connectors.left, []);
            assert.deepEqual(connectors.right, []);
        });

        it('should create only right connectors for smod', function () {
            const smodNode = createModule('SMOD_L', 1);
            createBackEndConnector(smodNode, 1);
            createFrontEndConnector(smodNode, 2);

            const connectors = factory.createConnectors(smodNode);

            assert.deepEqual(connectors.right, [{
                alarms: [],
                distName: '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/CONNECTOR_L-1',
                label: '1',
                number: 1,
                stateInfo: 'stateInfo'
            }, {
                alarms: [],
                distName: '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/CONNECTOR_L-2',
                label: '2',
                number: 2,
                stateInfo: 'stateInfo'
            }
            ]);
        });

        it('should create only right connectors for pwrmod', function () {
            const pwrmodNode = createModule('PWRMOD_L', 1);
            createBackEndConnector(pwrmodNode, 1);
            createFrontEndConnector(pwrmodNode, 2);

            const connectors = factory.createConnectors(pwrmodNode);

            assert.deepEqual(connectors.right, [{
                alarms: [],
                distName: '/MRBTS-1/RAT-1/EQM_L-1/PWRMOD_L-1/CONNECTOR_L-1',
                label: '1',
                number: 1,
                stateInfo: 'stateInfo'
            }, {
                alarms: [],
                distName: '/MRBTS-1/RAT-1/EQM_L-1/PWRMOD_L-1/CONNECTOR_L-2',
                label: '2',
                number: 2,
                stateInfo: 'stateInfo'
            }
            ]);
        });

        it('should left connectors be sorted by label', function () {
            const moduleNode = createModule('RMOD_L', 1);
            createFrontEndConnector(moduleNode, 4);
            createFrontEndConnector(moduleNode, 3);
            createFrontEndConnector(moduleNode, 2);
            createFrontEndConnector(moduleNode, 1);

            const connectors = factory.createConnectors(moduleNode);

            assert.deepEqual(connectors.left[0].label, '1');
            assert.deepEqual(connectors.left[1].label, '2');
            assert.deepEqual(connectors.left[2].label, '3');
            assert.deepEqual(connectors.left[3].label, '4');
        });

        it('should right connectors be sorted by number', function () {
            const moduleNode = createModule('SMOD_L', 1);
            createBackEndConnector(moduleNode, 4);
            createBackEndConnector(moduleNode, 3);
            createBackEndConnector(moduleNode, 2);
            createBackEndConnector(moduleNode, 1);

            const connectors = factory.createConnectors(moduleNode);

            assert.deepEqual(connectors.right[0].number, 1);
            assert.deepEqual(connectors.right[1].number, 2);
            assert.deepEqual(connectors.right[2].number, 3);
            assert.deepEqual(connectors.right[3].number, 4);
        });

        it('should create connectors representing qsfp', function () {
            const moduleNode = createModule('SMOD_L', 1);
            createQfspNode(moduleNode, 1);
            createQfspNode(moduleNode, 2);

            const connectors = factory.createConnectors(moduleNode);

            assert.deepEqual(connectors.right[0], {
                distName: '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/QSFP_L-1',
                label: 'BBEXT-1',
                alarms: [],
                stateInfo: 'stateInfo'
            });
            assert.deepEqual(connectors.right[1], {
                distName: '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/QSFP_L-2',
                label: 'BBEXT-2',
                alarms: [],
                stateInfo: 'stateInfo'
            });
        });

        function createModule(className, id, parameters, parentNode) {
            return imTreeNodeFactory.createModule(imTree, eqmL, className, id, parameters, parentNode);
        }

        function createFrontEndConnector(parentNode, connectorNumber) {
            return imTreeNodeFactory.createFrontEndConnector(imTree, parentNode, connectorNumber);
        }

        function createBackEndConnector(parentNode, connectorNumber) {
            return imTreeNodeFactory.createBackEndConnector(imTree, parentNode, connectorNumber);
        }

        function createQfspNode(module, id) {
            return imTreeNodeFactory.createQfspNode(imTree, module, id);
        }
    });

})();
