const siteMeasurementsFactoryProvider = require('../../../src/site/factories/siteMeasurementsFactory'),
    ImTree = require('../../../src/imtree');

describe('measurements', function () {
    'use strict';

    const DIST_NAME_MODULE_ROOT = '/SMOD_L-1';

    var siteMeasurementsFactory, imTree;
    beforeEach(function () {
        siteMeasurementsFactory = siteMeasurementsFactoryProvider;
        imTree = new ImTree();
        addParametersToModule({});
    });

    describe('all measurements', function () {

        beforeEach(inject(function () {
            addParametersToModule('/CCU_L-1/SOC_L-1/CPU_L-1', {
                cpuLoad: 65535 * 2
            });
            addParametersToModule('/CCU_L-1/SOC_L-1/CPU_L-1/MEMORY_L-1', {
                usedCapacity: 5,
                totalCapacity: 10
            });
            addParametersToModule('/CCU_L-1/STORAGE_L-1/FILE_SYSTEM_L-1', {
                usedCapacity: 2,
                totalCapacity: 10,
                mountPoint: '/ffs/fs1'
            });
            addParametersToModule('/CCU_L-1/STORAGE_L-1/FILE_SYSTEM_L-2', {
                usedCapacity: 1,
                totalCapacity: 10,
                mountPoint: '/ffs/fs2'
            });
        }));

        it('should calculate all measurements', function () {
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createMeasurements(moduleNode);
            // then
            assert.equal(measurements[0].name, 'CPU');
            assert.equal(measurements[0].value, 2);

            assert.equal(measurements[1].name, 'RAM');
            assert.equal(measurements[1].value, 50);

            assert.equal(measurements[2].name, '/ffs/fs1');
            assert.equal(measurements[2].value, 20);

            assert.equal(measurements[3].name, '/ffs/fs2');
            assert.equal(measurements[3].value, 10);
        });

    });

    describe('cpu', function () {

        it('should calculate cpu usage when CPU is direct child of module', function () {
            // given
            addParametersToModule('/CPU_L-1', {
                cpuLoad: 4 * siteMeasurementsFactory.CPU_LOAD_DIVIDER
            });
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createCpuUsage(moduleNode);
            // then
            assert.deepEqual(measurements, [{name: 'CPU', value: 4}]);
        });

        it('should calculate cpu usage when CPU is direct child of SOC_L', function () {
            // given
            addParametersToModule('/SOC_L-1/CPU_L-1', {
                cpuLoad: 4 * siteMeasurementsFactory.CPU_LOAD_DIVIDER
            });
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createCpuUsage(moduleNode);
            // then
            assert.deepEqual(measurements, [{name: 'CPU', value: 4}]);
        });

        it('should calculate cpu usage when CPU is direct child of CCU', function () {
            // given
            addParametersToModule('/CCU_L-1/CPU_L-1', {
                cpuLoad: 4 * siteMeasurementsFactory.CPU_LOAD_DIVIDER
            });
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createCpuUsage(moduleNode);
            // then
            assert.deepEqual(measurements, [{name: 'CPU', value: 4}]);
        });

        it('should not calculate cpu usage - no CCU_L object in moduleNode_L', function () {
            // given
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createCpuUsage(moduleNode);
            // then
            assert.lengthOf(measurements, 0);
        });

        it('should not calculate cpu usage - no SOC_L object in CCU_L', function () {
            // given
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createCpuUsage(moduleNode);
            // then
            assert.lengthOf(measurements, 0);
        });

        it('should not calculate cpu usage - no CPU_L object in SOC_L', function () {
            // given
            addParametersToModule('/CCU_L-1/SOC_L-1', {});
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createCpuUsage(moduleNode);
            // then
            assert.lengthOf(measurements, 0);
        });

        it('should not calculate cpu usage - param not measured', function () {
            // given
            var CPU_NOT_MEASURED = -1;
            addParametersToModule('/CCU_L-1/SOC_L-1/CPU_L-1', {cpuLoad: CPU_NOT_MEASURED});
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createCpuUsage(moduleNode);
            // then
            assert.lengthOf(measurements, 0);
        });
    });

    describe('ram', function () {

        it('should not calculate ram usage when CPU is missing', function () {
            // given
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createRamUsage(moduleNode);
            // then
            assert.lengthOf(measurements, 0);
        });

        it('should not calculate ram usage when MEMORY is missing', function () {
            // given
            addParametersToModule('/CPU_L-1', {});
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createRamUsage(moduleNode);
            // then
            assert.lengthOf(measurements, 0);
        });

        it('should calculate ram usage', function () {
            // given
            addParametersToModule('/CPU_L-1/MEMORY_L-1', {
                usedCapacity: 1,
                totalCapacity: 10
            });
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createRamUsage(moduleNode);
            // then
            assert.deepEqual(measurements, [{name: 'RAM', value: 10}]);
        });

    });

    describe('disk usage', function () {
        const FILE_SYSTEM_L_1_PARAMS = {
            usedCapacity: 2,
            totalCapacity: 10,
            mountPoint: '/ffs/fs2'
        };
        const FILE_SYSTEM_L_2_PARAMS = {
            usedCapacity: 1,
            totalCapacity: 10,
            mountPoint: '/ffs/fs1'
        };

        it('should not calculate disk usage - no CCU_L object under module', function () {
            // given
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createDiskUsage(moduleNode);
            // then
            assert.lengthOf(measurements, 0);
        });

        it('should not calculate disk usage - no STORAGE_L object under /CCU_L-1/', function () {
            // given
            addParametersToModule('/CCU_L-1', {});
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createDiskUsage(moduleNode);
            // then
            assert.lengthOf(measurements, 0);
        });

        it('should not calculate disk usage - no FILE_SYSTEM_L object under /CCU_L-1/STORAGE_L-1/', function () {
            // given
            addParametersToModule('/CCU_L-1/STORAGE_L-1', {});
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createDiskUsage(moduleNode);
            // then
            assert.lengthOf(measurements, 0);
        });

        it('should get multiple disk usages', function () {
            //given
            addParametersToModule('/CCU_L-1/STORAGE_L-1/FILE_SYSTEM_L-1', FILE_SYSTEM_L_1_PARAMS);
            addParametersToModule('/CCU_L-1/STORAGE_L-1/FILE_SYSTEM_L-2', FILE_SYSTEM_L_2_PARAMS);
            const moduleNode = getModuleNode();
            // when
            const measurements = siteMeasurementsFactory.createDiskUsage(moduleNode);
            // then
            assert.deepEqual(measurements, [{name: '/ffs/fs1', value: 10}, {name: '/ffs/fs2', value: 20}]);
        });

    });

    function getModuleNode() {
        return imTree.findByDistName(DIST_NAME_MODULE_ROOT);
    }

    function addParametersToModule(distNameSuffix, parameters) {
        if (arguments.length === 1) distNameSuffix = '';

        imTree.mergeNode(DIST_NAME_MODULE_ROOT + distNameSuffix, parameters);
    }
});
