const siteModelFactory = require('../../src/site/siteModelFactory'),
    ImTree = require('../../src/imtree'),
    imTreeNodeFactory = require('../../src/site/imTreeNodeFactory');

describe('Factory: siteModelFactory', function () {
    'use strict';

    var factory;

    var imTree, eqmL, hwtopL, lnbts;

    const HWTOP_L_DISTNAME = '/MRBTS-1/RAT-1/HWTOP_L-1';

    beforeEach(function () {
        factory = siteModelFactory;

        imTree = new ImTree();
        eqmL = imTree.mergeNode('/MRBTS-1/RAT-1/EQM_L-1', {});
        hwtopL = imTree.mergeNode(HWTOP_L_DISTNAME, {});
        imTree.mergeNode('/MRBTS-1/RAT-1/BBTOP_L-1', {});
        lnbts = imTree.mergeNode('/MRBTS-1/RAT-1/LNBTS-1', {});
    });

    it('should create empty hardware', function () {
        const hardware = factory.create(imTree);

        assert.lengthOf(hardware.smods, 0);
        assert.lengthOf(hardware.extSmods, 0);
        assert.lengthOf(hardware.rmods, 0);
        assert.lengthOf(hardware.mhas, 0);
        assert.lengthOf(hardware.rets, 0);
        assert.lengthOf(hardware.raes, 0);
        assert.lengthOf(hardware.phyants, 0);
        assert.lengthOf(hardware.cells, 0);
        assert.lengthOf(hardware.links, 0);
        assert.lengthOf(hardware.logLinks, 0);
        assert.lengthOf(hardware.alarms, 0);
    });

    it('should create site model', function () {
        const smodNode = createModule('SMOD_L', 1);
        const source = createBackEndConnector(smodNode, 1);
        const target = createFrontEndConnector(smodNode, 2);
        createLink(1, source, target);

        createModule('SMOD_EXT_L', 1);
        createModule('RMOD_L', 1);
        createModule('MHA_L', 1);
        createModule('RET_L', 1);
        createModule('RAE_L', 1);
        createModule('PHYANT_L', 1);

        const cellNo = 1;
        createCell(cellNo);
        createChannel(cellNo, 1, {direction: 'RX', AntennaDN: '/MRBTS-1/RAT-1/EQM_L-1/PHYANT_L-1'});

        const hardware = factory.create(imTree);

        assert.lengthOf(hardware.smods, 1);
        assert.lengthOf(hardware.extSmods, 1);
        assert.lengthOf(hardware.rmods, 1);
        assert.lengthOf(hardware.mhas, 1);
        assert.lengthOf(hardware.rets, 1);
        assert.lengthOf(hardware.raes, 1);
        assert.lengthOf(hardware.phyants, 1);
        assert.lengthOf(hardware.cells, 1);
        assert.lengthOf(hardware.links, 1);
        assert.lengthOf(hardware.logLinks, 1);
        assert.lengthOf(hardware.alarms, 0);
    });

    describe('modules', function () {
        it('should create smod', function () {
            createModule('SMOD_L', 1, {
                serialNumber: '111',
                productName: 'pn',
                productCode: '222'
            });

            const hardware = factory.create(imTree);

            assert.lengthOf(hardware.smods, 1);
            const smod = hardware.smods[0];
            assert.deepEqual(smod.distName, '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1');
            assert.deepEqual(smod.number, 1);
            assert.deepEqual(smod.name, 'SMOD_L-1');
            assert.deepEqual(smod.sn, '111');
            assert.deepEqual(smod.pc, '222');
            assert.deepEqual(smod.pn, 'pn');
            assert.deepEqual(smod.leftConnectors, []);
            assert.deepEqual(smod.rightConnectors, []);
        });

        it('should create smod ext', function () {
            createModule('SMOD_EXT_L', 1, {ratTechnology: 'WCDMA', serialNumber: '111', productCode: '222'});

            const hardware = factory.create(imTree);

            assert.lengthOf(hardware.extSmods, 1);
            const extSmod = hardware.extSmods[0];
            assert.equal(extSmod.distName, '/MRBTS-1/RAT-1/EQM_L-1/SMOD_EXT_L-1');
            assert.equal(extSmod.number, 1);
            assert.equal(extSmod.name, 'SMOD_EXT_L-1 WCDMA');
            assert.equal(extSmod.sn, '111');
            assert.equal(extSmod.pc, '222');
            assert.deepEqual(extSmod.leftConnectors, []);
            assert.deepEqual(extSmod.rightConnectors, []);
        });

        it('should create module with children', function () {
            const module = createModule('SMOD_L', 1, {serialNumber: '111', productCode: '222'});
            createModule('TRU_L', 1, {}, module);

            const smod = factory.create(imTree).smods[0];

            assert.deepEqual(smod.distName, '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1');

            const child = smod.children[0];
            assert.deepEqual(child.distName, '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/TRU_L-1');
        });

        it('should create rae with raeu', function () {
            const raeNode = createModule('RAE_L', 1);
            createModule('RAEU_L', 1, {}, raeNode);

            const hardware = factory.create(imTree);

            assert.lengthOf(hardware.raes, 1);
            assert.lengthOf(hardware.raes[0].children, 1);
        });
    });

    describe('commisioning data', function () {
        it('should add commissioning data to specific modules', function () {
            createModule('BBMOD_L', 1);
            const smodNode = createModule('SMOD_L', 1);
            createModule('TRU_L', 1, {}, smodNode);
            const rmodNode = createModule('RMOD_L', 1);
            createModule('RU_L', 1, {}, rmodNode);
            const mhaNode = createModule('MHA_L', 1);
            createModule('HWTOP_L', 1, {}, mhaNode);
            createModule('RET_L', 1);

            const model = factory.create(imTree);

            assert.isDefined(model.bbmods[0].commissioningData);
            assert.isDefined(model.smods[0].children[0].commissioningData);
            assert.isDefined(model.rmods[0].commissioningData);
            assert.isDefined(model.mhas[0].commissioningData);
            assert.isDefined(model.rets[0].commissioningData);
        });

        it('should add commissioning data to RMOD_L even when there is no RU_L', function () {
            createModule('RMOD_L', 1);

            const model = factory.create(imTree);

            assert.isDefined(model.rmods[0].commissioningData);
        });
    });

    describe('hardware sorting', function () {
        it('should create smods sorted by number', function () {
            createModule('SMOD_L', 4);
            createModule('SMOD_L', 2);
            createModule('SMOD_L', 3);
            createModule('SMOD_L', 1);

            const smods = factory.create(imTree).smods;

            assert.deepEqual(smods[0].number, 1);
            assert.deepEqual(smods[1].number, 2);
            assert.deepEqual(smods[2].number, 3);
            assert.deepEqual(smods[3].number, 4);
        });

        it('should create cells sorted by number', function () {
            createCell(1);
            createCell(4);
            createCell(2);
            createCell(3);

            const cells = factory.create(imTree).cells;

            assert.deepEqual(cells[0].number, 1);
            assert.deepEqual(cells[1].number, 2);
            assert.deepEqual(cells[2].number, 3);
            assert.deepEqual(cells[3].number, 4);
        });

        it('should sort raes by position', function () {
            const rmod = createModule('RMOD_L', 1);

            const rae1 = createModule('RAE_L', 1);
            const rae2 = createModule('RAE_L', 2);


            const source1 = createFrontEndConnector(rae1, 1);
            const source2 = createFrontEndConnector(rae2, 1);
            const target1 = createBackEndConnector(rmod, 2);
            const target2 = createBackEndConnector(rmod, 1);
            createLink(1, source1, target1);
            createLink(2, source2, target2);

            const raes = factory.create(imTree).raes;

            assert.deepEqual(raes[0].number, 2);
            assert.deepEqual(raes[1].number, 1);
        });
    });

    describe('distName maping', function () {
        it('should create map with distName as key', function () {
            createModule('SMOD_L', 1, {serialNumber: '111', productCode: '222'});

            const hardware = factory.create(imTree);

            assert(hardware.mapByDistName['/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1']);
        });

        it('should create distName links mapping', function () {
            const module = createModule('SMOD_L', 1);
            const source = createBackEndConnector(module, 1);
            const target = createFrontEndConnector(module, 2);
            const linkNode = createLink(1, source, target);

            const mapping = factory.create(imTree).mapByDistName;

            assert.ok(mapping[linkNode.distName]);
        });
    });

    function createFrontEndConnector(parentNode, connectorNumber) {
        return imTreeNodeFactory.createFrontEndConnector(imTree, parentNode, connectorNumber);
    }

    function createBackEndConnector(parentNode, connectorNumber) {
        return imTreeNodeFactory.createBackEndConnector(imTree, parentNode, connectorNumber);
    }

    function createModule(className, id, parameters, parentNode) {
        return imTreeNodeFactory.createModule(imTree, eqmL, className, id, parameters, parentNode);
    }

    function createCell(cellNumber, parameters) {
        return imTreeNodeFactory.createCell(imTree, lnbts, cellNumber, parameters);
    }

    function createChannel(parentCellNumber, channelNumber, parameters) {
        return imTreeNodeFactory.createChannel(imTree, lnbts, parentCellNumber, channelNumber, parameters);
    }

    function createLink(linkNumber, source, target) {
        return imTreeNodeFactory.createLink(imTree, hwtopL, linkNumber, source, target);
    }
});
