const timingServiceForOnlineImTree = require('../../src/timings/timingServiceForOnlineImTree');
const ImTree = require('../../src/imtree');

function createOnlineImTree(im) {
    return {
        getImTree() {
            return Promise.resolve(im);
        }
    };
}

describe('timingServiceForOnlineImTree', () => {
    let imTree;
    beforeEach(() => {
        imTree = new ImTree();
    });

    describe('findAllTimings', () => {

        it('should return empty array from empty tree ', () => {
            timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const timings = timingApi.findAllTimings();
                    assert.equal(timings.length, 0);
                });

        });

        it('should return all timings objects', () => {
            imTree.mergeNode('/MRBTS-1/TIMINGS_L-1', {});
            imTree.mergeNode('/MRBTS-1/XYZ/TIMINGS_L-2', {});

            return timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const timings = timingApi.findAllTimings();
                    assert.equal(timings.length, 2);
                });
        });
    });

    describe('findTimingsForProcedureName', () => {

        it('should return empty array from empty tree', () => {
            timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const timings = timingApi.findTimingsForProcedureName('any');
                    assert.equal(timings.length, 0);
                });
        });

        it('should return empty array from tree without timings objects having procedure name ', () => {
            imTree.mergeNode('/MRBTS-1/TIMINGS_L-1', {
                procedureName: 'ConfigureDevice1'
            });
            imTree.mergeNode('/MRBTS-1/XYZ/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice2'
            });

            return timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const timings = timingApi.findTimingsForProcedureName('any');
                    assert.equal(timings.length, 0);
                });
        });

        it('should return timing object with procedure name', () => {
            imTree.mergeNode('/MRBTS-1/TIMINGS_L-1', {
                procedureName: 'ConfigureDevice1'
            });
            imTree.mergeNode('/MRBTS-1/XYZ/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice2'
            });
            return timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const timings = timingApi.findTimingsForProcedureName('ConfigureDevice1');
                    assert.equal(timings.length, 1);
                });
        });

        it('should return all timing objects if procedure name is not present', () => {
            imTree.mergeNode('/MRBTS-1/TIMINGS_L-1', {
                procedureName: 'ConfigureDevice1'
            });
            imTree.mergeNode('/MRBTS-1/XYZ-1/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice2'
            });
            imTree.mergeNode('/MRBTS-1/XYZAA-1/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice3'
            });
            imTree.mergeNode('/MRBTS-1/XYZBB-1/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice4'
            });
            return timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const timings = timingApi.findTimingsForProcedureName('');
                    assert.equal(timings.length, 4);
                    const allTimings = timingApi.findAllTimings();
                    assert.equal(timings.length, allTimings.length);
                });
        });
    });

    describe('findTimingsForDistName', () => {

        it('should return empty array from empty tree', () => {
            timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const timings = timingApi.findTimingsForDistName('any');
                    assert.equal(timings.length, 0);
                });

        });

        it('should return empty array for tree without searched dn', () => {
            imTree.mergeNode('/MRBTS-1/TIMINGS_L-1', {
                procedureName: 'ConfigureDevice1'
            });
            imTree.mergeNode('/MRBTS-1/XYZ/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice2'
            });
            return timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const timings = timingApi.findTimingsForDistName('somedn');
                    assert.equal(timings.length, 0);
                });
        });

        it('should return array with timings objects', () => {
            imTree.mergeNode('/MRBTS-1/TIMINGS_L-1', {
                procedureName: 'ConfigureDevice1'
            });
            imTree.mergeNode('/MRBTS-1/XYZ/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice2'
            });
            imTree.mergeNode('/ABC-1/XYZ/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice2'
            });
            return timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const timings = timingApi.findTimingsForDistName('ABC');
                    assert.equal(timings.length, 1);
                });
        });

        it('should return same number of timings as "findAllTimings" if dist name is not presented', () => {
            imTree.mergeNode('/MRBTS-1/TIMINGS_L-1', {
                procedureName: 'ConfigureDevice1'
            });
            imTree.mergeNode('/MRBTS-1/XYZ/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice2'
            });
            imTree.mergeNode('/ABC-1/XYZ/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice2'
            });
            return timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const timings = timingApi.findTimingsForDistName('');
                    const allTimings = timingApi.findAllTimings();

                    assert.equal(timings.length, allTimings.length);
                });
        });
    });

    describe('getAllProcedureNames', () => {

        it('should return empty array from empty tree', () => {
            timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const procedures = timingApi.getAllProcedureNames();
                    assert.equal(procedures.length, 0);
                });

        });

        it('should return empty array for tree without TIMINGS_L objects', () => {
            imTree.mergeNode('/MRBTS-1/ABC-1', {});
            imTree.mergeNode('/MRBTS-1/XYZ/QWER-2', {});
            return timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const procedures = timingApi.getAllProcedureNames();
                    assert.equal(procedures.length, 0);
                });
        });

        it('should return array denoting set of all procedures names for tree with TIMINGS_L objects', () => {
            imTree.mergeNode('/MRBTS-1/TIMINGS_L-1', {
                procedureName: 'ConfigureDevice2'
            });
            imTree.mergeNode('/MRBTS-1/XYZ/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice2'
            });
            imTree.mergeNode('/MRBTS-1/XYZabc/TIMINGS_L-2', {
                procedureName: 'ConfigureDevice2'
            });
            imTree.mergeNode('/MRBTS-1/XYZqqq/TIMINGS_L-2', {
                procedureName: 'sync'
            });
            return timingServiceForOnlineImTree.newInstance(createOnlineImTree(imTree))
                .then((timingApi) => {
                    const procedures = timingApi.getAllProcedureNames();
                    assert.equal(procedures.length, 2);
                });
        });
    });
});
