describe('typeInfo v1', function () {
    var typeInfo;
    var meta;

    beforeEach(function () {
        meta = {p: []};
    });

    describe('legacy parameters type info', function () {

        it('should return [] when parameters not exist', function () {
            typeInfo = create({"p": null});
            assert.deepEqual(typeInfo.getParameters(), []);
        });

        it('should return name', function () {
            meta.p.push({"@name": "productCode"});
            assert.strictEqual(firstParameter().getName(), "productCode");
        });

        it('should return false when is not array - maxOccurs not exist', function () {
            meta.p.push({});
            assert.isFalse(firstParameter().isArray());
        });

        it('should return false when is not array - maxOccurs = 0', function () {
            meta.p.push({"@maxOccurs": "0"});
            assert.isFalse(firstParameter().isArray());
        });

        it('should return false when is not array - maxOccurs = 1', function () {
            meta.p.push({"@maxOccurs": "1"});
            assert.isFalse(firstParameter().isArray());
        });

        it('should return true when is array', function () {
            meta.p.push({"@maxOccurs": "2"});
            assert.isTrue(firstParameter().isArray());
        });

        it('should return true when is simple type', function () {
            meta.p.push({"simpleType": {}});
            assert.isTrue(firstParameter().isSimple());
        });

        it('should return false when is not simple type', function () {
            meta.p.push({});
            assert.isFalse(firstParameter().isSimple());
        });

        it('should return true when is complex type', function () {
            meta.p.push({"complexType": {}});
            assert.isTrue(firstParameter().isComplex());
        });

        it('should return true when is complex type - p inside of p', function () {
            meta.p.push({
                "p": {
                    "p": {}
                }
            });
            assert.isTrue(firstParameter().isComplex());
        });

        it('should return false when is not complex type', function () {
            meta.p.push({});
            assert.isFalse(firstParameter().isComplex());
        });

        it('should return true when is mandatory', function () {
            meta.p = [{}];
            assert.isTrue(firstParameter().isMandatory());

            meta.p = [{"creation": {"@priority": "mandatory"}}];
            assert.isTrue(firstParameter().isMandatory());

            meta.p = [{"creation": {"@priority": "other value"}}];
            assert.isFalse(firstParameter().isMandatory());
        });

        it('should return true when is set by system', function () {
            meta.p.push({"creation": {"@priority": "Value set by the system"}});
            assert.isTrue(firstParameter().isSetBySystem());
        });

        it('should return true when is integer', function () {
            meta.p.push({"simpleType": {"@base": "integer"}});

            assert.isTrue(firstParameter().isInteger());
            assert.isFalse(firstParameter().isEnum());
        });

        it('should return true when is enum', function () {
            meta.p.push({
                "simpleType": {
                    "@base": "integer",
                    "enumeration": [
                        {
                            "@value": "0",
                            "@text": "x"
                        },
                        {
                            "@value": "1",
                            "@text": "y"
                        }
                    ],
                    "default": {"@value": "0"}
                }
            });

            assert.isFalse(firstParameter().isInteger());
            assert.isTrue(firstParameter().isEnum());
        });

        it('should return true when is integer - decimal without divisor', function () {
            meta.p.push({
                "simpleType": {"@base": "decimal"}
            });

            assert.isTrue(firstParameter().isInteger());
        });

        it('should return true when is string', function () {
            meta.p.push({
                "simpleType": {"@base": "string"}
            });

            assert.isTrue(firstParameter().isString());
        });

        it('should return false when is ObjectId', function () {
            meta.p.push({
                "simpleType": {"@base": "ObjectId"}
            });

            assert.isFalse(firstParameter().isObjectId());
        });

        it('should return true when is boolean', function () {
            meta.p.push({
                "simpleType": {"@base": "boolean"}
            });

            assert.isTrue(firstParameter().isBoolean());
        });

        it('should return true when decimal is float', function () {
            meta.p.push({
                "simpleType": {
                    "@base": "decimal",
                    "editing": {
                        "@divisor": "10"
                    }
                }
            });

            assert.isTrue(firstParameter().isFloat());
        });

        it('should return true when integer is actually float', function () {
            meta.p.push({
                "simpleType": {
                    "@base": "integer",
                    "editing": {
                        "@divisor": "10"
                    }
                }
            });

            assert.isTrue(firstParameter().isFloat());
            assert.isFalse(firstParameter().isInteger());
        });

        it('should return false when integer is not float - has no divisor', function () {
            meta.p.push({
                "simpleType": {
                    "@base": "integer"
                }
            });

            assert.isFalse(firstParameter().isFloat());
            assert.isTrue(firstParameter().isInteger());
        });

    });

    describe('get parameters by name', function () {

        it('should get by names', function () {
            meta.p.push({"@name": "parameterA"});
            meta.p.push({"@name": "parameterB"});
            meta.p.push({"@name": "parameterC"});

            assert.ok(parameterByName("parameterA"));
            assert.ok(parameterByName("parameterB"));
            assert.ok(parameterByName("parameterC"));
        });

        it('should return undefined when parameter name not exist', function () {
            meta.p.push({"@name": "parameterA"});

            assert.notOk(parameterByName("parameterB"));
        });

    });

    describe('get parameters by id', function () {

        it('should get by names', function () {
            meta.p.push({"@name": "parameterA"});
            meta.p.push({"@name": "parameterB"});
            meta.p.push({"@name": "parameterC"});

            assert.strictEqual(parameterById(1).getName(), "parameterA");
            assert.strictEqual(parameterById(2).getName(), "parameterB");
            assert.strictEqual(parameterById(3).getName(), "parameterC");
        });

        it('should return undefined when parameter id not exist', function () {
            assert.notOk(parameterById(-1));
        });

    });

    function create(metaDefinition) {
        return infoModel.typeInfo(metaDefinition);
    }

    function firstParameter() {
        return create(meta).getParameters()[0];
    }

    function parameterById(id) {
        return create(meta).getParameterById(id);
    }

    function parameterByName(name) {
        return create(meta).getParameterByName(name);
    }

});
