(function (global) {
    'use strict';

    describe('typeInfo v2', function () {
        var typeInfo;
        var meta;

        beforeEach(function () {
            meta = {p: [], struct: [], enumeration: []};
        });

        describe('parameters type info', function () {
            it('should return [] when parameters not exist', function () {
                typeInfo = create({'p': null});
                assert.deepEqual(typeInfo.getParameters(), []);
            });

            it('should return name', function () {
                meta.p.push({'@id': 1, '@name': "productCode"});
                assert.strictEqual(firstParameter().getName(), "productCode");
            });

            it('should return false when is not array - recurrence not exist', function () {
                meta.p.push({'@id': 1});
                assert.isFalse(firstParameter().isArray());
            });

            it('should return false when is not array - recurrence = required', function () {
                meta.p.push({'@id': 1, '@recurrence': 'required'});
                assert.isFalse(firstParameter().isArray());
            });

            it('should return true when is array', function () {
                meta.p.push({'@id': 1, '@recurrence': 'repeated'});
                assert.isTrue(firstParameter().isArray());
            });

            it('should return true when is simple type', function () {
                meta.p.push({'@id': 1, '@type': 'integer'});
                assert.isTrue(firstParameter().isSimple());
            });

            it('should return false when is not simple', function () {
                meta.struct.push({'@name': 'myType'});
                meta.p.push({'@id': 1, '@type': 'myType'});
                assert.isFalse(firstParameter().isSimple());
            });

            it('should return true when is complex type', function () {
                meta.struct.push({'@name': 'myType'});
                meta.p.push({'@id': 1, '@type': 'myType'});
                assert.isTrue(firstParameter().isComplex());
            });

            it('should return false when is not complex type', function () {
                meta.p.push({'@id': 1, '@type': 'string'});
                assert.isFalse(firstParameter().isComplex());
            });

            it('should return true when is mandatory', function () {
                meta.p = [{'@id': 1, '@recurrence': 'required'}];
                assert.isTrue(firstParameter().isMandatory());

                meta.p = [{'@id': 1, '@recurrence': 'mandatory'}];
                assert.isTrue(firstParameter().isMandatory());

                meta.p = [{'@id': 1, '@recurrence': 'optional'}];
                assert.isFalse(firstParameter().isMandatory());
            });

            it('should return true when is set by system', function () {
                meta.p.push({'@id': 1, "@recurrence": "Value set by the system"});
                assert.isTrue(firstParameter().isSetBySystem());
            });

            it('should return true when is integer', function () {
                meta.p.push({'@id': 1, '@type': 'integer'});

                assert.isFalse(firstParameter().isString());
                assert.isFalse(firstParameter().isObjectId());
                assert.isTrue(firstParameter().isInteger());
                assert.isFalse(firstParameter().isEnum());
                assert.isFalse(firstParameter().isFloat());
                assert.isFalse(firstParameter().isBoolean());
                assert.isFalse(firstParameter().isComplex());
                assert.isFalse(firstParameter().isArray());
            });

            it('should return true when is enum', function () {
                meta.enumeration.push({
                    '@name': 'EmyEnum',
                    'enum': [
                        {'@value': '0', '@name': 'X'},
                        {'@value': '1', '@name': 'Y'}
                    ]
                });

                meta.p.push({
                    '@id': 1,
                    '@type': 'EMyEnum'
                });

                assert.isFalse(firstParameter().isString());
                assert.isFalse(firstParameter().isObjectId());
                assert.isFalse(firstParameter().isInteger());
                assert.isTrue(firstParameter().isEnum());
                assert.isFalse(firstParameter().isFloat());
                assert.isFalse(firstParameter().isBoolean());
                assert.isFalse(firstParameter().isComplex());
                assert.isFalse(firstParameter().isArray());
            });


            it('should return true when is string', function () {
                meta.p.push({'@id': 1, '@type': 'string'});
                assert.isTrue(firstParameter().isString());
                assert.isFalse(firstParameter().isObjectId());
                assert.isFalse(firstParameter().isInteger());
                assert.isFalse(firstParameter().isEnum());
                assert.isFalse(firstParameter().isFloat());
                assert.isFalse(firstParameter().isBoolean());
                assert.isFalse(firstParameter().isComplex());
                assert.isFalse(firstParameter().isArray());
            });

            it('should return true when is objectId', function () {
                meta.p.push({'@id': 1, '@type': 'objectid'});
                assert.isFalse(firstParameter().isString());
                assert.isTrue(firstParameter().isObjectId());
                assert.isFalse(firstParameter().isInteger());
                assert.isFalse(firstParameter().isEnum());
                assert.isFalse(firstParameter().isFloat());
                assert.isFalse(firstParameter().isBoolean());
                assert.isFalse(firstParameter().isComplex());
                assert.isFalse(firstParameter().isArray());
            });

            it('should return true when is boolean', function () {
                meta.p.push({'@id': 1, '@type': 'boolean'});
                assert.isFalse(firstParameter().isString());
                assert.isFalse(firstParameter().isObjectId());
                assert.isFalse(firstParameter().isInteger());
                assert.isFalse(firstParameter().isEnum());
                assert.isFalse(firstParameter().isFloat());
                assert.isTrue(firstParameter().isBoolean());
                assert.isFalse(firstParameter().isComplex());
                assert.isFalse(firstParameter().isArray());
            });

            it('should return true when is float', function () {
                meta.p.push({'@id': 1, '@type': 'double'});
                assert.isFalse(firstParameter().isString());
                assert.isFalse(firstParameter().isObjectId());
                assert.isFalse(firstParameter().isInteger());
                assert.isFalse(firstParameter().isEnum());
                assert.isTrue(firstParameter().isFloat());
                assert.isFalse(firstParameter().isBoolean());
                assert.isFalse(firstParameter().isComplex());
                assert.isFalse(firstParameter().isArray());
            });

        });

        describe('get parameters by name', function () {
            it('should get by names', function () {
                meta.p.push({'@id': 1, '@name': 'parameterA'});
                meta.p.push({'@id': 2, '@name': 'parameterB'});
                meta.p.push({'@id': 3, '@name': 'parameterC'});

                assert.ok(parameterByName("parameterA"));
                assert.ok(parameterByName("parameterB"));
                assert.ok(parameterByName("parameterC"));
            });

            it('should return undefined when parameter name not exist', function () {
                meta.p.push({'@id': 1, '@name': 'parameterA'});
                assert.notOk(parameterByName("parameterB"));
            });

        });

        describe('get parameters by id', function () {
            it('should get by names', function () {
                meta.p.push({'@id': 1, '@name': "parameterA"});
                meta.p.push({'@id': 2, '@name': "parameterB"});
                meta.p.push({'@id': 3, '@name': "parameterC"});

                assert.strictEqual(parameterById(1).getName(), "parameterA");
                assert.strictEqual(parameterById(2).getName(), "parameterB");
                assert.strictEqual(parameterById(3).getName(), "parameterC");
            });

            it('should return undefined when parameter id not exist', function () {
                assert.notOk(parameterById(-1));
            });

        });

        function create(metaDefinition) {
            return infoModel.typeInfoV2(metaDefinition);
        }

        function firstParameter() {
            return create(meta).getParameters()[0];
        }

        function parameterById(id) {
            return create(meta).getParameterById(id);
        }

        function parameterByName(name) {
            return create(meta).getParameterByName(name);
        }
    });
})(this);
