describe('Service: updateService#historyListener', function () {

    var updateService;
    var historyListener;

    beforeEach(function () {
        var metaService = {
            getLastMetaUid: function () {
                return 0;
            }
        };
        updateService = infoModel.updateService({metaService: metaService});
        historyListener = createHistoryListener();
    });

    it('should add and remove history listener', function (done) {
        updateService.addHistoryListener(historyListener)
            .then(function () {
                pushUpdate();
                assert.lengthOf(historyListener.entries, 1);

                return updateService.removeHistoryListener(historyListener);
            })
            .then(function () {
                pushUpdate();
                assert.lengthOf(historyListener.entries, 1);
            })
            .then(done).catch(done);
    });

    it('should notify history listener about new entry', function (done) {
        updateService.addHistoryListener(historyListener)
            .then(function () {
                assert.lengthOf(historyListener.entries, 0);

                pushUpdate();

                assert.lengthOf(historyListener.entries, 1);
                var entry = historyListener.entries[0];
                assert.equal(entry.id, 0);
                assert.equal(entry.numOfChanges, 1);
                assert.isTrue(entry.timestamp > 0);
            })
            .then(done).catch(done);
    });

    it('should clear all history', function (done) {
        pushUpdate();

        updateService.addHistoryListener(historyListener)
            .then(function () {
                return updateService.clearHistory();
            })
            .then(function () {
                assert.lengthOf(historyListener.entries, 0);
            })
            .then(done).catch(done);
    });

    it('multiple resets should be considered as one', function (done) {
        updateService.addHistoryListener(historyListener)
            .then(function () {
                pushUpdate();

                pushReset();
                pushReset();
                pushReset();

                var entries = historyListener.entries;
                assert.lengthOf(entries, 2);
                assert.ok(entries[0].change);
                assert.ok(entries[1].reset);
            })
            .then(done).catch(done);
    });

    it('bts disconnect should be treated as reset', function (done) {
        updateService.addHistoryListener(historyListener)
            .then(function () {
                pushUpdate();

                pushDisconnect('BTS');

                var entries = historyListener.entries;
                assert.lengthOf(entries, 2);
                assert.ok(entries[0].change);
                assert.ok(entries[1].reset);
            })
            .then(done).catch(done);
    });

    it('user disconnect should not be treated as reset', function (done) {
        updateService.addHistoryListener(historyListener)
            .then(function () {
                pushUpdate();

                pushDisconnect('USER');

                var entries = historyListener.entries;
                assert.lengthOf(entries, 1);
                assert.ok(entries[0].change);
            })
            .then(done).catch(done);
    });

    it('should clear all history', function (done) {
        pushUpdate();

        updateService.addHistoryListener(historyListener)
            .then(function () {
                return updateService.clearHistory();
            })
            .then(function () {
                assert.lengthOf(historyListener.entries, 0);
            })
            .then(done).catch(done);
    });

    it('clear history should clear also history marker', function (done) {
        pushUpdate();

        updateService.addHistoryListener(historyListener)
            .then(function () {
                return updateService.setHistoryMarker();
            })
            .then(function () {
                assert.ok(historyListener.historyMarker);

                return updateService.clearHistory();
            })
            .then(function () {
                assert.notOk(historyListener.historyMarker);
            })
            .then(done).catch(done);
    });

    it('should notify about history marker when adding listener', function (done) {
        pushUpdate();

        updateService.setHistoryMarker()
            .then(function () {
                return updateService.addHistoryListener(historyListener);
            })
            .then(function () {
                assert.ok(historyListener.historyMarker);
            })
            .then(done).catch(done);
    });


    it('should notify history listener about past entry', function (done) {
        pushUpdate();

        updateService.addHistoryListener(historyListener)
            .then(function () {
                assert.lengthOf(historyListener.entries, 1);
            })
            .then(done).catch(done);
    });

    it('should set history marker on last history entry', function (done) {
        pushUpdate();
        pushUpdate();

        updateService.addHistoryListener(historyListener)
            .then(function () {
                return updateService.setHistoryMarker();
            })
            .then(function (entry) {
                assert.deepEqual(_.last(historyListener.entries), entry);
            })
            .then(done).catch(done);
    });

    describe('current entry listener', function () {

        it('should notify current entry listener when history marker is set and cleared', function (done) {
            // given
            const entries = [];

            const entry1 = pushUpdate();
            const entry2 = pushUpdate();

            const listener = {
                changeEntry: function (entry) {
                    entries.push(entry.id);
                }
            };
            // when
            updateService.addCurrentEntryListener(listener)
                .then(function () {
                    return updateService.setHistoryMarker(entry1);
                })
                .then(function () {
                    return updateService.clearHistoryMarker();
                })
                .then(function () {
                    // then
                    assert.deepEqual(entries, [entry2.id, entry1.id, entry2.id]);
                })
                .then(done).catch(done);
        });
    });

    describe('merge history', function () {

        beforeEach(function (done) {
            _.times(10, pushUpdate);

            updateService.addHistoryListener(historyListener)
                .then(done).catch(done);
        });

        it('should merge history entries into one entry', function (done) {
            assert.isTrue(historyListener.entries.length > 1);

            updateService.mergeHistory()
                .then(function () {
                    assert.lengthOf(historyListener.entries, 1);
                })
                .then(done).catch(done);
        });

        it('should notify about merge', function (done) {
            historyListener.clear();

            updateService.mergeHistory()
                .then(function () {
                    assert.deepEqual(historyListener.events, ['clearEntries', 'newEntry']);
                })
                .then(done).catch(done);
        });

        it('when history marker was set then should notify about new history marker', function (done) {
            historyListener.clear();

            updateService.setHistoryMarker()
                .then(function () {
                    historyListener.clear();
                    return updateService.mergeHistory();
                })
                .then(function () {
                    assert.equal(historyListener.entries[0], historyListener.historyMarker);
                    assert.deepEqual(historyListener.events, [
                        'clearEntries',
                        'historyMarkerWasCleared',
                        'newEntry',
                        'historyMarkerWasSet'
                    ]);
                })
                .then(done).catch(done);
        });

    });

    describe('merge first entries', function () {
        const NUM_OF_ENTRIES = 5;

        beforeEach(function (done) {
            _.times(NUM_OF_ENTRIES, function (index) {
                updateService.pushUpdate('/A-' + index, new Uint8Array(1000));
            });

            updateService.addHistoryListener(historyListener)
                .then(function () {
                    historyListener.events = [];
                })
                .then(done).catch(done);
        });

        it('should merge first n entries when history marker is set', function (done) {
            updateService.setHistoryMarker()
                .then(function () {
                    assert.deepEqual(historyListener.events, ['historyMarkerWasSet']);
                    assert.lengthOf(historyListener.entries, NUM_OF_ENTRIES);
                    historyListener.events = [];
                })
                .then(function () {
                    return updateService.mergeFirstEntries(NUM_OF_ENTRIES);
                })
                .then(function () {
                    assert.deepEqual(historyListener.events, [
                        'clearEntries',
                        'historyMarkerWasCleared',
                        'newEntries',
                        'historyMarkerWasSet'
                    ]);
                    assert.lengthOf(historyListener.entries, 1);
                })
                .then(done).catch(done);
        });

        it('should merge first n entries when history marker is not set', function (done) {
            updateService.mergeFirstEntries(NUM_OF_ENTRIES)
                .then(function () {
                    assert.deepEqual(historyListener.events, [
                        'clearEntries',
                        'historyMarkerWasCleared',
                        'newEntries'
                    ]);
                    assert.lengthOf(historyListener.entries, 1);
                })
                .then(done).catch(done);
        });

        it('should merge first 3 entries when history marker is not set', function (done) {
            updateService.mergeFirstEntries(3)
                .then(function () {
                    assert.deepEqual(historyListener.events, [
                        'clearEntries',
                        'historyMarkerWasCleared',
                        'newEntries'
                    ]);
                    assert.lengthOf(historyListener.entries, 3);
                })
                .then(done).catch(done);
        });

        describe('setting history marker after merge', function () {
            var historyMarker;

            beforeEach(function (done) {
                var listener = {
                    historyMarkerWasSet: function (newHistoryMarker) {
                        historyMarker = newHistoryMarker;
                    }
                };
                updateService.addHistoryListener(listener)
                    .then(done)
                    .catch(done);
            });

            it('should set history marker as merged entry if it was disappeared entry', function (done) {
                updateService.setHistoryMarker(updateService.getHistoryEntry(1))
                    .then(function () {
                        return updateService.mergeFirstEntries(3);
                    })
                    .then(function () {
                        assert.deepEqual(historyMarker.id, 0);
                    })
                    .then(done)
                    .catch(done);
            });

            it('should reset history marker if entry still exists', function (done) {
                updateService.setHistoryMarker(updateService.getHistoryEntry(4))
                    .then(function () {
                        return updateService.mergeFirstEntries(3);
                    })
                    .then(function () {
                        assert.deepEqual(historyMarker.id, 2);
                    })
                    .then(done)
                    .catch(done);
            });

        });

        describe('find number of first entries of size', function () {
            var entries;
            var finder;

            beforeEach(function () {
                entries = historyListener.entries;
                finder = updateService.entriesFinder();
            });

            it('should find entries for given size', function () {
                var firstEntrySize = entries[0].estimatedSizeInBytes;

                var numOfEntriesBellow = finder.findOfSize(firstEntrySize - 1);
                assert.equal(numOfEntriesBellow, 1);

                var numOfEntriesEqual = finder.findOfSize(firstEntrySize);
                assert.equal(numOfEntriesEqual, 1);

                var numOfEntriesGreater = finder.findOfSize(firstEntrySize + 1);
                assert.equal(numOfEntriesGreater, 2);
            });

            it('should return 0 after skipping all entries', function () {
                var n = finder.skipFirst(NUM_OF_ENTRIES).findOfSize(1);
                assert.equal(n, 0);
            });

            it('should return number of all entries when given size of all entries', function () {
                var totalSize = totalSizeOfEntries();

                var n = finder.findOfSize(totalSize);
                assert.equal(n, NUM_OF_ENTRIES);
            });

            it('should skip given entries', function () {
                var numOfEntriesToSkip = 2;
                var totalSize = totalSizeOfEntries();

                var n = finder.skipFirst(numOfEntriesToSkip).findOfSize(totalSize);
                assert.equal(n, NUM_OF_ENTRIES - numOfEntriesToSkip);
            });

            function totalSizeOfEntries() {
                var totalSize = 0;
                _.each(historyListener.entries, function (entry) {
                    totalSize += entry.estimatedSizeInBytes;
                });
                return totalSize;
            }

        });

    });

    describe('merge range', function () {
        const NUM_OF_ENTRIES = 5;

        beforeEach(function (done) {
            _.times(NUM_OF_ENTRIES, function (index) {
                updateService.pushUpdate('/A-' + index, new Uint8Array(1000));
            });

            updateService.addHistoryListener(historyListener)
                .then(function () {
                    historyListener.events = [];
                })
                .then(done).catch(done);
        });

        describe('setting history marker after merge', function () {
            var historyMarker;

            beforeEach(function (done) {
                var listener = {
                    historyMarkerWasSet: function (newHistoryMarker) {
                        historyMarker = newHistoryMarker;
                    }
                };
                updateService.addHistoryListener(listener)
                    .then(done)
                    .catch(done);
            });

            it('should set history marker as merged entry if it was disappeared entry', function (done) {
                updateService.setHistoryMarker(updateService.getHistoryEntry(1))
                    .then(function () {
                        return updateService.mergeEntriesRange(0, 3);
                    })
                    .then(function () {
                        assert.deepEqual(historyMarker.id, 0);
                    })
                    .then(done)
                    .catch(done);
            });

            it('should not reset history marker it is entry before range', function (done) {
                updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                    .then(function () {
                        return updateService.mergeEntriesRange(1, 3);
                    })
                    .then(function () {
                        assert.deepEqual(historyMarker.id, 0);
                    })
                    .then(done)
                    .catch(done);
            });

            it('should reset history marker it is entry after range', function (done) {
                updateService.setHistoryMarker(updateService.getHistoryEntry(4))
                    .then(function () {
                        return updateService.mergeEntriesRange(1, 3);
                    })
                    .then(function () {
                        assert.deepEqual(historyMarker.id, 3);
                    })
                    .then(done)
                    .catch(done);
            });

        });

        describe('find number of first entries of size', function () {
            var entries;
            var finder;

            beforeEach(function () {
                entries = historyListener.entries;
                finder = updateService.entriesFinder();
            });

            it('should find entries for given size', function () {
                var firstEntrySize = entries[0].estimatedSizeInBytes;

                var numOfEntriesBellow = finder.findOfSize(firstEntrySize - 1);
                assert.equal(numOfEntriesBellow, 1);

                var numOfEntriesEqual = finder.findOfSize(firstEntrySize);
                assert.equal(numOfEntriesEqual, 1);

                var numOfEntriesGreater = finder.findOfSize(firstEntrySize + 1);
                assert.equal(numOfEntriesGreater, 2);
            });

            it('should return 0 after skipping all entries', function () {
                var n = finder.skipFirst(NUM_OF_ENTRIES).findOfSize(1);
                assert.equal(n, 0);
            });

            it('should return number of all entries when given size of all entries', function () {
                var totalSize = totalSizeOfEntries();

                var n = finder.findOfSize(totalSize);
                assert.equal(n, NUM_OF_ENTRIES);
            });

            it('should skip given entries', function () {
                var numOfEntriesToSkip = 2;
                var totalSize = totalSizeOfEntries();

                var n = finder.skipFirst(numOfEntriesToSkip).findOfSize(totalSize);
                assert.equal(n, NUM_OF_ENTRIES - numOfEntriesToSkip);
            });

            function totalSizeOfEntries() {
                var totalSize = 0;
                _.each(historyListener.entries, function (entry) {
                    totalSize += entry.estimatedSizeInBytes;
                });
                return totalSize;
            }

        });

    });

    function pushUpdate() {
        return updateService.pushUpdate('ANY_DIST_NAME-1', new Uint8Array(0));
    }

    function pushReset() {
        pushDisconnect("BTS");
    }

    function pushDisconnect(source) {
        updateService.pushDisconnect({autoReconnect: false, source: source});
    }

    function createHistoryListener() {
        return {
            entries: [],
            historyMarker: null,
            events: [],
            newEntry: function (entry) {
                this.entries.push(entry);
                this.events.push('newEntry');
            },
            newEntries: function (entries) {
                this.entries = entries;
                this.events.push('newEntries');
            },
            clearEntries: function () {
                this.entries = [];
                this.events.push('clearEntries');
            },
            historyMarkerWasSet: function (historyMarker) {
                this.historyMarker = historyMarker;
                this.events.push('historyMarkerWasSet');
            },
            historyMarkerWasCleared: function () {
                this.historyMarker = null;
                this.events.push('historyMarkerWasCleared');
            },
            clear: function () {
                this.entries = [];
                this.historyMarker = null;
                this.events = [];
            }
        };
    }
});
