describe('Service: updateService#imTreeListener', function () {

    var updateService;
    var listener;

    beforeEach(function () {
        var metaService = {
            getLastMetaUid: function () {
                return 0;
            }
        };
        updateService = infoModel.updateService({metaService: metaService});
    });

    describe('completeImTree', function () {

        beforeEach(function () {
            listener = {
                imTree: null,
                completeImTree: function (imTree) {
                    this.imTree = imTree;
                },
                hasUpdateFor: function (distName) {
                    return hasDistName(this.imTree, distName);
                }
            };
        });

        it('should notify imTree listener with info model tree', function (done) {
            pushUpdate('/A-1');

            updateService.addImTreeListener(listener)
                .then(function () {
                    assert.isTrue(listener.hasUpdateFor('/A-1'));
                })
                .then(done).catch(done);
        });

        it('should notify imTree listener with current info model tree', function (done) {
            pushUpdate('/A-1');
            pushUpdate('/A-1/B-1');

            updateService.addImTreeListener(listener)
                .then(function () {
                    assert.isTrue(listener.hasUpdateFor('/A-1/B-1'));
                })
                .then(done).catch(done);
        });

        it('after setting history marker should notify imTree listener with proper historical im tree', function (done) {
            pushUpdate('/A-1');
            pushUpdate('/A-1/B-1');

            var historyEntries = null;
            updateService.getHistoryEntries()
                .then(function (entries) {
                    historyEntries = entries;
                    assert.lengthOf(entries, 2);
                })
                .then(function () {
                    return updateService.addImTreeListener(listener);
                })
                .then(function () {
                    return updateService.setHistoryMarker(historyEntries[0]);
                })
                .then(function () {
                    assert.isTrue(listener.hasUpdateFor('/A-1'));
                    assert.isFalse(listener.hasUpdateFor('/A-1/B-1'));

                    return updateService.setHistoryMarker(historyEntries[1]);
                })
                .then(function () {
                    assert.isTrue(listener.hasUpdateFor('/A-1/B-1'));
                })
                .then(done).catch(done);
        });

        it('after clearing history marker should notify imTree listener with updated im tree', function (done) {
            pushUpdate('/A-1');

            updateService.setHistoryMarker()
                .then(function () {
                    return updateService.addImTreeListener(listener);
                })
                .then(function () {
                    pushUpdate('/B-1');

                    assert.isTrue(listener.hasUpdateFor('/A-1'));
                    assert.isFalse(listener.hasUpdateFor('/B-1'));

                    return updateService.clearHistoryMarker();
                })
                .then(function () {
                    assert.isTrue(listener.hasUpdateFor('/A-1'));
                    assert.isTrue(listener.hasUpdateFor('/B-1'));
                })
                .then(done).catch(done);
        });

        it('should remove given distName from imTree', function (done) {
            updateService.addImTreeListener(listener)
                .then(function () {
                    pushUpdate('/A-1');
                    assert.isTrue(listener.hasUpdateFor('/A-1'));

                    pushRemove('/A-1');
                    assert.isFalse(listener.hasUpdateFor('/A-1'));
                })
                .then(done).catch(done);
        });

        it('after reset should discard previous updates', function (done) {
            updateService.addImTreeListener(listener)
                .then(function () {
                    pushUpdate('/A-1');
                    pushUpdate('/A-1/B-1');
                    assert.isTrue(listener.hasUpdateFor('/A-1/B-1'));

                    pushReset();
                    pushUpdate('/A-2');

                    assert.isFalse(listener.hasUpdateFor('/A-1'));
                    assert.isTrue(listener.hasUpdateFor('/A-2'));
                })
                .then(done).catch(done);
        });

    });

    describe('imTree listener', function () {
        var listener;

        beforeEach(function () {
            listener = createListener();
        });

        it('after update should notify with new and complete imTree', function (done) {
            pushUpdate('/A-1');

            updateService.addImTreeListener(listener)
                .then(function () {
                    assert.sameMembers(listener.events, ['newImTree', 'completeImTree', 'onlineImTree']);
                })
                .then(done).catch(done);
        });

        it('should add and remove imTree listener', function (done) {
            pushUpdate('/A-1');

            updateService.addImTreeListener(listener)
                .then(function () {
                    assert.sameMembers(listener.events, ['newImTree', 'completeImTree', 'onlineImTree']);

                    listener.clear();
                    return updateService.removeImTreeListener(listener);
                })
                .then(function () {
                    pushUpdate('/B-1');

                    assert.lengthOf(listener.events, 0);
                })
                .then(done).catch(done);
        });

        it('should clear all history and notify with empty imTree', function (done) {
            updateService.addImTreeListener(listener)
                .then(function () {
                    pushUpdate('/A-1');
                    listener.clear();

                    return updateService.clearHistory();
                })
                .then(function () {
                    assert.sameMembers(listener.events, ['newImTree', 'completeImTree', 'onlineImTree']);
                    assert.isTrue(listener.imTree.isEmpty());
                })
                .then(done).catch(done);
        });

        it('after reset should notify with new and complete empty imTree', function (done) {
            var completeImTree = null;
            var newImTree = null;

            listener = {
                completeImTree: function (imTree) {
                    completeImTree = imTree;
                },
                newImTree: function (imTree) {
                    newImTree = imTree;
                }
            };

            updateService.addImTreeListener(listener)
                .then(function () {
                    pushUpdate('/A-1');
                    assert.isFalse(completeImTree.isEmpty());
                    assert.isFalse(newImTree.isEmpty());

                    pushReset();
                    assert.isTrue(completeImTree.isEmpty());
                    assert.isTrue(newImTree.isEmpty());

                })
                .then(done).catch(done);
        });

        it('after first update should notify with new imTree and then only updates and complete imTree', function (done) {
            pushUpdate('/A-1');

            updateService.addImTreeListener(listener)
                .then(function () {
                    assert.sameMembers(listener.events, ['newImTree', 'completeImTree', 'onlineImTree']);

                    listener.clear();

                    pushUpdate('/A-1/B-1');

                    assert.sameMembers(listener.events, ['updateOfImTree', 'completeImTree', 'onlineImTree']);
                })
                .then(done).catch(done);
        });

        it('after setting history marker should notify with new and complete imTree', function (done) {
            pushUpdate('/A-1');

            updateService.setHistoryMarker()
                .then(function () {
                    return updateService.addImTreeListener(listener);
                })
                .then(function () {
                    pushUpdate('/B-1');

                    assert.sameMembers(listener.events, ['newImTree', 'completeImTree', 'onlineImTree']);
                })
                .then(done).catch(done);
        });

        it('after clearing history marker should notify with new and complete imTree', function (done) {
            pushUpdate('/A-1');

            Promise.join(
                updateService.addImTreeListener(listener),
                updateService.setHistoryMarker())
                .then(function () {
                    pushUpdate('/A-1/B-1');
                    pushUpdate('/A-1/B-2');
                    listener.clear();

                    return updateService.clearHistoryMarker();
                })
                .then(function () {
                    assert.sameMembers(listener.events, ['newImTree', 'completeImTree', 'onlineImTree']);
                })
                .then(done).catch(done);
        });

    });

    describe('merge history', function () {
        const NUMBER_OF_ENTRIES = 3;

        beforeEach(function (done) {
            pushUpdate('/A-1');
            pushUpdate('/A-1/B-1');
            pushUpdate('/A-1/B-1/C-1');

            listener = createListener();
            updateService.addImTreeListener(listener)
                .then(done).catch(done);
            listener.clear();
        });

        it('should notify with imTree', function (done) {
            updateService.mergeHistory()
                .then(function () {
                    assert.deepEqual(listener.events, [
                        //notify before merge with empty imTree
                        {name: 'newImTree', imTreeSize: 0},
                        {name: 'completeImTree', imTreeSize: 0},
                        {name: 'onlineImTree', imTreeSize: 0},

                        //notify after merge with 'merged' imTree
                        {name: 'newImTree', imTreeSize: NUMBER_OF_ENTRIES},
                        {name: 'completeImTree', imTreeSize: NUMBER_OF_ENTRIES},
                        {name: 'onlineImTree', imTreeSize: NUMBER_OF_ENTRIES}
                    ]);
                })
                .then(done).catch(done);
        });

        it('should notify when is set history marker before merge', function (done) {
            updateService.setHistoryMarker()
                .then(function () {
                    listener.clear();
                    return updateService.mergeHistory();
                })
                .then(function () {
                    assert.deepEqual(listener.events, [
                        //notify before merge with empty imTree
                        {name: 'newImTree', imTreeSize: 0},
                        {name: 'completeImTree', imTreeSize: 0},
                        {name: 'onlineImTree', imTreeSize: 0},

                        //notify after merge with 'merged' imTree
                        {name: 'newImTree', imTreeSize: NUMBER_OF_ENTRIES},
                        {name: 'completeImTree', imTreeSize: NUMBER_OF_ENTRIES},
                        {name: 'onlineImTree', imTreeSize: NUMBER_OF_ENTRIES}
                    ]);

                })
                .then(done).catch(done);
        });

        describe('merge first entries', function () {
            const NOT_ENOUGH_ENTRIES_TO_MERGE = 1;

            it('should not merge when there is not enough bytes to merge', function (done) {
                updateService.setHistoryMarker()
                    .then(function () {
                        listener.events = [];
                    })
                    .then(function () {
                        return updateService.mergeFirstEntries(NOT_ENOUGH_ENTRIES_TO_MERGE);
                    })
                    .then(function () {
                        assert.deepEqual(listener.events, []);
                    })
                    .then(done).catch(done);
            });

            it('should notify with imTree after merge of first entries when history marker was set', function (done) {
                updateService.setHistoryMarker()
                    .then(function () {
                        listener.events = [];
                    })
                    .then(function () {
                        return updateService.mergeFirstEntries(NUMBER_OF_ENTRIES);
                    })
                    .then(function () {
                        assert.deepEqual(listener.events, [
                            {name: 'newImTree', imTreeSize: NUMBER_OF_ENTRIES},
                            {name: 'completeImTree', imTreeSize: NUMBER_OF_ENTRIES}
                        ]);
                    })
                    .then(done).catch(done);
            });

            it('should notify with imTree even when history marker was not set', function (done) {
                updateService.mergeFirstEntries(NUMBER_OF_ENTRIES)
                    .then(function () {
                        assert.deepEqual(listener.events, [
                            {name: 'newImTree', imTreeSize: NUMBER_OF_ENTRIES},
                            {name: 'completeImTree', imTreeSize: NUMBER_OF_ENTRIES},
                            {name: 'onlineImTree', imTreeSize: NUMBER_OF_ENTRIES}
                        ]);
                    })
                    .then(done).catch(done);
            });

        });

        function createListener() {
            return {
                events: [],
                completeImTree: function (imTree) {
                    this.events.push({name: 'completeImTree', imTreeSize: imTree.size()});
                },
                newImTree: function (imTree) {
                    this.events.push({name: 'newImTree', imTreeSize: imTree.size()});
                },
                updateOfImTree: function () {
                    this.events.push({name: 'updateOfImTree'});
                },
                onlineImTree: function (imTree) {
                    this.events.push({name: 'onlineImTree', imTreeSize: imTree.size()});
                },
                clear: function () {
                    this.events = [];
                }
            };
        }

    });

    function pushUpdate(distName, byteSize) {
        updateService.pushUpdate(distName, new Uint8Array(byteSize || 0));
    }

    function pushRemove(distName) {
        updateService.pushRemove(distName);
    }

    function pushReset() {
        pushDisconnect("BTS");
    }

    function pushDisconnect(source) {
        updateService.pushDisconnect({autoReconnect: false, source: source});
    }

    function hasDistName(imTree, distName) {
        return !!imTree.findByDistName(distName);
    }

    function createListener() {
        return {
            imTree: null,
            events: [],
            completeImTree: function (imTree) {
                this.events.push('completeImTree');
                this.imTree = imTree;
            },
            newImTree: function (imTree) {
                this.events.push('newImTree');
                this.imTree = imTree;
            },
            updateOfImTree: function () {
                this.events.push('updateOfImTree');
            },
            onlineImTree: function () {
                this.events.push('onlineImTree');
            },
            clear: function () {
                this.imTree = null;
                this.events = [];
            },
            hasUpdateFor: function (distName) {
                return hasDistName(this.imTree, distName);
            }
        };
    }


});
