describe('Service: updateService#notifier', function () {

    var updateService;
    var listener;

    beforeEach(function () {
        var metaService = {
            getLastMetaUid: () => 0
        };
        updateService = infoModel.updateService({metaService: metaService});

        listener = createListener();
    });

    describe('addImTreeListener', function () {

        it('should not notify when listeners disabled', function (done) {
            updateService.disableListeners();
            updateService.addImTreeListener(listener)
                .then(function () {
                    assert.deepEqual(listener.$events, []);
                })
                .then(done)
                .catch(done);
        });

        it('should notify correct listeners when history marker set', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                .then(function () {
                    return updateService.addImTreeListener(listener);
                })
                .then(function () {
                    assert.deepEqual(listener.$events, ['newImTree', 'completeImTree']);
                })
                .then(done)
                .catch(done);
        });

        it('should not notify other listeners', function (done) {
            const secondListener = createListener();
            updateService.addImTreeListener(secondListener)
                .then(function () {
                    secondListener.clearEvents();
                    return updateService.addImTreeListener(listener);
                })
                .then(function () {
                    assert.deepEqual(secondListener.$events, []);
                })
                .then(done)
                .catch(done);
        });

        it('should notify correct listeners when history marker not set', function (done) {
            updateService.addImTreeListener(listener)
                .then(function () {
                    assert.deepEqual(listener.$events, ['newImTree', 'completeImTree', 'onlineImTree']);
                })
                .then(done)
                .catch(done);
        });

    });

    describe('addHistoryListener', function () {

        it('should not notify other listeners', function (done) {
            const secondListener = createListener();
            updateService.addHistoryListener(secondListener)
                .then(function () {
                    secondListener.clearEvents();
                    return updateService.addHistoryListener(listener);
                })
                .then(function () {
                    assert.deepEqual(secondListener.$events, []);
                })
                .then(done)
                .catch(done);
        });

        it('should notify correct listeners when history marker not set', function (done) {
            updateService.addHistoryListener(listener)
                .then(function () {
                    assert.deepEqual(listener.$events, ['newEntries']);
                })
                .then(done)
                .catch(done);
        });

        it('should notify correct listeners when history marker set', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                .then(function () {
                    return updateService.addHistoryListener(listener);
                })
                .then(function () {
                    assert.deepEqual(listener.$events, ['newEntries', 'historyMarkerWasSet']);
                })
                .then(done)
                .catch(done);
        });

    });

    describe('addCurrentEntryListener', function () {

        it('should not notify other listeners', function (done) {
            const secondListener = createListener();
            updateService.addCurrentEntryListener(secondListener)
                .then(function () {
                    secondListener.clearEvents();
                    return updateService.addCurrentEntryListener(listener);
                })
                .then(function () {
                    assert.deepEqual(secondListener.$events, []);
                })
                .then(done)
                .catch(done);
        });

        it('should notify with changeEntry', function (done) {
            updateService.addCurrentEntryListener(listener)
                .then(function () {
                    assert.deepEqual(listener.$events, ['changeEntry']);
                })
                .then(done)
                .catch(done);
        });

    });

    describe('socket events', function () {

        beforeEach(function (done) {
            updateService.addHistoryListener(listener)
                .then(function () {
                    return updateService.addImTreeListener(listener);
                })
                .then(function () {
                    return updateService.addCurrentEntryListener(listener);
                })
                .then(function () {
                    listener.clearEvents();
                })
                .then(done)
                .catch(done);
        });

        describe('disconnected', function () {

            it('should notify correct listeners when history marker set', function (done) {
                pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
                pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
                updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                    .then(function () {
                        listener.clearEvents();
                        pushDisconnect('BTS');
                        assert.deepEqual(listener.$events, ['newEntry', 'onlineImTree']);
                    })
                    .then(done)
                    .catch(done);
            });

            it('should notify correct listeners when history marker set', function () {
                pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
                listener.clearEvents();
                pushDisconnect('BTS');
                assert.deepEqual(listener.$events, ['newEntry', 'newImTree', 'completeImTree', 'onlineImTree', 'changeEntry']);
            });

        });

        describe('data', function () {

            it('should notify correct listeners when history marker set', function (done) {
                pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
                pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
                updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                    .then(function () {
                        listener.clearEvents();
                        pushUpdate('MRBTS-1/SOME_CLASS_NAME-3');
                        assert.deepEqual(listener.$events, ['newEntry', 'onlineImTree']);
                    })
                    .then(done)
                    .catch(done);
            });

            it('should notify correct listeners when history marker not set', function () {
                pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
                assert.deepEqual(listener.$events, ['newEntry', 'completeImTree', 'onlineImTree', 'updateOfImTree', 'changeEntry']);
            });

        });

    });

    describe('setHistoryMarker', function () {

        beforeEach(function (done) {
            addAllListeners(done);
        });

        it('should notify correct listeners', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            listener.clearEvents();
            updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                .then(function () {
                    assert.deepEqual(listener.$events, ['historyMarkerWasSet', 'newImTree', 'completeImTree', 'changeEntry']);
                })
                .then(done)
                .catch(done);
        });

    });

    describe('clearHistoryMarker', function () {

        beforeEach(function (done) {
            addAllListeners(done);
        });

        it('should not notify when history marker is not set', function (done) {
            updateService.clearHistoryMarker()
                .then(function () {
                    assert.deepEqual(listener.$events, []);
                })
                .then(done)
                .catch(done);
        });

        it('should notify correct listeners', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                .then(function () {
                    listener.clearEvents();
                    return updateService.clearHistoryMarker();
                })
                .then(function () {
                    assert.deepEqual(listener.$events, ['historyMarkerWasCleared', 'newImTree', 'completeImTree', 'onlineImTree', 'changeEntry']);
                })
                .then(done)
                .catch(done);
        });

    });

    describe('clearHistory', function () {

        beforeEach(function (done) {
            addAllListeners(done);
        });

        it('should notify correct listeners when history marker was set', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                .then(function () {
                    listener.clearEvents();
                    return updateService.clearHistory();
                })
                .then(function () {
                    assert.deepEqual(listener.$events, ['clearEntries', 'historyMarkerWasCleared', 'newImTree', 'completeImTree', 'onlineImTree']);
                })
                .then(done)
                .catch(done);
        });

        it('should notify correct listeners when history marker was not set', function (done) {
            updateService.clearHistory()
                .then(function () {
                    assert.deepEqual(listener.$events, ['clearEntries', 'newImTree', 'completeImTree', 'onlineImTree']);
                })
                .then(done)
                .catch(done);
        });

    });

    describe('mergeHistory', function () {

        beforeEach(function (done) {
            addAllListeners(done);
        });

        it('should notify correct listeners when history marker was set', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                .then(function () {
                    listener.clearEvents();
                    return updateService.mergeHistory();
                })
                .then(function () {
                    assert.deepEqual(listener.$events, ['clearEntries', 'historyMarkerWasCleared', 'newImTree', 'completeImTree', 'onlineImTree',
                        'newEntry', 'historyMarkerWasSet', 'newImTree', 'completeImTree', 'onlineImTree', 'changeEntry']);
                })
                .then(done)
                .catch(done);
        });

        it('should notify correct listeners when history marker was not set', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            listener.clearEvents();
            updateService.mergeHistory()
                .then(function () {
                    assert.deepEqual(listener.$events, ['clearEntries', 'newImTree', 'completeImTree', 'onlineImTree',
                        'newEntry', 'newImTree', 'completeImTree', 'onlineImTree', 'changeEntry']);
                })
                .then(done)
                .catch(done);
        });

    });

    describe('mergeFirstEntries', function () {

        beforeEach(function (done) {
            addAllListeners(done);
        });

        it('should notify correct listeners when history marker was set', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                .then(function () {
                    listener.clearEvents();
                    return updateService.mergeFirstEntries(2);
                })
                .then(function () {
                    assert.deepEqual(listener.$events, ['clearEntries', 'historyMarkerWasCleared',
                        'newEntries', 'historyMarkerWasSet', 'newImTree', 'completeImTree', 'changeEntry']);
                })
                .then(done)
                .catch(done);
        });

        it('should notify correct listeners when history marker was not set', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            listener.clearEvents();
            updateService.mergeFirstEntries(2)
                .then(function () {
                    assert.deepEqual(listener.$events, ['clearEntries', 'historyMarkerWasCleared',
                        'newEntries', 'newImTree', 'completeImTree', 'onlineImTree', 'changeEntry']);
                })
                .then(done)
                .catch(done);
        });

    });

    describe('mergeEntriesRange', function () {

        beforeEach(function (done) {
            addAllListeners(done);
        });

        it('should notify correct listeners when history marker was set', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            updateService.setHistoryMarker(updateService.getHistoryEntry(0))
                .then(function () {
                    listener.clearEvents();
                    return updateService.mergeEntriesRange(0, 2);
                })
                .then(function () {
                    assert.deepEqual(listener.$events, ['clearEntries', 'historyMarkerWasCleared',
                        'newEntries', 'historyMarkerWasSet', 'newImTree', 'completeImTree', 'changeEntry']);
                })
                .then(done)
                .catch(done);
        });

        it('should notify correct listeners when history marker was not set', function (done) {
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-1');
            pushUpdate('MRBTS-1/SOME_CLASS_NAME-2');
            listener.clearEvents();
            updateService.mergeEntriesRange(0, 2)
                .then(function () {
                    assert.deepEqual(listener.$events, ['clearEntries', 'historyMarkerWasCleared',
                        'newEntries', 'newImTree', 'completeImTree', 'onlineImTree', 'changeEntry']);
                })
                .then(done)
                .catch(done);
        });

    });

    function addAllListeners(done) {
        updateService.addHistoryListener(listener)
            .then(function () {
                return updateService.addImTreeListener(listener);
            })
            .then(function () {
                return updateService.addCurrentEntryListener(listener);
            })
            .then(function () {
                listener.clearEvents();
            })
            .then(done)
            .catch(done);
    }

    function createListener() {
        return {
            $events: [],
            clearEvents: function () {
                this.$events = [];
            },
            // imTreeListener part
            completeImTree: function () {
                this.$events.push('completeImTree');
            },
            newImTree: function () {
                this.$events.push('newImTree');
            },
            updateOfImTree: function () {
                this.$events.push('updateOfImTree');
            },
            onlineImTree: function () {
                this.$events.push('onlineImTree');
            },
            // historyListener part
            newEntry: function () {
                this.$events.push('newEntry');
            },
            newEntries: function () {
                this.$events.push('newEntries');
            },
            clearEntries: function () {
                this.$events.push('clearEntries');
            },
            historyMarkerWasSet: function () {
                this.$events.push('historyMarkerWasSet');
            },
            historyMarkerWasCleared: function () {
                this.$events.push('historyMarkerWasCleared');
            },
            // currentEntryListener part
            changeEntry: function () {
                this.$events.push('changeEntry');
            }
        };
    }

    function pushUpdate(distName) {
        updateService.pushUpdate(distName, new Uint8Array(0));
    }

    function pushRemove(distName) {
        updateService.pushRemove(distName);
    }

    function pushDisconnect(source) {
        updateService.pushDisconnect({autoReconnect: false, source: source});
    }

});
