const ImTree = require('../../src/imtree');
const btsInfoServiceFactory = require('../../src/utils/btsInfoService');

describe('btsInfoService', () => {
    let updateService,
        btsInfoService,
        imTree;

    function updateServiceStub() {
        let imTreeListener = null;
        let data = null;

        return {
            addImTreeListener: listener => {
                imTreeListener = listener;
                if (data) imTreeListener.completeImTree(data);
            },
            removeImTreeListener: () => {
                imTreeListener = null;
            },
            pushData: tree => {
                data = tree;
                if (imTreeListener) imTreeListener.completeImTree(tree);
            }
        };
    }

    beforeEach(() => {
        updateService = updateServiceStub();
        btsInfoService = btsInfoServiceFactory.newInstance(updateService);
        imTree = new ImTree();
    });

    it('should notify when all parameters are null', done => {
        btsInfoService.watch(watcher);
        updateService.pushData(imTree);

        function watcher(btsInfo) {
            assert.isNull(btsInfo.id);
            assert.isNull(btsInfo.state);
            done();
        }
    });

    it('should notify when watching started after IM update', done => {
        updateService.pushData(imTree);
        btsInfoService.watch(watcher);

        function watcher(btsInfo) {
            assert.isNull(btsInfo.id);
            assert.isNull(btsInfo.state);
            done();
        }
    });

    it('should not be notified after unwatch', done => {
        imTree.mergeNode('/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1', {stateInfo: {proceduralState: 'commissioned'}});
        const firstUnwatcher = btsInfoService.watch(firstWatcher);
        btsInfoService.watch(secondWatcher);
        firstUnwatcher.unwatch();

        updateService.pushData(imTree);

        function firstWatcher() {
            assert.fail('Watcher should not be notified');
        }

        function secondWatcher() {
            done();
        }
    });

    describe('id', () => {

        beforeEach(() => {
            imTree.mergeNode('/MRBTS-1/RAT-1/BTS_L-1/BTS_CONF-1', {currentNp: '/MRBTS-1/RAT-1/BTS_L-1/BTS_CONF-1/NP-1'});
            imTree.mergeNode('/MRBTS-1/RAT-1/BTS_L-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-1', {mrbtsId: 20});
        });

        it('should notify with correct id', done => {
            btsInfoService.watch(btsInfoWatcher);

            updateService.pushData(imTree);

            function btsInfoWatcher(btsInfo) {
                assert.equal(btsInfo.id, 20);
                done();
            }
        });

        it('should notify when watched -> unwatched -> watched, but bts id doesn\'t change', done => {
            const unwatcher = btsInfoService.watch(firstWatcher);

            updateService.pushData(imTree);
            unwatcher.unwatch();
            btsInfoService.watch(secondWatcher);

            function firstWatcher(btsInfo) {
                assert.equal(btsInfo.id, 20);
            }

            function secondWatcher(btsInfo) {
                assert.equal(btsInfo.id, 20);
                done();
            }
        });

    });

    describe('state', () => {

        beforeEach(() => {
            imTree.mergeNode('/MRBTS-1/RAT-1/BTS_L-1/LNBTS-1', {stateInfo: {proceduralState: 'commissioned'}});
        });

        it('should notify with correct state', done => {
            btsInfoService.watch(btsInfoWatcher);

            updateService.pushData(imTree);

            function btsInfoWatcher(btsInfo) {
                assert.equal(btsInfo.state, 'commissioned');
                done();
            }
        });

        it('should notify when watched -> unwatched -> watched, but state doesn\'t change', done => {
            const unwatcher = btsInfoService.watch(firstWatcher);

            updateService.pushData(imTree);
            unwatcher.unwatch();
            btsInfoService.watch(secondWatcher);

            function firstWatcher(btsInfo) {
                assert.equal(btsInfo.state, 'commissioned');
            }

            function secondWatcher(btsInfo) {
                assert.equal(btsInfo.state, 'commissioned');
                done();
            }
        });

    });
});
