describe('Service: distNameService', function () {

    var dnService = adminCore.utils.distNameUtils;

    var DIST_NAME = "/ROOT-1/NAME-1/CHILD-1";

    describe('distNameUpToClassName', function () {

        it('should get distName of up to class name', function () {
            assert.equal("/ROOT-1/NAME-1/CHILD-1", dnService.distNameUpToClassName(DIST_NAME, "CHILD-1"));
            assert.equal("/ROOT-1/NAME-1", dnService.distNameUpToClassName(DIST_NAME, "NAME-1"));
            assert.equal("/ROOT-1", dnService.distNameUpToClassName(DIST_NAME, "ROOT-1"));
        });

        it('should get undefined when cannot find distName up to class name', function () {
            assert.isUndefined(dnService.distNameUpToClassName(DIST_NAME, "XXX"));
        });

        it('distName up to class name should search up to last occurrence of class name', function () {
            assert.equal("/A-1/B-1/C-1/B-1", dnService.distNameUpToClassName("/A-1/B-1/C-1/B-1", "B-1"));
            assert.equal("/A-1/A-1/A-1", dnService.distNameUpToClassName("/A-1/A-1/A-1", "A-1"));
        });

    });

    describe('distNameStartingFromClassName', function () {

        it('should get distName from given class name', function () {
            assert.equal("CHILD-1", dnService.distNameStartingFromClassName(DIST_NAME, "CHILD-1"));
            assert.equal("NAME-1/CHILD-1", dnService.distNameStartingFromClassName(DIST_NAME, "NAME-1"));
            assert.equal("ROOT-1/NAME-1/CHILD-1", dnService.distNameStartingFromClassName(DIST_NAME, "ROOT-1"));
        });

        it('should get undefined when cannot find class name in distName', function () {
            assert.isUndefined(dnService.distNameStartingFromClassName(DIST_NAME, "XXX"));
        });

    });

    describe('getName', function () {

        it('should get name from distName (distName last part)', function () {
            assert.equal('CHILD-1', dnService.getName(DIST_NAME));
        });

    });

    describe('getParent', function () {

        it('should get parent distName', function () {
            assert.equal('/ROOT-1/NAME-1', dnService.getParent(DIST_NAME));
        });

        it('should return root as parent - slash at the beginning of DN', function () {
            assert.equal('/ROOT-1', dnService.getParent('/ROOT-1'));
        });

        it('should return root as parent - without slash at the beginning of DN', () => {
            assert.equal('ROOT-1', dnService.getParent('ROOT-1'));
        });

    });

    describe('isRoot', function () {

        it('should detect if distName is a root', function () {
            assert.isTrue(dnService.isRoot('/ROOT-1'));
        });

    });

    describe('getClassName', function () {

        it('should get className of distName (className of distName last part)', function () {
            assert.equal('CHILD', dnService.getClassName(DIST_NAME));
        });

    });

    describe('getParentClassName', function () {

        it('should get className of parent distName', function () {
            assert.equal('NAME', dnService.getParentClassName(DIST_NAME));
        });

    });

    describe('getParsedName', function () {

        it('should return parsed name of distName (className and id of distName last part)', function () {
            var parsedName = dnService.getParsedName(DIST_NAME);
            assert.equal('CHILD', parsedName.className);
            assert.equal('1', parsedName.id);
        });

    });

    describe('getAncestor', function () {

        it('should return second level ancestor', function () {
            const level = 2;
            assert.strictEqual(dnService.getAncestor(DIST_NAME, level), '/ROOT-1');
        });

    });

    describe('isSubtree', function () {

        it('should test for subtree', function () {
            assert.isTrue(dnService.isSubtree('/root-1', '/root-1/child-1'));
            assert.isTrue(dnService.isSubtree('/root-1', '/root-1'));

            assert.isFalse(dnService.isSubtree('/root-1/child-1', '/root-1/child-12'));
            assert.isFalse(dnService.isSubtree('/root-1/child-1', '/root-1'));
        });

    });

    describe('getId', function () {

        it('should get one character id', function () {
            assert.deepEqual(dnService.getId('/THIS/IS-1/DIST_NAME-2'), 2);
        });

        it('should get multiple character id', function () {
            assert.deepEqual(dnService.getId('/THIS/IS-1/DIST_NAME-232'), 232);
        });

    });

    describe('compareDistNames', function () {

        it('should return 0 when distNames are the same', function () {
            const DN = '/FIRST/PARENT-3/DIST_NAME-1';
            assert.deepEqual(dnService.compareDistNames(DN, DN), 0);
        });

        it('should compare when distNames differ on node id', function () {
            const DN_1 = '/FIRST/PARENT-3/DIST_NAME-1';
            const DN_2 = '/FIRST/PARENT-3/DIST_NAME-2';
            assert.deepEqual(dnService.compareDistNames(DN_1, DN_2), -1);
            assert.deepEqual(dnService.compareDistNames(DN_2, DN_1), 1);
        });

        it('should compare when distNames differ on node id - number of characters', function () {
            const DN_1 = '/FIRST/PARENT-3/DIST_NAME-2';
            const DN_2 = '/FIRST/PARENT-3/DIST_NAME-11';
            assert.deepEqual(dnService.compareDistNames(DN_1, DN_2), -1);
            assert.deepEqual(dnService.compareDistNames(DN_2, DN_1), 1);
        });

        it('should compare when distNames differ on parent id', function () {
            const DN_1 = '/FIRST/PARENT-3/DIST_NAME-1';
            const DN_2 = '/FIRST/PARENT-2/DIST_NAME-1';
            assert.deepEqual(dnService.compareDistNames(DN_1, DN_2), 1);
            assert.deepEqual(dnService.compareDistNames(DN_2, DN_1), -1);
        });

        it('should compare when distNames differ on parent id - number of characters', function () {
            const DN_1 = '/FIRST/PARENT-22/DIST_NAME-1';
            const DN_2 = '/FIRST/PARENT-3/DIST_NAME-1';
            assert.deepEqual(dnService.compareDistNames(DN_1, DN_2), 1);
            assert.deepEqual(dnService.compareDistNames(DN_2, DN_1), -1);
        });

        it('should compare when distNames differ on parent & node id', function () {
            const DN_1 = '/FIRST/PARENT-2/DIST_NAME-2';
            const DN_2 = '/FIRST/PARENT-3/DIST_NAME-1';
            assert.deepEqual(dnService.compareDistNames(DN_1, DN_2), -1);
            assert.deepEqual(dnService.compareDistNames(DN_2, DN_1), 1);
        });

        it('should compare when distNames differ on node class name', function () {
            const DN_1 = '/FIRST/PARENT-2/CLASS_NAME-1';
            const DN_2 = '/FIRST/PARENT-2/SECOND_CLASS_NAME-1';
            assert.deepEqual(dnService.compareDistNames(DN_1, DN_2), -1);
            assert.deepEqual(dnService.compareDistNames(DN_2, DN_1), 1);
        });

        it('should compare when distNames differ on parent class name', function () {
            const DN_1 = '/FIRST/B_PARENT-2/DIST_NAME-1';
            const DN_2 = '/FIRST/A_PARENT-3/DIST_NAME-1';
            assert.deepEqual(dnService.compareDistNames(DN_1, DN_2), 1);
            assert.deepEqual(dnService.compareDistNames(DN_2, DN_1), -1);
        });

    });
});
