(function () {
    'use strict';

    const fileNameFactory = require('../../src/utils/fileNameFactory');
    const imCompatibilityService = require('../../src/utils/imCompatibilityService');
    const ImTree = require('../../src/imtree');

    describe('fileNameFactory', function () {

        const TEST_DATE = new Date(2016, 4, 24, 9, 4, 24).getTime(); // Tue May 24 2016 09:04:24
        const PREFIX = 'Abcd_efg';
        const EXTENSION = 'zip';

        var imTree;

        beforeEach(function () {

            var decoderService = {
                decodeByClassName: (_, payload) => {
                    return payload;
                }
            };
            imTree = new ImTree(decoderService);
            const swName = 'FL17_FSM3_9999_160518_030257';
            const btsId = 1;
            const btsName = 'IAV_WRO345';
            setupBtsConf(swName);
            setupMrbtsA(btsId);
            setupLnbtsA(btsName);
        });

        it('should create correct name when imtree doesn\'t exist', function () {
            // given
            imTree = null;
            // when
            const name = fileNameFactory.createName(imTree, PREFIX, EXTENSION, TEST_DATE);
            // then
            assert.match(name, /^Abcd_efg_20160524-0904\.zip$/);
        });

        it('should create correct name when MRBTS doesn\'t exist', function () {
            // given
            imTree.removeByDistName('/MRBTS-1');
            // when
            const name = fileNameFactory.createName(imTree, PREFIX, EXTENSION, TEST_DATE);
            // then
            assert.match(name, /^Abcd_efg_20160524-0904\.zip$/);
        });

        it('should create correct name when BTS_CONF doesn\'t exist', function () {
            // given
            imTree.removeByDistName('/MRBTS-1/RAT-1/BTS_L-1/BTS_CONF-1');
            // when
            const name = fileNameFactory.createName(imTree, PREFIX, EXTENSION, TEST_DATE);
            // then
            assert.match(name, /^Abcd_efg_20160524-0904\.zip$/);
        });

        it('should create correct name when MRBTS_A doesn\'t exist', function () {
            // given
            imTree.removeByDistName('/MRBTS-1/RAT-1/BTS_L-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-1');
            // when
            const name = fileNameFactory.createName(imTree, PREFIX, EXTENSION, TEST_DATE);
            // then
            assert.match(name, /^Abcd_efg_FL17_FSM3_9999_160518_030257_20160524-0904\.zip$/);
        });

        it('should create correct name when LNBTS_A doesn\'t exist', function () {
            // given
            imTree.removeByDistName('/MRBTS-1/RAT-1/BTS_L-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-1/LNBTS_A-1');
            // when
            const name = fileNameFactory.createName(imTree, PREFIX, EXTENSION, TEST_DATE);
            // then
            assert.match(name, /^Abcd_efg_MRBTS-1_FL17_FSM3_9999_160518_030257_20160524-0904\.zip$/);
        });

        it('should create correct name', function () {
            // when
            const name = fileNameFactory.createName(imTree, PREFIX, EXTENSION, TEST_DATE);
            // then
            assert.match(name, /^Abcd_efg_MRBTS-1_FL17_FSM3_9999_160518_030257_IAV_WRO345_20160524-0904\.zip$/);
        });

        it('should create correct name without prefix', function () {
            // when
            const name = fileNameFactory.createName(imTree, null, EXTENSION, TEST_DATE);
            // then
            assert.match(name, /^MRBTS-1_FL17_FSM3_9999_160518_030257_IAV_WRO345_20160524-0904\.zip$/);
        });

        function setupBtsConf(buildName) {
            const params = {
                activeSoftwareVersion: buildName,
                currentNp: '/MRBTS-1/RAT-1/BTS_L-1/BTS_CONF-1/NP-1'
            };
            imTree.mergeNode('/MRBTS-1/RAT-1/BTS_L-1/BTS_CONF-1', params, params, 1);
        }

        function setupMrbtsA(id) {
            const params = {mrbtsId: id};
            imTree.mergeNode('/MRBTS-1/RAT-1/BTS_L-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-1', params, params, 1);
        }

        function setupLnbtsA(btsName) {
            const params = {btsName: btsName};
            imTree.mergeNode('/MRBTS-1/RAT-1/BTS_L-1/BTS_CONF-1/NP-1/SCF-1/MRBTS_A-1/LNBTS_A-1', params, params, 1);
        }

    });

})(this);
