const qsfpLinksMergerProvider = require('../../src/utils/qsfpLinksMergerService');

describe('qsfpLinksMerger', function () {
    'use strict';
    var qsfpLinksMerger;

    beforeEach(
        function () {
            qsfpLinksMerger = qsfpLinksMergerProvider;
        });


    describe('qsfp parameter values ', function () {
        var CONNECTOR_DIST_NAME = '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/QSFP_L-1/CONNECTOR_L-';

        it('should be unique', function () {
            var parameter1 = {cablinkType: 'RP3Link'};
            var parameter2 = {cablinkType: 'RP3_01Link'};
            var parameters = [parameter1, parameter1, parameter1, parameter2];


            var result = qsfpLinksMerger.mergeParameters(parameters);
            expect(result).to.eql({cablinkType: 'RP3Link, RP3_01Link'});
        });

        it('undefined values should be ignored', function () {
            var parameter1 = {cablinkType: 'RP3Link'};
            var parameter2 = {cablinkType: 'RP3_01Link'};
            var parameter3 = {};
            var parameters = [parameter1, parameter1, parameter3, parameter2];


            var result = qsfpLinksMerger.mergeParameters(parameters);
            expect(result).to.eql({cablinkType: 'RP3Link, RP3_01Link'});
        });

        it('should be qsfp dist name for sourceDistName paramater ', function () {
            var parameter1 = {sourceDistName: CONNECTOR_DIST_NAME + 1};
            var parameter2 = {sourceDistName: CONNECTOR_DIST_NAME + 2};
            var parameter3 = {sourceDistName: CONNECTOR_DIST_NAME + 3};
            var parameter4 = {sourceDistName: CONNECTOR_DIST_NAME + 4};
            var parameters = [parameter1, parameter2, parameter3, parameter4];

            var result = qsfpLinksMerger.mergeParameters(parameters);
            expect(result).to.eql({sourceDistName: '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/QSFP_L-1'});
        });

        it('should be qsfp dist name for destinationDistName parameter', function () {
            var parameter1 = {destinationDistName: CONNECTOR_DIST_NAME + 1};
            var parameter2 = {destinationDistName: CONNECTOR_DIST_NAME + 2};
            var parameter3 = {destinationDistName: CONNECTOR_DIST_NAME + 3};
            var parameter4 = {destinationDistName: CONNECTOR_DIST_NAME + 4};
            var parameters = [parameter1, parameter2, parameter3, parameter4];

            var result = qsfpLinksMerger.mergeParameters(parameters);
            expect(result).to.eql({destinationDistName: '/MRBTS-1/RAT-1/EQM_L-1/SMOD_L-1/QSFP_L-1'});
        });

        it('should be merge recursively', function () {

            var stateInfoValues = {
                administrativeState: 'Unlocked',
                availabilityStatus: 'Online',
                operationalState: 'Enabled',
                usageState: 'Idle'
            };

            var parameter1 = {stateInfo: stateInfoValues};
            var parameters = [parameter1, parameter1, parameter1, parameter1];

            var result = qsfpLinksMerger.mergeParameters(parameters);
            expect(result).to.eql({stateInfo: stateInfoValues});
        });

        it('operational state should be set to Disabled if one of values is Disabled', function () {

            var stateInfoValuesEnabled = {operationalState: 'Enabled'};
            var stateInfoValuesDisabled = {operationalState: 'Disabled'};

            var parameter1 = {stateInfo: stateInfoValuesEnabled};
            var parameter2 = {stateInfo: stateInfoValuesDisabled};
            var parameters = [parameter1, parameter1, parameter1, parameter2];

            var result = qsfpLinksMerger.mergeParameters(parameters);
            expect(result).to.eql({stateInfo: stateInfoValuesDisabled});
        });

        it('should merge array values', function () {

            var reasoningStatus1 = {reasoningStatus: ['RTPMConnectionFailure']};
            var reasoningStatus2 = {reasoningStatus: ['RTPMDataGap']};

            var parameter1 = {stateInfo: reasoningStatus1};
            var parameter2 = {stateInfo: reasoningStatus2};
            var parameters = [parameter1, parameter1, parameter1, parameter2];

            var expectedReasoningStatus = {reasoningStatus: ['RTPMConnectionFailure', 'RTPMDataGap']};

            var result = qsfpLinksMerger.mergeParameters(parameters);
            expect(result).to.eql({stateInfo: expectedReasoningStatus});
        });

        it('should merge integers', function () {
            var parameter1 = {destinationPhysicalPortNo: 0};
            var parameter2 = {destinationPhysicalPortNo: 1};
            var parameters = [parameter1, parameter1, parameter1, parameter2];

            var result = qsfpLinksMerger.mergeParameters(parameters);
            expect(result).to.eql({destinationPhysicalPortNo: '0, 1'});
        });

        describe('should merge boolean values', function () {
            it('should merge true values', function () {

                var parameter1 = {configuration: true};
                var parameter2 = {configuration: true};
                var parameters = [parameter1, parameter1, parameter1, parameter2];

                var result = qsfpLinksMerger.mergeParameters(parameters);
                expect(result).to.eql({configuration: true});
            });

            it('should merge false values', function () {

                var parameter1 = {configuration: false};
                var parameter2 = {configuration: false};
                var parameters = [parameter1, parameter1, parameter1, parameter2];

                var result = qsfpLinksMerger.mergeParameters(parameters);
                expect(result).to.eql({configuration: false});
            });

            it('should merge different values', function () {

                var parameter1 = {configuration: true};
                var parameter2 = {configuration: false};
                var parameters = [parameter1, parameter1, parameter1, parameter2];

                var result = qsfpLinksMerger.mergeParameters(parameters);
                expect(result).to.eql({configuration: false});
            });

            it('should merge undefined value with true', function () {

                var parameter1 = {configuration: true};
                var parameter2 = {};
                var parameters = [parameter1, parameter1, parameter1, parameter2];

                var result = qsfpLinksMerger.mergeParameters(parameters);
                expect(result).to.eql({configuration: true});
            });

            it('should merge undefined value with false', function () {

                var parameter1 = {configuration: false};
                var parameter2 = {};
                var parameters = [parameter1, parameter1, parameter1, parameter2];

                var result = qsfpLinksMerger.mergeParameters(parameters);
                expect(result).to.eql({configuration: false});
            });

        });

    });
});
