const entriesValidatorFactory = require('../../src/validation/entriesValidator');
const updateServiceFactory = require('../../src/updateService');
const assert = require('assert');

describe('entriesValidator', () => {

    var updateService;
    var entriesValidator;
    var entriesTimestampProvider;

    beforeEach(() => {
        var metaService = {
            getLastMetaUid: () => 0
        };
        updateService = updateServiceFactory({metaService});
        entriesValidator = entriesValidatorFactory(updateService);
    });

    describe('timestamp decreasing', () => {

        var idsWithDecreasingTimestamp;
        var unwatcher;

        beforeEach(() => {
            idsWithDecreasingTimestamp = [];
            unwatcher = entriesValidator.watchOnTimestampDecreasing(event => {
                if (event.type === 'new') idsWithDecreasingTimestamp.push(event.entryId);
                else if (event.type === 'clear') idsWithDecreasingTimestamp = [];
            });
        });

        it('should not notify with one history entry', () => {
            sendNewEntry();

            assert.deepEqual(idsWithDecreasingTimestamp, []);
        });

        it('should notify when previous entry has lower timestamp', () => {
            sendNewEntry(2000);
            sendNewEntry(1000);

            assert.deepEqual(idsWithDecreasingTimestamp, [1]);
        });

        it('should not notify when previous entry has the same timestamp', () => {
            sendNewEntry(1000);
            sendNewEntry(1000);

            assert.deepEqual(idsWithDecreasingTimestamp, []);
        });

        it('should not notify when previous entry has higher timestamp', () => {
            sendNewEntry(1000);
            sendNewEntry(2000);

            assert.deepEqual(idsWithDecreasingTimestamp, []);
        });

        it('should notify with several entries with decreasing timestamp', () => {
            sendNewEntry(1000);
            sendNewEntry(2000);
            sendNewEntry(1000); // decreasing
            sendNewEntry(2000);
            sendNewEntry(3000);
            sendNewEntry(2000); // decreasing
            sendNewEntry(3000);

            assert.deepEqual(idsWithDecreasingTimestamp, [2, 5]);
        });

        it('should notify new watcher about previous entries with decreasing timestamp', () => {
            var newWatcherIds = [];

            sendNewEntry(2000);
            sendNewEntry(1000);

            entriesValidator.watchOnTimestampDecreasing(event => {
                if (event.type === 'new') newWatcherIds.push(event.entryId);
                else if (event.type === 'clear') newWatcherIds = [];
            });

            assert.deepEqual(newWatcherIds, [1]);
        });

        it('should unwatch', () => {
            var newWatcherIds = [];

            unwatcher.unwatch();
            sendNewEntry(2000);
            sendNewEntry(1000);

            assert.deepEqual(idsWithDecreasingTimestamp, []);
        });

        it('should clear validations after history clear', () => {
            sendNewEntry(2000);
            sendNewEntry(1000);
            updateService.clear();

            assert.deepEqual(idsWithDecreasingTimestamp, []);
        });

        it('should validate new entries after history clear', () => {
            sendNewEntry(2000);
            sendNewEntry(1000);
            updateService.clear();
            sendNewEntry(2000);
            sendNewEntry(1000);

            assert.deepEqual(idsWithDecreasingTimestamp, [1]);
        });

    });

    function sendNewEntry(timestamp) {
        updateService.pushData({
            timestamp,
            removedDistNames: [],
            managedObjects: [{distName: '/SOME/DIST-1/NAME-3', payload: new ArrayBuffer()}]
        });
    }

});
